/*
 * Copyright (c) 1999 by David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Stand-alone shell for system maintainance for Linux.
 * This program should NOT be built using shared libraries.
 */

#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <errno.h>

#include "sash.h"


static const char * const	version = "3.3";


/*
 * The special maximum argument value which means that there is
 * no limit to the number of arguments on the command line.
 */
#define	INFINITE_ARGS	0x7fffffff

static  void    do_exit(int argc, const char **argv);
static  void    do_help(int argc, const char **argv);
static  void    do_prompt(int argc, const char **argv);
static  void    do_exit(int argc, const char **argv);
static  void    do_passwd(int argc, const char **argv);
static  void    do_ping(int argc, const char **argv);
static  void    do_catsn(int argc, const char **argv);
static  void    alrm_catcher(int i);


/*
 * One entry of the command table.
 */
typedef struct
{
	const char *	name;
	void		(*func)(int argc, const char ** argv);
	int		minArgs;
	int		maxArgs;
	const char *	description;
	const char *	usage;
} CommandEntry;


/*
 * The table of built-in commands.
 * This is terminated wih an entry containing NULL values.
 */
static const CommandEntry	commandEntryTable[] =
{
	{
		"prompt",	do_prompt,	2,	INFINITE_ARGS,
		"Set the prompt string for sash",
		"string"
	},
	{
		"passwd",	do_passwd,	1,	1,
		"Change password",
		""
	},
	{
		"ping",	do_ping,	2,	2,
		"Ping a host",
		"host(ip address)"
	},
	{
		"catsn",	do_catsn,	1,	1,
		"Show serial number",
		""
	},
/*
	{
		"help",		do_help,	1,	2,
		"Print help about a command",
		"[word]"
	},
	{
		"?",		do_help,	1,	2,
		"Print help about a command",
		"[word]"
	},
*/
	{
		"exit",		do_exit,	1,	1,
		"Exit from sash",
		""
	},
	{
		"quit",		do_exit,	1,	1,
		"Exit from sash",
		""
	},
	{
		NULL,		0,		0,	0,
		NULL,
		NULL
	}
};


/*
 * The definition of an command alias.
 */
typedef struct
{
	char *	name;
	char *	value;
} Alias;


/*
 * Local data.
 */

static	FILE *	sourcefiles[MAX_SOURCE];
static	int	sourceCount;

static	BOOL	intCrlf = TRUE;
static	char *	prompt;


/*
 * Local procedures.
 */
static	void	catchInt(int);
static	void	catchQuit(int);
static	void	readFile(const char * name);
static	void	command(const char * cmd);
static	BOOL	tryBuiltIn(const char * cmd);
static	void	runCmd(const char * cmd);
//static	void	childProcess(const char * cmd);
static	void	showPrompt(void);
static	void	usage(void);
static  void	WriteCmdLine(const char * cmd); 

/*
 * Global interrupt flag.
 */
BOOL	intFlag;


int
main(int argc, const char ** argv)
{
	const char *	cp;
	const char *	singleCommand;
	char		buf[PATH_LEN];

	singleCommand = NULL;

	/*
	 * Look for options.
	 */
	argv++;
	argc--;

	while ((argc > 0) && (**argv == '-'))
	{
		cp = *argv++ + 1;
		argc--;

		while (*cp) switch (*cp++)
		{
			case 'c':
				/*
				 * Execute specified command.
				 */
				if ((argc != 1) || singleCommand)
					usage();

				singleCommand = *argv++;
				argc--;

				break;

			case 'p':
				/*
				 * Set the prompt string.
				 */
				if ((argc <= 0) || (**argv == '-'))
					usage();

				if (prompt)
					free(prompt);

				prompt = strdup(*argv++);
				argc--;

				break;

			case 'h':
			case '?':
				usage();
				break;

			default:
				fprintf(stderr, "Unknown option -%c\n", cp[-1]);

				return 1;
		}
	}

	/*
	 * No more arguments are allowed.
	 */
	if (argc > 0)
		usage();

	/*
	 * Default our path if it is not set.
	 */
	if (getenv("PATH") == NULL)
		putenv("PATH=/bin:/usr/bin:/sbin:/usr/sbin:/etc");

	/*Angus ++ 2000/9/2
	 *Set USER enviroment in /etc/profile
	 */
	system("/bin/sh /etc/profile");

	if (singleCommand)
	{
		command(singleCommand);

		return 0;
	}

	signal(SIGINT, catchInt);
	signal(SIGQUIT, catchQuit);

	/*
	 * Execute the user's alias file if present.
	 */
	cp = getenv("HOME");

	if (cp)
	{
		strcpy(buf, cp);
		strcat(buf, "/");
		strcat(buf, ".aliasrc");

		if ((access(buf, 0) == 0) || (errno != ENOENT))
			readFile(buf);
	}

	/*
	 * Read commands from stdin.
	 */
	readFile(NULL);

	return 0;
}

static void
alrm_catcher(int i)
{
  fprintf (stderr, "Timed out logout!\n");
  exit (0);
}


/*
 * Read commands from the specified file.
 * A null name pointer indicates to read from stdin.
 */
static void
readFile(const char * name)
{
	FILE *	fp;
	int	cc;
	BOOL	ttyFlag;
	char	buf[CMD_LEN];
	int     tmout_len = 60;
	void    *old_alrm;


	if (sourceCount >= MAX_SOURCE)
	{
		fprintf(stderr, "Too many source files\n");

		return;
	}

	fp = stdin;

	if (name)
	{
		fp = fopen(name, "r");

		if (fp == NULL)
		{
			perror(name);

			return;
		}
	}

	sourcefiles[sourceCount++] = fp;

	ttyFlag = isatty(fileno(fp));


	while (TRUE)
	{

		if (ttyFlag) {
			showPrompt();
                }

		if (intFlag && !ttyFlag && (fp != stdin))
		{
			fclose(fp);
			sourceCount--;

			return;
		} 	



    {
      old_alrm = (void *)NULL;

	    {
	      old_alrm = signal(SIGALRM, alrm_catcher);
	      alarm (tmout_len);
	    }
    }


		if (fgets(&buf[0], CMD_LEN - 1, fp) == NULL)
		{

			if (ferror(fp) && (errno == EINTR))
			{
				clearerr(fp);

				continue;
			}

			break;
		}

    		{
      			alarm(0);
      			signal (SIGALRM, old_alrm);
    		}

		cc = strlen(buf);

		if (buf[cc - 1] == '\n')
			cc--;

		while ((cc > 0) && isBlank(buf[cc - 1]))
			cc--;

		buf[cc] = '\0';

		command(buf);
	}

	if (ferror(fp))
	{
		perror("Reading command line");

		if (fp == stdin)
			exit(1);
	}

	clearerr(fp);

	if (fp != stdin)
		fclose(fp);

	sourceCount--;
}


/*
 * Parse and execute one null-terminated command line string.
 * This breaks the command line up into words, checks to see if the
 * command is an alias, and expands wildcards.
 */
static void
command(const char * cmd)
{
	const char *	endCmd;
	char		cmdName[CMD_LEN];

	/*
	 * Rest the interrupt flag and free any memory chunks that
	 * were allocated by the previous command.
	 */
	intFlag = FALSE;

	freeChunks();

	/*
	 * Skip leading blanks.
	 */
	while (isBlank(*cmd))
		cmd++;

	/*
	 * If the command is empty or is a comment then ignore it.
	 */
	if ((*cmd == '\0') || (*cmd == '#'))
		return;

	/*
	 * Look for the end of the command name and then copy the
	 * command name to a buffer so we can null terminate it.
	 */
	endCmd = cmd;

	while (*endCmd && !isBlank(*endCmd))
		endCmd++;

	memcpy(cmdName, cmd, endCmd - cmd);

	cmdName[endCmd - cmd] = '\0';

	/*
	 * Now look for the command in the builtin table, and execute
	 * the command if found.
	 */
	if (tryBuiltIn(cmd))
		return;

	/*
	 * The command is not a built-in, so run the program along
	 * the PATH list.
	 */
	runCmd(cmd);
}


/*
 * Try to execute a built-in command.
 * Returns TRUE if the command is a built in, whether or not the
 * command succeeds.  Returns FALSE if this is not a built-in command.
 */
static BOOL
tryBuiltIn(const char * cmd)
{
	const char *		endCmd;
	const CommandEntry *	entry;
	int			argc;
	const char **		argv;
	char			cmdName[CMD_LEN];

	/*
	 * Look for the end of the command name and then copy the
	 * command name to a buffer so we can null terminate it.
	 */
	endCmd = cmd;

	while (*endCmd && !isBlank(*endCmd))
		endCmd++;

	memcpy(cmdName, cmd, endCmd - cmd);

	cmdName[endCmd - cmd] = '\0';

	/*
	 * Search the command table looking for the command name.
	 */
	for (entry = commandEntryTable; entry->name != NULL; entry++)
	{
		if (strcmp(entry->name, cmdName) == 0)
			break;
	}

	/*
	 * If the command is not a built-in, return indicating that.
	 */
	if (entry->name == NULL)
		return FALSE;

	/*
	 * The command is a built-in.
	 * Break the command up into arguments and expand wildcards.
	 */
	if (!makeArgs(cmd, &argc, &argv))
		return TRUE;

	/*
	 * Give a usage string if the number of arguments is too large
	 * or too small.
	 */
	if ((argc < entry->minArgs) || (argc > entry->maxArgs))
	{
		fprintf(stderr, "usage: %s %s\n", entry->name, entry->usage);

		return TRUE;
	}

	/*
	 * Call the built-in function with the argument list.
	 */
	entry->func(argc, argv);

	return TRUE;
}

const char *cli_paser = "/usr/local/apache/cgi-bin/cli.pl";

#define CLI_CMD_LEN (CMD_LEN + 256)
/*
 * Execute the specified command either by forking and executing
 * the program ourself, or else by using the shell.
 */
static void
runCmd(const char * cmd)
{
	char cmdline[CLI_CMD_LEN];

	sprintf((char *)&cmdline[0], "%s %s", cli_paser, cmd);

	//WriteCmdLine(cmdline);

	system(cmdline);

	return;
}

void
do_help(int argc, const char ** argv)
{
	const CommandEntry *	entry;
	const char *		str;

	str = NULL;

	if (argc == 2)
		str = argv[1];

	/*
	 * Check for an exact match, in which case describe the program.
	 */
	if (str)
	{
		for (entry = commandEntryTable; entry->name; entry++)
		{
			if (strcmp(str, entry->name) == 0)
			{
				printf("%s\n", entry->description);

				printf("usage: %s %s\n", entry->name,
					entry->usage);

				return;
			}
		}
	}

	/*
	 * Print short information about commands which contain the
	 * specified word.
	 */
	for (entry = commandEntryTable; entry->name; entry++)
	{
		if ((str == NULL) || (strstr(entry->name, str) != NULL) ||
			(strstr(entry->usage, str) != NULL))
		{
			printf("%-10s %s\n", entry->name, entry->usage);
		}
	}
}

void
do_prompt(int argc, const char ** argv)
{
	char *	cp;
	char	buf[CMD_LEN];

	if (!makeString(argc - 1, argv + 1, buf, CMD_LEN))
		return;

	cp = malloc(strlen(buf) + 2);

	if (cp == NULL)
	{
		fprintf(stderr, "No memory for prompt\n");

		return;
	}

	strcpy(cp, buf);
	strcat(cp, " ");

	if (prompt)
		free(prompt);

	prompt = cp;
}


/*
 * Display the prompt string.
 */
static void
showPrompt(void)
{
	const char *	cp;

	cp = "->";

	if (prompt)
		cp = prompt;

	write(STDOUT, cp, strlen(cp));
}	


static void
catchInt(int val)
{
	signal(SIGINT, catchInt);

	intFlag = TRUE;

	if (intCrlf)
		write(STDOUT, "\n", 1);
}


static void
catchQuit(int val)
{
	signal(SIGQUIT, catchQuit);

	intFlag = TRUE;

	if (intCrlf)
		write(STDOUT, "\n", 1);
}


/*
 * Print the usage information and quit.
 */
static void
usage(void)
{
	fprintf(stderr, "Stand-alone shell (version %s)\n", version);
	fprintf(stderr, "\n");
	fprintf(stderr, "Usage: sash [-a] [-q] [-c command] [-p prompt]\n");

	exit(1);
}

/* END CODE */

#define CMD_LINE_CFG "/etc/sysconfig/.firewall/cli.conf"

static void 
WriteCmdLine(const char * cmd)
{
	FILE *fp;

	fp = fopen(CMD_LINE_CFG, "w");	

	if (fp == NULL)
	{
		perror(CMD_LINE_CFG);
		return;
	}
	fprintf(fp, "<Input>\n");
	fprintf(fp, "%s\n", cmd);
	fprintf(fp, "</Input>\n");

	fclose(fp);

}


static void
do_exit(int argc, const char ** argv)
{
	if (getpid() == 1)
	{
		fprintf(stderr, "You are the INIT process!\n");

		return;
	}
	system("/bin/login");

	exit(0);
}
const char *passwd = "htpasswd";
const char *usermod = "usermod -p";
const char *conf = "/etc/sysconfig/.firewall/.htpasswd";


static void
do_passwd(int argc, const char ** argv)
{
	const char *user;
	char cmd[256];
	char str[265];
	const char *crypt;
	FILE *fp;

	//user = getenv("USER");
	user = "admin";

	sprintf((char *)&cmd[0], "%s %s %s", passwd, conf, user);
	system(cmd);


	fp = fopen(conf, "r");	

	if (fp == NULL)
	{
		perror(conf);
		return;
	}

	while (!feof(fp))
	{
		fscanf(fp, "%s", &str[0]);
		if (memcmp(str, user, strlen(user)) == 0)
			break; 
	}

	crypt = strchr(str, ':');
	crypt++;

	sprintf((char *)&cmd[0], "%s %s %s", usermod, crypt, user);
	system(cmd);	

	fclose(fp);


	return;

}

const char *ping = "ping -c 4";

static void
do_ping(int argc, const char ** argv)
{
	char cmd[256];

	if (strchr(argv[1], '.') == NULL)
	{
		fprintf(stderr, "usage: ping host(ip address)\n"); 
		return;
	}
	sprintf((char *)&cmd[0], "%s %s", ping, argv[1]);
	system(cmd);

	return;
}

const char *sn = "/etc/sysconfig/.firewall/sn.conf";

static void
do_catsn(int argc, const char ** argv)
{
	char cmd[256];

	sprintf((char *)&cmd[0], "/bin/cat %s", sn);
	system(cmd);

	return;
}


