/* Shared library add-on to iptables to add ALARM support. */
#include <stdio.h>
#include <netdb.h>
#include <string.h>
#include <stdlib.h>
#include <syslog.h>
#include <getopt.h>
#include <iptables.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ipt_ALARM.h>

#define ALARM_DEFAULT_LEVEL LOG_NOTICE

/* Function which prints out usage message. */
static void
help(void)
{
	printf(
"ALARM v%s options:\n"
" --alarm-level level		Level of alarming (numeric or see syslog.conf)\n",
NETFILTER_VERSION);
}

static struct option opts[] = {
	{ "alarm-level", 1, 0, '!' },
	{ 0 }
};

/* Initialize the target. */
static void
init(struct ipt_entry_target *t, unsigned int *nfcache)
{
	struct ipt_alarm_info *alarminfo = (struct ipt_alarm_info *)t->data;

	alarminfo->level = ALARM_DEFAULT_LEVEL;

	/* Can't cache this */
	*nfcache |= NFC_UNKNOWN;
}

struct ipt_alarm_names {
	const char *name;
	unsigned int level;
};

static struct ipt_alarm_names ipt_alarm_names[]
= { { "alert", LOG_ALERT },
    { "crit", LOG_CRIT },
    { "debug", LOG_DEBUG },
    { "emerg", LOG_EMERG },
    { "error", LOG_ERR },		/* DEPRECATED */
    { "info", LOG_INFO },
    { "notice", LOG_NOTICE },
    { "panic", LOG_EMERG },		/* DEPRECATED */
    { "warning", LOG_WARNING }
};

static u_int8_t
parse_level(const char *level)
{
	unsigned int lev = -1;
	unsigned int set = 0;

	if (string_to_number(level, 0, 7, &lev) == -1) {
		unsigned int i = 0;

		for (i = 0;
		     i < sizeof(ipt_alarm_names) / sizeof(struct ipt_alarm_names);
		     i++) {
			if (strncasecmp(level, ipt_alarm_names[i].name,
					strlen(level)) == 0) {
				if (set++)
					exit_error(PARAMETER_PROBLEM,
						   "alarm-level `%s' ambiguous",
						   level);
				lev = ipt_alarm_names[i].level;
			}
		}

		if (!set)
			exit_error(PARAMETER_PROBLEM,
				   "alarm-level `%s' unknown", level);
	}

	return (u_int8_t)lev;
}

#define IPT_ALARM_OPT_LEVEL 0x01

/* Function which parses command options; returns true if it
   ate an option */
static int
parse(int c, char **argv, int invert, unsigned int *flags,
      const struct ipt_entry *entry,
      struct ipt_entry_target **target)
{
	struct ipt_alarm_info *alarminfo = (struct ipt_alarm_info *)(*target)->data;

	switch (c) {
	case '!':
		if (*flags & IPT_ALARM_OPT_LEVEL)
			exit_error(PARAMETER_PROBLEM,
				   "Can't specify --alarm-level twice");

		if (check_inverse(optarg, &invert))
			exit_error(PARAMETER_PROBLEM,
				   "Unexpected `!' after --alarm-level");

		alarminfo->level = parse_level(optarg);
		*flags |= IPT_ALARM_OPT_LEVEL;
		break;

	default:
		return 0;
	}

	return 1;
}

/* Final check; nothing. */
static void final_check(unsigned int flags)
{
}

/* Prints out the targinfo. */
static void
print(const struct ipt_ip *ip,
      const struct ipt_entry_target *target,
      int numeric)
{
	const struct ipt_alarm_info *alarminfo
		= (const struct ipt_alarm_info *)target->data;
	unsigned int i = 0;

	printf("ALARM ");
	if (numeric)
		printf("level %u ", alarminfo->level);
	else {
		for (i = 0;
		     i < sizeof(ipt_alarm_names) / sizeof(struct ipt_alarm_names);
		     i++) {
			if (alarminfo->level == ipt_alarm_names[i].level) {
				printf("level %s ", ipt_alarm_names[i].name);
				break;
			}
		}
		if (i == sizeof(ipt_alarm_names) / sizeof(struct ipt_alarm_names))
			printf("UNKNOWN level %u ", alarminfo->level);
	}
}

/* Saves the union ipt_targinfo in parsable form to stdout. */
static void
save(const struct ipt_ip *ip, const struct ipt_entry_target *target)
{
	const struct ipt_alarm_info *alarminfo
		= (const struct ipt_alarm_info *)target->data;

	if (alarminfo->level != ALARM_DEFAULT_LEVEL)
		printf("--alarm-level %u ", alarminfo->level);
}

static
struct iptables_target alarm
= { NULL,
    "ALARM",
    NETFILTER_VERSION,
    IPT_ALIGN(sizeof(struct ipt_alarm_info)),
    IPT_ALIGN(sizeof(struct ipt_alarm_info)),
    &help,
    &init,
    &parse,
    &final_check,
    &print,
    &save,
    opts
};

void _init(void)
{
	register_target(&alarm);
}
