/* Shared library add-on to iptables to add MARK target support. */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>

#include <iptables.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ipt_VIPMARK.h>

struct markinfo {
	struct ipt_entry_target t;
	struct ipt_vipmark_target_info mark;
};

/* Function which prints out usage message. */
static void
help(void)
{
	printf(
"VIPMARK target v%s options:\n"
"  --set-mark value                   Set nfmark value\n"
"  --and-mark value                    And nfmark value\n"
"  --or-mark value                   Or nfmark value\n"
"\n",
NETFILTER_VERSION);
}

static struct option opts[] = {
	{ "set-mark", 1, 0, '1' },
	{ "and-mark", 1, 0, '2' },
	{ "or-mark", 1, 0, '3' },
	{ 0 }
};

/* Initialize the target. */
static void
init(struct ipt_entry_target *t, unsigned int *nfcache)
{
}

/* Function which parses command options; returns true if it
   ate an option */
static int
parse(int c, char **argv, int invert, unsigned int *flags,
      const struct ipt_entry *entry,
      struct ipt_entry_target **target)
{
	struct ipt_vipmark_target_info *markinfo
		= (struct ipt_vipmark_target_info *)(*target)->data;

	switch (c) {
		char *end;
	case '1':
		markinfo->mark = strtoull(optarg, &end, 0);
		if (*end != '\0' || end == optarg)
			exit_error(PARAMETER_PROBLEM, "Bad VIPMARK value `%s'", optarg);
		if (*flags)
			exit_error(PARAMETER_PROBLEM,
			           "MARK target: Can't specify --set-mark twice");
		*flags = 1;
		break;

	case '2':
		markinfo->mark = strtoull(optarg, &end, 0);
		if (*end != '\0' || end == optarg)
			exit_error(PARAMETER_PROBLEM, "Bad VIPMARK value `%s'", optarg);
		if (*flags)
			exit_error(PARAMETER_PROBLEM,
			           "MARK target: Can't specify --set-mark twice");
		*flags = 1;
		break;

	case '3':
		markinfo->mark = strtoull(optarg, &end, 0);
		if (*end != '\0' || end == optarg)
			exit_error(PARAMETER_PROBLEM, "Bad VIPMARK value `%s'", optarg);
		if (*flags)
			exit_error(PARAMETER_PROBLEM,
			           "MARK target: Can't specify --set-mark twice");
		*flags = 1;
		break;

	default:
		return 0;
	}

	return 1;
}

static void
final_check(unsigned int flags)
{
	if (!flags)
		exit_error(PARAMETER_PROBLEM,
		           "VIPMARK target: Parameter --set-mark is required");
}

static void
print_mark(unsigned long long mark, int numeric)
{
	printf("0x%llx ", mark);
}

/* Prints out the targinfo. */
static void
print(const struct ipt_ip *ip,
      const struct ipt_entry_target *target,
      int numeric)
{
	const struct ipt_vipmark_target_info *markinfo =
		(const struct ipt_vipmark_target_info *)target->data;

	switch (markinfo->mode) {
	case IPT_MARK_SET:
		printf("VIPMARK set ");
		print_mark(markinfo->mark, numeric);
		break;
	case IPT_MARK_AND:
		printf("VIPMARK and ");
		print_mark(markinfo->mark, numeric);
		break;
	case IPT_MARK_OR:
		printf("VIPMARK or ");
		print_mark(markinfo->mark, numeric);
		break;
	default:
		printf("ERROR: UNKNOWN MARK MODE ");
		break;
	}
}

/* Saves the union ipt_targinfo in parsable form to stdout. */
static void
save(const struct ipt_ip *ip, const struct ipt_entry_target *target)
{
	const struct ipt_vipmark_target_info *markinfo =
		(const struct ipt_vipmark_target_info *)target->data;

        switch (markinfo->mode) {
        case IPT_MARK_SET:
		printf("--set-mark 0x%llx ", markinfo->mark);
		break;
        case IPT_MARK_AND:
		printf("--and-mark 0x%llx ", markinfo->mark);
		break;
        case IPT_MARK_OR:
		printf("--or-mark 0x%llx ", markinfo->mark);
		break;
        default:
		printf("ERROR: UNKNOWN MARK MODE ");
		break;
	}
}

static
struct iptables_target vipmark
= { NULL,
    "VIPMARK",
    NETFILTER_VERSION,
    IPT_ALIGN(sizeof(struct ipt_vipmark_target_info)),
    IPT_ALIGN(sizeof(struct ipt_vipmark_target_info)),
    &help,
    &init,
    &parse,
    &final_check,
    &print,
    &save,
    opts
};

void _init(void)
{
	register_target(&vipmark);
}
