/* Shared library add-on to iptables to add NFMARK matching support. */
#include <stdio.h>
#include <netdb.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>

#include <iptables.h>
#include <linux/netfilter_ipv4/ipt_connlimit.h>

/* Function which prints out usage message. */
static void
help(void)
{
	printf(
		"connlimit match v%s options:\n"
		"--connlimit-mark  value         Match nfmark value\n"
		"--connlimit-limit limit         Above connection limit\n"
		"\n",
		NETFILTER_VERSION);
}

static struct option opts[] = {
	{ "connlimit-mark", 1, 0, '1' },
	{ "connlimit-limit", 1, 0, '2' },
	{0}
};

#define		IPT_CONNLIMIT_MARK	1
#define		IPT_CONNLIMIT_LIMIT	2

/* Initialize the match. */
static void
init(struct ipt_entry_match *m, unsigned int *nfcache)
{
	/* Can't cache this. */
	*nfcache |= NFC_UNKNOWN;
}

/* Function which parses command options; returns true if it
   ate an option */
static int
parse(int c, char **argv, int invert, unsigned int *flags,
      const struct ipt_entry *entry,
      unsigned int *nfcache,
      struct ipt_entry_match **match)
{
	struct ipt_connlimit_info *markinfo = (struct ipt_connlimit_info *)(*match)->data;

	switch (c) {
		char *end;
		case '1':
			markinfo->mark = strtoul(optarg, &end, 0);
			if (*end != '\0' || end == optarg)
				exit_error(PARAMETER_PROBLEM, "Bad mark value `%s'", optarg);
			*flags |= IPT_CONNLIMIT_MARK;
			break;
			
		case '2':
			markinfo->limit = strtoul(optarg, &end, 0);
			if (*end != '\0' || end == optarg)
				exit_error(PARAMETER_PROBLEM, "Bad above-limit value `%s'", optarg);
			*flags |= IPT_CONNLIMIT_LIMIT;
			break;
			
		default:
			return 0;
	}
	
	return 1;
}

static void
print_connlimit(unsigned long mark, unsigned long limit, int numeric)
{
	printf("mark: 0x%lx limit: %lu ", mark, limit);
}

/* Final check; must have specified --mark. */
static void
final_check(unsigned int flags)
{
	if (!(flags & IPT_CONNLIMIT_MARK))
		exit_error(PARAMETER_PROBLEM,
			   "connlimit match: You must specify `--mark'");
	if (!(flags & IPT_CONNLIMIT_LIMIT))
		exit_error(PARAMETER_PROBLEM,
			   "connlimit match: You must specify `--above-limit'");
}

/* Prints out the matchinfo. */
static void
print(const struct ipt_ip *ip,
      const struct ipt_entry_match *match,
      int numeric)
{
	printf("connlimit match ");
	print_connlimit(((struct ipt_connlimit_info *)match->data)->mark,
					((struct ipt_connlimit_info *)match->data)->limit,
					numeric);
}

/* Saves the union ipt_matchinfo in parsable form to stdout. */
static void
save(const struct ipt_ip *ip, const struct ipt_entry_match *match)
{
	printf("--connlimit ");
	print_connlimit(((struct ipt_connlimit_info *)match->data)->mark,
		        ((struct ipt_connlimit_info *)match->data)->limit,
		        0);
}

static
struct iptables_match connlimit
= { NULL,
    "connlimit",
    NETFILTER_VERSION,
    IPT_ALIGN(sizeof(struct ipt_connlimit_info)),
    IPT_ALIGN(sizeof(struct ipt_connlimit_info)),
    &help,
    &init,
    &parse,
    &final_check,
    &print,
    &save,
    opts
};

void _init(void)
{
	register_match(&connlimit);
}
