/* ====================================================================
 * Copyright (c) 1995-1999 The Apache Group.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * 4. The names "Apache Server" and "Apache Group" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * THIS SOFTWARE IS PROVIDED BY THE APACHE GROUP ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Group and was originally based
 * on public domain software written at the National Center for
 * Supercomputing Applications, University of Illinois, Urbana-Champaign.
 * For more information on the Apache Group and the Apache HTTP server
 * project, please see <http://www.apache.org/>.
 *
 */

/*
 * http_auth: authentication
 * 
 * Rob McCool
 * 
 * Adapted to Apache by rst.
 *
 * dirkx - Added Authoritative control to allow passing on to lower
 *         modules if and only if the user-id is not known to this
 *         module. A known user with a faulty or absent password still
 *         causes an AuthRequired. The default is 'Authoritative', i.e.
 *         no control is passed along.
 */

#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"

/* 2002/3/19 Steve ++ */
#include "util_md5.h"
#include "ap_compat.h"

#ifndef FALSE
#define FALSE 0
#endif

#ifndef TRUE
#define TRUE !FALSE
#endif
/* Steve -- */

typedef struct auth_config_struct {
    char *auth_pwfile;
    char *auth_grpfile;
    int auth_authoritative;
/* 2002/3/19 Steve ++ */
    int auth_expire;
/* Steve -- */
/* 2003/3/17 Steve ++ */
    int auth_fw;
/* Steve -- */
} auth_config_rec;

static void *create_auth_dir_config(pool *p, char *d)
{
    auth_config_rec *sec =
    (auth_config_rec *) ap_pcalloc(p, sizeof(auth_config_rec));
    sec->auth_pwfile = NULL;	/* just to illustrate the default really */
    sec->auth_grpfile = NULL;	/* unless you have a broken HP cc */
    sec->auth_authoritative = 1;	/* keep the fortress secure by default */
/* 2002/3/19 Steve ++ */
    sec->auth_expire = 0;  
/* Steve -- */
/* 2003/3/17 Steve ++ */
    sec->auth_fw = 0;
/* Steve -- */
    return sec;
}

static const char *set_auth_slot(cmd_parms *cmd, void *offset, char *f, char *t)
{
    if (t && strcmp(t, "standard"))
	return ap_pstrcat(cmd->pool, "Invalid auth file type: ", t, NULL);

    return ap_set_file_slot(cmd, offset, f);
}

/* 2002/3/19 Steve ++ */
static const char *set_auth_expire(cmd_parms *cmd, void *offset, char *f, char *t)
{
    auth_config_rec *s = (auth_config_rec *)offset;
    if (t && strcmp(t, "standard"))
        return ap_pstrcat(cmd->pool, "Invalid auth file type: ", t, NULL);

    s->auth_expire = atoi(f);
    return NULL;
}
/* Steve -- */

static const command_rec auth_cmds[] =
{
    {"AuthUserFile", set_auth_slot,
     (void *) XtOffsetOf(auth_config_rec, auth_pwfile), OR_AUTHCFG, TAKE12,
     "text file containing user IDs and passwords"},
    {"AuthGroupFile", set_auth_slot,
     (void *) XtOffsetOf(auth_config_rec, auth_grpfile), OR_AUTHCFG, TAKE12,
     "text file containing group names and member user IDs"},
    {"AuthAuthoritative", ap_set_flag_slot,
     (void *) XtOffsetOf(auth_config_rec, auth_authoritative),
     OR_AUTHCFG, FLAG,
     "Set to 'no' to allow access control to be passed along to lower modules if the UserID is not known to this module"},
/* 2002/3/19 Steve ++ */
    {"AuthExpire", set_auth_expire,
     (void *) XtOffsetOf(auth_config_rec, auth_expire), OR_AUTHCFG, TAKE12,
     "Set auth expire time"},
/* Steve -- */
/* 2003/3/17 Steve ++ */
    {"AuthForward", ap_set_flag_slot,
     (void *) XtOffsetOf(auth_config_rec, auth_fw),
     OR_AUTHCFG, FLAG,
     "Set Forward Auth"},
/* Steve -- */
    {NULL}
};

module MODULE_VAR_EXPORT auth_module;

static char *get_pw(request_rec *r, char *user, char *auth_pwfile)
{
    configfile_t *f;
    char l[MAX_STRING_LEN];
    const char *rpw, *w;

    if (!(f = ap_pcfg_openfile(r->pool, auth_pwfile))) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
		    "Could not open password file: %s", auth_pwfile);
	return NULL;
    }
    while (!(ap_cfg_getline(l, MAX_STRING_LEN, f))) {
	if ((l[0] == '#') || (!l[0]))
	    continue;
	rpw = l;
	w = ap_getword(r->pool, &rpw, ':');

	if (!strcmp(user, w)) {
	    ap_cfg_closefile(f);
	    return ap_getword(r->pool, &rpw, ':');
	}
    }
    ap_cfg_closefile(f);
    return NULL;
}

static table *groups_for_user(pool *p, char *user, char *grpfile)
{
    configfile_t *f;
    table *grps = ap_make_table(p, 15);
    pool *sp;
    char l[MAX_STRING_LEN];
    const char *group_name, *ll, *w;

    if (!(f = ap_pcfg_openfile(p, grpfile))) {
/*add?	aplog_error(APLOG_MARK, APLOG_ERR, NULL,
		    "Could not open group file: %s", grpfile);*/
	return NULL;
    }

    sp = ap_make_sub_pool(p);

    while (!(ap_cfg_getline(l, MAX_STRING_LEN, f))) {
	if ((l[0] == '#') || (!l[0]))
	    continue;
	ll = l;
	ap_clear_pool(sp);

	group_name = ap_getword(sp, &ll, ':');

	while (ll[0]) {
	    w = ap_getword_conf(sp, &ll);
	    if (!strcmp(w, user)) {
		ap_table_setn(grps, ap_pstrdup(p, group_name), "in");
		break;
	    }
	}
    }
    ap_cfg_closefile(f);
    ap_destroy_pool(sp);
    return grps;
}

/* 2002/3/19 Steve ++ */
#define COOKIE_SIZE 1024
/* make a cookie based on secret + public information */
static char *
make_cookie(request_rec *r, time_t expires, const char *passwd, const char *string)
{
  char one[COOKIE_SIZE], two[COOKIE_SIZE];
  char *cookie = palloc(r->pool, COOKIE_SIZE);
  conn_rec *c = r->connection;
  server_rec *s = r->server;

  /*
   * Arg! We can't use 'ntohs(c->remote_addr.sin_port)', because I.E.
   * ignores keepalives, and opens a new connection on EVERY request!
   * This is a BAD security problem!  It allows multiple users on the
   * same machine to access the data.
   *
   * A derivative security problem is users authenticating from
   * behind a firewall.
   * All users appear to be coming from the firewall.  A malicious
   * agent working in the same company as the authorized user can sniff
   * the cookie, and and use it themselves.  Since they appear to be
   * coming from the same IP address (firewall), they're let in.
   * Oh well, at least the connection is traceable to a particular machine.
   */

  /*
   * Make a cookie based on secret + public information.
   *
   * cookie = MAC(M) = MD5(secret, MD5(secret, M))
   *
   * See Scheier, B, "Applied Cryptography" 2nd Ed., p.458
   * Also, RFC 2104.  I don't know if the HMAC gives any additional
   * benefit here.
   */
  ap_snprintf(one, COOKIE_SIZE, "%s%s%s%08x",
              c->user, passwd, c->remote_ip, expires);

  /* MD5 the cookie to make it secure, and add more secret information */
  ap_snprintf(two, COOKIE_SIZE, "%s", ap_md5(r->pool, one));
  if (string == NULL) {
    ap_snprintf(cookie, COOKIE_SIZE, "%s%08x",
                ap_md5(r->pool, two), expires);
  } else {
    ap_snprintf(cookie, COOKIE_SIZE, "%s%08x%s",
                ap_md5(r->pool, two), expires, string);
  }
  return cookie;
}

/* Add a cookie to an outgoing request */
static const char *cookie_name = "Firewall";
static void
add_cookie(request_rec *r, table *header, char *cookie, time_t expires)
{
  char *new_cookie = palloc(r->pool, COOKIE_SIZE); /* so it'll stick around */

  if (expires != 0) {
    ap_snprintf(new_cookie, 1024, "%s=%s; path=/;", cookie_name, cookie);
  } else {
    ap_snprintf(new_cookie, 1024,
                "%s=%s; path=/; expires=Wed, 01-Oct-97 01:01:01 GMT;",
                cookie_name, cookie);
  }

  table_set(header,"Set-Cookie", new_cookie);
}

static char *
spot_cookie(request_rec *r)
{
  const char *cookie;
  char *value;

  if ((cookie = table_get (r->headers_in, "Cookie"))) {
    if ((value=strstr(cookie, cookie_name))) {
      char *cookiebuf, *cookieend;

      value += strlen(cookie_name) + 1; /* skip '=' */
      cookiebuf = pstrdup( r->pool, value );
      cookieend = strchr(cookiebuf,';');
      if (cookieend) *cookieend = '\0'; /* Ignore anything after a ; */

      /* Set the cookie in a note, for logging */
      return cookiebuf;          /* Theres already a cookie, no new one */
    }
  }
  return NULL;                        /* no cookie was found */
}

static int
valid_cookie(request_rec *r, const char *cookie, const char *passwd)
{
  time_t expires, now;

  if (strlen(cookie) < (16 + 4)*2) { /* MD5 is 16 bytes, and expiry date is 4*/
    return FALSE;               /* invalid */
  }

  sscanf(&cookie[32], "%8lx", &expires);

  now = time(NULL);
  if (expires < now) {  /* valid only for a short window of time */
    return FALSE;               /* invalid: expired */
  }

  /* Is the returned cookie identical to one made from our secret? */
  if (strcmp(cookie, make_cookie(r, expires, passwd, NULL)) == 0)
    return TRUE;

  return FALSE;                 /* cookie doesn't match: re-validate */
}

#if 0
static void
note_challenge_auth_failure(request_rec *r, char *user, char *message)
{
    if (!*message) {            /* no message to print */
      ap_note_basic_auth_failure(r);
    } else {                    /* print our magic message */
      table_set (r->err_headers_out, "WWW-Authenticate",
                 pstrcat(r->pool, "Basic realm=\"", auth_name(r), " for ", user, " '", message, "'", NULL));
    }
}
#endif
/* Steve -- */

/* These functions return 0 if client is OK, and proper error status
 * if not... either AUTH_REQUIRED, if we made a check, and it failed, or
 * SERVER_ERROR, if things are so totally confused that we couldn't
 * figure out how to tell if the client is authorized or not.
 *
 * If they return DECLINED, and all other modules also decline, that's
 * treated by the server core as a configuration error, logged and
 * reported as such.
 */

/* Determine user ID, and check if it really is that user, for HTTP
 * basic authentication...
 */
/* 2003/3/17 Steve ++ */
static char *authifname[] =
{
        "internal",
        "external",
        "DMZ"
};
/* Steve -- */

static int authenticate_basic_user(request_rec *r)
{
    auth_config_rec *sec =
    (auth_config_rec *) ap_get_module_config(r->per_dir_config, &auth_module);
    conn_rec *c = r->connection;
    const char *sent_pw;
    char *real_pw;
    char *invalid_pw;
    int res;
    /* 2002/3/19 Steve ++ */
    time_t authSec = sec->auth_expire * 60;
    time_t expires = time(NULL) + authSec;
    char *cookie;
    static int eventlog = 0;
    char auth_remote_ip[256];
    /* Steve -- */

    if ((res = ap_get_basic_auth_pw(r, &sent_pw))) {
	eventlog = 1;
	return res;
    }

    if (!sec->auth_pwfile)
	return DECLINED;

/* 2002/3/19 Steve ++ */
	cookie = spot_cookie(r);
	if (cookie && !strncmp(cookie, "logout", 6)) {
		if (!strcmp(r->uri, "/cgi-bin/logout.cgi"))
			return OK;
		else
			return HTTP_LOGOUT;
	} else if (!strcmp(r->uri, "/cgi-bin/logout.cgi")) {
		if (!r->args || strcmp(r->args, "type=logout"))
			return HTTP_LOGOUT;
	}

	if (sec->auth_expire && cookie) {
		if (strncmp(cookie, "expire", 6) &&
		    !strcmp(r->uri, "/cgi-bin/expire.cgi"))
			return HTTP_AUTH_EXPIRE;

		if (!valid_cookie(r, cookie, sent_pw)) {
               		if (!strcmp(r->uri, "/cgi-bin/expire.cgi"))
                       		return OK;
               		else
                       		return HTTP_AUTH_EXPIRE;
		}
	}
#if 0
	if (sec->auth_expire) {
		if (cookie && !valid_cookie(r, cookie, sent_pw)) {
               		if (!strcmp(r->uri, "/cgi-bin/expire.cgi"))
                       		return OK;
               		else
                       		return HTTP_AUTH_EXPIRE;
		}
	}
#endif
/* Steve -- */

    if (!(real_pw = get_pw(r, c->user, sec->auth_pwfile))) {
	if (!(sec->auth_authoritative))
	    return DECLINED;
	ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
		    "user %s not found: %s", c->user, r->uri);
	ap_note_basic_auth_failure(r);
	if (!sec->auth_fw)
        	syslog(LOG_INFO | LOG_LOCAL0, "admin user %s [Login failure] from %s\n",
                	c->user,
                	c->remote_ip
        	);
	return AUTH_REQUIRED;
    }

    invalid_pw = ap_validate_password(sent_pw, real_pw);
    if (invalid_pw != NULL) {
	ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
		      "user %s: authentication failure for \"%s\": %s",
		      c->user, r->uri, invalid_pw);
	ap_note_basic_auth_failure(r);
/* 2002/8/7 Steve ++ */
	if (!sec->auth_fw)
	        syslog(LOG_INFO | LOG_LOCAL0, "admin user %s [Login failure] from %s\n",
        	        c->user,
        	        c->remote_ip
        	);
/* Steve -- */
	return AUTH_REQUIRED;
    }

    if (eventlog && !sec->auth_fw) {
    	syslog(LOG_INFO | LOG_LOCAL0, "admin user %s [Login success] from %s",
               c->user,
               c->remote_ip
        );
        eventlog = 0;
    }

/* 2002/3/19 Steve ++ */
	cookie = make_cookie(r, expires, sent_pw, NULL);
	add_cookie(r, r->headers_out, cookie, expires);
/* Steve -- */

/* 2003/3/17 Steve ++ */
	if (sec->auth_fw) {
		sprintf(auth_remote_ip, "echo \"+%s\" > /proc/net/ipt_recent/%s",
			c->remote_ip,
			authifname[c->localid]);
		system(auth_remote_ip);
	}
/* Steve -- */

    return OK;
}

/* Checking ID */

static int check_user_access(request_rec *r)
{
    auth_config_rec *sec =
    (auth_config_rec *) ap_get_module_config(r->per_dir_config, &auth_module);
    char *user = r->connection->user;
    int m = r->method_number;
    int method_restricted = 0;
    register int x;
    const char *t, *w;
    table *grpstatus;
    const array_header *reqs_arr = ap_requires(r);
    require_line *reqs;

    /* BUG FIX: tadc, 11-Nov-1995.  If there is no "requires" directive, 
     * then any user will do.
     */
    if (!reqs_arr)
	return (OK);
    reqs = (require_line *) reqs_arr->elts;

    if (sec->auth_grpfile)
	grpstatus = groups_for_user(r->pool, user, sec->auth_grpfile);
    else
	grpstatus = NULL;

    for (x = 0; x < reqs_arr->nelts; x++) {

	if (!(reqs[x].method_mask & (1 << m)))
	    continue;

	method_restricted = 1;

	t = reqs[x].requirement;
	w = ap_getword_white(r->pool, &t);
	if (!strcmp(w, "valid-user"))
	    return OK;
	if (!strcmp(w, "user")) {
	    while (t[0]) {
		w = ap_getword_conf(r->pool, &t);
		if (!strcmp(user, w))
		    return OK;
	    }
	}
	else if (!strcmp(w, "group")) {
	    if (!grpstatus)
		return DECLINED;	/* DBM group?  Something else? */

	    while (t[0]) {
		w = ap_getword_conf(r->pool, &t);
		if (ap_table_get(grpstatus, w))
		    return OK;
	    }
	} else if (sec->auth_authoritative) {
	    /* if we aren't authoritative, any require directive could be
	     * valid even if we don't grok it.  However, if we are 
	     * authoritative, we can warn the user they did something wrong.
	     * That something could be a missing "AuthAuthoritative off", but
	     * more likely is a typo in the require directive.
	     */
	    ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
		"access to %s failed, reason: unknown require directive:"
		"\"%s\"", r->uri, reqs[x].requirement);
	}
    }

    if (!method_restricted)
	return OK;

    if (!(sec->auth_authoritative))
	return DECLINED;

    ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
	"access to %s failed, reason: user %s not allowed access",
	r->uri, user);
	
    ap_note_basic_auth_failure(r);
    return AUTH_REQUIRED;
}

module MODULE_VAR_EXPORT auth_module =
{
    STANDARD_MODULE_STUFF,
    NULL,			/* initializer */
    create_auth_dir_config,	/* dir config creater */
    NULL,			/* dir merger --- default is to override */
    NULL,			/* server config */
    NULL,			/* merge server config */
    auth_cmds,			/* command table */
    NULL,			/* handlers */
    NULL,			/* filename translation */
    authenticate_basic_user,	/* check_user_id */
    check_user_access,		/* check auth */
    NULL,			/* check access */
    NULL,			/* type_checker */
    NULL,			/* fixups */
    NULL,			/* logger */
    NULL,			/* header parser */
    NULL,			/* child_init */
    NULL,			/* child_exit */
    NULL			/* post read-request */
};
