/*
----[  INTRODUCTION

Do you know if your system has been hacked?  If you found those funny user
accounts or that Trojaned program, its too late.  You're owned.  Chances are
that your systems were scanned for holes before your systems were cracked.
If you had just seen them coming you wouldn't be reloading that OS right now.
Programs like TCP Wrappers do some good, but they don't see the stealth scans
or DOS attacks.  You could by a nice commercial network intrusion detector,
but your wallet screams in agony.  What you need is a low cost (as in free)
fast, somewhat paranoid network monitor that watches all packets and uses
few resources.  Watcher provides this.

----[  IMPLEMENTATION

Watcher examines all packets on the network interface and assumes they all are
potentially hostile.  Watcher examines every packet within a 10 second window,
and, at the end of each window it will record any malicious activity it sees
using syslog.  Watcher currently detects the following attacks:

    - All TCP scans
    - All UDP scans
    - Synflood attacks
    - Teardrop attacks
    - Land attacks
    - Smurf attacks
    - Ping of death attacks

All parameters and thresholds are configurable through command line options.
You can also configure watcher to just look for scans or just look for DOS
attacks.  Watcher assumes any TCP packet other than a RST (which elicits no
response) may be used to scan for services.  If packets of any type are
received by more than 7 different ports within the window, an event is
logged.  The same criteria are used for UDP scans.  If watcher sees more than
8 SYN packets to the same port with no ACK's or FIN's associated with the
SYN's, a synflood event is logged.  If a fragmented UDP packet with an IP id
of 242 is seen, it is assumed to be a teardrop attack since the published code
uses an id of 242.  This is somewhat lame since anyone could change the
attacking code to use other id's.  The code should track all fragmented IP's
and check for overlapping offsets.  I may do this in a future version.  Any
TCP SYN packets with source and destination address and ports the same is a
identified as a land attack.  If more than 5 ICMP ECHO REPLIES are seen within
the window, Watcher assumes it may be a Smurf attack.  Note that this is not a
certainty, since someone your watching may just be pinging the hell out of
someone.  Watcher also assumes that any fragmented ICMP packet is bad, bad,
bad.  This catches attacks such as the ping of death.

Watcher has three modes of monitoring.  In the default mode, it just watches
for attacks against its own host.  The second monitoring mode is to watch all
hosts on it's class C subnet.  In the third mode, it watches all hosts whose
packets it sees.  Watching multiple hosts is useful if you put Watcher on your
border to external networks, or to have hosts watch out for each other in case
one gets cracked before you can react.  Even if log files are destroyed, the
other host has a record.

It must be noted that since Watcher treats every packet as potentially hostile,
it sometimes can report false positives.  There are some checks in the code
to minimize this by increasing its tolerance for certain activity.
Unfortunately this also increases the rate at which scans can be done before
Watcher notices.  The usual false positives are TCP scans and synfloods,
mostly resulting from WWW activity.  Some web pages have many URL's to GIF
files and other pretty stuff.  Each of these may cause the client to open a
separate TCP connection to download.  Watcher sees these and treats them as
a TCP scan of the client.  To minimize this, watcher will only log TCP scans
if more than 40 are received in the window AND the source port of the scan
was 80.  This of course can be configured higher or lower as desired.  As for
synfloods we will use the same WWW example above.  If the client opens a lot
of connections to the server right before the 10 second window expires and
Watcher does not see the ACK's or FIN's for those SYN packets, Watcher will
think the client is synflooding port 80 on the server.  This only happens
if watcher is watching the server, or if you are watching everyone.  You
may also get occasional false UDP scans if the system being watched makes
lots of DNS queries within the window.

The output for Watcher is pretty simple.  Every 10 seconds, any detected
attacks are logged to syslog.  The source and target IP's are logged along
with the type of attack.  Where appropriate, other information such as the
number of packets, or the port involved are logged.  If the attack is normally
associated with false IP addresses, the MAC address is also logged.  If the
attack is external, the MAC will be for the local router that handled the
packet.  If it was from your LAN, you'll have the source machine and you can
thank the sender in an appropriate manner.

----[  PROGRAM EXECUTION

Watcher was written to run on Linux systems.  Watcher has a variety of, most
of the self-explanatory.  To execute watcher, simply run it in the background,
usually from the system startup script.  The options are:

Usage: watcher [options]
  -d device       Use 'device' as the network interface device
                  The first non-loopback interface is the default
  -f flood        Assume a synflood attack occurred if more than
                  'flood' uncompleted connections are received
  -h              A little help here
  -i icmplimit    Assume we may be part of a smurf attack if more
                  than icmplimit ICMP ECHO REPLIES are seen
  -m level        Monitor more than just our own host.
                  A level of 'subnet' watches all addresses in our
                  subnet and 'all' watches all addresses
  -p portlimit    Logs a portscan alert if packets are received for
                  more than portlimit ports in the timeout period.
  -r reporttype   If reporttype is dos, only Denial Of Service
                  attacks are reported.  If reporttype is scan
                  then only scanners are reported.  Everything is
                  reported by default.
  -t timeout      Count packets and print potential attacks every
                  timeout seconds
  -w webcount     Assume we are being portscanned if more than
                  webcount packets are received from port 80

Hopefully, watcher will keep your systems a little better protected.  But
remember that good security is multiple layers, and no single defense tool will
save you by itself.  If you forget this, you'll be reloading that OS one day.

----[  THE CODE

<++> EX/Watcher.c
*/
/*********************************************************************
Program: watcher

A network level monitoring tool to detect incoming packets indicative of
potential attacks.

This software detects low level packet scanners and several DOS attacks.
Its primary use is to detect low level packet scans, since these are usually
done first to identify active systems and services to mount further attacks.

The package assumes every incoming packet is potentially hostile.  Some checks
are done to minimize false positives, but on occasion a site may be falsely
identified as having performed a packet scan or SYNFLOOD attack.  This usually
occurs if a large number of connections are done in a brief time right before
the reporting timeout period (i.e. when browsing a WWW site with lots of
little GIF's, each requiring a connection to download).  You can also get false
positives if you scan another site, since the targets responses will be viewed
as a potential scan of your system.

By default, alerts are printed to SYSLOG every 10 seconds.
***********************************************************************/

#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netdb.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <malloc.h>
#include <signal.h>
#include <sys/wait.h>

#define __FAVOR_BSD

#include <netinet/tcp.h>
#include <netinet/in_systm.h>
#include <net/if_arp.h>
#include <net/if.h>
#undef __FAVOR_BSD
#include <netinet/udp.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <linux/if_ether.h>
#include <linux/sockios.h>

#include <syslog.h>
#include <time.h>

#include "init.h"

/* 2001/10/16 Steve++ */
#ifdef __EMBEDDED__
#include <linux/if_pppox.h>
#include <linux/ppp_defs.h>
#endif

#define PKTLEN 96       /* Should be enough for what we want */
#ifndef IP_MF
#define IP_MF   0x2000
#endif


#define SSMTP "/bin/ssmtp"
#define MAILFILE "/var/tmp/mail"

/***** WATCH LEVELS ******/

#define MYSELFONLY      1
#define MYSUBNET        2
#define HUMANITARIAN    3

/***** REPORT LEVELS *****/

#define REPORTALL       1
#define REPORTDOS       2
#define REPORTSCAN      3

struct floodinfo {
    u_short sport;
    struct floodinfo *next;
};

/* Steve ++ 2001/10/24 */
struct portlist {
    u_short dport;
    struct portlist *next;
};

struct addrlist {
    u_long saddr;
    int cnt;
    int wwwcnt;
/* Steve ++ 2001/10/24 */
    struct portlist *portinfo;
    struct addrlist *next;
};

struct atk {
    u_long saddr;
    u_char eaddr[ETH_ALEN];
    time_t atktime;
};

struct pktin {
    u_long saddr;
    u_short sport;
    u_short dport;
    time_t timein;
    u_char eaddr[ETH_ALEN];
    struct floodinfo *fi;
    struct pktin *next;
};

struct scaninfo {
    u_long addr;
    struct atk teardrop;
    struct atk land;
    struct atk icmpfrag;
    struct pktin *tcpin;
    struct pktin *udpin;
    struct scaninfo *next;
    u_long icmpcnt;
//Angus ++
    struct pktin *icmpin;
} ;

//Angus ++
int stcp, sudp, sicmp;
fd_set set;
struct ethhdr ethnull;

//Angus --

/* 2001/11/19 Steve ++ */
#define MAX_SCANINFO_COUNT  1024
u_long count = 0;
/* Steve -- */

struct scaninfo *Gsilist = NULL, *Gsi, *Gsitmp, *Gsiprev;

u_long Gmaddr;
time_t Gtimer = 1, Gtimein;
int Gportlimit = 7;
int Gsynlimit = 8;
int Gwebcount = 40;
int Gicmplimit = 5;
int Gwatchlevel = HUMANITARIAN;
//int Gwatchlevel = MYSELFONLY;
int Greportlevel = REPORTALL;
char *Gprogramname, *Gdevice = "eth1", *Loopback = "127.0.0.1";

int Gsynflood = 0;
int Gudpflood = 0;
int Gicmpflood = 0;
int Gudplimit = 1000;
int Gpingofdeath = 0;
int Gland = 0;
int Gportscan = 0;
int Gteardrop = 0;

/* 2001/09/16 Steve ++ */
//int gEmail = 0;
int gEmail = 1;
char gEmail_addr1[MAXBUF];
char gEmail_addr2[MAXBUF];

char gEmail_Buffer[2048]; 
/* Angus ++ 2001/05/14
add for decide email address config file in argument 
Note: This is not compatible with previous version
      (Previous version use FW_CONFIG_FILE)
*/
char ssmtp_conf[80] = SSMTP_CONFIG_FILE;


/******** IP packet info ********/

u_long Gsaddr, Gdaddr;
int Giplen, Gisfrag, Gid;

/****** Externals *************/

extern int errno;
extern char *optarg;
extern int optind, opterr;

void do_tcp(), do_udp(), do_icmp(), print_info(), process_packet();
void addtcp(), addudp(), addicmp(), clear_pktin(), buildnet();
void doargs(), usage(), addfloodinfo(), rmfloodinfo();
void Email_Alter();

struct scaninfo *doicare(), *addtarget();
char *anetaddr(), *ether_ntoa();
u_char *readdevice();

main(argc, argv)
int argc;
char *argv[];
{
    int pktlen = 0, i, netfd;
    u_char *pkt;
    char hostname[32];
    struct hostent *hp;
    time_t t;

    doargs(argc, argv);
    InitConfig(ssmtp_conf);

    /* Angus ++ 2001/05/09 add for call in cgi */
    close (0);
    close (1);
    close (2);
    setsid ();
    /* -- */
    if ( !(Gsynflood | Gudpflood | Gicmpflood | Gpingofdeath | Gland 
           | Gportscan | Gteardrop)) {
		while (1)
			sleep(100);
	exit(0);
    }

#if 0
    if(fork() > 0)
	exit(0);
    /* Angus ++ 2001/05/09 add for call in cgi */
    close (0);
    close (1);
    close (2);
    setsid ();
    /* -- */
#endif
    openlog("WATCHER", 0, LOG_LOCAL1);
#if 0
    if(gethostname(hostname, sizeof(hostname)) < 0)
    {
        perror("gethostname");
        exit(-1);
    }
    if((hp = gethostbyname(hostname)) == NULL)
    {
        fprintf(stderr, "Cannot find own address\n");
        exit(-1);
    }
#endif
    /* 2001/09/16 Steve ++ */
    Gmaddr = inet_addr(Loopback);
//    memcpy((char *)&Gmaddr, hp->h_addr, hp->h_length);
    buildnet();
    if((netfd = initdevice(O_RDWR, 0)) < 0)
        exit(-1);

    /* Now read packets forever and process them. */

    t = time((time_t *)0);
    while(pkt = readdevice(netfd, &pktlen))
    {
        process_packet(pkt, pktlen);

        if(time((time_t *)0) - t >= Gtimer)
        {
            /* Times up.  Print what we found and clean out old stuff. */
//            for(Gsi = Gsilist, i = 0; Gsi; Gsi = Gsi->next, i++)
            for(Gsi = Gsilist, i = 0; Gsi; i++)
            {
		/* 2001/11/19 Steve ++ */
		Gsitmp = Gsi->next;
                clear_pktin(Gsi);
                print_info();

		Email_Alter();
		gEmail_Buffer[0] = '\0';
                Gsi->icmpcnt = 0;
	        if (count >= MAX_SCANINFO_COUNT)
			my_free(Gsi);

		Gsi = Gsitmp;
            }

	    /* 2001/11/19 Steve ++ */
	    if (count >= MAX_SCANINFO_COUNT) {
		count = 0;
		Gsilist = NULL;
	    }
	    /* Steve -- */

            t = time((time_t *)0);
        }
    }
}

/**********************************************************************
Function: doargs

Purpose:  sets values from environment or command line arguments.
**********************************************************************/
void doargs(argc, argv)
int argc;
char **argv;
{
    char c;

    Gprogramname = argv[0];
    while((c = getopt(argc,argv,"d:f:hi:m:p:r:t:w:")) != EOF)
    {
        switch(c)
        {
            case 'd':
                Gdevice = optarg;
                break;
            case 'f':
                Gsynlimit = atoi(optarg);
                break;
            case 'h':
                usage();
                exit(0);
            case 'i':
                Gicmplimit = atoi(optarg);
                break;
            case 'm':
                if(strcmp(optarg, "all") == 0)
                    Gwatchlevel = HUMANITARIAN;
                else if(strcmp(optarg, "subnet") == 0)
                    Gwatchlevel = MYSUBNET;
                else
                {
                    usage();
                    exit(-1);
                }
                break;
            case 'p':
                Gportlimit = atoi(optarg);
                break;
            case 'r':
                if(strcmp(optarg, "dos") == 0)
                    Greportlevel = REPORTDOS;
                else if(strcmp(optarg, "scan") == 0)
                    Greportlevel = REPORTSCAN;
                else
                {
                    exit(-1);
                }
                break;
            case 't':
                Gtimer = atoi(optarg);
                break;
            case 'w':
                Gwebcount = atoi(optarg);
                break;
/* Angus ++ 2001/05/14
add for decide email address config file in argument 
*/
	    case 's':
		strncpy(ssmtp_conf, optarg, sizeof(ssmtp_conf));
		break;
            default:
                usage();
                exit(-1);
        }
    }
}

/**********************************************************************
Function: usage

Purpose:  Display the usage of the program
**********************************************************************/
void usage()
{
printf("Usage: %s [options]\n", Gprogramname);
printf("  -d device       Use 'device' as the network interface device\n");
printf("                  The first non-loopback interface is the default\n");
printf("  -f flood        Assume a synflood attack occurred if more than\n");
printf("                  'flood' uncompleted connections are received\n");
printf("  -h              A little help here\n");
printf("  -i icmplimit    Assume we may be part of a smurf attack if more\n");
printf("                  than icmplimit ICMP ECHO REPLIES are seen\n");
printf("  -m level        Monitor more than just our own host.\n");
printf("                  A level of 'subnet' watches all addresses in our\n");
printf("                  subnet and 'all' watches all addresses\n");
printf("  -p portlimit    Logs a portscan alert if packets are received for\n");
printf("                  more than portlimit ports in the timeout period.\n");
printf("  -r reporttype   If reporttype is dos, only Denial Of Service\n");
printf("                  attacks are reported.  If reporttype is scan\n");
printf("                  then only scanners are reported.  Everything is\n");
printf("                  reported by default.\n");
printf("  -t timeout      Count packets and print potential attacks every\n");
printf("                  timeout seconds\n");
printf("  -w webcount     Assume we are being portscanned if more than\n");
printf("                  webcount packets are received from port 80\n");
}

/**********************************************************************
Function: buildnet

Purpose:  Setup for monitoring of our host or entire subnet.
**********************************************************************/
void buildnet()
{
    u_long addr;
    u_char *p;
    int i;

    if(Gwatchlevel == MYSELFONLY)               /* Just care about me */
    {
        (void) addtarget(Gmaddr);
    }
    else if(Gwatchlevel == MYSUBNET)            /* Friends and neighbors */
    {
        addr = htonl(Gmaddr);
        addr = addr & 0xffffff00;
        for(i = 0; i < 256; i++)
            (void) addtarget(ntohl(addr + i));
    }
}
/**********************************************************************
Function: doicare

Purpose:  See if we monitor this address
**********************************************************************/
struct scaninfo *doicare(addr)
u_long addr;
{
    struct scaninfo *si;
    int i;

    for(si = Gsilist; si; si = si->next)
    {
        if(si->addr == addr)
            return(si);
    }
    if(Gwatchlevel == HUMANITARIAN)     /* Add a new address, we always care */
    {
        si = addtarget(addr);
        return(si);
    }
    return(NULL);
}

/**********************************************************************
Function: addtarget

Purpose:  Adds a new IP address to the list of hosts to watch.
**********************************************************************/
struct scaninfo *addtarget(addr)
u_long addr;
{
    struct scaninfo *si;

    if((si = (struct scaninfo *)my_malloc(sizeof(struct scaninfo))) == NULL)
    {
        perror("malloc scaninfo");
        exit(-1);
    }
    memset(si, 0, sizeof(struct scaninfo));
    si->addr = addr;
    si->next = Gsilist;
    Gsilist = si;
    /* 2001/11/19 Steve ++ */
    count++;
    /* Steve -- */
    return(si);
}

/**********************************************************************
Function: process_packet

Purpose:  Process raw packet and figure out what we need to to with it.

Pulls the packet apart and stores key data in global areas for reference
by other functions.
**********************************************************************/
void process_packet(pkt, pktlen)
u_char *pkt;
int pktlen;
{
    struct ethhdr *ep;
    struct iphdr *ip;
    static struct align { struct iphdr ip; char buf[PKTLEN]; } a1;
    u_short off;

    /* 2001/10/16 Steve++
     * Accept PPPoE Packets
     */
#ifdef __EMBEDDED__
    struct pppoe_hdr *pppoehdr;
    unsigned short ppphdr;
#endif

    Gtimein = time((time_t *)0);
    ep = (struct ethhdr *) pkt;

#ifdef __EMBEDDED__
    switch (ntohs(ep->h_proto)) {
	case ETH_P_PPP_SES :
		pkt += (sizeof(struct ethhdr) + sizeof(struct pppoe_hdr));
		pktlen -= (sizeof(struct ethhdr) + sizeof(struct pppoe_hdr));
		memcpy(&ppphdr, pkt, sizeof(unsigned short));
		if (ntohs(ppphdr) == PPP_IP) {
			pkt += sizeof(unsigned short);
			pktlen -= sizeof(unsigned short);
			break;
		} else
			return;
	case ETH_P_IP :
		pkt += sizeof(struct ethhdr);
		pktlen -= sizeof(struct ethhdr);
		break;
	default :
		return;
    }
#else
    if(ntohs(ep->h_proto) != ETH_P_IP)
        return;

    pkt += sizeof(struct ethhdr);
    pktlen -= sizeof(struct ethhdr);
#endif
    memcpy(&a1, pkt, pktlen);
    ip = &a1.ip;
    Gsaddr = ip->saddr;
    Gdaddr = ip->daddr;

    if((Gsi = doicare(Gdaddr)) == NULL)
        return;

    off = ntohs(ip->frag_off);
    Gisfrag = (off & IP_MF);    /* Set if packet is fragmented */
    Giplen = ntohs(ip->tot_len);
    Gid = ntohs(ip->id);
    pkt = (u_char *)ip + (ip->ihl << 2);
    Giplen -= (ip->ihl << 2);
    switch(ip->protocol)
    {
        case IPPROTO_TCP:
            do_tcp(ep, pkt);
            break;
        case IPPROTO_UDP:
            do_udp(ep, pkt);
            break;
        case IPPROTO_ICMP:
            do_icmp(ep, pkt);
            break;
        default:
            break;
    }
}

/**********************************************************************
Function: do_tcp

Purpose:  Process this TCP packet if it is important.
**********************************************************************/
void do_tcp(ep, pkt)
struct ethhdr *ep;
u_char *pkt;
{
    struct tcphdr *thdr;
    u_short sport, dport;

    thdr = (struct tcphdr *) pkt;
    if(thdr->th_flags & TH_RST) /* RST generates no response */
        return;                 /* Therefore can't be used to scan. */
    sport = ntohs(thdr->th_sport);
    dport = ntohs(thdr->th_dport);

    if (Gland)
    {
        if(thdr->th_flags & TH_SYN)
        {
            if(Gsaddr == Gdaddr && sport == dport)
            {
                Gsi->land.atktime = Gtimein;
                Gsi->land.saddr = Gsaddr;
                memcpy(Gsi->land.eaddr, ep->h_source, ETH_ALEN);
            }
        }
    }
    addtcp(sport, dport, thdr->th_flags, ep->h_source);
}

/**********************************************************************
Function: addtcp

Purpose:  Add this TCP packet to our list.
**********************************************************************/
void addtcp(sport, dport, flags, eaddr)
u_short sport;
u_short dport;
u_char flags;
u_char *eaddr;
{
    struct pktin *pi, *last, *tpi;

    /* See if this packet relates to other packets already received. */

    for(pi = Gsi->tcpin; pi; pi = pi->next)
    {
        if (Gsynflood)
        {
            if(pi->saddr == Gsaddr && pi->dport == dport)
            {
                if(flags == TH_SYN)
                    addfloodinfo(pi, sport);
                else if((flags & TH_FIN) || (flags & TH_ACK))
                    rmfloodinfo(pi, sport);
                return;
            }
        }

        last = pi;
    }

    /* Must be new entry */

    if((tpi = (struct pktin *)my_malloc(sizeof(struct pktin))) == NULL)
    {
        perror("Malloc");
        exit(-1);
    }
    memset(tpi, 0, sizeof(struct pktin));
    memcpy(tpi->eaddr, eaddr, ETH_ALEN);
    tpi->saddr = Gsaddr;
    tpi->sport = sport;
    tpi->dport = dport;
    tpi->timein = Gtimein;

    if(Gsynflood)
    {
        if(flags == TH_SYN)
            addfloodinfo(tpi, sport);
    }

    if(Gsi->tcpin)
        last->next = tpi;
    else
        Gsi->tcpin = tpi;
}

/**********************************************************************
Function: addfloodinfo

Purpose:  Add floodinfo information
**********************************************************************/
void addfloodinfo(pi, sport)
struct pktin *pi;
u_short sport;
{
    struct floodinfo *fi;

    fi = (struct floodinfo *)my_malloc(sizeof(struct floodinfo));
    if(fi == NULL)
    {
        perror("Malloc of floodinfo");
        exit(-1);
    }
    memset(fi, 0, sizeof(struct floodinfo));
    fi->sport = sport;
    fi->next = pi->fi;
    pi->fi = fi;
}

/**********************************************************************
Function: rmfloodinfo

Purpose:  Removes floodinfo information
**********************************************************************/
void rmfloodinfo(pi, sport)
struct pktin *pi;
u_short sport;
{
    struct floodinfo *fi, *prev = NULL;

    for(fi = pi->fi; fi; fi = fi->next)
    {
        if(fi->sport == sport)
            break;
        prev = fi;
    }
    if(fi == NULL)
        return;
    if(prev == NULL)    /* First element */
        pi->fi = fi->next;
    else
        prev->next = fi->next;
    my_free(fi);
}

/**********************************************************************
Function: do_udp

Purpose:  Process this udp packet.

Currently teardrop and all its derivitives put 242 in the IP id field.
This could obviously be changed.  The truly paranoid might want to flag all
fragmented UDP packets.  The truly adventurous might enhance the code to
track fragments and check them for overlaping boundaries.
**********************************************************************/
void do_udp(ep, pkt)
struct ethhdr *ep;
u_char *pkt;
{
    struct udphdr *uhdr;
    u_short sport, dport;

    uhdr = (struct udphdr *) pkt;

    if (Gteardrop)
    {
        if(Gid == 242 && Gisfrag)   /* probable teardrop */
        {
            Gsi->teardrop.saddr = Gsaddr;
            memcpy(Gsi->teardrop.eaddr, ep->h_source, ETH_ALEN);
            Gsi->teardrop.atktime = Gtimein;
        }
    }

    sport = ntohs(uhdr->source);
    dport = ntohs(uhdr->dest);
    addudp(sport, dport, ep->h_source);
}

/**********************************************************************
Function: addudp

Purpose:  Add this udp packet to our list.
**********************************************************************/
void addudp(sport, dport, eaddr)
u_short sport;
u_short dport;
u_char *eaddr;
{
    struct pktin *pi, *last, *tpi;

    for(pi = Gsi->udpin; pi; pi = pi->next)
    {
        if (Gudpflood)
        {
            if(pi->saddr == Gsaddr && pi->dport == dport)
	    {
//Angus ++
                addfloodinfo(pi, sport);
//Angus --
//Angus Mask
                //pi->timein = Gtimein;
                return;
            }
        }

        last = pi;
    }

    /* Must be new entry */

    if((tpi = (struct pktin *)my_malloc(sizeof(struct pktin))) == NULL)
    {
        perror("Malloc");
        exit(-1);
    }
    memset(tpi, 0, sizeof(struct pktin));
    memcpy(tpi->eaddr, eaddr, ETH_ALEN);
    tpi->saddr = Gsaddr;
    tpi->sport = sport;
    tpi->dport = dport;
    tpi->timein = Gtimein;

//Angus ++
    if (Gudpflood)
        addfloodinfo(tpi, sport);
//Angus --

    if(Gsi->udpin)
        last->next = tpi;
    else
        Gsi->udpin = tpi;
}

/**********************************************************************
Function: do_icmp

Purpose:  Process an ICMP packet.

We assume there is no valid reason to receive a fragmented ICMP packet.
**********************************************************************/
void do_icmp(ep, pkt)
struct ethhdr *ep;
u_char *pkt;
{
    struct icmphdr *icmp;

    icmp = (struct icmphdr *) pkt;

    if (Gpingofdeath)
    {
        if(Gisfrag) /* probable ICMP attack (i.e. Ping of Death) */
        {
            Gsi->icmpfrag.saddr = Gsaddr;
            memcpy(Gsi->icmpfrag.eaddr, ep->h_source, ETH_ALEN);
            Gsi->icmpfrag.atktime = Gtimein;
        }
    }
#if 0
    if(icmp->type == ICMP_ECHOREPLY)
        Gsi->icmpcnt++;
#endif

//Angus ++
    if (Gicmpflood)
    {
        if (icmp->type == ICMP_ECHO)
            addicmp(ep->h_source);
    }	
//Angus --

    return;
}

/**********************************************************************
Function: addicmp

Purpose:  Add this icmp packet to our list.
**********************************************************************/
void addicmp(eaddr)
u_char *eaddr;
{
    struct pktin *pi, *last, *tpi;

    for(pi = Gsi->icmpin; pi; pi = pi->next)
    {
        if(pi->saddr == Gsaddr)
	{
            addfloodinfo(pi, 0);
            return;
        }
        last = pi;
    }
    /* Must be new entry */

    if((tpi = (struct pktin *)my_malloc(sizeof(struct pktin))) == NULL)
    {
        perror("Malloc");
        exit(-1);
    }
    memset(tpi, 0, sizeof(struct pktin));
    memcpy(tpi->eaddr, eaddr, ETH_ALEN);
    tpi->saddr = Gsaddr;
    tpi->timein = Gtimein;
    addfloodinfo(tpi, 0);
    if(Gsi->icmpin)
        last->next = tpi;
    else
        Gsi->icmpin = tpi;
}

/**********************************************************************
Function: clear_pkt

Purpose:  Delete and free space for any old packets.
**********************************************************************/
void clear_pktin(si)
struct scaninfo *si;
{
    struct pktin *pi;
    struct floodinfo *fi, *tfi;
    time_t t, t2;

    t = time((time_t *)0);
    while(si->tcpin)
    {
        t2 = t - si->tcpin->timein;
        if(t2 > Gtimer)
        {
            pi = si->tcpin;
            fi = pi->fi;
            while(fi)
            {
                tfi = fi;
                fi = fi->next;
                my_free(tfi);
            }
            si->tcpin = pi->next;
            my_free(pi);
        }
        else
            break;
    }
    while(si->udpin)
    {
        t2 = t - si->udpin->timein;
        if(t2 > Gtimer)
        {
            pi = si->udpin;
//Angus ++
            fi = pi->fi;
            while(fi)
            {
                tfi = fi;
                fi = fi->next;
                my_free(tfi);
            }
//Angus --
            si->udpin = pi->next;
            my_free(pi);
        }
        else
            break;
    }
//Angus ++
    while(si->icmpin)
    {
        t2 = t - si->icmpin->timein;
        if(t2 > Gtimer)
        {
            pi = si->icmpin;
            fi = pi->fi;
            while(fi)
            {
                tfi = fi;
                fi = fi->next;
                my_free(tfi);
            }
            si->icmpin = pi->next;
            my_free(pi);
        }
        else
            break;
    }
//Angus --

}

/**********************************************************************
Function: print_info

Purpose:  Print out any alerts.
**********************************************************************/
void print_info()
{
    struct pktin *pi;
    struct addrlist *tcplist = NULL, *udplist = NULL, *al;
    struct floodinfo *fi;
    char buf[1024], *eaddr, abuf[32];
    int i;
/* Steve ++ 2001/10/24 */
    struct portlist *uportlist = NULL, *tportlist = NULL, *pl;
   
    strcpy(abuf, anetaddr(Gsi->addr));
    if(Greportlevel == REPORTALL || Greportlevel == REPORTDOS)
    {
        if(Gsi->teardrop.atktime)
        {
            eaddr = ether_ntoa(Gsi->teardrop.eaddr);
            sprintf(buf, "Possible teardrop attack from %s (%s) against %s",
                anetaddr(Gsi->teardrop), eaddr, abuf);
//Angus ++
	    strcat(gEmail_Buffer, buf);

            syslog(LOG_INFO, buf);
            memset(&Gsi->teardrop, 0, sizeof(struct atk));
        }
        if(Gsi->land.atktime)
        {
            eaddr = ether_ntoa(Gsi->land.eaddr);
            sprintf(buf, "Possible land attack from (%s) against %s",
                eaddr, abuf);
//Angus ++
	    strcat(gEmail_Buffer, buf);

            syslog(LOG_INFO, buf);
            memset(&Gsi->land, 0, sizeof(struct atk));
        }
        if(Gsi->icmpfrag.atktime)
        {
            eaddr = ether_ntoa(Gsi->icmpfrag.eaddr);
            sprintf(buf, "Ping of Death detected from %s (%s) against %s",
                anetaddr(Gsi->icmpfrag), eaddr, abuf);
//Angus ++
	    strcat(gEmail_Buffer, buf);

            syslog(LOG_INFO, buf);
            memset(&Gsi->icmpfrag, 0, sizeof(struct atk));
        }
#if 0
        if(Gsi->icmpcnt > Gicmplimit)
        {
            sprintf(buf, "ICMP ECHO threshold exceeded, smurfs up.  I saw %d packets\n", Gsi->icmpcnt);
            syslog(LOG_INFO, buf);
            Gsi->icmpcnt = 0;
        }
#endif
    }
    for(pi = Gsi->tcpin; pi; pi = pi->next)
    {
        i = 0;
        for(fi = pi->fi; fi; fi = fi->next)
            i++;

        if(Greportlevel == REPORTALL || Greportlevel == REPORTDOS)
        {
            if(i > Gsynlimit)
            {
                eaddr = ether_ntoa(pi->eaddr);
                sprintf(buf, "Possible SYN FLOOD from %s (%s) against %s, recieved %d packets\n",
                    anetaddr(pi->saddr), eaddr, abuf, i);
//Angus ++
                strcat(gEmail_Buffer, buf);

                syslog(LOG_INFO, buf);
            }
        }
        for(al = tcplist; al; al = al->next)
        {
            if(pi->saddr == al->saddr)
            {
		/* Steve ++ 2001/10/24 */
		for (pl = tportlist; pl; pl = pl->next) {
			if (pi->dport == pl->dport)
				goto tcp_next;
		}
		if (pl == NULL) { /* new port */
			pl = (struct portlist *)my_malloc(sizeof(struct portlist));
			if (pl == NULL) {
		                perror("Malloc port list");
                		exit(-1);
            		}
		        memset(pl, 0, sizeof(struct portlist));
			pl->dport = pi->dport;
			pl->next = tportlist;
			tportlist = pl;
		}
		/* Steve -- */
                al->cnt++;
                if(pi->sport == 80)
                    al->wwwcnt++;
                break;
            }
        }

/* Steve ++ 2001/10/24 */
tcp_next:

        if(al == NULL)  /* new address */
        {
            al = (struct addrlist *)my_malloc(sizeof(struct addrlist));
            if(al == NULL)
            {
                perror("Malloc address list");
                exit(-1);
            }
            memset(al, 0, sizeof(struct addrlist));
            al->saddr = pi->saddr;
            al->cnt = 1;
            if(pi->sport == 80)
                al->wwwcnt = 1;
            al->next = tcplist;
            tcplist = al;
        }
    }

//Angus ++
    for(pi = Gsi->icmpin; pi; pi = pi->next)
    {
    	i = 0;
        for(fi = pi->fi; fi; fi = fi->next)
            i++;

        if(Greportlevel == REPORTALL || Greportlevel == REPORTDOS)
        {
           if(i > Gicmplimit)
           {
               eaddr = ether_ntoa(pi->eaddr);
               sprintf(buf, "Possible ICMP FLOOD from %s (%s) against %s, recieved %d packets\n",
                     anetaddr(pi->saddr), eaddr, abuf, i);

               //DisposeConnect(anetaddr(pi->saddr), 0);
//Angus ++
               strcat(gEmail_Buffer, buf);

               syslog(LOG_INFO, buf);
           }
        }
    }
//Angus --

    if(Greportlevel == REPORTALL || Greportlevel == REPORTSCAN)
    {
        if (Gportscan)
        {
            for(al = tcplist; al; al = al->next)
            {
                if((al->cnt - al->wwwcnt) > Gportlimit 
                   || al->wwwcnt > Gwebcount)
                {
                    sprintf(buf, "Possible TCP port scan from %s (%d ports) against %s\n",
                    anetaddr(al->saddr), al->cnt, abuf);
//Angus++
                    strcat(gEmail_Buffer, buf);
	
                    syslog(LOG_INFO, buf);
                }
            }
        }

        for(pi = Gsi->udpin; pi; pi = pi->next)
        {
//Angus ++
            i = 0;
            for(fi = pi->fi; fi; fi = fi->next)
                i++;

            if(Greportlevel == REPORTALL || Greportlevel == REPORTDOS)
            {
                if(i > Gudplimit)
                {
                    eaddr = ether_ntoa(pi->eaddr);
                    sprintf(buf, "Possible UDP FLOOD from %s (%s) against %s, recieved %d packets\n",
                    anetaddr(pi->saddr), eaddr, abuf, i);

            	    strcat(gEmail_Buffer, buf);

                    syslog(LOG_INFO, buf);
               }
            }
//Angus --
            for(al = udplist; al; al = al->next)
            {
                if(pi->saddr == al->saddr)
                {
		    /* Steve ++ 2001/10/24 */
		    for (pl = uportlist; pl; pl = pl->next) {
			if (pi->dport = pl->dport)
				goto udp_next;
		    }
		    if (pl == NULL) {
			pl = (struct portlist *)my_malloc(sizeof(struct portlist));
                        if (pl == NULL) {
                                perror("Malloc port list");
                                exit(-1);
                        }
                        memset(pl, 0, sizeof(struct portlist));
                        pl->dport = pi->dport;
                        pl->next = uportlist;
                        uportlist = pl;
		    }
		    /* Steve -- */

                    al->cnt++;
                    break;
                }
            }

/* Steve ++ 2001/10/24 */
udp_next:
            if(al == NULL)      /* new address */
            {
                al = (struct addrlist *)my_malloc(sizeof(struct addrlist));
                if(al == NULL)
                {
                    perror("Malloc address list");
                    exit(-1);
                }
                memset(al, 0, sizeof(struct addrlist));
                al->saddr = pi->saddr;
                al->cnt = 1;
                al->next = udplist;
                udplist = al;
            }
        }
        for(al = udplist; al; al = al->next)
        {
            if (Gportscan)
            {
                if(al->cnt > Gportlimit)
                {
                    sprintf(buf, "Possible UDP port scan from %s (%d ports) against %s\n",
                    anetaddr(al->saddr), al->cnt, abuf);
//Angus ++
           	    strcat(gEmail_Buffer, buf);

                    syslog(LOG_INFO, buf);
                }
            }
        }
    }

    while(tcplist)
    {
	tportlist = tcplist->portinfo;
	while(tportlist) {
		pl = tportlist->next;
		my_free(tportlist);
		tportlist = pl;
	}		
        al = tcplist->next;
        my_free(tcplist);
        tcplist = al;
    }
    while(udplist)
    {
	uportlist = udplist->portinfo;
	while(uportlist) {
		pl = uportlist->next;
		my_free(uportlist);
		uportlist = pl;
	}		
        al = udplist->next;
        my_free(udplist);
        udplist = al;
    }
}

/************************************************************************
Function:  anetaddr

Description:

Another version of the intoa function.
************************************************************************/

char *anetaddr(addr)
u_long addr;
{
    u_long naddr;
    static char buf[16];
    u_char b[4];
    int i;

    naddr = ntohl(addr);
    for(i = 3; i >= 0; i--)
    {
        b[i] = (u_char) (naddr & 0xff);
        naddr >>= 8;
    }
    sprintf(buf, "%d.%d.%d.%d", b[0], b[1], b[2], b[3]);
    return(buf);
}

/************************************************************************
Function:  initdevice

Description: Set up the network device so we can read it.

**************************************************************************/
initdevice(fd_flags, dflags)
int fd_flags;
u_long dflags;
{
    struct ifreq ifr;
    int fd, flags = 0;

    if((fd = socket(PF_INET, SOCK_PACKET, htons(0x0003))) < 0)
    {
        perror("Cannot open device socket");
        exit(-1);
    }

    /* Get the existing interface flags */

    strcpy(ifr.ifr_name, Gdevice);
    if(ioctl(fd, SIOCGIFFLAGS, &ifr) < 0)
    {
        perror("Cannot get interface flags");
        exit(-1);
    }

/* 2003/8/28 Steve ++ */
#if 0
    ifr.ifr_flags |= IFF_PROMISC;

    if(ioctl(fd, SIOCSIFFLAGS,  &ifr) < 0)
    {
        perror("Cannot set interface flags");
        exit(-1);
    }
#endif
/* Steve -- */
    return(fd);
}

/************************************************************************
Function:  readdevice

Description: Read a packet from the device.

**************************************************************************/
u_char *readdevice(fd, pktlen)
int fd;
int *pktlen;
{
    int cc = 0, from_len, readmore = 1;
    struct sockaddr from;
    static u_char pktbuffer[PKTLEN];
    u_char *cp;

    while(readmore)
    {
        from_len = sizeof(from);
        if((cc = recvfrom(fd, pktbuffer, PKTLEN, 0, &from, &from_len)) < 0)
        {
            if(errno != EWOULDBLOCK)
                return(NULL);
        }
        if(strcmp(Gdevice, from.sa_data) == 0)
            readmore = 0;
    }
    *pktlen = cc;

    return(pktbuffer);
}

/*************************************************************************
Function: ether_ntoa

Description:

Translates a MAC address into ascii.  This function emulates
the ether_ntoa function that exists on Sun and Solaris, but not on Linux.
It could probably (almost certainly) be more efficent, but it will do.
*************************************************************************/
char *ether_ntoa(etheraddr)
u_char etheraddr[ETH_ALEN];
{
    int i, j;
    static char eout[32];
    char tbuf[10];

    for(i = 0, j = 0; i < 5; i++)
    {
        eout[j++] = etheraddr[i] >> 4;
        eout[j++] = etheraddr[i] & 0xF;
        eout[j++] = ':';
    }
    eout[j++] = etheraddr[i] >> 4;
    eout[j++] = etheraddr[i] & 0xF;
    eout[j++] = '\0';
    for(i = 0; i < 17; i++)
    {
        if(eout[i] < 10)
            eout[i] += 0x30;
        else if(eout[i] < 16)
            eout[i] += 0x57;
    }
    return(eout);
}

void Email_Alter(void)
{
    char CmdLine[MAXBUF];
    FILE *fd;
    time_t cur_time;
/* 2002/11/24 Steve ++ */
        int pid = 0;
        int status = 0;
        static int id = 0;
	char host[80]; // 031105 cate
/* Steve -- */

    if (gEmail)
    {
/* 2002/11/24 Steve ++ */
        pid = waitpid((pid_t) -1, &status, WNOHANG);
        if (pid > 0) {
                kill(pid, SIGKILL);
                id = 0;
        }
	if(gethostname(host,60) != 0) //------------+
	    sprintf(host,"Firewall"); // 031105 cate|
	
/* Steve -- */
	if (strlen(gEmail_Buffer))
	{
            if ((fd = fopen(MAILFILE, "w+")) == NULL)
                return;

	    if (time(&cur_time) < 0)
		return;

	    fprintf(fd, "Subject: %s Hacker Attack Alarm!\n\n",host);
	    fprintf(fd, "Time: %s\n", ctime(&cur_time));
            fprintf(fd, "%s\n", gEmail_Buffer);
/* Steve -- */
	    fclose(fd);
            if(fork() == 0) {
                execl(SSMTP, SSMTP, MAILFILE, "nofork", NULL);
                exit(0);
            }
            id = 1;
/* Steve -- */

/* SuKent++, Mon 30 Jun 2003 10:27:31 AM GMT+8 */
	    {
    		    char	hostname[128], buf[512];

		    if (gethostname(hostname, 128) < 0)	strcpy(hostname, "Multihoming");
	
    		    sprintf(buf, "/usr/local/apache/cgi-bin/trap_msg -m \"%s hacker attack alarm : %s\"", hostname, gEmail_Buffer);
		    system(buf);
	    }
/* ******** */		    

	}
    }
}
