#include <linux/types.h>
#include <linux/errno.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <asm/arch/gpio.h>
#include <asm/hardware.h>
#include <linux/ioport.h>
#include <asm/io.h>
#include <asm/delay.h>
#include <asm/uaccess.h>
#include <linux/spinlock.h>
#include <linux/proc_fs.h>

#define	IXP425_EEPROM_CS_PIN		IXP425_GPIO_PIN_4
#define	IXP425_EEPROM_CLK_PIN		IXP425_GPIO_PIN_5
#define IXP425_EEPROM_IN_PIN		IXP425_GPIO_PIN_6
#define IXP425_EEPROM_OUT_PIN		IXP425_GPIO_PIN_7

#define MASK_EEPROM_CS			(1 << IXP425_EEPROM_CS_PIN)
#define MASK_EEPROM_CLK			(1 << IXP425_EEPROM_CLK_PIN)
#define MASK_EEPROM_IN			(1 << IXP425_EEPROM_IN_PIN)
#define MASK_EEPROM_OUT			(1 << IXP425_EEPROM_OUT_PIN)

#define IXP425_EEPROM_MAJOR		168
#define IXP425_EEPROM_MINOR		0
#define EEPROM_SIZE			0x80

/* AT93C66 op code */
#define OP_EWDS		0x00
#define OP_WRAL		0x40
#define OP_ERAL		0x80
#define OP_WRITE	0x100
#define OP_ERASE	0x300
#define OP_EWEN		0xC0		
#define OP_READ		0x200

#define EEPROM_IS_OPEN	1
static int eeprom_status = 0;
static spinlock_t ixp425_eeprom_lock = SPIN_LOCK_UNLOCKED;

static unsigned int
get_op_addr(int op_addr)
{
	unsigned int mask = 0x200;
	unsigned int addr = 0;

	*IXP425_GPIO_GPOUTR = addr;
	udelay(1);

	addr |= MASK_EEPROM_CS;
	*IXP425_GPIO_GPOUTR = addr;
	udelay(1);

	addr |= MASK_EEPROM_IN;
	*IXP425_GPIO_GPOUTR = addr;
	udelay(1);

	addr |= MASK_EEPROM_CLK;
	*IXP425_GPIO_GPOUTR = addr;
	udelay(1);

	while (mask) {
		addr &= ~(MASK_EEPROM_CLK + MASK_EEPROM_IN);
		if (mask & op_addr)
			addr |= MASK_EEPROM_IN;

		*IXP425_GPIO_GPOUTR = addr;
		udelay(1);

		addr |= MASK_EEPROM_CLK;
		*IXP425_GPIO_GPOUTR = addr;
		udelay(1);

		mask >>= 1;
	}

	return addr;
}

static unsigned int
read_data(int offset)
{
	unsigned int romdata = 0, mask = 0x8000, tmp;
	unsigned int addr = 0;

	addr = get_op_addr(OP_READ|offset);
	while (mask) {
		addr &= ~(MASK_EEPROM_CLK + MASK_EEPROM_IN);
		*IXP425_GPIO_GPOUTR = addr;
		udelay(1);
		
		addr |= MASK_EEPROM_CLK;
		*IXP425_GPIO_GPOUTR = addr;
		udelay(1);

		tmp = *IXP425_GPIO_GPINR;

		if (tmp & MASK_EEPROM_OUT)
			romdata |= mask;

		mask >>= 1;
	}
	return romdata;
}

static int
eeprom_read(struct file *file, char *buf, size_t count, loff_t *fpos)
{
	int size, offset = 0, i = 0;
	unsigned char eeprom_data[EEPROM_SIZE+1];
	unsigned long p = *fpos, eeprom_p;
	u_int16_t wdata;

	if (fpos != &file->f_pos)
		return -ESPIPE;

	if (p >= EEPROM_SIZE)
		return 0;

	if ((p + count) >= EEPROM_SIZE)
		size = (EEPROM_SIZE - p);
	else
		size = count;

	spin_lock(&ixp425_eeprom_lock);

	eeprom_p = p / 2;
	if ((eeprom_p * 2) < p) {
		wdata = (u_int16_t)read_data((eeprom_p+i));
		eeprom_data[offset] = (unsigned char)(wdata & 0x00ff);
		i++;
		offset++;
	}

	for (; offset < size; offset += 2, i++) {
		wdata = (u_int16_t)read_data((eeprom_p+i));
		eeprom_data[offset] = (unsigned char)((wdata & 0xff00) >> 8);
		eeprom_data[offset+1] = (unsigned char)(wdata & 0x00ff);
	}

	if (copy_to_user(buf, eeprom_data, size)) {
		spin_unlock(&ixp425_eeprom_lock);
		return -EFAULT;
	}

	*fpos += size;

	spin_unlock(&ixp425_eeprom_lock);
	return size;
}

static void
eeprom_ewen(void)
{
	unsigned int addr = 0;

	addr = get_op_addr(OP_EWEN);
	addr &= ~(MASK_EEPROM_CS + MASK_EEPROM_CLK + MASK_EEPROM_IN);
	*IXP425_GPIO_GPOUTR = addr;
	udelay(1);
}

static void
eeprom_ewds(void)
{
	unsigned int addr = 0;
	addr = get_op_addr(OP_EWDS);
	addr &= ~(MASK_EEPROM_CS + MASK_EEPROM_CLK + MASK_EEPROM_IN);
	*IXP425_GPIO_GPOUTR = addr;
	udelay(1);	
}

static int
check_eeprom_status(unsigned int addr, int mode)
{
	unsigned long poll = (mode) ? 7500000L : 100000L;

	addr &= ~(MASK_EEPROM_CS + MASK_EEPROM_CLK + MASK_EEPROM_IN);
	*IXP425_GPIO_GPOUTR = addr;
	udelay(1);

	addr |= MASK_EEPROM_CS;
	*IXP425_GPIO_GPOUTR = addr;
	udelay(1);

	while (--poll && !(*IXP425_GPIO_GPINR & MASK_EEPROM_OUT));

	addr &= ~(MASK_EEPROM_CS + MASK_EEPROM_CLK + MASK_EEPROM_IN);
	*IXP425_GPIO_GPOUTR = addr;
	udelay(1);
	udelay(1000);
	return (poll > 0);	
}

static void
erase_data(unsigned int offset)
{
	unsigned int addr = 0;
	addr = get_op_addr(OP_ERASE | offset);
	check_eeprom_status(addr, 0);
}

static void
write_data(unsigned int offset, u_int16_t edata)
{
	unsigned int addr = 0;
	unsigned int mask = 0x8000;

	addr = get_op_addr(OP_WRITE | offset);
	while (mask) {
		addr &= ~(MASK_EEPROM_CLK + MASK_EEPROM_IN);
		if (mask & edata)
			addr |= MASK_EEPROM_IN;
		*IXP425_GPIO_GPOUTR = addr;
		udelay(1);

		addr |= MASK_EEPROM_CLK;
		*IXP425_GPIO_GPOUTR = addr;
		udelay(1);

		mask >>= 1;
	}
	check_eeprom_status(addr, 1);
}

static int
eeprom_write(struct file *file, char *buf, size_t count, loff_t *fpos)
{
	u_int16_t romdata = 0, w1, w2;
	unsigned int offset, size = 0, i = 0;
	unsigned char eeprom_data[EEPROM_SIZE+1];
	unsigned long p = *fpos, eeprom_p;

	if (fpos != &file->f_pos)
		return -ESPIPE;

	if ((p % 2))
		return -ESPIPE;

	if (p >= EEPROM_SIZE)
		return 0;

	if ((p + count) >= EEPROM_SIZE)
		size = (EEPROM_SIZE - p);
	else
		size = count;

	spin_lock(&ixp425_eeprom_lock);
	if (copy_from_user(eeprom_data, buf, size)) {
		spin_unlock(&ixp425_eeprom_lock);
		return -EFAULT;
	}

	eeprom_ewen();

	eeprom_p = p / 2;
	for (offset = 0, i = 0; offset < size; offset += 2, i++) {
		w1 = eeprom_data[offset];
		w2 = eeprom_data[offset+1];
		romdata = (w1 << 8) + w2;
		erase_data(eeprom_p+i);
		write_data(eeprom_p+i, romdata);
	}

	eeprom_ewds();
	buf += size;
	*fpos += size;	
	spin_unlock(&ixp425_eeprom_lock);
	return size;
}

static loff_t
eeprom_llseek(struct file * file, loff_t offset, int orig)
{
        switch (orig) {
                case 0:
			if (offset < 0 || offset > EEPROM_SIZE)
				return -EINVAL;

                        file->f_pos = offset;
                        return file->f_pos;
                case 1:
	                if ((file->f_pos + offset) > EEPROM_SIZE)
                        	return -EINVAL;
                	if ((file->f_pos + offset) < 0)
                        	return -EINVAL;

                        file->f_pos += offset;
                        return file->f_pos;
                default:
                        return -EINVAL;
        }
}

static int
eeprom_release(struct inode *inode, struct file *file)
{
	spin_lock(&ixp425_eeprom_lock);
	if (MINOR(inode->i_rdev) == IXP425_EEPROM_MINOR)
		eeprom_status &= ~EEPROM_IS_OPEN;

	gpio_line_config(IXP425_EEPROM_CS_PIN, IXP425_GPIO_IN);
	gpio_line_config(IXP425_EEPROM_CLK_PIN, IXP425_GPIO_IN);
	gpio_line_config(IXP425_EEPROM_IN_PIN, IXP425_GPIO_IN);
	udelay(1);
	spin_unlock(&ixp425_eeprom_lock);
	return 0;
}

/*
 * We enforce only one user at a time here with the open/close.
 */
static int
eeprom_open(struct inode *inode, struct file *file)
{
	spin_lock(&ixp425_eeprom_lock);
	if (MINOR(inode->i_rdev) != IXP425_EEPROM_MINOR) {
		spin_unlock(&ixp425_eeprom_lock);
		return -ENODEV;
	}

	if (eeprom_status & EEPROM_IS_OPEN) {
		spin_unlock(&ixp425_eeprom_lock);
		return -EBUSY;
	}

	eeprom_status |= EEPROM_IS_OPEN;
	gpio_line_config(IXP425_EEPROM_CS_PIN, IXP425_GPIO_OUT);
	gpio_line_config(IXP425_EEPROM_CLK_PIN, IXP425_GPIO_OUT);
	gpio_line_config(IXP425_EEPROM_IN_PIN, IXP425_GPIO_OUT);
	udelay(1);
	spin_unlock(&ixp425_eeprom_lock);
	return 0;
}

static struct file_operations eeprom_fops = {
	owner:		THIS_MODULE,
	read:		eeprom_read,
	open:		eeprom_open,
	write:		eeprom_write,
	release:	eeprom_release,
	llseek:		eeprom_llseek,
};

static void __exit eeprom_cleanup_module(void)
{
	unregister_chrdev(IXP425_EEPROM_MAJOR, "ixp425-eeprom");
}

static int __init eeprom_init_module(void)
{
	int result = 0;

	result = register_chrdev(IXP425_EEPROM_MAJOR, "ixp425-eeprom", &eeprom_fops);
	if (result < 0) {
		printk("EEPROM register fails\n");
		goto register_error;
	}

	printk(KERN_INFO "EEPROM Program Drivers\n");
	return 0;

register_error:
	return result;
}

module_init(eeprom_init_module);
module_exit(eeprom_cleanup_module);
EXPORT_NO_SYMBOLS;
