/************************************************
 Cryptolib.c
 Cryptographic front end. 
*************************************************/

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/if_crypto.h>
#include <linux/slab.h>
#include <linux/errno.h>
#include <linux/random.h>
#include <linux/init.h>

/* S/W crypto includes */
#ifdef USE_RC4
#include <openssl/rc4.h>
#endif /* USE_RC4 */
#include <linux/des.h>
#include <linux/md5.h>
#include <linux/sha1.h>

#define ENGINE_MAX	5

#ifndef TRUE
#define TRUE 	(1==1)
#define FALSE	!TRUE
#endif

static	
struct crypto_engine_def *engine_db[ENGINE_MAX];

struct crypto_request *crypto_alloc_request(int flags)
{
	struct crypto_request *req;
	req = kmalloc(sizeof(struct crypto_request),flags);
	if(req == NULL)
		return NULL;
	memset(req,0,sizeof(struct crypto_request));
	return req;
}

struct crypto_engine_def *crypto_open_engine(crypto_cmd_t cap,unsigned int attrib)
{
	struct crypto_engine_def *eng;
	int t;

	for(t=0;t<ENGINE_MAX;t++)
	{
		eng = engine_db[t];
		if(eng == NULL)
			continue;
		if((cap & eng->capability) == cap
			&& (attrib & eng->attributes) == attrib)
		{
			if(eng->open != NULL)
				eng->open();
			return eng;
		}
	}
	return NULL;
}

void crypto_close_engine(struct crypto_engine_def *eng)
{
	if(eng != NULL && eng->close != NULL)
		eng->close();
}

#ifdef USE_RC4
#define BASIC_CMDS	(CRYPT_RC4|CRYPT_DES|CRYPT_3DES|CRYPT_MD5|CRYPT_SHA|CRYPT_RANDOM)
#else
#define BASIC_CMDS	(CRYPT_DES|CRYPT_3DES|CRYPT_MD5|CRYPT_SHA|CRYPT_RANDOM)
#endif /* USE_RC4 */
int crypto_execute(struct crypto_request *req)
{
	struct crypto_engine_def *eng;
	int t,ret;
/* Jack ++ 2003/11/12 */
#if 0
	if(req != NULL)
	{
		if(req->engine == NULL)
		{
			for(t=0;t<ENGINE_MAX;t++)
			{
				eng = engine_db[t];
				if(eng == NULL)
					continue;
			/* Engines are sorted in performance order, take first one that matches */
				if((req->op.cmd & eng->capability) == (req->op.cmd & BASIC_CMDS))
					break;
			}
		}
		else 
			eng = req->engine;

		if(eng != NULL && eng->request_handler != NULL)
		{
			ret = eng->request_handler(req);
			if(ret == CRYPT_READY && req->callback != NULL)
				req->callback(req);

			return ret;
		}
	}
#else
	return ixp425_crypto_request_fnc(req);
#endif
	return CRYPT_ERROR;
}

int crypto_query(crypto_cmd_t cap,unsigned int attrib)
{
	struct crypto_engine_def *eng;
	int t;

	for(t=0;t<ENGINE_MAX;t++)
	{
		eng = engine_db[t];
		if(eng == NULL)
			continue;
		if((cap & eng->capability) == cap
			&& (attrib & eng->attributes) == attrib)
			return TRUE;
	}
	return FALSE;
}



int crypto_register_engine(struct crypto_engine_def *eng)
{
	int t,r;
	struct crypto_engine_def **slot = NULL;

	if(eng == NULL || eng->name == NULL)
		return -EINVAL;

	eng->capability &= BASIC_CMDS;
	for(t=0;t<ENGINE_MAX;t++)
	{
		/* Put S/W last, H/W first */
		if(eng->attributes & IS_HARDWARE)
			r = (ENGINE_MAX-1) - t;
		else
			r = t;
		if(engine_db[r] == NULL)
		{
			slot = &engine_db[r];
		}
		else
		{
			if(strcmp(eng->name,engine_db[r]->name) == 0)
				return -EEXIST;
		}
	}
	if(slot == NULL)
		return -EMFILE;

	*slot = eng;
	printk(KERN_INFO "Crypto engine %s registered\n",eng->name);
	return 0;
}

int crypto_unregister_engine(char *name)
{
	int t;

	for(t=0;t<ENGINE_MAX;t++)
	{
		if(engine_db[t] != NULL && strcmp(engine_db[t]->name,name) == 0)
		{
			printk(KERN_INFO "Crypto engine %s unregistered\n",engine_db[t]->name);
			engine_db[t] = NULL;
			return 0;
		}
	}
	return -ENOENT;
}

/***************************************************
* S/W crypto engine
****************************************************/
static inline
int do_hmac(struct crypto_request *req,struct crypto_operation *op)
{
	unsigned char digest[20];
	int hash_len,i;
	MD5_CTX md5_ctx;
	SHA1_CTX	sha1_ctx;
	unsigned char k_ipad[65];    /* inner padding -
                                      * key XORd with ipad
                                      */
	unsigned char k_opad[65];    /* outer padding -
                                      * key XORd with opad
                                      */
	       /* start out by storing key in pads */
	memset(k_ipad, 0, sizeof k_ipad);
	memset(k_opad, 0, sizeof k_opad);
	memcpy(k_ipad, op->u.hash.hmac_key,op->u.hash.hmac_key_len);
	memcpy(k_opad, op->u.hash.hmac_key,op->u.hash.hmac_key_len);

        /* XOR key with ipad and opad values */
	for (i=0; i<64; i++) {
		k_ipad[i] ^= 0x36;
		k_opad[i] ^= 0x5c;
	}
	if(op->cmd & CRYPT_MD5)
	{
        /*
         * perform inner MD5
         */
		hash_len = 16;
		MD5Init(&md5_ctx);                   /* init md5_ctx for 1st
                                              * pass */
		MD5Update(&md5_ctx, k_ipad, 64);      /* start with inner pad */
		MD5Update(&md5_ctx, &req->outbuf[op->offset], op->len); /* then text of datagram */
		MD5Final(digest, &md5_ctx);          /* finish up 1st pass */
        /*
         * perform outer MD5
         */
		MD5Init(&md5_ctx);                   /* init md5_ctx for 2nd
                                              * pass */
		MD5Update(&md5_ctx, k_opad, 64);     /* start with outer pad */
		MD5Update(&md5_ctx, digest, 16);     /* then results of 1st
                                              * hash */
		MD5Final(digest, &md5_ctx);          /* finish up 2nd pass */
	}
	else if(op->cmd & CRYPT_SHA)
	{
		hash_len = 20;
		SHA1Init(&sha1_ctx);                   /* init sha1_ctx for 1st
                                              * pass */
		SHA1Update(&sha1_ctx, k_ipad, 64);      /* start with inner pad */
		SHA1Update(&sha1_ctx, &req->outbuf[op->offset], op->len); /* then text of datagram */
		SHA1Final(digest, &sha1_ctx);          /* finish up 1st pass */
        /*
         * perform outer 
         */
		SHA1Init(&sha1_ctx);                   /* init sha1_ctx for 2nd
                                              * pass */
		SHA1Update(&sha1_ctx, k_opad, 64);     /* start with outer pad */
		SHA1Update(&sha1_ctx, digest, 20);     /* then results of 1st
                                              * hash */
		SHA1Final(digest, &sha1_ctx);          /* finish up 2nd pass */
	}
	if(op->cmd & CRYPT_96)
	{
		hash_len = 96/8;	/* Truncate digest */
	}

	if(op->cmd & CRYPT_APPEND)
	{
		if(req->outlen < (req->inlen + hash_len))
			return CRYPT_ERROR;
		memcpy(req->outbuf + req->inlen, digest, hash_len);
		req->outlen = req->inlen + hash_len;
	}
	if(req->outlen < hash_len)
		return CRYPT_ERROR;
	memcpy(op->u.hash.result, digest, hash_len);
	op->u.hash.reslen= hash_len;
	return CRYPT_READY;
}

static inline
int do_regular_hash(struct crypto_request *req,struct crypto_operation *op)
{
	unsigned char digest[20];
	int hash_len;
	MD5_CTX md5_ctx;
	SHA1_CTX	sha1_ctx;

	if(op->cmd & CRYPT_MD5)
	{
		hash_len = 16;
		MD5Init(&md5_ctx);                  
		MD5Update(&md5_ctx, &req->outbuf[op->offset], op->len);
		MD5Final(digest, &md5_ctx);          
	}
	else if(op->cmd & CRYPT_SHA)
	{
		hash_len = 20;
		SHA1Init(&sha1_ctx);                  
		SHA1Update(&sha1_ctx, &req->outbuf[op->offset], op->len); /* then text of datagram */
		SHA1Final(digest, &sha1_ctx);          /* finish up 1st pass */
	}
	if(op->cmd & CRYPT_96)
	{
		hash_len = 96/8;	/* Truncate digest */
	}

	if(op->cmd & CRYPT_APPEND)
	{
		if(req->outlen < (req->inlen + hash_len))
			return CRYPT_ERROR;
		memcpy(req->outbuf + req->inlen, digest, hash_len);
		req->outlen = req->inlen + hash_len;
	}

	memcpy(op->u.hash.result, digest, hash_len);
	op->u.hash.reslen= hash_len;
	return CRYPT_READY;
}

static
int sw_request_handler(struct crypto_request *req)
{
	unsigned char *in, *out;
	int t,ret,maxout,reslen=0;
	struct crypto_operation *op;

	maxout = req->outlen;
	if(req->outlen  < req->inlen)
		return CRYPT_ERROR;

	if(req->inbuf != NULL && req->inbuf != req->outbuf)
		memcpy(req->outbuf,req->inbuf,req->inlen);

	for(op = &req->op;op != NULL;op=op->next)
	{
		in = &req->outbuf[op->offset];
		out = &req->outbuf[op->offset];

		if(op->cmd & CRYPT_RANDOM)
		{
			get_random_bytes(req->outbuf,req->outlen);
			reslen = req->outlen;
			break; /* Should be the only real operation... */
		}

#ifdef USE_RC4
		if(op->cmd & CRYPT_RC4)
		{
			RC4_KEY key;

			if(maxout < req->inlen || req->op.u.rc4.key == NULL)
				return CRYPT_ERROR;

			RC4_set_key(&key,op->u.rc4.key->len,op->u.rc4.key->raw);
			RC4(&key,op->len,in,out);
			req->outlen = req->inlen;
		}
#endif /* USE_RC4 */

		if(op->cmd & CRYPT_DES)
		{
			/* To do */
			return CRYPT_ERROR;
		}
		if(op->cmd & CRYPT_3DES)
		{
			if(maxout < req->inlen)
				return CRYPT_ERROR;

			if(op->cmd & CRYPT_CBC)
			{
				des_key_schedule sched[3];
				des_cblock ivec;

				memcpy(&ivec, op->u.des.ivec, 8);

				for(t=0;t<3;t++)
					des_set_key(&(op->u.des.key[t]),sched[t]);
				des_ede3_cbc_encrypt((des_cblock*)in,(des_cblock*)out,op->len,sched[0],sched[1],sched[2]
					,&ivec,(op->cmd & CRYPT_DECODE)?0:1);
				req->outlen = req->inlen;
			}
			if(op->cmd & CRYPT_ECB)
			{
				/* Todo */
				return CRYPT_ERROR;
			}
		}
		if(op->cmd & (CRYPT_MD5|CRYPT_SHA))
		{
			if(op->cmd & CRYPT_HMAC)
			{
				ret = do_hmac(req,op);
				if(ret != CRYPT_READY)
					return ret;
			}
			else
			{
				ret = do_regular_hash(req,op);
				if(ret != CRYPT_READY)
					return ret;
			}
		}
		reslen = req->outlen > reslen? req->outlen: reslen;
		req->outlen = maxout;
	}

	req->outlen = reslen;
	req->flags |= CRYPT_SOFTWARE;
	return CRYPT_READY;
}

/****** RANDOM NUMBERS HANDLING HERE ********/
static wait_queue_t *crypto_random_read_wait;

struct random_priv
{
	ssize_t	nbytes,count;
	char 		*buf;
};

static
void crypto_random_callback(struct crypto_request *req)
{
	struct random_priv 	*rp = (struct random_priv *) req->priv;

	rp->count += req->outlen;
	rp->buf += req->outlen;
	rp->nbytes -= req->outlen;
	crypto_free_request(req);
	wake_up_interruptible(&crypto_random_read_wait);
}	

static ssize_t
crypto_random_read(struct file * file, char * buf, size_t nbytes, loff_t *ppos)
{
	wait_queue_t       wait = { current, NULL };
	ssize_t                 n, retval = 0;
	struct crypto_request *req;
	struct random_priv	rpriv;

	if (nbytes == 0)
		return 0;

	rpriv.nbytes = nbytes;
	rpriv.count = 0;
	rpriv.buf = buf;

	add_wait_queue(&crypto_random_read_wait, &wait);
	current->state = TASK_INTERRUPTIBLE;

	n = rpriv.nbytes;
	if (file->f_flags & O_NONBLOCK) {
		return -EAGAIN;
	}
	if (signal_pending(current)) {
		return -ERESTARTSYS;
	}
		/* Go and get random data */
	req = crypto_alloc_request(GFP_KERNEL);
	req->op.cmd = CRYPT_RANDOM;
	req->op.offset = 0;
	req->op.len = n;
	req->inbuf = NULL;
	req->inlen = 0;
	req->outlen = n;
	req->outbuf = rpriv.buf;
	req->flags = 0;
	req->callback = crypto_random_callback;
	req->priv = (void*) &rpriv;
	crypto_execute(req);

	schedule();

	current->state = TASK_RUNNING;
	remove_wait_queue(&crypto_random_read_wait, &wait);


        /*
         * If we gave the user some bytes, update the access time.
         */
	if (rpriv.count != 0) {
		UPDATE_ATIME(file->f_dentry->d_inode);
	}

	return (rpriv.count ? rpriv.count : retval);
}

static struct file_operations crypto_random_fops = {
#if LINUX_VERSION_CODE >= 0x20300	/* Someone knows when these made their debut? */
	owner:		THIS_MODULE,
#endif
	read:		crypto_random_read	/* read */
};

static
void __init crypto_random_init(void)
{
	if (register_chrdev(121, "crypto_random", &crypto_random_fops))
	{
		printk(KERN_NOTICE "Can't allocate major number 121 for random.\n");
	}
}


static
struct crypto_engine_def sw_def = 
{
	name : "S/W",
#ifdef USE_RC4
	capability: (CRYPT_RC4|CRYPT_DES|CRYPT_3DES|CRYPT_MD5|CRYPT_SHA),
#else
	capability: (CRYPT_DES|CRYPT_3DES|CRYPT_MD5|CRYPT_SHA),
#endif /* USE_RC4 */
	attributes: IS_SOFTWARE | IS_SYNC,
	request_handler: sw_request_handler
};

void ixp425_inc_use_count()
{}

void ixp425_dec_use_count()
{}

/* Jack ++ 2003/09/26 */
#include <linux/interrupt.h>
#include <linux/spinlock.h>

int ixp425_crypto_request_fnc(struct crypto_request *req)
{
/*
        if(req->flags & CRYPT_SYNC)
        {
                return CRYPT_ERROR;
        }
*/
        ixCryptoAccCodeletMain(req); 
        return CRYPT_OK;
}

static
struct crypto_engine_def ixp425_def =
{
        name : "IXP425",
        capability: (CRYPT_DES|CRYPT_3DES|CRYPT_MD5|CRYPT_SHA),
        attributes: IS_HARDWARE | IS_ASYNC,
        request_handler: ixp425_crypto_request_fnc,
        open: ixp425_inc_use_count,
        close: ixp425_dec_use_count
};



int  __init crypto_init(void)
{
/*	crypto_register_engine(&sw_def); */
	crypto_register_engine(&ixp425_def); 
	crypto_random_init();
	return 0;
}

#ifdef MODULE
mod_exit_t crypto_cleanup (void)
{
	crypto_unregister_engine(sw.def.name);
}

module_exit(crypto_cleanup);
#endif

//#if LINUX_VERSION_CODE > 0x20300
module_init(crypto_init);
//#endif
