 /*
 * FileName:    IxAdslDriverInfo.c
 * Author: Intel Corporation
 * Created:     18-Dec-2000
 *
 * Note: 
 *    Extension to management interface code so that the ADSL
 * component supports the standard Atur MIB objects as set
 * out in RFC 2662 which are applicable to the Alcatel Phy
 * including some Atuc objects.
 * 
 * Description:
 *    IxAdsl Management Interface
 *
 * Design Notes:
 *
 * -- Intel Copyright Notice --
 * INTEL CONFIDENTIAL
 *
 * Copyright 2000-2002 Intel Corporation All Rights Reserved.
 *
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.  The Material
 * contains trade secrets and proprietary and confidential information of
 * Intel or its suppliers and licensors.  The Material is protected by
 * worldwide copyright and trade secret laws and treaty provisions. No
 * part of the Material may be used, copied, reproduced, modified,
 * published, uploaded, posted, transmitted, distributed, or disclosed in
 * any way without Intel's prior express written permission.
 *
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel or otherwise.  Any license under
 * such intellectual property rights must be express and approved by
 * Intel in writing.
 *
 * For further details, please see the file README.TXT distributed with
 * this software.
 * -- End Intel Copyright Notice --
 */



#include <stdio.h>
#include <stdlib.h>
#include <timers.h>
#include <string.h>
#include <tickLib.h>
#include <taskLib.h>
#include <sysLib.h>


#include "IxTimerCtrl.h"
#include "IxOsServices.h"
#include "IxAdsl.h"
#include "IxAdslMgmtInfo.h"
#include "IxInterfaceNums.h"
#include "IxManagerInfoTypes.h"

#include "ixCtrle_c.h"
#include "ixAdsl_p.h"


/* defines the number of interfaces supported by the ifTable */ 
#define IFENTRYSIZE 3

#define CURR_STATUS_NO_DEFECT 0x1
#define CURR_STATUS_LOSS_MARGIN 0x4
#define CURR_STATUS_LOSS_POWER 0x20
#define CURR_STATUS_LOSS_FRAME 0x40
#define CURR_STATUS_LOSS_SIGNAL 0x80
#define CURR_STATUS_LOSS_CD_INTERLEAVED 0x8
#define CURR_STATUS_LOSS_CD_NONINTERLEAVED 0x10 

#define SNMP_CURR_STATUS_NO_DEFECT 0x80
#define SNMP_CURR_STATUS_LOSS_FRAME 0x40
#define SNMP_CURR_STATUS_LOSS_SIGNAL 0x20
#define SNMP_CURR_STATUS_LOSS_POWER 0x10
#define SNMP_CURR_STATUS_LOSS_QUALITY 0x8

/* Translation from ifInterfaces to useful indexes */
#define ADSL_PHYS_INDEX 0
#define ADSL_ITLV_INDEX 1
#define ADSL_FAST_INDEX 2

/* Line coding type is always dmt mode */
#define CODING_DMT_MODE 2
/* Line Type is either fast or interleaved */ 
#define FAST_OR_INTERLEAVED 4

#define SERIALNO_STRING_LEN 60
#define NUM_ADSL_LINES 8

/* The number of seconds before a failure is declared */
#define IX_ADSL_FAILURE_SECS 3

/* Interval time in miliseconds for performance monitor */ 
#define IX_ADSL_PERF_TIMER 1000


typedef struct {
    INT32 IxAdslLineCoding;
    INT32 IxAdslLineType;
} IxAdslLineEntryData;

typedef struct {
    INT32 IxAdslAtucCurrSnrMgn;
    UINT32 IxAdslAtucCurrAtn;
    char IxAdslAtucCurrStatus;
    INT32 IxAdslAtucCurrOutputPwr;
    UINT32 IxAdslAtucCurrAttainableRate;
} IxAdslAtucPhysEntryData;

typedef struct {
    UINT8 *IxAdslAturInvVendorID;
    UINT8 *IxAdslAturInvVersionNumber;
    INT32 IxAdslAturCurrSnrMgn;
    UINT32 IxAdslAturCurrAtn;
    UINT8 IxAdslAturCurrStatus;
    T_adsl_def_bitmap_set *alcatelActualBitmap;
    INT32  IxAdslAturCurrOutputPwr;
    UINT32 IxAdslAturCurrAttainableRate;
} IxAdslAturPhysEntryData;

typedef struct {
    UINT32 IxAdslAturChanInterleaveDelay;
    T_adslChannelOperData channelOperData;
    UINT32 IxAdslAturChanCrcBlockLength;
} IxAdslAturChanTable;

/*
 * Phys Address need not be in the structure as for any of the three
 * interfaces it will always return an OCTET String of zero
 * length.
 */ 
typedef struct {
    UINT8* IxAdslIfDescr;
    INT32 IxAdslIfType;
    INT32 IxAdslIfMtu;
    UINT32 IxAdslIfSpeed;
    INT32 IxAdslIfAdminStatus;
    UINT32 IxAdslIfLastChange;
    UINT32 IxAdslIfInOctets;
    UINT32 IxAdslIfInUcastPkts;
    UINT32 IxAdslIfInDiscards;
    UINT32 IxAdslIfInErrors;
    UINT32 IxAdslIfInUnknownProtos;
    UINT32 IxAdslIfOutOctets;
    UINT32 IxAdslIfOutUcastPkts;
    UINT32 IxAdslIfOutDiscards;
    UINT32 IxAdslIfOutErrors;
    INT8* IxAdslIfName;
    UINT32 IxAdslIfInMulticastPkts;
    UINT32 IxAdslIfInBroadcastPkts;
    UINT32 IxAdslIfOutMulticastPkts;
    UINT32 IxAdslIfOutBroadcastPkts;
    UINT32 IxAdslIfHighSpeed;
    INT32 IxAdslIfPromiscuousMode;
    INT32 IxAdslIfConnectorPresent;
    UINT32 IxAdslIfCounterDiscontinuityTime;
    UINT32 IxAdslIfStackLowerLevel;
} IxIfEntryData;


PRIVATE IxAdslLineEntryData ixAdslLineEntryData[NUM_ADSL_LINES];

PRIVATE IxAdslAtucPhysEntryData ixAdslAtucPhysEntryData[NUM_ADSL_LINES];

PRIVATE IxAdslAturPhysEntryData ixAdslAturPhysEntryData[NUM_ADSL_LINES];

PRIVATE IxAdslAturChanTable ixAdslAturChanTable[NUM_ADSL_LINES];

IxadslAturPerfDataEntry ixAdslAturPerfDataEntry[NUM_ADSL_LINES];

IxadslAturPerfRetrainStats ixAdslAturPerfRetrainStats[NUM_ADSL_LINES];

PRIVATE IxIfEntryData ixIfEntryData[NUM_ADSL_LINES][IFENTRYSIZE];

PRIVATE IxAdslDriverTrapCallbackPtr_t IxAdslDriverTrapCallbackPtr = (IxAdslDriverTrapCallbackPtr_t)NULL;

PRIVATE unsigned char ixAdslMinNoiseMarginUpstr;
PRIVATE char ixAdslInvSerialNumber[SERIALNO_STRING_LEN];

PRIVATE struct T_Controller_SW_version ixAdslControllerSwVersion[NUM_ADSL_LINES];
/* TODO: initiailise in ManagementInfoinit */
PRIVATE T_LTadslNearEndLineOperData ixAdslNearOperData[NUM_ADSL_LINES];
PRIVATE T_LTadslFarEndLineOperData ixAdslFarOperData[NUM_ADSL_LINES];
PRIVATE T_adslChannelOperData ixAdslChannelOperData[NUM_ADSL_LINES];

/*
 * There is a subtle difference between 'Loss of' seconds and failures.
 * A 'Loss of' second is a defect which has occured in the last second,
 * whereas a failure is declared when a defect has been active for the
 * last 2.5 seconds +/- 0.5 seconds. A failure is a more terminal event
 * and would normally cause a modem resync event. - DH 04/04/01 
 */
PRIVATE BOOL ixAdslLofs[NUM_ADSL_LINES]={FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE};	/* Loss of Framing failure */
PRIVATE BOOL ixAdslLoss[NUM_ADSL_LINES]={FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE};	/* Loss of Signal failure */
PRIVATE BOOL ixAdslLprs[NUM_ADSL_LINES]={FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE};	/* Loss of Power failure */
PRIVATE BOOL ixAdslLcds[NUM_ADSL_LINES]={FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE};	/* Loss of Cell delineation second */
PRIVATE BOOL ixAdslLoms[NUM_ADSL_LINES]={FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE};	/* Loss of Margin in last second */

static BOOL adminInit[NUM_ADSL_LINES]={FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE};

static UCHAR ixAdslLofCounter[NUM_ADSL_LINES]={0,0,0,0,0,0,0,0}; /* Counters used to declare failures */
static UCHAR ixAdslLosCounter[NUM_ADSL_LINES]={0,0,0,0,0,0,0,0};
static UCHAR ixAdslLprCounter[NUM_ADSL_LINES]={0,0,0,0,0,0,0,0};
static UCHAR ixAdslLcdCounter[NUM_ADSL_LINES]={0,0,0,0,0,0,0,0};
static UCHAR ixAdslLomCounter[NUM_ADSL_LINES]={0,0,0,0,0,0,0,0};


/*
 * This function is used to verify if the index passed to 
 * DriverInfo Get or Set is one of the three valid
 * interface numbers IF_ADSL, IF_ADSL_ITLV and IF_ADSL_FAST
 * and adjusts the index to a range of 0-2.
 */
PRIVATE int
ixAdslIfIndexAdjust(UINT32 *ifIndex)
{
    int retVal;

    switch(*ifIndex)
    {
	case IF_ADSL:
	    *ifIndex = ADSL_PHYS_INDEX;
	    retVal = IX_SUCCESS;
	    break;
	case IF_ADSL_ITLV:
	    *ifIndex = ADSL_ITLV_INDEX;
	    retVal = IX_SUCCESS;
	    break;
	case IF_ADSL_FAST:
	    *ifIndex = ADSL_FAST_INDEX;
	    retVal = IX_SUCCESS;
	    break;
	default:
	    return IX_FAIL;
    }
    return retVal;
}

PRIVATE int 
ixAdslCheckOperStatus(unsigned int lineNum, int ifIndex)
{
    int retVal = IX_MGRINF_OPER_DOWN;

    if(IxADSL_GetLineState(lineNum) == IX_ADSL_LINE_UP)
    {
	if(ifIndex == ADSL_PHYS_INDEX)
	{
	    retVal = IX_MGRINF_OPER_UP;
	}
	else
	{
	    if(C_CONF_DATA_ACK != ADSL_get_ChannelOperData(lineNum, &ixAdslChannelOperData[lineNum]))
	    {
		retVal = IX_MGRINF_OPER_DOWN;
	    }
	    else
	    {
		if(ifIndex == ADSL_ITLV_INDEX)
		{
		    if(ixAdslChannelOperData[lineNum].actualBitrateNearInt != 0)
		    {
			retVal = IX_MGRINF_OPER_UP;
		    }
		}
		else if(ifIndex == ADSL_FAST_INDEX)
		{
		    if(ixAdslChannelOperData[lineNum].actualBitrateNearFast != 0)
		    {
			retVal = IX_MGRINF_OPER_UP;
		    }
		}
	    }
	}
    }
    return retVal;
}

PRIVATE void
ixAdslAdminInit(unsigned int lineNum)
{

    if(adminInit[lineNum] == FALSE &&
       ixAdslCheckOperStatus(lineNum, ADSL_PHYS_INDEX) == IX_MGRINF_OPER_UP)
    {
	if(ixAdslCheckOperStatus(lineNum, ADSL_ITLV_INDEX) == IX_MGRINF_OPER_UP)
	{
	    ixIfEntryData[lineNum][ADSL_ITLV_INDEX].IxAdslIfAdminStatus = IX_MGRINF_ADMIN_UP;
	}
	if(ixAdslCheckOperStatus(lineNum, ADSL_FAST_INDEX) == IX_MGRINF_OPER_UP)
	{
	    ixIfEntryData[lineNum][ADSL_FAST_INDEX].IxAdslIfAdminStatus = IX_MGRINF_ADMIN_UP;
	}
	adminInit[lineNum] = TRUE;
    }
}

PRIVATE int
ixAdslOrderBits(unsigned int lineNum)
{
    unsigned int rc;
    T_adsl_def_bitmap_set currStatus;

    ixAdslAturPhysEntryData[lineNum].IxAdslAturCurrStatus = 0;

    rc = ADSL_def_bitmap_poll(lineNum, &currStatus);
    if ( rc == C_CONF_DATA_ACK )/* The data is good */
    {
	if((currStatus.near_end.status & CURR_STATUS_LOSS_FRAME) ||
	   (currStatus.near_end.status & CURR_STATUS_LOSS_SIGNAL) ||
	   (currStatus.near_end.status & CURR_STATUS_LOSS_POWER) ||
	   (ixAdslNearOperData[lineNum].noiseMarginUpstr < ixAdslMinNoiseMarginUpstr))
	{

	    if(currStatus.near_end.status & CURR_STATUS_LOSS_FRAME)
	    {
		ixAdslAturPhysEntryData[lineNum].IxAdslAturCurrStatus |= SNMP_CURR_STATUS_LOSS_FRAME; 
	    }
	    
	    if(currStatus.near_end.status & CURR_STATUS_LOSS_SIGNAL)
	    {
		ixAdslAturPhysEntryData[lineNum].IxAdslAturCurrStatus |= SNMP_CURR_STATUS_LOSS_SIGNAL; 
	    }
	    
	    if(currStatus.near_end.status & CURR_STATUS_LOSS_POWER)
	    {
		ixAdslAturPhysEntryData[lineNum].IxAdslAturCurrStatus |= SNMP_CURR_STATUS_LOSS_POWER; 
	    }
	    
	    if(ixAdslNearOperData[lineNum].noiseMarginUpstr < ixAdslMinNoiseMarginUpstr)
	    {
		ixAdslAturPhysEntryData[lineNum].IxAdslAturCurrStatus |= SNMP_CURR_STATUS_LOSS_QUALITY;
	    }
	    
	    return IX_SUCCESS;
	}
	else
	{
	    ixAdslAturPhysEntryData[lineNum].IxAdslAturCurrStatus |= SNMP_CURR_STATUS_NO_DEFECT; 
	    
	    return IX_SUCCESS;
	}
    }
    return IX_FAIL;
}

PRIVATE int
ixAdslSyncWithModem(unsigned int lineNum)
{
    int rc;
    /*
     * GetLine Statistics.
     */

    rc = LT_ADSL_get_NearEndLineOperData(lineNum, & ixAdslNearOperData[lineNum]);
    
    if(rc != C_CONF_DATA_ACK )
    {
	return IX_FAIL;
    }

    
    rc = LT_ADSL_get_FarEndLineOperData(lineNum, & ixAdslFarOperData[lineNum]);
     
    if(rc != C_CONF_DATA_ACK )
    {
	return IX_FAIL;
    }


    /* Get Channel Oper Data */
    rc = ADSL_get_ChannelOperData(lineNum, &ixAdslChannelOperData[lineNum]);
    
    if(rc != C_CONF_DATA_ACK )
    {
	return IX_FAIL;
    }

    
    /* Get Controller Software version */
    
    if ( ADSL_get_controller_SWVersion(lineNum, &ixAdslControllerSwVersion[lineNum]) != 
	 C_CONF_DATA_ACK )
    {
	return IX_FAIL;
    }
    return IX_SUCCESS;
}


INT32
ixAdslDriverInfoGet(unsigned int lineNum,
		    const IxAdslManagerInfo entity,
		    UINT32 ifIndex,
		    ManagerInfoData *dataBuffer)
{   
    struct timespec now;
    PRIVATE struct timespec lastinvoked={0,0};
    INT32 retVal = IX_FAIL;
    t_ADSL_PHY_TYPE AdslPhyType;
    
    /*
     * Which end of the line are we on?
     */
     AdslPhyType=ixAdslPhyTypeGet(lineNum);

    /*
     * We only want to sync if the data we currently have is 
     * 10 seconds out of date or more, since it is an 
     * expensive operation
     */

    clock_gettime(CLOCK_REALTIME, &now);

    if((now.tv_sec - lastinvoked.tv_sec > 10) || 
       (now.tv_sec - lastinvoked.tv_sec < 0))
    {
	if(ixAdslSyncWithModem(lineNum) == IX_SUCCESS)
	{
	    lastinvoked.tv_sec=now.tv_sec;
	}
    }

    if(dataBuffer == NULL)
    {	
        return IX_FAIL;
    }    
       
    if(dataBuffer->dataPtr == NULL)
    {
	return IX_FAIL;
    }

    switch(entity)
    {
        case IxAdslLineCoding_TRO: /*The value here is always dmt set at init*/
            if(ifIndex == IF_ADSL)
            {
	        if(dataBuffer->dataLength != sizeof(INT32))
	        {
		    retVal = IX_FAIL;
	        }
                else
                {
                    *(INT32*)dataBuffer->dataPtr = ixAdslLineEntryData[lineNum].IxAdslLineCoding;
		    retVal = IX_SUCCESS;
                } 
            }
            break;

        case IxAdslLineType_TRO: /*The value here is always interleavedorfast*/
             if(ifIndex == IF_ADSL)
	     {
		 if(dataBuffer->dataLength != sizeof(INT32))
		 {
		     retVal = IX_FAIL;
		 }
		 else
		 {
		     *(INT32*)dataBuffer->dataPtr = ixAdslLineEntryData[lineNum].IxAdslLineType;
		     retVal = IX_SUCCESS;
		 } 
	     }
	     break;
	     
	case IxAdslAtucInvSerialNumber_TRO:
	    if(ifIndex == IF_ADSL)
            {
	        if(dataBuffer->dataLength < strlen(ixAdslInvSerialNumber))
	        {
		    retVal = IX_FAIL;
	        }
                else
                {
		    if(AdslPhyType!=IX_ADSL_CO)
		    {
			return IX_MGRINF_NOTSUPPORTED;
		    }
                    strcpy(dataBuffer->dataPtr, ixAdslInvSerialNumber);
                    dataBuffer->dataLength = strlen((char*)dataBuffer->dataPtr);
		    retVal = IX_SUCCESS;
                } 

            }
            break;

	case IxAdslAtucInvVendorID_TRO: 
            if(ifIndex == IF_ADSL)
            {
	        if(dataBuffer->dataLength < strlen("Vendor Id"))
	        {
		    retVal = IX_FAIL;
	        }
                else
                {
		    if(AdslPhyType!=IX_ADSL_CO)
		    {
			return IX_MGRINF_NOTSUPPORTED;
		    }
                    strcpy(dataBuffer->dataPtr,"Vendor Id");
                    dataBuffer->dataLength = strlen((char*)dataBuffer->dataPtr);
		    retVal = IX_SUCCESS;
                } 
            }
            break;
	case IxAdslAtucInvVersionNumber_TRO:
	    if(ifIndex == IF_ADSL)
            {
	        if(dataBuffer->dataLength < strlen("Version Number"))
	        {
		    retVal = IX_FAIL;
	        }
                else
                {
		    if(AdslPhyType!=IX_ADSL_CO)
		    {
			return IX_MGRINF_NOTSUPPORTED;
		    }
                    strcpy(dataBuffer->dataPtr,"Version Number");
                    dataBuffer->dataLength = strlen((char*)dataBuffer->dataPtr);
		    retVal = IX_SUCCESS;
                } 
            }
            break;
   
        case IxAdslAtucCurrSnrMgn_TRO:
	    if(ifIndex == IF_ADSL)
            {
	        if(dataBuffer->dataLength != sizeof(INT32))
	        {
		    retVal = IX_FAIL;
	        }
                else
                {
		    if(AdslPhyType!=IX_ADSL_CO)
		    {
			return IX_MGRINF_NOTSUPPORTED;
		    }
                    *(INT32*)dataBuffer->dataPtr = ixAdslNearOperData[lineNum].noiseMarginUpstr;
		    retVal = IX_SUCCESS;
                } 
            }
            break;
        case IxAdslAtucCurrAtn_TRO:
	    if(ifIndex == IF_ADSL)
            {
 	        if(dataBuffer->dataLength != sizeof(UINT32))
	        {
	    	    retVal = IX_FAIL;
	        }
                else
                {
		    if(AdslPhyType!=IX_ADSL_CO)
		    {
			return IX_MGRINF_NOTSUPPORTED;
		    }
                    *(UINT32*)dataBuffer->dataPtr = ixAdslNearOperData[lineNum].attenuationUpstr;
		    retVal = IX_SUCCESS;
                } 
            }
            break;
        case IxAdslAtucCurrStatus_TRO:
	    return IX_MGRINF_NOTSUPPORTED;
        case IxAdslAtucCurrOutputPwr_TRO:
            if(ifIndex == IF_ADSL)
            {
	        if(dataBuffer->dataLength != sizeof(INT32))
	        {
		    retVal = IX_FAIL;
	        }
                else
                {
		    if(AdslPhyType!=IX_ADSL_CO)
		    {
			return IX_MGRINF_NOTSUPPORTED;
		    }
		    ixAdslAtucPhysEntryData[lineNum].IxAdslAtucCurrOutputPwr= ixAdslNearOperData[lineNum].outputPowerDnstr;
                    *(INT32*)dataBuffer->dataPtr = ixAdslAtucPhysEntryData[lineNum].IxAdslAtucCurrOutputPwr;
		    retVal = IX_SUCCESS;
                } 
            }
            break;
        case IxAdslAtucCurrAttainableRate_TRO:
            if(ifIndex == IF_ADSL)
            {
 	        if(dataBuffer->dataLength != sizeof(UINT32))
	        {
	    	    retVal = IX_FAIL;
	        }
                else
                {
		    if(AdslPhyType!=IX_ADSL_CO)
		    {
			return IX_MGRINF_NOTSUPPORTED;
		    }
		     ixAdslAtucPhysEntryData[lineNum].IxAdslAtucCurrAttainableRate=(ixAdslFarOperData[lineNum].attainableBitrate)*1000;
                    *(UINT32*)dataBuffer->dataPtr = ixAdslAtucPhysEntryData[lineNum].IxAdslAtucCurrAttainableRate;
		    retVal = IX_SUCCESS;
                } 
            }
            break;

	    /*
	     * Begin ATUR stats
	     */
	case IxAdslAturInvSerialNumber_TRO:
            if(ifIndex == IF_ADSL)
            {
	        if(dataBuffer->dataLength < strlen(ixAdslInvSerialNumber))
	        {
		    retVal = IX_FAIL;
	        }
                else
                {
		    if(AdslPhyType!=IX_ADSL_CPE)
		    {
			return IX_MGRINF_NOTSUPPORTED;
		    }
                    strcpy(dataBuffer->dataPtr, ixAdslInvSerialNumber);
                    dataBuffer->dataLength = strlen(dataBuffer->dataPtr);
		    retVal = IX_SUCCESS;
                } 
            }
            break;

	case IxAdslAturInvVendorID_TRO:
	    if(ifIndex == IF_ADSL)
            {
	        if(dataBuffer->dataLength < strlen((char*)ixAdslAturPhysEntryData[lineNum].IxAdslAturInvVendorID))
	        {
		    retVal = IX_FAIL;
	        }
                else
                {
		    if(AdslPhyType!=IX_ADSL_CPE)
		    {
			return IX_MGRINF_NOTSUPPORTED;
		    }
		    /* The SNMP admin string for vendorId will always be 16bits or 16 char's */
		    strcpy((char*)dataBuffer->dataPtr,
			    (char*)ixAdslAturPhysEntryData[lineNum].IxAdslAturInvVendorID);
                    dataBuffer->dataLength = strlen((char*)dataBuffer->dataPtr);
		    retVal = IX_SUCCESS;
                } 
            }
            break;
	case IxAdslAturInvVersionNumber_TRO:
	    if(ifIndex == IF_ADSL)
            {
	        if(dataBuffer->dataLength < strlen((char*)ixAdslAturPhysEntryData[lineNum].IxAdslAturInvVersionNumber))
	        {
		    retVal = IX_FAIL;
	        }
                else
                {
		    if(AdslPhyType!=IX_ADSL_CPE)
		    {
			return IX_MGRINF_NOTSUPPORTED;
		    }
		    strcpy((char*)dataBuffer->dataPtr,
			    (char*)ixAdslAturPhysEntryData[lineNum].IxAdslAturInvVersionNumber);
                    dataBuffer->dataLength = strlen((char*)dataBuffer->dataPtr);
		    retVal = IX_SUCCESS;
                } 
            }
            break;

        case IxAdslAturCurrSnrMgn_TRO:
            if(ifIndex == IF_ADSL)
            {
	        if(dataBuffer->dataLength != sizeof(INT32))
	        {
		    retVal = IX_FAIL;
	        }
                else
                {
		    if(AdslPhyType!=IX_ADSL_CPE)
		    {
			return IX_MGRINF_NOTSUPPORTED;
		    }
                    *(INT32*)dataBuffer->dataPtr = ixAdslNearOperData[lineNum].noiseMarginUpstr;
		    retVal = IX_SUCCESS;
                } 
            }
            break;

        case IxAdslAturCurrAtn_TRO:
            if(ifIndex == IF_ADSL)
            {
 	        if(dataBuffer->dataLength != sizeof(UINT32))
	        {
	    	    retVal = IX_FAIL;
	        }
                else
                {
		    if(AdslPhyType!=IX_ADSL_CPE)
		    {
			return IX_MGRINF_NOTSUPPORTED;
		    }
                    *(UINT32*)dataBuffer->dataPtr = ixAdslNearOperData[lineNum].attenuationUpstr;
		    retVal = IX_SUCCESS;
                } 
            }
            break;

        case IxAdslAturCurrStatus_TRO:
	    if(ifIndex == IF_ADSL)
            {
		if(AdslPhyType!=IX_ADSL_CPE)
		{
		    return IX_MGRINF_NOTSUPPORTED;
		}
		/* 
		 * SNMP expects one byte with 5 bits (0-4) toggled depending on the the error conditions
		 * present or not present.
		 */
		if((retVal = ixAdslOrderBits(lineNum)) == IX_SUCCESS)
		{
		    if(dataBuffer->dataLength < sizeof(ixAdslAturPhysEntryData[lineNum].IxAdslAturCurrStatus))
		    {
			retVal = IX_FAIL;
		    }
		    else
		    {
			*(UINT8*)dataBuffer->dataPtr = ixAdslAturPhysEntryData[lineNum].IxAdslAturCurrStatus;
			dataBuffer->dataLength = sizeof(ixAdslAturPhysEntryData[lineNum].IxAdslAturCurrStatus);
			retVal = IX_SUCCESS;
		    }
                } 
            }
            break;

        case IxAdslAturCurrOutputPwr_TRO:
            if(ifIndex == IF_ADSL)
            {
	        if(dataBuffer->dataLength != sizeof(INT32))
	        {
		    retVal = IX_FAIL;
	        }
                else
                {
		    if(AdslPhyType!=IX_ADSL_CPE)
		    {
			return IX_MGRINF_NOTSUPPORTED;
		    }
                    *(INT32*)dataBuffer->dataPtr = ixAdslNearOperData[lineNum].outputPowerDnstr;
		    retVal = IX_SUCCESS;
                } 
            }
            break;

        case IxAdslAturCurrAttainableRate_TRO:
            if(ifIndex == IF_ADSL)
            {
 	        if(dataBuffer->dataLength != sizeof(UINT32))
	        {
	    	    retVal = IX_FAIL;
	        }
                else
                {
		    if(AdslPhyType!=IX_ADSL_CPE)
		    {
			return IX_MGRINF_NOTSUPPORTED;
		    }
                    *(UINT32*)dataBuffer->dataPtr = (ixAdslNearOperData[lineNum].attainableBitrate)*1000;
		    retVal = IX_SUCCESS;
                } 
            }
            break;

	    /*
	     * Atur Chan Table
	     */
	case IxAdslAturChanInterleaveDelay_TRO:
	    if(ifIndex == IF_ADSL_ITLV)
	    {
		if(dataBuffer->dataLength != sizeof(UINT32))
		{
		    retVal = IX_FAIL;
		}
		else
		{
		    *(UINT32*)dataBuffer->dataPtr = ixAdslAturChanTable[lineNum].IxAdslAturChanInterleaveDelay;
		    retVal = IX_SUCCESS;
		}
	    }
	    break;

	case IxAdslAturChanCurrTxRate_TRO:
	case IxAdslAturChanPrevTxRate_TRO:
	    /*
	     * The transmit rate will not change while the line is up
	     * therefore the Curr and Prev Tx rates will always be the 
	     * same.
	     */
	    if(dataBuffer->dataLength != sizeof(UINT32))
	    {
		retVal = IX_FAIL;
	    }
	    else
	    {
		if(AdslPhyType!=IX_ADSL_CPE)
		{
		    retVal = IX_MGRINF_NOTSUPPORTED;
		}
		else
		{
		    if(ADSL_get_ChannelOperData(lineNum,&ixAdslAturChanTable[lineNum].channelOperData) == C_CONF_DATA_ACK)
		    {
			if(ifIndex == IF_ADSL_ITLV)
			{
			    *(UINT32*)dataBuffer->dataPtr = 
				ixAdslAturChanTable[lineNum].channelOperData.actualBitrateNearInt;
			    retVal = IX_SUCCESS;
			}
			
			if(ifIndex == IF_ADSL_FAST)
			{
			    *(UINT32*)dataBuffer->dataPtr = 
				ixAdslAturChanTable[lineNum].channelOperData.actualBitrateNearFast;
			    retVal = IX_SUCCESS;
			}
		    }
		}
	    }
	    break;

	case IxAdslAturChanCrcBlockLength_TRO:
	    if((ifIndex == IF_ADSL_ITLV) || (ifIndex == IF_ADSL_FAST))
	    {
		if(dataBuffer->dataLength != sizeof(UINT32))
	        {
	    	    retVal = IX_FAIL;
	        }
		else
		{
		    *(UINT32*)dataBuffer->dataPtr = ixAdslAturChanTable[lineNum].IxAdslAturChanCrcBlockLength;
		    retVal = IX_SUCCESS;
		}
	    }
	    else
	    {
		retVal = IX_FAIL;
	    }
	    break;
	    
	    /*
	     * Atur Perf Data Table
	     */
	case IxAdslAturPerfLofs_TRO:
	    if(ifIndex == IF_ADSL)
	    {
		if(dataBuffer->dataLength != sizeof(UINT32))
		{
		    retVal = IX_FAIL;
		}
		else
		{
		    if(AdslPhyType!=IX_ADSL_CPE)
		    {
			retVal = IX_MGRINF_NOTSUPPORTED;
		    }
		    else
		    {
			*(UINT32*)dataBuffer->dataPtr = ixAdslAturPerfDataEntry[lineNum].IxadslAturPerfLofs;
			retVal = IX_SUCCESS;
		    }
		}
	    }
	    break;

	case IxAdslAturPerfLoss_TRO:
	    if(ifIndex == IF_ADSL)
	    {
		if(dataBuffer->dataLength != sizeof(UINT32))
		{
		    retVal = IX_FAIL;
		}
		else
		{
		    if(AdslPhyType!=IX_ADSL_CPE)
		    {
			retVal = IX_MGRINF_NOTSUPPORTED;
		    }
		    else
		    {
			*(UINT32*)dataBuffer->dataPtr = ixAdslAturPerfDataEntry[lineNum].IxadslAturPerfLoss;
			retVal = IX_SUCCESS;
		    }
		}
	    }
	    break;

	case IxAdslAturPerfLprs_TRO:
	    if(ifIndex == IF_ADSL)
	    {
		if(dataBuffer->dataLength != sizeof(UINT32))
		{
		    retVal = IX_FAIL;
		}
		else
		{
		    if(AdslPhyType!=IX_ADSL_CPE)
		    {
			retVal = IX_MGRINF_NOTSUPPORTED;
		    }
		    else
		    {
			*(UINT32*)dataBuffer->dataPtr = ixAdslAturPerfDataEntry[lineNum].IxadslAturPerfLprs;
			retVal = IX_SUCCESS;
		    }
		}
	    }
	    break;

	case IxAdslAturPerfESs_TRO:
	    if(ifIndex == IF_ADSL)
	    {
		if(dataBuffer->dataLength != sizeof(UINT32))
		{
		    retVal = IX_FAIL;
		}
		else
		{
		    if(AdslPhyType!=IX_ADSL_CPE)
		    {
			retVal = IX_MGRINF_NOTSUPPORTED;
		    }
		    else
		    {
			*(UINT32*)dataBuffer->dataPtr = ixAdslAturPerfDataEntry[lineNum].IxadslAturPerfESs;
			retVal = IX_SUCCESS;
		    }
		}
	    }
	    break;

	    /*
	     * Atur Chan Perf Data Table Support
	     */
	case IxAdslAturChanReceivedBlks_TRO:
	case IxAdslAturChanTransmittedBlks_TRO:
	case IxAdslAturChanCorrectedBlks_TRO:
	case IxAdslAturChanUncorrectBlks_TRO:
	    retVal = IX_MGRINF_NOTSUPPORTED;
	    break;

	    /*
	     * Begin IF entries: at present there is only limited support 
	     * in the interfaces table for channels. There is full support 
	     * for the Physical
	     */
        case IxAdslIfDescr_TRO:
	    if((retVal =  ixAdslIfIndexAdjust(&ifIndex)) == IX_SUCCESS)
	    {
		if(dataBuffer->dataLength < (strlen((char*)ixIfEntryData[lineNum][ifIndex].IxAdslIfDescr)))
		{
		    retVal = IX_FAIL;
		}
		else
		{
		    strcpy((char*)dataBuffer->dataPtr, (char*)ixIfEntryData[lineNum][ifIndex].IxAdslIfDescr);
		    dataBuffer->dataLength = strlen((char*)dataBuffer->dataPtr);
		} 
	    }
            break;
        case IxAdslIfType_TRO:
	    if((retVal = ixAdslIfIndexAdjust(&ifIndex)) == IX_SUCCESS)
	    {
		if(dataBuffer->dataLength != sizeof(INT32))
		{
		    retVal = IX_FAIL;
		}
		else
		{
		    *(INT32*)dataBuffer->dataPtr = ixIfEntryData[lineNum][ifIndex].IxAdslIfType;
		} 
	    }
            break;
        case IxAdslIfMtu_TRO:
	    if((retVal = ixAdslIfIndexAdjust(&ifIndex)) == IX_SUCCESS)
	    {
	        if(dataBuffer->dataLength != sizeof(INT32))
	        {
		    retVal = IX_FAIL;
	        }
                else
                {
                    *(INT32*)dataBuffer->dataPtr = ixIfEntryData[lineNum][ifIndex].IxAdslIfMtu;
                } 
            }
            break;
        case IxAdslIfSpeed_TRO:
            if((retVal = ixAdslIfIndexAdjust(&ifIndex)) == IX_SUCCESS)
	    {
 	        if(dataBuffer->dataLength != sizeof(UINT32))
	        {
	    	    retVal = IX_FAIL;
	        }
                else
                {
		    if(ifIndex == ADSL_PHYS_INDEX)
		    {
			if(AdslPhyType==IX_ADSL_CO)
			{
			    ixIfEntryData[lineNum][ifIndex].IxAdslIfSpeed =
				(ixAdslFarOperData[lineNum].fastBitRate+ixAdslFarOperData[lineNum].interleavedBitRate)*1000;
			} 
			else 
			{
			    ixIfEntryData[lineNum][ifIndex].IxAdslIfSpeed =
				(ixAdslNearOperData[lineNum].fastBitRate+ixAdslNearOperData[lineNum].interleavedBitRate)*1000;
			}  
		    } 
		    else
		    {
			if(AdslPhyType==IX_ADSL_CO)
			{
			    if(ifIndex == ADSL_ITLV_INDEX)
			    {
				ixIfEntryData[lineNum][ifIndex].IxAdslIfSpeed = 
				    ixAdslChannelOperData[lineNum].actualBitrateFarInt*1000;
			    }
			    else
			    {
				ixIfEntryData[lineNum][ifIndex].IxAdslIfSpeed = 
				    ixAdslChannelOperData[lineNum].actualBitrateFarFast*1000;
			    }
			}
			else
			{
			    if(ifIndex == ADSL_ITLV_INDEX)
			    {
				ixIfEntryData[lineNum][ifIndex].IxAdslIfSpeed = 
				    ixAdslChannelOperData[lineNum].actualBitrateNearInt*1000;
			    }
			    else
			    {
				ixIfEntryData[lineNum][ifIndex].IxAdslIfSpeed = 
				    ixAdslChannelOperData[lineNum].actualBitrateNearFast*1000;
			    }
			}
		    }
		  
                    *(UINT32*)dataBuffer->dataPtr = ixIfEntryData[lineNum][ifIndex].IxAdslIfSpeed;
                } 
            }
            break;
        case IxAdslIfPhysAddress_TRO:
            if((retVal = ixAdslIfIndexAdjust(&ifIndex)) == IX_SUCCESS)
	    {
		/* 
		 * According to RFC 2662 this object should be an OCTET String of
		 * zero length
		 */
		dataBuffer->dataLength = 0; 
            }
            break;
        case IxAdslIfAdminStatus_TRW:
            if((retVal = ixAdslIfIndexAdjust(&ifIndex)) == IX_SUCCESS)
            {
	        if(dataBuffer->dataLength != sizeof(INT32))
	        {
		    retVal = IX_FAIL;
	        }
                else
                {
		    ixAdslAdminInit(lineNum);
                    *(INT32*)dataBuffer->dataPtr = ixIfEntryData[lineNum][ifIndex].IxAdslIfAdminStatus;
                } 
            }
            break;
        case IxAdslIfOperStatus_TRO:
            if((retVal = ixAdslIfIndexAdjust(&ifIndex)) == IX_SUCCESS)
            {
	        if(dataBuffer->dataLength != sizeof(UINT32))
	        {
		    retVal = IX_FAIL;
	        }
                else
                {
		    *(INT32*)dataBuffer->dataPtr = (INT32)ixAdslCheckOperStatus(lineNum, ifIndex);
		} 
            }
            break;
        case IxAdslIfLastChange_TRO:
            if((retVal = ixAdslIfIndexAdjust(&ifIndex)) == IX_SUCCESS)
            {
 	        if(dataBuffer->dataLength != sizeof(UINT32))
	        {
	    	    retVal = IX_FAIL;
	        }
                else
                {
                    *(UINT32*)dataBuffer->dataPtr = ixIfEntryData[lineNum][ifIndex].IxAdslIfLastChange;
		} 
            }
            break;
        case IxAdslIfInOctets_TRO:
            if((retVal = ixAdslIfIndexAdjust(&ifIndex)) == IX_SUCCESS)
            {
 	        if(dataBuffer->dataLength != sizeof(UINT32))
	        {
	    	    retVal = IX_FAIL;
	        }
                else
                {
                    *(UINT32*)dataBuffer->dataPtr = ixIfEntryData[lineNum][ifIndex].IxAdslIfInOctets;
		} 
            }
            break;
        case IxAdslIfInUcastPkts_TRO:
            if((retVal = ixAdslIfIndexAdjust(&ifIndex)) == IX_SUCCESS)
            {
 	        if(dataBuffer->dataLength != sizeof(UINT32))
	        {
	    	    retVal = IX_FAIL;
	        }
                else
                {
                    *(UINT32*)dataBuffer->dataPtr = ixIfEntryData[lineNum][ifIndex].IxAdslIfInUcastPkts;
		} 
            }
            break;
        case IxAdslIfInDiscards_TRO:
            if((retVal = ixAdslIfIndexAdjust(&ifIndex)) == IX_SUCCESS)
            {
 	        if(dataBuffer->dataLength != sizeof(UINT32))
	        {
	    	    retVal = IX_FAIL;
	        }
                else
                {
                    *(UINT32*)dataBuffer->dataPtr = ixIfEntryData[lineNum][ifIndex].IxAdslIfInDiscards;
		} 
            }
            break;
        case IxAdslIfInErrors_TRO:
            if((retVal = ixAdslIfIndexAdjust(&ifIndex)) == IX_SUCCESS)
            {
 	        if(dataBuffer->dataLength != sizeof(UINT32))
	        {
	    	    retVal = IX_FAIL;
	        }
                else
                {
                    *(UINT32*)dataBuffer->dataPtr = ixIfEntryData[lineNum][ifIndex].IxAdslIfInErrors;
		} 
            }
            break;
        case IxAdslIfInUnknownProtos_TRO:
            if((retVal = ixAdslIfIndexAdjust(&ifIndex)) == IX_SUCCESS)
            {
 	        if(dataBuffer->dataLength != sizeof(UINT32))
	        {
	    	    retVal = IX_FAIL;
	        }
                else
                {
                    *(UINT32*)dataBuffer->dataPtr = ixIfEntryData[lineNum][ifIndex].IxAdslIfInUnknownProtos;
		} 
            }
            break;
        case IxAdslIfOutOctets_TRO:
            if((retVal = ixAdslIfIndexAdjust(&ifIndex)) == IX_SUCCESS)
            {
 	        if(dataBuffer->dataLength != sizeof(UINT32))
	        {
	    	    retVal = IX_FAIL;
	        }
                else
                {
                    *(UINT32*)dataBuffer->dataPtr = ixIfEntryData[lineNum][ifIndex].IxAdslIfOutOctets;
		} 
            }
            break;
        case IxAdslIfOutUcastPkts_TRO:
            if((retVal = ixAdslIfIndexAdjust(&ifIndex)) == IX_SUCCESS)
            {
 	        if(dataBuffer->dataLength != sizeof(UINT32))
	        {
	    	    retVal = IX_FAIL;
	        }
                else
                {
                    *(UINT32*)dataBuffer->dataPtr = ixIfEntryData[lineNum][ifIndex].IxAdslIfOutUcastPkts;
		} 
            }
            break;
        case IxAdslIfOutDiscards_TRO:
            if((retVal = ixAdslIfIndexAdjust(&ifIndex)) == IX_SUCCESS)
            {
 	        if(dataBuffer->dataLength != sizeof(UINT32))
	        {
	    	    retVal = IX_FAIL;
	        }
                else
                {
                    *(UINT32*)dataBuffer->dataPtr = ixIfEntryData[lineNum][ifIndex].IxAdslIfOutDiscards;
		} 
            }
            break;
        case IxAdslIfOutErrors_TRO:
            if((retVal = ixAdslIfIndexAdjust(&ifIndex)) == IX_SUCCESS)
            {
 	        if(dataBuffer->dataLength != sizeof(UINT32))
	        {
	    	    retVal = IX_FAIL;
	        }
                else
                {
                    *(UINT32*)dataBuffer->dataPtr = ixIfEntryData[lineNum][ifIndex].IxAdslIfOutErrors;
		} 
            }
            break;
        case IxAdslIfName_TRO:
            if((retVal = ixAdslIfIndexAdjust(&ifIndex)) == IX_SUCCESS)
            {
	        if(dataBuffer->dataLength < strlen((char*)ixIfEntryData[lineNum][ifIndex].IxAdslIfName))
	        {
		    retVal = IX_FAIL;
	        }
                else
                {
                    strcpy(dataBuffer->dataPtr,(char*)ixIfEntryData[lineNum][ifIndex].IxAdslIfName);
                    dataBuffer->dataLength = strlen((char*)ixIfEntryData[lineNum][ifIndex].IxAdslIfName);
		} 
            }
            break;
        case IxAdslIfInMulticastPkts_TRO:
            if((retVal = ixAdslIfIndexAdjust(&ifIndex)) == IX_SUCCESS)
            {
 	        if(dataBuffer->dataLength != sizeof(UINT32))
	        {
	    	    retVal = IX_FAIL;
	        }
                else
                {
                    *(UINT32*)dataBuffer->dataPtr = ixIfEntryData[lineNum][ifIndex].IxAdslIfInMulticastPkts;
		} 
            }
            break;
        case IxAdslIfInBroadcastPkts_TRO:
            if((retVal = ixAdslIfIndexAdjust(&ifIndex)) == IX_SUCCESS)
            {
 	        if(dataBuffer->dataLength != sizeof(UINT32))
	        {
	    	    retVal = IX_FAIL;
	        }
                else
                {
                    *(UINT32*)dataBuffer->dataPtr = ixIfEntryData[lineNum][ifIndex].IxAdslIfInBroadcastPkts;
		} 
            }
            break;
        case IxAdslIfOutMulticastPkts_TRO:
            if((retVal = ixAdslIfIndexAdjust(&ifIndex)) == IX_SUCCESS)
            {
 	        if(dataBuffer->dataLength != sizeof(UINT32))
	        {
	    	    retVal = IX_FAIL;
	        }
                else
                {
                    *(UINT32*)dataBuffer->dataPtr = ixIfEntryData[lineNum][ifIndex].IxAdslIfOutMulticastPkts;
		} 
            }
            break;
        case IxAdslIfOutBroadcastPkts_TRO:
            if((retVal = ixAdslIfIndexAdjust(&ifIndex)) == IX_SUCCESS)
            {
 	        if(dataBuffer->dataLength != sizeof(UINT32))
	        {
	    	    retVal = IX_FAIL;
	        }
                else
                {
                    *(UINT32*)dataBuffer->dataPtr = ixIfEntryData[lineNum][ifIndex].IxAdslIfOutBroadcastPkts;
		} 
            }
            break;
        case IxAdslIfHighSpeed_TRO:
            if((retVal = ixAdslIfIndexAdjust(&ifIndex)) == IX_SUCCESS)
            {
 	        if(dataBuffer->dataLength != sizeof(UINT32))
	        {
	    	    retVal = IX_FAIL;
	        }
                else
		{
		    if(ifIndex == ADSL_PHYS_INDEX)
		    {
			if(AdslPhyType==IX_ADSL_CO)
			{
			    ixIfEntryData[lineNum][ifIndex].IxAdslIfHighSpeed =
				(ixAdslFarOperData[lineNum].fastBitRate+ixAdslFarOperData[lineNum].interleavedBitRate) / 1000;
			} 
			else 
			{
			    ixIfEntryData[lineNum][ifIndex].IxAdslIfHighSpeed =
				(ixAdslNearOperData[lineNum].fastBitRate+ixAdslNearOperData[lineNum].interleavedBitRate) / 1000;
			}  
		    } 
		    else
		    {
			if(AdslPhyType==IX_ADSL_CO)
			{
			    if(ifIndex == ADSL_ITLV_INDEX)
			    {
				ixIfEntryData[lineNum][ifIndex].IxAdslIfHighSpeed = 
				    ixAdslChannelOperData[lineNum].actualBitrateFarInt / 1000;
			    }
			    else
			    {
				ixIfEntryData[lineNum][ifIndex].IxAdslIfHighSpeed = 
				    ixAdslChannelOperData[lineNum].actualBitrateFarFast / 1000;
			    }
			}
			else
			{
			    if(ifIndex == ADSL_ITLV_INDEX)
			    {
				ixIfEntryData[lineNum][ifIndex].IxAdslIfHighSpeed = 
				    ixAdslChannelOperData[lineNum].actualBitrateNearInt / 1000;
			    }
			    else
			    {
				ixIfEntryData[lineNum][ifIndex].IxAdslIfHighSpeed = 
				    ixAdslChannelOperData[lineNum].actualBitrateNearFast / 1000;
			    }
			}
		    }
		  
                    *(UINT32*)dataBuffer->dataPtr = ixIfEntryData[lineNum][ifIndex].IxAdslIfHighSpeed;
		} 
            }
            break;
        case IxAdslIfPromiscuousMode_TRW:
            if((retVal = ixAdslIfIndexAdjust(&ifIndex)) == IX_SUCCESS)
            {
	        if(dataBuffer->dataLength != sizeof(INT32))
	        {
		    retVal = IX_FAIL;
	        }
                else
                {
                    *(INT32*)dataBuffer->dataPtr = ixIfEntryData[lineNum][ifIndex].IxAdslIfPromiscuousMode;
		} 
            }
            break;
        case IxAdslIfConnectorPresent_TRO:
            if((retVal = ixAdslIfIndexAdjust(&ifIndex)) == IX_SUCCESS)
            {
	        if(dataBuffer->dataLength != sizeof(INT32))
	        {
		    retVal = IX_FAIL;
	        }
                else
                {
                    *(INT32*)dataBuffer->dataPtr = ixIfEntryData[lineNum][ifIndex].IxAdslIfConnectorPresent;
		} 
            }
            break;
        case IxAdslIfCounterDiscontinuityTime_TRO:
            if((retVal = ixAdslIfIndexAdjust(&ifIndex)) == IX_SUCCESS)
            {
 	        if(dataBuffer->dataLength != sizeof(UINT32))
	        {
	    	    retVal = IX_FAIL;
	        }
                else
                {
                    *(UINT32*)dataBuffer->dataPtr = ixIfEntryData[lineNum][ifIndex].IxAdslIfCounterDiscontinuityTime;
		} 
            }
            break;
        case IxAdslIfStackLowerLevel_TRO:
            if((retVal = ixAdslIfIndexAdjust(&ifIndex)) == IX_SUCCESS)
            {
 	        if(dataBuffer->dataLength != sizeof(UINT32))
	        {
	    	    retVal = IX_FAIL;
	        }
                else
                {
                    *(UINT32*)dataBuffer->dataPtr = 
			ixIfEntryData[lineNum][ifIndex].IxAdslIfStackLowerLevel;
		} 
            }
            break;
	default:
	    return IX_MGRINF_NOTSUPPORTED;
    }
    return retVal;
}

INT32
ixAdslDriverInfoSet(unsigned int lineNum,
		    const IxAdslManagerInfo entity,
		    UINT32 ifIndex,
		    ManagerInfoData *dataBuffer)
{
    INT32 retVal = IX_FAIL;
    ixAdslState_t ixAdslState;

    if(dataBuffer == NULL)
    {
        return IX_FAIL;
    }

    if(dataBuffer->dataPtr == NULL)
    {
	return IX_FAIL;
    }
    
    switch(entity)
    {
        case IxAdslLineCoding_TRO:
        case IxAdslLineType_TRO:
	case IxAdslAtucInvSerialNumber_TRO:
	case IxAdslAtucInvVendorID_TRO:
	case IxAdslAtucInvVersionNumber_TRO:
        case IxAdslAtucCurrSnrMgn_TRO:
        case IxAdslAtucCurrAtn_TRO:
        case IxAdslAtucCurrStatus_TRO:
        case IxAdslAtucCurrOutputPwr_TRO:
        case IxAdslAtucCurrAttainableRate_TRO:
	case IxAdslAturInvSerialNumber_TRO:
	case IxAdslAturInvVendorID_TRO:
	case IxAdslAturInvVersionNumber_TRO:
	case IxAdslAturCurrSnrMgn_TRO:
	case IxAdslAturCurrAtn_TRO:
	case IxAdslAturCurrStatus_TRO:
	case IxAdslAturCurrOutputPwr_TRO:
	case IxAdslAturCurrAttainableRate_TRO:
	case IxAdslAturChanInterleaveDelay_TRO:
	case IxAdslAturChanCurrTxRate_TRO:
	case IxAdslAturChanPrevTxRate_TRO:
	case IxAdslAturChanCrcBlockLength_TRO:
	case IxAdslAturPerfLofs_TRO:
	case IxAdslAturPerfLoss_TRO:
	case IxAdslAturPerfLprs_TRO:
	case IxAdslAturPerfESs_TRO:
	case IxAdslAturChanReceivedBlks_TRO:
	case IxAdslAturChanTransmittedBlks_TRO:
	case IxAdslAturChanCorrectedBlks_TRO:
	case IxAdslAturChanUncorrectBlks_TRO:
        case IxAdslIfDescr_TRO:
        case IxAdslIfType_TRO:
        case IxAdslIfMtu_TRO:
        case IxAdslIfSpeed_TRO:
        case IxAdslIfPhysAddress_TRO:
	    return IX_FAIL;
            break;
        case IxAdslIfAdminStatus_TRW:
            if((retVal = ixAdslIfIndexAdjust(&ifIndex)) == IX_SUCCESS)
            {
	        if(dataBuffer->dataLength != sizeof(INT32))
	        {
		    return IX_FAIL;
	        }
                
		if((*(INT32*)dataBuffer->dataPtr != IX_MGRINF_OPER_DOWN) &&
		    (*(INT32*)dataBuffer->dataPtr != IX_MGRINF_OPER_UP))
		{
		    return IX_FAIL;
		}

		if(ixAdslPhyTypeGet(lineNum) == IX_ADSL_CPE)
                {
		    ixAdslAdminInit(lineNum);

		    if((ixIfEntryData[lineNum][ifIndex].IxAdslIfAdminStatus != *(INT32*)dataBuffer->dataPtr))
		    {
			if((ifIndex ==  ADSL_ITLV_INDEX) ||
			   (ifIndex == ADSL_FAST_INDEX))
			{
			    ixIfEntryData[lineNum][ifIndex].IxAdslIfAdminStatus = *(INT32*)dataBuffer->dataPtr;
			}
			else
			{
			    
			    if((ixIfEntryData[lineNum][ADSL_PHYS_INDEX].IxAdslIfAdminStatus = 
				*(INT32*)dataBuffer->dataPtr) == IX_MGRINF_OPER_DOWN)
			    {
				if(ixAdslCheckOperStatus(lineNum, ADSL_PHYS_INDEX) == IX_MGRINF_OPER_UP)
				{
				    if(IxADSL_CPE_LineClose(lineNum) != IX_ADSL_SUCCESS)
				    {			      
				    }
				}    
			    }

			    if(ixIfEntryData[lineNum][ADSL_PHYS_INDEX].IxAdslIfAdminStatus == IX_MGRINF_OPER_UP)
			    {
				if(ixAdslCheckOperStatus(lineNum, ADSL_PHYS_INDEX) == IX_MGRINF_OPER_DOWN)
				{
				    ixAdslState = ixAdslStateGet(lineNum);
				    if(IxADSL_CPE_OpenInternal(lineNum, ixAdslState.lineType) != IX_ADSL_SUCCESS)
				    {
				    }
				}    
			    }
			}
		    }
		    else
		    {
			return IX_SUCCESS;
		    }
		}
	    }
            break;
        case IxAdslIfOperStatus_TRO:
        case IxAdslIfLastChange_TRO:
        case IxAdslIfInOctets_TRO:
        case IxAdslIfInUcastPkts_TRO:
        case IxAdslIfInDiscards_TRO:
        case IxAdslIfInErrors_TRO:
        case IxAdslIfInUnknownProtos_TRO:
        case IxAdslIfOutOctets_TRO:
        case IxAdslIfOutUcastPkts_TRO:
        case IxAdslIfOutDiscards_TRO:
        case IxAdslIfOutErrors_TRO:
        case IxAdslIfName_TRO:
        case IxAdslIfInMulticastPkts_TRO:
        case IxAdslIfInBroadcastPkts_TRO:
        case IxAdslIfOutMulticastPkts_TRO:
        case IxAdslIfOutBroadcastPkts_TRO:
        case IxAdslIfHighSpeed_TRO:
	    return IX_FAIL;
            break;
        case IxAdslIfPromiscuousMode_TRW:
	    /*
	     * This value should always be true(1) as the DSL interface accepts
	     * all frames transmitted on the physical line
	     */ 
            retVal = IX_FAIL;
            break;
        case IxAdslIfConnectorPresent_TRO:
        case IxAdslIfCounterDiscontinuityTime_TRO:
        case IxAdslIfStackLowerLevel_TRO:
	    return IX_FAIL;
            break;
	default:
	    return IX_MGRINF_NOTSUPPORTED;
    }
    return retVal;
}

INT32
ixAdslDriverTrapBind(IxAdslDriverTrapCallbackPtr_t funcPtr)
{
     IxAdslDriverTrapCallbackPtr = funcPtr;
     return IX_SUCCESS;
}

PRIVATE void
ixAdslNotificationCallback (unsigned int lineNum, t_ADSL_LINE_STATE lineState)
{
    UINT32 timeNow;
    unsigned int index = 0;
    UINT32 ticks;
    int    clk_rate;

    ticks = tickGet();
    clk_rate = sysClkRateGet();
    
    /* 
     * The Link state has changed, therefore the ifLastChange 
     * object must be updated for each interface.
     */
    timeNow = ticks / clk_rate;
    
    for( ; index <= ADSL_FAST_INDEX; index++)
    {
	ixIfEntryData[lineNum][index].IxAdslIfLastChange = timeNow;
    }

    /* 
     * Link Up/Down traps will only ever occur on the physical
     * line. It is recommended in RFC 2233 that Link Up/Down traps
     * only be used at a physical layer ifEntry
     */
    if(IxAdslDriverTrapCallbackPtr != NULL)
    {
	if(lineState == IX_ADSL_LINE_DOWN)
	{
	    IxAdslDriverTrapCallbackPtr(IxAdslLinkDown,IF_ADSL);	
	}
	if(lineState == IX_ADSL_LINE_UP)
	{
	    IxAdslDriverTrapCallbackPtr(IxAdslLinkUp,IF_ADSL);
	}
    }
	
}


PRIVATE void 
ixAdslCheckESs(unsigned int lineNum)
{
    int rc;
    T_adsl_def_counter_set ixPerf_counter;
    unsigned int ixCRCCurrStatus=0;
    unsigned int ixFECCurrStatus=0;
    unsigned int ixHECCurrStatus=0;
    unsigned static int ixCRCLastStatus=0; /* Latest number of CRC errors */
    unsigned static int ixHECLastStatus=0;
    unsigned static int ixFECLastStatus=0;


    rc = ADSL_def_counter_poll(lineNum, &ixPerf_counter);
    if ( rc == C_CONF_DATA_ACK )
    {
	/* We need to check both Interleaved and NonInterleaved */
	ixCRCCurrStatus=ixPerf_counter.near_end.CrcNotInterleaved;
	ixCRCCurrStatus+=ixPerf_counter.near_end.CrcInterleaved;

	if(ixAdslLofs[lineNum] || ixAdslLcds[lineNum] || ixAdslLoss[lineNum] || ixAdslLoms[lineNum] || ixAdslLprs[lineNum] || (ixCRCCurrStatus!=ixCRCLastStatus))
	{
	    ixAdslAturPerfDataEntry[lineNum].IxadslAturPerfESs++;
	}
	ixCRCLastStatus=ixCRCCurrStatus;
        ixFECCurrStatus=ixPerf_counter.near_end.FecNotInterleaved;
        ixFECCurrStatus+=ixPerf_counter.near_end.FecInterleaved;
        
        if(ixFECCurrStatus!=ixFECLastStatus)
        {
            ixAdslAturPerfDataEntry[lineNum].IxadslAturPerfFEC++;
        }
        ixFECLastStatus=ixFECCurrStatus;
        
        ixHECCurrStatus=ixPerf_counter.near_end.HecNotInterleaved;
        ixHECCurrStatus+=ixPerf_counter.near_end.HecInterleaved;
        
        if(ixHECCurrStatus!=ixHECLastStatus)
        {
            ixAdslAturPerfDataEntry[lineNum].IxadslAturPerfHEC++;
        }
        ixHECLastStatus=ixHECCurrStatus;
    }
}

PRIVATE void 
ixAdslPerfMonitor(void *ln)
{
    unsigned int rc;
    unsigned char ixStatus;
    unsigned lineNum = (unsigned)ln;

    rc = ADSL_def_bitmap_poll(lineNum, ixAdslAturPhysEntryData[lineNum].alcatelActualBitmap);

    if( rc == C_CONF_DATA_ACK )	/* The data is good */
    {
	ixStatus=ixAdslAturPhysEntryData[lineNum].alcatelActualBitmap->near_end.status;

	if(ixStatus&CURR_STATUS_LOSS_FRAME)
	{
	    ixAdslLofCounter[lineNum]++;
	    /* If TRUE then has been active in previous sec */
	    if(ixAdslLofs[lineNum]==TRUE && ixAdslLofCounter[lineNum]>=IX_ADSL_FAILURE_SECS)
	    {
		ixAdslAturPerfDataEntry[lineNum].IxadslAturPerfLoff=TRUE;
	    }
	    ixAdslAturPerfDataEntry[lineNum].IxadslAturPerfLofs++;
	    ixAdslLofs[lineNum]=TRUE;
	}
	else
	{
	    ixAdslLofCounter[lineNum]=0;
	    ixAdslLofs[lineNum]=FALSE;
	    ixAdslAturPerfDataEntry[lineNum].IxadslAturPerfLoff=FALSE;
	}

	if(ixStatus&CURR_STATUS_LOSS_CD_INTERLEAVED ||
	   ixStatus&CURR_STATUS_LOSS_CD_NONINTERLEAVED)
	{
	    ixAdslLcdCounter[lineNum]++;
	    if(ixAdslLcds[lineNum]==TRUE && ixAdslLcdCounter[lineNum]>=IX_ADSL_FAILURE_SECS)
	    {
		ixAdslAturPerfDataEntry[lineNum].IxadslAturPerfLcdf=TRUE;
	    }
	    ixAdslAturPerfDataEntry[lineNum].IxadslAturPerfLcds++;
	    ixAdslLcds[lineNum]=TRUE;
	}
	else
	{
	    ixAdslLcdCounter[lineNum]=0;
	    ixAdslLcds[lineNum]=FALSE;
	    ixAdslAturPerfDataEntry[lineNum].IxadslAturPerfLcdf=FALSE;
	}

	if(ixStatus&CURR_STATUS_LOSS_SIGNAL)
	{
	    ixAdslLosCounter[lineNum]++;
	    if(ixAdslLoss[lineNum]==TRUE && ixAdslLosCounter[lineNum]>=IX_ADSL_FAILURE_SECS)
	    {
		ixAdslAturPerfDataEntry[lineNum].IxadslAturPerfLosf=TRUE;
	    }
	    ixAdslAturPerfDataEntry[lineNum].IxadslAturPerfLoss++;
	    ixAdslLoss[lineNum]=TRUE;
	}
	else
	{
	    ixAdslLosCounter[lineNum]=0;
	    ixAdslLoss[lineNum]=FALSE;
	    ixAdslAturPerfDataEntry[lineNum].IxadslAturPerfLosf=FALSE;
	}

	if(ixStatus&CURR_STATUS_LOSS_POWER)
	{
	    ixAdslLprCounter[lineNum]++;
	    if(ixAdslLprs[lineNum]==TRUE && ixAdslLprCounter[lineNum]>=IX_ADSL_FAILURE_SECS)
	    {
		ixAdslAturPerfDataEntry[lineNum].IxadslAturPerfLprf=TRUE;
	    }
	    ixAdslAturPerfDataEntry[lineNum].IxadslAturPerfLprs++;
	    ixAdslLprs[lineNum]=TRUE;
	}
	else
	{
	    ixAdslLprCounter[lineNum]=0;
	    ixAdslLprs[lineNum]=FALSE;
	    ixAdslAturPerfDataEntry[lineNum].IxadslAturPerfLprf=FALSE;
	}

	if(ixStatus&CURR_STATUS_LOSS_MARGIN)
	{
	    ixAdslLomCounter[lineNum]++;
	    if(ixAdslLoms[lineNum]==TRUE && ixAdslLomCounter[lineNum]>=IX_ADSL_FAILURE_SECS)
	    {
		ixAdslAturPerfDataEntry[lineNum].IxadslAturPerfLomf=TRUE;
	    }
	    ixAdslAturPerfDataEntry[lineNum].IxadslAturPerfLoms++;
	    ixAdslLoms[lineNum]=TRUE;
	}
	else
	{
	    ixAdslLomCounter[lineNum]=0;
	    ixAdslLoms[lineNum]=FALSE;
	    ixAdslAturPerfDataEntry[lineNum].IxadslAturPerfLomf=FALSE;
	}

	ixAdslCheckESs(lineNum);
    }
}

void
ixAdslManagementInfoInit(unsigned int lineNum)
{
    UINT8* adslPhy = (UINT8*)"aadsl-line";
    UINT8* adslFast = (UINT8*)"aadsl-fast";
    UINT8* adslIntlv = (UINT8*)"aadsl-intlv";
    int index;
    unsigned ixAdslPerfMonitorId;
    T_adsl_ConfigurationData cpeConfigData;

    char physLine[] = "Physical Line";
    char fastChan[] = "Fast Channel";
    char interLvedChan[] = "Interleaved Channel";

    
    char phyDescr[MAX_STRING_LENGTH];
    char intlvDescr[MAX_STRING_LENGTH];
    char fastDescr[MAX_STRING_LENGTH];

    T_VendorId vendorId;
    char vendorIdString[17];

    int vendorMask = 0x1;

    T_swVersion vendorVersion;
    char vendorVerString[9];

    /* Update controller software version */
    ixAdslSyncWithModem(lineNum);

    sprintf(phyDescr,"Alcatel ADSL %s, S/W version: %d.%d.%d",
	    physLine,
	    ixAdslControllerSwVersion[lineNum].number1,
	    ixAdslControllerSwVersion[lineNum].number2,
	    ixAdslControllerSwVersion[lineNum].number3);

    
    sprintf(intlvDescr,"Alcatel ADSL %s, S/W version: %d.%d.%d",
	    interLvedChan,
	    ixAdslControllerSwVersion[lineNum].number1,
	    ixAdslControllerSwVersion[lineNum].number2,
	    ixAdslControllerSwVersion[lineNum].number3);
    
    
    sprintf(fastDescr,"Alcatel ADSL %s, S/W version: %d.%d.%d",
	    fastChan, 
	    ixAdslControllerSwVersion[lineNum].number1,
	    ixAdslControllerSwVersion[lineNum].number2,
	    ixAdslControllerSwVersion[lineNum].number3);

    /*
     * adslLineTable
     */
    ixAdslLineEntryData[lineNum].IxAdslLineCoding = CODING_DMT_MODE; /* we are always in dmt mode */
    ixAdslLineEntryData[lineNum].IxAdslLineType = FAST_OR_INTERLEAVED;  /* always fast or interleaved */

    /*
     * adslAtucPhysEntryTable
     */
    ixAdslAtucPhysEntryData[lineNum].IxAdslAtucCurrSnrMgn = 0;
    ixAdslAtucPhysEntryData[lineNum].IxAdslAtucCurrAtn = 0;
    ixAdslAtucPhysEntryData[lineNum].IxAdslAtucCurrOutputPwr = 0;
    ixAdslAtucPhysEntryData[lineNum].IxAdslAtucCurrAttainableRate = 0;

    /* 
     * Atur and Atuc Serial Number limited by the code above to SERIALNO_STRING_LEN char's
     */
    strncpy((char*)ixAdslInvSerialNumber,"Alcatel MTC20156 ADSL PHY",
	    (SERIALNO_STRING_LEN - 1));
    ixAdslInvSerialNumber[SERIALNO_STRING_LEN - 1] = '\0';

    /*
     * adslAturPhysEntryTable
     */

    /* The SNMP admin string for vendorId will always be 16bits or 2 bytes */
    ADSL_getVendorId(lineNum, &vendorId);
    for(index =0; index < 16; index++)
    {
	if((vendorId.near_end >> index) & vendorMask)
	{
	    vendorIdString[15 - index] = '1'; 
	}
	else
	{
	    vendorIdString[15 - index] = '0';
	}
    }
    vendorIdString[16] = '\0';
    ixAdslAturPhysEntryData[lineNum].IxAdslAturInvVendorID = 
	(UINT8*)malloc(strlen(vendorIdString)+1);
    strcpy((char*)ixAdslAturPhysEntryData[lineNum].IxAdslAturInvVendorID,vendorIdString);

    /* The SNMP admin string for vendorVersion will always be 16bits or 2 bytes */
    ADSL_getVersion(lineNum, &vendorVersion);
    for(index =0; index < 8; index++)
    {
	if((vendorId.near_end >> index) & vendorMask)
	{
	    vendorVerString[7 - index] = '1'; 
	}
	else
	{
	    vendorVerString[7 - index] = '0'; 
	}
    }
    vendorVerString[8] = '\0'; 
    ixAdslAturPhysEntryData[lineNum].IxAdslAturInvVersionNumber = 
	(UINT8*)malloc(strlen(vendorVerString)+1);
    strcpy((char*)ixAdslAturPhysEntryData[lineNum].IxAdslAturInvVersionNumber,vendorVerString);

    ixAdslAturPhysEntryData[lineNum].IxAdslAturCurrSnrMgn = 1;
    ixAdslAturPhysEntryData[lineNum].IxAdslAturCurrAtn = 2;

    /* Alloc and initialise Alcatel bitmap for defect bitmaps */
    ixAdslAturPhysEntryData[lineNum].alcatelActualBitmap = 
	(T_adsl_def_bitmap_set *)malloc(sizeof(T_adsl_def_bitmap_set));
    ixAdslAturPhysEntryData[lineNum].alcatelActualBitmap->near_end.status = 0;
    ixAdslAturPhysEntryData[lineNum].alcatelActualBitmap->near_end.change = 0;
    ixAdslAturPhysEntryData[lineNum].alcatelActualBitmap->far_end.status = 0;
    ixAdslAturPhysEntryData[lineNum].alcatelActualBitmap->far_end.change = 0;
    
    ixAdslAturPhysEntryData[lineNum].IxAdslAturCurrOutputPwr = 1;
    ixAdslAturPhysEntryData[lineNum].IxAdslAturCurrAttainableRate = 2;
   
    /* 
     * adslAturChanTable
     */
    cpeConfigData = ixAdslAturConfigDataGet(lineNum);
    ixAdslMinNoiseMarginUpstr = cpeConfigData.MinNoiseMarginUpstr;
    ixAdslAturChanTable[lineNum].IxAdslAturChanInterleaveDelay = (UINT32)cpeConfigData.InterlMaxDelayUpstr;
    
    ixAdslAturChanTable[lineNum].IxAdslAturChanCrcBlockLength = 0;
    
    /*
     * adslAturPerfDataTable
     */ 
    ixAdslAturPerfDataEntry[lineNum].IxadslAturPerfLofs=0;
    ixAdslAturPerfDataEntry[lineNum].IxadslAturPerfLoss=0;
    ixAdslAturPerfDataEntry[lineNum].IxadslAturPerfLprs=0;
    ixAdslAturPerfDataEntry[lineNum].IxadslAturPerfESs=0;
    ixAdslAturPerfDataEntry[lineNum].IxadslAturPerfLcds=0;
    ixAdslAturPerfDataEntry[lineNum].IxadslAturPerfLoms=0;
    ixAdslAturPerfDataEntry[lineNum].IxadslAturPerfLoff=FALSE;
    ixAdslAturPerfDataEntry[lineNum].IxadslAturPerfLosf=FALSE;
    ixAdslAturPerfDataEntry[lineNum].IxadslAturPerfLprf=FALSE;
    ixAdslAturPerfDataEntry[lineNum].IxadslAturPerfLcdf=FALSE;
    ixAdslAturPerfDataEntry[lineNum].IxadslAturPerfLomf=FALSE;

    /*
     * adslAturPerfRetrainStats. Stats will be incremented if a 
     * re-train occurs as a result of one of the relevant failure
     * flags being detected.
     */
    ixAdslAturPerfRetrainStats[lineNum].IxadslAturPerfLoffRetrainInitiated=0;
    ixAdslAturPerfRetrainStats[lineNum].IxadslAturPerfLosfRetrainInitiated=0;
    ixAdslAturPerfRetrainStats[lineNum].IxadslAturPerfLprfRetrainInitiated=0;
    ixAdslAturPerfRetrainStats[lineNum].IxadslAturPerfLcdfRetrainInitiated=0;
    ixAdslAturPerfRetrainStats[lineNum].IxadslAturPerfLomfRetrainInitiated=0;
    ixAdslAturPerfRetrainStats[lineNum].IxadslAturPerfSwRetrainInitiated=0;

    /* 
     * ifTable
     */
    ixIfEntryData[lineNum][ADSL_PHYS_INDEX].IxAdslIfDescr = 
	(UINT8*)malloc(strlen(phyDescr)+1);
    strcpy((char*)ixIfEntryData[lineNum][ADSL_PHYS_INDEX].IxAdslIfDescr,phyDescr);

    ixIfEntryData[lineNum][ADSL_ITLV_INDEX].IxAdslIfDescr = 
	(UINT8*)malloc(strlen(intlvDescr)+1);
    strcpy((char*)ixIfEntryData[lineNum][ADSL_ITLV_INDEX].IxAdslIfDescr,intlvDescr);
    
    ixIfEntryData[lineNum][ADSL_FAST_INDEX].IxAdslIfDescr = 
	(UINT8*)malloc(strlen(fastDescr)+1);
    strcpy((char*)ixIfEntryData[lineNum][ADSL_FAST_INDEX].IxAdslIfDescr,fastDescr);

    ixIfEntryData[lineNum][ADSL_PHYS_INDEX].IxAdslIfType = IF_TYPE_ADSL_PHYS; /*adsl physical*/
    ixIfEntryData[lineNum][ADSL_ITLV_INDEX].IxAdslIfType = IF_TYPE_ADSL_INTLV; /*adsl interleaved*/
    ixIfEntryData[lineNum][ADSL_FAST_INDEX].IxAdslIfType = IF_TYPE_ADSL_FAST; /*adsl fast*/

    ixIfEntryData[lineNum][ADSL_PHYS_INDEX].IxAdslIfName = 
	(INT8*)malloc(strlen((char*)adslPhy)+1);
    strcpy((char*)ixIfEntryData[lineNum][ADSL_PHYS_INDEX].IxAdslIfName,(char*)adslPhy);

    ixIfEntryData[lineNum][ADSL_ITLV_INDEX].IxAdslIfName = 
	(INT8*)malloc(strlen((char*)adslIntlv)+1);
    strcpy((char*)ixIfEntryData[lineNum][ADSL_ITLV_INDEX].IxAdslIfName,(char*)adslIntlv);

    ixIfEntryData[lineNum][ADSL_FAST_INDEX].IxAdslIfName = 
	(INT8*)malloc(strlen((char*)adslFast)+1);
    strcpy((char*)ixIfEntryData[lineNum][ADSL_FAST_INDEX].IxAdslIfName,(char*)adslFast);

    ixIfEntryData[lineNum][ADSL_PHYS_INDEX].IxAdslIfConnectorPresent = 1;
    ixIfEntryData[lineNum][ADSL_ITLV_INDEX].IxAdslIfConnectorPresent = 2;
    ixIfEntryData[lineNum][ADSL_FAST_INDEX].IxAdslIfConnectorPresent = 2;

    ixIfEntryData[lineNum][ADSL_PHYS_INDEX].IxAdslIfStackLowerLevel = 0;
    ixIfEntryData[lineNum][ADSL_ITLV_INDEX].IxAdslIfStackLowerLevel = IF_ADSL;
    ixIfEntryData[lineNum][ADSL_FAST_INDEX].IxAdslIfStackLowerLevel = IF_ADSL;

    ixIfEntryData[lineNum][ADSL_PHYS_INDEX].IxAdslIfAdminStatus = IX_MGRINF_ADMIN_UP;
    ixIfEntryData[lineNum][ADSL_ITLV_INDEX].IxAdslIfAdminStatus = IX_MGRINF_ADMIN_DOWN;
    ixIfEntryData[lineNum][ADSL_FAST_INDEX].IxAdslIfAdminStatus = IX_MGRINF_ADMIN_DOWN;


    for(index = 0; index < 3; index++)
    {
	ixIfEntryData[lineNum][index].IxAdslIfMtu = 0;   
	ixIfEntryData[lineNum][index].IxAdslIfSpeed = 0;
	ixIfEntryData[lineNum][index].IxAdslIfLastChange = 0;
	ixIfEntryData[lineNum][index].IxAdslIfInOctets = 0;
	ixIfEntryData[lineNum][index].IxAdslIfInUcastPkts = 0;
	ixIfEntryData[lineNum][index].IxAdslIfInDiscards = 0;
	ixIfEntryData[lineNum][index].IxAdslIfInErrors = 0;
	ixIfEntryData[lineNum][index].IxAdslIfInUnknownProtos = 0;
	ixIfEntryData[lineNum][index].IxAdslIfOutOctets = 0;
	ixIfEntryData[lineNum][index].IxAdslIfOutUcastPkts = 0;
	ixIfEntryData[lineNum][index].IxAdslIfOutDiscards = 0;
	ixIfEntryData[lineNum][index].IxAdslIfOutErrors = 0;
	ixIfEntryData[lineNum][index].IxAdslIfInMulticastPkts = 0;
	ixIfEntryData[lineNum][index].IxAdslIfInBroadcastPkts = 0;
	ixIfEntryData[lineNum][index].IxAdslIfOutMulticastPkts = 0;
	ixIfEntryData[lineNum][index].IxAdslIfOutBroadcastPkts = 0;
	ixIfEntryData[lineNum][index].IxAdslIfHighSpeed = 0;
	ixIfEntryData[lineNum][index].IxAdslIfPromiscuousMode = 1;
	ixIfEntryData[lineNum][index].IxAdslIfCounterDiscontinuityTime = 0;
    }
    /* Register a Link Up/Down callback with the Phy */
    if((IxADSL_LineStateReg(lineNum,ixAdslNotificationCallback)) == IX_FAIL)
    {
    }
    
    /* Timer callback for Performance counters */
    if(ixTimerCtrlScheduleRepeating (ixAdslPerfMonitor,
				     (void *)lineNum,
				     IxTimerCtrlAdslPurpose,
				     IX_ADSL_PERF_TIMER,
				     &ixAdslPerfMonitorId) != IX_SUCCESS)
    {
	ixOsServLog(LOG_ERROR, "Couldn't schedule performance monitor timer", 0, 0, 0, 0, 0, 0);
    }
						   
}

IxadslAturPerfDataEntry* 
ixAdslGetPerfData(unsigned int lineNum)
{
    return(&ixAdslAturPerfDataEntry[lineNum]);
}

IxadslAturPerfRetrainStats*
ixAdslPerfRetrainStatsGet(unsigned int lineNum)
{
    return(&ixAdslAturPerfRetrainStats[lineNum]);
}






