/* 
 * FileName:    IxTskcTimers.c
 * Author: Intel Corporation
 * Created:     25-APR-2000
 * Description: 
 *    This is the implementiation file for task controller timers.
 * 
 * Design Notes:
 *
 * -- Intel Copyright Notice --
 * INTEL CONFIDENTIAL
 *
 * Copyright 2000-2002 Intel Corporation All Rights Reserved.
 *
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.  The Material
 * contains trade secrets and proprietary and confidential information of
 * Intel or its suppliers and licensors.  The Material is protected by
 * worldwide copyright and trade secret laws and treaty provisions. No
 * part of the Material may be used, copied, reproduced, modified,
 * published, uploaded, posted, transmitted, distributed, or disclosed in
 * any way without Intel's prior express written permission.
 *
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel or otherwise.  Any license under
 * such intellectual property rights must be express and approved by
 * Intel in writing.
 *
 * For further details, please see the file README.TXT distributed with
 * this software.
 * -- End Intel Copyright Notice --
 */

/*
 * Put the system defined include files required
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <vxWorks.h>
#include <timers.h>
#include <sysLib.h>
#include <taskLib.h>
#include <limits.h>
#include <semLib.h>
#include "ixAdslUtil.h"

/*
 * Put the user defined include files required
 */

typedef void (*TimerCallback)(void *);

typedef struct
{
    BOOL     inUse;
    unsigned id;
    BOOL isRepeating;
    struct timespec period;
    struct timespec expires;
    TimerCallback callback;
    void *callbackParam;
} TimerRec;

#define MAX_TIMERS 100
#define CRITICAL_SECTION_TIMER 1
#define PRIVATE static
#define TIMER_WARNING_THRESHOLD (MAX_TIMERS - 20)

PRIVATE TimerRec timers[MAX_TIMERS];
PRIVATE SEM_ID timerRecalcSem;
PRIVATE int lastTimerId = 0;
PRIVATE unsigned highestTimerSlotUsed = 0;
PRIVATE BOOL thresholdErrorOccurred = FALSE;

/*
 *  This function allocates a new timer.  It is used by Schedule and 
 *  ScheduleRepeating.  The value returned is the id of the timer which
 *  can be used to cancel timers.  Errors are logged but failure is not
 *  reported to the calling application.
 */

PRIVATE unsigned 
ixTskcNewTimer(TimerCallback func, void *param,
	       UINT32 interval, BOOL isRepeating)
{
    unsigned i;
    unsigned timerId = ++lastTimerId;

    if (interval < (UINT32)(1000 / sysClkRateGet()))
    {
	return timerId;
    }
	
    ixTskcCriticalSectionBegin(CRITICAL_SECTION_TIMER);

    for (i = 0; i < MAX_TIMERS; i++)
    {
        if (!timers[i].inUse)
        {
            break;
        }
    }

    if ((i >= TIMER_WARNING_THRESHOLD) && (thresholdErrorOccurred == FALSE))
    {
        /*
         * This error serves as an early indication that the number of
         * available timer slots will need to be increased. This is done
         * by increasing MAX_TIMERS
         */
        thresholdErrorOccurred = TRUE;
    }

    if (i == MAX_TIMERS)
    {
        /* 
	 *  If you get this error, increase MAX_TIMERS above 
	 */
    }
    else
    {
        timers[i].inUse = TRUE;
        timers[i].period = msToTimespec(interval);

	if (isRepeating)
	{
	    timers[i].isRepeating = TRUE;
	}
	else
	{
	    timers[i].isRepeating = FALSE;
	}

	clock_gettime(CLOCK_REALTIME, &timers[i].expires);
	timers[i].expires = timespecAdd(timers[i].expires, timers[i].period);

        timers[i].callback = func;
	timers[i].callbackParam = param;
	timers[i].id = timerId;

	if ((i+1) > highestTimerSlotUsed)
	    highestTimerSlotUsed = i+1;
    
        semGive(timerRecalcSem);
    }

    ixTskcCriticalSectionEnd(CRITICAL_SECTION_TIMER);

    return timerId;
}


unsigned 
ixTskcScheduleRepeating(TimerCallback func, void *param, UINT32 interval)
{
    return ixTskcNewTimer(func, param, interval, TRUE);
}

struct timespec
msToTimespec(UINT32 ms)
{
    struct timespec retVal;

    retVal.tv_sec = ms / 1000;
    retVal.tv_nsec = (ms % 1000) * 1000000;
    return retVal;
}

struct timespec
timespecAdd(struct timespec a, struct timespec b)
{
    struct timespec retVal;

    retVal.tv_sec = a.tv_sec + b.tv_sec;
    retVal.tv_nsec = a.tv_nsec + b.tv_nsec;
    if(retVal.tv_nsec > 1000000000)
    {
        retVal.tv_sec++;
        retVal.tv_nsec -= 1000000000;
    }
    return retVal;
}

