/**
* @file IxAtmdPortMgmt.c
*
 * @author Intel Corporation
* @date 17 March 2002
*
* @brief ATM Port configuration and management
*
 * -- Intel Copyright Notice --
 *
 * @par
 * INTEL CONFIDENTIAL
 *
 * @par
 * Copyright 2002 Intel Corporation All Rights Reserved.
 *
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.  The Material
 * contains trade secrets and proprietary and confidential information of
 * Intel or its suppliers and licensors.  The Material is protected by
 * worldwide copyright and trade secret laws and treaty provisions. No
 * part of the Material may be used, copied, reproduced, modified,
 * published, uploaded, posted, transmitted, distributed, or disclosed in
 * any way without Intel's prior express written permission.
 *
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel or otherwise.  Any license under
 * such intellectual property rights must be express and approved by
 * Intel in writing.
 *
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * -- End Intel Copyright Notice --
*/

/*
* Sytem defined include files
*/
#include <stdio.h>
#include <string.h>

/*
* User defined include files
*/
#include "IxOsServices.h"
#include "IxQMgr.h"
#include "IxNpeMh.h"

#include "IxAtmdAccCtrl.h"
#include "IxAtmdDefines_p.h"
#include "IxAtmdNpe_p.h"
#include "IxAtmdAssert_p.h"
#include "IxAtmdUtil_p.h"
#include "IxAtmdPortMgmt_p.h"

/*
* #defines
*/
#define NPE_RESP_REQ NPE_RESP_REQ_OFF

#define IX_ATMDACC_PORTMGMT_LOCK_INIT() do{\
    IX_STATUS returnStatus = IX_SUCCESS;\
    IX_ATMDACC_ENSURE (ixAtmdAccPortMgmtInitDone == FALSE, "Initialisation Error");\
    returnStatus = ixOsServMutexInit (&portLock);\
    IX_ATMDACC_ENSURE (returnStatus == IX_SUCCESS, "Initialisation Error");\
    } while(0)

#define IX_ATMDACC_PORTMGMT_LOCK_GET() do{\
    ixOsServMutexLock (&portLock);\
    } while(0)

#define IX_ATMDACC_PORTMGMT_LOCK_RELEASE() do{\
    ixOsServMutexUnlock (&portLock);\
    } while(0)

/* The npeResponse lock will serialise the excution of the functions
* ixAtmdAccUtopiaStatusRead, ixAtmdAccUtopiaConfigRead and
* ixAtmdAccUtopiaStatusRead with the IxNpeMh solicited callbacks.
* Each of these functions will again call IX_ATMDACC_PORTMGMT_LOCK_GET and
* block until the apropriate IxNpeMh solicited callback calls
* IX_ATMDACC_PORTMGMT_LOCK_RELEASE.
*
* NOTE: It would be more desirable if IxNpeMh provided an interface that would
* block until the response message from the NPE was available and return this
* message to the client instead of using the current callback mechanism.
*/
#define IX_ATMDACC_NPE_RESPONSE_LOCK_INIT() do{\
    IX_STATUS returnStatus = IX_SUCCESS;\
    IX_ATMDACC_ENSURE (ixAtmdAccPortMgmtInitDone == FALSE, "Initailisation Error");\
    returnStatus = ixOsServMutexInit (&npeResponseLock);\
    IX_ATMDACC_ENSURE (returnStatus == IX_SUCCESS, "Initailisation Error");\
    IX_ATMDACC_NPE_RESPONSE_LOCK_GET ();\
    } while(0)

#define IX_ATMDACC_NPE_RESPONSE_LOCK_GET() do{\
    ixOsServMutexLock (&npeResponseLock);\
    } while(0)

#define IX_ATMDACC_NPE_RESPONSE_LOCK_RELEASE() do{\
    ixOsServMutexUnlock (&npeResponseLock);\
    } while(0)

/*
* Function prototypes.
*/
PRIVATE void ixAtmdAccDummySetupNotifyHandler (unsigned int numPort);
PRIVATE IX_STATUS ixAtmdAccDummyStateChangeHandler (IxAtmLogicalPort port,
                                                    IxAtmdAccPortState state);
PRIVATE BOOL ixAtmdAccDummyStateQuery (IxAtmLogicalPort port);
PRIVATE IX_STATUS ixAtmdAccPortStateQuery (IxAtmLogicalPort port,
                                           IxAtmdAccPortState state,
                                           BOOL *paramError);
PRIVATE IX_STATUS ixAtmdAccPortStateChange (IxAtmLogicalPort port,
                                            IxAtmdAccPortState newState);

PRIVATE IX_STATUS ixAtmdAccUtopiaConfigWrite (const IxAtmdAccUtopiaConfig *utConfig);
PRIVATE IX_STATUS ixAtmdAccUtopiaConfigLoad (void);
PRIVATE void ixAtmdAccUtopiaConfigLoadCallback (IxNpeMhNpeId npeMhNpeId,
                                                  IxNpeMhMessage npeMhMessage);

PRIVATE void ixAtmdAccUtopiaStatusUploadCallback (IxNpeMhNpeId npeMhNpeId,
                                                  IxNpeMhMessage npeMhMessage);
PRIVATE IX_STATUS ixAtmdAccUtopiaStatusUpload (void);
PRIVATE void ixAtmdAccNpeUtopiaStatusReadCallback (IxNpeMhNpeId npeMhNpeId,
                                                   IxNpeMhMessage npeMhMessage);
PRIVATE IX_STATUS ixAtmdAccUtopiaStatusRead (IxAtmdAccUtopiaStatus* utStatus);

PRIVATE BOOL ixAtmdAccUtopiaConfigSetParamsValidate (const IxAtmdAccUtopiaConfig
                                                     *utConfig);

                                                     /*
                                                     * Variables Private To This File.
*/
static IxMutex portLock;
static IxMutex npeResponseLock;
static BOOL ixAtmdAccPortMgmtInitDone = FALSE;
static IxAtmdAccPortSetupNotifyHandler setupNotify = ixAtmdAccDummySetupNotifyHandler;
static IxAtmdAccPortStateChangeHandler stateChangeRequest = ixAtmdAccDummyStateChangeHandler;
static IxAtmdAccPortStateQuery             isEnabledQuery = ixAtmdAccDummyStateQuery;
static IxAtmdAccPortStateQuery          isDisableComplete = ixAtmdAccDummyStateQuery;
static unsigned int numTxVcQueues; /**< number of Tx Vc queues in the system */
static unsigned int numberOfPortsConfigured; /**< number of ports in the system */
                                             /*
                                             * The following UINT32 variables are defined as volatile because they are accessed
                                             * from multiple threads. Declaring these as volatile indicates to the compiler that
                                             * these variables should not be placed in registers, which would result in incorrect
                                             * operation of the code.
                                            */
static volatile UINT32 npeRespWordRead = 0;
static volatile UINT32 npeRespOffsetRead = 0;
static BOOL utopiaConfigSetDone = FALSE;
static unsigned int portRequestStats[IX_UTOPIA_MAX_PORTS];
/*
* Function implementations
*/

/* -----------------------------------------
* API functions
* ----------------------------------------- */
PUBLIC IX_STATUS
ixAtmdAccPortEnable(IxAtmLogicalPort port)
{
    if (!ixAtmdAccPortMgmtInitDone)
    {
        return IX_FAIL;
    }
    else
    {
        return ixAtmdAccPortStateChange(port, IX_ATMD_PORT_ENABLED);
    }
}

/* ----------------------------------------------
*/
PUBLIC IX_STATUS
ixAtmdAccPortDisable(IxAtmLogicalPort port)
{
    if (!ixAtmdAccPortMgmtInitDone)
    {
        return IX_FAIL;
    }
    else
    {
        return ixAtmdAccPortStateChange(port, IX_ATMD_PORT_DISABLED);
    }
}

/* ----------------------------------------------
*/
PUBLIC BOOL
ixAtmdAccPortDisableComplete(IxAtmLogicalPort port)
{
    if (!ixAtmdAccPortMgmtInitDone)
    {
        return TRUE;
    }
    else
    {
        IX_STATUS retval;
        BOOL paramError;

        retval = ixAtmdAccPortStateQuery(port, IX_ATMD_PORT_DISABLED, &paramError);

        if ((retval != IX_SUCCESS) || (paramError))
        {
            return FALSE;
        }
    }

    return TRUE;
}

/* ----------------------------------------------
*/
PUBLIC IX_STATUS
ixAtmdAccUtopiaConfigSet (const IxAtmdAccUtopiaConfig *utConfig)
{
    IX_STATUS returnStatus = IX_SUCCESS;

    if (!ixAtmdAccPortMgmtInitDone || utopiaConfigSetDone)
    {
        returnStatus = IX_FAIL;
    }

    if (returnStatus == IX_SUCCESS)
    {
        if (!ixAtmdAccUtopiaConfigSetParamsValidate (utConfig))
        {
            returnStatus = IX_FAIL;
        }
    }

    if (returnStatus == IX_SUCCESS)
    {
        IX_ATMDACC_PORTMGMT_LOCK_GET ();

        returnStatus = ixAtmdAccUtopiaConfigWrite (utConfig);
        if (returnStatus == IX_SUCCESS)
        {
            returnStatus = ixAtmdAccUtopiaConfigLoad ();
        }

        if (returnStatus == IX_SUCCESS)
        {
            utopiaConfigSetDone = TRUE;
            numberOfPortsConfigured = utConfig->utTxConfig.txAddrRange + 1;
            setupNotify(numberOfPortsConfigured);
        }

        IX_ATMDACC_PORTMGMT_LOCK_RELEASE ();
    } /* end of if(returnStatus) */

    return returnStatus;
}

/* ----------------------------------------------
*/
PUBLIC IX_STATUS
ixAtmdAccUtopiaStatusGet (IxAtmdAccUtopiaStatus* utStatus)
{
    IX_STATUS returnStatus = IX_SUCCESS;

    if (!ixAtmdAccPortMgmtInitDone
        || utStatus == NULL)
    {
        returnStatus = IX_FAIL;
    }
    else
    {
        IX_ATMDACC_PORTMGMT_LOCK_GET ();
        
        /* Copy config & status registers from the Utopia coprocessor registers
        * to NPE memory
        */
        returnStatus = ixAtmdAccUtopiaStatusUpload ();
        
        if (returnStatus == IX_SUCCESS)
        {
            /* Copy the status table from Npe memory to Xscale memory.
            * NOTE: This is a blocking call
            */
            returnStatus = ixAtmdAccUtopiaStatusRead (utStatus);
        }
        
        IX_ATMDACC_PORTMGMT_LOCK_RELEASE ();
    }

    return returnStatus;
}

/* -----------------------------------------
* Internal functions
* ----------------------------------------- */
IX_STATUS
ixAtmdAccPortMgmtInit (void)
{
    IX_STATUS returnStatus = IX_SUCCESS;

    if (!ixAtmdAccPortMgmtInitDone)
    {
        utopiaConfigSetDone = FALSE;

        /* initialise internal lock */
        IX_ATMDACC_PORTMGMT_LOCK_INIT ();
        IX_ATMDACC_NPE_RESPONSE_LOCK_INIT ();

        ixAtmdAccPortMgmtInitDone = TRUE;
    } /* end of if(ixAtmdAccPortMgmtInitDone) */
    else
    {
        utopiaConfigSetDone = FALSE;
        returnStatus = IX_FAIL;
    } /* end of if-else(ixAtmdAccPortMgmtInitDone) */

    return returnStatus;
}

/* ----------------------------------------------
*/
IX_STATUS
ixAtmdAccPortMgmtNumTxVcQueuesSet (unsigned int numTxQueues)
{
    IX_STATUS returnStatus = IX_SUCCESS;

    /* Check that the number of transmit VC queues is valid */
    if (numTxQueues > IX_UTOPIA_MAX_PORTS)
    {
        returnStatus = IX_FAIL;
    }
    else
    {
        numTxVcQueues = numTxQueues;
    } /* end of if-else(numTxQueues) */

    return returnStatus;
}

/* ----------------------------------------------
*/
IX_STATUS
ixAtmdAccPortMgmtNumTxVcQueuesGet (unsigned int *numTxQueuesPtr)
{
    IX_STATUS returnStatus = IX_SUCCESS;

    if (numTxQueuesPtr == NULL)
    {
        returnStatus = IX_FAIL;
    }
    else
    {
        *numTxQueuesPtr = numTxVcQueues;
    } /* end of if-else(numTxQueuesPtr) */

    return returnStatus;
}

/* ----------------------------------------------
*/
BOOL
ixAtmdAccPortIsEnabled(IxAtmLogicalPort port)
{
    IX_STATUS retval;
    BOOL paramError;

    retval = ixAtmdAccPortStateQuery(port, IX_ATMD_PORT_ENABLED, &paramError);

    if ((retval != IX_SUCCESS) || (paramError))
    {
        return FALSE;
    }

    return TRUE;
}

/* ----------------------------------------------
*/
void
ixAtmdAccPortStateHandlersRegister(
                                   IxAtmdAccPortSetupNotifyHandler setupNotifyHandler,
                                   IxAtmdAccPortStateChangeHandler stateChangeHandler,
                                   IxAtmdAccPortStateQuery         enabledQuery,
                                   IxAtmdAccPortStateQuery         disabledQuery)
{
    setupNotify        = setupNotifyHandler;
    stateChangeRequest = stateChangeHandler;
    isEnabledQuery     = enabledQuery;
    isDisableComplete  = disabledQuery;
}

/* ----------------------------------------------
*/
BOOL
ixAtmdAccPortConfigured (IxAtmLogicalPort port)
{
    /* check if the port valid and is configured */
    BOOL result = (port >= IX_UTOPIA_PORT_0) &&
        (port < IX_UTOPIA_MAX_PORTS) &&
        ((unsigned int)port < numberOfPortsConfigured);

    return result;
}

/* ----------------------------------------------
*/
void
ixAtmdAccPortMgmtStatsShow (void)
{
    int port;

    if (ixAtmdAccPortMgmtInitDone)
    {
        printf("\n");
        for (port = 0; port < IX_UTOPIA_MAX_PORTS; port++)
        {
            printf("Port %d : state change requested %u times\n",
                port,
                portRequestStats[port]);
        }
    }
    printf("\n");
}

/* ----------------------------------------------
*/
void
ixAtmdAccPortMgmtStatsReset (void)
{
    memset(portRequestStats, 0, sizeof(portRequestStats));
}

/* ----------------------------------------------
*/
void
ixAtmdAccUtopiaControlStatsShow (void)
{
    IxAtmdAccUtopiaStatus utStatus;

    if (ixAtmdAccPortMgmtInitDone)
    {
        ixAtmdAccUtopiaStatusGet (&utStatus);

        printf ("Utopia Data\n");
        printf ("           utTxCellCount ....... : %10u\n",
            (unsigned int) utStatus.utTxCellCount);
        printf ("           utTxIdleCellCount ... : %10u\n",
            (unsigned int) utStatus.utTxIdleCellCount);
        printf ("           txFIFO2Underflow .... : %10u\n",
            (unsigned int) utStatus.utTxCellConditionStatus.txFIFO2Underflow);
        printf ("           txFIFO1Underflow .... : %10u\n",
            (unsigned int) utStatus.utTxCellConditionStatus.txFIFO1Underflow);
        printf ("           txFIFO2Overflow ..... : %10u\n",
            (unsigned int) utStatus.utTxCellConditionStatus.txFIFO2Overflow);
        printf ("           txFIFO1Overflow ..... : %10u\n",
            (unsigned int) utStatus.utTxCellConditionStatus.txFIFO1Overflow);
        printf ("           txIdleCellCountOvr .. : %10u\n",
            (unsigned int) utStatus.utTxCellConditionStatus.txIdleCellCountOvr);
        printf ("           txCellCountOvr ...... : %10u\n",
            (unsigned int) utStatus.utTxCellConditionStatus.txCellCountOvr);
        printf ("           utRxCellCount ....... : %10u\n",
            (unsigned int) utStatus.utRxCellCount);
        printf ("           utRxIdleCellCount ... : %10u\n",
            (unsigned int) utStatus.utRxIdleCellCount);
        printf ("           utRxInvalidHECount .. : %10u\n",
            (unsigned int) utStatus.utRxInvalidHECount);
        printf ("           utRxInvalidParCount . : %10u\n",
            (unsigned int) utStatus.utRxInvalidParCount);
        printf ("           utRxInvalidSizeCount  : %10u\n",
            (unsigned int) utStatus.utRxInvalidSizeCount);
        printf ("           rxCellCountOvr ...... : %10u\n",
            (unsigned int) utStatus.utRxCellConditionStatus.rxCellCountOvr);
        printf ("           invalidHecCountOvr .. : %10u\n",
            (unsigned int) utStatus.utRxCellConditionStatus.invalidHecCountOvr);
        printf ("           invalidParCountOvr .. : %10u\n",
            (unsigned int) utStatus.utRxCellConditionStatus.invalidParCountOvr);
        printf ("           invalidSizeCountOvr . : %10u\n",
            (unsigned int) utStatus.utRxCellConditionStatus.invalidSizeCountOvr);
        printf ("           rxIdleCountOvr ...... : %10u\n",
            (unsigned int) utStatus.utRxCellConditionStatus.rxIdleCountOvr);
        printf ("           rxFIFO2Underflow .... : %10u\n",
            (unsigned int) utStatus.utRxCellConditionStatus.rxFIFO2Underflow);
        printf ("           rxFIFO1Underflow .... : %10u\n",
            (unsigned int) utStatus.utRxCellConditionStatus.rxFIFO1Underflow);
        printf ("           rxFIFO2Overflow ..... : %10u\n",
            (unsigned int) utStatus.utRxCellConditionStatus.rxFIFO2Overflow);
        printf ("           rxFIFO1Overflow ..... : %10u\n",
            (unsigned int) utStatus.utRxCellConditionStatus.rxFIFO1Overflow);
    }

    return;
}

/* ----------------------------------------------
*/
void
ixAtmdAccUtopiaControlStatsReset (void)
{
}

/* -----------------------------------------
* Functions visable in this file only
* ----------------------------------------- */
PRIVATE IX_STATUS
ixAtmdAccPortStateChange (IxAtmLogicalPort port, IxAtmdAccPortState newState)
{
    IX_STATUS returnStatus = IX_SUCCESS;

    /* Check parameters */
    if ((port < IX_UTOPIA_PORT_0)
        ||(port >= IX_UTOPIA_MAX_PORTS)
        || !ixAtmdAccPortConfigured(port))
    {
        returnStatus = IX_FAIL;
    }

    if (returnStatus == IX_SUCCESS)
    {
        IX_ATMDACC_PORTMGMT_LOCK_GET();

        /* Notify the interested party */
        returnStatus = stateChangeRequest(port, newState);

        /* update Stats */
        portRequestStats[port]++;

        IX_ATMDACC_PORTMGMT_LOCK_RELEASE ();
    } /* end of if(returnStatus) */
    return returnStatus;
}

/* ----------------------------------------------
*/
PRIVATE IX_STATUS
ixAtmdAccPortStateQuery (IxAtmLogicalPort port,
                         IxAtmdAccPortState state,
                         BOOL *paramError)
{
    IX_STATUS returnStatus = IX_SUCCESS;

    *paramError = FALSE;

    /* Check parameters */
    if ((port < IX_UTOPIA_PORT_0)
        ||(port >= IX_UTOPIA_MAX_PORTS)
        || !ixAtmdAccPortConfigured(port))
    {
        *paramError = TRUE;
        returnStatus = IX_SUCCESS;
    }
    else if ((state != IX_ATMD_PORT_ENABLED) &&
        (state != IX_ATMD_PORT_DISABLED))
    {
        *paramError = TRUE;
        returnStatus = IX_FAIL;
    }
    else if (returnStatus == IX_SUCCESS)
    {
        IX_ATMDACC_PORTMGMT_LOCK_GET();

        if (state == IX_ATMD_PORT_ENABLED)
        {
            if (!isEnabledQuery (port))
            {
                returnStatus = IX_FAIL;
            }
        }
        else
        {
            if (!isDisableComplete (port))
            {
                returnStatus = IX_FAIL;
            }
        } /* end of if-else(state) */

        IX_ATMDACC_PORTMGMT_LOCK_RELEASE();
    } /* end of if(returnStatus) */

    return returnStatus;
}

/* ----------------------------------------------
*/
PRIVATE IX_STATUS
ixAtmdAccDummyStateChangeHandler (IxAtmLogicalPort port,
                                  IxAtmdAccPortState newState)
{
    IX_ATMDACC_ENSURE(0, "PortMgmt No StateChange Handler Registered");
    return IX_FAIL;
}

/* ----------------------------------------------
*/
PRIVATE void
ixAtmdAccDummySetupNotifyHandler (unsigned int numPort)
{
    IX_ATMDACC_ENSURE(0, "PortMgmt No Setup Handler Registered");
}

/* ----------------------------------------------
*/
PRIVATE BOOL
ixAtmdAccDummyStateQuery (IxAtmLogicalPort port)
{
    IX_ATMDACC_ENSURE(0, "PortMgmt No State Check Registered");
    return FALSE;
}

/* ----------------------------------------------
*/
PRIVATE IX_STATUS
ixAtmdAccUtopiaConfigWrite (const IxAtmdAccUtopiaConfig *utConfig)
{
    IX_STATUS returnStatus = IX_SUCCESS;
    UINT32* npeWords = (UINT32 *)utConfig;
    int numWords = sizeof (IxAtmdAccUtopiaConfig) / sizeof (UINT32);
    int wordCnt;

    IX_ATMDACC_ENSURE (numWords < 64, "out of bound structure");

    /* Write one word of the config at a time to NPE-A */
    for (wordCnt = 0;
    (wordCnt < numWords) && (returnStatus == IX_SUCCESS);
    wordCnt++)
    {
        returnStatus = ixAtmdAccUtilNpeMsgSend (IX_NPE_A_MSSG_ATM_UTOPIA_CONFIG_WRITE,
            wordCnt * sizeof(UINT32),
            npeWords[wordCnt]);
    } /* end of for(wordCnt) */

    return returnStatus;
}

/* ----------------------------------------------
*/
PRIVATE IX_STATUS
ixAtmdAccUtopiaConfigLoad (void)
{
    IX_STATUS returnStatus = IX_SUCCESS;

    returnStatus = ixAtmdAccUtilNpeMsgWithResponseSend
        (IX_NPE_A_MSSG_ATM_UTOPIA_CONFIG_LOAD,
        NPE_IGNORE,
        NPE_IGNORE,
        ixAtmdAccUtopiaConfigLoadCallback);;

    if (returnStatus == IX_SUCCESS)
    {
        /* Wait for ixAtmdAccUtopiaConfigGetCallback to release */
        IX_ATMDACC_NPE_RESPONSE_LOCK_GET ();
    }

    return returnStatus;
}

/* ----------------------------------------------
*/
PRIVATE void
ixAtmdAccUtopiaConfigLoadCallback (IxNpeMhNpeId npeMhNpeId,
                                   IxNpeMhMessage npeMhMessage)
{
    UINT32 id;
    UINT32 status;

    IX_ATMDACC_CONVERT_FROM_BIG_ENDIAN (npeMhMessage.data[0]);

    /* Check NpeId */
    IX_ATMDACC_ENSURE (npeMhNpeId == IX_NPEMH_NPEID_NPEA, "wrong npeMhNpeId");

    /* Check the ID */
    id = npeMhMessage.data[0] & NPE_RESP_ID_MASK;

    IX_ATMDACC_ENSURE (id == NPE_UT_CONFIG_LOAD_EXPECTED_ID, "wrong id");

    /* Check the status */
    status = npeMhMessage.data[0] & NPE_RESP_STATUS_MASK;
    IX_ATMDACC_ENSURE (status == NPE_SUCCESS, "wrong status");

    /* Unblock the config get context */
    IX_ATMDACC_NPE_RESPONSE_LOCK_RELEASE ();
}

/* ----------------------------------------------
*/
PRIVATE IX_STATUS
ixAtmdAccUtopiaStatusUpload (void)
{
    IX_STATUS returnStatus;

    /* Setup message for UtopiaStatus command */
    returnStatus = ixAtmdAccUtilNpeMsgWithResponseSend(
        IX_NPE_A_MSSG_ATM_UTOPIA_STATUS_UPLOAD,
        NPE_IGNORE,
        NPE_IGNORE,
        ixAtmdAccUtopiaStatusUploadCallback);

    if (returnStatus == IX_SUCCESS)
    {
        /* Wait for ixAtmdAccUtopiaConfigGetCallback to release */
        IX_ATMDACC_NPE_RESPONSE_LOCK_GET ();
    }

    return returnStatus;
}

/* ----------------------------------------------
*/
PRIVATE void
ixAtmdAccUtopiaStatusUploadCallback (IxNpeMhNpeId npeMhNpeId,
                                     IxNpeMhMessage npeMhMessage)
{
    UINT32 id;
    UINT32 status;

    IX_ATMDACC_CONVERT_FROM_BIG_ENDIAN (npeMhMessage.data[0]);

    /* Check NpeId */
    IX_ATMDACC_ENSURE (npeMhNpeId == IX_NPEMH_NPEID_NPEA, "wrong npeMhNpeId");

    /* Check the ID */
    id = npeMhMessage.data[0] & NPE_RESP_ID_MASK;

    IX_ATMDACC_ENSURE (id == NPE_UT_STATUS_UPLOAD_EXPECTED_ID, "wrong id");

    /* Check the status */
    status = npeMhMessage.data[0] & NPE_RESP_STATUS_MASK;
    IX_ATMDACC_ENSURE (status == NPE_SUCCESS, "wrong status");

    /* Unblock the config get context */
    IX_ATMDACC_NPE_RESPONSE_LOCK_RELEASE ();
}

/* ----------------------------------------------
*/
PRIVATE void
ixAtmdAccNpeStatusReadCallback (IxNpeMhNpeId npeMhNpeId,
                                IxNpeMhMessage npeMhMessage)
{
    UINT32 id;

    IX_ATMDACC_CONVERT_FROM_BIG_ENDIAN (npeMhMessage.data[0]);
    IX_ATMDACC_CONVERT_FROM_BIG_ENDIAN (npeMhMessage.data[1]);

    /* Check NpeId */
    IX_ATMDACC_ENSURE (npeMhNpeId == IX_NPEMH_NPEID_NPEA, "wrong npeMhNpeId");

    /* Check Id */
    id = npeMhMessage.data[0] & NPE_RESP_ID_MASK;
    IX_ATMDACC_ENSURE (id == NPE_UT_STATUS_READ_EXPECTED_ID, "wrong id");

    npeRespOffsetRead = npeMhMessage.data[0] & NPE_RESP_OFFSET_MASK;
    npeRespWordRead = npeMhMessage.data[1];

    /* Unblock the reading context */
    IX_ATMDACC_NPE_RESPONSE_LOCK_RELEASE ();
}

/* ----------------------------------------------
*/
PRIVATE IX_STATUS
ixAtmdAccUtopiaStatusRead (IxAtmdAccUtopiaStatus *utStatus)
{
    IX_STATUS returnStatus = IX_SUCCESS;
    UINT32* utStatusWords = (UINT32 *)utStatus;
    int numWords = sizeof(IxAtmdAccUtopiaStatus)/sizeof(UINT32);
    int wordCnt;
    UINT32 readOffset;

    IX_ATMDACC_ENSURE (numWords < 64, "out of bound structure");

    memset (utStatus, 0, sizeof (IxAtmdAccUtopiaStatus));

    /* Read the status table from NPE memory */

    for (wordCnt = 0;
        (wordCnt < numWords) && (returnStatus == IX_SUCCESS);
        wordCnt++)
    {
        /* offset to read */
        readOffset = wordCnt * sizeof(UINT32);
        /* read the word of the status table at this offset */
        returnStatus = ixAtmdAccUtilNpeMsgWithResponseSend(
            IX_NPE_A_MSSG_ATM_UTOPIA_STATUS_READ,
            readOffset,
            NPE_IGNORE,
            ixAtmdAccNpeStatusReadCallback);

        if (returnStatus == IX_SUCCESS)
        {
            /* Wait for ixAtmdAccNpeReadCallback to release */
            IX_ATMDACC_NPE_RESPONSE_LOCK_GET ();

            /* Check the address matches the address from ixAtmdAccNpeReadCallback */
            IX_ATMDACC_ENSURE (npeRespOffsetRead == readOffset, "Read error");

            /* Copy the word read from NPE memory into the utopia status
            * structure. NOTE: npewordRead is set by ixAtmdAccNpeReadCallback
            */
            utStatusWords[wordCnt] = npeRespWordRead;

        } /* end of if(returnStatus) */
    } /* end of for(wordCnt) */

    return returnStatus;
}

/* ----------------------------------------------
*/
PRIVATE BOOL
ixAtmdAccUtopiaConfigSetParamsValidate (const IxAtmdAccUtopiaConfig *utConfig)
{
    unsigned int numberOfPortsToConfigure = 0;
    BOOL isValid = TRUE;

    if (utConfig == NULL)
    {
        isValid = FALSE;
    }

    if (isValid)
    {
        /* txAddrRange  = (numberOfPorts - 1) */
        numberOfPortsToConfigure = utConfig->utTxConfig.txAddrRange + 1;

        /* Number of Rx ports must be equal the the number of Tx ports */
        if (numberOfPortsToConfigure != (utConfig->utRxConfig.rxAddrRange + 1))
        {
            isValid = FALSE;
        }
    } /* end of if(isValid) */

    if (isValid)
    {
    /* Check the number of TxVc Queues is equal to the number of configured
    * ports
    */
        if (numTxVcQueues < numberOfPortsToConfigure)
        {
            isValid = FALSE;;
        }

        if (isValid)
        {
            /* Check that the number of ports is valid */
            if (numberOfPortsToConfigure > IX_UTOPIA_MAX_PORTS)
            {
                isValid = FALSE;
            }
        } /* end of if(isValid) */
    } /* end of if(isValid) */

    return isValid;
}

