/**
 * @file IxAdslCodelet.c
 *
 * @date 10/12/2002
 *
 * @brief IXP425 codelet to bring up ADSL lines
 *
 * This codelet is for Linux @em only.  When loaded, it will bring up
 * any ADSL lines which were requested by the @c ixp425_adsl_line_num
 * parameter.
 *
 * It's also possible to use this module to bring up CO lines using
 * the @c ixp425_adsl_board_type parameter.  In
 * that case, @c ixp425_adsl_line_num is not used - line 0 is always
 * brought up.
 *
 * -- Intel Copyright Notice -- 
 *
 * @par
 * INTEL CONFIDENTIAL
 * 
 * @par
 * Copyright 200x-200x Intel Corporation All Rights Reserved.
 * 
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.  The Material
 * contains trade secrets and proprietary and confidential information of
 * Intel or its suppliers and licensors.  The Material is protected by
 * worldwide copyright and trade secret laws and treaty provisions. No
 * part of the Material may be used, copied, reproduced, modified,
 * published, uploaded, posted, transmitted, distributed, or disclosed in
 * any way without Intel's prior express written permission.
 * 
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel or otherwise.  Any license under
 * such intellectual property rights must be express and approved by
 * Intel in writing.
 * 
 * -- End Intel Copyright Notice --
 */

/*
 * Put the system defined include files required.
 */
#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <asm/hardware.h>

/*
 * Put the user defined include files required.
 */
#include "IxTypes.h"
#include "IxAdsl.h"

/*
 * #defines and macros used in this file.
 */
/**
 * @def LINE_MAX
 *
 * @brief Maximum length of the string parameter
 * @c ixp425_adsl_board_type
 */
#define LINE_MAX 80

/**
 * @def ADSL_MAX_LINES
 *
 * @brief Maximum number of ADSL lines
 */
#define ADSL_MAX_LINES 12

/*
 * Variable declarations global to this file only.  Externs are followed by
 * static variables.
 */
/**
 * @var ixp425_adsl_board_type
 * @brief The board type in string form
 */
static char ixp425_adsl_board_type[LINE_MAX] = { 0 };

/**
 * @var ixp425_adsl_board_type_enum
 * @brief The board type in enum form
 */
static enum { CPE, CO } ixp425_adsl_board_type_enum = CPE;

/**
 * @var ixp425_adsl_line_num
 * @brief The list of lines to open
 */
static int ixp425_adsl_line_num[ADSL_MAX_LINES] = {
	-1, -1, -1, -1,
	-1, -1, -1, -1,
	-1, -1, -1, -1
};

MODULE_PARM(ixp425_adsl_board_type, "s");
MODULE_PARM_DESC(ixp425_adsl_board_type, "The evaluation board type, CPE (the default) or CO");
MODULE_PARM(ixp425_adsl_line_num, "0-12i");
MODULE_PARM_DESC(ixp425_adsl_line_num, "The ADSL line numbers to open, for CPE only");

/*
 * Function definition.
 */
/**
 * @fn ixAdslCodeletInitModule
 *
 * @brief <brief description>
 *
 * Global Data  :
 * - @c ixp425_adsl_board_type - Reads this, if it exists
 * - @c ixp425_adsl_board_type_enum - Sets this based on the value of
 *   ixp425_adsl_board_type_enum
 * - @c ixp425_adsl_line_num - Opens any of these which are set (i.e. not -1)
 *
 * This is the module load callback.  It brings ADSL lines up.
 *
 * @return 0 for success
 */
static int __init
ixAdslCodeletInitModule(void)
{
	/* Allow access to the CS1 on the expansion BUS */
	*(UINT32 *)IXP425_EXP_BUS_CS1_BASE_VIRT |= 0xbfff3c03;

	if(strcmp(ixp425_adsl_board_type, "CO") == 0) {
		ixp425_adsl_board_type_enum = CO;
	} else {
		ixp425_adsl_board_type_enum = CPE;
	}
	if(ixp425_adsl_board_type_enum == CPE) {
		unsigned i = 0;
		while(ixp425_adsl_line_num[i] != -1) {
			IxADSL_CPE_Open(ixp425_adsl_line_num[i], ADSL_NORMAL);
			i++;
		}
	} else {
		IxADSL_CO_Open(0, ADSL_NORMAL);
	}
	return 0;
}

/**
 * @fn ixAdslCodeletExitModule
 *
 * @brief Module unload callback which brings down any ADSL lines
 *
 * Global Data  :
 * - @c ixp425_adsl_board_type_enum - The board type
 * - @c ixp425_adsl_line_num - For CPE, which lines to close.
 * These will both have been set when the module was loaded.
 *
 * This is the module unload callback.  It brings ADSL lines down.
 *
 * @return none
 */
static void __exit
ixAdslCodeletExitModule(void)
{
	if(ixp425_adsl_board_type_enum == CPE) {
		unsigned i = 0;
		while(ixp425_adsl_line_num[i] != -1) {
			IxADSL_CPE_LineClose(ixp425_adsl_line_num[i]);
			i++;
		}
	} else {
		IxADSL_CO_LineClose(0);
	}
}

module_init(ixAdslCodeletInitModule);
module_exit(ixAdslCodeletExitModule);
