/**
 * @file    IxAtmCodelet.h
 *
 * @date    17-May-2002
 *
 * @brief   Header file for the IXP4xx Atm codelet
 *
 * -- Intel Copyright Notice --
 *
 * @par
 * INTEL CONFIDENTIAL
 *
 * @par
 * Copyright 2002 Intel Corporation All Rights Reserved.
 *
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.  The Material
 * contains trade secrets and proprietary and confidential information of
 * Intel or its suppliers and licensors.  The Material is protected by
 * worldwide copyright and trade secret laws and treaty provisions. No
 * part of the Material may be used, copied, reproduced, modified,
 * published, uploaded, posted, transmitted, distributed, or disclosed in
 * any way without Intel's prior express written permission.
 *
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel or otherwise.  Any license under
 * such intellectual property rights must be express and approved by
 * Intel in writing.
 *
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * -- End Intel Copyright Notice --
 *
 * @sa IxAtmCodelet.h
 *
 */

/**
 * @defgroup Codelets IXP4xx Codelets
 *
 * @brief IXP4xx demo codelets
 *
 * @{
 */

/**
 * @ingroup Codelets
 *
 * @defgroup IxAtmCodelet IXP4xx Atm Codelet (IxAtmCodelet) API
 *
 * @brief IXP4xx codelet ATM component API
 *
 * This codelet demonstrates an example implementation of a working Atm
 * driver that makes use of the AtmdAcc component, as well as demonstrating how
 * the lower layer IxAtmdAcc component can be used for configuration and control.
 *
 * <b>Features</b>
 * @li An interface is provided to setup Aal5/Aal0 Transmit and Recieve VCs on one or 
 *     more Ports. A maximum of 32 Tx and 32 Rx VCs can be setup on a maximum of 12 Ports. 
 *     Only UBR VCs can be setup. When a channel is setup using this interface both a Transmit
 *     and a Recieve VC is created.
 *     Aal0 and Aal5 VCs can be setup at the same time.
 * 
 * @li An interface is provided to remove all registered Aal5/Aal0 VCs.
 *
 * @li Both remote and local loopback of Aal5/Aal0 is provided by this codelet. 
 *     Local loopback refers to loopback provided by the UTOPIA interface. In local loopback
 *     packets generated by the IXP4xx are looped back to the Atm driver by the UTOPIA
 *     hardware.
 *     Software loopback refers to the software looping all packets received on the wire back out
 *     onto the wire.
 *     Remote loopback refers to where the far end is expected to performa a software loopback, i.e. any
 *     packets sent by the codelet are expected to be looped back by the far end into the codelet.
 *      
 *
 * @li Both interrupt and polled mode of operation is provided by this codelet.
 *
 * @li An interface is provided to send Aal5 SDUs specifying the sdu size and the
 *     number of Aal5 sdus to send.
 *
 * @li An interface is provided to send Aal0 packets specifying the packet size and the
 *     number of Aal0 packets to send.
 *
 * @li Both the Transmit port rate and the Recieve port rate can be modified. The Tx port rate is
 *     used by IxAtmSch in performing the shaping functions. The Rx port rate is not used by any
 *     component.
 *
 * @li An interface is provided to allow the querying of the ATM ports and registered VCs.
 *
 * <b>IXP4xx ATM Components used by this codelet</b>
 * @li IxAtmdAcc. This component is the low level interface by which AAL packets get transmitted
 *                to, and received from the UTOPIA bus.
 *
 * <b>IXP4xx ATM Codelet components used by this codelet</b>
 * @li IxAtmSch. This component demonstrates an ATM Traffic Shaper implementation. IxAtmdAcc
 *               gets scheduling information from this component.
 * 
 * @li IxAtmm. This component provides ATM port and VC management facilities. This component
 *             also manages the configuration of the UTOPIA co-processor.
 *
 * <b>IxAtmCodelet modes of operation</b><br>
 * This codelet can be initialised to operate in one of three configurations.
 * 
 * @li IX_ATMCODELET_UTOPIA_LOOPBACK. In this mode the UTOPIA interface will loopback all traffic
 *     transmitted.
 *     <b>Buffer management</b>
 *     In this mode a simple buffering mechanism is used; mbufs are allocated from
 *     the vxWorks pool as needed for RxFree replenishing/Tx and are returned to the vxWorks pool
 *     for TxDone/Rx. 
 *
 *     <b>Interrupt/Task based processing</b>
 *     In this mode of operation the IxQMgrDispatcher is hooked to interupts. This means that TxDone/RxLo
 *     will be perfomed from a task level and IxAtmdAcc Tx processing/RxHi and RxFree will be precessed
 *     from an interrupt level.
 * 
 *     <b>Sending Aal5 PDUs</b>
 *     In this mode of operation an interface is available to send Aal5 PDUs of specified size in bytes.
 *
 *     <b>Sending Aal0 Packets</b>
 *     In this mode of operation an interface is available to send Aal0 Packets of specified size in cells.
 *
 * @li IX_ATMCODELET_REMOTE_LOOPBACK.
 *     In this mode packets are sent and are expected to be looped back on the remote end.
 *
 * @li IX_ATMCODELET_SOFTWARE_LOOPBACK. In this mode of operation all traffic sent are looped back to receive
 *     in the IXP4xx device.
 *     <b>Buffer management</b>
 *     In this mode a more complex buffering mechanism is used; mbufs are allocated from
 *     the vxWorks pool and stored in a software queue. These mbufs are fetched from this software
 *     queue as needed for RxFree replenishing/Tx and are returned to the software queue for TxDone/Rx. 
 *
 *     <b>Interrupt/Task based processing</b>
 *     In this mode of operation the IxQMgrDispatcher is called form a task every 1 msec. This means that
 *     TxDone/RxLo/IxAtmdAcc Tx processing/RxHi and RxFree will be processed from a task level.
 * 
 *     <b>Sending PDUs/Packets</b>
 *     This mode of operation does not provide an interface for sending Aal5 PDUs/Aal0 Packets.
 *
 *     <b>IxAtmCodelet sub-components</b>
 *     <pre>
 *                           +--------------------+
 *                           |      AtmCodelet    |
 *                           +--------------------+
 *                               |          |
 *                         ,-----'          `-----,
 *                         |                      |
 *                         V                      V
 *                    +------------+        +------------+
 *                    |  RxTxDemo  |        | SwLoopback |
 *                    +------------+        +------------+

 *     </pre>
 * <b>AtmCodelet</b>
 * This implements the API functions.
 *
 * <b>RxTxDemo</b>
 * This sub-component implements the IX_ATMCODELET_UTOPIA_LOOPBACK and IX_ATMCODELET_REMOTE_LOOPBACK modes
 * of operation.
 *
 * <b>SwLoopback</b>
 * This sub-component implements the IX_ATMCODELET_SOFTWARE_LOOPBACK mode operation.
 *
 * <b>BufMan</b>
 * This sub-component implements the intefaces used internally for getting and returning vxWorks mbufs.
 *
 * <b>UTOPIA Recieve and Transmit PHY addresses</b>
 * In this codelet UTOPIA Phy Addresses are assigned numbers starting at UTOPIA_PHY_ADDR and
 * are incremented for each port used. These addresses _WILL_ need to be changed depending on
 * the hardware setup.
 * 
 * @par
 * Modification history
 * 21 Aug 02 Ported functionality from IxAal5DemoCodelet.h. Added Aal0 + OAM support
 *
 * @{
 */

#ifndef IXATMCODELET_H
#define IXATMCODELET_H

/*
 * User defined includes
 */
#include "IxAtmTypes.h"
#include "IxAtmdAcc.h"

/*
 * #defines and macros used in this file.
 */

/*
 * Typedefs
 */

/**
 *
 * @enum IxAtmCodeletMode
 *
 * @ingroup IxAtmCodelet
 *
 * @brief Mode of operation of the AtmCodelet codelet.
 *
 */
typedef enum
{
    IX_ATMCODELET_UTOPIA_LOOPBACK = 0, /**< All cells sent out are looped back in by the
                                        *   UTOPIA hardware. Useful for demonstrating traffic
                                        *   on a board with no ADSL PHY or external test equipment */
    IX_ATMCODELET_SOFTWARE_LOOPBACK,   /**< All cells received are looped back out again by the software */
    IX_ATMCODELET_REMOTE_LOOPBACK      /**< All cells sent are anticipated to be loopbed back by external
                                        * equipment */
} IxAtmCodeletMode;

/*
 * Function declarations
 */

/** @brief This function is used to initialize the lower level system component
 *         used by the IxAtmCodelet. It should be called before any other
 *         IxAtmCodelet API function UNLESS the system initialisation is done
 *         elsewhere. 
 *
 * @param "unsigned int numPorts" The number of ATM ports to configure.
 *         Valid values range from 1 to 12.
 *
 * @param "IxAtmDemoCodeletMode mode" This indicates the codelet mode.
 *
 *
 * @return
 * - <b>IX_SUCCESS :</b> indicates that the Atm Codelet system init successfull.
 *
 * - <b>IX_FAIL :</b> Some internal error has prevented the Atm Codelet
 *          sstem components from initialising.
 */
PUBLIC IX_STATUS
ixAtmCodeletSystemInit (unsigned int numPorts,
			IxAtmCodeletMode mode);

/** @brief This function is used to initialize the IxAtmCodelet. It
 *         should be called before any other IxAtmCodelet API function.
 *         This function can be called from the VxWorks windSh.
 *
 * @param "unsigned int numPorts" The number of ATM ports to configure.
 *         Valid values range from 1 to 12.
 *
 * @param "unsigned int rxToTxRatio" Ratio of PDUs received to attempted PDU sends
 *        when softwareLoopbackEnabled := TRUE. e.g. if rxToTxRatio is 6 then 1 PDU
 *        will be sent for every 6 PDUs received.
 *
 * @return
 * - <b>IX_SUCCESS :</b> indicates tha the Atm Codelet has been successfully 
 *          initialized.
 * - <b>IX_FAIL :</b> Some internal error has prevented the Atm Codelet
 *          from initialising.
 */
IX_STATUS
ixAtmCodeletInit (unsigned int numPorts, 
		  unsigned int rxToTxRatio);

/** @brief This function is used to provision a number of bidirectional
 *         Atm UBR channels on the specified number of ports.
 *         This function may be called a number of times for different Aal
 *         types until all of the available 32 channels are used.
 *
 * @param "unsigned int numPorts" The number of ATM ports to configure
 *        channels on. This must be less than or equal to the number of
 *        ports configured in the call to ixAtmCodeletInit().
 *
 * @param "unsigned int numChannels" The number of bidirectional Atm
 *        channels to provision. Channels are provisioned accross numPorts
 *        in the following manner, for 2 ports, 4 VCs,
 *        Channel 0 on port 0, channel 1 on port 1, channel 2 on port 0,
 *        channel 3 on port 1.
 *        Valid values range from 1 to 32.
 *        Subsequent calls will assign channel and port numbers in ascending order.
 *
 * @param "IxAtmdAccAalType aalType" The aalType for the channels to
 *        provision, i.e. Aal5, Aal0 48 byte mode or Aal0 52 byte cell mode.
 *
 * @return
 * - <b>IX_SUCCESS :</b> indicates tha the Atm Codelet has successfully provisioned the requested
 *          number of channels.
 * - <b>IX_FAIL :</b> Some internal error has prevented the ATM codelet from
 *          from provisioning the channels.
 */
IX_STATUS
ixAtmCodeletUbrChannelsProvision (unsigned int numPorts,
				  unsigned int numChannels,
				  IxAtmdAccAalType aalType);

/** @brief This function is used to remove all channels provisioned earlier
 *         in a call to ixAtmCodeletUbrChannelsProvision().
 *
 * @param none
 *
 * @return
 * - <b>IX_SUCCESS :</b> indicates that the Atm Codelet has successfully removed
 *          all provisioned channels.
 * - <b>IX_FAIL :</b> Some internal error has prevented the Atm Codelet
 *          from removing provisioned channels.
 */
IX_STATUS
ixAtmCodeletUbrChannelsRemove (void);


/** @brief This function is used to send the specified number of Aal0 packets,
 *         with the number of cells per packet specified. The packets are multiplexed
 *         across all provisioned Aal0 48/52 channels.
 *         This function will block until the transmission is complete.
 *
 * @param "unsigned int cellsPerPacket" The number of cells per packet to send.
 *
 * @param "unsigned int numPackets" The total number of packets to send.
 *
 *
 * @return
 * - <b>IX_SUCCESS :</b> indicates tha the Atm Codelet has successfully sent
 *          all of the packets.
 * - <b>IX_FAIL :</b> Some internal error has prevented the Atm Codelet
 *          from sending the packets successfully.
 */
IX_STATUS
ixAtmCodeletAal0PacketsSend (unsigned int cellsPerPacket,
			     unsigned int numPackets);

/** @brief This function is used to send the specified number of Aal5 CPCS SDUs
 *         of specified SDU size multiplexed accross all Aal5 provisioned channels.
 *         This function will block until the transmission is complete.
 *
 * @param "unsigned int sduSize" The SDU size of the sdus to send in bytes. The Atm
 *         trailer + padding will be appended to this.
 *
 * @param "unsigned int numsdus" The total number of sdus to send.
 *
 *
 * @return
 * - <b>IX_SUCCESS :</b> indicates tha the Atm Codelet has successfully sent
 *          all of the sdus.
 * - <b>IX_FAIL :</b> Some internal error has prevented the Atm Codelet
 *          from sending the sdus successfully.
 */
IX_STATUS
ixAtmCodeletAal5CpcsSdusSend (unsigned int sduSize,
			      unsigned int numSdus);

/** @brief This function is called to modify the portRate on a
 *         previously initialized port.
 *
 * @param "IxAtmLogicalPort port" Specifies the ATM port which is to be
 *          modified.
 *
 * @return
 * - <b>IX_SUCCESS :</b> The port rate has been successfully modified.<br>
 * - <b>IX_FAIL :</b> The port rate could not be modified, either
 *      because the input data was invalid, or the new port rate is
 *      insufficient to support established ATM VC contracts on this
 *      port.
 */
IX_STATUS
ixAtmCodeletPortRateModify (IxAtmLogicalPort port,
			    unsigned int portRate);

/** @brief This function is used to dump VC configuration and sdu send and
 *         receive counters.
 *
 * @param none
 *
 * @return void
 */
void
ixAtmCodeletShow (void);
#endif
/* IXATMCODELET_H */

/** @} defgroup IxAtmCodelet*/

/** @} defgroup Codelets*/




