/**
 * @file    IxAtmCodeletRxTxDemo.c
 *
 * @date    20-May-2002
 *
 * @brief   Rx and Tx Transport for IXP425 Atm Codelet (IxAtmCodelet)
 *
 * Functions implemented in this file:
 *    ixAtmRxTxDemoInit
 *    ixAtmRxTxDemoStatsGet
 *    ixAtmRxTxDemoChannelsProvision
 *    ixAtmRxTxDemoSdusSend
 * -- Intel Copyright Notice --
 *
 * @par
 * INTEL CONFIDENTIAL
 *
 * @par
 * Copyright 2002 Intel Corporation All Rights Reserved.
 *
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.  The Material
 * contains trade secrets and proprietary and confidential information of
 * Intel or its suppliers and licensors.  The Material is protected by
 * worldwide copyright and trade secret laws and treaty provisions. No
 * part of the Material may be used, copied, reproduced, modified,
 * published, uploaded, posted, transmitted, distributed, or disclosed in
 * any way without Intel's prior express written permission.
 *
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel or otherwise.  Any license under
 * such intellectual property rights must be express and approved by
 * Intel in writing.
 *
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * -- End Intel Copyright Notice --
 *
 */

/*
 * System defined include files.
 */
#include <stdio.h>

/*
 * User defined include files.
 */
#include "IxTypes.h"
#include "ix_ossl.h"
#include "IxAssert.h"
#include "IxOsServices.h"
#include "IxAtmdAcc.h"
#include "IxDemoUtils.h"
#include "IxAtmCodelet_p.h"
#include "IxAtmCodeletRxTxDemo_p.h"

/*
 * #defines and macros
 */
#define IX_ATMCODELET_TX_RETRY_CNT         (1000) /* Number of retries before failing to transmit a PDU */
/*
  Aal5 PDU                      <------- trailer -------->
  +------------------------------------------------------+
  |    Payload           | PAD  | UU | CPI | Length | CRC|
  +------------------------------------------------------+
  |     1 - 65,535       | 0-47 | 1  |  1  |   2    |  4 |
 */
#define IX_ATMCODELET_TRAILER_SIZE         (8) /* Atm trailer size in bytes. UU+CPI+Length+CRC          */
#define IX_ATMCODELET_SDU_LEN_LO_BYTE      (3) /* Offset in trailer to the low byte of SDU length       */
#define IX_ATMCODELET_SDU_LEN_HI_BYTE      (2) /* Offset in trailer to the high bye of SDU length field */
#define IX_ATMCODELET_CPI_OFFSET           (1) /* Offset to Common Part Indicator field                 */
#define IX_ATMCODELET_UUI_OFFSET           (0) /* Offset to User-User field                             */
#define IX_ATMCODELET_DEFAULT_CPI          (0) /* Default Common Part Indicator field value             */
#define IX_ATMCODELET_DEFAULT_UUI          (9) /* Default User-User field value                         */
#define IX_ATMCODELET_SHIFT_EIGHT_BITS     (8) /* Operand to << and >> operators                        */
#define IX_ATMCODELET_MAX_SDU_LEN          (65535) /* Max length of Aal5 SDU                            */
#define IX_ATMCODELET_MAX_PACKET_LEN       (65535) /* Max length of Aal0 Packet                         */
/* Max offset in an ATM cell                         */
#define IX_ATMCODELET_FIRST_BYTE_MASK      (0x00ff)
#define IX_ATMCODELET_SECOND_BYTE_MASK     (0xff00)

#define IX_ATMCODELET_SAME_CELL_MAX_OFFSET (IX_ATM_CELL_PAYLOAD_SIZE-IX_ATMCODELET_TRAILER_SIZE)
#define IX_ATMCODELET_NEW_CELL_THRESHOLD (IX_ATMCODELET_SAME_CELL_MAX_OFFSET+1)
#define IX_ATMCODELET_NEW_CELL_MAX_OFFSET (IX_ATMCODELET_SAME_CELL_MAX_OFFSET+IX_ATM_CELL_PAYLOAD_SIZE)

/*
 * Typedefs
 */

typedef struct {
    IxAtmConnId connId;
    IxAtmdAccAalType aalType;
    unsigned int bytesPerCell;
    unsigned int atmHeader;
} VcInfo;

/*
 * Static variables
 */
static VcInfo ixAtmRxTxDemoTxVcInfoTable[IX_ATM_MAX_NUM_VC];
static VcInfo ixAtmRxTxDemoRxVcInfoTable[IX_ATM_MAX_NUM_VC];

/* Number of VCs configured */
static unsigned int ixAtmRxTxDemoNumChannelsConfigured = 0;

/* Statistics */
static IxAtmCodeletStats *ixAtmRxTxDemoStats = NULL;

/*
 * Function definitions
 */
PRIVATE void
ixAtmRxTxDemoStatsReset (void);


/* Recieve callback */
PRIVATE void
ixAtmRxTxDemoRxCallback (IxAtmLogicalPort port,
			 IxAtmdAccUserId userId,
			 IxAtmdAccPduStatus pduStatus,
			 IxAtmdAccClpStatus clp,
			 IX_MBUF *mbufPtr);

/* Transmit done callback */
PRIVATE void
ixAtmRxTxDemoTxDoneCallback (IxAtmdAccUserId userId,
			     IX_MBUF * mbufPtr);

/* Tansmit an Aal5 packet */
PRIVATE IX_STATUS
ixAtmRxTxDemoAal5CpcsTx (IxAtmConnId connId,
			 unsigned int sduLen);

/* Tansmit an Aal0 packet */
PRIVATE IX_STATUS
ixAtmRxTxDemoAal0Tx (int channelNum,
		     unsigned int packetSize,
		     unsigned int sizeInCells);

/* Generic send with retries function */
PRIVATE IX_STATUS
ixAtmRxTxDemoWithRetiesSend (IxAtmConnId connId, 
			     IX_MBUF* mBuf, 
			     unsigned int sizeInCells);

/* --------------------------------------------------------------
   Initialise the RxTx Demo.
   -------------------------------------------------------------- */
IX_STATUS
ixAtmRxTxDemoInit (IxAtmCodeletStats *stats)
{
    ixAtmRxTxDemoStats = stats;

    /* Reset statistics */
    ixAtmRxTxDemoStatsReset();

    return IX_SUCCESS;
}

/* --------------------------------------------------------------
   Get the IxAtmCodeletRxTxDemo counters.
   -------------------------------------------------------------- */
void
ixAtmRxTxDemoStatsGet (IxAtmCodeletStats *stats)
{
    *stats = *ixAtmRxTxDemoStats;
}

/* --------------------------------------------------------------
   Provision channels specifying the number of ports and channels.
   -------------------------------------------------------------- */
IX_STATUS
ixAtmRxTxDemoChannelsProvision (unsigned int numPorts,
				unsigned int numChannels,
				IxAtmdAccAalType aalType)
{
    IX_STATUS retval;
    static IxAtmLogicalPort port = 0; /* Start with port 0 */
    static unsigned int vci = IX_ATMCODELET_START_VCI;
    static unsigned int vpi = IX_ATMCODELET_START_VPI;    
    static unsigned int channelIdx = 0;
    unsigned int i;
    
    /* Parameter validation */
    if ((numPorts == 0) || (numChannels == 0))
    {
	IX_ATMCODELET_LOG("Invalid parameters to channesls provision \n");
	return IX_FAIL;
    }

    /* N.B. channelIdx is static */
    for (i=0; i < numChannels; i++, channelIdx++)
    {

	/* Save the aalType and bytesPerCell in the Rx & Tx Vc Info tables */
	ixAtmRxTxDemoRxVcInfoTable[channelIdx].aalType = aalType;
	ixAtmRxTxDemoTxVcInfoTable[channelIdx].aalType =
	    ixAtmRxTxDemoRxVcInfoTable[channelIdx].aalType;
	if (aalType == IX_ATMDACC_AAL0_52)
	{
	    ixAtmRxTxDemoTxVcInfoTable[channelIdx].bytesPerCell = 
		IX_ATM_AAL0_52_CELL_SIZE_NO_HEC;
	    ixAtmRxTxDemoRxVcInfoTable[channelIdx].bytesPerCell = 
		ixAtmRxTxDemoTxVcInfoTable[channelIdx].bytesPerCell;
	    
	    /* build the cell header for transmission */
	    ixAtmRxTxDemoTxVcInfoTable[channelIdx].atmHeader = (((vci & 0xFFFF) << 4) | 
								((vpi & 0xFF) << 20));
	    ixAtmRxTxDemoRxVcInfoTable[channelIdx].atmHeader = ixAtmRxTxDemoTxVcInfoTable[channelIdx].atmHeader;
	}
	else
	{
	    ixAtmRxTxDemoTxVcInfoTable[channelIdx].bytesPerCell = IX_ATM_AAL0_48_CELL_PAYLOAD_SIZE;
	    ixAtmRxTxDemoRxVcInfoTable[channelIdx].bytesPerCell = 
		ixAtmRxTxDemoTxVcInfoTable[channelIdx].bytesPerCell;
	}

	/*
	 * Register this VC with IxAtmm and connect it to IxAtmdAcc
	 */
	retval = ixDemoUtilsAtmVcRegisterConnect(port,
						 vpi,
						 vci,
						 aalType,
						 IX_ATM_UBR,
						 IX_ATM_RX_A,
						 ixAtmRxTxDemoRxCallback,
						 IX_ATMCODELET_RX_FREE_Q_LOW_THRESHOLD,
						 ixAtmRxTxDemoTxDoneCallback,
						 ixAtmRxTxDemoRxFreeLowReplenish,
						 channelIdx,
						 &ixAtmRxTxDemoRxVcInfoTable[channelIdx].connId,
						 &ixAtmRxTxDemoTxVcInfoTable[channelIdx].connId);
	
	if (retval != IX_SUCCESS)
	{
	    IX_ATMCODELET_LOG("VC setup failed for port %2u vpi %3u vci %5u\n", 
			      port, vpi, vci);
	    return IX_FAIL;
	}
	else
	{
	    IX_ATMCODELET_LOG("VC configured for port %2u vpi %3u vci %5u\n", 
			      port, vpi, vci);
	}
	
	/* Use a different port per channel, depending on the maximum
	 * number of ports configured
	 */
	port++;

	/* Wrap the port back to 0 when numPorts used already */
	if (port >= numPorts)
        {
	    port = 0;
        }

	/* Use the next vci */
	vci++;
       
	ixAtmRxTxDemoNumChannelsConfigured++;
    }

    return IX_SUCCESS;
}

/* --------------------------------------------------------------
   Send a number of PDUs on the registered channels.
   -------------------------------------------------------------- */
IX_STATUS
ixAtmRxTxDemoAal0PacketsSend (unsigned int cellsPerPacket,
			      unsigned int numPackets)
{
    unsigned int pktCnt;
    IX_STATUS retval;
    unsigned int channelIdx;
    unsigned int packetSize;

    for (pktCnt=0; pktCnt<numPackets;)
    {
	/* For each Aal0 VC */
	for (channelIdx=0; (pktCnt<numPackets) &&
		 (channelIdx<ixAtmRxTxDemoNumChannelsConfigured); channelIdx++, pktCnt++)
	{
	    if ((ixAtmRxTxDemoTxVcInfoTable[channelIdx].aalType == IX_ATMDACC_AAL0_48) ||
		(ixAtmRxTxDemoTxVcInfoTable[channelIdx].aalType == IX_ATMDACC_AAL0_52))
	    {
		packetSize = ixAtmRxTxDemoTxVcInfoTable[channelIdx].bytesPerCell * cellsPerPacket;

		retval = ixAtmRxTxDemoAal0Tx (channelIdx, /* ixAtmRxTxDemoTxVcInfoTable[channelIdx].connId */
					      packetSize,
					      cellsPerPacket );
		
		if (retval != IX_SUCCESS)
		{
		    IX_ATMCODELET_LOG ("Failed to send a pdu on channel...... %u\n", channelIdx);
		}
	    }
	    else
	    {
		/* decrement the pcktCnt since a packet was not sent on this vc */
		pktCnt--;
	    }
	}
    }

    return IX_SUCCESS;    
}

/* --------------------------------------------------------------
   Send a number of PDUs on the registered channels.
   -------------------------------------------------------------- */
IX_STATUS
ixAtmRxTxDemoAal5CpcsSdusSend (unsigned int sduSize,
			       unsigned int numSdus)
{
    unsigned int sduCnt;
    IX_STATUS retval;
    unsigned int channelIdx;

    for (sduCnt=0; sduCnt<numSdus;)
    {
	/* For each Aal5 VC */
	for (channelIdx=0; (sduCnt<numSdus) &&
		 (channelIdx < ixAtmRxTxDemoNumChannelsConfigured); channelIdx++, sduCnt++)
	{
	    if (ixAtmRxTxDemoTxVcInfoTable[channelIdx].aalType == IX_ATMDACC_AAL5)
	    {
		retval = ixAtmRxTxDemoAal5CpcsTx (ixAtmRxTxDemoTxVcInfoTable[channelIdx].connId,
						  sduSize);
		
		if (retval != IX_SUCCESS)
		{
		    IX_ATMCODELET_LOG ("Failed to send a pdu on channel...... %u\n", channelIdx);
		}
	    }
	    else
	    {
		/* decrement the sduCnt a sdu was not sent on this vc */
		sduCnt--;
	    }
	}
    }    

    return IX_SUCCESS;    
}

/* --------------------------------------------------------------
   Replenish buffers to the RxFree queue. Buffers are taken from
   the mbuf pool
   -------------------------------------------------------------- */
void
ixAtmRxTxDemoRxFreeLowReplenish (IxAtmdAccUserId userId)
{
    IX_MBUF *mBuf;
    IX_STATUS retval;
    IxAtmConnId connId;
    unsigned int numFreeEntries;
    int cnt;
    
    connId = ixAtmRxTxDemoRxVcInfoTable[userId].connId;
    
    /* Get the number of free entries in the RX Free queue */
    retval = ixAtmdAccRxVcFreeEntriesQuery (connId, &numFreeEntries);
    if (retval != IX_SUCCESS)
    {
	IX_ATMCODELET_LOG("Failed to query depth of Rx Free Q for connection %u\n",
			  connId);
	return;
    }
    
    /* Replenish Rx buffers  */
    for (cnt=0; cnt<numFreeEntries; cnt++)
    {
	/* Get an MBUF */
	ixDemoUtilsMbufGet(IX_ATMCODELET_MBUF_SIZE, &mBuf);
	
	if (mBuf == NULL)
	{
	    IX_ATMCODELET_LOG("Failed to get rx free mbuffers\n");
	    return;
	}

	IX_ASSERT(mBuf->m_next == 0);

	/* 
	 * Set the number of bytes of data in this MBUF to the number of
	 * payloads that will fit evenly into IX_ATMCODELET_MBUF_SIZE.
	 */
	mBuf->m_len = 
	    (IX_ATMCODELET_MBUF_SIZE / ixAtmRxTxDemoRxVcInfoTable[userId].bytesPerCell) *
	    ixAtmRxTxDemoRxVcInfoTable[userId].bytesPerCell;
	
	/* Send free buffers to NPE */
	retval = ixAtmdAccRxVcFreeReplenish (connId, mBuf);
	
	if (retval != IX_SUCCESS)
	{
	    /* Free the allocated buffer */
	    ixDemoUtilsMbufFree(mBuf);
	    IX_ATMCODELET_LOG("Failed to pass Rx free buffers to Atmd \n");
	    return;
	}
	
	/* Update stats */
	ixAtmRxTxDemoStats->rxFreeBuffers++;
    }
}

/* --------------------------------------------------------------
   Return an Aal0 packet. This assumes that the Packet will fit into a
   single mbuf. i.e. No chaining.
   -------------------------------------------------------------- */
PRIVATE IX_STATUS
ixAtmRxTxDemoAal0PacketBuild (int channelNum,
			      unsigned int packetLen,
			      IX_MBUF **mBuf)
{
    unsigned char *dataPtr;
    unsigned char headerChar1, headerChar2, headerChar3, headerChar4;
    unsigned int tmpLen;

    if (IX_ATMCODELET_MAX_PACKET_LEN < packetLen)
    {
	IX_ATMCODELET_LOG ("ixAtmRxTxDemoAal0PacketBuild(): IX_ATMCODELET_MAX_PACKET_LEN < packetLen\n");
	return IX_FAIL;
    }

    /* Initialise the returned pointer */
    *mBuf = NULL;

    /* Get the first buffer */
    ixDemoUtilsMbufGet (packetLen, mBuf);

    if (*mBuf == NULL)
    {
	IX_ATMCODELET_LOG ("ixAtmRxTxDemoAal0PacketBuild() failed to get mBuf\n");
        return IX_FAIL;
    }

    /* if the aaltype is aal0_52 then insert the header into the buffer */
    if (ixAtmRxTxDemoTxVcInfoTable[channelNum].aalType == IX_ATMDACC_AAL0_52)
    {
	dataPtr = (*mBuf)->m_data;
	tmpLen = packetLen;
	
	headerChar1 = (char)(ixAtmRxTxDemoTxVcInfoTable[channelNum].atmHeader >> 24);
	headerChar2 = (char)((ixAtmRxTxDemoTxVcInfoTable[channelNum].atmHeader >> 16) & 0xFF);
	headerChar3 = (char)((ixAtmRxTxDemoTxVcInfoTable[channelNum].atmHeader >> 8) & 0xFF);
	headerChar4 = (char)(ixAtmRxTxDemoTxVcInfoTable[channelNum].atmHeader & 0xFF);

	IX_ATMCODELET_LOG ("AtmHeader = %u for channelNum %u\n", ixAtmRxTxDemoTxVcInfoTable[channelNum].atmHeader, channelNum);
	
	while(tmpLen > 0)
	{
	    /* get the dataPtr and insert the header */
	    *dataPtr++ = headerChar1;
	    *dataPtr++ = headerChar2;
	    *dataPtr++ = headerChar3;
	    *dataPtr++ = headerChar4;

	    /* increment to the next header position */
	    dataPtr = dataPtr + 48;

	    /* decrement the packetlen */
	    tmpLen = tmpLen - ixAtmRxTxDemoTxVcInfoTable[channelNum].bytesPerCell;
	}
    }

    IX_ASSERT((*mBuf)->m_next == 0);

    /* Setup the MBUF,
     * N.B. for simplicity not filling in payload
     */
    (*mBuf)->m_pkthdr.len = packetLen;

    return IX_SUCCESS;
}

/* --------------------------------------------------------------
   Return an Aal5 PDU. This assumes that the PDU will fit into a
   single mbuf. i.e. No chaining.
   -------------------------------------------------------------- */
PRIVATE IX_STATUS
ixAtmRxTxDemoAal5CpcsPduBuild (unsigned int sduLen,
			       IX_MBUF **mBuf)
{
    int pduLen;
    char *trailer = NULL;

    if (IX_ATMCODELET_MAX_SDU_LEN < sduLen)
    {
	IX_ATMCODELET_LOG ("ixAtmRxTxDemoAal5CpcsPduBuild(): IX_ATMCODELET_MAX_SDU_LEN < sduLen\n");
	return IX_FAIL;
    }

    /* Initialise the returned pointer */
    *mBuf = NULL;

    /* Get the first buffer */
    ixDemoUtilsMbufGet (sduLen, mBuf);
;
    if (*mBuf == NULL)
    {
	IX_ATMCODELET_LOG ("ixAtmRxTxDemoAal5CpcsPduBuild() failed to get mBuf\n");
        return IX_FAIL;
    }

    IX_ASSERT((*mBuf)->m_next == 0);

    /* CASE 1: Enough bytes left in CELL to store trailer */
    if ((sduLen % IX_ATM_CELL_PAYLOAD_SIZE) ==
	IX_ATMCODELET_SAME_CELL_MAX_OFFSET)
    {
	pduLen = sduLen + IX_ATMCODELET_TRAILER_SIZE;
    }
    /* CASE 2: More than enough bytes in CELL to store trailer */
    else if ((sduLen % IX_ATM_CELL_PAYLOAD_SIZE) <
	     IX_ATMCODELET_SAME_CELL_MAX_OFFSET)
    {
	pduLen = (sduLen - (sduLen % IX_ATM_CELL_PAYLOAD_SIZE)) +
	    IX_ATM_CELL_PAYLOAD_SIZE;
    }
    /* CASE 3: Not enough bytes in CELL to store trailer. Need another CELL */
    else
    {
	pduLen = (sduLen - (sduLen % IX_ATM_CELL_PAYLOAD_SIZE)) +
	    (IX_ATM_CELL_PAYLOAD_SIZE * 2/* Need another CELL*/);
    }
    
    trailer = (*mBuf)->m_data + pduLen - 8;
    
    /*Fill out the relevant inf/trailero in the trailer*/
    trailer[IX_ATMCODELET_UUI_OFFSET]      = IX_ATMCODELET_DEFAULT_UUI;
    trailer[IX_ATMCODELET_CPI_OFFSET]      = IX_ATMCODELET_DEFAULT_CPI;
    trailer[IX_ATMCODELET_SDU_LEN_LO_BYTE] = sduLen & IX_ATMCODELET_FIRST_BYTE_MASK;
    trailer[IX_ATMCODELET_SDU_LEN_HI_BYTE] = (sduLen & IX_ATMCODELET_SECOND_BYTE_MASK)
	>> IX_ATMCODELET_SHIFT_EIGHT_BITS;
        
    (*mBuf)->m_pkthdr.len = pduLen;
    (*mBuf)->m_len = pduLen;

    return IX_SUCCESS;
}

/* --------------------------------------------------------------
   Reset the counters.
   -------------------------------------------------------------- */
PRIVATE void
ixAtmRxTxDemoStatsReset (void)
{
    if (NULL != ixAtmRxTxDemoStats)
    {
	ixAtmRxTxDemoStats->txPdus = 0;
	ixAtmRxTxDemoStats->txBytes = 0;
	ixAtmRxTxDemoStats->rxPdus = 0;
	ixAtmRxTxDemoStats->rxBytes = 0;
	ixAtmRxTxDemoStats->txDonePdus = 0;    
	ixAtmRxTxDemoStats->rxFreeBuffers = 0;
    }
}

/* --------------------------------------------------------------
   The callback is called whenever the TxDone queue reaches a
   certain level. Mbufs are simply returned to the vxWorks pool.
   -------------------------------------------------------------- */
PRIVATE void
ixAtmRxTxDemoTxDoneCallback (IxAtmdAccUserId userId,
			     IX_MBUF *mbufPtr)
{
    ixAtmRxTxDemoStats->txDonePdus++;

    /* Return the buffer to the vxWorks pool */
    ixDemoUtilsMbufFree(mbufPtr);
}


/* --------------------------------------------------------------
   This callback is called when a PDU is recieved. The Mbuf
   is freed to the vxWorks pool.
   -------------------------------------------------------------- */
PRIVATE void
ixAtmRxTxDemoRxCallback (IxAtmLogicalPort port,
			 IxAtmdAccUserId userId,
			 IxAtmdAccPduStatus pduStatus,
			 IxAtmdAccClpStatus clp,
			 IX_MBUF *mbufPtr)
{
    unsigned int sizeInCells;
    IxAtmdAccAalType aalType;

    /* Update stats */
    ixAtmRxTxDemoStats->rxPdus++;

    if (mbufPtr == NULL)
    {
	IX_ATMCODELET_LOG("rxCallback called with NULL mbufPtr\n");
	return;
    }

    /* Size of the data transmitted in CELLS */
    sizeInCells = mbufPtr->m_pkthdr.len / 
	ixAtmRxTxDemoRxVcInfoTable[userId].bytesPerCell;

    if (pduStatus == IX_ATMDACC_MBUF_RETURN)
    {
	/* Return this buffer to the pool */
	ixDemoUtilsMbufFree (mbufPtr);
	return;
    }

    aalType = ixAtmRxTxDemoRxVcInfoTable[userId].aalType;

    if ((aalType == IX_ATMDACC_AAL0_48) || (aalType == IX_ATMDACC_AAL0_52))
    {
	/* Aal0 should always be valid */
	IX_ASSERT(pduStatus == IX_ATMDACC_AAL0_VALID);
    }
    else 
    {
	/* Must be Aal5 */
	IX_ASSERT(aalType == IX_ATMDACC_AAL5);

	/* Check PDU status, could be valid, partial or crc error */
	if (pduStatus != IX_ATMDACC_AAL5_VALID)
	{	
	    IX_ATMCODELET_LOG("rxCallback called with CRC error\n");
	}
    }
    ixAtmRxTxDemoStats->rxBytes += mbufPtr->m_pkthdr.len;
    
    /* Return this buffer to the pool */
    ixDemoUtilsMbufFree (mbufPtr);

    return;
}

/* --------------------------------------------------------------
   Transmit an Atm PDU of the specified length on the channel
   associated with connId.
   -------------------------------------------------------------- */
PRIVATE IX_STATUS
ixAtmRxTxDemoAal5CpcsTx (IxAtmConnId connId,
			 unsigned int sduLen)
{
    int sizeInCells = 0;
    IX_MBUF *mBuf;

    /* Build an Aal5 PDU for this SDU */
    if (ixAtmRxTxDemoAal5CpcsPduBuild (sduLen, &mBuf) != IX_SUCCESS)
    {
	return IX_FAIL;
    }

    /* Size of the data transmitted in CELLS */
    sizeInCells = mBuf->m_pkthdr.len / IX_ATM_CELL_PAYLOAD_SIZE;

    return ixAtmRxTxDemoWithRetiesSend (connId, mBuf, sizeInCells);
}

/* --------------------------------------------------------------
   Transmit an Aal0 Packet of the specified length on the channel
   associated with connId.
   -------------------------------------------------------------- */
PRIVATE IX_STATUS
ixAtmRxTxDemoAal0Tx (int channelNum,
		     unsigned int packetSize,
		     unsigned int sizeInCells)
{
    IX_MBUF *mBuf;

    /* Build an Aal0 Packet */
    if (ixAtmRxTxDemoAal0PacketBuild (channelNum, packetSize, &mBuf) != IX_SUCCESS)
    {
	return IX_FAIL;
    }

    return ixAtmRxTxDemoWithRetiesSend (ixAtmRxTxDemoTxVcInfoTable[channelNum].connId, mBuf, sizeInCells);
}

PRIVATE IX_STATUS
ixAtmRxTxDemoWithRetiesSend (IxAtmConnId connId, IX_MBUF* mBuf, unsigned int sizeInCells)
{
    IX_STATUS status;
    unsigned int retryCount;

    /* Transmit with retries */
    for (status=IX_FAIL, retryCount=0; ((retryCount<IX_ATMCODELET_TX_RETRY_CNT) &&
					(status!=IX_SUCCESS)); retryCount++)
    {
	status = ixAtmdAccTxVcPduSubmit(connId,
					mBuf,
					IX_ATMCODELET_DEFAULT_CLP,
					sizeInCells);

        if (status != IX_SUCCESS)
        {
            if (status != IX_ATMDACC_BUSY)
            {
		IX_ATMCODELET_LOG("txAndRetry() failed when sending pkt\n");
                break;
            }

	    ix_ossl_sleep(1);
        }
    }

    if (status == IX_SUCCESS)
    {	
	/* Update stats */
	ixAtmRxTxDemoStats->txPdus++;
	ixAtmRxTxDemoStats->txBytes += mBuf->m_pkthdr.len;
    }
    else
    {
	status = IX_FAIL;
    }
    
    return status;
}


