/**
 * @file IxCryptoAccCodelet.h
 *
 * @date October-31-2002
 *
 * @brief This is the header file for the Crypto Access Compenent Codelet.
 *
 * -- Intel Copyright Notice -- 
 *
 * @par
 * INTEL CONFIDENTIAL
 * 
 * @par
 * Copyright 2002 Intel Corporation All Rights Reserved.
 * 
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.  The Material
 * contains trade secrets and proprietary and confidential information of
 * Intel or its suppliers and licensors.  The Material is protected by
 * worldwide copyright and trade secret laws and treaty provisions. No
 * part of the Material may be used, copied, reproduced, modified,
 * published, uploaded, posted, transmitted, distributed, or disclosed in
 * any way without Intel's prior express written permission.
 * 
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel or otherwise.  Any license under
 * such intellectual property rights must be express and approved by
 * Intel in writing.
 * 
 * -- End Intel Copyright Notice -- 
 */

/**
 * @defgroup Codelets IXP4xx Codelets
 *
 * @brief IXP4xx demo codelets
 *
 * @{
 */

/**
 *
 * @ingroup Codelets
 *
 * @defgroup IxCryptoAccCodelet IXP4xx Crypto Access Codelet (IxCryptoAccCodelet) API
 *
 * @brief IXP4xx demo Crypto Access component API
 *
 * Functionality of the cryptoAcc Codelet
 *
 * The codelet shall demonstrate how the Security Hardware Accelerator API 
 * can be used for cryptographic purposes.
 *
 * Three different implementations are demonstrated. 
 * 
 * 1. 	Encryption and decryption with DES in CBC mode.
 * 2.	Authentication calculation and check for SHA1.
 * 3.	Combined service of encryption/authentication calculation and
 *	authentication check/decryption. The combined services uses DES
 *	CBC mode with SHA1.
 * 
 * Here is the list of functionality of the codelet and they are in the
 * order in which the codelet must follow in order to utilise the Security
 * Hardware Accelerator, i.e. Initialise the related component and followed by
 * registration, and then perform followed by unregistration, etc.
 *
 * 1. 	The codelet shall demonstrate how the Network Processor Engine (NPE)
 * 	is initialised and started. It will also show how the latest NPE crypto
 *	image is downloaded to the NPE.
 *
 * 2.	The codelet shall demonstrate how the Queue Manager is initialised to
 *	polling or interrupt mode.
 *
 * 3.	The codelet will show an example of register and perform callback
 *	functions and how they are set up.
 *
 * 4.	The codelet will demonstrate how the registration should be done for
 *	the three different implementation using the register API.
 *	Note that the authentication data doesn't matter when encrypt/decrypt 
 *	only operation is chosen and vice-versa.
 *
 * 5.	The codelet shall demonstrate how the perform API can be used after
 *	successful registration.
 *
 * 6.	The codelet shall demonstrate the use of the unregister API in the
 *	event of re-starting a new functionality or implementation.
 *
 * 7.	The codelet will measure the performance of the Security Hardware
 *	Accelerator for the three implementation stated above.
 *
 * For COEXIST functionality, -DCOEXIST should be defined in cryptoAcc codelet
 * component.mk. The COEXIST codelet shall:
 * 1. 	Demonstrate that cryptographic process can be run simultenously with 
 *	ethernet packet tranmission in port 2 in vxWorks only.
 *
 * Note:For interrupt mode: 
 *	1.	Packets could be lost due to the usage of log messages, in
 *	        crypto and COEXIST mode. Longer packets will eliminate this
 *		effect due to the fact that the NPE is processing faster than
 *		XScale core.
 *	2.	Longer packets for the crypto is required (>1024 bytes) for
 *		COEXIST to work and the line rate utilization of approximately
 *		50% is required. Higher utilization rate will result in the
 *		cryptoAcc callback not being serviced.
 *
 * Note:For linux, only one callback is demonstrated per buffer. Perform
 *	 function is not executed in the Perform callback function.
 *
 * @{
 */
#ifndef IxCryptoAccCodelet_H
#define IxCryptoAccCodelet_H

#include <stdlib.h>

#include "IxTypes.h"
#include "IxOsBuffMgt.h"
#include "IxNpeDl.h"

#define INLINE __inline__


/**
 * @ingroup IxCryptoAccCodelet
 *
 * @fn  ixCryptoAccCodeletInit(void)
 * @brief This function is the entry point to the Crypto Access codelet.
 * 	  It must be called before any of the demos as it initialises the 
 *	  neccessary component.
 *
 * @note The initialisation sequence is as follows:
 * <UL>
 *	<LI> Initialise Network Processor Engine (NPE)
 *	<LI> Initialise Queue manager
 *	<LI> Initialise Crypto Access component
 *	<LI> Start NPE
 *	<LI> Start Queue manager dispatcher. 
 * </UL>  
 *
 * @param None
 *
 * @return @li IX_SUCCESS - Successful initialisation.
 *	   @li IX_FAIL - Initialisation failed.
 *
 * Reentrant     : no
 * ISR Callable  : no
 */

IX_STATUS ixCryptoAccCodeletInit (void);


/**
 * @ingroup IxCryptoAccCodelet
 *
 * @fn ixCryptoAccCodeletMbufToQTailAdd(IX_MBUF *pMbufToAdd)
 *
 * @brief This function will add an mbuf to the queue. The mbuf will
 *	  be added to the end of the queue.  
 *
 * @param IX_MBUF* [in] pMbufToAdd  - Pointer of the mbuf to be added  to the
 *				      tail of the queue.
 *
 * @return None.
 *
 * Reentrant     : yes
 * ISR Callable  : no
 */

void ixCryptoAccCodeletMbufToQTailAdd (IX_MBUF *pMbufToAdd);


/**
 * @ingroup IxCryptoAccCodelet
 *
 * @fn ixCryptoAccCodeletMbufFromQHeadRemove(IX_MBUF **pMbufToRem)
 *
 * @brief This function will remove a mbuf from the queue. The mbuf will
 *	  be removed from the head of the queue.
 *
 * @param IX_MBUF** [in] pMbufToRem  - Pointer to a pointer of the mbuf to be
 *				       removed from the head of the queue.
 *
 * @return None.
 *
 * Reentrant     : yes
 * ISR Callable  : no
 */

void ixCryptoAccCodeletMbufFromQHeadRemove (IX_MBUF **pMbufToRem);


/**
 * @ingroup IxCryptoAccCodelet
 *
 * @fn ixCryptoAccCodeletMemPoolInit(void)
 *
 * @brief This function will initialise mbuf pool for the Crypto
 *	  Access Codelet.
 *
 * @param None
 *
 * @return @li IX_SUCCESS - Successful mbuf pool initialisation.
 *	   @li IX_FAIL - Mbuf pool initialisation failed.
 *
 * Reentrant     : no
 * ISR Callable  : no
 */

PRIVATE IX_STATUS ixCryptoAccCodeletMemPoolInit (void);


/**
 * @ingroup IxCryptoAccCodelet
 *
 * @fn ixCryptoAccCodeletNpeInit(IxNpeDlNpeId npeId)
 *
 * @brief This function is to download microcode to the NPE. 
 * 	  This function always downloads the latest crypto
 *        version of NPE microcode available.
 *
 * @param IxNpeDlNpeId [in] npeId - The ID of the NPE which the microcode
 *				    downloaded to.
 *
 * @return @li IX_SUCCESS - Successful NPE initialisation.
 *	   @li IX_FAIL - NPE initialisation failed.
 *
 * Reentrant     : no
 * ISR Callable  : no
 */

PRIVATE IX_STATUS ixCryptoAccCodeletNpeInit (IxNpeDlNpeId npeId);


/**
 * @ingroup IxCryptoAccCodelet
 *
 * @fn ixCryptoAccCodeletDispatcherStart(BOOL useInterrupt)
 *
 * @brief  This function starts the Queue manager dispatcher.
 *
 * @param BOOL [in] useInterrupt - start in interrupt or polled mode.
 *		    		   TRUE is interrupt mode while FALSE is
 *				   poll mode.
 * 
 * @return IX_SUCCESS - Dispatcher successfully started.
 * @return IX_FAIL - Error starting dispatcher.
 *
 * Reentrant     : no
 * ISR Callable  : no
 */

PRIVATE IX_STATUS ixCryptoAccCodeletDispatcherStart (BOOL useInterrupt);


/**
 * @ingroup IxCryptoAccCodelet
 *
 * @fn ixCryptoAccCodeletDispatcherPoll(void)
 *
 * @brief This function polls the Queue manager for the poll mode
 *	  set in the function ixCryptoAccCodeletDispatcherStart.
 *
 * @param None
 * 
 * @return None
 *
 * Reentrant     : no
 * ISR Callable  : no
 */

PRIVATE void ixCryptoAccCodeletDispatcherPoll (void);


/**
 * @ingroup IxCryptoAccCodelet
 *
 * @fn ixCryptoAccCodeletRegisterCtx(UINT32 opType)
 *
 * @brief  This function will register to the Crypto Access Component for
 * 	   three different operation which will demonstrate the capability
 *	   of the security hardware accelerator.
 *	   The types of operation are:
 *	   1 - Encryption and Decryption
 *	   2 - Authentication: Calculation and Check
 *	   3 - Encryption/Decryption with authentication
 *
 * @param UINT32 [in] opType - Type of operation to be registered. This
 *			       parameter will only take 3 different unsigned
 *			       integer as shown above in the brief, e.g. 1 for
 *			       encryption and decryption only and so on.
 * 
 * @return IX_SUCCESS - Register successfully
 * @return IX_FAIL - Error in registering
 *
 * Reentrant     : no
 * ISR Callable  : no
 */

IX_STATUS ixCryptoAccCodeletRegisterCtx (UINT32 opType);


/**
 * @ingroup IxCryptoAccCodelet
 *
 * @fn test_main(void)
 *
 * @brief  This is the main function of the Codelet where registration
 *	   and perform is done. This function will also rotate between
 *	   encryption/authentication and authentication/decryption
 *
 * @return integer.
 *
 * Reentrant     : no
 * ISR Callable  : no
 */

void test_main (void);


/**
 * @ingroup IxCryptoAccCodelet
 *
 * @fn ixCryptoAccCodeletBufIsQEmpty(void)
 *
 * @brief This inline function is used to check whether all the buffer
 *        queue is empty.
 *
 * @return BOOL
 *             - TRUE - the queue is full.
 *             - FALSE - the queue is not full.
 */

INLINE BOOL 
ixCryptoAccCodeletBufIsQEmpty(void);


/**
 * @ingroup IxCryptoAccCodelet
 *
 * @fn ixCryptoAccCodeletBufIsQFull(void)
 *
 * @brief This inline function is used to check whether all the buffer
 *        queue is full.
 *
 * @return BOOL
 *             - TRUE - the queue is full.
 *             - FALSE - the queue is not full.
 */

INLINE BOOL 
ixCryptoAccCodeletBufIsQFull (void);

/**
 * @ingroup IxCryptoAccCodelet
 *
 * @def IX_CRYPTO_ACC_CODELET_DRV_DMA_FREE
 *
 * @brief To free the memory allocated through IX_CRYPTO_ACC_DRV_DMA_MALLOC
 *        function.
 *
 * @note Example of usage: IX_CRYPTO_ACC_CODELET_DRV_DMA_FREE(pData)
 *
 * @param UINT32 [in] *pData - memory block pointer to be freed.
 *
 * @return None
 *
 */
#ifdef __vxworks
 
#define IX_CRYPTO_ACC_CODELET_DRV_DMA_FREE(pData) cacheDmaFree (pData)            

#else

/* Note: Do nothing in  Linux platform. Will be ported to Linux platform.*/
#define IX_CRYPTO_ACC_CODELET_DRV_DMA_FREE(pData) (pData) 

#endif /* def __vxworks */

#endif /* IxCryptoAccCodelet_H */

/** @} defgroup IxCryptoAccCodelet*/

/** @} defgroup Codelet*/
