/**
 * @file    IxDemoUtils.h
 *
 * @date    17-May-2002
 *
 * @brief   Header file for the IXP4xx demo utils for ATM
 *
 * -- Intel Copyright Notice --
 *
 * @par
 * INTEL CONFIDENTIAL
 *
 * @par
 * Copyright 2002 Intel Corporation All Rights Reserved.
 *
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.  The Material
 * contains trade secrets and proprietary and confidential information of
 * Intel or its suppliers and licensors.  The Material is protected by
 * worldwide copyright and trade secret laws and treaty provisions. No
 * part of the Material may be used, copied, reproduced, modified,
 * published, uploaded, posted, transmitted, distributed, or disclosed in
 * any way without Intel's prior express written permission.
 *
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel or otherwise.  Any license under
 * such intellectual property rights must be express and approved by
 * Intel in writing.
 *
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * -- End Intel Copyright Notice --
 *
 * @sa IxAtmDemoUtils.h
 *
 */

/**
 * @defgroup Codelets IXP4xx Codelets
 *
 * @brief IXP4xx demo codelets
 *
 * @{
 */

/**
 * @ingroup Codelets
 *
 * @defgroup IxAal5Demo IXP4xx Demo Utils (IxDemoUtils) API
 *
 * @brief IXP4xx Demo Utils component API
 *
 * These utilities provide support for
 *   - adding/removing ATM VCs via the IxAtmdAcc component
 *   - downloading the NPEA image
 *
 * @{
 */
#ifndef IXDEMOUTILSATM_H
#define IXDEMOUTILSATM_H

/* User defined header files */
#include "IxTypes.h"
#include "IxAtmTypes.h"
#include "IxAtmdAcc.h"
#include "IxAtmm.h"
#include "IxNpeDl.h"

/**
 *
 * @ingroup IxAtmDemoUtilsAPI
 *
 * @brief Register and connect an Aal Pdu receive and transmit VC for a
 *        particular port/vpi/vci.
 *
 * This function allows a user to connect to an Aal5 Pdu receive and transmitservice
 * for a particular port/vpi/vci. It registers the callback and allocates
 * internal resources and a Connection Id to be used in further API calls
 * related to this VCC.
 *
 * @param port (in) VC identification : logical PHY port [0..7]
 * @param vpi (in) VC identification : ATM Vpi [0..255]
 * @param vci (in) VC identification : ATM Vci [0..65535]
 * @param atmService (in) type of service AAL5 @b only in ixp425 1.0
 * @param rxQueueId (in) this identifieds which of two Qs the VC
 *     should use.when icoming traffic is processed
 * @param rxCallback (in) function called when mbufs are received.
 *     This parameter cannot be a null pointer.
 * @param minimumReplenishCount (in) number of free mbufs to be used with
 *     this channel. Use a high number when the expected traffic rate on
 *     this channel is high, or when the user's mbufs are small, or when
 *     the RxVcFreeLow Notification has to be invoked less often. 
 * @param bufferFreeCallback (in) function to be called to return
 *     ownership of buffers to IxAtmDemoUtils user.
 * @param userId (in) user Id to use in callback communications
 * @param rxConnId (out) Rx Conn Id passed back from IxAtmdAcc
 * @param txConnId (out) Tx Conn Id passed back from IxAtmdAcc
 *
 * @return @li IX_SUCCESS successful call to ixDemoUtilsAtmVcRegisterConnect
 * @return @li IX_FAIL parameter error, VC already in use or port is
 * not initialised or some other error occurs during processing.
 *
 */
PUBLIC IX_STATUS
ixDemoUtilsAtmVcRegisterConnect (IxAtmLogicalPort port,
				 unsigned vpi,
				 unsigned vci,
				 IxAtmdAccAalType aalType,
				 IxAtmServiceCategory atmService,
				 IxAtmRxQueueId rxQueueId,
				 IxAtmdAccRxVcRxCallback rxCallback,
				 unsigned int minimumReplenishCount,
				 IxAtmdAccTxVcBufferReturnCallback bufferFreeCallback,
				 IxAtmdAccRxVcFreeLowCallback rxFreeLowCallback,
				 IxAtmdAccUserId userId,
				 IxAtmConnId *rxConnId,
				 IxAtmConnId *txConnId);

/**
 *
 * @ingroup IxAtmDemoUtilsAPI
 *
 * @brief Disconnect and unregister an Aal transmit and receive VC.
 *
 * @param rxConnId (in) conn Id of the Rx VC to disconnect and unregister
 * @param txConnId (in) conn Id of the Tx VC to disconnect and unregister
 *
 * @return @li IX_SUCCESS successful call to ixDemoUtilsAtmVcUnregisterDisconnect
 * @return @li IX_FAIL parameter error, failed to disconnect,
 * not initialised or some other error occurs during processing.
 *
 */
PUBLIC IX_STATUS
ixDemoUtilsAtmVcUnregisterDisconnect (IxAtmConnId rxConnId, IxAtmConnId txConnId);


/**
 *
 * @ingroup IxAtmDemoUtilsAPI
 *
 * @brief Disconnect and unregister all registered VCs.
 *
 * @param none
 *
 * @return @li IX_SUCCESS successful call to ixDemoUtilsAtmAllVcsDisconnect
 * @return @li IX_FAIL parameter error, failed to disconnect,  
 * not initialised or some other error occurs during processing.
 *
 */
PUBLIC IX_STATUS
ixDemoUtilsAtmAllVcsDisconnect (void);

/**
 *
 * @ingroup IxAtmDemoUtilsAPI
 *
 * @brief Disconnect and unregister all registered VCs.
 *
 * @param none
 *
 * @return @li IX_SUCCESS successful call to ixDemoUtilsAtmAllVcsDisconnect
 * @return @li IX_FAIL parameter error, failed to disconnect,  
 * not initialised or some other error occurs during processing.
 *
 */
PUBLIC IX_STATUS
ixDemoUtilsAtmAllVcsRemove (void);

/**
 *
 * @ingroup IxAtmDemoUtilsAPI
 *
 * @brief Download an ATM image to the NPEA
 *
 * @param numports (in) Number of ports in the system. A value > 1 requires 
 *        a MPHY_MULTIPORT to NPEA image to be downloaded. A value of 1 can use
 *        a SPHY or MPHY NPEA image.
 *
 * @param imagePtr (in) Pointer to image file that can contain multiple images.
 *
 * @param phyMode (out) This parameter will be set depending on which images is 
 *        available in imagePtr. See description for numports above.
 *
 *
 * @return @li IX_SUCCESS successful call to ixDemoUtilsAtmNpeImagedownload
 * @return @li IX_FAIL parameter error, failed to disconnect,  
 * not initialised or some other error occurs during processing.
 *
 */
PUBLIC IX_STATUS
ixDemoUtilsAtmImageDownload (unsigned numPorts, 
			     IxAtmmPhyMode *phyMode);

/**
 *
 * @ingroup IxAtmDemoUtilsAPI
 *
 * @brief Download an image to the NPE
 *
 * @note
 * The latest version which has the requested capability will be downloaded.
 *
 * @param npeId (in) Identifier of the NPE to download the image to.
 *
 * @param buildId (in) Identifier of the build within the microCodeImage
 *
 *
 * @return @li IX_SUCCESS successful call to ixDemoUtilsNpeImageDownload
 * @return @li IX_FAIL parameter error, failed to disconnect,  
 * not initialised or some other error occurs during processing.
 *
 */
PUBLIC IX_STATUS
ixDemoUtilsNpeImageDownload ( IxNpeDlNpeId npeId,
			      IxNpeDlBuildId buildId);

/**
 *
 * @ingroup IxAtmDemoUtilsAPI
 *
 * @brief Configure test test and stimulus FPGA clocks for IXP4xx as ATM master
 *
 * @param none
 *
 *
 * @return @li IX_SUCCESS successful call to ixDemoUtilsUtopiaFpgaStimulusAsMasterSet
 * @return @li IX_FAIL parameter error, failed to configure stimulus FPGA,
 * not initialised or some other error occurs during processing.
 *
 */
PUBLIC BOOL 
ixDemoUtilsUtopiaFpgaStimulusAsMasterSet(void);

/**
 *
 * @ingroup IxAtmDemoUtilsAPI
 *
 * @brief Initialize the demo utils buffer pool.
 *
 * @param none
 *
 *
 * @return @li IX_SUCCESS successful call to ixDemoUtilsMbufPoolInit
 * @return @li IX_FAIL parameter error, failed initialize buffer pool
 * or some other error occurs during processing.
 *
 */
IX_STATUS
ixDemoUtilsMbufPoolInit (void);

/**
 *
 * @ingroup IxAtmDemoUtilsAPI
 *
 * @brief Disconnect and unregister all registered VCs.
 *
 * @param none
 *
 * @return @li IX_SUCCESS successful call to ixDemoUtilsAtmAllVcsDisconnect
 * @return @li IX_FAIL parameter error, failed to disconnect,  
 * not initialised or some other error occurs during processing.
 *
 */
PUBLIC IX_STATUS
ixDemoUtilsAtmAllVcsDisconnect (void);

/**
 *
 * @ingroup IxAtmDemoUtilsAPI
 *
 * @brief Disconnect and unregister all registered VCs.
 *
 * @param none
 *
 * @return @li IX_SUCCESS successful call to ixDemoUtilsAtmAllVcsDisconnect
 * @return @li IX_FAIL parameter error, failed to disconnect,  
 * not initialised or some other error occurs during processing.
 *
 */
PUBLIC IX_STATUS
ixDemoUtilsAtmAllVcsRemove (void);

/**
 *
 * @ingroup IxAtmDemoUtilsAPI
 *
 * @brief Get an unchained buffer
 *
 * @param bufSize (in) Buffer capacity.
 *
 * @param buf (out) Pointer to buffer; will be set the new buffer
 *
 *
 * @return @li IX_SUCCESS successful call to ixDemoUtilsMbufGet
 * @return @li IX_FAIL parameter error, failed to get a buffer,  
 * or some other error occurs during processing.
 *
 */
void
ixDemoUtilsMbufGet (unsigned int bufSize, IX_MBUF **buf);

/**
 *
 * @ingroup IxAtmDemoUtilsAPI
 *
 * @brief Free a buffer chain
 *
 * @param buf (out) Pointer to buffer to be freed.
 *
 *
 * @return @li IX_SUCCESS successful call to ixDemoUtilsMbufGet
 * @return @li IX_FAIL parameter error, failed to get a buffer,  
 * or some other error occurs during processing.
 *
 */
void
ixDemoUtilsMbufFree (IX_MBUF *buf);

/**
 *
 * @ingroup IxAtmDemoUtilsAPI
 *
 * @brief Show statistics for the buffer pool; free, used buffer counts etc.
 *
 * @return none
 *
 */
void
ixDemoUtilsMbufShow (void);

#endif
/* IXDEMOUTILSATM_H */

/** @} defgroup IxAal5Demo*/

/** @} defgroup Codelets*/
