/**
 * @file IxDmaAccCodelet.c
 *
 * @author Intel Corporation
 * @date 18 November 2002
 *
 * @brief This file contains the implementation of the Dma Access Codelet.
 *
 * -- Intel Copyright Notice --
 *
 * @par
 * INTEL CONFIDENTIAL
 *
 * @par
 * Copyright 2002 Intel Corporation All Rights Reserved.
 *
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.  The Material
 * contains trade secrets and proprietary and confidential information of
 * Intel or its suppliers and licensors.  The Material is protected by
 * worldwide copyright and trade secret laws and treaty provisions. No
 * part of the Material may be used, copied, reproduced, modified,
 * published, uploaded, posted, transmitted, distributed, or disclosed in
 * any way without Intel's prior express written permission.
 *
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel or otherwise.  Any license under
 * such intellectual property rights must be express and approved by
 * Intel in writing.
 *
 * -- End Intel Copyright Notice --
*/

/*
 * System defined include files.
 */
#include <stdio.h>

/*
 * User defined include files.
 */
#include "IxTypes.h"
#include "IxQMgr.h"
#include "IxNpeDl.h"
#include "IxOsServices.h"
#include "IxOsCacheMMU.h"
#include "IxNpeMicrocode.h"
#include "ix_ossl.h"
#include "IxDmaAcc.h"
#include "IxDmaAccCodelet.h"
#include "IxDmaAccCodelet_p.h"
#include "ixp425.h"

/*
 * Variable declarations global to this file only.
 */
static BOOL ixDmaAccCodeletInitialised = FALSE;
static UINT32 ixDmaAccCodeletDispatchId;

static UINT8 *ixDmaAccCodeletSrcBlock;
             /**< Pointer to memory allocation for storing DMA source test patterns */
static UINT8 *ixDmaAccCodeletDestBlock;
             /**< Pointer to memory allocation for storing DMA destination test patterns */
/*
 * Static function prototypes.
 */
PRIVATE void ixDmaAccCodeletShow(void);
PRIVATE void ixDmaAccCodeletTestPatternReset(void);
PRIVATE void ixDmaAccCodeletDispatcherPoll(void* argUnused, void** ptrRetObjUnused);
PRIVATE IX_STATUS ixDmaAccCodeletDispatcherStart(BOOL useInterrupt);
PRIVATE IX_STATUS ixDmaAccCodeletNpeInit(IxNpeDlNpeId npeId);
PRIVATE void ixDmaAccCodeletTestPerform( UINT16 transferLength,
                                         IxDmaTransferMode transferMode,
                                         IxDmaAddressingMode addressingMode,
                                         IxDmaTransferWidth transferWidth);
PRIVATE void ixDmaAccCodeletCallback(IX_STATUS status);

/*
 * Function definition: ixDmaAccCodeletInit()
 * See header file for documentation.
 */
IX_STATUS ixDmaAccCodeletInit(IxNpeDlNpeId npeId)
{
    /* Block reinitialisation if already done so */
    if(ixDmaAccCodeletInitialised)
    {
        printf("\nDma codelet already initialised");
	    return(IX_FAIL);
    }

    /* get the memory for source area */
    ixDmaAccCodeletSrcBlock = (UINT8*) IX_ACC_DRV_DMA_MALLOC(IX_DMA_CODELET_TEST_MAXLENGTH);

    /* get the memory for destination area */
    ixDmaAccCodeletDestBlock = (UINT8*) IX_ACC_DRV_DMA_MALLOC(IX_DMA_CODELET_TEST_MAXLENGTH);

    /* Initialise Queue Manager */
    printf("\nInitialising Queue Manager...");
    if (ixQMgrInit() != IX_SUCCESS)
    {
	    printf("\nError initialising queue manager!");
	    return (IX_FAIL);
    }
    /* Start the Queue Manager dispatcher loop :
       Parameter is TRUE for Interrupt mode else poll mode */
    if(ixDmaAccCodeletDispatcherStart(FALSE) != IX_SUCCESS)
    {
	    printf("\nError starting queue manager dispatch loop!");
	    return (IX_FAIL);
    }
    /* Initialise NPE and download Image */
    printf("\nInitialising NPE %d...", npeId);
    if(ixDmaAccCodeletNpeInit(npeId) != IX_SUCCESS)
    {
	    printf("\nError initialising NPE %d!", npeId);
	    return (IX_FAIL);
    }
    /* Start NPEs */
    printf("\nStarting NPEs...");
    if(ixNpeDlNpeExecutionStart(npeId) != IX_SUCCESS)
    {
	    printf("\nError starting NPE %d!", npeId);
	    return (IX_FAIL);
    }
    /***********************************************************************
     * System initialisation done. Now initialise Dma Access component.
     ***********************************************************************/
    if (ixDmaAccInit(npeId) != IX_SUCCESS)
    {
	    printf("\nError initialising Dma access driver!");
	    return (IX_FAIL);
    }
    ixDmaAccCodeletInitialised = TRUE;
    return (IX_SUCCESS);
}

/*
 * Function definition: ixDmaAccCodeletDispatcherPoll()
 * See header file for documentation.
 */
PRIVATE void ixDmaAccCodeletDispatcherPoll (void* argUnused, void** ptrRetObjUnused)
{
    while (1)
    {
	    /* Dma NPE A: queues 19 & 20; NPE B: 24 & 26; NPE B: 25 & 27 */
	    ixQMgrDispatcherLoopRun (IX_QMGR_QUELOW_GROUP);
            /* Sleep for MS_SLEEP milliseconds */
            ixOsServTaskSleep(IX_DMA_CODELET_MS_SLEEP);
    }
}

/*
 * Function definition : ixDmaAccCodeletDispatcherStart()
 * See header file for documentation.
 */
PRIVATE IX_STATUS ixDmaAccCodeletDispatcherStart(BOOL useInterrupt)
{

    if(useInterrupt)	/* Interrupt mode */
    {
	/*
	 * Hook the QM QLOW dispatcher to the interrupt controller.
	 * IX_QMGR_QUELOW_GROUP refers to Queues 0-31
	 * Dma NPE A: queues 19 & 20; NPE B: 24 & 26; NPE B: 25 & 27
	 */
	if (ixOsServIntBind(IXP425_INT_LVL_QM1,
			    (IxVoidFnPtr)ixQMgrDispatcherLoopRun,
			    (void *)IX_QMGR_QUELOW_GROUP) != IX_SUCCESS)
	{
	    printf("\nFailed to bind to QM1 interrupt");
	    return (IX_FAIL);
	}
    }
    else	/* Polled mode */
    {
 	    if (ix_ossl_thread_create((ix_ossl_thread_entry_point_t)
				  ixDmaAccCodeletDispatcherPoll,
				  NULL,
				  &ixDmaAccCodeletDispatchId) != IX_OSSL_ERROR_SUCCESS)
	    {
	        printf("\nError spawning dispatch task");
	        return (IX_FAIL);
	    }
	    ix_ossl_thread_set_priority(ixDmaAccCodeletDispatchId, IX_DMA_CODELET_QMR_PRIORITY);
    }
    return (IX_SUCCESS);
}

/*
 * Function definition : ixDmaAccCodeletNpeInit()
 * See header file for documentation.
 */
PRIVATE IX_STATUS ixDmaAccCodeletNpeInit(IxNpeDlNpeId npeId)
{
    IxNpeDlVersionId npeIdList[IX_NPE_MICROCODE_AVAILABLE_VERSIONS_COUNT]; /* Storage for NPE microcode Id list */
    UINT32 npeListSize = IX_NPE_MICROCODE_AVAILABLE_VERSIONS_COUNT;        /* Size for Id list */
    BOOL imageIdFound = FALSE;     /* Flag for Id found */
    IxNpeDlVersionId tmpIdVersion;    /* Storage for single version Id */
    UINT32 i;                      /* Counter for ImageId "for" loop */
    tmpIdVersion.npeId = npeId;       /* Choice of NPE to be initialised provided by client */

    tmpIdVersion.major = 0;
    tmpIdVersion.minor = 0;

    /* Check for invalid npeId value */
    /* Set the build Id for DMA component */
    switch (npeId)
    {
        case IX_NPEDL_NPEID_NPEA :
             tmpIdVersion.buildId = IX_DMA_CODELET_NPE_A_BUILDID;
             break;
        case IX_NPEDL_NPEID_NPEB :
             tmpIdVersion.buildId = IX_DMA_CODELET_NPE_A_BUILDID;
             break;
        case IX_NPEDL_NPEID_NPEC :
             tmpIdVersion.buildId = IX_DMA_CODELET_NPE_A_BUILDID;
             break;
        default:
            /* Invalid NPE ID */
            ixOsServLog ( LOG_ALL,
                          "\nDMAAccDemoNpeInit : invalid Npe ID.",
                          0,0,0,0,0,0);
            return (IX_FAIL);
    }

    /* Stop the NPE before attempting a download */
    if(ixNpeDlNpeStopAndReset(npeId) != IX_SUCCESS)
    {
	    printf("\nNPE stop and reset failed");
	    return (IX_FAIL);
    }

    /* Get list of image versions from the NPE image file */
    if(ixNpeDlAvailableVersionsListGet(npeIdList,
				       &npeListSize) != IX_SUCCESS)
    {
	    printf("\nUnable to get NPE version information");
   	    return (IX_FAIL);
    }

    /* Get the latest version of NPE code available */
    for(i = 0; i < npeListSize; i++)
    {
        if(npeIdList[i].npeId == npeId)
	    {
	        if(npeIdList[i].buildId == tmpIdVersion.buildId)
	        {
                imageIdFound = TRUE;
                if(npeIdList[i].major > tmpIdVersion.major)
		        {
		            tmpIdVersion.major = npeIdList[i].major;
		            tmpIdVersion.minor = npeIdList[i].minor;
		        }
		        else
		        if(npeIdList[i].major == tmpIdVersion.major)
		        {
		            if(npeIdList[i].minor >= tmpIdVersion.minor)
		            {
			            tmpIdVersion.minor = npeIdList[i].minor;
		            }
		        }
	        }
	    }
    }

    /* Quit this function if microcode for Dma function not found */
    if( FALSE==imageIdFound )
    {
	    printf("\nDMA microcode for NPE %d not found ", npeId);
	    return (IX_FAIL);
    }
    printf("\nNPE %d Version information", tmpIdVersion.npeId);
    printf("\nMajor ID:\t 0x%x", tmpIdVersion.major);
    printf("\nMinor ID:\t 0x%x", tmpIdVersion.minor);
    printf("\nBuild ID:\t 0x%x", tmpIdVersion.buildId);
     /* Download NPE code */
    if(ixNpeDlVersionDownload(&tmpIdVersion, TRUE) != IX_SUCCESS)
    {
	    printf("\nNPE download failed");
	    return(IX_FAIL);
    }
    return(IX_SUCCESS);
}

/*
 * Function definition : ixDmaAccCodeletShow()
 * See header file for documentation.
 */

void
ixDmaAccCodeletShow(void)
{
    UINT32 i;       /* Counter for memory dump loop */

    /* Invalidate cache for source block before a read to ensure consistency */
    IX_ACC_DATA_CACHE_INVALIDATE(ixDmaAccCodeletSrcBlock,IX_DMA_CODELET_MEMDUMPSIZE);

    /* Show memory dump for source test array */
    for(i = 0; i < IX_DMA_CODELET_MEMDUMPSIZE; i++)
    {
        if ( i%16 == 0 )       /* Show newline and address for every 16 bytes */
        {
            printf ("\n Src Addr %02x : ", (UINT32) &ixDmaAccCodeletSrcBlock[i]);
        }
        printf ("%02x ", (UINT32) ixDmaAccCodeletSrcBlock[i]);  /* Show raw memory byte */
    }

    /* Invalidate cache for destination block before a read access to ensure consistency */
    IX_ACC_DATA_CACHE_INVALIDATE(ixDmaAccCodeletDestBlock,IX_DMA_CODELET_MEMDUMPSIZE);

    /* Show memory dump for destination test array */
    for(i = 0; i < IX_DMA_CODELET_MEMDUMPSIZE; i++)
    {
        if ( i%16 == 0 )       /* Show newline and address for every 16 bytes */
        {
            printf ("\n Dst Addr %02x : ", (UINT32) &ixDmaAccCodeletDestBlock[i]);
        }
        printf ("%02x ", (UINT32) ixDmaAccCodeletDestBlock[i]);  /* Show raw memory byte */
    }
    printf ("\n");
    return;
}

/*
 * Function definition : ixDmaAccCodeletCallback()
 * See header file for documentation.
 */
void
ixDmaAccCodeletCallback(IX_STATUS status)
{
    /* Status message to indicate if the transfer had succeeded */
    if (IX_SUCCESS != status)
    {
        printf("\nDma Transfer Failed");
    }
    else
    {
        printf("\nDma Transfer Succeeded");
    }
    printf ("\n\n%d byte memory dump after transfer :", IX_DMA_CODELET_MEMDUMPSIZE );
    /* Memory dump to show result of the current transfer */
    ixDmaAccCodeletShow();
    return;
}

/*
 * Function definition : ixDmaAccCodeletTestPatternReset()
 * See header file for documentation.
 */
PRIVATE void ixDmaAccCodeletTestPatternReset(void)
{
    /* Counter for Test For-Loops */
    UINT32 i,j;

    /* Test pattern is loaded into a temporary array here */
    UINT8 testPattern[IX_DMA_CODELET_TESTPATTERN_LENGTH]
          = IX_DMA_CODELET_TESTPATTERN_LIST;

    /* Flush the cache before performing a write to memory block */
    IX_ACC_DATA_CACHE_FLUSH(ixDmaAccCodeletSrcBlock,IX_DMA_CODELET_TESTPATTERN_LENGTH);

    /* Initialise the values in the Source Array */
    for ( i = 0;
          i < IX_DMA_CODELET_TEST_MAXLENGTH;
          i += IX_DMA_CODELET_TESTPATTERN_LENGTH)
    {
        for ( j = 0;
              j < IX_DMA_CODELET_TESTPATTERN_LENGTH;
              j += 1)
        {
            ixDmaAccCodeletSrcBlock[i+j] = testPattern[j];
        }
    }

    /* Flush the cache before performing a write to memory block */
    IX_ACC_DATA_CACHE_FLUSH(ixDmaAccCodeletDestBlock,IX_DMA_CODELET_TESTPATTERN_LENGTH);

    /* Initialise the values in the Destination Array */
    memset( ixDmaAccCodeletDestBlock,
            0xFF,
            IX_DMA_CODELET_TEST_MAXLENGTH);
    return;
}

/*
 * Function definition : ixDmaAccCodeletTestPerform()
 * See header file for documentation.
 */
void ixDmaAccCodeletTestPerform( UINT16 transferLength,
                                 IxDmaTransferMode transferMode,
                                 IxDmaAddressingMode addressingMode,
                                 IxDmaTransferWidth transferWidth)
{
    /* Initialise Source Address to ixDmaAccCodeletSrcBlock */
    UINT32 sourceAddr = (UINT32) ixDmaAccCodeletSrcBlock;

    /* Initialise Destination Address to ixDmaAccCodeletDestBlock */
    UINT32 destinationAddr = (UINT32) ixDmaAccCodeletDestBlock;

    /* Initialise Test Patterns */
    ixDmaAccCodeletTestPatternReset();

    printf ("\nStarting transfer with the following parameters");
    printf ("\nSource Address      : 0x%x", sourceAddr);
    printf ("\nDestination Address : 0x%x", destinationAddr);
    printf ("\nTransfer Length     : %d", transferLength);
    printf ("\nTransfer Mode       : %d", transferMode);
    printf ("\nAddressing Mode     : %d", addressingMode);
    printf ("\nTransfer Width      : %d", transferWidth);
    printf ("\n\n%d byte memory dump before transfer :", IX_DMA_CODELET_MEMDUMPSIZE );

    /* Show memory dump before transfer */
    ixDmaAccCodeletShow();
    /* Perform Dma Transfer */
    /* Callback function will show Memory dump after transfer */
    /* Callback function will clear status flag for transfer status */
    if ( IX_DMA_SUCCESS != ixDmaAccDmaTransfer( ixDmaAccCodeletCallback,
                                                sourceAddr,
                                                destinationAddr,
                                                transferLength,
                                                transferMode,
                                                addressingMode,
                                                transferWidth ) )
    {
        printf("\nDma Transfer Request Failed");
        printf("\nUsage :");
        printf("\n ixDmaAccCodeletTestPerform ");
        printf("(transferLength,transferMode,addressMode,transferWidth) ");
        printf("\n");
        printf("\n  transferLength : 1 to 65535 bytes         ");
        printf("\n");
        printf("\n  transferMode   : 0 = Copy and Clear Source   2 = Copy and Byte Swap");
        printf("\n                   1 = Copy                    3 = Copy and Reverse");
        printf("\n");
        printf("\n  addressMode    : 0 = Increment Source Addr and Increment Destination Addr");
        printf("\n                   1 = Increment Source Addr and Fix Destination Addr");
        printf("\n                   2 = Fix Source Addr and Increment Destination Addr");
        printf("\n");
        printf("\n  transferWidth  : 0 = 32 bit src, 32 bit dest   8  =  8 bit src,  8 bit dest");
        printf("\n                   1 = 32 bit src, 16 bit dest   9  =  8 bit src,  Burst dest");
        printf("\n                   2 = 32 bit src,  8 bit dest   10 = 16 bit src,  Burst dest");
        printf("\n                   3 = 16 bit src, 32 bit dest   11 = 32 bit src,  Burst dest");
        printf("\n                   4 = 16 bit src, 16 bit dest   12 =  Burst src,  8 bit dest");
        printf("\n                   5 = 16 bit src,  8 bit dest   13 =  Burst src, 16 bit dest");
        printf("\n                   6 =  8 bit src, 32 bit dest   14 =  Burst src, 32 bit dest");
        printf("\n                   7 =  8 bit src, 16 bit dest   15 =  Burst src,  Burst dest");
        printf("\n");
    }
    return;
}

