 /**
 * @file IxDmaAccCodelet.h
 *
 * @date 18 November 2002
 *
 * @brief This file contains the interface for the Dma Access Codelet.
 *
 * -- Intel Copyright Notice --
 *
 * @par
 * INTEL CONFIDENTIAL
 *
 * @par
 * Copyright 2002 Intel Corporation All Rights Reserved.
 *
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.  The Material
 * contains trade secrets and proprietary and confidential information of
 * Intel or its suppliers and licensors.  The Material is protected by
 * worldwide copyright and trade secret laws and treaty provisions. No
 * part of the Material may be used, copied, reproduced, modified,
 * published, uploaded, posted, transmitted, distributed, or disclosed in
 * any way without Intel's prior express written permission.
 *
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel or otherwise.  Any license under
 * such intellectual property rights must be express and approved by
 * Intel in writing.
 *
 * -- End Intel Copyright Notice --
 *
 */

/**
 * @defgroup Codelets IXP4xx Codelets
 *
 * @brief IXP4xx demo codelets
 *
 * @{
 */

/**
 * @ingroup Codelets
 *
 * @defgroup IxDmaAccCodelet IXP4xx DMA Access Codelet (IxAal5Demo) API
 *
 * @brief IXP4xx demo DMA Access component API
 *
 * This file contains the interface of the Dma Access Codelet.
 *
 * The API ixDmaAccCodeletTestPerform() allows the user to perform a
 * Dma transfer of block size 0 to 65535 bytes between two locations
 * in the SRAM. The user can specify any combination of the following modes.   
 *
 *            DMA Transfer Modes 
 *	         1. Copy
 *	      	 2. Copy and Clear Source
 *	      	 3. Copy with Bytes Swap
 *	      	 4. Copy with Bytes Reversed
 *
 *	      DMA Addressing Modes
 *	      	 1. Incremental Source to Incremental Destination Addressess
 *	         2. Fixed Source to Incremental Destination Addressess
 *	      	 3. Incremental Source to Fixed Destination Addressess
 *
 *	      DMA Transfer Widths
 *	       	 1. 32-bit Transfer
 *	      	 2. 16-bit Transfer
 *	      	 3. 8-bit Transfer
 *	      	 4. Burst Transfer
 *
 *  NOTE : The user must initialise the system with ixDmaAccCodeletInit prior to 
 *         calling the function ixDmaAccCodeletiTestPerform()
 *
 * @{
 */

#ifndef IXDMAACCCODELET_H
#define IXDMAACCCODELET_H

#include "IxOsBuffMgt.h"
#include "IxDmaAcc.h"
/*
 * Prototypes for interface functions.
 */

/**
 * @fn IX_STATUS ixDmaAccCodeletInit(IxDmaNpeId npeId)
 *
 * This function is the entry point to the Dma Access codelet.
 * It must be called before any of the demos as it initialises the necessary
 * components.
 *
 * @param npeID             Id for the Npe to setup for Dma function
 *                   @li    IX_DMA_NPEID_A for Npe A
 *                   @li    IX_DMA_NPEID_B for Npe B
 *                   @li    IX_DMA_NPEID_C for Npe C
 *
 * The initialisation sequence is as follows:
 *
 *	 Initialise Queue manager
 *	 Start Queue manager dispatcher loop
 *	 Download NPE microcode
 *	 Start NPEs
 *	 Initialise Dma Access component
 *
 * @return IX_SUCCESS - Codelet successfully initialised
 * @return IX_FAIL - Error initialising codelet
 */

IX_STATUS
ixDmaAccCodeletInit(IxDmaNpeId npeId);

/**
 * @fn void ixDmaAccCodeletTestPerform( UINT16 transferLength, IxDmaTransferMode transferMode,IxDmaAddressingMode addressingMode, IxDmaTransferWidth transferWidth)
 *
 * This function performs a dma transfer with parameters provided by the client 
 * This function can only function after ixDmaAccCodeletInit is performed.
 *
 * @param transferLength 1 to 65535 bytes         
 * 
 * @param transferMode 
 *                  @li 0 = Copy 
 *                  @li 1 = Copy and Clear Source  
 *                  @li 2 = Copy and Byte Swap  
 *                  @li 3 = Copy and Reverse
 * 
 * @param addressMode 
 *                  @li 0 = Increment Source Addr and Increment Destination Addr");
 *                  @li 1 = Increment Source Addr and Fix Destination Addr");
 *                  @li 2 = Fix Source Addr and Increment Destination Addr");
 * 
 * @param transferWidth  : 
 *                  @li  0 = 32 bit src, 32 bit dest   
 *                  @li  1 = 32 bit src, 16 bit dest   
 *                  @li  2 = 32 bit src,  8 bit dest   
 *                  @li  3 = 16 bit src, 32 bit dest   
 *                  @li  4 = 16 bit src, 16 bit dest   
 *                  @li  5 = 16 bit src,  8 bit dest   
 *                  @li  6 =  8 bit src, 32 bit dest   
 *                  @li  7 =  8 bit src, 16 bit dest   
 *                  @li  8 =  8 bit src,  8 bit dest
 *                  @li  9 =  8 bit src,  Burst dest
 *                  @li 10 = 16 bit src,  Burst dest
 *                  @li 11 = 32 bit src,  Burst dest
 *                  @li 12 =  Burst src,  8 bit dest
 *                  @li 13 =  Burst src, 16 bit dest
 *                  @li 14 =  Burst src, 32 bit dest
 *                  @li 15 =  Burst src,  Burst dest
 *
 * @return none 
 */

void ixDmaAccCodeletTestPerform( UINT16 transferLength,
                                 IxDmaTransferMode transferMode,
                                 IxDmaAddressingMode addressingMode,
                                 IxDmaTransferWidth transferWidth);

#endif /* IXDMAACCCODELET_H */

/** @} defgroup IxDmaAccCodelet*/

/** @} defgroup Codelet*/
