/**
 * @file IxEthAccCodelet.c
 *
 * @author Intel Corporation
 * @date 22 April 2002
 *
 * @brief This file contains the implementation of the Ethernet Access Codelet.
 *
 * -- Intel Copyright Notice --
 *
 * @par
 * INTEL CONFIDENTIAL
 *
 * @par
 * Copyright 2002 Intel Corporation All Rights Reserved.
 *
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.  The Material
 * contains trade secrets and proprietary and confidential information of
 * Intel or its suppliers and licensors.  The Material is protected by
 * worldwide copyright and trade secret laws and treaty provisions. No
 * part of the Material may be used, copied, reproduced, modified,
 * published, uploaded, posted, transmitted, distributed, or disclosed in
 * any way without Intel's prior express written permission.
 *
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel or otherwise.  Any license under
 * such intellectual property rights must be express and approved by
 * Intel in writing.
 *
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * -- End Intel Copyright Notice --
*/

/*
 * Put the system defined include files required.
 */

#include <stdio.h>
#include <stdlib.h>

#include <vxWorks.h>
#include <taskLib.h>
#include <sysLib.h>

/*
 * Put the user defined include files required.
 */

#include "IxTypes.h"
#include "IxQMgr.h"
#include "IxNpeDl.h"
#include "IxNpeMh.h"
#include "IxOsServices.h"
#include "IxOsCacheMMU.h"
#include "ix_ossl.h"
#include "IxEthDB.h"
#include "IxEthAccCodelet.h"
#include "IxNpeMicrocode.h"

#include "IxDemoUtils.h"

#include "ixp425.h"	/* Chip level definitions required */
#include "IxEthAccCodelet_p.h"

#ifdef IX_ETHACC_CODELET_USE_NVRAM_MAC 
#include "config.h"
#endif

#ifdef __linux
/* use interrupts for performances */
#define IX_ETH_CODELET_QMGR_DISPATCH_MODE TRUE
#else
/* use polled mode for performances */
#define IX_ETH_CODELET_QMGR_DISPATCH_MODE FALSE
#endif

/*
 * defines and macros used in this file.
 */
/**
 * @def NPE_B_BUILD
 * @brief NPE B build to load for the required features
 *
 * 1 means Ethernet and Fast Path
 */
#define NPE_B_BUILD 1

/**
 * @def NPE_C_BUILD
 * @brief NPE C build to load for the required features
 *
 * 0 means Ethernet
 */
#define NPE_C_BUILD 0

/*
 * Variable declarations global to this file only.
 */

PRIVATE IX_MBUF *ixEthAccCodeletBufPool;

PRIVATE UINT8 *ixEthAccCodeletBufData;

PRIVATE UINT8 compData[IX_ETHACC_CODELET_PCK_LEN];

PRIVATE IX_MBUF *ixEthAccCodeletFreeBufQ;

PRIVATE volatile UINT32 memPoolFreeCnt;

PRIVATE UINT32 phyAddresses[IX_ETHACC_CODELET_MAX_PORT];
PRIVATE UINT32 maxPhyNo;

PRIVATE IxEthAccCodeletPhyConf phyConf = 
{
    TRUE,	/* 100 Mbits */
    TRUE,	/* Full duplex */
    TRUE	/* Autonegotiate */
};

PRIVATE struct _IxEthAccCodeletStats
{
    UINT32 rxCount;
    UINT32 txCount;
} IxEthAccCodeletStats[IX_ETHACC_CODELET_MAX_PORT];
  
PRIVATE BOOL ixEthAccCodeletInitialised = FALSE;
PRIVATE BOOL ixEthAccCodeletNpeInitialised = FALSE;

PRIVATE volatile BOOL ixEthAccCodeletPollEnabled = FALSE;

/*
 * Static function prototypes.
 */

PRIVATE IX_STATUS ixEthAccCodeletDispatcherStart(BOOL useInterrupt);
PRIVATE void ixEthAccCodeletDispatcherPoll(void* arg, void** ptrRetObj);
PRIVATE void ixEthAccCodeletDBMaintenanceTask(void* arg, void** ptrRetObj);
PRIVATE IX_STATUS ixEthAccCodeletPhyInit(void);
PRIVATE IX_STATUS ixEthAccCodeletMemPoolInit(void);
PRIVATE IX_STATUS ixEthAccCodeletMacSet(void);
PRIVATE IX_STATUS ixEthAccCodeletRecoverBuffers(void);
PRIVATE IX_STATUS ixEthAccCodeletReplenishBuffers(IxEthAccPortId portNo, UINT32 num); 
PRIVATE IX_STATUS ixEthAccCodeletLoop(void);
PRIVATE void ixEthAccCodeletStatsPollTask(void* arg, void** ptrRetObj);

/* Ethernet callbacks */

PRIVATE void ixEthAccCodeletRxSinkCB(UINT32 cbTag, IX_MBUF* mBufPtr, IxEthAccPortId portId);
PRIVATE void ixEthAccCodeletTxGenRxSinkLoopbackRxCB(UINT32 cbTag, IX_MBUF* mBufPtr, IxEthAccPortId portId);
PRIVATE void ixEthAccCodeletTxGenRxSinkLoopbackTxCB(UINT32 cbTag, IX_MBUF* mBufPtr);
PRIVATE void ixEthAccCodeletSwLoopbackRxCB(UINT32 cbTag, IX_MBUF* mBufPtr, IxEthAccPortId portId);
PRIVATE void ixEthAccCodeletSwLoopbackTxCB(UINT32 cbTag, IX_MBUF* mBufPtr);
PRIVATE void ixEthAccCodeletBridgeRxCB(UINT32 cbTag, IX_MBUF* mBufPtr, IxEthAccPortId portId);
PRIVATE void ixEthAccCodeletBridgeTxCB(UINT32 cbTag, IX_MBUF* mBufPtr);
PRIVATE void ixEthAccCodeletMemPoolFreeRxCB(UINT32 cbTag, IX_MBUF* mBufPtr, IxEthAccPortId portId);
PRIVATE void ixEthAccCodeletMemPoolFreeTxCB(UINT32 cbTag, IX_MBUF* mBufPtr);

/*
 * Function definition: ixEthAccCodeletInit()
 *
 * See header file for documentation.
 */

IX_STATUS ixEthAccCodeletInit(void)
{  
    unsigned int statstid;

    if(ixEthAccCodeletInitialised) 
    {
	printf("Ethernet codelet already initialised\n");
	return(IX_SUCCESS);
    }

    /* Initialise Queue Manager */
    printf("Initialising Queue Manager...\n");
    if (ixQMgrInit() != IX_SUCCESS)
    {
	printf("Error initialising queue manager!\n");
	return (IX_FAIL);
    }

    /* Start the Queue Manager dispatcher */   
    if(ixEthAccCodeletDispatcherStart(IX_ETH_CODELET_QMGR_DISPATCH_MODE) != IX_SUCCESS)
    {
	printf("Error starting queue manager dispatch loop!\n");
	return (IX_FAIL);
    }

    /* Initialise NPEs */
    printf("Initialising NPEs...\n");
    if(ixDemoUtilsNpeImageDownload(IX_NPEDL_NPEID_NPEB, NPE_B_BUILD) != IX_SUCCESS)
    {
	printf("Error initialising NPE B!\n");
	return (IX_FAIL);
    }
 
    if(ixDemoUtilsNpeImageDownload(IX_NPEDL_NPEID_NPEC, NPE_C_BUILD) != IX_SUCCESS)
    {
	printf("Error initialising NPE C!\n");
	return (IX_FAIL);
    }

    /* Initialise NPE Message handler */
    printf("\nStarting NPE message handler...\n");
    if(ixNpeMhInitialize(IX_NPEMH_NPEINTERRUPTS_YES) != IX_SUCCESS)
    {
	printf("Error initialising NPE Message handler!\n");
	return (IX_FAIL);
    }

    /* Start NPEs */
    printf("Starting NPEs...\n");
    if(ixNpeDlNpeExecutionStart(IX_NPEDL_NPEID_NPEB) != IX_SUCCESS)
    {
	printf("Error starting NPE B!\n");
	return (IX_FAIL);
    }

    if(ixNpeDlNpeExecutionStart(IX_NPEDL_NPEID_NPEC) != IX_SUCCESS)
    {
	printf("Error starting NPE C!\n");
	return (IX_FAIL);
    }

    ixEthAccCodeletNpeInitialised = TRUE;

    /***********************************************************************
     *
     * System initialisation done. Now initialise Ethernet Access components. 
     *
     ***********************************************************************/

    if (ixEthAccInit() != IX_ETH_ACC_SUCCESS)
    {
	printf("Error initialising Ethernet access driver!\n");
	return (IX_FAIL);
    }

    /* Initialise Ethernet ports */
    if(ixEthAccPortInit(IX_ETH_PORT_1) != IX_ETH_ACC_SUCCESS)
    {
	printf("Error initialising Ethernet port 1!\n");	
	return (IX_FAIL);
    }

    if(ixEthAccPortInit(IX_ETH_PORT_2) != IX_ETH_ACC_SUCCESS)
    {
	printf("Error initialising Ethernet port 2!\n");	
	return (IX_FAIL);
    }

    /* Find and initialise all available PHYs */
    if(ixEthAccCodeletPhyInit() != IX_SUCCESS)
    {
	printf("Error initialising Ethernet phy(s)!\n");
	return (IX_FAIL);
    }

    /* Program MAC addresses for available PHYs */
    if(ixEthAccCodeletMacSet() != IX_SUCCESS)
    {
	printf("Error programming MAC address for Ethernet phy(s)!\n");
	return (IX_FAIL);
    }

    /* Set ports to promiscuous mode  */
    if(ixEthAccPortPromiscuousModeSet(IX_ETH_PORT_1) != IX_ETH_ACC_SUCCESS)
    {
	printf("Error setting promiscuous mode for port 1\n");
	return (IX_FAIL);
    }

    if(ixEthAccPortPromiscuousModeSet(IX_ETH_PORT_2) != IX_ETH_ACC_SUCCESS)
    {
	printf("Error setting promiscuous mode for port 2\n");
	return (IX_FAIL);
    }

    /* Initialise MBUF pool */
    if(ixEthAccCodeletMemPoolInit() != IX_SUCCESS)
    {
	printf("Error initialising mBuf pool\n");
	return (IX_FAIL);
    }

    /* Set scheduling discipline */
    ixEthAccTxSchedulingDisciplineSet(IX_ETH_PORT_1, FIFO_NO_PRIORITY);
    ixEthAccTxSchedulingDisciplineSet(IX_ETH_PORT_2, FIFO_NO_PRIORITY);

    ixEthAccCodeletPollEnabled = FALSE;
    if (ix_ossl_thread_create((ix_ossl_thread_entry_point_t)
                              ixEthAccCodeletStatsPollTask,
                              NULL,
                              &statstid) != IX_OSSL_ERROR_SUCCESS)
    {
        printf("Error spawning stats task\n");
        return (IX_FAIL);
    }

    ixEthAccCodeletInitialised = TRUE;
    return (IX_SUCCESS);
}

/*
 * Function definition: ixEthAccCodeletLinkUpCheck()
 *
 * Check the phys are ready
 */

IX_STATUS ixEthAccCodeletLinkUpCheck(unsigned int phyNo)
{
   IX_STATUS returnStatus = IX_SUCCESS;
   BOOL fullDuplex;
   BOOL linkUp;
   BOOL speed;
   BOOL autoneg;

   ixEthAccMiiLinkStatus(phyAddresses[phyNo], &linkUp, &speed, &fullDuplex, &autoneg);
   if (linkUp == FALSE)
   {
          unsigned int retry = 20; /* 20 retries */
          printf("Wait for PHY %u to be ready ...", phyNo);
          while (linkUp == FALSE && retry-- > 0)
          {
            ixOsServTaskSleep(100);  /* 100 milliseconds */
            ixEthAccMiiLinkStatus(phyAddresses[phyNo], &linkUp, &speed, &fullDuplex, &autoneg);
          }
          if (linkUp == FALSE) 
          {
            returnStatus = IX_FAIL;
          }
   }
   /* return fail if one of the links is not up */
   return returnStatus;
}

/*
 * Function definition: ixEthAccCodeletStart()
 *
 * See header file for documentation.
 */

IX_STATUS ixEthAccCodeletStart(IxEthAccPortRxCallback port1RxCB,
			       IxEthAccPortRxCallback port2RxCB,
			       IxEthAccPortTxDoneCallback port1TxDoneCB,
			       IxEthAccPortTxDoneCallback port2TxDoneCB)
{
    unsigned int phyNo;

    IX_ETHACC_IS_CODELET_INITIALISED();

    /* Clear stats */
    memset(&IxEthAccCodeletStats[0], 0, sizeof(IxEthAccCodeletStats[0]));
    memset(&IxEthAccCodeletStats[1], 0, sizeof(IxEthAccCodeletStats[1]));

    /* check links are up and continue, even on link failure */
   for(phyNo=0; phyNo<maxPhyNo; phyNo++)
   {
      (void)ixEthAccCodeletLinkUpCheck(phyNo);
   }

    if (port1RxCB != NULL)
    {
	if(ixEthAccPortRxCallbackRegister(IX_ETH_PORT_1,
					  port1RxCB,
					  IX_ETH_PORT_1) != IX_ETH_ACC_SUCCESS)
	{
	    printf("Failed to register Rx callback for port 1\n");
	    return (IX_FAIL);
	}
    }

    if (port2RxCB != NULL)
    {
	if(ixEthAccPortRxCallbackRegister(IX_ETH_PORT_2,
					  port2RxCB,
					  IX_ETH_PORT_2) != IX_ETH_ACC_SUCCESS)
	{
	    printf("Failed to register Rx callback for port 2\n");
	    return (IX_FAIL);
	}
    }

    if(port1TxDoneCB != NULL)
    {
	if (ixEthAccPortTxDoneCallbackRegister(IX_ETH_PORT_1,
					       port1TxDoneCB,
					       IX_ETH_PORT_1) != IX_ETH_ACC_SUCCESS)
	{
	    printf("Failed to register Tx done callback for port 1\n");
	    return (IX_FAIL);
	}
    }

    if(port2TxDoneCB != NULL)
    {
	if(ixEthAccPortTxDoneCallbackRegister(IX_ETH_PORT_2,
					      port2TxDoneCB,
					      IX_ETH_PORT_2) != IX_ETH_ACC_SUCCESS)
	{
	    printf("Failed to register Tx done callback for port 2\n");
	    return (IX_FAIL);
	}
    }

    if (ixEthAccCodeletNpeInitialised == FALSE)
    {
        printf("Initialising NPEs...\n");
        if(ixDemoUtilsNpeImageDownload(IX_NPEDL_NPEID_NPEB, NPE_B_BUILD) != IX_SUCCESS)
        {
        printf("Error initialising NPE B!\n");
        return (IX_FAIL);
        }

        if(ixDemoUtilsNpeImageDownload(IX_NPEDL_NPEID_NPEC, NPE_C_BUILD) != IX_SUCCESS)
        {
        printf("Error initialising NPE C!\n");
        return (IX_FAIL);
        }
	
	/* Start NPEs */
	printf("Starting NPEs...\n");
	if(ixNpeDlNpeExecutionStart(IX_NPEDL_NPEID_NPEB) != IX_SUCCESS)
	{
	    printf("Error starting NPE B!\n");
	    return (IX_FAIL);
	}
	
	if(ixNpeDlNpeExecutionStart(IX_NPEDL_NPEID_NPEC) != IX_SUCCESS)
	{
	    printf("Error starting NPE C!\n");
	    return (IX_FAIL);
	}
	ixEthAccCodeletNpeInitialised = TRUE;
    }
 
    if(ixEthAccCodeletReplenishBuffers(IX_ETH_PORT_1, 
				       IX_ETHACC_CODELET_RX_MBUF_POOL_SIZE) 
       != IX_SUCCESS)
    {
	printf("Error replenish port 1\n");
	return (IX_FAIL);
    }

    if(ixEthAccCodeletReplenishBuffers(IX_ETH_PORT_2,
				       IX_ETHACC_CODELET_RX_MBUF_POOL_SIZE)
       != IX_SUCCESS)
    {
	printf("Error replenish port 2\n");
	return (IX_FAIL);
    }

#ifdef IX_ETHACC_CODELET_RX_FCS_STRIP
    if(ixEthAccPortRxFrameAppendFCSDisable(IX_ETH_PORT_1) != IX_ETH_ACC_SUCCESS)
    {
	printf("Error disabling FCS for port 1\n");
	return (IX_FAIL);
    }

    if(ixEthAccPortRxFrameAppendFCSDisable(IX_ETH_PORT_2) != IX_ETH_ACC_SUCCESS)
    {
	printf("Error disabling FCS for port 2\n");
	return (IX_FAIL);
    }
#else
    if(ixEthAccPortRxFrameAppendFCSEnable(IX_ETH_PORT_1) != IX_ETH_ACC_SUCCESS)
    {
	printf("Error enabling FCS for port 1\n");
	return (IX_FAIL);
    }

    if(ixEthAccPortRxFrameAppendFCSEnable(IX_ETH_PORT_2) != IX_ETH_ACC_SUCCESS)
    {
	printf("Error enabling FCS for port 2\n");
	return (IX_FAIL);
    }
#endif

#ifdef IX_ETHACC_CODELET_TX_FCS_APPEND
    if(ixEthAccPortTxFrameAppendFCSEnable(IX_ETH_PORT_1) != IX_ETH_ACC_SUCCESS)
    {
	printf("Error enabling FCS for port 1\n");
	return (IX_FAIL);
    }

    if(ixEthAccPortTxFrameAppendFCSEnable(IX_ETH_PORT_2) != IX_ETH_ACC_SUCCESS)
    {
	printf("Error enabling FCS for port 2\n");
	return (IX_FAIL);
    }
#else
    if(ixEthAccPortTxFrameAppendFCSDisable(IX_ETH_PORT_1) != IX_ETH_ACC_SUCCESS)
    {
	printf("Error disabling FCS for port 1\n");
	return (IX_FAIL);
    }

    if(ixEthAccPortTxFrameAppendFCSDisable(IX_ETH_PORT_2) != IX_ETH_ACC_SUCCESS)
    {
	printf("Error disabling FCS for port 2\n");
	return (IX_FAIL);
    }
#endif

    /* Enable ports */
    if(ixEthAccPortEnable(IX_ETH_PORT_1) != IX_ETH_ACC_SUCCESS)
    {
	printf("Error enabling port 1\n");
	return (IX_FAIL);
    }

    if(ixEthAccPortEnable(IX_ETH_PORT_2) != IX_ETH_ACC_SUCCESS)
    {
	printf("Error enabling port 2\n");
	return (IX_FAIL);
    }

    return IX_SUCCESS;
}

IX_STATUS ixEthAccCodeletStop(IxEthAccPortRxCallback port1RxCB,
			       IxEthAccPortRxCallback port2RxCB,
			       IxEthAccPortTxDoneCallback port1TxDoneCB,
			       IxEthAccPortTxDoneCallback port2TxDoneCB)
{
    if (port1RxCB != NULL)
    {
	if(ixEthAccPortRxCallbackRegister(IX_ETH_PORT_1,
					  port1RxCB,
					  IX_ETH_PORT_1) != IX_ETH_ACC_SUCCESS)
	{
	    printf("Failed to register Rx callback for port 1\n");
	    return (IX_FAIL);
	}
    }

    if (port2RxCB != NULL)
    {
	if(ixEthAccPortRxCallbackRegister(IX_ETH_PORT_2,
					  port2RxCB,
					  IX_ETH_PORT_2) != IX_ETH_ACC_SUCCESS)
	{
	    printf("Failed to register Rx callback for port 2\n");
	    return (IX_FAIL);
	}
    }

    if(port1TxDoneCB != NULL)
    {
	if (ixEthAccPortTxDoneCallbackRegister(IX_ETH_PORT_1,
					       port1TxDoneCB,
					       IX_ETH_PORT_1) != IX_ETH_ACC_SUCCESS)
	{
	    printf("Failed to register Tx done callback for port 1\n");
	    return (IX_FAIL);
	}
    }

    if(port2TxDoneCB != NULL)
    {
	if(ixEthAccPortTxDoneCallbackRegister(IX_ETH_PORT_2,
					      port2TxDoneCB,
					      IX_ETH_PORT_2) != IX_ETH_ACC_SUCCESS)
	{
	    printf("Failed to register Tx done callback for port 2\n");
	    return (IX_FAIL);
	}
    }

    if (ixEthAccCodeletNpeInitialised == TRUE)
    {
	/* Disable ports - unused MBufs will be returned */
	if(ixEthAccPortDisable(IX_ETH_PORT_2) != IX_ETH_ACC_SUCCESS)
	{
	    printf("Error disabling port 2\n");
	    return (IX_FAIL);
	}
	
	if(ixEthAccPortDisable(IX_ETH_PORT_1) != IX_ETH_ACC_SUCCESS)
	{
	    printf("Error disabling port 1\n");
	    return (IX_FAIL);
	}
    
	ixEthAccCodeletNpeInitialised = FALSE;
    }

    return IX_SUCCESS;
}

/*
 * Function definition: ixEthAccCodeletRxSink()
 *
 * See header file for documentation.
 */

IX_STATUS ixEthAccCodeletRxSink(void)
{
    IX_ETHACC_IS_CODELET_INITIALISED();

    if (ixEthAccCodeletStart(ixEthAccCodeletRxSinkCB, 
			     ixEthAccCodeletRxSinkCB,
			     ixEthAccCodeletMemPoolFreeTxCB,
                             ixEthAccCodeletMemPoolFreeTxCB
			     ) != IX_ETH_ACC_SUCCESS)
    {
	printf("Failed to start the RX sink demo\n");
	return IX_FAIL;
    }

    printf("Rx Sink demo started. Begin sending packets in from external source\n");

    if(ixEthAccCodeletLoop() != IX_SUCCESS)
    {
	return (IX_FAIL);
    }
    
    /* Recover buffers */
    if(ixEthAccCodeletRecoverBuffers() != IX_SUCCESS)
    {
	printf("Error recovering buffers for Rx Sink demo\n");
	return (IX_FAIL);
    }

    return (IX_SUCCESS);
}


/*
 * Function definition: ixEthAccCodeletSwLoopback()
 *
 * See header file for documentation.
 */

IX_STATUS ixEthAccCodeletSwLoopback(void)
{
    IX_ETHACC_IS_CODELET_INITIALISED();

    if ( ixEthAccCodeletStart(ixEthAccCodeletSwLoopbackRxCB, 
			      ixEthAccCodeletSwLoopbackRxCB, 
			      ixEthAccCodeletSwLoopbackTxCB,
			      ixEthAccCodeletSwLoopbackTxCB) != IX_SUCCESS)
    {
	printf("Failed to start the Sw Loopback Demo\n");
	return (IX_FAIL);
    }


    printf("Software loopback successfully started. Begin sending packets from external source\n");  

    if(ixEthAccCodeletLoop() != IX_SUCCESS)
    {
	return (IX_FAIL);
    }

    /* Recover buffers */
    if(ixEthAccCodeletRecoverBuffers() != IX_SUCCESS)
    {
	printf("Error recovering buffers for Sw Loopback demo\n");
	return (IX_FAIL);
    }    

    return (IX_SUCCESS);
}


/*
 * Function definition: ixEthAccCodeletTxGenRxSinkLoopback()
 *
 * See header file for documentation.
 */

IX_STATUS ixEthAccCodeletTxGenRxSinkLoopback(void)
{
    IX_MBUF *mBufPtr;
    UINT32 numBufs;

    IX_ETHACC_IS_CODELET_INITIALISED();

    if (ixEthAccCodeletStart(ixEthAccCodeletMemPoolFreeRxCB, 
			     ixEthAccCodeletTxGenRxSinkLoopbackRxCB,
			     ixEthAccCodeletTxGenRxSinkLoopbackTxCB,
                             ixEthAccCodeletMemPoolFreeTxCB
			     ) != IX_ETH_ACC_SUCCESS)
    {
	printf("Failed to start the Tx Gen RX sink demo\n");
	return IX_FAIL;
    }

    /* Generate our random data for the payload */
    IX_ETHACC_CODELET_DATAGEN(compData);

    /* Now start the loopback by transmitting the first few frames */
    for (numBufs=0; numBufs<IX_ETHACC_CODELET_TXGEN_PCKS; numBufs++)
    {
	IX_ETHACC_CODELET_REMOVE_MBUF_FROM_Q_HEAD(ixEthAccCodeletFreeBufQ, mBufPtr);
	
	if (mBufPtr == NULL)
	{
	    printf("Buffer queue empty. Not enough free buffers to transmit in TxGen-RxSink Loopback!\n");
	    return (IX_FAIL);
	}

	IX_MBUF_MLEN(mBufPtr) = IX_MBUF_PKT_LEN(mBufPtr) = IX_ETHACC_CODELET_TXGEN_PCK_LEN;
	memcpy(IX_MBUF_MDATA(mBufPtr), &compData[0], IX_ETHACC_CODELET_TXGEN_PCK_LEN);
	IX_ACC_DATA_CACHE_FLUSH(IX_MBUF_MDATA(mBufPtr), IX_MBUF_MLEN(mBufPtr));

	if(ixEthAccPortTxFrameSubmit(IX_ETH_PORT_1, mBufPtr,
				     IX_ETH_ACC_TX_DEFAULT_PRIORITY) != IX_ETH_ACC_SUCCESS)
	{
	    printf("Error starting TxGen-RxSink Loopback\n");
	    return (IX_FAIL);
	}
    }

    printf("Port %d Tx pool has %d buffers\n", IX_ETH_PORT_1 + 1, numBufs);
    printf("TxGen-RxSink demo successfully started\n");

    if(ixEthAccCodeletLoop() != IX_SUCCESS)
    {
	return (IX_FAIL);
    }
    
    /* Recover buffers */
    if(ixEthAccCodeletRecoverBuffers() != IX_SUCCESS)
    {
	printf("Error recovering buffers for TxGen-RxSink Loopback demo\n");
	return (IX_FAIL);
    } 
    
    return (IX_SUCCESS);
}


/*
 * Function definition: ixEthAccCodeletBridge()
 *
 * See header file for documentation.
 */

IX_STATUS ixEthAccCodeletBridge(void)
{
    IX_ETHACC_IS_CODELET_INITIALISED();

    if ( ixEthAccCodeletStart(ixEthAccCodeletBridgeRxCB, 
			      ixEthAccCodeletBridgeRxCB, 
			      ixEthAccCodeletBridgeTxCB,
			      ixEthAccCodeletBridgeTxCB) != IX_SUCCESS)
    {
	printf("Failed to start the Bridge Demo\n");
	return (IX_FAIL);
    }

    printf("Bridge demo started. Begin sending packets from external source\n");

    if(ixEthAccCodeletLoop() != IX_SUCCESS)
    {
	return (IX_FAIL);
    }

    /* Recover buffers */
    if(ixEthAccCodeletRecoverBuffers() != IX_SUCCESS)
    {
	printf("Error recovering buffers for Bridge demo\n");
	return (IX_FAIL);
    } 

    return (IX_SUCCESS);
}


/*
 * Function definition: ixEthAccCodeletDBLearning()
 *
 * See header file for documentation.
 */

IX_STATUS ixEthAccCodeletDBLearning(void)
{
    IxEthAccPortId portNo, portPtr;
    ix_ossl_thread_t dbMaintenanceid;

    IxEthDBMacAddr staticMacAddress[IX_ETHACC_CODELET_MAX_PORT] = 
	{{{0,1,2,3,4,5}}, {{6,7,8,9,0,1}}};

    IxEthDBMacAddr dynamMacAddress[IX_ETHACC_CODELET_MAX_PORT] = 
	{{{0xa,0xb,0xc,0xd,0xe,0xf}}, {{0x0,0xb,0x3,0xc,0x4,0xd}}};

    IX_ETHACC_IS_CODELET_INITIALISED();

    if (ix_ossl_thread_create((ix_ossl_thread_entry_point_t) 
			      ixEthAccCodeletDBMaintenanceTask,
			      NULL,
			      &dbMaintenanceid) != IX_OSSL_ERROR_SUCCESS)
    {
	printf("Error spawning DB maintenance task\n");
	return (IX_FAIL);
    }

    ix_ossl_thread_set_priority(dbMaintenanceid, IX_ETHACC_CODELET_DB_PRIORITY);

    /* Enable ports */
    if(ixEthAccPortEnable(IX_ETH_PORT_1) != IX_ETH_ACC_SUCCESS)
    {
	printf("Error enabling port 1\n");
	return (IX_FAIL);
    }

    if(ixEthAccPortEnable(IX_ETH_PORT_2) != IX_ETH_ACC_SUCCESS)
    {
	printf("Error enabling port 2\n");
	return (IX_FAIL);
    }

    /* Add some static entries */
    printf("Adding static entries...\n");

    for (portNo=IX_ETH_PORT_1; portNo<IX_ETHACC_CODELET_MAX_PORT; portNo++)
    {
	if (ixEthDBFilteringStaticEntryProvision(portNo,
						  &staticMacAddress[portNo])
	    != IX_ETH_DB_SUCCESS)
	{
	    printf("Failed to add static MAC to port %d\n", portNo+1);
	    return (IX_FAIL);
	}

	/* Verify that the address has successfully been added */
	if (ixEthDBFilteringDatabaseSearch(&portPtr, 
					    &staticMacAddress[portNo]) 
	    == IX_ETH_DB_NO_SUCH_ADDR)
	{
	    printf("Static MAC address not found in Database\n");
	    return (IX_FAIL);
	}
    }

    /* Add some dynamic entries */
    printf("Adding dynamic entries...\n");

    for (portNo=IX_ETH_PORT_1; portNo<IX_ETHACC_CODELET_MAX_PORT; portNo++)
    {
	if (ixEthDBFilteringDynamicEntryProvision(portNo,
						   &dynamMacAddress[portNo])
	    != IX_ETH_DB_SUCCESS)
	{
	    printf("Failed to add dynamic MAC to port %d\n", portNo+1);
	}

	/* Verify that the address has successfully been added */
	if (ixEthDBFilteringDatabaseSearch(&portPtr, 
					    &dynamMacAddress[portNo]) 
	    == IX_ETH_DB_NO_SUCH_ADDR)
	{
	    printf("Dynamic MAC address not found in Database\n");
	    return (IX_FAIL);
	}
    }

    ixEthDBFilteringDatabaseShowAll();

    /* Now age the dynamic entries */
    for (portNo=IX_ETH_PORT_1; portNo<IX_ETHACC_CODELET_MAX_PORT; portNo++)
    {
	printf("Enabling aging for port %d...\n", portNo+1);

	if (ixEthDBPortAgingEnable(portNo) != IX_ETH_DB_SUCCESS)
	{
	    printf("Failed to enable aging for port %d\n", portNo+1);
	    return (IX_FAIL);
	}
    }

    /* Wait aging time and verify that the dynamic entries have been removed */
    printf("Aging entries.\nWaiting for 16 minutes...");

    ix_ossl_sleep((IX_ETH_DB_LEARNING_ENTRY_AGE_TIME + 60) * 1000);

    ixEthDBFilteringDatabaseShowAll();

    /* Remove static entries so they can be used for another demo */
    for (portNo=IX_ETH_PORT_1; portNo<IX_ETHACC_CODELET_MAX_PORT; portNo++)
    {
	if(ixEthDBFilteringEntryDelete(&staticMacAddress[portNo])
	   != IX_ETH_DB_SUCCESS)
	{
	    printf("Failed to remove static MAC on port %d\n", portNo+1);
	}
    }
    
    /* Disable ports */
    if(ixEthAccPortDisable(IX_ETH_PORT_1) != IX_ETH_ACC_SUCCESS)
    {
	printf("Error disabling port 1\n");
	return (IX_FAIL);
    }

    if(ixEthAccPortDisable(IX_ETH_PORT_2) != IX_ETH_ACC_SUCCESS)
    {
	printf("Error disabling port 2\n");
	return (IX_FAIL);
    }

    return (IX_SUCCESS);
}


/*
 * Function definition: ixEthAccCodeletShow()
 *
 * See header file for documentation.
 */

void ixEthAccCodeletShow(void)
{
    static IxEthEthObjStats *portStats = NULL;
    UINT32 portNo;

    if(portStats == NULL)
    {
	/*
	 * The statistics are gathered by the NPE therefore we use DMA_MALLOC
	 * to make it cache safe for us to read.
	 */
	if((portStats = IX_ACC_DRV_DMA_MALLOC(sizeof(IxEthEthObjStats))) == NULL)
	{
	    printf("Unable to create port stats!\n");
	    return;
	}
    }

    for(portNo=0; portNo<IX_ETHACC_CODELET_MAX_PORT; portNo++)
    {
	memset(portStats, 0, sizeof(IxEthEthObjStats));

	printf("\nStatistics for port %d:\n", portNo);
	if(ixEthAccMibIIStatsGetClear(portNo, portStats) != IX_ETH_ACC_SUCCESS)
	{
	    printf("Unable to retrieve statistics for port %d!\n", portNo);
	}
	else
	{
	    printf("dot3StatsAlignmentErrors: \t\t%u\n",
		   portStats->dot3StatsAlignmentErrors);
	    printf("dot3StatsFCSErrors: \t\t\t%u\n",
		   portStats->dot3StatsFCSErrors);
	    printf("dot3StatsFrameTooLongs: \t\t%u\n",
		   portStats->dot3StatsFrameTooLongs);
	    printf("dot3StatsInternalMacReceiveErrors: \t%u\n",
		   portStats->dot3StatsInternalMacReceiveErrors);
	    printf("dot3StatsSingleCollisionFrames: \t%u\n",
		   portStats->dot3StatsSingleCollisionFrames);
	    printf("dot3StatsMultipleCollisionFrames: \t%u\n",
		   portStats->dot3StatsMultipleCollisionFrames);
	    printf("dot3StatsDeferredTransmissions: \t%u\n",
		   portStats->dot3StatsDeferredTransmissions);
	    printf("dot3StatsLateCollisions: \t\t%u\n",
		   portStats->dot3StatsLateCollisions);
	    printf("dot3StatsExcessiveCollisions: \t\t%u\n",
		   portStats->dot3StatsExcessiveCollsions);
	    printf("dot3StatsInternalMacTransmitErrors: \t%u\n",
		   portStats->dot3StatsInternalMacTransmitErrors);
	    printf("dot3StatsCarrierSenseErrors: \t\t%u\n",
		   portStats->dot3StatsCarrierSenseErrors);
	    printf("LearnedEntryDiscards: \t\t\t%u\n",
		   portStats->LearnedEntryDiscards);

	    printf("\n");
	}
    }
}


/******************************************************************************
 *
 * Ethernet Codelet support functions
 *
 ******************************************************************************/

/**
 * @fn void ixEthAccCodeletDBMaintenanceTask
 *
 * This function calls the ixEthAccDatabaseMaintenance every 
 * IX_ETH_DB_MAINTENANCE_TIME seconds.
 * 
 * @return void
 */

PRIVATE void ixEthAccCodeletDBMaintenanceTask (void* arg, void** ptrRetObj)
{
    while (1)
    {
	/* 
	 * The Database maintenance function must be called at a period of
	 * approximately IX_ETH_ACC_MAINTENANCE_TIME seconds regardless of 
	 * whether learning is enabled or not.
	 */
	ix_ossl_sleep(IX_ETH_DB_MAINTENANCE_TIME * 1000);
	ixEthDBDatabaseMaintenance();
    }
}


/**
 * @fn void ixEthAccCodeletDispatcherPoll
 *
 * This function polls the Queue manager loop.
 * 
 * @return void
 */

PRIVATE void ixEthAccCodeletDispatcherPoll (void* arg, void** ptrRetObj)
{
    while (1)
    {
	ix_ossl_sleep(1);
	ixQMgrDispatcherLoopRun (IX_QMGR_QUELOW_GROUP);
    }
}


/**
 * @fn IX_STATUS ixEthAccCodeletDispatcherStart()
 *
 * @param BOOL useInterrupt - start in interrupt or polled mode
 *
 * This function starts the Queue manager dispatch timer.
 * 
 * @return IX_SUCCESS - Dispatch timer successfully started
 * @return IX_FAIL - Error starting dispatch timer
 */

PRIVATE IX_STATUS ixEthAccCodeletDispatcherStart(BOOL useInterrupt)
{
    unsigned int dispatchtid;

    if(useInterrupt)	/* Interrupt mode */
    {
	/* 
	 * Hook the QM QLOW dispatcher to the interrupt controller. 
	 */
	if (ixOsServIntBind(IXP425_INT_LVL_QM1,
			    (void *)(ixQMgrDispatcherLoopRun),
			    (void *)IX_QMGR_QUELOW_GROUP) != IX_SUCCESS)
	{
	    printf("Failed to bind to QM1 interrupt\n");
	    return (IX_FAIL);
	}
    }
    else	/* Polled mode */
    {
 	if (ix_ossl_thread_create((ix_ossl_thread_entry_point_t)
				  ixEthAccCodeletDispatcherPoll,
				  NULL,
				  &dispatchtid) != IX_OSSL_ERROR_SUCCESS)
	{
	    printf("Error spawning dispatch task\n");
	    return (IX_FAIL);
	}

	ix_ossl_thread_set_priority(dispatchtid, IX_ETHACC_CODELET_QMR_PRIORITY);
    }

    return (IX_SUCCESS);
}


/**
 * @fn IX_STATUS ixEthAccCodeletPhyInit()
 *
 * This function scans for and then initialises any available PHYs on
 * the board. It uses the EthAcc MII library routines to do so.
 * 
 * @return IX_SUCCESS - PHY(s) detected and initialised
 * @return IX_FAIL - Error initialising PHY(s)
 */

PRIVATE IX_STATUS ixEthAccCodeletPhyInit(void)
{
   BOOL phyPresent[IXP425_ETH_ACC_MII_MAX_ADDR];

   UINT32 phyNo;
   UINT32 phyNoAddr;

   /* Scan for available Ethernet PHYs on the board */
   if(ixEthAccMiiPhyScan(phyPresent) == ERROR)
   {	
       return (IX_FAIL);
   }
   else
   {
       maxPhyNo = 0;
       for(phyNoAddr=0, phyNo=0; phyNoAddr<IXP425_ETH_ACC_MII_MAX_ADDR; phyNoAddr++)
       {
	   if(phyPresent[phyNoAddr] == TRUE)
	   {
	       phyAddresses[phyNo++] = phyNoAddr;
               maxPhyNo = phyNo;

	       if(phyNo == IX_ETHACC_CODELET_MAX_PORT)
	       {
		   break;
	       }
	   }
       }
   }
    
   for(phyNo=0; phyNo<maxPhyNo; phyNo++)
   {
       BOOL speed, linkUp, fullDuplex, autoneg;

       /* Reset each phy */
       ixEthAccMiiPhyReset(phyAddresses[phyNo]);

       /* Set each phy properties */
       ixEthAccMiiPhyConfig(phyAddresses[phyNo],
			    phyConf.speed100, 
			    phyConf.fullDuplex, 
			    phyConf.autonegotiate);

       /* wait until the link is up before setting the MAC duplex
        * mode, the PHY duplex mode may change after autonegotiation 
        */
       (void)ixEthAccCodeletLinkUpCheck(phyNo);
       ixEthAccMiiLinkStatus(phyAddresses[phyNo], 
                                &linkUp,
                                &speed,
                                &fullDuplex,
                                &autoneg);

       /* Set the MAC duplex mode */
       if(fullDuplex)
       {
	   ixEthAccPortDuplexModeSet (phyNo, IX_ETH_ACC_FULL_DUPLEX);
       }
       else
       {
	   ixEthAccPortDuplexModeSet (phyNo, IX_ETH_ACC_HALF_DUPLEX);
       }
       printf("\nPHY %d configuration:\n", phyNo);
       ixEthAccMiiShow(phyAddresses[phyNo]);
   }

   return (IX_SUCCESS);
}


/**
 * @fn IX_STATUS ixEthAccCodeletMemPoolInit()
 *
 * Initialise Ethernet Access Codelet MBUF pool
 *
 * @return IX_SUCCESS - MBuf pool successfully initialised
 * @return IX_FAIL - Error initialising MBuf pool
 */

PRIVATE IX_STATUS ixEthAccCodeletMemPoolInit(void)
{
    UINT32 bufNo;
    IX_MBUF *mBufPtr;

    if ((ixEthAccCodeletBufPool = 
	 IX_ACC_DRV_DMA_MALLOC(IX_ETHACC_CODELET_MBUF_POOL_SIZE*sizeof(IX_MBUF)))
	== NULL)
    {
	printf("Error allocating mBuf pool for Ethernet Codelet!\n");
	return (IX_FAIL);
    }
    
    mBufPtr = ixEthAccCodeletBufPool;

    if ((ixEthAccCodeletBufData = 
	 IX_ACC_DRV_DMA_MALLOC(IX_ETHACC_CODELET_MBUF_DATA_POOL_SIZE)) == NULL)
    {
	printf("Error allocating mBuf data pool for Ethernet Codelet!\n");
	return (IX_FAIL);
    }

    /* Format our Receive mBufs */
    for(bufNo = 0;
	bufNo < IX_ETHACC_CODELET_RX_MBUF_POOL_SIZE*IX_ETHACC_CODELET_MAX_PORT;
	bufNo++)
    {
	/* Initialise payload */ 
	IX_MBUF_MDATA(mBufPtr) = (char *)&ixEthAccCodeletBufData[bufNo*IX_ETHACC_CODELET_PCK_SIZE];

        /* Initialise length */
	IX_MBUF_MLEN(mBufPtr) = IX_MBUF_PKT_LEN(mBufPtr) = IX_ETHACC_CODELET_PCK_SIZE;
	
	/* No chains */
        IX_MBUF_NEXT_BUFFER_IN_PKT_PTR(mBufPtr) = NULL;
        mBufPtr->m_flags = 0x0;

	/* Add newly formated buffer to our Receive free queue */
	IX_ETHACC_CODELET_ADD_MBUF_TO_Q_TAIL(ixEthAccCodeletFreeBufQ, mBufPtr);
	mBufPtr++;
    }

    /* Format our Transmit mBufs */
    for(bufNo = IX_ETHACC_CODELET_RX_MBUF_POOL_SIZE*IX_ETHACC_CODELET_MAX_PORT; 
	bufNo < (IX_ETHACC_CODELET_MBUF_POOL_SIZE); 
	bufNo++)
    {
	/* Initialise payload */ 
	IX_MBUF_MDATA(mBufPtr) = (char *)&ixEthAccCodeletBufData[bufNo*IX_ETHACC_CODELET_PCK_SIZE];

        /* Initialise length */
	IX_MBUF_MLEN(mBufPtr) = IX_MBUF_PKT_LEN(mBufPtr) = IX_ETHACC_CODELET_PCK_SIZE;
	
        /* No chains */
        IX_MBUF_NEXT_BUFFER_IN_PKT_PTR(mBufPtr) = NULL;
        mBufPtr->m_flags = 0x0;

	/* Add newly formated buffer to our free queue */
	IX_ETHACC_CODELET_ADD_MBUF_TO_Q_TAIL(ixEthAccCodeletFreeBufQ, mBufPtr);
	mBufPtr++;
    }

    return (IX_SUCCESS);
}


/**
 * @fn IX_STATUS ixEthAccCodeletMacSet()
 *
 * Set the MAC address for available Ports on the board.
 *
 * @return IX_SUCCESS - MAC address successfuly set
 * @return IX_FAIL - Error setting MAC address
 */

PRIVATE IX_STATUS ixEthAccCodeletMacSet(void)
{
    IxEthAccMacAddr npeMacAddr1 = IX_ETHACC_CODELET_NPE1_MAC;
    IxEthAccMacAddr npeMacAddr2 = IX_ETHACC_CODELET_NPE2_MAC;

#ifdef IX_ETHACC_CODELET_USE_NVRAM_MAC
    printf("\nReading MAC address information from non-volatile storage...\n");

    if(sysNvRamGet((UINT8 *)&npeMacAddr1.macAddress, 
		   IX_IEEE803_MAC_ADDRESS_SIZE, 
		   NV_MAC_ADRS_NPE1) == ERROR)
    {
	printf("Unable to read NPE1 MAC address from non-volatile storage!\n");
	return (IX_FAIL);
    }

    if(sysNvRamGet((UINT8 *)&npeMacAddr2.macAddress, 
		   IX_IEEE803_MAC_ADDRESS_SIZE, 
		   NV_MAC_ADRS_NPE2) == ERROR)
    {
	printf("Unable to read NPE2 MAC address from non-volatile storage!\n");
	return (IX_FAIL);
    }
#endif

    if(ixEthAccPortUnicastMacAddressSet(IX_ETH_PORT_1, &npeMacAddr1) != IX_ETH_ACC_SUCCESS)
    {
	return (IX_FAIL);
    }

    printf("PHY %d MAC address is:\t", IX_ETH_PORT_1);
    ixEthAccPortUnicastAddressShow(IX_ETH_PORT_1);

    if(ixEthAccPortUnicastMacAddressSet(IX_ETH_PORT_2, &npeMacAddr2) != IX_ETH_ACC_SUCCESS)
    {
	return (IX_FAIL);
    }

    printf("PHY %d MAC address is:\t", IX_ETH_PORT_2);
    ixEthAccPortUnicastAddressShow(IX_ETH_PORT_2);

    return (IX_SUCCESS);
}

/**
 * @fn IX_STATUS ixEthAccCodeletRecoverBuffers()
 *
 * Recover buffers used by a demo into the free queue.
 *
 * @return IX_SUCCESS - Recovery succeeded
 * @return IX_FAIL - Error recovering buffers
 */

IX_STATUS ixEthAccCodeletRecoverBuffers(void)
{
    unsigned int count = 0;

    memPoolFreeCnt = 0;

    printf("Recovering buffers...\n");

    /* Register callbacks to recover the buffers used */
    if(ixEthAccCodeletStop(ixEthAccCodeletMemPoolFreeRxCB,
			   ixEthAccCodeletMemPoolFreeRxCB,
			   ixEthAccCodeletMemPoolFreeTxCB,
			   ixEthAccCodeletMemPoolFreeTxCB) != IX_SUCCESS)
    {
	printf("Failed to stop the codelet\n");
	return (IX_FAIL);
    }

    /* wait 1 second for mbufs to be received */
    while ((count++ < 10) &&
	   ( memPoolFreeCnt != IX_ETHACC_CODELET_RX_MBUF_POOL_SIZE 
                               + IX_ETHACC_CODELET_TX_MBUF_POOL_SIZE))
    {
	ixOsServTaskSleep(100); /* 100 millisecs */
    }
    
    printf("%d buffers recovered\n", memPoolFreeCnt);

    return (IX_SUCCESS);
}


/**
 * @fn IX_STATUS ixEthAccCodeletReplenishBuffers()
 *
 * @param portNo - port number to replenish buffers into.
 * @param num - number of buffers to replenish.
 *
 * Function to replenish buffers into the Rx free queue.
 *
 * @return IX_SUCCESS - Replenish succeeded
 * @return IX_FAIL - Error replenishing buffers in Rx queue for port
 */

PRIVATE IX_STATUS ixEthAccCodeletReplenishBuffers(IxEthAccPortId portNo,
						 UINT32 num)
{
    IX_MBUF *mBufPtr;
    UINT32 numBufs;

    for(numBufs=0; numBufs<num; numBufs++)
    {
	if(IX_ETHACC_CODELET_IS_Q_EMPTY(ixEthAccCodeletFreeBufQ))
	{
	    printf("Buffer queue empty."
		   "Not enough free buffers for port %d!\n", portNo+1);
	    return (IX_FAIL);
	}
	
	IX_ETHACC_CODELET_REMOVE_MBUF_FROM_Q_HEAD(ixEthAccCodeletFreeBufQ, mBufPtr);
	IX_MBUF_MLEN(mBufPtr) = IX_MBUF_PKT_LEN(mBufPtr) = IX_ETHACC_CODELET_PCK_SIZE;
	IX_ACC_DATA_CACHE_INVALIDATE(IX_MBUF_MDATA(mBufPtr), IX_MBUF_MLEN(mBufPtr));
	
	if(ixEthAccPortRxFreeReplenish(portNo, mBufPtr) != IX_SUCCESS)
	{
	    printf("Error replenishing free queue on port %d\n", portNo+1);
	    return (IX_FAIL);
	}
    }

    printf("Port %d Rx Free pool has %d buffers\n", portNo+1, numBufs);
    return (IX_SUCCESS);
}

/**
 * @fn void ixEthAccCodeletStatsPollTask
 *
 * This task polls the Codelet Stats and displays the rate of Rx and 
 * Tx packets per second.
 * 
 * @return void
 */

PRIVATE void ixEthAccCodeletStatsPollTask (void* arg, void** ptrRetObj)
{
    static char stillRunning[] = "|/-\\";
    static int  stillRunningIndex = 0;

    while (1)
    {
	while (ixEthAccCodeletPollEnabled == FALSE)
	{
	    /* Sleep 1 sec */
	    ix_ossl_sleep(1000);
	}

	printf("\nPort1 Rates:              | Port2 Rates:\n"
	       "=====================================================\n");
	while (ixEthAccCodeletPollEnabled == TRUE)
	{
	    /* Clear stats */
	    IxEthAccCodeletStats[0].rxCount=0;
	    IxEthAccCodeletStats[0].txCount=0;
	    IxEthAccCodeletStats[1].rxCount=0;
	    IxEthAccCodeletStats[1].txCount=0;

	    /* Sleep 1 sec */
	    ix_ossl_sleep(1000);
	    if  (ixEthAccCodeletPollEnabled == TRUE)
	    {
		/* Display stats */
		printf("\rRx: %6u     Tx: %6u | Rx: %6u     Tx: %6u %c",
		       IxEthAccCodeletStats[0].rxCount,
		       IxEthAccCodeletStats[0].txCount,
		       IxEthAccCodeletStats[1].rxCount,
		       IxEthAccCodeletStats[1].txCount,
		       stillRunning[stillRunningIndex++ & 3]);
	    }
	}
	printf("\n");
    }
}


/**
 * @fn void ixEthAccCodeletLoop()
 *
 * This function is called once a demo has started. It spawns the
 * stats poll task and waits for use interrupt to end the codelet.
 * 
 * @return IX_SUCCESS - demo successfully run
 * @return IX_FAIl - error spawning stats poll task
 */

PRIVATE IX_STATUS ixEthAccCodeletLoop(void)
{
    int ch = 0;

    printf("Press ESC at any time to stop the demo ... \n");

    ixEthAccCodeletPollEnabled = TRUE;

    do
    {
	ch = getchar();
    }while(ch != 27);
    
    ixEthAccCodeletPollEnabled = FALSE;

    /* Dump Stats */
    ixEthAccCodeletShow();

    return (IX_SUCCESS);
}

/******************************************************************************
 *
 * Ethernet Codelet callback functions
 *
 ******************************************************************************/

/**
 * @fn void ixEthAccCodeletRxSinkCB()
 *
 * Callback for the Rx Sink demo. Simply replenish the buffer back into the rx queue.
 * 
 * @return void
 */ 

PRIVATE void ixEthAccCodeletRxSinkCB(UINT32 cbTag, IX_MBUF* mBufPtr, IxEthAccPortId portId)
{
    /* Put the mbuf back on the rx free pool */
    if(cbTag == IX_ETH_PORT_1 || cbTag == IX_ETH_PORT_2)
    {
	IxEthAccCodeletStats[cbTag].rxCount++;

	IX_MBUF_MLEN(mBufPtr) = IX_MBUF_PKT_LEN(mBufPtr) = IX_ETHACC_CODELET_PCK_SIZE;
	if(ixEthAccPortRxFreeReplenish(cbTag, mBufPtr) != IX_SUCCESS)
	{
	    ixOsServLog(LOG_ERROR, "Error replenishing RX free queue on port %d\n", 
			cbTag+1, 0, 0, 0, 0, 0);
	}    
    }
    else
    {
	ixOsServLog(LOG_ERROR, "Invalid Rx callback tag %d in RxSink demo\n", 
		    cbTag, 0, 0, 0, 0, 0);
    }
}


/**
 * @fn void ixEthAccCodeletSwLoopbackTxCB()
 *
 * Transmit callback for Sw Loopback demo. Transmitted frames are put on
 * the receive free pool.
 * 
 * @return void
 */

PRIVATE void ixEthAccCodeletSwLoopbackTxCB(UINT32 cbTag, IX_MBUF* mBufPtr)
{
    /* Put the mbuf back on the rx free pool */
    if(cbTag == IX_ETH_PORT_1 || cbTag == IX_ETH_PORT_2)
    {
	IxEthAccCodeletStats[cbTag].txCount++;

	IX_MBUF_MLEN(mBufPtr) = IX_MBUF_PKT_LEN(mBufPtr) = IX_ETHACC_CODELET_PCK_SIZE;
	if(ixEthAccPortRxFreeReplenish(cbTag, mBufPtr) != IX_SUCCESS)
	{
	    ixOsServLog(LOG_ERROR, "Error replenishing RX free queue on port %d\n", 
			cbTag+1, 0, 0, 0, 0, 0);
	}    
    }
    else
    {
	ixOsServLog(LOG_ERROR, "Invalid Tx callback tag %d in RxSink demo\n", 
		    cbTag, 0, 0, 0, 0, 0);
    }
}


/**
 * @fn void ixEthAccCodeletSwLoopbackRxCB()
 *
 * Receive callback for the Sw Loopback demo. Any frames received are transmited
 * back out on the same port.
 * 
 * @return void
 */

PRIVATE void ixEthAccCodeletSwLoopbackRxCB(UINT32 cbTag, IX_MBUF* mBufPtr, IxEthAccPortId portId)
{
    /* Transmit the buffer back */
    if(cbTag == IX_ETH_PORT_1 || cbTag == IX_ETH_PORT_2)
    {
	IxEthAccCodeletStats[cbTag].rxCount++;

	if(ixEthAccPortTxFrameSubmit(cbTag, mBufPtr,
				     IX_ETH_ACC_TX_DEFAULT_PRIORITY) != IX_ETH_ACC_SUCCESS)
	{
	    ixOsServLog(LOG_ERROR, "Tx Buffer submission failure port %d\n", 
			cbTag+1, 0, 0, 0, 0, 0);
	}
    }
    else
    {
	ixOsServLog(LOG_ERROR, "Invalid Rx callback tag %d in Sw Loopback demo\n", 
		    cbTag, 0, 0, 0, 0, 0);
    }
}


/**
 * @fn void ixEthAccCodeletBridgeTxCB()
 *
 * Transmit callback for Bridge demo. Transmitted frames are put on
 * the receive queue of the other port.
 * 
 * @return void
 */

PRIVATE void ixEthAccCodeletBridgeTxCB(UINT32 cbTag, IX_MBUF* mBufPtr)
{
    /* 
     * Put the mbuf on the rx free q of the other port since it was received
     * there initially.
     */
    if(cbTag == IX_ETH_PORT_1)
    {
	IxEthAccCodeletStats[cbTag].txCount++;

	IX_MBUF_MLEN(mBufPtr) = IX_MBUF_PKT_LEN(mBufPtr) = IX_ETHACC_CODELET_PCK_SIZE;
	if(ixEthAccPortRxFreeReplenish(IX_ETH_PORT_2, mBufPtr)!=IX_SUCCESS)
	{
	    ixOsServLog(LOG_ERROR, "Error replenishing RX free q for port 2\n", 
			0, 0, 0, 0, 0, 0);
	}    
    }
    else if(cbTag == IX_ETH_PORT_2)
    {
	IxEthAccCodeletStats[cbTag].txCount++;

	IX_MBUF_MLEN(mBufPtr) = IX_MBUF_PKT_LEN(mBufPtr) = IX_ETHACC_CODELET_PCK_SIZE;
	if(ixEthAccPortRxFreeReplenish(IX_ETH_PORT_1, mBufPtr)!=IX_SUCCESS)
	{
	    ixOsServLog(LOG_ERROR, "Error replenishing RX free q for port 1\n", 
			0, 0, 0, 0, 0, 0);
	}    
    }
    else
    {
	ixOsServLog(LOG_ERROR, "Invalid Tx callback tag %d in Bridge demo\n", 
		    cbTag, 0, 0, 0, 0, 0);
    }
}


/**
 * @fn void ixEthAccCodeletBridgeRxCB()
 *
 * Receive callback for Bridge demo. Received frames are retransmited on
 * the other port.
 * 
 * @return void
 */

PRIVATE void ixEthAccCodeletBridgeRxCB(UINT32 cbTag, IX_MBUF* mBufPtr, IxEthAccPortId portId)
{
    /* Transmit the buffer on the other port */
    if(cbTag == IX_ETH_PORT_1)
    {
	IxEthAccCodeletStats[cbTag].rxCount++;

	if(ixEthAccPortTxFrameSubmit(IX_ETH_PORT_2, mBufPtr,
				     IX_ETH_ACC_TX_DEFAULT_PRIORITY)!=IX_ETH_ACC_SUCCESS)
	{
	    ixOsServLog(LOG_ERROR, "Tx Buffer submission failed on port 2\n", 
			0, 0, 0, 0, 0, 0);
	}
    }
    else if(cbTag == IX_ETH_PORT_2)
    {
	IxEthAccCodeletStats[cbTag].rxCount++;

	if(ixEthAccPortTxFrameSubmit(IX_ETH_PORT_1, mBufPtr,
				     IX_ETH_ACC_TX_DEFAULT_PRIORITY)!=IX_ETH_ACC_SUCCESS)
	{
	    ixOsServLog(LOG_ERROR, "Tx Buffer submission failed on port 1\n", 
			0, 0, 0, 0, 0, 0);
	}
    }
    else
    {
	ixOsServLog(LOG_ERROR, "Invalid Rx callback tag %d in Bridge demo\n", 
		    cbTag, 0, 0, 0, 0, 0);
    }
}


/**
 * @fn void ixEthAccCodeletTxGenRxSinkLoopbackRxCB()
 *
 * Receive callback for TxGen-RxSink loopback demo. Received frames are put back on
 * the free queue.
 * 
 * @return void
 */

PRIVATE void ixEthAccCodeletTxGenRxSinkLoopbackRxCB(UINT32 cbTag, IX_MBUF* mBufPtr, IxEthAccPortId portId)
{
    /* Return mBuf to free queue on same port */
    IxEthAccCodeletStats[cbTag].rxCount++;

#ifdef IX_ETHACC_CODELET_TXGENRXSINK_VERIFY 
    IX_ETHACC_CODELET_DATA_VERIFY(IX_MBUF_MDATA(mBufPtr), compData);
#endif
    IX_MBUF_MLEN(mBufPtr) = IX_MBUF_PKT_LEN(mBufPtr) = IX_ETHACC_CODELET_PCK_SIZE;
    if(ixEthAccPortRxFreeReplenish(portId, mBufPtr) != IX_SUCCESS)
    {
	ixOsServLog(LOG_ERROR, "Error replenishing RX free q in TxGen-RxSink\n", 
		    0, 0, 0, 0, 0, 0);
    }
}


/**
 * @fn void ixEthAccCodeletTxGenRxSinkLoopbackTxCB()
 *
 * Transmit callback for TxGenRxSink Loopback demo. Transmitted frames are re-
 * transmitted.
 * 
 * @return void
 */

PRIVATE void ixEthAccCodeletTxGenRxSinkLoopbackTxCB(UINT32 cbTag, IX_MBUF* mBufPtr)
{
    IxEthAccCodeletStats[cbTag].txCount++;

    /* Re-transmit the frame on same port */
    if(ixEthAccPortTxFrameSubmit(cbTag, mBufPtr,
				 IX_ETH_ACC_TX_DEFAULT_PRIORITY) != IX_ETH_ACC_SUCCESS)
    {
	ixOsServLog(LOG_ERROR, "Tx Buffer submission failure in TxGen-RxSink\n", 
		    0, 0, 0, 0, 0, 0);
    }
}


/*
 * @fn ixEthAccCodeletMemPoolFreeRxCb
 *
 * This callback is used at the end of each ethernet demo.
 * It recovers the buffers used and places them back in the free queue.
 * This way the MBUF pool is free to be used by another ethernet demo.
 *
 * @return void
 */

PRIVATE void ixEthAccCodeletMemPoolFreeRxCB(UINT32 cbTag, IX_MBUF* mBufPtr, IxEthAccPortId portId)
{
    IX_ETHACC_CODELET_ADD_MBUF_TO_Q_TAIL(ixEthAccCodeletFreeBufQ, mBufPtr);
    memPoolFreeCnt++;
}

/*
 * @fn ixEthAccCodeletMemPoolFreeTxCb
 *
 * This callback is used at the end of each ethernet demo.
 * It recovers the buffers used and places them back in the free queue.
 * This way the MBUF pool is free to be used by another ethernet demo.
 *
 * @return void
 */

PRIVATE void ixEthAccCodeletMemPoolFreeTxCB(UINT32 cbTag, IX_MBUF* mBufPtr)
{
    IX_ETHACC_CODELET_ADD_MBUF_TO_Q_TAIL(ixEthAccCodeletFreeBufQ, mBufPtr);
    memPoolFreeCnt++;
}
