 /**
 * @file IxEthAccCodelet.h
 *
 * @date 22 April 2002
 *
 * @brief This file contains the interface for the Ethernet Access Codelet.
 *
 * -- Intel Copyright Notice --
 *
 * @par
 * INTEL CONFIDENTIAL
 *
 * @par
 * Copyright 2002 Intel Corporation All Rights Reserved.
 *
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.  The Material
 * contains trade secrets and proprietary and confidential information of
 * Intel or its suppliers and licensors.  The Material is protected by
 * worldwide copyright and trade secret laws and treaty provisions. No
 * part of the Material may be used, copied, reproduced, modified,
 * published, uploaded, posted, transmitted, distributed, or disclosed in
 * any way without Intel's prior express written permission.
 *
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel or otherwise.  Any license under
 * such intellectual property rights must be express and approved by
 * Intel in writing.
 *
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * -- End Intel Copyright Notice --
 *
 */

/**
 * @defgroup Codelets IXP4xx Codelets
 *
 * @brief IXP4xx demo codelets
 *
 * @{
 */

/**
 * @ingroup Codelets
 *
 * @defgroup IxEthAccCodelet IXP4xx Ethernet Access Codelet (IxEthAccCodelet) API
 *
 * @brief IXP4xx demo Ethernet Access component API
 *
 * This file contains the interface of the Ethernet Access Codelet.
 * <P>
 * <B>Demos</B>
 * The following top-level scenarios are supported:
 * <UL>
 *   <LI> Demonstrate Ethernet Data and Control plane services by:
 *   <UL>
 *        <LI> Configuring both ports as a receiver sink from an external source (such 
 *            as Smartbits).
 *        <LI> Connecting a port to external test equipment performing a loopback.
 *        <LI> Configuring and performing a software loopback on each of the two ethernet ports.
 *	  <LI> Configuring both ports to act as a bridge so that frames received on one port are retransmitted on the other.
 *   </UL>
 * </UL>
 *
 *   <UL>
 *	<LI> Demonstrate Ethernet Management services by:
 *      <UL>
 *	<LI> Adding and removing static/dynamic entries.
 *	<LI> Calling the maintenance interface (shall be run as a separate background task)
 *      <LI> Calling the show routine to display the MAC address filtering tables.    
 *      </UL>
 *   </UL>
 *
 * All demos will be conducted using using different packet sizes and varying
 * source MAC addresses. Where external test equipment is used to transmit data it should
 * verify that data received is the same as that transmitted.
 *
 * A list of statistics will be available for all demos and shall be display-able by using
 * the Codelet's show function.
 *
 * <P>
 * <B>MAC Setup</B>
 * <P>
 * The default MAC setup will be:
 * <UL>
 *	<LI>Promiscuous mode enabled (for learning example)
 *	<LI>Frame Check Sequence appended for all frames generated on the XScale
 * </UL>
 *
 * <P>
 * <B>PHY Setup</B>
 * <P>
 * This demo uses two PHYs as defined by IX_ETHACC_CODELET_MAX_PHY 
 * The default PHY setup will be: 
 * <UL>
 *	<LI>100Mbits,
 *	<LI>full duplex,
 *	<LI>auto-negotiation on.
 * </UL>
 *
 * <P>
 * <B>Loopback</B>
 * <P>
 * In the loopback scenarios the Codelet will offer the facility
 * to perform a loopback.
 * <UL>
 *   <LI> The external test equipment will create and transmit packets, and
 * ensure that received packets match those transmitted.  The Codelet will
 * re-transmit any received packets.
 *   <LI> In the scenario where two ports are involved in the loopback the received
 * packets will be passed to the alternative port much like a bridge would do.
 * </UL>
 *
 * <P>
 * <B>Test Equipment</B>
 * <P>
 * The test harness will consist of external test equipment capable of
 * generating Ethernet packets (e.g. SmartBits).
 * <P>
 * The test equipment must be capable of performing at least the following
 * actions to support the scenarios outlined for the Codelet:
 * <UL>
 *   <LI> Send/receive an Ethernet data stream.
 *   <LI> Send/receive frames of different length.
 *   <LI> Detect CRC errors.
 *   <LI> Append FCS.
 *   <LI> Support 100Mbit full duplex mode.
 * </UL>
 *
 * @{
 */

#ifndef IXETHACCCODELET_H
#define IXETHACCCODELET_H

#include "IxEthAcc.h"

/*
 * #defines for function return types, etc.
 */

/**< Hard-encoded MAC address for NPE1 */
#define IX_ETHACC_CODELET_NPE1_MAC {{0x2, 0x0, 0xa, 0xb, 0xc, 0xd}}

/**< Hard-encoded MAC address for NPE2 */
#define IX_ETHACC_CODELET_NPE2_MAC {{0x2, 0x0, 0xe, 0xf, 0xa, 0xb}}

/**< Recommended size of receive MBuf pool */
#define IX_ETHACC_CODELET_RX_MBUF_POOL_SIZE   128

/**< Recommended size of transmit MBuf pool */
#define IX_ETHACC_CODELET_TX_MBUF_POOL_SIZE   128

/**< Number of Ports supported for this demo */
#define IX_ETHACC_CODELET_MAX_PORT IX_ETH_ACC_NUMBER_OF_PORTS

/**< Size of MBuf pool */
#define IX_ETHACC_CODELET_MBUF_POOL_SIZE \
((IX_ETHACC_CODELET_RX_MBUF_POOL_SIZE + IX_ETHACC_CODELET_TX_MBUF_POOL_SIZE) \
  * IX_ETHACC_CODELET_MAX_PORT)

/**< Size of MBuf packet */
#define IX_ETHACC_CODELET_PCK_SIZE IX_ETHACC_RX_MBUF_MIN_SIZE

/**< Length of MBuf payload (in bytes) */
#define IX_ETHACC_CODELET_PCK_LEN 1514

/**< Size of MBuf data pool */
#define IX_ETHACC_CODELET_MBUF_DATA_POOL_SIZE \
(IX_ETHACC_CODELET_MBUF_POOL_SIZE * IX_ETHACC_CODELET_PCK_SIZE)

/**<Size of packets for TxGenRxSink demo */
#define IX_ETHACC_CODELET_TXGEN_PCK_LEN 60

/**< Number of packets to generate for the TxGenRxSink demo */
#define IX_ETHACC_CODELET_TXGEN_PCKS 128

#if IX_ETHACC_CODELET_TXGEN_PCKS > IX_ETHACC_CODELET_TX_MBUF_POOL_SIZE
    #error Number of TX Gen packets is greater than Tx MBuf pool
#endif

/**< Verify payload for TxGenRxSink demo - defining this will affect performance */
#define IX_ETHACC_CODELET_TXGENRXSINK_VERIFY 

typedef struct
{
    BOOL speed100;	/**< 100 Mbits */
    BOOL fullDuplex;	/**< Full Duplex */
    BOOL autonegotiate;	/**< Autonegotiation */
} IxEthAccCodeletPhyConf;


/*
 * Prototypes for interface functions.
 */

/**
 * @fn IX_STATUS ixEthAccCodeletInit()
 *
 * This function is the entry point to the Ethernet Access codelet.
 * It must be called before any of the demos as it initialises the neccessary 
 * components.
 * 
 * The initialisation sequence is as follows:
 * <UL>
 *	<LI> Initialise Queue manager
 *	<LI> Start Queue manager dispatcher loop
 *	<LI> Download NPE microcode
 *	<LI> Start NPE message handler
 *	<LI> Start NPEs
 *	<LI> Initialise Ethernet Access component
 *	<LI> Initialise Ethernet ports
 *	<LI> Initialise Ethernet PHYs
 *	<LI> Program MAC addresses
 *	<LI> Set ports to promiscuous mode
 *	<LI> Initialise MBUF pool
 *	<LI> Set the Tx scheduling priority to FIFO_NO_PRIORITY 
 * </UL>  
 *
 * At the end of the initialisation sequence ports are not enabled.
 * They are enabled as appropriate for each demo that's run.
 *
 * @return IX_SUCCESS - Codelet successfully initialised 
 * @return IX_FAIL - Error initialising codelet
 */

IX_STATUS
ixEthAccCodeletInit(void);


/**
 * @fn IX_STATUS ixEthAccCodeletRxSink()
 *
 * Receive sink demo.
 *
 * This function sinks received packets as fast as possible on both ports.
 * An external source must be connected to both ports transmitting
 * Ethernet packets.  
 *
 * @return IX_SUCCESS - Rx Sink demo successfully started
 * @return IX_FAIL - Error starting Rx Sink demo
 */ 

IX_STATUS
ixEthAccCodeletRxSink(void);


/**
 * @fn IX_STATUS ixEthAccCodeletSwLoopback()
 *
 * Software loopback demo.
 *
 * This function sets up some callbacks so that any frames received are
 * looped back and retransmitted unmodified on the same port.
 * An external Ethernet source must be connected to both ports transmitting
 * packets.
 * 
 * @return IX_SUCCESS - Software Loopback demo successfully started
 * @return IX_FAIL - Software started external Loopback demo
 */ 

IX_STATUS
ixEthAccCodeletSwLoopback(void);


/**
 * @fn IX_STATUS ixEthAccCodeletTxGenRxSinkLoopback()
 *
 * Port-to-Port (Tx to Rx Sink) loopback demo.
 *
 * This function performs a software loopback on the two ethernet ports.
 * Frames are generated on the XScale and sent from port 1 and received on port 
 * 2 via a cross-over cable. Received frames on port 2 are put back on the free 
 * queue which is shared between the two ports. A verify function ensures that the
 * frames transmitted are the same as those received. A cross-over cable connecting
 * port 1 and port 2 is needed to perform this demo.
 *
 * @return IX_SUCCESS - SW Loopback successfully started
 * @return IX_FAIL - Error starting SW loopback demo
 */

IX_STATUS
ixEthAccCodeletTxGenRxSinkLoopback(void);


/**
 * @fn IX_STATUS ixEthAccCodeletBridge()
 *
 * Bridge demo.
 *
 * This demo transmits any frames received on one port through the other one,
 * as a bridge would do. An external Ethernet source must be connected to
 * both ports transmitting frames.
 * 
 * @return IX_SUCCESS - Bridge demo successfully started.
 * @return IX_FAIL - Error starting bridge demo.
 */

IX_STATUS
ixEthAccCodeletBridge(void);


/**
 * @fn IX_STATUS ixEthAccCodeletDBLearning()
 *
 * Ethernet MAC address learning demo.
 *
 * This function demonstrates the use of the Ethernet MAC learning facility.
 * It adds some static and dynamic entries. Dynamic entries are then aged and
 * verified that they no longer appear in the database.
 *
 * @return IX_SUCCESS - DB Learning demo successfully started
 * @return IX_FAIL - Error starting DB Learning demo.
 */ 

IX_STATUS
ixEthAccCodeletDBLearning(void);


/*
 * @fn ixEthAccCodeletShow()
 *
 * This function is called at the end of each ethernet demo.
 * It displays the MIB statistics for each port.
 *
 * @return void
 */

void
ixEthAccCodeletShow(void);

#endif /* IXETHACCCODELET_H */

/** @} defgroup IxEthAccCodelet*/

/** @} defgroup Codelets*/
