/**
 * @file IxFpathAccCodelet.h
 * @date 18 June 2002
 * @brief This file contains the API for the Fpath Access Codelet.
 *
 * -- Intel Copyright Notice --
 *
 * @par
 * INTEL CONFIDENTIAL
 *
 * @par
 * Copyright 2002 Intel Corporation All Rights Reserved.
 *
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.  The Material
 * contains trade secrets and proprietary and confidential information of
 * Intel or its suppliers and licensors.  The Material is protected by
 * worldwide copyright and trade secret laws and treaty provisions. No
 * part of the Material may be used, copied, reproduced, modified,
 * published, uploaded, posted, transmitted, distributed, or disclosed in
 * any way without Intel's prior express written permission.
 *
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel or otherwise.  Any license under
 * such intellectual property rights must be express and approved by
 * Intel in writing.
 *
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * -- End Intel Copyright Notice --
 *
 */

/**
 * @defgroup Codelets IXP4xx Codelets
 *
 * @brief IXP4xx demo codelets
 *
 * @{
 */

/**
 * @ingroup Codelets
 *
 * @defgroup IxFpathAccCodelet IXP4xx Fast Path Access Codelet (IxFpathAccCodelet) API
 *
 * @brief IXP4xx demo Fast Path Access component API
 *
 * This module contains the implementation of the FPATH Access Codelet.
 * <P>
 * The following top-level scenarios are supported:
 * <UL>
 *   <LI> Provide a Ethernet Bridge fast path Demo on a single VC.
 *   <UL>
 *     <LI> Configures the system to pass Ethernet frames with a given
 *          RFC1483 (LLC/VCMUX) encapsulation on the fast path from the
 *          Utopia to a single Ethernet port.
 *     <LI> Provide example classifier and modifier templates for LLC and
 *          VCMUX.
 *     <LI> Simple demonstration of the Wan Mac address learning functionality.
 *     <LI> The Utopia port is connected to a piece of test equipment
 *          capabale of generating encapsulated ethernet Frames.
 *     <LI> The Ethernet port is connected to a piece of test equipment
 *          capabale of accepting and verifying ethernet Frames.
 *   </UL>
 *   The ATM test equipment will inject traffic on the Utopia and the Ethernet
 *   test equipment will ensure the integrity of the traffic. No packet loss
 *   should be observed. The learned Mac address database will be displayed
 *   periodically on standard out.
 *
 *   Since Ethernet Frame sequencing between hosts must be observed all packets
 *   are conveyed on the Fast Path. Only errored packets are set to the slow
 *   path and at that are dropped.
 *
 *   No Ethernet to ATM traffic will be supported.
 *
 *
 *   <LI> Provide a Ip forwarding fast path demo on a single VC.
 *   <UL>
 *     <LI> Configures the system to pass Routed IP packet with a given
 *          RFC1483 (LLC/VCMUX) encapsulation on the fast path from the
 *          Utopia to a single Ethernet port.
 *     <LI> Provide example classifier and modifier templates for Routed IP LLC
 *          and VCMUX.
 *     <LI> Simple demonstration of how IP Flow detection and aging can be
 *          implemented.
 *     <LI> The Utopia port is connected to a piece of test equipment
 *          capabale of generating encapsulated IP packets.
 *     <LI> The Ethernet port is connected to a piece of test equipment
 *          capabale of accepting ethernet frames containing IP packets.
 *   </UL>
 *   The ATM test equipment will inject traffic on the Utopia and the Ethernet
 *   test equipment will ensure the integrity of the traffic. No packet loss
 *   should be observed.
 *
 *   Initially packets will flow on the slow path until the demonstration learning
 *   and aging function identifies Ip packet flows. As flows are identified and
 *   aged the codelet will provide feedback on standard out.
 *
 *   Ip packet forwarding is provided in the ATM to Ehternet direction only and
 *   will not involve an IP stack. Packets travelling onthe slow path will be
 *   forwarded to a hard coded Ethernet address over the Ethernet port.
 *
 * </UL>
 * Further scenarios to include other encapsulation types are considered outside
 * the scope of this demonstration Codelet.  It will, however, be a design goal
 * of this Codelet to be re-usable in such extended scenarios e.g. NAPT, PPPoe.
 *
 * <P>
 * For the current release only one Utopia port will be supported.  However,
 * the Codelet will be developed in such a way that the scenarios are
 * easily extendible to use upto 12 ports.
 *
 * <P>
 * <B>Test Equipment </B>
 * <P>
 * <B> Atm: </B>
 * The test harness will consist of external test equipment capable of
 * generating RFC1483 encapsulated Ethernet and Ip traffic over ATM. It should
 * provide Utopia connectivity and be capable of operating in MPhy, Utopia
 * Slave mode.
 *
 * <P>
 * <B> Ethernet: </B>
 * The Ethernet test equipement should be attached to the IX_ETH_PORT_1. By
 * default both ethernet ports are brought up with auto negotiation enabled.
 * The test equipment should have the capability to count and inspect Ehternet
 * frames and should ideally support 100Meg Ethernet.
 *
 * <P>
 * <B>Buffer Management:</B>
 * <P>
 * The demo code will use a single buffer pool for supply of all buffers to
 * IxAtmdAcc and IxFpathAcc. The pool will contain a selection of buffer sizes,
 * the dimensions of which are available in IxFpathAccCodeletBufMan.c.
 *
 * <P>
 * <B>Buffers and Caching:</B>
 * The allocation of buffers uses the MACROS availiable in IxOsCacheMMU.h.
 * If these routines are populated will calls to the cache driver then buffer
 * memory will be allocated from cache. Special attention is made in the demo
 * code to invalidating and flushing cached memory when manipulating packets
 * within buffers on the datapath.
 * <P>
 * More specifically the data portion of buffers manipulated by the xscale are
 * flushed before transmit and invalidated after receive. Flushing the buffers
 * before transmit ensures the NPE reads and transmits the correct data.
 * Invalidating the buffers after receive ensures the XScale reads
 * and processes the correct data.  In the case of the Codelet, all data is
 * flushed and invalidated as every byte is being written to on transmit
 * and every byte is verified on receive.  In a real application flushing
 * or invalidating all the data may not be necessary, only the data that
 * the application has written before transmit or will read after receive.
 * Note, regarding the packetised service, the IxHssAcc component itself
 * takes care of flushing and invalidating mbuf headers.  The application
 * needs only to concern itself with the mbuf data.
 * <P>
 * <B> IxFpathAcc: </B>
 * Buffers will be supplied at initialisation to IxFpathAcc to populate the
 * the fast path buffer free pool. No further buffers need be supplied to
 * IxFpathAcc there after.
 * <P>
 * <B> IxAtmdAcc: </B>
 * Buffers will be supplied at initialisation to IxFpathAcc to populate the
 * the rx free Q for each active VC. Buffers will be further replenished
 * to each rx VC free Q each time a Q becomes empty.
 * @note Each buffer has space reserved at the beginning to allow header
 * information to be prepended. An example of when this might be used is
 * in the routing demo where IP packets are prepended with ethernet frame
 * headers before being submitted to the Ethernet driver.
 *
 * <P>
 * <B> Connection Database:</B>
 * <P>
 * The demo codelet maintains a simple connection database so that multiple VCs
 * can be enabled simultaneously. Each vc has a VcNumber which identifies it.
 * The codelet is design to support a maximum of 8 VCs. The VcNumber is set
 * as the userId in any per VC callbacks set on IxFpathAcc or IxAtmdAcc.
 *
 * <B> System Configuration:</B>
 * <P>
 * The following is the static configuration of the system for both
 * bridged and routed demos. See IxFpathAccCodeletDefines.h for corresponding
 * #defines for some of the items below.
 * <UL>
 *   <LI> Utopia Interface Mode = MPhy, Bus Master.
 *   <LI> Utopia PHY Addr = 0,
 *   <LI> ATM Rx Free Q Buffer Replenishment = Triggered on Rx Free Q Empty.
 *   <LI> Ethernet Port         = IX_ETH_PORT_1.
 *   <LI> Ethernet Tx Priority  = IX_ETH_ACC_TX_DEFAULT_PRIORITY.
 *   <LI> Ethernet PHY Config   = 100Meg, Full Duplex, Auto-negtiate.
 *   <LI> Ethernet Port Mac Addr= 00:00:00:00:00:01
 *   <LI> QMgr Dispatch         = Executed from task
 *                              (Priority = IX_OSSL_THREAD_PRI_LOW)
 * </UL>
 * <B> Endianness:</B>
 * <P>
 * The demo codelet assumes that IP and MAC addresses are
 * formatted in network byte order. A UINT32 containing the value
 * 0x12345678 represents the IP address 12.34.56.78. A UINT8 array
 * containing the sequence {0x01,0x02,0x03,0x04,0x05,0x06} represents
 * the MAC address 01:02:03:04:05:06.
 *
 * <P>
 * <B> Source Code Overview:</B>
 * <P>
 * Brief description of the functionality of each of the src code demo files
 * <UL>
 *  <LI> <B> IxFpathAccCodeletMain.c:</B>
 *   <UL>
 *    <LI> Contains the main codelet initialisation routine.
 *    <LI> Implements the main show routine.
 *    <LI> Maintains the codelt connection table and statistics.
 *    <LI> Provides the primary ATM receive callback which strips aal5 PDU
 *         encapsulation and fowards the packet to an encapsulation specific
 *         handler in either the router or bridge demo files.
 *    <LI> Demonstrates the following IxFpathAcc API :-
 *         <OL>
 *           <LI> ixFpathAccInit()
 *           <LI> ixFpathAccBufferSupply()
 *           <LI> ixFpathAccBufferPoolInfoGet()
 *           <LI> ixFpathAccStatisticsGet()
 *         </OL>
 *   </UL>
 * <LI> <B> IxFpathAccCodeletTemplates.c: template generation utilities.</B>
 *  <UL>
 *   <LI> Contains the base templates for each the support encapsulation types.
 *   <LI> Engine to construct, store and print encapsulation specific templates
 *        for each connection.
 *  </UL>
 * <LI> <B> IxFpathAccCodeletRoutedDemo.c: </B> Routed VC demo.
 *  <UL>
 *   <LI> Flow learning and aging example.
 *   <LI> Demonstrates IxFpathAcc API :-
 *         <OL>
 *           <LI> ixFpathAccTemplatePairSet()
 *         </OL>
 *  </UL>
 * <LI> <B> IxFpathAccCodeletBridgeDemo.c: </B> Bridged VC demo
 *  <UL>
 *   <LI> Wan Mac learning example.
 *   <LI> Demonstrates IxFpathAcc API :-
 *         <OL>
 *           <LI> ixFpathAccTemplatePairSet()
 *         </OL>
 *  </UL>
 * <LI> <B> IxFpathAccCodeletBufMan.c:</B>
 *  <UL>
 *   <LI> Buffer pool creation.
 *   <LI> Buffer Alloc and Free routines.
 *  </UL>
 * <LI> <B> IxFpathAccCodeletUtils.c:</B>
 *  <UL>
 *   <LI> Configuration of s/w components other than IxFpathAcc.
 *   <LI> Routines to init and connect/disconnect to/from ATM and Utopia drivers
 *   <LI> Routines to init and connect to the Ethernet Driver
 *   <LI> Routines to Download Npe Firmware
 *  </UL>
 *  <LI> <B> IxFpathAccCodeletFPGA.c: </B>
 *  <UL>
 *   <LI> Fpga stimulus setup - to provide a clock src.
 *  </UL>
 * </UL>
 *
 * @{
 */

#ifndef IXFPATHACCCODELET_H
#define IXFPATHACCCODELET_H
#include "IxFpathAcc.h"

/**
 * @brief Start the Single VC Ethernet Bridge Fast Path Demo
 *
 * This function initialises the s/w components, sets up a single ATM VC and
 * configures the NPEs with a pair of fast path templates which support the
 * transportation of encapsulated ethernet frames on the Fast Path. Incoming
 * Encapsulated Ethernet Frames are classfied on the ATM NPE and
 * are directly passed to ETH NPE-B where they are modified and sent
 * over the Ethernet port.
 *
 * The demo automatically configures the classifier template with the WAN
 * MAC learning opcode so that a MAC database is populated with a log of
 * the last 16 (default) MAC addresses notified by the ATM NPE.
 *
 * Packets which fail classification are dropped in the codelet and the
 * buffer(s) associated with the packet is recycled to the buffer pool.
 *
 * Once this function completes, packets can be injected via the ATM equipment
 * and will be received on the Ethernet test equipment.
 *
 * <B> Configuration;</B>
 * <UL>
 *   <LI> VPI = 1
 *   <LI> VCI = 32
 *   <LI> VcNumber = 0 (codelet connection number)
 *   <LI> Encapsulation = VCMUX
 *   <LI> Max Encapsulated Ethernet Frame Size = 1516.
 *   <LI> Slow Path Packet Policy              = All slow path packets dropped.
 *   <LI> Wan Mac Addr database size           = 16
 *   <LI> Wan Mac Addr database display rate   = 5sec.
 * </UL>
 * <P>
 * <B> Bridge Data Flow:</B>
 * <P>
 * @verbatim
                                IXP4xx
   +----------------------------------------------------------------+
   |                            XSCALE                              |
   |                                                                |
   |                          +----------+ [Exception Packets       |
   |                          |  Demo    |  are are dropped]        |
   |                          |  Codelet | <----------------+       |
   |                          |----------|                  |       |
   |                          |  Init    |                  |       |
   |                          +----------+                  |       |
   |                  [Single      v                        |       |
   |                   Template    v                        |       |
   |                   Pair]       v                        |       |
   |                               v                        |       |
   | +----------+             +----------+            +----------+  |
   | | IxEthAcc |             |IxFpathAcc|            |IxAtmdAcc |  |
   | |          |             |          |            |          |  |
   | +----------+             +----------+            +----------+  |
   |                               v                        ^       |
   |           [Modifier Template] v[Classifier Template]   |       |
   |           +<<<<<<<<<<<<<<<<<<<+>>>>>>>>>>>>>>>>>>>>+   |       |
   |           v                                        v   |       |
   |...........v........................................v...|.......|
   |           v                 NPEs                   v   |       |
   |           v                                        v   |       |
   | +-----------+                                    +-----------+ |
   | |  NPE-B    |         [Classified Packets]       |  NPE-A    | |
   | |[Packet    |<-----------------------------------|[Packet    | |
   | | Modifier] |                                    |Classifier]| |
   | +-----------+                                    +-----------+ |
   |      |                                                 ^       |
   +------|-------------------------------------------------|-------+
          |                                                 |
          |                                                 |
          v                                                 |
     +-----------+                                    +-----------+
     | Ethernet  |                                    |ATM PHY    |
     |   PHY     |                                    |(Or Utopia)|
     +-----------+                                    +-----------+
           |                                                ^
           |                                                |
           v                                                |
     +-----------+                                    +-----------+
     | Ethernet  |                                    | ATM       |
     | Test      |                                    | Test      |
     | Equipment |                                    | Equipment |
     +-----------+                                    +-----------+

 * @endverbatim
 *
 */
void
ixFpCodeletEthBDemoStart( void );

/**
 * @brief Stop the Single VC Ethernet Bridge Fast Path Demo
 *
 * This function demonstrates the correct procedure for shutting down a
 * fast path VC in conjunction with disconnecting the VC from IxAtmdAcc.
 *
 */
void
ixFpCodeletEthBDemoStop( void );

/**
 * @brief Start the Single VC Routed Ip Fast Path Demo
 *
 * This function initialises the s/w components, sets up a single ATM VC and
 * configures the NPEs with a pair of fast path templates which support fast
 * path transportation of encapsulated Ip packets from aal5 to ethernet.
 *
 * Initially all IP packets will travel on the slow path until such time as
 * a packet flow is determined. The atomatic flow learning example function
 * reconfigures the classifier/modifier templates to allow packets for
 * learned destination IP address be forwarded on the fast path. The fast path
 * functionality supports 4 Ip flows per VC and this is demonstrated.
 *
 * Encapsulated IP packets are classfied on the ATM NPE and are directly passed
 * to ETH NPE-B where they are placed in Ethernet Frames and sent over
 * the Ethernet port.
 *
 * Packets which fail classification continue to the learning function and are
 * subsequently forwarded to the Ethernet driver (IxEthAcc) to a single fixed
 * MAC address.
 *
 * Once this function completes, packets can be injected via the ATM equipment
 * and will be received on the Ethernet test equipment.
 *
 * <B> Configuration;</B>
 * <UL>
 *   <LI> VPI = 1
 *   <LI> VCI = 33
 *   <LI> VcNumber = 1 (codelet connection number)
 *   <LI> Encapsulation = LLC/VCMUX (default).
 *   <LI> Max Encapsulated Ethernet Frame Size = Based on encapsulation.
 *   <LI> Slow Path Packet Policy = Packets have RFC1483/Aal5 encapsulation
 *        removed to IxEthAcc and are reencasulated in Ethernet frames then
 *        forwarded to IxEthAcc.
 *   <LI> Destination Mac for all IP packets = 00:00:00:00:00:03
 * </UL>
 *
 * <P>
 * <B> Routed Data Flow:</B>
 * <P>
 * @verbatim
                                IXP4xx
   +----------------------------------------------------------------+
   |                            XSCALE                              |
   | [Ethernet Encasulated                  [1483/Aal5 Encapsulated |
   |  IP Packets]             +----------+   Ip Packets]            |
   |                          |  Demo    |                          |
   |      +-------------------|  Codelet | <----------------+       |
   |      |                   |----------|                  |       |
   |      |                   | Flow     |                  |       |
   |      |                   | Learning&|                  |       |
   |      |                   | Aging    |                  |       |
   |      |                   +----------+                  |       |
   |      |                        v                        |       |
   |      |                        v [Template Pairs]       |       |
   |      v                        v                        |       |
   | +----------+             +----------+            +----------+  |
   | | IxEthAcc |             |IxFpathAcc|            |IxAtmdAcc |  |
   | |          |             |          |            |          |  |
   | +----------+             +----------+            +----------+  |
   |      |                        v                        ^       |
   |      |                        v                        |       |
   |      |    [Modifier Templates]v[Classifier Templates]  |       |
   |      |    +<<<<<<<<<<<<<<<<<<<+>>>>>>>>>>>>>>>>>>>>+   |       |
   |      |    v                                        v   |       |
   |......|....v........................................v...|.......|
   |      |    v                 NPEs                   v   |       |
   |      v    v                                        v   |       |
   | +-----------+        [4 IP Flows Per VC ]        +-----------+ |
   | |  NPE-B    |        [Classified Packets]        |  NPE-A    | |
   | |[Packet    |<-----------------------------------|[Packet    | |
   | | Modifier] |                                    |Classifier]| |
   | +-----------+                                    +-----------+ |
   |      |  SRC MAC=00:00:00:00:00:01                      ^       |
   +------|-------------------------------------------------|-------+
          |                                                 |
          |                                                 |
          v                                                 |
     +-----------+                                    +-----------+
     | Ethernet  |                                    |ATM PHY    |
     |   PHY     |                                    |(Or Utopia)|
     +-----------+                                    +-----------+
          |                                                 ^
          |                                                 |
          v                                                 |
     +-----------+                                    +-----------+
     | Ethernet  |                                    | ATM       |
     | Test      |                                    | Test      |
     | Equipment |                                    | Equipment |
     +-----------+                                    +-----------+

 @endverbatim
 *
 */
void
ixFpCodeletRoutedIpDemoStart( void );

/**
 * @brief Stop the Single VC Routed IP Fast Path Demo
 *
 * This function demonstrates the correct procedure for shutting down a
 * fast path VC in conjunction with disconnecting the VC from IxAtmdAcc.
 *
 */
void
ixFpCodeletRoutedIpDemoStop( void );

/**
 * @brief Display Demo Config and Statistics Per VC
 *
 * This function displays the configuration and statistics for each active VC.
 * It also demonstrates the use of the ixFpathAccStatisticsGet API function.
 */
void
ixFpathAccCodeletShow(void);

#endif /* IXFPATHACCCODELET_H */

/** @} defgroup IxFpathAccCodelet*/

/** @} defgroup Codelets*/
