/**       
 * @file  IxFpathAccCodeletBufMan.c
 * @date  June 23 2002 
 * @brief This file contains an implementation of an
 *        mbuf pool for use with the fpathAcc codelet.
 * 
 * -- Intel Copyright Notice --
 *
 * @par
 * INTEL CONFIDENTIAL
 *
 * @par
 * Copyright 2002 Intel Corporation All Rights Reserved.
 *
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.  The Material
 * contains trade secrets and proprietary and confidential information of
 * Intel or its suppliers and licensors.  The Material is protected by
 * worldwide copyright and trade secret laws and treaty provisions. No
 * part of the Material may be used, copied, reproduced, modified,
 * published, uploaded, posted, transmitted, distributed, or disclosed in
 * any way without Intel's prior express written permission.
 *
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel or otherwise.  Any license under
 * such intellectual property rights must be express and approved by
 * Intel in writing.
 *
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * -- End Intel Copyright Notice --
 */

#include <taskLib.h>
#include <logLib.h>
#include <ioLib.h>


#include "IxAssert.h"
#include "IxOsCacheMMU.h"
#include "IxOsBuffMgt.h"
#include "IxOsBuffPoolMgt.h"
#include "IxFpathAccCodeletDefines_p.h"

/**
 * The mbuf pool 
 */

__extension__ static struct {
    IX_MBUF_POOL *pool;
    UINT32 bufSize;
    UINT32 bufCount;
} ixFpCodeletNetPools[] =
    {
	{ NULL,   64,    50},
	{ NULL,  128,   100},
	{ NULL,  256,   150},
	{ NULL,  512,  4000},
	{ NULL,  1024,  200},
	{ NULL,  2048, 5200}
    };

/**
 * Initialise a pool of mbufs for use by the Access layer components
 * used in the fpathAcc codelet
 */
IX_STATUS
ixFpCodeletBufManMbufPoolInit(void)
{
    int i;
    int poolNumEnt = NELEMENTS(ixFpCodeletNetPools);

    /* Setup each pool */
    for (i = 0; i < poolNumEnt; i++)
    {
        IX_MBUF_POOL_INIT(&ixFpCodeletNetPools[i].pool, 
			  ixFpCodeletNetPools[i].bufCount,
			  ixFpCodeletNetPools[i].bufSize,
			  "Fpath codelet pool");
    } /* end of for(i) */
    
    return IX_SUCCESS;
}

/**
 * @brief Get an mbuf back from the pool
 *
 * Get an mbuf from the pool and initialise the necessary fields.
 *
 */
IX_MBUF*
ixFpCodeletBufManMbufGet(unsigned int size)
{
    IX_MBUF *buf;
    int i;
    int poolNumEnt = NELEMENTS(ixFpCodeletNetPools);

    for (i = 0; i < poolNumEnt; i++)
    {
	if (ixFpCodeletNetPools[i].bufSize >= size)
	{
	    IX_MBUF_POOL_GET(ixFpCodeletNetPools[i].pool, &buf);
	    
	    if (buf)
	    {
		IX_MBUF_MLEN(buf) = size;
		IX_MBUF_NEXT_BUFFER_IN_PKT_PTR(buf) = NULL;
		IX_MBUF_NEXT_PKT_IN_CHAIN_PTR(buf) = NULL;

		/* N.B. Adjust the mData to make room for 
		   the addition of extra header information.
		*/
		IX_MBUF_POOL_MDATA_RESET(buf);
		IX_MBUF_MDATA(buf) += MBUF_HEADROOM;
		IX_MBUF_PKT_LEN(buf) = size;
                IX_ACC_DATA_CACHE_INVALIDATE(IX_MBUF_MDATA(buf), IX_MBUF_MLEN(buf));

		return buf;
	    }
	}
    }

    return NULL;
}

/** 
 * @brief Free the mbuf back to the pool.
 */
void    
ixFpCodeletBufManMbufFree(IX_MBUF* buf)
{
    IX_MBUF_POOL_PUT_CHAIN(buf);
}


