/**
 * @file IxHssAccCodelet.c
 *
 * @date 26 Mar 2002
 *
 * @brief This file contains the main routine for the HSS Access Codelet.
 *
 * -- Intel Copyright Notice --
 *
 * @par
 * INTEL CONFIDENTIAL
 *
 * @par
 * Copyright 2002 Intel Corporation All Rights Reserved.
 *
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.  The Material
 * contains trade secrets and proprietary and confidential information of
 * Intel or its suppliers and licensors.  The Material is protected by
 * worldwide copyright and trade secret laws and treaty provisions. No
 * part of the Material may be used, copied, reproduced, modified,
 * published, uploaded, posted, transmitted, distributed, or disclosed in
 * any way without Intel's prior express written permission.
 *
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel or otherwise.  Any license under
 * such intellectual property rights must be express and approved by
 * Intel in writing.
 *
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * -- End Intel Copyright Notice --
 *
 * @sa IxHssAccCodelet.h
 */

/*
 * Put the system defined include files required.
 */

#include <ioLib.h>
#include <logLib.h>
#include <stdio.h>

/*
 * Put the user defined include files required.
 */

#include "ixp425.h"

#include "IxHssAcc.h"
#include "IxNpeDl.h"
#include "IxNpeMh.h"
#include "IxOsServices.h"
#include "IxQMgr.h"
#include "ix_ossl.h"

#include "IxDemoUtils.h"

#include "IxHssAccCodelet.h"

/*
 * #defines and macros used in this file.
 */

/* #define NO_CHANNELISED */
/* #define NO_PACKETISED */

/*
 * Typedefs whose scope is limited to this file.
 */

/*
 * Variable declarations global to this file only.  Externs are followed by
 * static variables.
 */

BOOL stopTest;

/*
 * Extern function prototypes.
 */

/*
 * Static function prototypes.
 */

/**
 * @fn void ixHssAccCodeletTestServices (
           IxHssAccHssPort hssPortId,
           BOOL codeletLoopback,
           BOOL hssLoopback)
 *
 * @param IxHssAccHssPort hssPortId (in) - the HSS port ID (0 or 1).
 * @param BOOL codeletLoopback (in) - the Codelet loopback state.
 * @param BOOL hssLoopback (in) - the HSS loopback state.
 *
 * Four clients of the Packetised Service will be tested simultaneously,
 * one client per T1/E1 stream.
 * <P>
 * One client of the Channelised Service will be tested, using 32 timeslots
 * spread across the four T1/E1 streams to carry a combination of both
 * 64Kbps and 56Kbps voice traffic.
 * <P>
 * For each client, the service will be started, exercised for a number of
 * iterations, and then stopped.  After stopping the service the transmit
 * and receive statistics, as well as any HSS errors, will be reported.
 */

PRIVATE
ix_error
ixHssAccCodeletThreadMain (
    void* arg,
    void** ptrRetObj)
{
    IxHssAccHssPort hssPortId = (IxHssAccHssPort)arg;
    IxHssAccHdlcPort hdlcPortId;
    unsigned i;

    /**********************/
    /* CONFIGURE HSS PORT */
    /**********************/

    /* configure the HSS port - note the loopback flag signifies whether */
    /* or not the Codelet performs a loopback, we want the hardware to */
    /* do the reverse */
    ixHssAccCodeletConfigure (hssPortId);

    /******************/
    /* START SERVICES */
    /******************/

    /* start the Packetised Service for all clients */
    for (hdlcPortId = 0; hdlcPortId < IX_HSSACC_HDLC_PORT_MAX;
         hdlcPortId++)
    {
#ifndef NO_PACKETISED
        ixHssAccCodeletPacketisedServiceStart (hssPortId, hdlcPortId);
#endif
    }

    /* start the Channelised Service for the single client */
#ifndef NO_CHANNELISED
    ixHssAccCodeletChannelisedServiceStart (hssPortId);
#endif

    /*********************/
    /* EXERCISE SERVICES */
    /*********************/

    /* enter a 10 second (50 * 200ms) loop to exercise the services */
    /* if stopTest is false then keep running (set to true to stop test) */
    for (i = 0; (i < 50) || (!stopTest); i++)
    {
        /* let each packetised client exercise the Packetised Service */
        for (hdlcPortId = 0; hdlcPortId < IX_HSSACC_HDLC_PORT_MAX;
             hdlcPortId++)
        {
#ifndef NO_PACKETISED
            ixHssAccCodeletPacketisedServiceRun (hssPortId, hdlcPortId);
#endif
        }

        /* let the channelised client exercise the Channelised Service */
#ifndef NO_CHANNELISED
        ixHssAccCodeletChannelisedServiceRun (hssPortId);
#endif

        /* sleep for 200ms */
        ixOsServTaskSleep (200);
    } /* for (i ... */

    /*****************/
    /* STOP SERVICES */
    /*****************/

    /* stop the Packetised Service for all clients */
    for (hdlcPortId = 0; hdlcPortId < IX_HSSACC_HDLC_PORT_MAX;
         hdlcPortId++)
    {
#ifndef NO_PACKETISED
        ixHssAccCodeletPacketisedServiceStop (hssPortId, hdlcPortId);
#endif
    }

    /* stop the Channelised Service for the single client */
#ifndef NO_CHANNELISED
    ixHssAccCodeletChannelisedServiceStop (hssPortId);
#endif

    ixHssAccCodeletShow(hssPortId);

    return IX_OSSL_ERROR_SUCCESS;
}

/*
 * Function definition: ixHssAccCodeletInit
 */

void
ixHssAccCodeletInit (
    BOOL runForever)
{
    IX_STATUS status;
    ix_ossl_thread_t threadId;

    /* initialise the HssAcc component */
    status = ixHssAccInit ();

    /* if there was any problem then print an error */
    if (status != IX_SUCCESS)
    {
        printf ("ixHssAccInit failed\n");
    }

    ixHssAccCodeletCodeletLoopbackSet (FALSE);
    ixHssAccCodeletHssLoopbackSet (TRUE);

    /* should we stop the test automatically after some time? */
    stopTest = !runForever;

    /* start the main thread */
    (void) ix_ossl_thread_create (
        ixHssAccCodeletThreadMain, /* entryPoint */
        IX_HSSACC_HSS_PORT_0,      /* arg */
        &threadId);                /* ptrTid */

    (void) ix_ossl_thread_set_priority (
        threadId,                /* tid */
        IX_OSSL_THREAD_PRI_LOW); /* priority */
}

/*
 * Function definition: test_main
 */

int
test_main (void)
{
    IX_STATUS status;

    printf ("Setting the debug level & redirecting output\n");

    /* set the debug level */
    logInit (ioTaskStdGet (0,1), 1);
    logFdSet (ioTaskStdGet (0,1));
    ixOsServLogLevelSet (LOG_ERROR); /* LOG_ALL, LOG_MESSAGE, ... */

    printf ("Performing ixDemoUtilsNpeImageDownload\n");

    /* download software to NPE-A */
    status = ixDemoUtilsNpeImageDownload (
        IX_NPEDL_NPEID_NPEA,      /* npeId */
        1);                       /* buildId */

    /* if there was any problem then print an error */
    if (status != IX_SUCCESS)
    {
        printf ("Failed to download image to NPE\n");
    }

    printf ("Performing ixNpeDlNpeExecutionStart\n");

    /* start the NPE execution */
    status = ixNpeDlNpeExecutionStart (IX_NPEDL_NPEID_NPEA);
    if (status != IX_SUCCESS)
    {
        printf ("Failed to start NPE execution\n");
    }

    /* initialise the IxNpeMh component */
    status = ixNpeMhInitialize (IX_NPEMH_NPEINTERRUPTS_YES);

    /* if there was any problem then print an error */
    if (status != IX_SUCCESS)
    {
        printf ("ixNpeMhInitialize failed\n");
    }

    /* initialise the QMgr component */
    status = ixQMgrInit ();

    /* if there was any problem then print an error */
    if (status != IX_SUCCESS)
    {
        printf ("ixQMgrInit failed\n");
    }

    /* Hook the QM QUELOW dispatcher to interrupt */
    status = ixOsServIntBind (
        IXP425_INT_LVL_QM1,               /* vector */
        (IxVoidFnPtr) ixQMgrDispatcherLoopRun, /* routine */
        (void *) IX_QMGR_QUELOW_GROUP);   /* parameter */

    if (status != IX_SUCCESS)
    {
        printf ("ixOsServIntBind failed to bind to QM1 interrupt\n");
    }

    /* initialise the codelet itself and start the demo */
    ixHssAccCodeletInit (
        FALSE); /* runForever */

    return 0;
}
