/* 
 * @file    IxOamCodelet.c
 *
 * @date    21-July-2002
 *
 * @brief   The module implements the loopback functionality
 *          functions of the Oam Codelet.
 *
 *
 * IxOamCodelet API functions:
 *       ixOamCodeletInit
 *       ixOamCodeletOamF5EtePing
 *       ixOamCodeletOamF5SegPing
 *       ixOamCodeletOamF4EtePing
 *       ixOamCodeletOamF4SegPing
 *       ixOamCodeletShow
 *
 * @par
 * INTEL CONFIDENTIAL
 *
 * @par
 * Copyright 2002 Intel Corporation All Rights Reserved.
 *
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.  The Material
 * contains trade secrets and proprietary and confidential information of
 * Intel or its suppliers and licensors.  The Material is protected by
 * worldwide copyright and trade secret laws and treaty provisions. No
 * part of the Material may be used, copied, reproduced, modified,
 * published, uploaded, posted, transmitted, distributed, or disclosed in
 * any way without Intel's prior express written permission.
 *
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel or otherwise.  Any license under
 * such intellectual property rights must be express and approved by
 * Intel in writing.
 *
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * -- End Intel Copyright Notice --
 * 
 */

/*
 * Put the system defined include files required
 */

#include <string.h>

/*
 * Put the user defined include files required
 */
#include "IxTypes.h"
#include "IxAtmTypes.h"
#include "ix_ossl.h"
#include "IxAssert.h"
#include "IxAtmdAcc.h"
#include "IxAtmm.h"
#include "IxDemoUtils.h"
#include "IxOsServices.h"
#include "IxOamITU610Types_p.h"
#include "IxOamCodelet.h"

/*
 * #defines and macros used in this file.
 */

/* IRQ safe logs, for use in callbacks which could be in the context of an IRQ */
#define IX_OAM_CODELET_IRQ_SAFE_LOG(msg, param1, param2, param3, param4) \
 (void) ixOsServLog(LOG_USER, msg, param1, param2, param3, param4, 0, 0);
#define IX_OAM_CODELET_IRQ_SAFE_LOG_ERROR(msg) \
 (void) ixOsServLog (LOG_ERROR, msg, 0, 0, 0, 0, 0, 0);


/* Normal Log */
#define IX_OAM_CODELET_LOG printf
#define IX_OAM_CODELET_LOG_ERROR printf


#define IX_OAM_REPLENISH_WATERMARK  2

/*
 * Typedefs whose scope is limited to this file.
 */

/*
 * Global Variables
 */
static UINT32 lbCorrelationTag = 0; /* Correlation Tag of Loopback in progress, see I-610 */

static unsigned int lbVci = 0;    /* VCI of Loopback in progress */
static unsigned int lbVpi = 0;    /* VPI of Loopback in progress */
static unsigned int lbPti = 0;    /* PTI of Loopback in progress */
static unsigned int lbPort = 0;   /* Port of Loopback in progress */

/* Cell statistics */
static unsigned int parentLbCellTxCount[IX_UTOPIA_MAX_PORTS]; /* Count of parent cells sent */
static unsigned int childLbCellTxCount[IX_UTOPIA_MAX_PORTS]; /* Count of LB child cells sent */

static unsigned int parentLbCellRxCount[IX_UTOPIA_MAX_PORTS]; /* Count of Rx Parent Loopback cells, 
							  i.e. Loopbacks initiated at far end */

static unsigned int childLbCellRxCount[IX_UTOPIA_MAX_PORTS]; /* Count of expected Rx Child Loopback cells */

static unsigned int childLbCellRxErrCount[IX_UTOPIA_MAX_PORTS]; /* Count of unexpected Rx Child Loopback cells, 
							    e.g. loopback not in progress, wrong VPI, VCI etc. */

static unsigned int parentLbCellRxErrCount[IX_UTOPIA_MAX_PORTS]; /* Count of unexpected Rx Parent Loopback cells,
							     e.g. LLID didn't match this CPID */

static unsigned int unsupportedOamRxCount[IX_UTOPIA_MAX_PORTS]; /* Count of OAM Rx cells that are not supported,
							    i.e. any cells != OAM LB */

static unsigned int txDoneCallbackCount[IX_UTOPIA_MAX_PORTS]; /* Count of tx done callbacks for a port */
static unsigned int replenishCallbackCount; /* Count of Rx Q replenish callbacks, all ports */

/* This CPID is set to all 1's, i.e. no specific coding structure I-610 */
static UINT8 oamCpid[IX_OAM_ITU610_LOCATION_ID_LEN] = {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,  0xFF, 0xFF, 
				  0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};

/* A location Id of all 1's corresponds to end point */
static UINT8 allOnesLocId[IX_OAM_ITU610_LOCATION_ID_LEN] = {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,  0xFF, 0xFF, 
				       0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};

/* A location Id of all 0's corresponds to segment end point */
static UINT8 allZerosLocId[IX_OAM_ITU610_LOCATION_ID_LEN] = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
					0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};

static unsigned int ixOamCodeletNumPortsConfigured = 0;

static IxAtmConnId oamRxConnId;
static IxAtmConnId oamTxConnId[IX_UTOPIA_MAX_PORTS];
static IxAtmdAccUserId oamRxUserId = 99; /* For debug purposes */

static BOOL ixOamCodeletInitialized = FALSE;


/*
 * Static function prototypes
 */
PRIVATE void
ixOamChildLbCellRx (IxAtmLogicalPort port, IX_MBUF *rxMbuf);

PRIVATE IX_STATUS
ixOamParentLbCellTx (IxAtmLogicalPort port, unsigned int vpi, unsigned int vci, unsigned int pti);

PRIVATE IX_STATUS
ixOamChildLbCellTx (IxAtmLogicalPort port, IX_MBUF *rxMbuf);

PRIVATE void
ixOamParentLbCellRx (IxAtmLogicalPort port, IX_MBUF *rxMbuf);

PRIVATE IX_STATUS
ixOamRxLlidCheck (IxOamITU610Cell *oamCell);

PRIVATE void
ixOamCellRxCallback (IxAtmLogicalPort port,
		     IxAtmdAccUserId userId,
		     IxAtmdAccPduStatus status,
		     IxAtmdAccClpStatus clp,
		     IX_MBUF *rxMbuf);

PRIVATE IX_STATUS
ixOamCodeletOamPing(IxAtmLogicalPort port, 
		    unsigned int vpi,
		    unsigned int vci, 
		    unsigned int pti,
		    unsigned int numCells);

PRIVATE void
ixOamTxDoneCallback (IxAtmdAccUserId userId,
		     IX_MBUF *mbufPtr);

PRIVATE void
ixOamRxFreeLowReplenishCallback (IxAtmdAccUserId userId);


PUBLIC IX_STATUS
ixOamCodeletInit (unsigned int numPorts)
{
    IX_STATUS retval;
    unsigned int txRate[IX_UTOPIA_MAX_PORTS];
    unsigned int rxRate;    
    IxAtmmVc txVc;
    IxAtmmVc rxVc;
    IxAtmLogicalPort port;
    IxAtmSchedulerVcId rxVcId;
    IxAtmSchedulerVcId txVcId;
    IxAtmNpeRxVcId npeRxVcId;

    if (ixOamCodeletInitialized)
    {
	IX_OAM_CODELET_LOG_ERROR("Already initialized");
	return IX_FAIL;
    }

    /* Check parameters */
    if (numPorts < 1 || numPorts > IX_UTOPIA_MAX_PORTS)
    {
	IX_OAM_CODELET_LOG_ERROR("ixOamCodeletInit(): numPorts (%u) invalid\n", numPorts);
	return IX_FAIL;
    }

    /* Check how many ports have been configured, rxRate not used */
    for (port =0; port<numPorts; port++)
    {
	if (ixAtmmPortQuery (port, &txRate[port], &rxRate) != IX_SUCCESS)
	{
	    IX_OAM_CODELET_LOG_ERROR("ixOamCodeletInit(): numPorts (%u) not configured, %u ports configured\n", 
				     numPorts, port);	    
	}
    }

    /* Initialize stats */
    for (port = 0; port < IX_UTOPIA_MAX_PORTS; port++)
    {
	parentLbCellTxCount[port] = 0;
	childLbCellTxCount[port] = 0;
	parentLbCellRxCount[port] = 0;
	childLbCellRxCount[port] = 0;
	childLbCellRxErrCount[port] = 0;
	parentLbCellRxErrCount[port] = 0;
	unsupportedOamRxCount[port] = 0;
	txDoneCallbackCount[port] = 0;	
    }
    replenishCallbackCount = 0; /* 1 for all ports */

    /* Setup the OAM Tx Port Channels */
    for (port=0; port<numPorts; port++)
    {
	memset(&txVc, 0, sizeof(txVc));

	/* Setup Tx Vc descriptor */
	txVc.vpi = IX_ATMDACC_OAM_TX_VPI;
	txVc.vci = IX_ATMDACC_OAM_TX_VCI;
	txVc.direction = IX_ATMM_VC_DIRECTION_TX;
	txVc.trafficDesc.atmService = IX_ATM_UBR;
	txVc.trafficDesc.pcr = txRate[port];
	
	/* Setup tx VC, N.B. TxVcId not used in this codelet, 
	 * would typically be used for Vc Deregister
	 */
	retval = ixAtmmVcRegister (port,
				   &txVc,
				   &txVcId);

	if (retval != IX_SUCCESS)
	{
	    IX_OAM_CODELET_LOG_ERROR("Failed to register Tx Port VC\n");
	    return IX_FAIL;
	}


	retval = ixAtmdAccTxVcConnect (port,
				       IX_ATMDACC_OAM_TX_VPI,
				       IX_ATMDACC_OAM_TX_VCI,
				       IX_ATMDACC_OAM,
				       port, /* set userId to port */
				       ixOamTxDoneCallback,
				       &oamTxConnId[port]);

	if (retval != IX_SUCCESS)
	{
	    IX_OAM_CODELET_LOG_ERROR ("Failed to connect Tx Port Channel\n");
	    return IX_FAIL;
	}
    }

    /* 
     * Setup OAM Rx Channel
     * N.B. OAM traffic for all ports is received on this VC
     */

    memset(&rxVc, 0, sizeof(rxVc));

    /* Setup Rx Vc descriptor */
    rxVc.vpi = IX_ATMDACC_OAM_RX_VPI;
    rxVc.vci = IX_ATMDACC_OAM_RX_VCI;
    rxVc.direction = IX_ATMM_VC_DIRECTION_RX;
    rxVc.trafficDesc.atmService = IX_ATM_UBR;
    rxVc.trafficDesc.pcr = 0; /* Ignored for Rx */

    /* Setup rx VC, N.B. RxVcId not used in this codelet, 
     * would typically be used for Vc Deregister
     */
    retval = ixAtmmVcRegister (IX_ATMDACC_OAM_RX_PORT,
			       &rxVc,
			       &rxVcId);
    
    if (retval != IX_SUCCESS)
    {
	IX_OAM_CODELET_LOG_ERROR("Failed to register Rx Channel\n");
	return IX_FAIL;
    }

    /* Connect Rx VC, N.B. npeRxVcId not again,
     * this would typically be used for FastPath APIS
     * which are not relevant to OAM
     */
    retval = ixAtmdAccRxVcConnect (IX_ATMDACC_OAM_RX_PORT,
				   IX_ATMDACC_OAM_RX_VPI,
				   IX_ATMDACC_OAM_RX_VCI,
				   IX_ATMDACC_OAM,
				   IX_ATM_RX_B, /* low priority Q */
				   oamRxUserId,
				   ixOamCellRxCallback,
				   IX_ATMDACC_DEFAULT_REPLENISH_COUNT,
				   &oamRxConnId,
				   &npeRxVcId);

    if (retval != IX_SUCCESS)
    {
	IX_OAM_CODELET_LOG_ERROR("Failed to connect Rx Channel\n");
	return IX_FAIL;
    }

    /* Replenish Rx then register replenish callback */
    ixOamRxFreeLowReplenishCallback (oamRxUserId);
    
    retval = ixAtmdAccRxVcFreeLowCallbackRegister (oamRxConnId,
						   IX_OAM_REPLENISH_WATERMARK,
						   ixOamRxFreeLowReplenishCallback);

    if (retval != IX_SUCCESS)
    {
	IX_OAM_CODELET_LOG_ERROR("Failed to register Rx replenish callback\n");
	return IX_FAIL;
    }

    /* enable the oam vc */
    retval = ixAtmdAccRxVcEnable (oamRxConnId);

    if (retval != IX_SUCCESS)
    {
	ixOsServLog(LOG_ERROR, "Failed to enable Rx VC \n", 0, 0, 0, 0, 0, 0);
	return retval;
    }

    /* Set the number of ports configured */
    ixOamCodeletNumPortsConfigured = numPorts;

    /* set initialised flag */
    ixOamCodeletInitialized = TRUE;

    IX_OAM_CODELET_LOG ("Initialization Phase Complete\n");

    return IX_SUCCESS;
}

PUBLIC IX_STATUS
ixOamCodeletOamF5EtePing(IxAtmLogicalPort port, 
			 unsigned int vpi,
			 unsigned int vci, 
			 unsigned int numCells)
{
    return ixOamCodeletOamPing(port,
			       vpi,
			       vci,
			       IX_OAM_ITU610_F5_ETE_PTI,
			       numCells);
}

PUBLIC IX_STATUS
ixOamCodeletOamF5SegPing(IxAtmLogicalPort port, 
			 unsigned int vpi,
			 unsigned int vci, 
			 unsigned int numCells)
{
    return ixOamCodeletOamPing(port,
			       vpi,
			       vci,
			       IX_OAM_ITU610_F5_SEG_PTI,
			       numCells);
}


PUBLIC IX_STATUS
ixOamCodeletOamF4EtePing(IxAtmLogicalPort port, 
			 unsigned int vpi,
			 unsigned int numCells)
{
    return ixOamCodeletOamPing(port,
			       vpi,
			       IX_OAM_ITU610_F4_ETE_VCI,
			       0, /* PTI is don't care for F4, use 0 */
			       numCells);
}

PUBLIC IX_STATUS
ixOamCodeletOamF4SegPing(IxAtmLogicalPort port, 
			 unsigned int vpi,
			 unsigned int numCells)
{
    return ixOamCodeletOamPing(port,
			       vpi,
			       IX_OAM_ITU610_F4_SEG_VCI,
			       0, /* PTI is don't care for F4, use 0 */
			       numCells);
}



PRIVATE IX_STATUS
ixOamCodeletOamPing(IxAtmLogicalPort port, 
		    unsigned int vpi,
		    unsigned int vci, 
		    unsigned int pti,
		    unsigned int numCells)
{
    unsigned int sendCount = 0;
    IX_STATUS retval = IX_FAIL;
    ix_ossl_error_code osRetval;

    if (!ixOamCodeletInitialized)
    {
	IX_OAM_CODELET_LOG_ERROR("Codelet not initialized\n");
	return IX_FAIL;
    }

    if (port >= ixOamCodeletNumPortsConfigured)
    {
	IX_OAM_CODELET_LOG_ERROR("Port not configured\n");
	return IX_FAIL;
    }

    if (numCells == 0)
    {
	IX_OAM_CODELET_LOG_ERROR("numCells should be > 0\n");
	return IX_FAIL;
    }

    /* Setup LB info */
    lbVpi = vpi;
    lbVci = vci;
    lbPti = pti;
    lbPort = port;

    while (numCells--)
    {
	/* Notify user that a loopback cell is being sent */
	IX_OAM_CODELET_LOG ("Sending loopback cell %u\n", sendCount++);

	/* Send the first cell */
	retval = ixOamParentLbCellTx (port, vpi, vci, pti);

	if (retval != IX_SUCCESS)
	{
	    IX_OAM_CODELET_LOG_ERROR("Failed to send parent loopback cell #%u",sendCount);
	    return IX_FAIL;
	}

	/*
	 * A message is displayed at the windSh  if a correct loopback child cell has been received in
	 * response to the sending of the parent cell. No message is displayed if no correct child cell
	 * has been received
	 */

	/* Wait for 5 seconds for response, see ITU-610 */
	osRetval = ix_ossl_sleep (IX_OAM_ITU610_LB_TIMEOUT_PERIOD_MSECS);
	IX_ASSERT(osRetval == IX_OSSL_ERROR_SUCCESS);
    }

    return IX_SUCCESS;
}

void
ixOamCodeletShow (void)
{
    unsigned int i;

    IX_OAM_CODELET_LOG("replenishCallbackCount = %u\n",
		       replenishCallbackCount);

    for (i = 0; i < ixOamCodeletNumPortsConfigured; i++)
    {
	IX_OAM_CODELET_LOG("OAM codelet stats for Port %u\n", i);

	IX_OAM_CODELET_LOG("parentLbCellTxCount[i] = %u\n",
			   parentLbCellTxCount[i]);
	IX_OAM_CODELET_LOG("childLbCellTxCount[i] = %u\n",
			   childLbCellTxCount[i]);
	IX_OAM_CODELET_LOG("parentLbCellRxCount[i] = %u\n",
			   parentLbCellRxCount[i]);
	IX_OAM_CODELET_LOG("childLbCellRxCount[i] = %u\n",
			   childLbCellRxCount[i]);
	IX_OAM_CODELET_LOG("childLbCellRxErrCount[i] = %u\n",
			   childLbCellRxErrCount[i]);
	IX_OAM_CODELET_LOG("parentLbCellRxErrCount[i] = %u\n",
			   parentLbCellRxErrCount[i]);
	IX_OAM_CODELET_LOG("unsupportedOamRxCount[i] = %u\n",
			   unsupportedOamRxCount[i]);
	IX_OAM_CODELET_LOG("txDoneCallbackCount[i] = %u\n",
			   txDoneCallbackCount[i]);

	IX_OAM_CODELET_LOG("\n");
    }
}

PRIVATE void
ixOamCellRxCallback (IxAtmLogicalPort port,
		     IxAtmdAccUserId userId,
		     IxAtmdAccPduStatus status,
		     IxAtmdAccClpStatus clp,
		     IX_MBUF *rxMbuf)
{
    IxOamITU610Cell *rxCell;
    unsigned int pti;
    unsigned int vci;

    IX_ASSERT(userId == oamRxUserId);
    
    /* Status should always be OAM Valid */
    IX_ASSERT(status == IX_ATMDACC_OAM_VALID);

    rxCell = (IxOamITU610Cell *)(rxMbuf->m_data);

    /* Read the VCI & Payload Type */
    vci = rxCell->header.vci;
    pti = rxCell->header.pti;

    /* Ensure access layer delivered a correct OAM cell */
    IX_ASSERT ((vci == IX_OAM_ITU610_F4_SEG_VCI) || (vci == IX_OAM_ITU610_F4_ETE_VCI) || 
	       (pti == IX_OAM_ITU610_F5_ETE_PTI) || (pti == IX_OAM_ITU610_F5_SEG_PTI));

    /* Is it an Loopback cell ? */
    if ((rxCell->payload.genericPayload.oamTypeAndFunction == IX_OAM_ITU610_TYPE_FAULT_MAN_LB))
    {
	/* Is Parent Loopback Indication field in cell set? */
	if (rxCell->payload.lbPayload.loopbackIndication == IX_OAM_ITU610_LB_INDICATION_PARENT)
	{
	    ixOamParentLbCellRx (port, rxMbuf);
	}
	/* Otherwise child cell, i.e. response to loopback cell sent earlier */
	else 
	{
	    ixOamChildLbCellRx(port, rxMbuf);
	}
    }
    else
    {
	unsupportedOamRxCount[port]++;
	/* free the buffer */
	ixDemoUtilsMbufFree (rxMbuf);
    }
}    


PRIVATE void
ixOamChildLbCellRx (IxAtmLogicalPort port, IX_MBUF *rxMbuf)
{
    unsigned int pti;
    unsigned int vpi;
    unsigned int vci;
    IxOamITU610Cell *rxCell;
    IxOamITU610LbPayload *lbPayload;

    rxCell = (IxOamITU610Cell *)(rxMbuf->m_data);

    /* Read the VPI, VCI & Payload Type */
    vci = rxCell->header.vci;
    vpi = rxCell->header.vpi;
    pti = rxCell->header.pti;

    /* Check if the VPI, VCI and Port match */
    if ((vpi == lbVpi) && (vci == lbVci) && (port == lbPort))
    {
	/* Check if F4 flow (F4 VCI) or PTI matches */
	if ((vci == IX_OAM_ITU610_F4_SEG_VCI) || (vci == IX_OAM_ITU610_F4_ETE_VCI) || (pti == lbPti))
	{
	    /* Setup the data pointers */
	    lbPayload = &(((IxOamITU610Cell *)(rxMbuf->m_data))->payload.lbPayload);
	    
	    /* Source Id correct, all Ones supported only in this codelet */
	    if (memcmp(lbPayload->sourceId, allOnesLocId, IX_OAM_ITU610_LOCATION_ID_LEN) == 0)
	    {
		/* Correlation tag correct? */
		if (memcmp(lbPayload->correlationTag, &lbCorrelationTag, IX_OAM_ITU610_LB_CORRELATION_TAG_LEN) == 0)
		{
		    /* Notify user that a ping has been successful,
		     * N.B. the user is not notified if ping has failed
		     */
		    IX_OAM_CODELET_IRQ_SAFE_LOG ("OAM ping reply for VPI:%u VCI:%u PTI:%u on Port:%u",
						 lbVpi, lbVci, lbPti, lbPort);

		    /* free the buffer */
		    ixDemoUtilsMbufFree (rxMbuf);
		    childLbCellRxCount[port]++;
		    return;
		}
	    }
	}
    }

    /* free the buffer */
    ixDemoUtilsMbufFree (rxMbuf);
    childLbCellRxErrCount[port]++;
}

/*
 *
 * Transmit loopback parent cell
 */
PRIVATE IX_STATUS
ixOamParentLbCellTx (IxAtmLogicalPort port, unsigned int vpi, unsigned int vci, unsigned int pti)
{    
    IxOamITU610Cell *txCell;
    IxOamITU610LbPayload *lbPayload;
    IX_MBUF *txMbuf;
    IX_STATUS retval = IX_FAIL;
    unsigned int lockKey;

    /* Get a Tx mbuf */
    ixDemoUtilsMbufGet(IX_ATM_OAM_CELL_SIZE_NO_HEC, &txMbuf);
    if (txMbuf != NULL)
    {		
	/* set the packet header len */
	txMbuf->m_pkthdr.len = txMbuf->m_len;

	txCell = (IxOamITU610Cell *)(txMbuf->m_data);

	/* Setup shortcut pointers */
	lbPayload = &(txCell->payload.lbPayload);
	
	/* Set the OAM function type to LB */
	lbPayload->oamTypeAndFunction = IX_OAM_ITU610_TYPE_FAULT_MAN_LB;
	
	/* Setup the loopback indication */
	lbPayload->loopbackIndication = IX_OAM_ITU610_LB_INDICATION_PARENT;
	
	/* Increment the correlation tag and write this into the cell */
 	lbCorrelationTag++;
	memcpy(lbPayload->correlationTag, &lbCorrelationTag, IX_OAM_ITU610_LB_CORRELATION_TAG_LEN);
	
	/* Set Loopback Location Id */
	memcpy(lbPayload->llid, allOnesLocId, IX_OAM_ITU610_LOCATION_ID_LEN);
	
	/* Store the source Id as the CPID */
	memcpy(lbPayload->sourceId, oamCpid, IX_OAM_ITU610_LOCATION_ID_LEN);
	
	/* Set the reserved fields to 0x6a */
	memset(lbPayload->reserved, 
	       IX_OAM_ITU610_RESERVED_BYTE_VALUE, 
	       IX_OAM_ITU610_LB_RESERVED_BYTES_LEN);
	
	/* Set the transmit VPI */
	txCell->header.vpi = vpi;

	/* Set the transmit VCI TX Vci */
	txCell->header.vci = vci;
	
	/* Set the PTI */
	txCell->header.pti = pti;

	/* Transmit the cell, not reentrant on a VC basis so protect */
	lockKey = ixOsServIntLock();
	retval = ixAtmdAccTxVcPduSubmit (oamTxConnId[port],
					 txMbuf,
					 0, /* CLP 0 */
					 1); /* Cells per packet */
	ixOsServIntUnlock (lockKey);

	if (retval == IX_SUCCESS)
	{
	    /* Increment the loopback out counter */
	    parentLbCellTxCount[port]++;
	}
	else
	{
	    /* Release the MBUF */
	    ixDemoUtilsMbufFree (txMbuf);
	}
    }
    else
    {
	IX_OAM_CODELET_IRQ_SAFE_LOG_ERROR("ixOamParentLbCellTx: Failed to get a buffer for transmit");
    }

    return retval;
}

PRIVATE void
ixOamParentLbCellRx (IxAtmLogicalPort port, IX_MBUF *rxMbuf)
{     
    /* Loopback location ID correct? */
    if (ixOamRxLlidCheck (((IxOamITU610Cell *)(rxMbuf->m_data))) == IX_SUCCESS)
    {
	parentLbCellRxCount[port]++;

	/* Send loopback child cell back to originator */
	ixOamChildLbCellTx (port, rxMbuf);
    }
    else
    {
	parentLbCellRxErrCount[port]++;

	/* free the buffer */
	ixDemoUtilsMbufFree (rxMbuf);
    }
}

/*
 * Reuse the received parent LB cell
 * and modify some of the fields
 */
PRIVATE IX_STATUS
ixOamChildLbCellTx (IxAtmLogicalPort port, IX_MBUF *rxMbuf)
{    
    IxOamITU610Cell *txCell;
    IxOamITU610LbPayload *lbPayload;
    IX_STATUS retval = IX_FAIL;
    unsigned int lockKey;

    /* Setup a pointer to the cell data */
    txCell = (IxOamITU610Cell *)rxMbuf->m_data;
        
    /* Setup pointers to Lb fields */
    lbPayload = &(txCell->payload.lbPayload);
    
    /* Change the loopback indication flag */
    lbPayload->loopbackIndication = IX_OAM_ITU610_LB_INDICATION_CHILD;

    /* Set the LLID to the CPID */
    memcpy(lbPayload->llid, oamCpid, IX_OAM_ITU610_LOCATION_ID_LEN);

    /* Transmit the cell, not reentrant on a VC basis so protect */
    lockKey = ixOsServIntLock();
    retval =  ixAtmdAccTxVcPduSubmit (oamTxConnId[port],
				      rxMbuf,
				      0, /* CLP 0 */
				      1); /* Cells per packet */
    ixOsServIntUnlock (lockKey);

    if (retval == IX_SUCCESS)
    {
	/* Increment the loopback out counter */
	childLbCellTxCount[port]++;
    }
    else
    {
	/* Release the MBUF */
	ixDemoUtilsMbufFree (rxMbuf);
    }

    return retval;
}

PRIVATE IX_STATUS
ixOamRxLlidCheck (IxOamITU610Cell *oamCell)
{
    UINT8 *llid;
    unsigned int pti;
    unsigned int vci;

    /* Get the LLID from the payload */
    llid = oamCell->payload.lbPayload.llid;

    /* Get the Pt and Vci */
    pti = oamCell->header.pti;
    vci = oamCell->header.vci;

    /* Compare LLID to the CPID */
    if (memcmp (llid, oamCpid, IX_OAM_ITU610_LOCATION_ID_LEN) == 0)
    {
	return IX_SUCCESS;
    }
    /* Compare to all 1's, i.e. end point  */
    else if (memcmp (llid, allOnesLocId, IX_OAM_ITU610_LOCATION_ID_LEN) == 0)
    {
	return IX_SUCCESS;
    }
    /* is this a segment loopback cell */
    else if ((vci == IX_OAM_ITU610_F4_SEG_VCI) || (pti == IX_OAM_ITU610_F5_SEG_PTI))
    {
	/* Compare to all 0's */
	if (memcmp (llid, allZerosLocId, IX_OAM_ITU610_LOCATION_ID_LEN) == 0)
	{
	    return IX_SUCCESS;
	}
    }

    return IX_FAIL;
}

PRIVATE void
ixOamTxDoneCallback (IxAtmdAccUserId userId,
		     IX_MBUF *mbufPtr)
{
    ixDemoUtilsMbufFree(mbufPtr);

    txDoneCallbackCount[userId]++; /* userId == port */
}

PRIVATE void
ixOamRxFreeLowReplenishCallback (IxAtmdAccUserId userId)
{
    IX_MBUF *mBuf;
    unsigned int numFreeEntries;
    unsigned int cnt;
    IX_STATUS retval;

    IX_ASSERT(userId == oamRxUserId);

    retval = ixAtmdAccRxVcFreeEntriesQuery (oamRxConnId, &numFreeEntries);
    if (retval != IX_SUCCESS)
    {
	IX_OAM_CODELET_IRQ_SAFE_LOG_ERROR("Failed to query depth of Oam Rx Free Q");
	return;
    }

    /* Replenish Rx buffers  */
    for (cnt=0; cnt<numFreeEntries; cnt++)
    {
	ixDemoUtilsMbufGet(IX_ATM_OAM_CELL_SIZE_NO_HEC, &mBuf);

	if (mBuf == NULL)
	{
	    IX_OAM_CODELET_IRQ_SAFE_LOG_ERROR("Failed to get rx free buffer");
	    return;
	}
	
	/* 
	 * Set the number of bytes of data in this MBUF to 1 OAM cell
	 */
	mBuf->m_len = IX_ATM_OAM_CELL_SIZE_NO_HEC;

	/* Send free buffers to NPE */
	retval = ixAtmdAccRxVcFreeReplenish (oamRxConnId, mBuf);

	if (retval != IX_SUCCESS)
	{
	    /* Free the allocated buffer */
	    ixDemoUtilsMbufFree(mBuf);
	    IX_OAM_CODELET_IRQ_SAFE_LOG_ERROR("Failed to pass Oam Rx free buffers to Atmd");
	    return;
	}
	
    }
    replenishCallbackCount++;
}

