/**
 * @file    IxOamCodelet.h
 *
 * @date    6-July-2002
 *
 * @brief   Header file for the IXP4xx OAM codelet
 *
 * -- Intel Copyright Notice --
 *
 * @par
 * INTEL CONFIDENTIAL
 *
 * @par
 * Copyright 2002 Intel Corporation All Rights Reserved.
 *
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.  The Material
 * contains trade secrets and proprietary and confidential information of
 * Intel or its suppliers and licensors.  The Material is protected by
 * worldwide copyright and trade secret laws and treaty provisions. No
 * part of the Material may be used, copied, reproduced, modified,
 * published, uploaded, posted, transmitted, distributed, or disclosed in
 * any way without Intel's prior express written permission.
 *
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel or otherwise.  Any license under
 * such intellectual property rights must be express and approved by
 * Intel in writing.
 *
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * -- End Intel Copyright Notice --
 *
 * @sa IxOamCodelet.h
 *
 */

/**
 * @defgroup Codelets IXP4xx Codelets
 *
 * @brief IXP4xx demo codelets
 *
 * @{
 */

/**
 * @ingroup Codelets
 *
 * @defgroup IxOamCodelet IXP4xx Oam Codelet (IxOamCodelet) API
 *
 * @brief IXP4xx codelet Oam demo compoent API
 *
 * This codelet demonstrates an example implementation of OAM F4 Segment,
 * F4 End-To-End (ETE), F5 Segment and F5 ETE loopback that makes use of the 
 * AtmdAcc component, as well as demonstrating how the lower layer IxAtmdAcc 
 * component can be used for configuration and control.
 *
 * <b>Features</b>
 * @li An interface is provided to configure the access layer to enable OAM traffic.
 * 
 * @li OAM Loopback responses are sent automatically on receipt of OAM F4 and F5
 *     parent Segment and ETE loopback cells, i.e. externally initiated OAM Loopbacks.
 *
 * @li Interfaces are provided to initiate an OAM F4/F5 Segment or ETE Loopback, 
 *     see ITU-610.
 *     One loopback can be initiated at a time, this is to keep the codelet simple,
 *     this is independant of sending responses to parent loopback cells received.
 *
 *
 * @li An interface is provided to query OAM traffic statistics.
 *
 * <b>IXP4xx ATM Components used by this codelet</b>
 * @li IxAtmdAcc. This component is the low level interface by which OAM cells get transmitted
 *                to, and received from the UTOPIA bus.
 *
 * @li IxAtmm. This component provides ATM port and VC management facilities. This component
 *             also contains the configuration of the UTOPIA co-processor.
 *
 * @li DemoUtils. This codelet provides VC setup facilities using IxAtmdAcc. 
 *                
 * @{
 */

#ifndef IXOAMCODELET_H
#define IXOAMCODELET_H

/*
 * User defined includes
 */
#include "IxAtmTypes.h"

/*
 * #defines and macros used in this file.
 */

/*
 * Typedefs
 */

/*
 * Function prototypes
 */

/** 
 * @fn ixOamCodeletInit (unsigned int numPorts)
 * 
 * @brief This function is used to configure OAM accross a number of previously configured ports.
 *         N.B. IxQmgr, IxNpeMh, IxAtmSch, IxAtmdAcc, IxAtmm must be initialized
 *         before calling this API, e.g. see AtmCodelet:ixAtmCodeletSystemInit()
 *
 * @param "unsigned int numPorts" The number of ATM ports to configure
 *        on. This must be less than or equal to the number of
 *        ports configured in e.g. the call to ixAtmCodeletInit().
 *
 * @return
 * - IX_SUCCESS : Indicates that the Atm Codelet has successfully provisioned the OAM Tx
 *          Port channels and the OAM Rx channel.
 * - IX_FAIL : Some internal error has prevented the ATM codelet from
 *          from provisioning the channels.
 */
IX_STATUS
ixOamCodeletInit (unsigned int numPorts);

/** 
 * @fn ixOamCodeletOamF5EtePing (IxAtmLogicalPort port,
			  unsigned int vpi,
			  unsigned int vci,
			  unsigned int numCells)
 * 
 * @brief This function is used to initiate an OAM F5 ETE loopback, i.e. OAM ping,
 *         on the specified Port, VPI and VCI for a number of cells.
 *         ixOamCodeletInit must be called before calling this function.
 * 
 *         OAM F5 ETE Parent Loopback cells are sent to the F5 connection 
 *         end point (LLID = all one's).
 *
 *         If an OAM F5 ETE Child Loopback cell has been received in response to a Parent
 *         cell sent, a message will be displayed at the WindSh prompt.
 *
 *         The loopback is complete and a subsequent loopback is allowed only when all
 *         the loopback parent cells have been sent and either A) a child loopback cell has been 
 *         received in response to the last cell OR B) 5 seconds have passed since the last cell 
 *         was sent.
 *         N.B. the function blocks until the loopback has completed.
 *
 * @param "IxAtmLogicalPort" The port to send the OAM loopback cells on. This port must have been
 *        configured previously.
 *
 * @param "unsigned int VPI" The VPI of the channel to send the OAM loopback cells on. An Aal0/Aal5 
 *        channel need not be configured previously on the same VPI/VCI.
 *
 * @param "unsigned int VCI" The VCI of the channel to send the OAM loopback cells on. An Aal0/Aal5 
 *        channel need not be configured previously on the same VPI/VCI.
 *
 * @param "unsigned int numCells" The number of parent loopback cells to send. Not checked against
 *        some max value, only limited by size of type.
 *
 * @return
 * - IX_SUCCESS : Indicates that the Oam Codelet has successfully performed the OAM ping.
 * - IX_FAIL : Some internal error has prevented the Oam Codelet performing the OAM ping.
 */
IX_STATUS
ixOamCodeletOamF5EtePing (IxAtmLogicalPort port,
			  unsigned int vpi,
			  unsigned int vci,
			  unsigned int numCells);

/** 
 * @fn ixOamCodeletOamF5SegPing (IxAtmLogicalPort port,
			  unsigned int vpi,
			  unsigned int vci,
			  unsigned int numCells)
 * 
 * @brief This function is used to initiate an OAM F5 Segment loopback, i.e. OAM ping,
 *         on the specified Port, VPI and VCI for a number of cells.
 *         ixOamCodeletInit must be called before calling this function.
 * 
 *         OAM F5 Segment Parent Loopback cells are sent to the F5 connection 
 *         end point (LLID = all one's).
 *
 *         If an OAM F5 Segment Child Loopback cell has been received in response to a Parent
 *         cell sent, a message will be displayed at the WindSh prompt.
 *
 *         The loopback is complete and a subsequent loopback is allowed only when all
 *         the loopback parent cells have been sent and either A) a child loopback cell has been 
 *         received in response to the last cell OR B) 5 seconds have passed since the last cell 
 *         was sent.
 *         N.B. the function blocks until the loopback has completed.
 *
 * @param "IxAtmLogicalPort" The port to send the OAM loopback cells on. This port must have been
 *        configured previously.
 *
 * @param "unsigned int VPI" The VPI of the channel to send the OAM loopback cells on. An Aal0/Aal5 
 *        channel need not be configured previously on the same VPI/VCI.
 *
 * @param "unsigned int VCI" The VCI of the channel to send the OAM loopback cells on. An Aal0/Aal5 
 *        channel need not be configured previously on the same VPI/VCI.
 *
 * @param "unsigned int numCells" The number of parent loopback cells to send. Not checked against
 *        some max value, only limited by size of type.
 *
 * @return
 * - IX_SUCCESS : Indicates that the Oam Codelet has successfully performed the OAM ping.
 * - IX_FAIL : Some internal error has prevented the Oam Codelet performing the OAM ping.
 */
IX_STATUS
ixOamCodeletOamF5SegPing (IxAtmLogicalPort port,
			  unsigned int vpi,
			  unsigned int vci,
			  unsigned int numCells);

/** 
 * @fn ixOamCodeletOamF4EtePing (IxAtmLogicalPort port,
			  unsigned int vpi,
			  unsigned int numCells)
 * 
 * @brief This function is used to initiate an OAM F4 ETE loopback, i.e. OAM ping,
 *         on the specified Port, VPI for a number of cells. F4 flows are VPC
 *         based and as such don't require a VCI.
 *         ixOamCodeletInit must be called before calling this function.
 * 
 *         OAM F4 ETE Parent Loopback cells are sent to the F4 connection 
 *         end point (LLID = all one's).
 *
 *         If an OAM F4 ETE Child Loopback cell has been received in response to a Parent
 *         cell sent, a message will be displayed at the WindSh prompt.
 *
 *         The loopback is complete and a subsequent loopback is allowed only when all
 *         the loopback parent cells have been sent and either A) a child loopback cell has been 
 *         received in response to the last cell OR B) 5 seconds have passed since the last cell 
 *         was sent.
 *         N.B. the function blocks until the loopback has completed.
 *
 * @param "IxAtmLogicalPort" The port to send the OAM loopback cells on. This port must have been
 *        configured previously.
 *
 * @param "unsigned int VPI" The VPI of the channel to send the OAM loopback cells on. An Aal0/Aal5 
 *        channel need not be configured previously on the same VPI/VCI.
 *
 * @param "unsigned int numCells" The number of parent loopback cells to send. Not checked against
 *        some max value, only limited by size of type.
 *
 * @return
 * - IX_SUCCESS : Indicates that the Oam Codelet has successfully performed the OAM ping.
 * - IX_FAIL : Some internal error has prevented the Oam Codelet performing the OAM ping.
 */
IX_STATUS
ixOamCodeletOamF4EtePing (IxAtmLogicalPort port,
			  unsigned int vpi,
			  unsigned int numCells);

/** 
 * @fn ixOamCodeletOamF4SegPing (IxAtmLogicalPort port,
			  unsigned int vpi,
			  unsigned int numCells)
 * 
 * @brief This function is used to initiate an OAM F4 Segment loopback, i.e. OAM ping,
 *         on the specified Port, VPI for a number of cells. F4 flows are VPC
 *         based and as such don't require a VCI.
 *         ixOamCodeletInit must be called before calling this function.
 * 
 *         OAM F4 Segment Parent Loopback cells are sent to the F4 connection 
 *         end point (LLID = all one's).
 *
 *         If an OAM F4 Segment Child Loopback cell has been received in response to a Parent
 *         cell sent, a message will be displayed at the WindSh prompt.
 *
 *         The loopback is complete and a subsequent loopback is allowed only when all
 *         the loopback parent cells have been sent and either A) a child loopback cell has been 
 *         received in response to the last cell OR B) 5 seconds have passed since the last cell 
 *         was sent.
 *         N.B. the function blocks until the loopback has completed.
 *
 * @param "IxAtmLogicalPort" The port to send the OAM loopback cells on. This port must have been
 *        configured previously.
 *
 * @param "unsigned int VPI" The VPI of the channel to send the OAM loopback cells on. An Aal0/Aal5 
 *        channel need not be configured previously on the same VPI/VCI.
 *
 * @param "unsigned int numCells" The number of parent loopback cells to send. Not checked against
 *        some max value, only limited by size of type.
 *
 * @return
 * - IX_SUCCESS : indicates that the Oam Codelet has successfully performed the OAM ping.
 * - IX_FAIL : Some internal error has prevented the Oam Codelet performing the OAM ping.
 */
IX_STATUS
ixOamCodeletOamF4SegPing (IxAtmLogicalPort port,
			  unsigned int vpi,
			  unsigned int numCells);

/** 
 * @fn ixOamCodeletShow (void)
 *
 * @brief This function is used to dump OAM statistics, i.e. parent and child cell receive
 * and transmit counts etc.
 *
 * @param none
 *
 * @return void
 */
void
ixOamCodeletShow (void);
#endif
/* IXOAMCODELET_H */

/** @} defgroup IxOamCodelet*/

/** @} defgroup Codelets */

