/** 
 * @file:	 IxUARTAccCodelet.c
 * @author Intel Corporation
 * @date:	 16-OCT-01
 *
 * @brief: Codelet for the Intel IXP425 internal UART.
 * 
 * Design Notes:
 *    
 * -- Intel Copyright Notice --
 *
 * @par
 * INTEL CONFIDENTIAL
 *
 * @par
 * Copyright 2002 Intel Corporation All Rights Reserved.
 *
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.  The Material
 * contains trade secrets and proprietary and confidential information of
 * Intel or its suppliers and licensors.  The Material is protected by
 * worldwide copyright and trade secret laws and treaty provisions. No
 * part of the Material may be used, copied, reproduced, modified,
 * published, uploaded, posted, transmitted, distributed, or disclosed in
 * any way without Intel's prior express written permission.
 *
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel or otherwise.  Any license under
 * such intellectual property rights must be express and approved by
 * Intel in writing.
 *
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * -- End Intel Copyright Notice --
*
 * This file contains the implementation of the UART Access Codelet.
 */


/*
 * Put the system defined include files required.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
 * Put the user defined include files required.
 */

#include <IxTypes.h>
#include <IxUART.h>
#include <ixp425.h>		 /* Chip level definitions required */
#include <ix_ossl.h>
#include "IxUARTAccCodelet.h"

/*
 * defines and macros used in this file.
 */

#define IX_UARTACC_CODELET_CTRL_Z 26
#define IX_UARTACC_CODELET_INT_SIZE 11

/*
 * Static function prototypes.
 */

PRIVATE void ixUARTAccCodeletIntToChar(UINT32 num, char* string);

/*
 * Variable declarations global to this file only.
 */

PRIVATE ixUARTDev uart1;
PRIVATE ixUARTDev uart2;

/*
 * Function definition: ixEthAccCodeletDevConfig()
 *
 * See header file for documentation.
 */
 
IX_STATUS ixUARTAccCodeletDevConfig(ixUARTAccPort uartNo)
{
    ixUARTDev* uart;

    switch(uartNo)
    {
	case IX_UART_PORT1:
	    uart1.addr = (unsigned char *)IXP425_UART1_BASE;
	    uart = &uart1;
	    break;

	case IX_UART_PORT2:
	    uart2.addr = (unsigned char *)IXP425_UART2_BASE;
	    uart = &uart2;
	    break;

	default:
	    return (IX_FAIL);
    }
 
    uart->baudRate = IX_UART_DEF_BAUD;

    /* Initialise device */
    if (ixUARTInit(uart) != IX_SUCCESS)
    {
	return (IX_FAIL);
    }

    ixUARTAccCodeletWrite(uartNo, "UART succesfully initialised\n\n");
    ixUARTAccCodeletShow(uartNo);
    
    return (IX_SUCCESS);
}


/*
 * Function definition: ixUARTAccCodeletSwLoopback()
 *
 * See header file for documentation.
 */

IX_STATUS ixUARTAccCodeletSwLoopback(ixUARTAccPort uartNo)
{
    int charCnt = 0;
    char buf[2];

    switch(uartNo)
    {
	case IX_UART_PORT1:
	    if(uart1.addr != 0)
	    {
		ixUARTAccCodeletWrite(uartNo, "UART S/W Loopback demo started on UART 1\n");
	    }
	    break;
    
	case IX_UART_PORT2:
	    if(uart2.addr != 0)
	    {
		ixUARTAccCodeletWrite(uartNo, "UART S/W Loopback demo started on UART 2\n");
	    }
	    break;
    
	default:
	    return (IX_FAIL);
    }

    ixUARTAccCodeletWrite(uartNo, "Press CTRL+Z at any time to end\n");

    while(1)
    {
	/* Read 1 character at a time */
	charCnt = ixUARTAccCodeletRead(uartNo, buf, 1);

	if (!charCnt)
	    continue;

	buf[1] = 0;
	
	/* User escaped */
	if((int)buf[0] == IX_UARTACC_CODELET_CTRL_Z)
	{
	    break;
	}

	/* Echo character to the terminal */
	ixUARTAccCodeletWrite(uartNo, buf);
    }

    ixUARTAccCodeletWrite(uartNo, "UART S/W Loopback demo ended\n");
    return (IX_SUCCESS);
}


/*
 * Function definition: ixUARTAccCodeletBaudDemo
 *
 * See header file for documentation.
 */

IX_STATUS ixUARTAccCodeletBaudDemo(ixUARTAccPort uartNo)
{
    ixUARTDev* uart;
    int i = 0;
    int baudTable[] = {9600, 19200, 38400, 56000, 115200,
		       230400, 460800, 921600};
    
    char *baudText[] = {"9600 Baud\n", "19200 Baud\n", "38400 Baud\n",
			"56000 Baud\n", "115200 Baud\n", "230400 Baud\n",
			"460800 Baud\n", "921600 Baud\n"};

    switch(uartNo)
    {
	case IX_UART_PORT1:
	    uart = &uart1;
	    break;

	case IX_UART_PORT2:
	    uart = &uart2;
	    break;

	default:
	    return (IX_FAIL);
    }

    for (i=0; i<8; i++)
    {
	if(ixUARTIoctl(uart, IX_BAUD_SET, (int *)&baudTable[i]) != IX_SUCCESS)
	{
	    ixUARTAccCodeletWrite(uartNo, "Error setting baud rate to: ");
	    ixUARTAccCodeletWrite(uartNo, baudText[i]);
	    return (IX_FAIL);
	}

	ixUARTAccCodeletWrite(uartNo, "Transmitting at: ");
	ixUARTAccCodeletWrite(uartNo, baudText[i]); 

	ix_ossl_sleep(1000);
    }
    
    return (IX_SUCCESS);
}


/*
 * Function definition: ixUARTAccCodeletRead()
 *
 * See header file for documentation.
 */

int ixUARTAccCodeletRead(ixUARTAccPort uartNo, char* buf, unsigned int count)
{
    ixUARTDev* pUart;
    int charCnt = 0;

    switch(uartNo)
    {
	case IX_UART_PORT1:
	    pUart = &uart1;
	    break;

	case IX_UART_PORT2:
	    pUart = &uart2;
	    break;

	default:
	    return (IX_FAIL);
    }

    while(charCnt < count)
    {
	if(ixUARTPollInput(pUart, &buf[charCnt]) != IX_SUCCESS)
	{
 	    break;
	}

#ifdef TRANSLATE_IN_CHARS	/* Character translations are common */
	if(*buf == '\r')
	{
	    *buf = '\n';
	}
#endif
	
	++buf;
	++charCnt;
    }

    buf[charCnt] = '\0';
    return (charCnt);
}

/*
 * Function definition: ixUARTAccCodeletWrite()
 *
 * See header file for documentation.
 */

int ixUARTAccCodeletWrite(ixUARTAccPort uartNo, char* buf)
{
    ixUARTDev* pUart;
    UINT32 charCnt = 0;
    int len = strlen(buf);

    switch(uartNo)
    {
	case IX_UART_PORT1:
	    pUart = &uart1;
	    break;

	case IX_UART_PORT2:
	    pUart = &uart2;
	    break;

	default:
	    return (IX_FAIL);
    }

    while(charCnt < len || buf[charCnt] != '\0')
    {
#ifdef TRANSLATE_OUT_CHARS
	if(buf[charCnt] == '\n')	/* Character translations are common */
	{
	    while(ixUARTPollOutput(pUart, '\r') != IX_SUCCESS);
	}
#endif

	while(ixUARTPollOutput(pUart, buf[charCnt]) != IX_SUCCESS); /* Keep trying */
	++charCnt;
    }
    
    return (charCnt);
}


/*
 * Function definition: ixUARTAccCodeletShow()
 *
 * See header file for documentation.
 */

IX_STATUS ixUARTAccCodeletShow(ixUARTAccPort uartNo)
{
    ixUARTDev* uart;
    int baudRate;
    int mode;
    ixUARTStats uartStats;
    char string[IX_UARTACC_CODELET_INT_SIZE];

    switch(uartNo)
    {
	case IX_UART_PORT1:
	    uart = &uart1;
	    ixUARTAccCodeletWrite(uartNo, "UART 1 information:\n");
	    break;

	case IX_UART_PORT2:
	    uart = &uart2;
	    ixUARTAccCodeletWrite(uartNo, "UART 2 information:\n");
	    break;

	default:
	    return (IX_FAIL);
    }

    ixUARTAccCodeletWrite(uartNo, "=================\n");

    ixUARTIoctl(uart, IX_BAUD_GET, &baudRate);
    ixUARTAccCodeletWrite(uartNo, "Baud rate:\t");
    ixUARTAccCodeletIntToChar(baudRate, string);
    ixUARTAccCodeletWrite(uartNo, string);

    ixUARTIoctl(uart, IX_MODE_GET, &mode);
    ixUARTAccCodeletWrite(uartNo, "\nMode:\t\t");

    switch(mode)
    {
	case INTERRUPT:
	    ixUARTAccCodeletWrite(uartNo, "interrupt");
	    break;

	case POLLED:
	    ixUARTAccCodeletWrite(uartNo, "polled");
	    break;

	case LOOPBACK:
	    ixUARTAccCodeletWrite(uartNo, "loopback");
	    break;

	default:
	    break;
    }

    ixUARTIoctl(uart, IX_STATS_GET, &uartStats);

    ixUARTAccCodeletWrite(uartNo, "\nRx Count:\t");
    ixUARTAccCodeletIntToChar(uartStats.rxCount, string);
    ixUARTAccCodeletWrite(uartNo, string);

    ixUARTAccCodeletWrite(uartNo, "\nTx Count:\t");
    ixUARTAccCodeletIntToChar(uartStats.txCount, string);
    ixUARTAccCodeletWrite(uartNo, string);

    ixUARTAccCodeletWrite(uartNo, "\nOverrun Errors:\t");
    ixUARTAccCodeletIntToChar(uartStats.overrunErr, string);
    ixUARTAccCodeletWrite(uartNo, string);

    ixUARTAccCodeletWrite(uartNo, "\nParity Errors:\t");
    ixUARTAccCodeletIntToChar(uartStats.parityErr, string);
    ixUARTAccCodeletWrite(uartNo, string);

    ixUARTAccCodeletWrite(uartNo, "\nFraming Errors:\t");
    ixUARTAccCodeletIntToChar(uartStats.framingErr, string);
    ixUARTAccCodeletWrite(uartNo, string);

    ixUARTAccCodeletWrite(uartNo, "\nBreak Errors:\t");
    ixUARTAccCodeletIntToChar(uartStats.breakErr, string);
    ixUARTAccCodeletWrite(uartNo, string);

    ixUARTAccCodeletWrite(uartNo, "\n");
    return (IX_SUCCESS);
}


/******************************************************************************
 *
 * UART Codelet support functions
 *
 ******************************************************************************/

/**
 * @fn ixUARTAccCodeletIntToChar()
 * 
 * @param num - number to convert
 * @param *string - string to hold the number
 *
 * Converts a number into an ASCII string.
 *
 * @return void
 ***************************************************************************/

PRIVATE void ixUARTAccCodeletIntToChar(UINT32 num, char* string)
{
    int charOffset = '0';
    int numDigit = 0;
    int cnt = 0;
    char tmpString[IX_UARTACC_CODELET_INT_SIZE];

    BOOL trailingBlank = TRUE;

    for (numDigit=IX_UARTACC_CODELET_INT_SIZE-1; numDigit>=0; numDigit--)
    {
	tmpString[numDigit] = (num%10) + charOffset;
	num = num/10;
    }
    
    /* String now contains our number but with padded zeros */
    for (numDigit=0; numDigit<IX_UARTACC_CODELET_INT_SIZE; numDigit++)
    {
	if(tmpString[numDigit] == '0' && trailingBlank == TRUE)
	{
	    /* Do nothing */
	}
	else
	{
	    trailingBlank = FALSE;
	    string[cnt] = tmpString[numDigit];
	    cnt++;
	}
    }

    if (cnt == 0)
    {
	string[0] = '0';
	cnt++;
    }

    string[cnt] = '\0';
}

