/**
 * @file IxCryptoAcc.c
 *
 * @date October-03-2002
 *
 * @brief  Source file for Public API (Component Interface module)
 *
 *
 * Design Notes:
 *
 * -- Intel Copyright Notice -- 
 *
 * @par
 * INTEL CONFIDENTIAL
 * 
 * @par
 * Copyright 2000-2002 Intel Corporation All Rights Reserved.
 * 
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.  The Material
 * contains trade secrets and proprietary and confidential information of
 * Intel or its suppliers and licensors.  The Material is protected by
 * worldwide copyright and trade secret laws and treaty provisions. No
 * part of the Material may be used, copied, reproduced, modified,
 * published, uploaded, posted, transmitted, distributed, or disclosed in
 * any way without Intel's prior express written permission.
 * 
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel or otherwise.  Any license under
 * such intellectual property rights must be express and approved by
 * Intel in writing.
 * 
 * -- End Intel Copyright Notice -- 
 */


/*
* Put the user defined include files required.
*/
#include "IxOsServices.h"
#include "IxCryptoAcc.h"
#include "IxCryptoAcc_p.h"
#include "IxCryptoAccCCDMgmt_p.h"
#include "IxCryptoAccDescMgmt_p.h"
#include "IxCryptoAccQAccess_p.h"
#include "IxCryptoAccCryptoRegister_p.h"
#include "IxCryptoAccCryptoPerform_p.h"
#include "IxCryptoAccUtilities_p.h"


#define IX_CRYPTO_ACC_DELAY_IN_MS   500    /**< Time taken in task delay */
 
#define IX_CRYPTO_ACC_TIMER_COUNT   100    /**< Timer counter to prevent
                                            * deadlock
                                            */

/*
* Variable declarations global to this file only.  Externs are followed by
* static variables.
*/
static BOOL ixCryptoAccServiceInit = FALSE; /**< TRUE if component 
                                             * initialized 
                                             */
static BOOL ixCryptoAccServiceStop = FALSE; /**< TRUE if crypto service 
                                             * is halt 
                                             */


/**
 * @def IX_CRYPTO_IS_SERVICE_INITIALIZED
 * @brief Check if component has been initialized
 *
 * return none
 *
 */
#define IX_CRYPTO_IS_SERVICE_INITIALIZED() (TRUE == ixCryptoAccServiceInit)



/**
 * @def IX_CRYPTO_IS_SERVICE_STOPPED
 * @brief Check if component service has been stopped
 *
 * return none
 *
 */
#define IX_CRYPTO_IS_SERVICE_STOPPED() (TRUE == ixCryptoAccServiceStop)



/**
 * @fn ixCryptoAccInit
 * @brief Initialise the Security Access component
 *
 */
PUBLIC IxCryptoAccStatus
ixCryptoAccInit (void)
{
    /* Check if the access component has been intialized, if not, 
     * initialize Descriptor Mgmt, CCD Mgmt, Statistics and QAccess 
     * modules.
     */
    if (!IX_CRYPTO_IS_SERVICE_INITIALIZED ()
        && !IX_CRYPTO_IS_SERVICE_STOPPED ())
    {       
        /* Initializing Descriptor Management Module */
        if (IX_CRYPTO_ACC_STATUS_SUCCESS != ixCryptoDescMgmtInit ())
        {
            /* Log error message in debugging mode */
            IX_CRYPTO_ACC_LOG (LOG_ERROR,
                "ixCryptoDescMgmtInit FAILED.\n", 
                0, 0, 0, 0, 0, 0);
             
            return IX_CRYPTO_ACC_STATUS_FAIL;
        } /* end of if (ixCryptoDescMgmtInit) */
            
        /* Initializing Crypto Context Database Management Module */
        if (IX_CRYPTO_ACC_STATUS_SUCCESS != ixCryptoCCDMgmtInit ())
        {
            /* Log error message in debugging mode */
            IX_CRYPTO_ACC_LOG (LOG_ERROR,
                "ixCryptoCCDMgmtInit FAILED.\n", 
                0, 0, 0, 0, 0, 0);
                
            return IX_CRYPTO_ACC_STATUS_FAIL;
        } /* end of if (ixCryptoCCDMgmtInit) */
            
        /* Initializing Queue Access Module */
        if (IX_CRYPTO_ACC_STATUS_SUCCESS != ixCryptoQAccessInit ())
        {
            /* Log error message in debugging mode */
            IX_CRYPTO_ACC_LOG (LOG_ERROR,
                "ixCryptoQAccessInit FAILED.\n", 
                0, 0, 0, 0, 0, 0);
                
            return IX_CRYPTO_ACC_STATUS_FAIL;
        } /* end of if (ixCryptoQAccessInit) */
        
        /* The initialisation is complete hence set the flag to TRUE */
        ixCryptoAccServiceInit = TRUE;
        
        /* Initialization successfully */
        return IX_CRYPTO_ACC_STATUS_SUCCESS;
    }
    else /* end of if (IX_CRYPTO_IS_SERVICE_INITIALIZED) &&
            (IX_CRYPTO_IS_SERVICE_STOPPED) */
    {
        /* Log error message in debugging mode */
        IX_CRYPTO_ACC_LOG (LOG_ERROR,
            "IxCrypto is initialized or stopped before this \
            function call.\n", 
            0, 0, 0, 0, 0, 0);

        return IX_CRYPTO_ACC_STATUS_FAIL;
    } /* end of if-else (IX_CRYPTO_IS_SERVICE_INITIALIZED) &&
     (IX_CRYPTO_IS_SERVICE_STOPPED) */       
} /* end of ixCryptoAccInit () function */

 
/**
 * @fn ixCryptoAccCtxRegister
 * @brief This function is for the client to pass all the required  
 *        information (eg. key, cipher algorithm, etc) for the hardware 
 *        accelerator context. All the information will be converted into 
 *        a cryptographic parameters structure to be shared with NPE. 
 *        The structure will be associated with cryptoCtxId.
 *
 */
PUBLIC IxCryptoAccStatus 
ixCryptoAccCtxRegister (
    IxCryptoAccCtx *pAccCtx,
    IX_MBUF *pMbufPrimaryChainVar,
    IX_MBUF *pMbufSecondaryChainVar,
    IxCryptoAccRegisterCompleteCallback registerCallbackFn,
    IxCryptoAccPerformCompleteCallback performCallbackFn,
    UINT32 *pCryptoCtxId)
{
    IxCryptoAccStatus status = IX_CRYPTO_ACC_STATUS_SUCCESS;
    UINT32 npeCryptoParamIndex = 0;
    BOOL callbackFlag = FALSE;

    /* Check if the client context and CryptoCtxId pointer are NULL, 
     * return error if either one is NULL 
     */
    if ((NULL == pAccCtx) || (NULL == pCryptoCtxId))
    {
        /* Log error message in debugging mode */
        IX_CRYPTO_ACC_LOG (
            LOG_MESSAGE, 
            "Cleint's Ctx is %x\n pCryptoCtxId is %x\n", 
            (int) pAccCtx, 
            (int) pCryptoCtxId,
            0, 0, 0, 0);

        return IX_CRYPTO_ACC_STATUS_FAIL;
    } /* end of if (pAccCtx) || (pCryptoCtxId) */
    
    /* Check if the callback functions provided by client is NULL, return
     * error if either one is NULL
     */
    if ((NULL == registerCallbackFn) || (NULL == performCallbackFn))
    {
        /* Log error message in debugging mode */
        IX_CRYPTO_ACC_LOG (
            LOG_MESSAGE, 
            "Register Callback: %x\nPerform Callback: %x\n", 
            (int) registerCallbackFn,
            (int) performCallbackFn, 
            0, 0, 0, 0);

        return IX_CRYPTO_ACC_STATUS_FAIL;
    } /* end of if (registerCallbackFn) || (performCallbackFn) */

    /* Check if the access component has been intialized and the crypto 
     * service is running 
     */
    if (IX_CRYPTO_IS_SERVICE_INITIALIZED() 
        && !IX_CRYPTO_IS_SERVICE_STOPPED())
    {
        /* Get an empty Crypto Context from CCD */
        status = ixCryptoCCDMgmtCryptoCtxGet (pCryptoCtxId);
    
        if (IX_CRYPTO_ACC_STATUS_SUCCESS == status)
        {
            /* Store callback function pointers into Crypto Context */
            ixCryptoCtx[*pCryptoCtxId].registerCallbackFn 
                = registerCallbackFn;
            ixCryptoCtx[*pCryptoCtxId].performCallbackFn
                = performCallbackFn;    
        }
        else /* No Crypto Context is allocated */
        {
            return status;
        } /* end of if-else (status) */
    
        /* Switch case for Operation registered by client */
        switch (pAccCtx->operation)
        {
            /* Encryption operation, notes: the code for the case of 
             * IX_CRYPTO_ACC_OP_ENCRYPT fall through to the case 
             * IX_CRYPTO_ACC_OP_DECRYPT
             */
            case IX_CRYPTO_ACC_OP_ENCRYPT:
            
            /* Deceryption operation */
            case IX_CRYPTO_ACC_OP_DECRYPT:
                /* Register cipher algorithm */
                status = ixCryptoRegisterCipherAlgoRegister (
                             *pCryptoCtxId,
                             pAccCtx, 
                             &npeCryptoParamIndex,
                             IX_CRYPTO_OP_REGISTER);
                             
                /* Check if cipher context registration success */
                if (IX_CRYPTO_ACC_STATUS_SUCCESS == status)
                {
                    /* Assign NPE operation to CRYPT for the Crypto Context */
                    ixCryptoCtx[*pCryptoCtxId].npeOperationMode.npeOperation
                        = IX_CRYPTO_NPE_OP_CRYPT;
    
                    /* Set direction fo cryption operation, default is decrypt,
                     * OR the NpeOperation with the mask will set the crypt
                     * direction bit to 1 and turn it to be encrypt operation
                     */
                    if (IX_CRYPTO_ACC_OP_ENCRYPT == pAccCtx->operation)
                    {
                        ixCryptoCtx[*pCryptoCtxId].npeOperationMode.npeOperation
                            = (((ixCryptoCtx[*pCryptoCtxId].npeOperationMode.
                                npeOperation)
                                | IX_CRYPTO_NPE_OP_CRYPT_DIR_MASK) 
                                & IX_CRYPTO_NPE_OP_MASK);
                    } /* end of if (pAccCtx->operation) */
                    
                    /* Set callbackFlag to TRUE to trigger client's 
                     * callback for the case of DES/3DES encryption and 
                     * decryption operation only
                     */                 
                    if (IX_CRYPTO_ACC_CIPHER_AES 
                        != pAccCtx->cipherCtx.cipherAlgo)
                    {
                        /* Callback is called only if DES or 3DES, callback 
                         * for AES will be called in QMgr callback context
                         */
                        callbackFlag = TRUE;
                    } /* end of if (pAccCtx->cipherCtx.cipherAlgo) */
                } /* end of if (status) */
                break;
    
            /* Authentication Computation operation, notes: the code for the  
             * case of IX_CRYPTO_ACC_OP_AUTH_CALC fall through to the case 
             * IX_CRYPTO_ACC_OP_AUTH_CHECK 
             */
            case IX_CRYPTO_ACC_OP_AUTH_CALC:
            
            /* Authentication Verification operation */
            case IX_CRYPTO_ACC_OP_AUTH_CHECK: 
                /* Register authentication algorithm */
                status = ixCryptoRegisterAuthAlgoRegister (
                             *pCryptoCtxId,
                             pAccCtx, 
                             pMbufPrimaryChainVar,
                             pMbufSecondaryChainVar, 
                             &npeCryptoParamIndex,
                             IX_CRYPTO_OP_REGISTER);
    
                /* Check if authentication context registration
                 * successful 
                 */
                if (IX_CRYPTO_ACC_STATUS_SUCCESS == status)
                {
                    /* Assign NPE operation */
                    if (IX_CRYPTO_ACC_OP_AUTH_CALC == pAccCtx->operation)
                    {
                        ixCryptoCtx[*pCryptoCtxId].npeOperationMode.npeOperation
                            = IX_CRYPTO_NPE_OP_HMAC_GEN_ICV;
                    }
                    else /* IX_CRYPTO_ACC_OP_AUTH_CHECK */
                    {
                        ixCryptoCtx[*pCryptoCtxId].npeOperationMode.npeOperation
                            = IX_CRYPTO_NPE_OP_HMAC_VER_ICV;            
                    } /* end of if-else (pAccCtx->Operation) */
                } /* end of if (status) */
                break;
    
            /* Combined service, encryption followed by authentication,
             * notes: the code for the case of IX_CRYPTO_ACC_OP_ENCRYPT_AUTH
             * fall through to the case IX_CRYPTO_ACC_OP_AUTH_DECRYPT
             */
            case IX_CRYPTO_ACC_OP_ENCRYPT_AUTH: 
            
            /* Combined service, authentication followed by decryption */
            case IX_CRYPTO_ACC_OP_AUTH_DECRYPT:
                /* Register cipher algorithm */
                status = ixCryptoRegisterCipherAlgoRegister (
                             *pCryptoCtxId,
                             pAccCtx, 
                             &npeCryptoParamIndex,
                             IX_CRYPTO_OP_WAIT);
    
                /* Check if cipher context registration successful */
                if (IX_CRYPTO_ACC_STATUS_SUCCESS == status)
                {
                    /* Register authentication algorithm */
                    status = ixCryptoRegisterAuthAlgoRegister (
                                 *pCryptoCtxId,
                                 pAccCtx, 
                                 pMbufPrimaryChainVar,
                                 pMbufSecondaryChainVar, 
                                 &npeCryptoParamIndex,
                                 IX_CRYPTO_OP_REGISTER);
                
                    /* Check if authentication context registration 
                     * successful 
                     */
                    if (IX_CRYPTO_ACC_STATUS_SUCCESS == status)
                    {
                        /* Assign NPE operation */
                        if (IX_CRYPTO_ACC_OP_ENCRYPT_AUTH == pAccCtx->operation)
                        {
                            ixCryptoCtx[*pCryptoCtxId].npeOperationMode.
                                npeOperation 
                                    = IX_CRYPTO_NPE_OP_ENCRYPT_HMAC_GEN_ICV;
                        }
                        else /* IX_CRYPTO_ACC_OP_AUTH_DECRYPT */
                        {
                            ixCryptoCtx[*pCryptoCtxId].npeOperationMode.
                                npeOperation
                                    = IX_CRYPTO_NPE_OP_HMAC_VER_ICV_DECRYPT;            
                        } /* end of if-else (pAccCtx->operation) */
                    } /* end of if (status) */
                } /* end of if (status) */
                break;
    
            /* all others case */
            default:
                /* operation not supported */
                status = IX_CRYPTO_ACC_STATUS_OPERATION_NOT_SUPPORTED;  

        } /* end of switch (pAccCtx->operation) */
                
        if (IX_CRYPTO_ACC_STATUS_SUCCESS == status) 
        {
            /* Set transfer mode, default is in-place transfer mode, OR
             * the NpeOperation with the mask will set the transfer mode
             * to 1 and turn it to non in-place operation
             */
            if (pAccCtx->useDifferentSrcAndDestMbufs)
            {
                ixCryptoCtx[*pCryptoCtxId].npeOperationMode.npeOperation
                    = (((ixCryptoCtx[*pCryptoCtxId].npeOperationMode.
                          npeOperation)
                          | IX_CRYPTO_NPE_OP_TRANSFER_MODE_MASK) 
                          & IX_CRYPTO_NPE_OP_MASK);
            } /* end of if (pAccCtx->useDifferentSrcAndDestMbufs) */
    
            /* Assign Init Length for NPE Crypto Parameters structure in 
             * Crypto Context 
             */
            ixCryptoCtx[*pCryptoCtxId].npeOperationMode.initLength 
                = npeCryptoParamIndex;

             /* Check if client's callback need to be called here */
             if (callbackFlag) /* Callback needed */
             {
                    /* Mark valid bit in Crypto Context to indicate the 
                     * Crypto Context has been registered successfully
                     */
                    ixCryptoCtx[*pCryptoCtxId].valid = TRUE;
                        
                    /* Call client's callback */
                    ixCryptoCtx[*pCryptoCtxId].registerCallbackFn(
                        *pCryptoCtxId,
                        NULL, 
                        status);                
            } /* end of if (callbackFlag) */                

            return status;
        }
        else /* Registration failed */
        {
            /* Release Crypto Context allocated previously */
            ixCryptoCCDMgmtCryptoCtxRelease (*pCryptoCtxId);
    
            return status;
        } /* end of if (status) */
    }
    else /* (IX_CRYPTO_IS_SERVICE_INITIALIZED) &&
          * (!IX_CRYPTO_IS_SERVICE_STOPPED)
          */
    {
        
        /* Log error message in debugging mode */
            IX_CRYPTO_ACC_LOG (LOG_ERROR, 
                "Component has not been initialized\n", 
                0, 0, 0, 0, 0, 0);
                
        return IX_CRYPTO_ACC_STATUS_FAIL;
    } /* end of if-else (IX_CRYPTO_IS_SERVICE_INITIALIZED) &&
       * (!IX_CRYPTO_IS_SERVICE_STOPPED) 
       */
} /* End of ixCryptoAccCtxRegister () function */



/**
 * @fn   ixCryptoAccAuthCryptPerform
 * @brief Perform Authentication and Decryption/Encryption functionality
 *
 */

PUBLIC IxCryptoAccStatus
ixCryptoAccAuthCryptPerform (
    UINT32 cryptoCtxId,
    IX_MBUF *pSrcMbuf,
    IX_MBUF *pDestMbuf,
    UINT16 authStartOffset,
    UINT16 authDataLen,
    UINT16 cryptStartOffset,
    UINT16 cryptDataLen,
    UINT16 icvOffset,
    UINT8  *pIV)
{
    IxCryptoQDescriptor *pQDesc = NULL;
    UINT32 operationMode;
    UINT32 npeOperation;
    IxCryptoNpeOperationStatus operStatus = IX_CRYPTO_OP_PERFORM;
    IxCryptoNpeCryptCfgWord *pCipherCfg;
    IxCryptoAccStatus status;
    IX_MBUF *pMbuf = pSrcMbuf;
    INT32 newIcvOffset = icvOffset;
    BOOL useDiffBuf = FALSE;
    UINT8 *tempIcvAddr;
   
    /* Check if the access component has been intialized and the crypto service
       is running */
    if (IX_CRYPTO_IS_SERVICE_INITIALIZED ()
        && !IX_CRYPTO_IS_SERVICE_STOPPED ())
    {
        /* Check the validity of CryptoCtxId */
        status = ixCryptoCCDMgmtCtxValidCheck (cryptoCtxId);
        if (IX_CRYPTO_ACC_STATUS_SUCCESS != status)
        {
            /* Log error message in debugging mode */
            IX_CRYPTO_ACC_LOG (LOG_MESSAGE,
                "cryptoCtxId %d is not valid or registered.\n", 
                cryptoCtxId, 0, 0, 0, 0, 0);
            return status;          
        }       
     
        /* Now we know CtxId valid, look up record for the NPE operationMode */
        operationMode = (ixCryptoCtx[cryptoCtxId].npeOperationMode.npeOperation
                            & MASK_8_BIT);
           
        /* Check for NULL MBufs, and the destination mbuf will only be checked 
         * if the transfer mode is non in place */
        if  ((NULL == pSrcMbuf)||
             ((IX_CRYPTO_NPE_OP_TRANSFER_MODE_IS_IN_PLACE != (operationMode 
             & IX_CRYPTO_NPE_OP_TRANSFER_MODE_MASK) &&
             (NULL==pDestMbuf))))
        {
            /* Log error message in debugging mode */
            IX_CRYPTO_ACC_LOG (
                LOG_MESSAGE,
                "Input Mbufs not valid. pSrcMbuf=0x%X\n \
                pDestMbuf=0x%X\n",
                (int) pSrcMbuf,
                (int) pDestMbuf, 
                0, 0, 0, 0);
            return IX_CRYPTO_ACC_STATUS_FAIL;       
        }
    
#ifdef DEBUG    
        /* Perform boundary check. 
         * Note that the icvOffset is not checked against the mbuf length.
         * The actual length of the mbuf will not be known. Only the length
         * of the data stored in a mbuf is known.
         */
        status = ixCryptoPerformBufferOffsetBoundaryCheck (
                     operationMode,
                     pSrcMbuf,
                     pDestMbuf,
                     authStartOffset,
                     authDataLen,
                     cryptStartOffset,
                     cryptDataLen);
        if (IX_CRYPTO_ACC_STATUS_SUCCESS != status)
        {
            return status;
        } /* End of Boundary Check */       
#endif

        /* Get a free Q descriptor from Q descriptors pool. Return if 
         * unsuccessful */
        if (IX_CRYPTO_ACC_STATUS_SUCCESS != 
            ixCryptoDescMgmtQDescriptorGet (&pQDesc))
        {
            /* Error logging is done in the DescMgmt module */
            return IX_CRYPTO_ACC_STATUS_FAIL;
        } /* End of ixCryptoDescMgmtQDescriptorGet */

        /* Get NPE operation */
        npeOperation = operationMode & (~IX_CRYPTO_NPE_OP_TRANSFER_MODE_MASK);
        
        /* Get operation transfer mode */
        if (IX_CRYPTO_NPE_OP_TRANSFER_MODE_IS_IN_PLACE != (operationMode 
             & IX_CRYPTO_NPE_OP_TRANSFER_MODE_MASK))
        {
            /* set to TRUE if non in-place operation */
            useDiffBuf = TRUE;                 
        }
        
        /* Switch case based on NPE operation. Two tasks need to be 
         * executed before sending the request to NPE.
         * 1. Check if the ICV is in the middle of data. If the ICV
         *    is in the middle of data and the NPE operation is 
         *    IX_CRYPTO_NPE_OP_HMAC_VER_ICV or 
         *    IX_CRYPTO_NPE_OP_HMAC_VER_ICV_DECRYPT; copy the original 
         *    ICV value into queue descriptor and store the original ICV 
         *    offset into queue descriptor too. Pass in the new ICV address to 
         *    NPE with the ICV field in the source mbuf is cleared to zero.
         * 2. Check if the operation is non in-place operation, if it
         *    is non in-place operation, shift the mdata pointer in 
         *    destination mbuf back to authStartOffset/cryptStartOffset
         *    depending on the NPE operation. As NPE always start writing to 
         *    offset 0 instead of authStartOffset and cryptStartOffset. Thus 
         *    the mdata pointer need to be shifted in order to get the data 
         *    write to the correct offset
         */        
        switch (npeOperation)
        {
            /* HMAC verification only, or combined service with
             * HMAC verification operation. Notes : the code for 
             * HMAC verification fall thorugh to combined service
             * IX_CRYPTO_NPE_OP_HMAC_VER_ICV_DECRYPT
             */
            case IX_CRYPTO_NPE_OP_HMAC_VER_ICV:
            case IX_CRYPTO_NPE_OP_HMAC_VER_ICV_DECRYPT:
                if (useDiffBuf) /* If non in-place operation */
                {
                    pMbuf = pDestMbuf;
                    
                    /* Shift mbuf mdata pointer to right by authStartOffset */
                    IX_MBUF_MDATA (pDestMbuf) += authStartOffset;
                    
                    /* Get new ICV offset */
                    newIcvOffset -= authStartOffset;
                } /* end of if (useDiffBuf) */

                /* Store original ICV offset into queue descriptor */
                pQDesc->originalIcvOffset = icvOffset;
                
                if ((icvOffset >= authStartOffset) &&
                    (icvOffset < (authStartOffset + authDataLen)))
                {                    
                    /* Get ICV address */
                    tempIcvAddr = (UINT8 *)
                        ixCryptoUtilMbufOffsetToAddressConvert (
                        pSrcMbuf,
                        icvOffset,
                        useDiffBuf);
                        
                    /* Store original ICV value into queue descriptor */
                    memcpy (
                        pQDesc->integrityCheckValue, 
                        tempIcvAddr, 
                        ixCryptoCtx[cryptoCtxId].digestLength);
                        
                    /* Clear ICV field in mbuf to zero */    
                    memset (
                        tempIcvAddr, 
                        0, 
                        ixCryptoCtx[cryptoCtxId].digestLength);
                    
                    /* Assign ICV value location to queue descriptor */    
                    tempIcvAddr = pQDesc->integrityCheckValue;
                }
                else /* ICV not in the middle of data */
                {
                    /* Get ICV address */
                    tempIcvAddr = (UINT8 *)
                        ixCryptoUtilMbufOffsetToAddressConvert (
                            pMbuf,
                            newIcvOffset,
                            useDiffBuf);
                } /* end of if-else (icvOffset) */
                
                /* Convert ICV address to physical address */
                pQDesc->npeQDesc.address.icvAddr
                    = (UINT32)IX_CRYPTO_VIRTUAL_TO_PHYSICAL_TRANSLATION (
                          (UINT32) tempIcvAddr);
                break; 
            
            /* HMAC generation only, or combined service with
             * HMAC generation operation. Notes : the code for 
             * HMAC generation fall thorugh to combined service
             * IX_CRYPTO_NPE_OP_ENCRYPT_HMAC_GEN_ICV
             */
            case IX_CRYPTO_NPE_OP_HMAC_GEN_ICV:
            case IX_CRYPTO_NPE_OP_ENCRYPT_HMAC_GEN_ICV:
            
                if (useDiffBuf) /* If non in-place */
                {
                    pMbuf = pDestMbuf;
                    IX_MBUF_MDATA (pDestMbuf) += authStartOffset;
                    newIcvOffset -= authStartOffset;
                } /* end of if (useDiffBuf) */
                
                /* Get ICV address */
                tempIcvAddr = (UINT8 *)
                    ixCryptoUtilMbufOffsetToAddressConvert (
                        pMbuf,
                        newIcvOffset,
                        useDiffBuf);
                
                /* Convert ICV address into physical address */
                pQDesc->npeQDesc.address.icvAddr
                    = (UINT32)IX_CRYPTO_VIRTUAL_TO_PHYSICAL_TRANSLATION (
                          tempIcvAddr);
                break; 
           
            /* Encrypt/Decrypt onlyoperation */
            default :
                if (useDiffBuf) /* If non in-place operation */
                {
                    
                    IX_MBUF_MDATA (pDestMbuf) += cryptStartOffset;
                } /* end of if (useDiffBuf) */
                break;
        } /* end of switch case (npeOperation) */ 
        
// Jack ++
printk("677 src address is %x\n",pSrcMbuf);

        /* 
         * If we are doing crypto we need to set up IV. No IV needed for 
         * Authentication. The cipher block length check is only performed 
         * when the operationMode is crypt */   
        if(IX_CRYPTO_NPE_OP_CRYPT_MODE_IS_DISABLED != 
           (operationMode & IX_CRYPTO_NPE_OP_CRYPT_ENABLE_MASK )) 
        { 
            /* 
             * We are doing some Crypto: Now we can be sure that the 
             * NpeCryptCfgWord is at start of UINT8 NpeCryptoInfo array
             */ 
#ifdef DEBUG
            /* Check for the correct cipher text and plain text length */          
            status = ixCryptoPerformCipherBlkLengthCheck (
                         operationMode,
                         ixCryptoCtx[cryptoCtxId].cipherBlockLength,
                         cryptStartOffset,
                         cryptDataLen);
                
            if (IX_CRYPTO_ACC_STATUS_SUCCESS != status)
            {
                /* Log error message in debugging mode */
                IX_CRYPTO_ACC_LOG (LOG_MESSAGE, 
                           "IX_CRYPTO_ACC_STATUS_CIPHER_INVALID_BLOCK_LEN.\n", 
                           0, 0, 0, 0, 0, 0);
                /* Need to release descriptor before returning. 
                 * Descriptor release will log a message when fail */
                ixCryptoDescMgmtQDescriptorRelease(pQDesc);
                return status;          
            } /* End of invalid block length */ 
#endif
    
            pCipherCfg = (IxCryptoNpeCryptCfgWord*)ixCryptoCtx[cryptoCtxId].
            pNpeCryptoParam->npeCryptoInfo;
                                    
            /* Check that IV needed, for either CBC or CTR  mode */
            if(IX_CRYPTO_NPE_CRYPT_CFG_ECB != pCipherCfg->cipherMode)
            {
                /* First check IV not NULL */
                if(NULL==pIV)
                {   
                    /* Log error message in debugging mode */
                    IX_CRYPTO_ACC_LOG (LOG_MESSAGE,
                               "Input IV not valid == NULL.\n", 
                               0, 0, 0, 0, 0, 0);
                    /* Descriptor release will log a message when fail */
                    ixCryptoDescMgmtQDescriptorRelease(pQDesc);
                    
                    return IX_CRYPTO_ACC_STATUS_FAIL;  
                } /* End of pIV Null check */
                
                /*
                 * Check if this is AES CTR. CTR Mode must be handled 
                 * differently
                 */
                if (IX_CRYPTO_NPE_CRYPT_CFG_CTR != pCipherCfg->cipherMode)  
                {
                    memcpy (pQDesc->npeQDesc.IV, pIV, 
                        ixCryptoCtx[cryptoCtxId].cipherIvLength);
                }
                else /* Create block for AES */
                {           
                    ixCryptoPerformAesCtrBlkCreate (
                        pQDesc, 
                        pIV, 
                        ixCryptoCtx[cryptoCtxId].
                        securityParameterIndex);
                } /* End of if-else CTR Check */    
            } /* End of if(ECB)  */      
        } /* End of if (Crypto) */
                  
        /* Construct Q descriptor message */   
        pQDesc->cryptoCtxId = cryptoCtxId;
        pQDesc->operStatus = operStatus;
    
        /* Set to NPE operation and initial values in Crypto Param structure */
        pQDesc->npeQDesc.npeOperationMode
            = ixCryptoCtx[cryptoCtxId].npeOperationMode;
      
        /* Convert source mbuf into NPE format. Virtual address must be used when
         * flushing to external memory */
        pQDesc->npeQDesc.srcMbufAddr = (UINT32)
            (ixCryptoUtilMbufToNpeFormatConvert(pSrcMbuf));
          
        /* If operationMode is in place, DestMbuf will be empty and don't need to
         * be written to the pQDesc or flushed to memory.
         */
        if (IX_CRYPTO_NPE_OP_TRANSFER_MODE_IS_IN_PLACE != 
            (operationMode & IX_CRYPTO_NPE_OP_TRANSFER_MODE_MASK))
        {
            /* Convert destination mbuf into NPE format */
            pQDesc->npeQDesc.destMbufAddr = (UINT32)
                (ixCryptoUtilMbufToNpeFormatConvert (pDestMbuf));
        }
        
        /* Offset and length of data */
        pQDesc->npeQDesc.authStartOffset = authStartOffset;
    
        pQDesc->npeQDesc.authLength = authDataLen;
        pQDesc->npeQDesc.cryptStartOffset = cryptStartOffset;
        pQDesc->npeQDesc.cryptLength = cryptDataLen;
        
        /* Write Crypto Param Address and flush*/
        pQDesc->npeQDesc.cryptoCtxNpeAddr 
            = (UINT32)IX_CRYPTO_VIRTUAL_TO_PHYSICAL_TRANSLATION (
            ixCryptoCtx[cryptoCtxId].pNpeCryptoParam);
        
        /* Flush Q descriptor message from cache */
        IX_CRYPTO_DATA_CACHE_FLUSH (pQDesc, sizeof (IxCryptoQDescriptor));
        
        /* translate pQDesc after flush */    
        pQDesc = (IxCryptoQDescriptor*) 
            IX_CRYPTO_VIRTUAL_TO_PHYSICAL_TRANSLATION (pQDesc);
    
        /* Enqueue request to hardware queue using QMgr */
        status = ixCryptoQAccessQueueWrite (IX_CRYPTO_ACC_CRYPTO_REQ_Q,
                     (UINT32 *) &pQDesc);               
    
        if (IX_CRYPTO_ACC_STATUS_SUCCESS == status)
        {
            ixCryptoCtx[cryptoCtxId].reqCount++;
// Jack ++
printk("816 src address is %x\n",pSrcMbuf);


        }
        else /* End of status check for Queue Write */ 
        {
            /* Log error message in debugging mode */
            IX_CRYPTO_ACC_LOG (LOG_MESSAGE,
                "ixCryptoQAccessQueueWrite failed.\n", 
                0, 0, 0, 0, 0, 0);
                       
            /* Convert mbuf back to host format from NPE format */
            pSrcMbuf = ixCryptoUtilMbufFromNpeFormatConvert
                            ((IX_MBUF *) pQDesc->npeQDesc.srcMbufAddr);
            pDestMbuf = ixCryptoUtilMbufFromNpeFormatConvert
                            ((IX_MBUF *) pQDesc->npeQDesc.destMbufAddr);

            /* Convert Q descriptor address back to virtual address 
             * before releasing it
             */
            pQDesc = (IxCryptoQDescriptor *) 
                         IX_CRYPTO_PHYSICAL_TO_VIRTUAL_TRANSLATION(
                         (UINT32) pQDesc);
    
            /* When the Queue write fails, pQDesc needs to be released */
            ixCryptoDescMgmtQDescriptorRelease(pQDesc);
     
        } /* End of if-else for Queue Write check */
        return status;
    }
    else /* end of if (IX_CRYPTO_IS_SERVICE_INITIALIZED) &&
        (IX_CRYPTO_IS_SERVICE_STOPPED) */
    {
        /* Log error message in debugging mode */
        IX_CRYPTO_ACC_LOG (LOG_ERROR,
                   "IxCrypto is not initialized or it is stopped.\n", 
                   0, 0, 0, 0, 0, 0);   
        return IX_CRYPTO_ACC_STATUS_FAIL;
    }/* end of if-else (IX_CRYPTO_IS_SERVICE_INITIALIZED) &&
    (IX_CRYPTO_IS_SERVICE_STOPPED) */  
} /* end of ixCryptoAccAuthCryptPerform () function */


/**
 * @fn ixCryptoAccCtxUnregister
 * @brief Unregister the CryptoCtx from Cryptographic Context Database.
 *
 */
PUBLIC IxCryptoAccStatus
ixCryptoAccCtxUnregister (UINT32 cryptoCtxId)
{
    IxCryptoAccStatus status;

    /* Check if the access component has been intialized and the crypto 
     * service is running 
     */
    if (IX_CRYPTO_IS_SERVICE_INITIALIZED()
        && !IX_CRYPTO_IS_SERVICE_STOPPED())
    {
        /* Check if the Crypto Context ID provided points to a valid Crypto
         * Context 
         */
        status = ixCryptoCCDMgmtCtxValidCheck (cryptoCtxId);
        
        if (IX_CRYPTO_ACC_STATUS_SUCCESS == status) 
        {
            /* Check if all requests from this cryptoCtxId have been
             * completed 
             */
            if (IX_CRYPTO_CTX_IS_ALL_TASK_DONE (
                ixCryptoCtx[cryptoCtxId].reqCount,
                ixCryptoCtx[cryptoCtxId].reqDoneCount))
            {
                /* Unregister the Crypto Context */
                return ixCryptoCCDMgmtCryptoCtxRelease (cryptoCtxId);              
            }
            else /* IX_CRYPTO_CTX_IS_ALL_TASK_DONE == FALSE */
            {
                /* There are requests pending in the queue for this Crypto 
                 * Context, cancel the unregistration and send Retry 
                 * message back to the client 
                 */
                return IX_CRYPTO_ACC_STATUS_RETRY;
            } /* end of if-else (IX_CRYPTO_CTX_IS_ALL_TASK_DONE) */
        }
        else /* Crypto context does not exist */
        {
            /* Log error message in debugging mode */
            IX_CRYPTO_ACC_LOG (LOG_MESSAGE, 
                "Crypto Context %d does not exist.\n", 
                cryptoCtxId, 
                0, 0, 0, 0, 0);
                
            return status;
        } /*end of if (status) for (ixCryptoCCDMgmtCtxValidCheck) */
    }
    else /* (IX_CRYPTO_IS_SERVICE_INITIALIZED) &&
          * (!IX_CRYPTO_IS_SERVICE_STOPPED) 
          */
    {
        /* Log error message in debugging mode */
            IX_CRYPTO_ACC_LOG (LOG_ERROR, 
                "Component has not been initialized\n", 
                0, 0, 0, 0, 0, 0);
                
        return IX_CRYPTO_ACC_STATUS_FAIL;
    } /* end of if-else (IX_CRYPTO_IS_SERVICE_INITIALIZED) &&
       * (!IX_CRYPTO_IS_SERVICE_STOPPED) 
       */
} /* End of ixCryptoAccCtxUnregister () */



/**
 * @fn   ixCryptoAccCryptoServiceStop
 * @brief API to stop the security hardware accelerator services.
 *
 */
PUBLIC IxCryptoAccStatus
ixCryptoAccCryptoServiceStop (void)
{
    UINT32 timer = 0;
        
    /* Set the flag to TRUE to prevent new service request */
    ixCryptoAccServiceStop = TRUE;
    
    while (IX_CRYPTO_ACC_STATUS_SUCCESS 
        != ixCryptoDescMgmtAllQDescriptorInPoolCheck())
    {
        /* Put the task into sleep mode and wait for the NPE to complete
         * all the tasks in the queue 
         */
        ixOsServTaskSleep(IX_CRYPTO_ACC_DELAY_IN_MS);

        /* Timer to avoid infinite lock */
        timer++;
        if (IX_CRYPTO_ACC_TIMER_COUNT < timer )
        {
            /* Log error message in debugging mode */
            IX_CRYPTO_ACC_LOG (
            LOG_FATAL,
            "Stop function failed!\n", 
            0, 0, 0, 0, 0, 0);
            return IX_CRYPTO_ACC_STATUS_FAIL;
    
        } /* End of if timer loop */

    } /* end of while (ixCryptoDescMgmtAllQDescriptorInPoolCheck) */
    
    /* Release all Crypto Contexts and this will not fail */
    ixCryptoCCDMgmtCryptoCtxReleaseAll ();

    /* Release memeory pool allocated to NPE Crypto Param */
    ixCryptoCCDMgmtNpeCryptoParamPoolFree ();
    
    /* Release memory pool allocated to Desc pool */
    ixCryptoDescMgmtDescPoolFree ();
    
    return IX_CRYPTO_ACC_STATUS_SUCCESS;
} /* end of ixCryptoAccCryptoServiceStop () function */



/**
 * @fn   ixCryptoAccShow
 * @brief API for printing statistic and status.
 *
 */

PUBLIC void
ixCryptoAccShow (void)
{
    /* Show request queue and done queue status */
    ixCryptoQAccessQueueStatusShow (IX_CRYPTO_ACC_CRYPTO_REQ_Q); 
    ixCryptoQAccessQueueStatusShow (IX_CRYPTO_ACC_CRYPTO_DONE_Q);
       
    /* To show the number of descriptors have been used, and the pool 
     * usage
     */
    ixCryptoDescMgmtShow ();

    /* To show number of Crypto Contexts have been registered */
    ixCryptoCCDMgmtShow ();


} /* end of ixCryptoAccShow () function */


/**
 * @fn   ixCryptoAccShowWithId
 * @brief API for printing statistic and status for particular Context.
 *
 */

PUBLIC void
ixCryptoAccShowWithId (UINT32 cryptoCtxId)
{
    if (IX_CRYPTO_ACC_STATUS_SUCCESS == 
    ixCryptoCCDMgmtCtxValidCheck (cryptoCtxId))
    {
        /* Show request queue and done queue status, number of descriptors have 
         * been used, and the pool usage and show number of Crypto Contexts
         * that have been registered */
        ixCryptoAccShow ();

        /* To show contents of Crypto Context which has been registered */
        ixCryptoCCDMgmtCryptoCtxShow (cryptoCtxId);
    }
    else
    {
        /* Log error message in debugging mode */
        IX_CRYPTO_ACC_LOG (LOG_MESSAGE,
                   "Invalid cryptoCtxId.\n", 
                   0, 0, 0, 0, 0, 0);
    } /* End of if-else Ctx Check */
    
    
} /* end of ixCryptoAccShowWithId () */
