/**
 * @file IxCryptoAccDescMgmt.c
 *
 * @date October-03-2002
 *
 * @brief Source file for Descriptors Management module
 *
 *
 * Design Notes:
 *
 * -- Intel Copyright Notice -- 
 *
 * @par
 * INTEL CONFIDENTIAL
 * 
 * @par
 * Copyright 2000-2002 Intel Corporation All Rights Reserved.
 * 
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.  The Material
 * contains trade secrets and proprietary and confidential information of
 * Intel or its suppliers and licensors.  The Material is protected by
 * worldwide copyright and trade secret laws and treaty provisions. No
 * part of the Material may be used, copied, reproduced, modified,
 * published, uploaded, posted, transmitted, distributed, or disclosed in
 * any way without Intel's prior express written permission.
 * 
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel or otherwise.  Any license under
 * such intellectual property rights must be express and approved by
 * Intel in writing.
 * 
 * -- End Intel Copyright Notice -- 
 */


/*
* Put the system defined include files required.
*/
#include <stdio.h>
#include <stdlib.h>


/*
 * Put the user defined include files required.
 */
#include "IxTypes.h"
#include "IxOsServices.h"
#include "IxCryptoAccDescMgmt_p.h"
#include "IxCryptoAccUtilities_p.h"


#define IX_CRYPTO_Q_DESC_ADDR_RANGE_MASK    0xF0000000
                                                /**< Mask for queue descriptor
                                                 * address range, the upper 
                                                 * 4-bit (MSB) are unused
                                                 */

#define IX_CRYPTO_Q_DESC_ADDR_VALID_RANGE   0   /**< Upper 4-nit are zeroes */                                                 

/*
 * Variable declarations global to this file only.  Externs are followed by
 * static variables.
 */
static IxCryptoQDescriptor* qDescList[MAX_DESCRIPTORS_NUM_IN_POOL];  
                                 /**< array of Q descriptors pointers,
                                  * this array is used as a FILO push-pop 
                                  * list in descriptor get and release 
                                  * funciton
                                  */
static UINT32 cryptoQDescCount = 0;   /**< Q descriptor index to qDescList 
                                         * array (Q descriptor list head 
                                         * pointer)
                                         */
static INT32 descMgmtLock;       /**< Protect critical sections in this 
                                  * module
                                  */
static UINT8 *pDescPool = NULL;  /**< Descriptor pool pointer */
static BOOL initDone = FALSE;    /**< Flag to prevent multiple 
                                  * initialization
                                  */

#ifdef DEBUG
static UINT32 qDescDrainCount = 0; /**< Number of times run out of Q 
                                    * descriptors 
                                    */
#endif /* ndef DEBUG */



/**
 * @fn      ixCryptoDescMgmtInit
 * @brief   Initialize descriptor pool
 *
 */
IxCryptoAccStatus
ixCryptoDescMgmtInit (void)
{
    UINT8 *pQDescPool = NULL;
    UINT32 i;
    UINT32 qDescPhyAddr;
    
    if (!initDone)
    {
        /* Allocate memory to Q Descriptors Pool */
        pQDescPool = IX_CRYPTO_ACC_DRV_DMA_MALLOC (
                         MAX_DESCRIPTORS_NUM_IN_POOL 
                         * IX_CRYPTO_Q_DESC_SIZE);

        if (NULL == pQDescPool) /* if NULL */
        {
            /* Log error message in debugging mode */
            IX_CRYPTO_ACC_LOG (
                LOG_ERROR, 
                "Memory allocation for Descriptors Pool failed\n", 
                0, 0, 0, 0, 0, 0);
    
            /* Memory allocation failed */
            return IX_CRYPTO_ACC_STATUS_FAIL;
        } /* end of if (pQDescPool) */
        
        /* Store descriptor pool pointer, this pointer will be 
         * used  in descriptor pool free function
         */
        pDescPool = pQDescPool;
        
        for (i = 0;
            i < MAX_DESCRIPTORS_NUM_IN_POOL;
            i++)
        {
            /* Check if the memory allocated to descriptor pool is within the
             * 28-bit address range, MSB 4-bit have been reserved for other 
             * usage, effective Q descriptor  address will be lower 28-bit ONLY
             */
            qDescPhyAddr = IX_CRYPTO_VIRTUAL_TO_PHYSICAL_TRANSLATION(
                                (UINT32) pQDescPool);
             
            if (IX_CRYPTO_Q_DESC_ADDR_VALID_RANGE 
                != (qDescPhyAddr & IX_CRYPTO_Q_DESC_ADDR_RANGE_MASK))
            {
                /* Log error message in debugging mode */
                IX_CRYPTO_ACC_LOG (
                    LOG_ERROR, 
                    "Invalid address range %x is allocated to DescPool\n", 
                    (int) qDescPhyAddr, 
                    0, 0, 0, 0, 0);
        
                /* Memory allocation failed */
                return IX_CRYPTO_ACC_STATUS_FAIL;
            } /* end of if (Q Desc address range check) */
        
            /* Allocate a Q descriptor from the Q Descriptors Pool
             * Assign Q descriptor to qDescList 
             */
            qDescList[i] = (IxCryptoQDescriptor *) pQDescPool;
        
            /* Move Q descriptor pool pointer to point to next Q 
             * descriptor element 
             */
            pQDescPool += IX_CRYPTO_Q_DESC_SIZE;                        
        } /* end of for (i) */

        /* Set qDescList head pointer to last element in the list */
        cryptoQDescCount = MAX_DESCRIPTORS_NUM_IN_POOL;

        /* Set flag to indicate this module has been initialized */
        initDone = TRUE;
        
        return IX_CRYPTO_ACC_STATUS_SUCCESS;
    }
    else /* Desc Mgmt module has been initialized */
    {
        return IX_CRYPTO_ACC_STATUS_FAIL;
    } /* end of if-else (initDone) */
} /* end of function ixCryptoDescMgmtInit () function */



/**
 * @fn      ixCryptoDescMgmtQDescriptorGet
 * @brief   Get a queue descriptor from a descriptor pool
 *
 */
IxCryptoAccStatus
ixCryptoDescMgmtQDescriptorGet (IxCryptoQDescriptor **pQDescriptor)
{
    IxCryptoAccStatus status;
    
    /* Lock qDescList to protect critical section */
    IX_CRYPTO_ACC_LOCK (descMgmtLock);

    if (cryptoQDescCount > 0) /* Unused Q descriptor available */
    {
        /* Get a Q descriptor from the pool */
        *pQDescriptor = qDescList[--cryptoQDescCount];
        
        /* Set src and dest mbuf to NULL */
        (*pQDescriptor)->npeQDesc.srcMbufAddr = (UINT32) NULL;
        (*pQDescriptor)->npeQDesc.destMbufAddr = (UINT32) NULL;

        status = IX_CRYPTO_ACC_STATUS_SUCCESS;
    }
    else /* cryptoQDescCount <= 0 */
    {
#ifdef DEBUG
        qDescDrainCount++;
#endif /* ndef DEBUG */
        /* Log error message in debugging mode */
        IX_CRYPTO_ACC_LOG(LOG_MESSAGE, "Run out of Q descriptor.\n",
            0, 0, 0, 0, 0, 0);
            
        /* Run out of Q descriptor in the pool */
        status = IX_CRYPTO_ACC_STATUS_FAIL;
    } /* end of if-else (cryptoQDescCount) */

    /* Unlock given mutex */
    IX_CRYPTO_ACC_UNLOCK (descMgmtLock);
    
    return status;
} /* end of ixCryptoDescMgmtQDescriptorGet () function */



/**
 * @fn      ixCryptoDescMgmtQDescriptorRelease
 * @brief   Release a queue descriptor to a descriptor pool
 *
 */
IxCryptoAccStatus
ixCryptoDescMgmtQDescriptorRelease (IxCryptoQDescriptor *pQDescriptor)
{
    IxCryptoAccStatus status;

    /* Lock qDescList to protect critical section */
    IX_CRYPTO_ACC_LOCK (descMgmtLock);

    if (cryptoQDescCount < MAX_DESCRIPTORS_NUM_IN_POOL)
    {
        /* Push the Q descriptor back to the pool */
        qDescList[cryptoQDescCount++] = pQDescriptor;

        status = IX_CRYPTO_ACC_STATUS_SUCCESS;
    }
    else /* cryptoQDescCount >= MAX_DESCRIPTORS_NUM_IN_POOL */
    {
        /* Log error message in debugging mode */
        IX_CRYPTO_ACC_LOG(LOG_ERROR, "Q descriptor release failed.\n",
            0, 0, 0, 0, 0, 0);

        /* Q Descriptor Pool overflow */
        status = IX_CRYPTO_ACC_STATUS_FAIL;
    } /* end of if-else (cryptoQDescCount) */

    /* Unlock given mutex */
    IX_CRYPTO_ACC_UNLOCK (descMgmtLock);

    return status;
} /* end of ixCryptoDescMgmtQDescriptorRelease () function */



/**
 * @fn      ixCryptoDescMgmtShow
 * @brief   Show the numbers of descriptor have been used
 *
 */
void
ixCryptoDescMgmtShow (void)
{
    /* Descriptor Management statistics display */
    printf ("\n\nQ Descriptors Pool Statistics\n");
    printf ("-----------------------------\n");
    printf ("Pool size in bytes ......... : %10lu bytes\n",
        IX_CRYPTO_Q_DESC_SIZE * MAX_DESCRIPTORS_NUM_IN_POOL);
    printf ("Q Descriptor size .......... : %10lu bytes\n",
        IX_CRYPTO_Q_DESC_SIZE);
    printf ("Pool depth ................. : %10u descriptors\n",
        MAX_DESCRIPTORS_NUM_IN_POOL);
    printf ("Number of Q descriptors used : %10u\n",
        (MAX_DESCRIPTORS_NUM_IN_POOL - cryptoQDescCount));

#ifdef DEBUG
    printf ("No of times run out of Q descriptors : %10u\n", 
        qDescDrainCount);
#endif /* ndef DEBUG */
} /* end of ixCryptoDescMgmtShow () function */



/**
 * @fn ixCryptoDescMgmtAllQDescriptorInPoolCheck
 *
 * @brief This inline function is used to check whether all the allocated 
 *        descriptors have been pushed back to the descriptors pool. If all 
 *        descriptors are unused, it indicates that there is no pending task 
 *        in either Crypto Req Q or Crypto Done Q.
 *
 */
INLINE IxCryptoAccStatus
ixCryptoDescMgmtAllQDescriptorInPoolCheck (void)
{
    /* default to fail */
    IxCryptoAccStatus status = IX_CRYPTO_ACC_STATUS_FAIL;
    
    if (MAX_DESCRIPTORS_NUM_IN_POOL == cryptoQDescCount)
    {
        status = IX_CRYPTO_ACC_STATUS_SUCCESS;
    }
    
    return status;
} /* end of ixCryptoDescMgmtAllQDescriptorInPoolCheck () function */



/**
 * @fn ixCryptoDescMgmtDescPoolFree
 *
 * @brief To free the memory allocated to descriptor pool through malloc 
 *        function.
 *
 */
void
ixCryptoDescMgmtDescPoolFree (void)
{
    /* Check if the pool has been allocated */
    if (NULL != pDescPool)
    {
        /* free memory allocated */
        IX_CRYPTO_ACC_DRV_DMA_FREE (pDescPool);
        
        /* Set flag to FALSE to indicate the Desc pool has been freed */
        initDone = FALSE;
        
        printf ("Descriptor pool has been freed.\n");
    } /* end of (pDescPool) */
} /* end of ixCryptoDescMgmtDescPoolFree () function */
