/**
 * @file IxCryptoAccUtilities.c
 *
 * @date October-03-2002
 *
 * @brief Source file for the Utilities module.
 *
 *
 * Design Notes:
 *
 * -- Intel Copyright Notice -- 
 *
 * @par
 * INTEL CONFIDENTIAL
 * 
 * @par
 * Copyright 2000-2002 Intel Corporation All Rights Reserved.
 * 
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.  The Material
 * contains trade secrets and proprietary and confidential information of
 * Intel or its suppliers and licensors.  The Material is protected by
 * worldwide copyright and trade secret laws and treaty provisions. No
 * part of the Material may be used, copied, reproduced, modified,
 * published, uploaded, posted, transmitted, distributed, or disclosed in
 * any way without Intel's prior express written permission.
 * 
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel or otherwise.  Any license under
 * such intellectual property rights must be express and approved by
 * Intel in writing.
 * 
 * -- End Intel Copyright Notice -- 
 */


/*
 * Put the user defined include files required.
 */
#include "IxCryptoAccUtilities_p.h"


/**
 * @fn ixCryptoUtilMbufOffsetToAddressConvert
 * @brief To convert offset to data block of mbuf into 32-bit address.
 *
 */
UINT32 
ixCryptoUtilMbufOffsetToAddressConvert (
    IX_MBUF *pMbuf,
    INT32 offset,
    BOOL useDiffBuf)
{
    UINT32 AccumPktLen;
    UINT32 offsetDiff;
    IX_MBUF *pBuffer = pMbuf;
    
    /* Check if in-place or non in-place operation */    
    if (!useDiffBuf)
    {
        /* Get accumulated data length in chained mbuf */
        AccumPktLen = IX_MBUF_MLEN (pBuffer);
        
        /* While offset > AccumPktLen, checking for mbuf header fields and 
         * length are done in IxCryptoAcc.c and IxCryptoAccPerform.c
         */
        while (offset > AccumPktLen)
        {
            /* Traverse to next mbuf buffer */
            pBuffer = IX_MBUF_NEXT_BUFFER_IN_PKT_PTR (pBuffer);
            
            /* Get accumulated data length in chained mbuf */
            AccumPktLen += IX_MBUF_MLEN (pBuffer);
        }
        
        /* Get the difference between the mbuf data offset with the accumulated
         * packet length, in order to move the pointer to the place pointed
         * by offset
         */
        offsetDiff = IX_MBUF_MLEN (pBuffer) + (offset - AccumPktLen);
        
        /* Return address of data pointer in mbuf */
        return (((UINT32) IX_MBUF_MDATA (pBuffer)) + offsetDiff);
    }
    else /* non in-place operation */
    {
        return (((UINT32) IX_MBUF_MDATA (pBuffer)) + offset);
    }
} /* end of ixCryptoUtilMbufOffsetToAddressConvert () function */                                        



/**
 * @fn ixCryptoUtilMbufToNpeFormatConvert
 * @brief To convert chained mbuf to NPE format, including endianess
 *        conversion (to Big Endian) and virtual address to physical 
 *        address conversion.
 *
 */

IX_MBUF*
ixCryptoUtilMbufToNpeFormatConvert (IX_MBUF *pMbuf)
{
    IX_MBUF *pChain = pMbuf;
    IX_MBUF *pTempBuf = NULL;
   
    /* Check if the mbuf pointer is NULL, if NULL, return the pointer 
     * without doing anything 
     */
    if (NULL == pMbuf)
    {
        return pMbuf;
    } /* end of if (pMbuf) */
      
    /* Convert the address of the mbuf into physical adddress */
    pMbuf = (IX_MBUF *) IX_CRYPTO_VIRTUAL_TO_PHYSICAL_TRANSLATION 
                            ((UINT32) pMbuf);
                            
    /* Convert the address of mbuf into network order (big Endian) */
    IX_CRYPTO_CONVERT_WORD_TO_NETWORK_ORDER ((UINT32) pMbuf);
   
    while (pChain) /* while (pChain != NULL) */
    {
        /* Save the address of the next buffer in the chain and the pointer 
         * to the data 
         */
        pTempBuf = IX_MBUF_NEXT_BUFFER_IN_PKT_PTR (pChain);
      
        /* Flush the cache lines associated with cluster into external 
         * memory 
         */
        IX_CRYPTO_DATA_CACHE_FLUSH (
            IX_MBUF_MDATA (pChain), 
            IX_MBUF_MLEN (pChain));

        /* Convert the relevant addresses */
        IX_MBUF_NEXT_BUFFER_IN_PKT_PTR (pChain) 
            = (IX_MBUF *) IX_CRYPTO_VIRTUAL_TO_PHYSICAL_TRANSLATION
                   ((UINT32) IX_MBUF_NEXT_BUFFER_IN_PKT_PTR (pChain));
        IX_MBUF_MDATA (pChain) 
            = (char *) IX_CRYPTO_VIRTUAL_TO_PHYSICAL_TRANSLATION
                   ((UINT32) IX_MBUF_MDATA (pChain));

        /* Convert mbuf pointer and data into newwork order (Big Endian) */
        IX_CRYPTO_CONVERT_WORD_TO_NETWORK_ORDER (
            (UINT32) IX_MBUF_NEXT_BUFFER_IN_PKT_PTR (pChain));
        IX_CRYPTO_CONVERT_WORD_TO_NETWORK_ORDER (
            (UINT32) IX_MBUF_MDATA (pChain));
        IX_CRYPTO_CONVERT_WORD_TO_NETWORK_ORDER (
            (UINT32) IX_MBUF_MLEN (pChain));

        /* Flush the cache lines associated with the mbuf header into 
         * external memory 
         */
        IX_CRYPTO_DATA_CACHE_FLUSH (pChain, sizeof(IX_MBUF));

        /* next */
        pChain  = pTempBuf;
    }
   
    return pMbuf;
} /* end of ixCryptoUtilMbufToNpeFormatConvert () function */



/**
 * @fn ixCryptoUtilMbufFromNpeFormatConvert
 * @brief To convert chained mbuf from NPE format, including endianess
 *        conversion to host endianess and physical address to virtual 
 *        address conversion.
 *
 */
IX_MBUF*
ixCryptoUtilMbufFromNpeFormatConvert (IX_MBUF *pMbuf)
{
    IX_MBUF *pChain = NULL;
    IX_MBUF *pTempBuf = NULL;
    
    /* Check if the mbuf pointer is NULL, if NULL, return the pointer 
     * without doing anything 
     */
    if (NULL == pMbuf)
    {
        return pMbuf;
    } /* end of if (pMbuf) */
    
    /* Convert the address of mbuf into host order */
    IX_CRYPTO_CONVERT_WORD_TO_HOST_ORDER ((UINT32) pMbuf);
    
    /* Convert the address of the mbuf into virtual address */
    pMbuf = (IX_MBUF *) IX_CRYPTO_PHYSICAL_TO_VIRTUAL_TRANSLATION
                            ((UINT32) pMbuf);
                            
    pChain = pMbuf; /* Assign pMbuf to pChain */
    
    while (pChain) /* while (pChain != NULL) */
    {
        /* Invalidate the mbuf header prior to operating on it */
        IX_ACC_DATA_CACHE_INVALIDATE (pChain, sizeof(IX_MBUF));
        
        /* Convert m_data pointer in mbuf */
        IX_CRYPTO_CONVERT_WORD_TO_HOST_ORDER (
            (UINT32) IX_MBUF_MDATA (pChain));
        IX_MBUF_MDATA (pChain)
            = (char *) IX_CRYPTO_PHYSICAL_TO_VIRTUAL_TRANSLATION
                    (IX_MBUF_MDATA (pChain));

        /* Convert to host order */
        IX_CRYPTO_CONVERT_WORD_TO_HOST_ORDER (
            (UINT32) IX_MBUF_MLEN (pChain));
                
        /* Invalidate the cluster if present in the cache */
        IX_CRYPTO_DATA_CACHE_INVALIDATE (
            IX_MBUF_MDATA (pChain), 
            IX_MBUF_MLEN (pChain));

        /* Convert mbuf pointer into host order */
        IX_CRYPTO_CONVERT_WORD_TO_HOST_ORDER (
            (UINT32) IX_MBUF_NEXT_BUFFER_IN_PKT_PTR (pChain));

        /* Convert mbuf address */
        IX_MBUF_NEXT_BUFFER_IN_PKT_PTR (pChain)
            = (IX_MBUF *) IX_CRYPTO_PHYSICAL_TO_VIRTUAL_TRANSLATION
                  (IX_MBUF_NEXT_BUFFER_IN_PKT_PTR (pChain));

        /* Save the address of the next buffer in the chain */
        pTempBuf = IX_MBUF_NEXT_BUFFER_IN_PKT_PTR (pChain);

        /* next */
        pChain  = pTempBuf;
    }
    
    return pMbuf;
} /* end of ixCryptoUtilMbufFromNpeFormatConvert () function */
