/**
 * @file IxCryptoAccUtilities_p.h
 *
 * @date October-03-2002
 *
 * @brief Header file for the utilities module.
 *
 *
 * Design Notes:
 *
 * -- Intel Copyright Notice -- 
 *
 * @par
 * INTEL CONFIDENTIAL
 * 
 * @par
 * Copyright 2000-2002 Intel Corporation All Rights Reserved.
 * 
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.  The Material
 * contains trade secrets and proprietary and confidential information of
 * Intel or its suppliers and licensors.  The Material is protected by
 * worldwide copyright and trade secret laws and treaty provisions. No
 * part of the Material may be used, copied, reproduced, modified,
 * published, uploaded, posted, transmitted, distributed, or disclosed in
 * any way without Intel's prior express written permission.
 * 
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel or otherwise.  Any license under
 * such intellectual property rights must be express and approved by
 * Intel in writing.
 * 
 * -- End Intel Copyright Notice -- 
 */


#ifndef IXCRYPTOACCUTILITIES_P_H
#define IXCRYPTOACCUTILITIES_P_H

/*
 * Os/System dependancies.
 */
#include "IxTypes.h"
#include "IxOsServices.h"
#include "IxOsCacheMMU.h"
#include "IxOsBuffMgt.h"

#ifdef __vxworks

#include <vxWorks.h>
#include <cacheLib.h>

#endif /* def __vxworks */


/**
 * @def IX_CRYPTO_CONVERT_WORD_TO_NETWORK_ORDER
 *
 * @brief Convert words into network order (Big Endian). If the host order 
 *        is also the same as network order, this macro will do nothing or 
 *        else it will convert words from Little Endian to Big Endian.
 *
 * @note Example of usage: IX_CRYPTO_CONVERT_WORD_TO_NETWORK_ORDER(x)
 *  	 This macro does not do anything for the current CSR version, 
 *  	 assuming Big Endian
 * @param UINT32 [inout] data - word to be converted.
 *
 * @return: none
 *
 */ 
#define IX_CRYPTO_CONVERT_WORD_TO_NETWORK_ORDER(data) (data)


/**
 * @def IX_CRYPTO_CONVERT_WORD_TO_HOST_ORDER
 *
 * @brief Convert words into host order (Little Endian or Big Endian). 
 *        If the host order is also the same to network order, this macro 
 *        will do nothing or else it will convert words into Little Endian
 *
 * @note Example of usage: IX_CRYPTO_CONVERT_WORD_TO_HOST_ORDER(x).
 * 	     This macro does not do anything for the current CSR version, 
 *  	 assuming Big Endian
 *
 * @param UINT32 [inout] data - words to be converted.
 *
 * @return none
 *
 */
#define IX_CRYPTO_CONVERT_WORD_TO_HOST_ORDER(data) (data)


/**
 * @def IX_CRYPTO_VIRTUAL_TO_PHYSICAL_TRANSLATION
 *
 * @brief To convert logical/virtual address to physical address. It will
 *        return a virtual address for the provided  physical address.
 *
 * @note Example of usage x = IX_CRYPTO_VIRTUAL_TO_PHYSICAL_TRANSLATION(x)
 * 
 * @param INT32 [in] address - virtual address to be translated.
 *
 * @return virtual address for the provided  physical address.
 *
 */
#define IX_CRYPTO_VIRTUAL_TO_PHYSICAL_TRANSLATION(address)\
             IX_MMU_VIRTUAL_TO_PHYSICAL_TRANSLATION(address)


/**
 * @def IX_CRYPTO_PHYSICAL_TO_VIRTUAL_TRANSLATION
 *
 * @brief To convert physical address to logical/virtual address. It will
 *        return a physical address for the provided virtual.
 *
 * @note Example of usage: x = IX_CRYPTO_PHYSICAL_TO_VIRTUAL_TRANSLATION(x)
 *
 * @param UINT32 [in] address - physical address to be translated.
 *
 * @return physical address for the provided virtual address
 *
 */
#define IX_CRYPTO_PHYSICAL_TO_VIRTUAL_TRANSLATION(address)\
             IX_MMU_PHYSICAL_TO_VIRTUAL_TRANSLATION (address)


/**
 * @def IX_CRYPTO_ACC_DRV_DMA_MALLOC
 *
 * @brief To allocate memory for driver use, that will be shared between
 *        XSCale and NPE.
 *
 * @note Example of usage: x = IX_CRYPTO_ACC_DRV_DMA_MALLOC(x)
 *
 * @param UINT32 [in] size - memory size in bytes to be allocated.
 *
 * @return  void * Pointer to memory that can be used between XScale and NPE's.
 *
 */
#define IX_CRYPTO_ACC_DRV_DMA_MALLOC(size)\
             IX_ACC_DRV_DMA_MALLOC(size)
             

/**
 * @def IX_CRYPTO_ACC_DRV_DMA_FREE
 *
 * @brief To free the memory allocated through IX_CRYPTO_ACC_DRV_DMA_MALLOC
 *        function.
 *
 * @note Example of usage: IX_CRYPTO_ACC_DRV_DMA_FREE(pData)
 *
 * @param UINT8 [in] *pData - memory block pointer to be freed.
 *
 * @return None
 *
 */
#ifdef __vxworks
 
#define IX_CRYPTO_ACC_DRV_DMA_FREE(pData) cacheDmaFree (pData)            

#else

#define IX_CRYPTO_ACC_DRV_DMA_FREE(pData) (pData) /* Note: Do nothing in 
                                                   * Linux platform. Will be 
                                                   * ported to Linux platform.
                                                   */
#endif /* def __vxworks */
             
/**
 * @def IX_CRYPTO_DATA_CACHE_INVALIDATE
 *
 * @brief To invalidate a cache range.
 *
 * @param UINT32 [in] address - cache address to be invalidated.
 * @param UINT32 [in] size - the size in bytes of cache to be invalidated.
 *
 * @return none
 *
 */
#define IX_CRYPTO_DATA_CACHE_INVALIDATE(address, size)\
             IX_ACC_DATA_CACHE_INVALIDATE(address, size)


/**
 * @def IX_CRYPTO_DATA_CACHE_FLUSH
 *
 * @brief To flush a cache range to physical memory. 
 *
 * @param UINT32 [in] address - cache address to be flushed to the physical 
 *	  memory.
 * @param UINT32 [in] size - the size in bytes of cache to be flushed to the 
 *	  physical memory.
 *
 * @return none
 *
 */
#define IX_CRYPTO_DATA_CACHE_FLUSH(address, size)\
             IX_ACC_DATA_CACHE_FLUSH(address, size)



/**
 * @fn ixCryptoUtilMbufOffsetToAddressConvert
 *
 * @brief To convert offset to data block of mbuf into 32-bit address.
 *
 * @param IX_MBUF [in] *pMbuf - pointer to the mbuf.
 * @param INT32 [in] offset - offset of the mbuf.
 * @param BOOL [in] useDiffBuf - TRUE if non in-place operation
 *                               FALSE if in-place operation 
 *
 * @return UINT32 address of data pointer in mbuf.
 *
 * Reentrant    : no
 * ISR Callable : no
 *
 */
UINT32 
ixCryptoUtilMbufOffsetToAddressConvert (
    IX_MBUF *pMbuf,
    INT32 offset,
    BOOL useDiffBuf);
                                        

/**
 * @fn ixCryptoUtilMbufToNpeFormatConvert
 *
 * @brief To convert chained mbuf to NPE format, including endianess
 *        conversion (to Big Endian) and virtual address to physical address
 *        conversion.  And then flush the data to physical memory.
 *
 * @param IX_MBUF [in] *pMbuf - pointer to the mbuf data block.
 *
 * @return IX_MBUF* - the converted mbuf is returned.
 *
 * Reentrant    : no
 * ISR Callable : no
 *
 */
IX_MBUF*
ixCryptoUtilMbufToNpeFormatConvert (IX_MBUF *pMbuf);


/**
 * @fn ixCryptoUtilMbufFromNpeFormatConvert
 *
 * @brief To convert chained mbuf from NPE format, including endianess
 *        conversion to host endianess and physical address to virtual 
 *        address conversion. And invalidate the cache.
 *
 * @param IX_MBUF [in] *pMbuf - pointer to the mbuf data block.
 *
 * @return IX_MBUF* - the converted mbuf is returned.
 *
 * Reentrant    : no
 * ISR Callable : no
 *
 */
IX_MBUF*
ixCryptoUtilMbufFromNpeFormatConvert (IX_MBUF *pMbuf);


/**
 * @def IX_CRYPTO_ACC_LOCK
 *
 * @brief Locks out IRQs
 *
 * @return INT32 lockVal - lock key used by IX_CRYPTO_ACC_UNLOCK to unlock 
 *         IRQs later
 *
 */
#define IX_CRYPTO_ACC_LOCK(lockVal) lockVal = ixOsServIntLock();


/**
 * @def IX_CRYPTO_ACC_UNLOCK
 *
 * @brief Unlock IRQs
 *
 * @param INT32 [in] lockVal - lock key previously obtained with 
 *        IX_CRYPTO_ACC_LOCK
 *
 * @return None
 *
 */
#define IX_CRYPTO_ACC_UNLOCK(lockVal) ixOsServIntUnlock(lockVal);


/**
 * @def IX_CRYPTO_ACC_LOG
 *
 * @brief logs a formatted message
 *
 * @param int [in] level - trace level.
 * @param char [in] format - format string, similar to printf().
 * @param int [in] a - first argument to display.
 * @param int [in] b - second argument to display.
 * @param int [in] c - third argument to display.
 * @param int [in] d - fourth argument to display.
 * @param int [in] e - fifth argument to display.
 * @param int [in] f - sixth argument to display.
 *
 * @return none
 *
 */
#ifdef DEBUG
#define IX_CRYPTO_ACC_LOG(level, str, a, b, c, d, e, f)\
             (ixOsServLog (level, str, a, b, c, d, e, f))
#else
#define IX_CRYPTO_ACC_LOG(level, str, a, b, c, d, e, f) /* do nothing */
#endif /* def DEBUG */


#endif /* ndef IXCRYPTOACCUTILITIES_P_H */
