/***
 * @file IxFpathBufferPoolManager.c
 *
 * @author Intel Corporation
 * @date 15 April 2002
 *
 * @brief Fast Path Buffer Manager
 *
 * Responsible for controlling the allocation and initialization of
 * Npe descriptors, and the assigment of user provided mbufs to them.
 * Manages the  process by which the Npe FreeFP queue is populated
 *
 * Design Notes:
 *    
 *
 * -- Intel Copyright Notice --
 *
 * @par
 * INTEL CONFIDENTIAL
 *
 * @par
 * Copyright 2002 Intel Corporation All Rights Reserved.
 *
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.  The Material
 * contains trade secrets and proprietary and confidential information of
 * Intel or its suppliers and licensors.  The Material is protected by
 * worldwide copyright and trade secret laws and treaty provisions. No
 * part of the Material may be used, copied, reproduced, modified,
 * published, uploaded, posted, transmitted, distributed, or disclosed in
 * any way without Intel's prior express written permission.
 *
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel or otherwise.  Any license under
 * such intellectual property rights must be express and approved by
 * Intel in writing.
 *
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * -- End Intel Copyright Notice --
 */

/*
 * System defined include files
 */

#include <stdio.h>

/*
 * Put the user defined include files required.
 */

#include "IxAssert.h"
#include "IxTypes.h"
#include "IxOsServices.h"
#include "IxNpeA.h"
#include "IxQMgr.h"
#include "IxFpathAcc.h"
#include "IxFpathDefines_p.h"
#include "IxFpathDescriptorManager_p.h"
#include "IxFpathBufferPoolManager_p.h"

#define IX_FPATHACC_MAX_PDU_LEN 65535

/* ------------------------------------------------------
 * local variables go here
 */
static BOOL ixFpathBPMInitDone = FALSE;

static IxFpathAccBufPoolUnderRunCallback 
    ixFpathBPMClientUnderrunCb =  NULL;
static unsigned 
    ixFpathBPMClientUnderrunCbId = 0;

static int requiredMbufSize = 0;

/* ------------------------------------------------------
 * dummy client callback
 */
PRIVATE void
ixFpathBPMDummyUnderRunCallback(IxFpathAccClientId clientId)
{
    ixOsServLog( LOG_WARNING, 
                 "Fpath free Q underrun ( no client callback )",
                 0,0,0,0,0,0);
}

/* ------------------------------------------------------
 * default handler for QMgr call back
 */
PRIVATE void 
ixFpathBPMUnderRunCallback(
    IxQMgrQId qId,
    IxQMgrCallbackId cbId )
{
    (*ixFpathBPMClientUnderrunCb)(ixFpathBPMClientUnderrunCbId);    
}



/* ------------------------------------------------------
 * Initialise
 */
IX_STATUS 
ixFpathBPMInit( void )                                      
{

    if( ixFpathBPMInitDone )
    {
        return IX_FAIL;
    }

    /* configure the Qmgr Q */
    if( ixQMgrQConfig ("Fast Path FreeQ",
	               IX_NPE_A_QMQ_FPATH_FREE_Q,
	               IXFPATH_FREE_Q_SIZE,
	               IX_QMGR_Q_ENTRY_SIZE1) == IX_SUCCESS )
    {
        /* Init DM */
        if( ixFpathDMInit(IXFPATH_FREE_Q_SIZE) == IX_SUCCESS )
        {
            /* set up the dummy callback */
            ixFpathBPMClientUnderrunCb = ixFpathBPMDummyUnderRunCallback;

            /* set up the underrun call back */
            if( ixQMgrNotificationCallbackSet( 
                    IX_NPE_A_QMQ_FPATH_FREE_Q,
                    ixFpathBPMUnderRunCallback,
                    0 ) == IX_SUCCESS )   /* 0 client Id,  we are not interested in it */
            {
                ixFpathBPMInitDone = TRUE;   
                return IX_SUCCESS;
            }
        }
    }

    return IX_FAIL;
}


/* ------------------------------------------------------
 * Validate the mbuf/chain to ensure that it is of
 * the correct size mbuf(s), and that the number of
 * mbufs do not exceed available resources.
 */
PRIVATE BOOL 
ixFpathBPMMbufValidate( int requiredSize,
                        int maxChainLength,
                        IX_MBUF *mbufPtr )
{
    /* we can never receive a null mbuf due to service i/f
     * policing 
     */

    int chainCount = maxChainLength;
    IX_MBUF *next = mbufPtr;

    do 
    {
      if(( IX_MBUF_MLEN(next) != requiredSize ) ||
         ( chainCount <= 0 ))
      {
          return FALSE;
      }
      next = IX_MBUF_NEXT_BUFFER_IN_PKT_PTR(next);
      chainCount--;
    } while(next != NULL );
    return TRUE;
}

/* -----------------------------------------------------
 * Allocate Npe descriptor(s) for each mbuf in a chain
 * bind the mbuf and descriptor, format the descriptor,
 * and place in the QMgr fast path free queue
 */
PRIVATE BOOL 
ixFpathBPMMbufDescriptorFormatAndQueue( IX_MBUF *mbufPtr )
{
    IX_MBUF *next = mbufPtr;
    IX_MBUF *prev = mbufPtr;
    IxNpeA_AtmVcFp *descriptor;

    do 
    {
        if( ixFpathDMDescriptorGet( &descriptor ) != IX_SUCCESS )
        {
            return FALSE;
        }

        /* Format the descriptor */
        descriptor->npe.rxBitField.vcId = 0;
        descriptor->npe.atmCellHeader.gfc = 0;
        descriptor->npe.atmCellHeader.vpi = 0;
        descriptor->npe.atmCellHeader.vci = 0;
        descriptor->npe.atmCellHeader.pti = 0;
        descriptor->npe.atmCellHeader.clp = 0;

        descriptor->npe.fpMatchIndex = 0;
        descriptor->npe.fpIndex = 0;
        descriptor->npe.currMbufLen = 0; 

        /* Flush the data from the cache */
        IX_ACC_DATA_CACHE_FLUSH(IX_MBUF_MDATA(next), IX_MBUF_MLEN(next) );
        IX_MBUF_MDATA(next)  = (char *) IX_FPATHACC_TO_PHY_ADDR_TRANSLATE(IX_MBUF_MDATA(next));

        descriptor->npe.pFpMbufData = (unsigned char *)IX_MBUF_MDATA(next);
        descriptor->npe.pCurrMbuf = (IX_MBUF *)IX_FPATHACC_TO_PHY_ADDR_TRANSLATE(next);
        descriptor->npe.pCurrMbufData = (unsigned char *)IX_MBUF_MDATA(next);
        descriptor->npe.rxBitField.currMbufSize = IX_MBUF_MLEN(next);
        descriptor->npe.pNextMbuf = NULL;

        descriptor->npe.totalLen  = 0; 

        descriptor->npe.aal5CrcResidue = 0xffffffff;

        descriptor->pRootMbuf = next;
        descriptor->masterMbufLen = descriptor->npe.rxBitField.currMbufSize;

        /* step on and Break the chain */
        prev = next;
        next = IX_MBUF_NEXT_BUFFER_IN_PKT_PTR(next);
        IX_MBUF_NEXT_PKT_IN_CHAIN_PTR(prev) = NULL;
        IX_MBUF_NEXT_BUFFER_IN_PKT_PTR(prev) = NULL;
 
        /* Flush the data from the cache */
        /* desc */
        IX_ACC_DATA_CACHE_FLUSH(descriptor, sizeof(descriptor->npe));
        /* mbuf header */
        IX_ACC_DATA_CACHE_FLUSH(prev, sizeof(IX_MBUF));
        
        /* Atmd deals with endianess */
        descriptor = (IxNpeA_AtmVcFp *) IX_FPATHACC_TO_PHY_ADDR_TRANSLATE(descriptor);

        if ( ixQMgrQWrite( IX_NPE_A_QMQ_FPATH_FREE_Q,
                           (UINT32 *) &descriptor ) != IX_SUCCESS)
        {
            /* No point in cleaning up */
            return FALSE;
        }

    } while(next != NULL );
    return TRUE;
}



/* ------------------------------------------------------
 * provide MBUfs to the fast path free queue
 */
IX_STATUS 
ixFpathBPMBufferSupply( IX_MBUF *mbuf, 
                       int *poolFreeEntriesRemaining ) 
{
    int maxSize;
    int populatedEntryCount;
    IX_STATUS returnStatus = IX_FAIL;

    IX_ENSURE( ixFpathBPMInitDone, "ixFpathBPM used before initialised");

    if ( mbuf != NULL )
    {
        if  (IX_MBUF_MLEN(mbuf) > 0) 
        {
            /* check required descriptor resources are available */
            ixFpathDMInfoGet( &maxSize, &populatedEntryCount );

            IX_ENSURE((populatedEntryCount <= maxSize), " Pool seems to be over populated ");

            if ( populatedEntryCount >= maxSize ) 
            {
                returnStatus = IX_FPATH_POOL_FULL;
            }
            else
            {
                /* store the mbuf size */
                if ( requiredMbufSize == 0 ) 
                {
                    requiredMbufSize = IX_MBUF_MLEN(mbuf);
                }
                
                /* validate the mbuf/chain */
                if ( ixFpathBPMMbufValidate( requiredMbufSize,
                                        maxSize - populatedEntryCount,
                                        mbuf ))
                {
                    if ( ixFpathBPMMbufDescriptorFormatAndQueue( mbuf ) )
                    {
                        returnStatus = IX_SUCCESS;
                    }
                }
            }
        }
    }
    else
    {
        returnStatus = IX_SUCCESS;
    }

    /* return the number of free entries remaining */    
    ixFpathDMInfoGet( &maxSize, &populatedEntryCount );
    *poolFreeEntriesRemaining = (maxSize - populatedEntryCount);
    
    return returnStatus;
}


/* ------------------------------------------------------
 * return the dimensions of the buffer pool
 */
IX_STATUS 
ixFpathBPMInfoGet( int *maxEntries, 
                   int *populatedEntryCnt )
{
    IX_ENSURE( ixFpathBPMInitDone, "ixFpathBPM used before initialised");

    /* check required descriptor resources are available */
    ixFpathDMInfoGet( maxEntries, populatedEntryCnt );
    return IX_SUCCESS;
}


/* -------------------------------------------------------
 * Set the client buffer underrun callback
 */
IX_STATUS 
ixFpathBPMUnderRunCallbackSet( 
    IxFpathAccBufPoolUnderRunCallback callback,
    IxFpathAccClientId clientId)
{
    IX_ENSURE( ixFpathBPMInitDone, "ixFpathBPM used before initialised");

    if ( ixQMgrNotificationDisable(IX_NPE_A_QMQ_FPATH_FREE_Q) 
             == IX_SUCCESS)
    {
        ixFpathBPMClientUnderrunCb = callback;
        ixFpathBPMClientUnderrunCbId = clientId;

        if ( callback == NULL ) 
        {
            /* user is disabling callbacks */
            ixFpathBPMClientUnderrunCb = ixFpathBPMDummyUnderRunCallback;
            ixFpathBPMClientUnderrunCbId = 0;
            return IX_SUCCESS;   
        }
        
        if (ixQMgrNotificationEnable( 
                IX_NPE_A_QMQ_FPATH_FREE_Q,
                IX_QMGR_Q_SOURCE_ID_E) == IX_SUCCESS)
        {
            return IX_SUCCESS;
        }
    }
    return IX_FAIL;
}


/* ----------------------------------------------------
 * Show
 */
void
ixFpathBPMShow( void )
{

    if(!ixFpathBPMInitDone)
    {
        printf("\nNot Initialized");    
    }
    else
    {
        printf("\nixFpathBPMClientUnderrunCbId %i",
               ixFpathBPMClientUnderrunCbId);
        printf("\nminimumMbufSize %u",requiredMbufSize);
        ixFpathDMShow();
    }
}

