/***
 * @file IxFpathDescriptorManager.c
 *
 * @author Intel Corporation
 * @date 15 April 2002
 *
 * @brief Fast Path Descriptor Manager
 *
 *      Implements a pool of Npe descriptors.
 *
 * Design Notes:
 *      
 *
 * -- Intel Copyright Notice --
 *
 * @par
 * INTEL CONFIDENTIAL
 *
 * @par
 * Copyright 2002 Intel Corporation All Rights Reserved.
 *
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.  The Material
 * contains trade secrets and proprietary and confidential information of
 * Intel or its suppliers and licensors.  The Material is protected by
 * worldwide copyright and trade secret laws and treaty provisions. No
 * part of the Material may be used, copied, reproduced, modified,
 * published, uploaded, posted, transmitted, distributed, or disclosed in
 * any way without Intel's prior express written permission.
 *
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel or otherwise.  Any license under
 * such intellectual property rights must be express and approved by
 * Intel in writing.
 *
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * -- End Intel Copyright Notice --
 */

/*
 * Sytem defined include files
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>


/*
 * Put the user defined include files required.
 */
#include "IxOsCacheMMU.h"
#include "IxAssert.h"
#include "IxFpathDescriptorManager_p.h"

#define IX_FPATHACC_DESC_ADDR_LIMIT 0xf0000000

static struct 
{
    IxNpeA_AtmVcFp **descriptorPtr;
    unsigned int size;
    unsigned int allocatedCnt;    
} ixFpathDMDescriptorPool = { NULL, 0 , 0 };

/*
 * Local variables
 */
static BOOL ixFpathDMInitDone = FALSE;
static unsigned int rawMemorySize = 0;
static unsigned int descriptorSize = 0;
static unsigned int cacheLinesPerDescriptor = 0; 
static unsigned char *rawMemoryPtr = NULL;

/* N.B. Npes in CSR1.0 reserve the use of the bottom 2 bits of the decriptor 
        address. Hence decripotrs must aligned at least on a 4 byte boundary. */
#if (IX_XSCALE_CACHE_LINE_SIZE < 4)
#error cache line size too small for use with Npe descriptors
#endif

/* ------------------------------------------------------
 * Initialise
 */
IX_STATUS
ixFpathDMInit( unsigned int numberOfDescriptors )
{
    unsigned int count; 

    if( ixFpathDMInitDone )
    {
        return IX_FAIL;
    }

    /* adjust the required memory size to allow enough memory
     * to enusre that every descriptor starts on a cache
     * line boundary
     */

    /* calcuclate the number of cache lines needed for
     * a descriptor, and ajust the descriptor size
     */
    descriptorSize = sizeof(IxNpeA_AtmVcFp);

    cacheLinesPerDescriptor 
       = ( descriptorSize / IX_XSCALE_CACHE_LINE_SIZE );

    if((descriptorSize % IX_XSCALE_CACHE_LINE_SIZE) != 0 )
    {
        cacheLinesPerDescriptor++;
        descriptorSize = ( cacheLinesPerDescriptor 
                           * IX_XSCALE_CACHE_LINE_SIZE );
    } 

    /* caclulate the actual raw memory required */
    /* should be a whole number of cache lines */
    rawMemorySize = ( numberOfDescriptors * descriptorSize ); 

    /* Get the memory for the desciptors themselves*/
    /* ASSUMPTION: IX_ACC_DRV_DMA_MALLOC allocates memory on cache line
                   boundary so there is no need to adjust the pointer. */
    rawMemoryPtr 
      = (unsigned char *) IX_ACC_DRV_DMA_MALLOC( rawMemorySize );

    /* The upper 4 bits of the descriptor address are reserved for use by NPES */
    IX_ASSERT(((unsigned int)(rawMemoryPtr+rawMemorySize) < IX_FPATHACC_DESC_ADDR_LIMIT));

    /* get memory for our pool of pointers to
     * descriptors
     */
    ixFpathDMDescriptorPool.descriptorPtr 
      = (IxNpeA_AtmVcFp **) IX_ACC_DRV_DMA_MALLOC(
                              numberOfDescriptors * sizeof(IxNpeA_AtmVcFp *));


    if ((ixFpathDMDescriptorPool.descriptorPtr != NULL) &&
        ( rawMemoryPtr != NULL )) 
    {
        /* Zero all of this memory */
        memset( ixFpathDMDescriptorPool.descriptorPtr,
                0,
                numberOfDescriptors * sizeof(void *));

        memset( rawMemoryPtr,
                0,
                rawMemorySize );

        /* assign the memory we have allocated to our pool */
        for( count = 0; count < numberOfDescriptors; count++)
        {
            ixFpathDMDescriptorPool.descriptorPtr[count]
              = (IxNpeA_AtmVcFp *) rawMemoryPtr;
            rawMemoryPtr += descriptorSize; 
        }

        ixFpathDMDescriptorPool.size = numberOfDescriptors;
        ixFpathDMDescriptorPool.allocatedCnt = 0;
        ixFpathDMInitDone = TRUE;
        return IX_SUCCESS;
    }

    /* Not cleaning up on fail */
    return IX_FAIL;
}


/* ------------------------------------------------------
 * return the dimensions of the descriptor pool
 */
IX_STATUS
ixFpathDMInfoGet( int *maxNoEntries, int *allocatedCount )
{
    IX_ENSURE( ixFpathDMInitDone, "ixFpathDM used before initialised");

    if( (maxNoEntries != NULL) &&
        (allocatedCount != NULL))
    {
       *maxNoEntries = ixFpathDMDescriptorPool.size;
       *allocatedCount = ixFpathDMDescriptorPool.allocatedCnt;
       return IX_SUCCESS;
    }
    return IX_FAIL;
}


/* ------------------------------------------------------
 * get and return a descriptor
 */
IX_STATUS 
ixFpathDMDescriptorGet( IxNpeA_AtmVcFp **descriptor )
{
    IX_ENSURE( ixFpathDMInitDone, "ixFpathDM used before initialised");

    if ( ixFpathDMDescriptorPool.allocatedCnt >= 
         ixFpathDMDescriptorPool.size ) 
    {
        return IX_FAIL;
    }

    IX_ASSERT(ixFpathDMDescriptorPool.descriptorPtr[
                  ixFpathDMDescriptorPool.allocatedCnt] != NULL);

    *descriptor = ixFpathDMDescriptorPool.descriptorPtr[
                      ixFpathDMDescriptorPool.allocatedCnt++];
    return IX_SUCCESS;
}


/* ----------------------------------------------------
 * Show
 */
void
ixFpathDMShow( void )
{
    printf("\n---------------------------------------"
           "\nixFpathDMShow()"
           "\n---------------------------------------");

    if( !ixFpathDMInitDone )
    {
        printf("\nNot Initialized");    
    }
    else
    {
        printf("\nixFpathDMDescriptorPool.size %u"
               "\nixFpathDMDescriptorPool.allocatedCnt %u"
               "\nrawMemorySize %u"     
               "\ndescriptorSize %u"    
               "\ncacheLinesPerDescriptor %u",
               ixFpathDMDescriptorPool.size,
               ixFpathDMDescriptorPool.allocatedCnt,
               rawMemorySize,
               descriptorSize,
               cacheLinesPerDescriptor );
    }
}

