/***
 * @file ixFpathNpeInterface.c
 *
 * @author Intel Corporation
 * @date 15 April 2002
 *
 * @brief Fast Path Npe Interface
 *
 * Provides a fast path specific interface to the Npe's. 
 * Hides ixNpeMh API, performs npeMh message formatting,
 * and transparently handles the processing of  both solicited 
 * and unsolicited acknowledgements from the Npe(s) .
 *
 * Design Notes:
 *    
 *
 * -- Intel Copyright Notice --
 *
 * @par
 * INTEL CONFIDENTIAL
 *
 * @par
 * Copyright 2002 Intel Corporation All Rights Reserved.
 *
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.  The Material
 * contains trade secrets and proprietary and confidential information of
 * Intel or its suppliers and licensors.  The Material is protected by
 * worldwide copyright and trade secret laws and treaty provisions. No
 * part of the Material may be used, copied, reproduced, modified,
 * published, uploaded, posted, transmitted, distributed, or disclosed in
 * any way without Intel's prior express written permission.
 *
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel or otherwise.  Any license under
 * such intellectual property rights must be express and approved by
 * Intel in writing.
 *
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * -- End Intel Copyright Notice --
 */

/*
 * Sytem defined include files
 */

#include <stdio.h>

/*
 * Put the user defined include files required.
 */
#include "IxAssert.h"
#include "IxOsServices.h"
#include "IxNpeMh.h"
#include "IxEthNpe.h"
#include "IxNpeA.h"
#include "IxFpathDefines_p.h"
#include "IxFpathNpeInterface_p.h"

/*
 * local variables
 */
static BOOL ixFpathNpeIfInitDone = FALSE;

static volatile BOOL ixFpathNpeIfMhResponseFlag = FALSE;

static IxNpeMhMessage ixFpathNpeIfMhResponseMsg;

const static unsigned char *ixFpathNpeIfRespMsgBytePtr 
    = (unsigned char *)&ixFpathNpeIfMhResponseMsg;

static IxFpathNpeIfWanMacMsgCallback ixFpathNpeIfWanMacCallback = NULL;
 
unsigned int ixFpathNpeIfRetryCount = 100;

/* 
 * defines 
 */
#define IX_FPATHACC_ONE_MILLISECOND 1
#define IX_FPATH_NPEMH_RESPONSE_TIMEOUT 10 
#define IX_FPATH_NPEMH_TIMEOUT_RESOLUTION 20*IX_FPATHACC_ONE_MILLISECOND
#define IX_FPATH_WANMAC_ADDR_SIZE 6

/* 
 * Byte offset into an array of bytes. PSM messages are Big Endian. Hence
 * extracting bytes from messages requires the correct offset based on the 
 * xscale endianess.
 */
#ifdef LITTLE_ENDIAN

/* LITTLE-ENDIAN */
#define NPEIF_NETWORK_BYTE_0 0x03
#define NPEIF_NETWORK_BYTE_1 0x02
#define NPEIF_NETWORK_BYTE_2 0x01
#define NPEIF_NETWORK_BYTE_3 0x00
#define NPEIF_NETWORK_BYTE_4 0x07
#define NPEIF_NETWORK_BYTE_5 0x06
#define NPEIF_NETWORK_BYTE_6 0x05
#define NPEIF_NETWORK_BYTE_7 0x04


#else 

/* common to all messages */
#define NPEIF_NETWORK_BYTE_0 0x00
#define NPEIF_NETWORK_BYTE_1 0x01
#define NPEIF_NETWORK_BYTE_2 0x02
#define NPEIF_NETWORK_BYTE_3 0x03
#define NPEIF_NETWORK_BYTE_4 0x04
#define NPEIF_NETWORK_BYTE_5 0x05
#define NPEIF_NETWORK_BYTE_6 0x06
#define NPEIF_NETWORK_BYTE_7 0x07

#endif

/* ------------------------------------------------------
 * Initialise the Mh response semaphore
 */
PRIVATE void  
ixFpathNpeIfMhResponseSemInit( void )
{
    ixFpathNpeIfMhResponseFlag = FALSE;
}


/* ------------------------------------------------------
 * pend on the mh response semaphore
 */
/* ToDo SCR - 862. Remove this workaround */
unsigned int pendCnt = 0;
unsigned int pendFailedCnt = 0;
unsigned int pendIterationCnt = 0;
unsigned int maxPendIterationCnt = 0;

PRIVATE BOOL
ixFpathNpeIfMhResponseSemPend( void )
{
    BOOL result;
    unsigned int _maxIterations = 0;

    int timeout = IX_FPATH_NPEMH_RESPONSE_TIMEOUT;


    /* wait for the response flag to be set, or we timeout */
    while( !ixFpathNpeIfMhResponseFlag && timeout-- )
    {
        /* we only sleep if the flag is not set */
        ixOsServTaskSleep (IX_FPATH_NPEMH_TIMEOUT_RESOLUTION);
        _maxIterations += 1;
    }

    /* Update Statistics */
    if ( ixFpathNpeIfMhResponseFlag != TRUE ) 
    {
        pendFailedCnt += 1;
    }
    if ( _maxIterations > maxPendIterationCnt ) 
    {
        maxPendIterationCnt = _maxIterations;
    }
    if ( _maxIterations > 0 ) 
    {
        pendIterationCnt += 1;
    }
    pendCnt += 1;

    result = ixFpathNpeIfMhResponseFlag;
    

    ixFpathNpeIfMhResponseFlag  = FALSE;
    return result;
}


/* ------------------------------------------------------
 * post to the mh reponse semaphore
 */
PRIVATE void
ixFpathNpeIfMhResponseSemPost( void )
{
    ixFpathNpeIfMhResponseFlag = TRUE; 
    return;
}


PRIVATE void 
ixFpathNpeIfMhResponseCallback (IxNpeMhNpeId npeId, IxNpeMhMessage msg)
{
    /* copy the message so that it is in scope for the request thread 
     * after this function returns
     */
    ixFpathNpeIfMhResponseMsg.data[0] = msg.data[0];
    ixFpathNpeIfMhResponseMsg.data[1] = msg.data[1];

    /* indicate to the request thread that the message has
     * been received
     */
    ixFpathNpeIfMhResponseSemPost();
    return;     
}


PRIVATE void
ixFpathNpeIfMhWanMacCallback(IxNpeMhNpeId npeId, IxNpeMhMessage msg)
{
    IxFpathIndex fpIndex;
    int i=0;
    unsigned char macAddr[ IX_FPATH_WANMAC_ADDR_SIZE ];
    unsigned char *msgPtr = (unsigned char *) &msg;

    /* get the fpath index from the message */
    fpIndex = msgPtr[ NPEIF_NETWORK_BYTE_1 ];
    
    /* get the macAddr from the message  */
    macAddr[i]   = msgPtr[ NPEIF_NETWORK_BYTE_2 ]; /* Mac 5 */
    macAddr[++i] = msgPtr[ NPEIF_NETWORK_BYTE_3 ];

    macAddr[++i] = msgPtr[ NPEIF_NETWORK_BYTE_4 ];
    macAddr[++i] = msgPtr[ NPEIF_NETWORK_BYTE_5 ];
    macAddr[++i] = msgPtr[ NPEIF_NETWORK_BYTE_6 ];
    macAddr[++i] = msgPtr[ NPEIF_NETWORK_BYTE_7 ]; /* Mac 0 */

    /* pass the message up to the NpeIf client */
    (*ixFpathNpeIfWanMacCallback)( fpIndex, macAddr );
}


PRIVATE void
ixFpathNpeIfDummyClientWanMacCb(IxFpathIndex fpIndex, unsigned char *macAddr)
{
    /* Should never happen */
    ixOsServLog( LOG_ERROR, 
                 "ixFpathAcc NpeIf Mac Addr callback not set",
                 0,0,0,0,0,0);
}

/* ------------------------------------------------------
 * Initialise
 */
IX_STATUS
ixFpathNpeIfInit( void )
{

    if( ixFpathNpeIfInitDone )
    {
        return IX_FAIL;
    }

    /* initialize the response semaphore */
    ixFpathNpeIfMhResponseSemInit();

    /* 
     * Set up function pointer to dummy value so invocation of pointer 
     */
    ixFpathNpeIfWanMacCallback = ixFpathNpeIfDummyClientWanMacCb;

    /* register NpeMh autonomous message callbacks */

    /* shutdwon ACk */
    /* Note this uses the generic response callback because
     * this can be handled as a solicited response
     */
    if( ixNpeMhUnsolicitedCallbackRegister (
                   IX_NPEMH_NPEID_NPEB,
                   IX_ETHNPE_P2X_FP_SHUTDOWNVCACK,
                   ixFpathNpeIfMhResponseCallback ) != IX_SUCCESS )
     {
        return IX_FAIL;
     }

    /* Wan Mac */
    if(  ixNpeMhUnsolicitedCallbackRegister (
                   IX_NPEMH_NPEID_NPEA,
                   IX_NPE_A_MSSG_ATM_FP_WAN_MAC,
                   ixFpathNpeIfMhWanMacCallback ) != IX_SUCCESS )
    {
        return IX_FAIL;
    }

    ixFpathNpeIfInitDone = TRUE;

    return IX_SUCCESS;
}

/* ------------------------------------------------------
 * Register the Wan Mac callback for the NpeIf client
 */
IX_STATUS
ixFpathNpeIfWanMacCallbackRegister(IxFpathNpeIfWanMacMsgCallback callback)
{
    IX_ENSURE( ixFpathNpeIfInitDone, "\nixFpathNpeIf used before initialised\n");

    if(callback != NULL )
    {
        ixFpathNpeIfWanMacCallback = callback;
        return IX_SUCCESS;
    }
    return IX_FAIL;
}

/* ToDo SCR - 862. */
unsigned int msgSendMaxRetryCnt = 0xffffffff;
unsigned int msgSendRetryDelay  = 100;
unsigned int msgSendRetryCnt    = 30;
/* --------------------------------------------------------
 * Send with retry until success or retry count has
 * reach a maximum value
 */
PRIVATE IX_STATUS 
ixFpathNpeIfMsgSend (
    IxNpeMhNpeId npeId,
    IxNpeMhMessage message)
{
    unsigned int retryCount = ixFpathNpeIfRetryCount;

    while ( retryCount !=0 )
    { 
        if( ixNpeMhMessageSend (npeId,
                                message,
                                msgSendRetryCnt ) == IX_SUCCESS )
        {
            break;
        }

        retryCount--;
    }

    if( retryCount < msgSendMaxRetryCnt ) 
    {
        msgSendMaxRetryCnt = retryCount;
    }

    if( retryCount == 0 ) 
    {
        return IX_FAIL;
    }
    return IX_SUCCESS;
}


/* ToDo SCR - 862. */
unsigned int msgSendWithRespMaxRetryCnt = 0xffffffff;
unsigned int msgSendWithRespRetryDelay  = 100;

/* --------------------------------------------------------
 * Send with retry until success or retry count has
 * reach a maximum value
 */
PRIVATE IX_STATUS 
ixFpathIfMsgWithResponseSend (
    IxNpeMhNpeId npeId,
    IxNpeMhMessage message,
    IxNpeMhMessageId solicitedMessageId,
    IxNpeMhCallback solicitedCallback)
{
    unsigned int retryCount = ixFpathNpeIfRetryCount;

    while ( retryCount != 0 )
    {
        if (ixNpeMhMessageWithResponseSend (
                npeId,
                message,
                solicitedMessageId,
                solicitedCallback,
                IX_NPEMH_SEND_RETRIES_DEFAULT) == IX_SUCCESS)
        {
            break; 
        }
        retryCount--;
        ixOsServSleep(msgSendWithRespRetryDelay);
    }

    if( retryCount < msgSendWithRespMaxRetryCnt ) 
    {
        msgSendWithRespMaxRetryCnt = retryCount;
    }

    if( retryCount == 0 ) 
    {
        return IX_FAIL;
    }
    return IX_SUCCESS;
}




/* ------------------------------------------------------
 * Disable fast path processing for the specified fast path
 * index
 */
IX_STATUS
ixFpathNpeIfDisableSend( IxNpeMhNpeId npeId, 
                         IxFpathIndex fpathIndex, /* Needed for Ack message for Npe */
                         IxAtmNpeRxVcId npeVcId )
{
    IxNpeMhMessage reqMsg;
    IxFpathIndex rxFpathIndex;
    unsigned char *reqMsgPtr = (unsigned char *) &reqMsg;

    IX_ENSURE( ixFpathNpeIfInitDone, "\nixFpathNpeIf used before initialised\n");

    /* first initialise the message to 0 */
    reqMsg.data[0] = 0;
    reqMsg.data[1] = 0;

    /* Format send requestMsg */
    reqMsgPtr[NPEIF_NETWORK_BYTE_0] = IX_NPE_A_MSSG_ATM_FP_DISABLE;
    reqMsgPtr[NPEIF_NETWORK_BYTE_3] = (UINT8) npeVcId;

    /* Send requestMsg */
    if ( ixFpathNpeIfMsgSend( npeId, reqMsg ) == IX_SUCCESS ) 
    {
        /* Pend on the response */
        if (ixFpathNpeIfMhResponseSemPend())
        {
            /* Check the fpIndex matches the msg we just sent */
            /* This assumes - max fpindex = 255 hence byte 3 is ignored*/
            rxFpathIndex = ixFpathNpeIfRespMsgBytePtr[NPEIF_NETWORK_BYTE_3];   

            IX_ENSURE( ( ixFpathNpeIfRespMsgBytePtr[NPEIF_NETWORK_BYTE_2] == 0) , 
                        "\nInvalid fpIndex from Eth Npe");

            if ( rxFpathIndex == fpathIndex ) 
            {
                /* Clear the message on success. */
                ixFpathNpeIfMhResponseMsg.data[0] = 0;
                ixFpathNpeIfMhResponseMsg.data[1] = 0;
                return IX_SUCCESS;
            }
        }
    }

    ixOsServLog( LOG_ERROR, 
                 "FpDisable failed on VcId %i, FpIndex, %i",
                 npeVcId,
                 fpathIndex,
                 0,0,0,0);

    return IX_FAIL;
}

/* ------------------------------------------------------
 * Enable fast path processing for the specified fast path
 * index
 */
IX_STATUS
ixFpathNpeIfEnableSend( IxNpeMhNpeId npeId, 
                        IxFpathIndex fpathIndex, 
                        IxAtmNpeRxVcId npeVcid )
{
    IxNpeMhMessage reqMsg;
    unsigned char *reqMsgPtr = (unsigned char *) &reqMsg;

    IX_ENSURE( ixFpathNpeIfInitDone, "\nixFpathNpeIf used before initialised\n");

    /* first initialise the message to 0 */
    reqMsg.data[0] = 0;
    reqMsg.data[1] = 0;

    /* Format send requestMsg */
    reqMsgPtr[NPEIF_NETWORK_BYTE_0] = IX_NPE_A_MSSG_ATM_FP_ENABLE;
    reqMsgPtr[NPEIF_NETWORK_BYTE_2] = (UINT8) fpathIndex;
    reqMsgPtr[NPEIF_NETWORK_BYTE_3] = (UINT8) npeVcid;

    /* Send requestMsg */
    if ( ixFpathNpeIfMsgSend( npeId, reqMsg ) == IX_SUCCESS ) 
    {
        return IX_SUCCESS;
    }
    
    ixOsServLog( LOG_ERROR, 
                 "FpEnable failed on VcId %i, FpIndex, %i",
                 npeVcid,
                 fpathIndex,
                 0,0,0,0);

    return IX_FAIL;
}

/* ------------------------------------------------------
 * Download a modifer template to the specified Npe
 */
IX_STATUS
ixFpathNpeIfModifierDownload(IxNpeMhNpeId npeId, 
                             IxFpathIndex fpathIndex, 
                             unsigned char *template, 
                             int templateLength )
{
    IxNpeMhMessage reqMsg;
    unsigned char *reqMsgPtr = (unsigned char *) &reqMsg;

    IX_ENSURE( ixFpathNpeIfInitDone, "\nixFpathNpeIf used before initialised\n");

    /* first initialise the message to 0 */
    reqMsg.data[0] = 0;
    reqMsg.data[1] = 0;

    /* Format send requestMsg */
    reqMsgPtr[NPEIF_NETWORK_BYTE_0] = IX_ETHNPE_X2P_FP_WRITETEMPMDF;
    reqMsgPtr[NPEIF_NETWORK_BYTE_3] = (UINT8) fpathIndex;

    reqMsg.data[1] = ( UINT32 ) IX_FPATHACC_TO_PHY_ADDR_TRANSLATE(template);

    /* Send requestMsg */
    if ( ixFpathIfMsgWithResponseSend( 
             npeId, 
             reqMsg,
             IX_ETHNPE_P2X_FP_WRITETEMPMDFACK,
             ixFpathNpeIfMhResponseCallback ) == IX_SUCCESS ) 
    {
        /* Pend on the response */
        if (ixFpathNpeIfMhResponseSemPend())
        {
            /* Check the response matches the msg we just sent */
            /* This assumes - max fpindex = 255 hence byte 3 is ignored*/
            if (( reqMsgPtr[NPEIF_NETWORK_BYTE_3] == 
                  ixFpathNpeIfRespMsgBytePtr[NPEIF_NETWORK_BYTE_3] ) &&
                ( reqMsg.data[1] ==  ixFpathNpeIfMhResponseMsg.data[1] )) 
            {
                /* Clear the message on success. */
                ixFpathNpeIfMhResponseMsg.data[0] = 0;
                ixFpathNpeIfMhResponseMsg.data[1] = 0;
                return IX_SUCCESS;
            }
        }
    }
    
    ixOsServLog( LOG_ERROR, 
                 "Modifier download failed on FpIndex, %i",
                 fpathIndex,
                 0,0,0,0,0);

    return IX_FAIL;
}

/* ------------------------------------------------------
 * Download a classifier template to the specifier Npe
 */
IX_STATUS
ixFpathNpeIfClassifierDownload( IxNpeMhNpeId npeId, 
                                IxFpathIndex fpathIndex, 
                                unsigned char *template, 
                                int templateLength )
{

    IX_STATUS returnStatus = IX_SUCCESS;
    IxNpeMhMessage reqMsg;
    unsigned char *reqMsgPtr = (unsigned char *) &reqMsg;
    UINT32 offset;

    IX_ENSURE( ixFpathNpeIfInitDone, "\nixFpathNpeIf used before initialised\n");

    for(offset = 0;  
        (offset < (UINT32) templateLength) &&  (returnStatus == IX_SUCCESS); 
        offset += sizeof(UINT32))
    {
        /* first initialise the message to 0 */
        reqMsg.data[0] = 0;
        reqMsg.data[1] = 0;

        /* Format send requestMsg */
        reqMsgPtr[NPEIF_NETWORK_BYTE_0] = IX_NPE_A_MSSG_ATM_FP_TEMPLATE_WRITE;
        reqMsgPtr[NPEIF_NETWORK_BYTE_1] = (UINT8) fpathIndex;
        reqMsgPtr[NPEIF_NETWORK_BYTE_3] = (UINT8) offset;
        reqMsg.data[1] = *(( UINT32 *) (template+offset));

        /* Send requestMsg */
        returnStatus |= ixFpathNpeIfMsgSend( npeId, reqMsg );
    }
    if( returnStatus != IX_SUCCESS )
    {
        ixOsServLog( LOG_ERROR, 
                 "Classifier download failed on npeId %i, FpIndex, %i\n"
                 "templateLength %i",
                 npeId,
                 fpathIndex,
                 templateLength,
                 0,0,0);
        returnStatus = IX_FAIL;
    }

    return returnStatus;
 
}


/* ----------------------------------------------
 * Set the ethernet type
 */
IX_STATUS
ixFpathNpeIfEtherTypeSend( IxNpeMhNpeId npeId,
                           UINT16 etherType )
{
    IxNpeMhMessage reqMsg;
    unsigned char *reqMsgPtr = (unsigned char *) &reqMsg;
    unsigned char *etherTypePtr = ( unsigned char *) &etherType;

    IX_ENSURE( ixFpathNpeIfInitDone, "\nixFpathNpeIf used before initialised\n");

    /* first initialise the message to 0 */
    reqMsg.data[0] = 0;
    reqMsg.data[1] = 0;

    /* Format send requestMsg */
    reqMsgPtr[NPEIF_NETWORK_BYTE_0] = IX_ETHNPE_X2P_FP_SETETHERNETTYPE;
    reqMsgPtr[NPEIF_NETWORK_BYTE_2] = etherTypePtr[0];
    reqMsgPtr[NPEIF_NETWORK_BYTE_3] = etherTypePtr[1];

    /* Send requestMsg */
    if ( ixFpathNpeIfMsgSend( npeId, reqMsg ) == IX_SUCCESS ) 
    {
        return IX_SUCCESS;
    }
    
    ixOsServLog( LOG_ERROR, 
                 "Set Ether Type failed on Npe %i",
                 npeId,
                 0,0,0,0,0);

    return IX_FAIL;
}

/* ------------------------------------------------------
 * Disable fast path processing for the specified fast path
 * index
 */
IX_STATUS
ixFpathNpeIfStatisticsGet( IxNpeMhNpeId npeId, 
                           IxFpathIndex fpathIndex, 
                           int *packetCnt )
{
    IxNpeMhMessage reqMsg;
    IxFpathIndex rxFpathIndex;
    unsigned char *reqMsgPtr = (unsigned char *) &reqMsg;

    IX_ENSURE( ixFpathNpeIfInitDone, "\nixFpathNpeIf used before initialised\n");

    /* first initialise the message to 0 */
    reqMsg.data[0] = 0;
    reqMsg.data[1] = 0;

    /* Format send requestMsg */
    reqMsgPtr[NPEIF_NETWORK_BYTE_0] = IX_NPE_A_MSSG_ATM_FP_STATISTICS_READ;
    reqMsgPtr[NPEIF_NETWORK_BYTE_3] = (UINT8) fpathIndex;

    /* Send requestMsg */
    if ( ixFpathIfMsgWithResponseSend( 
             npeId, 
             reqMsg,
             IX_NPE_A_MSSG_ATM_FP_STATISTICS_READ,
             ixFpathNpeIfMhResponseCallback ) == IX_SUCCESS ) 
    {
        /* Pend on the response */
        if (ixFpathNpeIfMhResponseSemPend())
        {
            /* Check the fpIndex matches the msg we just sent */
            rxFpathIndex = ixFpathNpeIfRespMsgBytePtr[NPEIF_NETWORK_BYTE_3];   

            if ( rxFpathIndex == fpathIndex ) 
            {
                *packetCnt = ixFpathNpeIfMhResponseMsg.data[1];
                /* Clear the message on success. */
                ixFpathNpeIfMhResponseMsg.data[0] = 0;
                ixFpathNpeIfMhResponseMsg.data[1] = 0;
                return IX_SUCCESS;
            }
        }
    }

    ixOsServLog( LOG_ERROR, 
                 "Fp Stats retrieveal failed on FpIndex, %i",
                 fpathIndex,
                 0,0,0,0,0);

    return IX_FAIL;
}

/* ------------------------------------------------------
 * Clear the fast path counters for a particluar fpIndex.
 *  */
IX_STATUS
ixFpathNpeIfStatisticsClear( IxNpeMhNpeId npeId, 
                             IxFpathIndex fpathIndex)
{
    IxNpeMhMessage reqMsg;
    unsigned char *reqMsgPtr = (unsigned char *) &reqMsg;

    IX_ENSURE( ixFpathNpeIfInitDone, "\nixFpathNpeIf used before initialised\n");

    /* first initialise the message to 0 */
    reqMsg.data[0] = 0;
    reqMsg.data[1] = 0;

    /* Format send requestMsg */
    reqMsgPtr[NPEIF_NETWORK_BYTE_0] = IX_NPE_A_MSSG_ATM_FP_STATISTICS_CLEAR;
    reqMsgPtr[NPEIF_NETWORK_BYTE_3] = (UINT8) fpathIndex;

    /* Send requestMsg */
    if ( ixFpathNpeIfMsgSend(npeId, reqMsg) == IX_SUCCESS ) 
    {    
        return IX_SUCCESS;
    }

    ixOsServLog( LOG_ERROR, 
                 "Fp Stats clear failed for FpIndex, %i",
                 fpathIndex,
                 0,0,0,0,0);

    return IX_FAIL;
}

/* ----------------------------------------------------
 * Shown
 */
void
ixFpathNpeIfShow( void )
{
    if(!ixFpathNpeIfInitDone)
    {
        printf("\nNot Initialized");    
    }
    else
    {
        printf("\nixFpathNpeIfMhResponseFlag %i"
               "\nixFpathNpeIfMhResponseMsg.data[0] %8.8x"
               "\nixFpathNpeIfMhResponseMsg.data[1] %8.8x",
               ixFpathNpeIfMhResponseFlag,
               ixFpathNpeIfMhResponseMsg.data[0],
               ixFpathNpeIfMhResponseMsg.data[1] );
    }
}

