/***
 * @file ixFpathServiceInterface.c
 *
 * @author Intel Corporation
 * @date 15 April 2002
 *
 * @brief Fast Path Service Interface
 *
 * Provides the fpath access layer  client API,  
 * Performs parameter validation.
 * Serializes accesses to fpath access layer by means of a lock.
 *
 * Design Notes:
 *     Return Types are guaranteed to be mapped to the interface return types
 * at any point where an external component is used. This is true for all
 * FpathAcc sub components. i.e. no subcomponent call will return a type
 * which isn't defined in FpathAcc.
 *
 * -- Intel Copyright Notice --
 *
 * @par
 * INTEL CONFIDENTIAL
 *
 * @par
 * Copyright 2002 Intel Corporation All Rights Reserved.
 *
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.  The Material
 * contains trade secrets and proprietary and confidential information of
 * Intel or its suppliers and licensors.  The Material is protected by
 * worldwide copyright and trade secret laws and treaty provisions. No
 * part of the Material may be used, copied, reproduced, modified,
 * published, uploaded, posted, transmitted, distributed, or disclosed in
 * any way without Intel's prior express written permission.
 *
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel or otherwise.  Any license under
 * such intellectual property rights must be express and approved by
 * Intel in writing.
 *
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * -- End Intel Copyright Notice --
 */

/*
 * Sytem defined include files
 */
#include <stdio.h>


/*
 * Put the user defined include files required.
 */
#include "IxTypes.h"
#include "IxAtmTypes.h"
#include "IxFpathAcc.h"
#include "IxEthAccFpathDep.h"
#include "IxAssert.h"
#include "IxOsServices.h"
#include "IxFpathBufferPoolManager_p.h"
#include "IxFpathConfigControl_p.h"

/*
 *Constants
 */
#define IX_FPATH_ACC_MAX_CLS_LEN 128
#define IX_FPATH_ACC_MAX_MDF_LEN 56

/* 
 * local variables
 */

/* this mutex is used to serialize accesses to IxFpathAcc */
static IxMutex ixFpathSvIfLock;
/* this flag is used to indicate that initialization has been completed */
static BOOL ixFpathSvIfInitDone = FALSE;
/* this flag will allow Init to complete irrespective of whether Ethernet
   is aware of the Fpath Component.
   WARNING; this is only provided to support unit testing. */
static BOOL ixFpathInitOveride = FALSE;

/* ------------------------------------------------------
 * Initialse the lock
 */
PRIVATE IX_STATUS
ixFpathSvIfLockInit( void )
{
    if( ixOsServMutexInit(&ixFpathSvIfLock) != IX_SUCCESS )
    {
        return IX_FAIL;
    }
    return IX_SUCCESS;
}

/* ------------------------------------------------------
 * get the lock
 */
PRIVATE void
ixFpathSvIfLockGet( void )
{
    IX_STATUS returnStatus;

    returnStatus = ixOsServMutexLock(&ixFpathSvIfLock);
    IX_ENSURE( returnStatus == IX_SUCCESS,"failed to get mutex");

    return;
}

/* ------------------------------------------------------
 * Release the lock
 */
PRIVATE void 
ixFpathSvIfLockRelease( void )
{
    IX_STATUS returnStatus;
                   
    returnStatus = ixOsServMutexUnlock(&ixFpathSvIfLock);
    IX_ENSURE( returnStatus == IX_SUCCESS,"failed to release mutex");

    return;
}

/* ------------------------------------------------------
 * Initialise
 */
PUBLIC IX_STATUS
ixFpathAccInit(void)
{
    IX_STATUS returnStatus = IX_FAIL;

# if (IX_ETH_ACC_FPATH_AWARE==0)
    if ( ixFpathInitOveride == FALSE ) 
    {
        ixOsServLog( LOG_ERROR, 
            "ixFpathAccInit failed because IX_ETH_ACC_FPATH_AWARE is 0 ",
            0,0,0,0,0,0);
        
        return returnStatus;
    }
    else
    {
        ixOsServLog( LOG_WARNING, 
            "ixFpathAccInit override - IxEthAcc not aware of IxFpath",
            0,0,0,0,0,0);
    }
# endif

    if(!ixFpathSvIfInitDone)
    {
        /* Initialize subcomponents */
        if(( ixFpathBPMInit() == IX_SUCCESS ) &&
           ( ixFpathCfgCtrlInit() == IX_SUCCESS ) &&
           ( ixFpathSvIfLockInit() == IX_SUCCESS ))
        {
            ixFpathSvIfInitDone = TRUE;
            returnStatus = IX_SUCCESS;
        }
    }
    return returnStatus;
}

/* ------------------------------------------------------
 * Set the ethernet type for the specified port
 */
PUBLIC IX_STATUS
ixFpathAccEtherTypeSet(IxFpathEthernetPort port, 
                       UINT16 etherType)
{
    IX_STATUS returnStatus = IX_FAIL;

    /* Validate the parameters */
    if ( ( ixFpathSvIfInitDone ) &&
         ( port >= IX_FPATHACC_ETHERNET_PORT0 ) &&
         ( port < IX_FPATHACC_ETHERNET_MAX_PORTS))
    {
        ixFpathSvIfLockGet();
        returnStatus = ixFpathCfgCtrlEtherTypeSet(port, etherType);
        ixFpathSvIfLockRelease();
    }

    return returnStatus;
}

/* ------------------------------------------------------
 * Supply mbufs to the fast path free queue
 */
PUBLIC IX_STATUS
ixFpathAccBufferSupply(IX_MBUF *mBuf, 
                       int *unpopulatedEntryCnt)
{
    IX_STATUS returnStatus = IX_FAIL;

    /* Validate the parameters */
    if ( ( ixFpathSvIfInitDone ) &&
         ( unpopulatedEntryCnt != NULL))
    {
        ixFpathSvIfLockGet();
        returnStatus = ixFpathBPMBufferSupply(mBuf, unpopulatedEntryCnt);
        ixFpathSvIfLockRelease();
    }

    return returnStatus;
}

/* ------------------------------------------------------
 * Get the maximum and current size of the buffer pool
 */
PUBLIC IX_STATUS
ixFpathAccBufferPoolInfoGet(int *maxEntries, 
                            int *populatedEntryCnt)
{
    IX_STATUS returnStatus = IX_FAIL;
    
    /* Validate the parameters */
    if ( ( ixFpathSvIfInitDone ) &&
         ( maxEntries != NULL) &&
         ( populatedEntryCnt != NULL))
    {
        ixFpathSvIfLockGet();
        returnStatus = ixFpathBPMInfoGet(maxEntries, populatedEntryCnt);
        ixFpathSvIfLockRelease();
    }

    return returnStatus;
}

/* ------------------------------------------------------
 * Set the callback for fast path free queue underrun
 */
PUBLIC IX_STATUS
ixFpathAccBufferPoolUnderRunCallbackSet(
    IxFpathAccBufPoolUnderRunCallback callback,
    IxFpathAccClientId clientId)
{
    IX_STATUS returnStatus = IX_FAIL;

    /* Validate the parameters */
    if ( ixFpathSvIfInitDone )
    {
        ixFpathSvIfLockGet();
        returnStatus = ixFpathBPMUnderRunCallbackSet( callback,
                                                      clientId);
        ixFpathSvIfLockRelease();
    }

    return returnStatus;
}

/* ------------------------------------------------------
 * Enable fast path  by settting for the specified npeVcid
 * a classifier and modifier template
 */
PUBLIC IX_STATUS
ixFpathAccTemplatePairSet(IxAtmNpeRxVcId npeVcId,
                          unsigned  char *classifierTemplate,
                          int classifierTemplateLength,
                          unsigned  char *modifierTemplate,
                          int modifierTemplateLength)
{
    IX_STATUS returnStatus = IX_FAIL;

    /* Validate the parameters */
    if ( ( ixFpathSvIfInitDone ) &&
         ( npeVcId >= 0 ) && 
         ( npeVcId < IX_ATM_MAX_NUM_VC ) &&
         ( classifierTemplate != NULL ) &&
         ( classifierTemplateLength > 0 ) &&
         ( classifierTemplateLength <= IX_FPATH_ACC_MAX_CLS_LEN ) &&
         ( modifierTemplate != NULL ) &&
         ( modifierTemplateLength > 0) &&
         ( modifierTemplateLength <= IX_FPATH_ACC_MAX_MDF_LEN) )
    {
        ixFpathSvIfLockGet();
        
        returnStatus = ixFpathCfgCtrlTemplatePairSet(npeVcId,
                                                 classifierTemplate,
                                                 classifierTemplateLength,
                                                 modifierTemplate,
                                                 modifierTemplateLength);
        ixFpathSvIfLockRelease();
    }

    return returnStatus;
}

/* ------------------------------------------------------
 * Disable fast path for the specified npeVcid by clearing 
 * the classifier modifier template
 */
PUBLIC IX_STATUS
ixFpathAccTemplatePairClear(IxAtmNpeRxVcId npeVcId)
{
    IX_STATUS returnStatus = IX_FAIL;

    /* Validate the parameters */
    if ( ( ixFpathSvIfInitDone ) &&
         ( npeVcId >= 0 ) && 
         ( npeVcId < IX_ATM_MAX_NUM_VC )) 
    {
        ixFpathSvIfLockGet();
        returnStatus = ixFpathCfgCtrlTemplatePairClear(npeVcId);
        ixFpathSvIfLockRelease();
    }
    return returnStatus;
}

/* ------------------------------------------------------
 * Set the callback for new Wan Mac Addresses
 */
PUBLIC IX_STATUS
ixFpathAccWanMacAddrCallbackSet( 
    IxFpathAccNewWanMacAddrCallback macAddrCallback,
    IxAtmNpeRxVcId npeVcId,
    IxFpathAccClientId clientId)
{
    IX_STATUS returnStatus = IX_FAIL;

    /* Validate the parameters */
    if ( ( ixFpathSvIfInitDone ) &&
         ( npeVcId >= 0 ) && 
         ( npeVcId < IX_ATM_MAX_NUM_VC )) 
    {
        ixFpathSvIfLockGet();
        returnStatus = 
            ixFpathCfgCtrlWanMacAddrCallbackSet( macAddrCallback,
                                                 npeVcId,
                                                 clientId);

        ixFpathSvIfLockRelease();
    }
    return returnStatus;
}


/* ------------------------------------------------------
 * Get fast path statistics
 */
PUBLIC IX_STATUS 
ixFpathAccStatisticsGet( IxAtmNpeRxVcId npeVcId, int *packetCnt)
{
    IX_STATUS returnStatus = IX_FAIL;

    /* Validate the parameters */
    if ( ( ixFpathSvIfInitDone ) &&
         ( npeVcId >= 0 ) && 
         ( npeVcId < IX_ATM_MAX_NUM_VC ) &&
         ( packetCnt != NULL )) 
    {
        ixFpathSvIfLockGet();
        returnStatus = 
            ixFpathCfgCtrlStatisticsGet( npeVcId, packetCnt );

        ixFpathSvIfLockRelease();
    }
    return returnStatus;

}


/* ------------------------------------------------------
 * Show interesting things about the fast path component
 */
PUBLIC IX_STATUS
ixFpathAccShow(void)
{

# if (IX_ETH_ACC_FPATH_AWARE==0)
    ixOsServLog( LOG_WARNING, 
        "ixFpathAccInit override - IxEthAcc not aware of IxFpath",
        0,0,0,0,0,0);
# endif

    printf("\n---------------------------------------"
           "\nixFpathAccShow()"
           "\n---------------------------------------");

    if(!ixFpathSvIfInitDone)
    {
        printf("\nNot Initialized");
        return IX_FAIL;
    }
    else
    {
       ixFpathBPMShow();
       ixFpathCfgCtrlShow();
       return IX_SUCCESS;
    }
}

/* Called before init - will ensure init will work */
void
ixFpathInitOverRideSet(void)
{
    ixFpathInitOveride = TRUE;
}

