/**
 * @file IxHssAccPktTx.c
 *
 * @author Intel Corporation
 * @date 10 Jan 2002
 *
 * @brief This file contains the implementation of the private API for the 
 * Transmit Module.
 *
 * -- Intel Copyright Notice --
 *
 * @par
 * INTEL CONFIDENTIAL
 *
 * @par
 * Copyright 2002 Intel Corporation All Rights Reserved.
 *
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.  The Material
 * contains trade secrets and proprietary and confidential information of
 * Intel or its suppliers and licensors.  The Material is protected by
 * worldwide copyright and trade secret laws and treaty provisions. No
 * part of the Material may be used, copied, reproduced, modified,
 * published, uploaded, posted, transmitted, distributed, or disclosed in
 * any way without Intel's prior express written permission.
 *
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel or otherwise.  Any license under
 * such intellectual property rights must be express and approved by
 * Intel in writing.
 *
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * -- End Intel Copyright Notice --
*/

/**
 * Put the system defined include files required.
 */

#include <stdio.h>

/**
 * Put the user defined include files required.
 */
#include "IxHssAcc.h"
#include "IxNpeA.h"
#include "IxQMgr.h"
#include "IxHssAccError_p.h"
#include "IxHssAccPktTx_p.h"
#include "IxHssAccPDM_p.h"
#include "IxHssAccPCM_p.h"
#include "IxHssAccCommon_p.h"

/*
 * Defines
 */
/* the following define adds error checking of descriptor contents on */
/* the transmit done data path */
#define IX_HSSACC_TXDESC_CONTENTS_DEBUG_CHECK FALSE
/* the following define adds error checking of mbuf lenfth on */
/* the transmit data path */
#define IX_HSSACC_TXMBUF_LENGTH_DEBUG_CHECK FALSE
/* The follwing define allows the Nearly Empty watermark level */
/* of the TxDone Q to be configurable.*/
#define	IX_HSSACC_PKT_TXDONE_Q_WATERMARK IX_QMGR_Q_WM_LEVEL0

/* 
 * The folling macro is used to calculate the remaining free space in an
 * mbuf that can be used to store data.  buffer->pClBlk->clSize can no
 * longer be used as pClBlk is a vxWorks dependent implementation, and is
 * not present in the POSIX mbuf.  The macro M_TRAILINGSPACE is present in
 * both vxWorks and POSIX.
 */
#define IX_HSSACC_MBUF_TRAILINGSPACE(buffer) M_TRAILINGSPACE (buffer);

/*
 * Typedefs
 */
typedef struct
{
    unsigned txs;     
    unsigned txDones;
    unsigned qWriteFails;
    unsigned qWriteOverflows;
    unsigned txDoneQReadFails;
    unsigned maxEntriesInTxDoneQ;
    unsigned invalidDescs;
} IxHssAccPktTxStats;


/**
 * Variable declarations global to this file only. 
 *  Externs are followed by static variables.
 */
static IxHssAccPktTxStats ixHssAccPktTxStats;


/**
 * Function definition:ixHssAccPktTxInternal
 */
__inline__ 
IX_STATUS 
ixHssAccPktTxInternal (IxHssAccHssPort hssPortId, IxHssAccHdlcPort hdlcPortId,
			IX_MBUF *buffer)
{
    IxHssAccPDMDescriptor *desc;   /* A desc ptr that we will write the 
				      clients mbuf to*/
    IxHssAccPDMDescriptor *physDesc;
    IX_STATUS status = IX_SUCCESS;
#if IX_HSSACC_TXMBUF_LENGTH_DEBUG_CHECK
    unsigned packetLength;
    IX_MBUF *tmpBuffer;
#endif
    IX_HSSACC_TRACE0 (IX_HSSACC_FN_ENTRY_EXIT, "Entering "
		      "ixHssAccPktTxInternal\n");
    
    if (ixHssAccPCMCheckTxOk (hssPortId, hdlcPortId))
    {
#if IX_HSSACC_TXMBUF_LENGTH_DEBUG_CHECK
	packetLength = IX_MBUF_MLEN(buffer);
	tmpBuffer = buffer;
	while (IX_MBUF_NEXT_BUFFER_IN_PKT_PTR(tmpBuffer) != NULL)
	{
	    tmpBuffer = IX_MBUF_NEXT_BUFFER_IN_PKT_PTR(tmpBuffer);
	    packetLength += IX_MBUF_MLEN(tmpBuffer);
	}
	
	if (IX_MBUF_PKT_LEN(buffer) != packetLength) 
	{
	    IX_HSSACC_REPORT_ERROR ("ixHssAccPktTxInternal:Invalid packet length,"
				    " packetLength not set correctly\n");
	    return IX_FAIL;
	}
#endif
	
	status = ixHssAccPDMDescGet (hssPortId, hdlcPortId, IX_HSSACC_PDM_TX_POOL,
				     &desc);
	if (status != IX_SUCCESS)
	{
	    return status;
	}
	
	/* Set desc up as required by the NPE*/
	desc->npeDesc.pRootMbuf    = ixHssAccPDMMbufToNpeFormatConvert(buffer);
	desc->npeDesc.status       = IX_HSSACC_PKT_OK;
	desc->npeDesc.chainCount   = 0;
	desc->npeDesc.packetLength = IX_MBUF_PKT_LEN(buffer);
	desc->npeDesc.pMbufData    = (unsigned char *) IX_MBUF_MDATA(buffer);
	desc->npeDesc.mbufLength   = IX_MBUF_MLEN(buffer);
	desc->npeDesc.pNextMbuf    = IX_MBUF_NEXT_BUFFER_IN_PKT_PTR(buffer); 
	desc->npeDesc.errorCount   = 0;
	
	IX_HSSACC_PKT_DATA_CACHE_FLUSH (desc, sizeof (IxHssAccPDMDescriptor));
	physDesc = (IxHssAccPDMDescriptor *) IX_HSSACC_PKT_MMU_VIRT_TO_PHY (desc);
	/* Write the desc to the Tx Q for this client */
	status = ixQMgrQWrite (ixHssAccPCMTxQIdGet (hssPortId, hdlcPortId), 
			       (UINT32 *) &physDesc);
	if (status != IX_SUCCESS)
	{
	    if (status == IX_FAIL)
	    {
		IX_HSSACC_REPORT_ERROR ("ixHssAccPktTxInternal:Writing "
					"a Tx descriptor "
					"to the Tx Queue failed\n");
		ixHssAccPktTxStats.qWriteFails++;
	    }
	    else if (status == IX_QMGR_Q_OVERFLOW)
	    {
		ixHssAccPktTxStats.qWriteOverflows++;
		status = IX_HSSACC_Q_WRITE_OVERFLOW;
	    }
	    ixHssAccPDMDescFree (desc, IX_HSSACC_PDM_TX_POOL);

	    /* we need to restore the buffer to the original (virtual) address
	     * before we return
	     */
	    ixHssAccPDMMbufFromNpeFormatConvert(desc->npeDesc.pRootMbuf);
	    return status;
	}
    }
    else 
    {
	IX_HSSACC_REPORT_ERROR ("ixHssAccPktTxInternal:This HDLC Port is either"
				" not connected or not started\n");
	return IX_FAIL;
    }

    ixHssAccPktTxStats.txs++;
    IX_HSSACC_TRACE0 (IX_HSSACC_FN_ENTRY_EXIT, "Exiting "
		      "ixHssAccPktTxInternal\n");
    return status;
}


/**
 * Function definition: ixHssAccPktTxDoneCallback
 */
void 
ixHssAccPktTxDoneCallback (IxQMgrQId ixHssAccPktTxDoneQId, 
			   IxQMgrCallbackId cbId)
{
    /* This function is called from within an ISR */
    IX_STATUS status;
    IxHssAccPDMDescriptor *desc;
    IxHssAccHssPort hssPortId =  cbId;
    unsigned numOfTxDoneQReads = 0;
    IX_HSSACC_DEBUG_OFF (IX_HSSACC_TRACE0 (IX_HSSACC_FN_ENTRY_EXIT, "Entering "
					   "ixHssAccPktTxDoneCallback\n"));

    /* Do a Q read to remove the desc ptr from the Q */
    do 
    {
	status = ixQMgrQRead (ixHssAccPktTxDoneQId, (UINT32*)&desc); 
	numOfTxDoneQReads++;
	if (status == IX_SUCCESS)
	{

#if IX_HSSACC_TXDESC_CONTENTS_DEBUG_CHECK
	    if (desc != NULL)
	    {
#endif
		desc = IX_HSSACC_PKT_MMU_PHY_TO_VIRT (desc);
		IX_HSSACC_PKT_DATA_CACHE_INVALIDATE(desc, sizeof (IxHssAccPDMDescriptor));
		desc->npeDesc.pRootMbuf = ixHssAccPDMMbufFromNpeFormatConvert (
		    desc->npeDesc.pRootMbuf);

#if IX_HSSACC_TXDESC_CONTENTS_DEBUG_CHECK
		/* Is the hss Port Id lower than the max hssPortId */
		/* AND is the hdlc Port Id lower than the max hdlcPortId */
		if (!(IX_HSSACC_ENUM_INVALID (hssPortId, IX_HSSACC_HSS_PORT_MAX)) && 
		    !(IX_HSSACC_ENUM_INVALID (desc->hdlcPortId, IX_HSSACC_HDLC_PORT_MAX)))
		{
#endif
		    ixHssAccPCMTxDoneCallbackRun (hssPortId,
						  desc->hdlcPortId,
						  desc->npeDesc.pRootMbuf, 
						  desc->npeDesc.errorCount,
						  desc->npeDesc.status);
		    
		    ixHssAccPDMDescFree (desc, IX_HSSACC_PDM_TX_POOL);
		    ixHssAccPktTxStats.txDones++;
#if IX_HSSACC_TXDESC_CONTENTS_DEBUG_CHECK		
		}
		else
		{
		    IX_HSSACC_DEBUG_OFF 
			(IX_HSSACC_REPORT_ERROR ("ixHssAccPktTxDoneCallback: Invalid "
						 "descriptor was recieved.\n"));
		    ixHssAccPktTxStats.invalidDescs++;
		}
	    }
	    else
	    {
		IX_HSSACC_DEBUG_OFF 
		    (IX_HSSACC_REPORT_ERROR ("ixHssAccPktTxDoneCallback: Invalid "
					     "descriptor was recieved.\n"));
		    ixHssAccPktTxStats.invalidDescs++;
	    }
#endif
	}
	else if (status == IX_FAIL)
	{
	    IX_HSSACC_DEBUG_OFF 
		(IX_HSSACC_REPORT_ERROR ("ixHssAccPktTxDoneCallback:"
					 "TxDone QRead failed\n"));
	    ixHssAccPktTxStats.txDoneQReadFails++;
	}
#ifdef IX_HSSACC_NO_QMQ_READ_DRAINING
	status = IX_QMGR_Q_UNDERFLOW;
#endif
    } while (status == IX_SUCCESS);
    
    if (numOfTxDoneQReads > ixHssAccPktTxStats.maxEntriesInTxDoneQ)
    {
	ixHssAccPktTxStats.maxEntriesInTxDoneQ = numOfTxDoneQReads;
    }
    IX_HSSACC_DEBUG_OFF (IX_HSSACC_TRACE0 (IX_HSSACC_FN_ENTRY_EXIT, "Exiting "
					   "ixHssAccPktTxDoneCallback\n"));
}


/**
 * Function definition: ixHssAccPktTxShow
 */
void 
ixHssAccPktTxShow (void)
{
    printf ("\nixHssAccPktTxShow:\n");
    printf ("           txs: %d \t     txDones: %d\n",
	    ixHssAccPktTxStats.txs,
	    ixHssAccPktTxStats.txDones);
    printf ("    writeFails: %d \t   readFails: %d \t  invalidDescs: %d\n",
	    ixHssAccPktTxStats.qWriteFails,
	    ixHssAccPktTxStats.txDoneQReadFails,
	    ixHssAccPktTxStats.invalidDescs);
    printf ("writeOverflows: %d \tmaxInTxDoneQ: %d\n",
	    ixHssAccPktTxStats.qWriteOverflows,
	    ixHssAccPktTxStats.maxEntriesInTxDoneQ);
}


/**
 * Function definition: ixHssAccPktTxStatsInit
 */
void 
ixHssAccPktTxStatsInit (void)
{
    ixHssAccPktTxStats.txs         = 0;
    ixHssAccPktTxStats.txDones     = 0;
    ixHssAccPktTxStats.qWriteFails = 0;
    ixHssAccPktTxStats.qWriteOverflows  = 0;
    ixHssAccPktTxStats.txDoneQReadFails = 0;
    ixHssAccPktTxStats.maxEntriesInTxDoneQ = 0;
    ixHssAccPktTxStats.invalidDescs = 0;
}

/**
 * Function definition: ixHssAccPktRxInit
 */
IX_STATUS 
ixHssAccPktTxInit (void)
{
    IX_STATUS status = IX_SUCCESS;

    IX_HSSACC_TRACE0 (IX_HSSACC_FN_ENTRY_EXIT, 
		      "Entering ixHssAccPktTxInit\n");

    /* initialise stats */
    ixHssAccPktTxStatsInit ();

    /* Set the callback for the TxDone Q */
    status = ixQMgrNotificationCallbackSet (
	IX_NPE_A_QMQ_HSS_PKT_TX_DONE,
	ixHssAccPktTxDoneCallback,
	IX_HSSACC_HSS_PORT_MAX - 1);
    
    if (status != IX_SUCCESS)
    {
	IX_HSSACC_REPORT_ERROR("ixHssAccPktTxInit:"
			       "Notification CallbackSet failed for TxDone Q\n");
	return IX_FAIL;
    }
    
    /* Set the Watermark for the TxDone Q for this client */
    status = ixQMgrWatermarkSet (
	IX_NPE_A_QMQ_HSS_PKT_TX_DONE,
	IX_HSSACC_PKT_TXDONE_Q_WATERMARK/*NE flag*/,
	0/*NF flag*/);
    
    if (status != IX_SUCCESS)
    {
	IX_HSSACC_REPORT_ERROR("ixHssAccPktTxInit:"
			       "ixQMgrWatermarkSet failed for this"
				       " clients TxDone Q\n");
	return IX_FAIL;
    }    

    /* Enable notification for the TxDone Q */
    status = ixQMgrNotificationEnable (IX_NPE_A_QMQ_HSS_PKT_TX_DONE,
				       IX_QMGR_Q_SOURCE_ID_NOT_NE);
    if (status != IX_SUCCESS)
    {
	IX_HSSACC_REPORT_ERROR("ixHssAccPktTxInit:"
			       "Notification enable failed for TxDone Q\n");
	return IX_FAIL;
    }
    
    IX_HSSACC_TRACE0 (IX_HSSACC_FN_ENTRY_EXIT, 
		      "Exiting ixHssAccPktTxInit\n");
    return status;
}

