/** 
 * @file IxAdsl.h
 * 
 * @brief Header file for the ADSL Driver
 * 
 * @par
 * -- Intel Copyright Notice --
 *
 * @par
 * INTEL CONFIDENTIAL
 * 
 * @par
 * Copyright 1999-2002 Intel Corporation All Rights Reserved.
 *
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.  The Material
 * contains trade secrets and proprietary and confidential information of
 * Intel or its suppliers and licensors.  The Material is protected by
 * worldwide copyright and trade secret laws and treaty provisions. No
 * part of the Material may be used, copied, reproduced, modified,
 * published, uploaded, posted, transmitted, distributed, or disclosed in
 * any way without Intel's prior express written permission.
 *
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel or otherwise.  Any license under
 * such intellectual property rights must be express and approved by
 * Intel in writing.
 *
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * -- End Intel Copyright Notice --
 */

/**
 * @defgroup IxAdsl IXP4xx ADSL Driver (IxAdsl) API
 *
 * @brief The public API for the IXP4xx ADSL Driver
 *
 * @{
 */

#ifndef IX_ADSL_H
#define IX_ADSL_H

#include "IxAdslMgmtInfo.h"

/**
 * @ingroup IxAdsl
 *
 * @def IX_ADSL_SUCCESS		  
 *
 * @brief Line was opened and is in 'showtime state' 	
 */
#define IX_ADSL_SUCCESS		  0

/**
 * @ingroup IxAdsl
 *
 * @def IX_ADSL_OPEN_FAILED
 *
 * @brief Line did not open properly. 	
 */
#define	IX_ADSL_OPEN_FAILED	  1

/**
 * @ingroup IxAdsl
 *
 * @def IX_ADSL_UNSUPPORTED_MODE
 *
 * @brief Illegal mode type for OPEN state.
 */
#define	IX_ADSL_UNSUPPORTED_MODE  2

/**
 * @ingroup IxAdsl
 *
 * @def IX_ADSL_COMMS_FAILURE
 *
 * @brief Serial read/write failure for Serial CTRLE only
 */
#define	IX_ADSL_COMMS_FAILURE	  3

/** 
 * @ingroup IxAdsl
 *
 * @def IX_ADSL_CLOSE_FAILED	  
 *
 * @brief Line failed to close properly.    
 */
#define IX_ADSL_CLOSE_FAILED	  4

/**
 * @ingroup IxAdsl
 *
 * @def IX_ADSL_ALREADY_DOWN	  
 *
 * @brief Line is already down
 */
#define IX_ADSL_ALREADY_DOWN	  5

/**
 * @ingroup IxAdsl
 *
 * @def IF_TYPE_ADSL_PHYS
 *  
 * @brief Physical interface channel type used for ADSL
 */
#define IF_TYPE_ADSL_PHYS  94

/**
 * @ingroup IxAdsl
 *
 * @def IF_TYPE_ADSL_INTLV   
 *
 * @brief Interleave interface channel type used for ADSL
 */
#define IF_TYPE_ADSL_INTLV 124

/**
 * @ingroup IxAdsl
 *
 * @def IF_TYPE_ADSL_FAST  
 *
 * @brief The fast interface channel type used for ADSL
 */
#define IF_TYPE_ADSL_FAST  125	

/**
 * @ingroup IxAdsl
 *
 * @def IX_ADSL_SELFTEST_OK		
 *
 * @brief Selftest passed.
 */
#define IX_ADSL_SELFTEST_OK		0	

/**
 * @ingroup IxAdsl
 *
 * @def IX_ADSL_SELFTEST_FAILED	
 *
 * @brief Selftest failed.
 */
#define IX_ADSL_SELFTEST_FAILED		1	

/**
 * @ingroup IxAdsl
 *
 * @def IX_ADSL_SELFTEST_UNIMPLEMENTED		
 *
 * @brief Selftest un-implemented.
 */
#define IX_ADSL_SELFTEST_UNIMPLEMENTED		2

/**
 * @ingroup IxAdsl
 *
 * @brief Used to indicate the type of ADSL line type used
 *
 * typedef enum t_ADSL_LINE_TYPE { \n
 *              ADSL_NORMAL =       0,    - Normal line type \n
 *              ADSL_GLITE =        1,    - GLite line type \n
 *              ADSL_DMT = 	    2,    - DMT line type \n
 *              ADSL_ANSI =         3,    - ANSI line type \n
 *              ADSL_MULTIMODE =    4,    - Multimode line type \n
 *              ADSL_INVALID_MODE = 5     - Invalid line type \n
 * }t_ADSL_LINE_TYPE;
 */
typedef enum t_ADSL_LINE_TYPE
{
		ADSL_NORMAL = 	0,
		ADSL_GLITE = 	1,
		ADSL_DMT = 	2,
		ADSL_ANSI =     3,
		ADSL_MULTIMODE =4,
		ADSL_INVALID_MODE = 5 /* Used internally to indicate last valid enum  */
} t_ADSL_LINE_TYPE;

/**
 * @ingroup IxAdsl
 *
 * @brief Used to indicate the ADSL line state either UP or DOWN
 * 
 * typedef enum t_ADSL_LINE_STATE { \n
 *        IX_ADSL_LINE_UP,       - ADSL line state is up \n
 *        IX_ADSL_LINE_DOWN,     - ADSL line state is down \n
 *        IX_ADSL_INVALID_STATE  - ADSL line state is invalid \n
 * } t_ADSL_LINE_STATE;
 */
typedef enum t_ADSL_LINE_STATE
{
	IX_ADSL_LINE_UP,
	IX_ADSL_LINE_DOWN,
	IX_ADSL_INVALID_STATE
} t_ADSL_LINE_STATE;

/**
 * @ingroup IxAdsl
 *
 * @brief Used to indicate whether the ADSL physical is either CO or CPE
 *
 * typedef enum t_ADSL_PHY_TYPE { \n
 *      IX_ADSL_CO,  - ADSL type is CO \n
 *      IX_ADSL_CPE   - ADSL type is CPE \n
 * } t_ADSL_PHY_TYPE;
 */
typedef enum t_ADSL_PHY_TYPE
{
        IX_ADSL_CO, 
        IX_ADSL_CPE
} t_ADSL_PHY_TYPE;

/**
 * @ingroup IxAdsl
 *
 * @brief Type definition callback function to indicate of the changes on the 
 * line state
 */
typedef void ( *IX_ADSL_LINE_CHANGE_FNC )(unsigned int lineNum, t_ADSL_LINE_STATE lineState);

/** 
 * @ingroup IxAdsl
 *
 * @fn ixAdslDriverInfoGet(unsigned int lineNum,
		    const IxAdslManagerInfo entity,
		    UINT32 ifIndex,
		    ManagerInfoData *dataBufferToWriteInto);
 *
 * @brief Retrieve information for the given ADSL line.
 *
 * This function returns to its caller the information for the given ADSL 
 * line associated with the management entity in the form of a ManagerInfoData 
 * structure.
 *        
 * @param "int lineNum" Parameter showing which ADSL module is being used.
 * @param "IxAdslManagerInfo entity" Is the enum value that corresponds to the
 *  MIB entity being set.
 * @param "UINT32 index" A variable containing the required index for the 
 * entity.
 * @param "ManagerInfoData *dataBuffer" A buffer to fill with the pertinent 
 * data.
 *
 * @return 
 *		- IX_SUCCESS: request completed successfully.
 *		- IX_FAILED: request failed for unspecified reason.
 *		- IX_MGRINF_INVALID: request invalid, unspecified reason.
 *		- IX_MGRINF_NOSUCHINDEX: instance of index given doesn't exist.
 *		- IX_MGRINF_NOSUCHENTITY: entity passed in does not exist.
 *		- IX_MGRINF_INDEXOUTOFRANGE: index passed not within valid range.
 *		- IX_MGRINF_NOMORE: index passed is the last valid index.
 *		- IX_MGRINF_NOTSUPPORTED: state to be supported in future.
 */
INT32
ixAdslDriverInfoGet(unsigned int lineNum,
		    const IxAdslManagerInfo entity,
		    UINT32 ifIndex,
		    ManagerInfoData *dataBufferToWriteInto);

/** 
 * @ingroup IxAdsl
 *
 * @fn ixAdslDriverInfoSet(unsigned int lineNum,
		    const IxAdslManagerInfo entity,
		    UINT32 ifIndex,
		    ManagerInfoData *dataBufferToRead);
 *
 * @brief Set operation for the given ADSL line
 *
 * The set operation changes, for the given ADSL line, the contents of the 
 * component's local store of information, pertaining to the specified entity, 
 * using the input dataBuffer contents.
 *        
 * @param "int lineNum" parameter showing which ADSL module is being used.
 * @param "IxAdslManagerInfo entity" Is the enum value that corresponds to the MIB entity being set.
 * @param "UINT32 index" A variable containing the required index for the entity.
 * @param "ManagerInfoData *dataBuffer" A buffer to fill with the pertinent data.
 *
 *  @return 
 *		- IX_SUCCESS: request completed successfully.
 *		- IX_FAILED: request failed for unspecified reason.
 *		- IX_MGRINF_INVALID: request invalid, unspecified reason.
 *		- IX_MGRINF_NOSUCHINDEX: instance of index given doesn't exist.
 *		- IX_MGRINF_NOSUCHENTITY: entity passed in does not exist.
 *		- IX_MGRINF_INDEXOUTOFRANGE: index passed not within valid range.
 *		- IX_MGRINF_NOMORE: index passed is the last valid index.
 *		- IX_MGRINF_NOTSUPPORTED: state to be supported in future.
 */
INT32
ixAdslDriverInfoSet(unsigned int lineNum,
		    const IxAdslManagerInfo entity,
		    UINT32 ifIndex,
		    ManagerInfoData *dataBufferToRead);

/**
 * @ingroup IxAdsl
 *
 * @brief Type definition function pointer to a callback in the Snmp component
 */
typedef void (*IxAdslDriverTrapCallbackPtr_t)(const IxAdslManagerTraps entity, UINT32 ifIndex);

/** 
 * @ingroup IxAdsl
 *
 * @fn ixAdslDriverTrapBind(IxAdslDriverTrapCallbackPtr_t funcPtr);
 *
 * @brief The Trap interface is used to prevent circular dependencies
 *
 * ADSL Management registers the function pointer as a subscriber to all events. 
 * IxAdslDriverTrapBind stores this function pointer internally and invokes it 
 * when an event listed in IxAdslManagerTraps occurs to notify the SNMP agent.
 *        
 * @param "IxAdslDriverTrapCallbackPtr_t funcPtr" This function takes a function pointer to a 
 * callback in the Snmp component. Where the prototype for the function pointer @ref IxAdslTrapCallbackPtr_t 
 *
 *
 *  @return 
 *		- IX_SUCCESS: request completed successfully.
 *		- IX_FAILED: request failed for unspecified reason.
 */
INT32
ixAdslDriverTrapBind(IxAdslDriverTrapCallbackPtr_t funcPtr);




/*
 * Attempt to bring up the link up.
 */

/** 
 * @ingroup IxAdsl
 *
 * @fn IxADSL_CPE_Open(unsigned int lineNum,enum t_ADSL_LINE_TYPE)
 *
 * @brief Opens the given ADSL line in ATU-R mode 
 *        
 * @param "int lineNum" parameter showing which ADSL module is being used.
 * @param "enum t_ADSL_LINE_TYPE" Indicates type of ADSL to be opened.
 *
 * Opens the given ADSL line in ATU-R mode and puts it in the 'showtime' state, 
 * i.e. available to carry user data. Note that when the CPE is asked to open 
 * its line in NORMAL mode and if the CO is configured for G.LITE, then the CPE 
 * will come up in G.LITE mode
 *
 * Blocking : This call may block for several seconds while the link is established.
 *
 * Impacts On Global Data: Sets the admin state.
 *
 * Pre-Conditions: The code must only be called once the operating system is running, 
 * i.e. do not call as part of the hardware init as this code requires base services. The code 
 * should only be called from task level.
 *
 * Post-Conditions: No cleanup after this call is required.
 * 
 * Exceptions: None.
 *
 * @return 
 *		- IX_ADSL_SUCCESS: Line was opened and is in 'showtime' state 
 *		- IX_ADSL_OPEN_FAILED: Line did not open properly. 
 *		- IX_ADSL_UNSUPPORTED_MODE: Illegal mode type for OPEN. 
 */
unsigned int IxADSL_CPE_Open(unsigned int lineNum,enum t_ADSL_LINE_TYPE );

/** 
 * @ingroup IxAdsl
 *
 * @fn IxADSL_CPE_LineClose(unsigned int lineNum)
 *
 * @brief This closes a previously opened ATU-R ADSL line. 
 *
 * @param "int lineNum" parameter showing which ADSL module is being used.
 *
 * If line-close is successful notify the callback that the line is now closed. The line
 * will be put in the idle state.
 *        
 * Blocking : This call may block for several seconds, as the negotiation to bring 
 * down the line takes place.
 *
 * Impacts On Global Data:  Sets the admin state.
 *
 * Pre-Conditions:  The code must only be called once the operating system is running, 
 * i.e. do not call as part of the hardware init as this code requires base services. The code 
 * should only be called from task level.
 *
 * Post-Conditions:  No cleanup after this call is required.
 *
 * Exceptions:  None.
 * 
 * @return 
 *		- IX_ADSL_SUCCESS: Line was closed
 *		- IX_ADSL_CLOSE_FAILED: Line did not close properly.
 *		- IX_ADSL_ALREADY_DOWN: Line was not open before close.
 */
unsigned int IxADSL_CPE_LineClose(unsigned int lineNum);

/** 
 * @ingroup IxAdsl
 *
 * @fn IxADSL_CO_Open(unsigned int lineNum,enum t_ADSL_LINE_TYPE )
 *
 * @brief Opens the given ADSL line in ATU-C mode 
 *
 * @param "int lineNum" Parameter showing which ADSL module is being used.
 * @param "enum t_ADSL_LINE_TYPE" Indicates type of ADSL to be opened.
 * 
 * Opens the given ADSL line in ATU-C mode and puts it in 'showtime' state, i.e. available to 
 * carry user data. There are a substantial number of line parameters that would normally be 
 * configurable as an ATU-C device. The IxADSL_CO_Open API does not allow for any control of 
 * these parameters. It will open the line with a default recommended set for demonstration usage. 
 * The user may be given access to the ADSL driver layer should our device be targeted towards 
 * a DSLAM type applications. 
 *  
 *  Blocking : This call may block for several seconds while the link is established.
 *
 *  Impacts On Global Data: Sets the admin state.
 *
 * Pre-Conditions: The code must only be called once the operating system is running, 
 * i.e. do not call as part of the hardware init as this code requires base services. The code 
 * should only be called from task level.
 *
 *  Post-Conditions: No cleanup after this call is required.
 *
 * Exceptions: None.
 * 
 * @return 
 *		- IX_ADSL_SUCCESS: Line was opened and is in 'showtime' state.
 *		- IX_ADSL_OPEN_FAILED: Line did not open properly.
 *		- IX_ADSL_UNSUPPORTED_MODE: Illegal mode type for OPEN.
 */
unsigned int IxADSL_CO_Open(unsigned int lineNum,enum t_ADSL_LINE_TYPE );

/** 
 * @ingroup IxAdsl
 *
 * @fn IxADSL_CO_LineClose(unsigned int lineNum)
 *
 * @brief Closes a previously opened ATU-C ADSL line. The line will be put in the idle state.
 *
 * @param "int lineNum" parameter showing which ADSL module is being used.
 *  
 * Blocking : Non-blocking   
 *
 * Impacts On Global Data: Sets the admin state. Notifies a callback routine.
 *
 * Pre-Conditions: Code should only be called from task level.
 *
 * Post-Conditions: No cleanup after this call is required.
 *
 * @return 
 *              - IX_ADSL_SUCCESS: Line was closed.
 *              - IX_ADSL_CLOSE_FAILED: Line did not close properly.
 *              - IX_ADSL_ALREADY_DOWN: Line was not open before close.
 *
 */
unsigned int IxADSL_CO_LineClose(unsigned int lineNum);

/** 
 * @ingroup IxAdsl
 *
 * @fn unsigned int IxADSL_LineOpen( unsigned int lineNum,
			      t_ADSL_PHY_TYPE phyType,
			      unsigned char *initFirmware,
			      unsigned int initFirmwarewLen,
			      unsigned char *phyFirmware,
			      unsigned int phyFirmwareLen,
			      t_ADSL_LINE_TYPE lineType )
 *
 * @brief Opens the given ADSL line in the specified mode 
 *
 * @param "int lineNum" parameter showing which ADSL module is being used.
 * @param "t_ADSL_PHY_TYPE phyType" the type of Phy used CO or CPE
 * @param "unsigned char *initFirmware" Pointer to an array containing firmware to initialise Alcatel PHY.
 * @param "unsigned int initFwLen" Indicates the length in bytes of Acatel PHY initialisation firmware.
 * @param "unsigned char *phyFirmware" Pointer to an array containing the CO or CPE firmware as indicated 
 *		  by the phyType parameter.
 * @param "unsigned int phyFwLen" Indicates length in bytes of the main CO or CPE firmware load to be 
 *		  downloaded onto the Alcatel PHY.
 * @param "t_ADSL_LINE_TYPE lineType" Indicates type of ADSL to be opened.
 *
 * Opens the given ADSL line in the specified mode and puts it in the 'showtime' state, i.e. available 
 * to carry user data. Called by IxADSL_CO_Open and IxADSL_CPE_Open. 
 *
 * Blocking : This call may block for several seconds while the link is established.
 *
 * Impacts On Global Data: Sets the admin state.
 *
 * Pre-Conditions: The code must only be called once the operating system is running, 
 * i.e. do not call as part of the hardware init as this code requires base services. The code 
 * should only be called from task level.
 *
 * Post-Conditions: No cleanup after this call is required.
 *
 * Exceptions: None.
 * 
 * @return 
 *		- IX_ADSL_SUCCESS: Line was opened and is in 'showtime' state
 *		- IX_ADSL_OPEN_FAILED: Line did not open properly.
 *		- IX_ADSL_UNSUPPORTED_MODE: Illegal mode type for OPEN.
 */
unsigned int IxADSL_LineOpen( unsigned int lineNum,
			      t_ADSL_PHY_TYPE phyType,
			      unsigned char *initFirmware,
			      unsigned int initFirmwarewLen,
			      unsigned char *phyFirmware,
			      unsigned int phyFirmwareLen,
			      t_ADSL_LINE_TYPE lineType );

/** 
 * @ingroup IxAdsl
 *
 * @fn xADSL_LineStateReg(unsigned int lineNum, IX_ADSL_LINE_CHANGE_FNC callback_fn)
 *
 * @brief This is a notification procedure that gets called if the line state of the given ADSL 
 * line goes from UP to DOWN (or visa versa). 
 *
 * @param "int lineNum" parameter showing which ADSL module is being used.
 * @param "IX_ADSL_LINE_CHANGE_FNC callback_fn" Indicates the line change 
 * 
 * The callback may be used for SNMP trap generation. The state of the line is also passed back. 
 *
 * Blocking : This call is a non-blocking function.
 *
 * Impacts On Global Data: This sets a global callback handler.
 *
 * Pre-Conditions: There are no pre conditions to this call.
 *
 * Post-Conditions: A global line state handler for ADSL line state changes shall be registered. 
 * It is advisable to register the callback before the given ADSL line is opened.
 *
 *  Exceptions: None.
 * 
 * @return 
 *		- IX_ADSL_SUCCESS: callback was registered.
 *		- IX_ADSL_FAILED: Internal error, registration failed.
 */
unsigned int IxADSL_LineStateReg(unsigned int lineNum, IX_ADSL_LINE_CHANGE_FNC callback_fn);

/** 
 * @ingroup IxAdsl
 *
 * @fn IxADSL_GetLineState(unsigned int lineNum)
 *
 * @brief Returns the current state of the given ADSL line.
 *
 * @param "int lineNum" Parameter showing which ADSL module is being used.
 * 
 * Blocking : This call is a non-blocking
 *
 * Impacts On Global Data: None
 *
 * Pre-Conditions: Task level calls only.
 *
 * Post-Conditions: None.
 *
 * Exceptions: None.
 * 
 * @return 
 *		- IX_ADSL_LINE_UP: The line is in show time state.
 *		- IX_ADSL_LINE_WAIT_FOR_ACTIVATING: The line is waiting for the peer to activate.
 *		- IX_ADSL_LINE_ACTIVATING: The line is negotiating with its peer.
 *		- IX_ADSL_LINE_DOWN: The line is idle.
 */
unsigned int IxADSL_GetLineState(unsigned int lineNum);

/** 
 * @ingroup IxAdsl
 *
 * @fn IxADSL_GetLineRateUpstream(unsigned int lineNum)
 *
 * @brief Return the current upstream line speed of the given ADSL line. 
 *
 * @param "int lineNum" parameter showing which ADSL module is being used.
 * 
 * This represents the speed of the line to the ATU-C devices assuming that the device is 
 * configured as an ATU-R device, i.e. the Tx speed from this device, and visa versa when 
 * the device is configured as an ATU-C.
 *
 * Blocking : This call is a non-blocking
 *
 * Impacts On Global Data: None
 *
 * Pre-Conditions: Task level calls only.
 *
 * Post-Conditions: None.
 *
 * Exceptions: None.
 * 
 * @return 
 *		- Integer: bit rate in kbits/second. N.B. Returns zero if the line is not in 'showtime' state.
 */
unsigned int IxADSL_GetLineRateUpstream(unsigned int lineNum);

/** 
 * @ingroup IxAdsl
 *
 * @fn IxADSL_GetLineRateDownstream(unsigned int lineNum)
 *
 * @brief Return the current downstream line speed of the given ADSL line. 
 *
 * @param "int lineNum" parameter showing which ADSL module is being used.
 *
 * This is the speed of the line from the ATU-C devices assuming that the device is configured 
 * as an ATU-R device, i.e. the Rx speed of this device, and, and visa versa when the device 
 * is configured as an ATU-C.
 * 
 * Blocking : This call is a non-blocking
 *
 * Impacts On Global Data: None
 *
 * Pre-Conditions: Task level calls only.
 *
 * Post-Conditions: None.
 *
 * Exceptions: None.
 * 
 * @return 
 *		- Integer: bit rate in kbits/second. N.B. Returns zero if the line is not in 'showtime' state.
 */
unsigned int IxADSL_GetLineRateDownstream(unsigned int lineNum);

/**
 * @ingroup IxAdsl
 *
 * @fn IxADSL_EnableDyingGasp(unsigned int lineNum)
 * 
 * @brief Enables the function that informs ATU-C when condition that leads to shutdown of the 
 * given Adsl line has been detected. 
 *
 * @param "int lineNum" parameter showing which ADSL module is being used.
 *
 * Blocking : This call is a non-blocking
 *
 * Impacts On Global Data: None
 *
 * Pre-Conditions: Task level calls only. Only Available in ATU-R.
 *
 * Post-Conditions: None.
 *
 * Exceptions: None.
 * 
 * @return None
 */
void IxADSL_EnableDyingGasp(unsigned int lineNum);

/**
 * @ingroup IxAdsl
 *
 * @def IX_ADSL_LINE_UNKNOWN_STATE              
 *
 * @brief ADSL Line unknown state
 */
#define IX_ADSL_LINE_UNKNOWN_STATE               (0) 

/**
 * @ingroup IxAdsl
 *
 * @def IX_ADSL_LINE_UP
 *
 * @brief The line is in show time state.
 */
#define IX_ADSL_LINE_UP				(1)

/**
 * @ingroup IxAdsl
 *
 * @def IX_ADSL_LINE_WAIT_FOR_ACTIVATING	
 *
 * @brief The line is waiting for the peer to activate.
 */
#define	IX_ADSL_LINE_WAIT_FOR_ACTIVATING	(2)

/**
 * @def IX_ADSL_LINE_ACTIVATING			
 *
 * @brief The line is negotiating with its peer.
 */
#define	IX_ADSL_LINE_ACTIVATING			(3)	

/**
 * @ingroup IxAdsl
 *
 * @def IX_ADSL_LINE_DOWN			
 *
 * @brief The line is down.
 */
#define	IX_ADSL_LINE_DOWN			(4)


/** 
 * @ingroup IxAdsl
 *
 * @fn void IxADSL_setVendorCode(unsigned int lineNum,
			  unsigned char itu_vendorid_countrycode,
			  unsigned char itu_vendorid_vendorcode_1,
			  unsigned char itu_vendorid_vendorcode_2,
			  unsigned char itu_vendorid_vendorcode_3,
			  unsigned char itu_vendorid_vendorcode_4,
			  unsigned char itu_vendorid_vendorspecific_1,
			  unsigned char itu_vendorid_vendorspecific_2);
 *
 * @brief Set the vendor specific bytes in the given ADSL line. 
 *
 * The vendor ID must be set before the line is open, if not the default vendor ID shall 
 * be sent to the peer modem upon request.
 *
 * @param "int lineNum" parameter showing which ADSL module is being used.
 * @param "unsigned char itu_vendorid_countrycode" Vendor country code that are predefined in 
 * standards. 
 * @param "unsigned char itu_vendorid_vendorcode_1" Vendor code 1 that are predefined in 
 * standards. 
 * @param "unsigned char itu_vendorid_vendorcode_2" Vendor code 2 that are predefined in 
 * standards. 
 * @param "unsigned char itu_vendorid_vendorcode_3" Vendor code 3 that are predefined in 
 * standards. 
 * @param "unsigned char itu_vendorid_vendorcode_4" Vendor code 4 that are predefined in 
 * standards. 
 * @param "unsigned char itu_vendorid_vendorspecific_1" Vendor specific 1 that are predefined in 
 * standards. 
 * @param "unsigned char itu_vendorid_vendorspecific_12 Vendor specific 2 that are predefined in 
 * standards. 
 *
 * Note: Vendor specific values are taken by the phy and linked together to form a single code
 *
 * The vendor ID must be set before the line is open, if not the default vendor ID shall 
 * be sent to the peer modem upon request.
 *
 * Blocking : This call is a non-blocking
 *
 * Impacts On Global Data: Sets an internal ADSL global data structure. This shall be 
 * used for all subsequent ADSL line open commands.
 *
 * Pre-Conditions: Task level calls only.
 *
 * Post-Conditions: None.
 *
 * Exceptions: None.
 * 
 * @return None
 */
void IxADSL_setVendorCode(unsigned int lineNum,
			  unsigned char itu_vendorid_countrycode,
			  unsigned char itu_vendorid_vendorcode_1,
			  unsigned char itu_vendorid_vendorcode_2,
			  unsigned char itu_vendorid_vendorcode_3,
			  unsigned char itu_vendorid_vendorcode_4,
			  unsigned char itu_vendorid_vendorspecific_1,
			  unsigned char itu_vendorid_vendorspecific_2);


/** 
 * @ingroup IxAdsl
 *
 * @fn IxADSL_SelfTest(unsigned int lineNum)
 *
 * @brief Runs the Modem self test.
 *
 * @param "int lineNum" parameter showing which ADSL module is being used. 
 *
 * This should only be called when the modem is off-line (line is closed).
 *
 * Blocking : This call may block for several seconds as the selftest completes.
 *
 * Impacts On Global Data: None
 *
 * Pre-Conditions: Task level calls only. If the given ADSL line is open when the call is made, 
 * the line shall be forced closed prior to the selftest execution.
 *
 * Post-Conditions: This shall leave the line in the closed state.
 *
 * Exceptions: None.
 *
 * @return 
 *                   - IX_ADSL_SELFTEST_OK: Selftest passed.
 *                   - IX_ADSL_SELFTEST_FAILED: Selftest failed.
 *                   -IX_ADSL_SELFTEST_UNIMPLEMENTED: Selftest unimplemented.
 *
 */
unsigned int IxADSL_SelfTest(unsigned int lineNum);

/*
 * Generic Print routing to display line stats. ( same function for all calls)
 */

/** 
 * @ingroup IxAdsl
 *
 * @fn IxADSL_DumpStats (unsigned int lineNum)
 *
 * @brief This function will show the current statistics associated with the given ADSL Line.
 *
 * The functionality is the same as @ref ixAdslShow
 */
void IxADSL_DumpStats (unsigned int lineNum);

/** 
 * @ingroup IxAdsl
 *
 * @fn ixAdslShow(unsigned int lineNum)
 *
 * @brief This function will show the current statistics associated with the given ADSL Line.
 *
 * @param "int lineNum" parameter showing which ADSL module is being used.
 *
 * Blocking : This call is non-blocking.
 *
 * Impacts On Global Data: None
 *
 * Pre-Conditions: Task level calls only.
 *
 * Post-Conditions: None.
 *
 * Exceptions: None.
 * 
 * @return None
 */
void ixAdslShow(unsigned int lineNum);

/** 
 * @ingroup IxAdsl
 *
 * @fn ixAdslStatsShow (unsigned int lineNum
 * 
 * @brief This function will show the current statistics associated with the given ADSL Line.
 *
 * The functionality is the same as @ref ixAdslShow
 */
void ixAdslStatsShow (unsigned int lineNum); /* Prefered interface */



#endif /* IX_ADSL_H */

/**
 *@}
 */





