/**
 * @file   IxCryptoAcc.h
 *
 * @date   2-OCT-2002
 *
 * @brief   Header file for the IXP4xx Security component (IxCryptoAcc)
 *
 * -- Intel Copyright Notice --
 *
 * @par
 * INTEL CONFIDENTIAL
 *
 * @par
 * Copyright 2002 Intel Corporation All Rights Reserved.
 *
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.  The Material
 * contains trade secrets and proprietary and confidential information of
 * Intel or its suppliers and licensors.  The Material is protected by
 * worldwide copyright and trade secret laws and treaty provisions. No
 * part of the Material may be used, copied, reproduced, modified,
 * published, uploaded, posted, transmitted, distributed, or disclosed in
 * any way without Intel's prior express written permission.
 *
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials either expressly, by
 * implication, inducement, estoppel or otherwise.  Any license under
 * such intellectual property rights must be express and approved by
 * Intel in writing.
 *
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * -- End Intel Copyright Notice --
 */


/**
 * @defgroup IxCryptoAcc IXP4xx Security (IxCryptoAcc) API
 *
 * @brief IXP4xx Security component Public API
 *
 * @{
 */

#ifndef IXCRYPTOACC_H
#define IXCRYPTOACC_H

/*
 * Put the user defined include files required
 */

#include "IxOsBuffMgt.h"
#include "IxTypes.h"



/*
 * #defines and macros used in this file.
 */
 
#define IX_CRYPTO_ACC_MAX_CIPHER_KEY_LENGTH 32  /**<
                                                 * Max length (byte) of cipher
                                                 * key for DES (64 bit), 
                                                 * 3DES (192 bit), AES (128, 192
                                                 * & 256 bit). 
                                                 */
#define IX_CRYPTO_ACC_MAX_CIPHER_IV_LENGTH 16   /**<
                                                 * Max IV length in byte
                                                 */
#define IX_CRYPTO_ACC_MAX_AUTH_KEY_LENGTH 20    /**<
                                                 * Max length (byte) of  
                                                 * authentication key for 
                                                 * SHA1 (160 bit) and
                                                 * MD5 (128 bit).
                                                 */
#define IX_CRYPTO_ACC_MAX_AUTH_IV_LENGTH 20     /**<
                                                 * Max length (byte) of initial 
                                                 * chaining variable for 
                                                 * SHA1 (160 bit) and 
                                                 * MD5 (128 bit).
                                                 */
#define  IX_CRYPTO_ACC_MAX_QUEUE_DEPTH 64       /**<
                                                 * Max queue depth supported
                                                 * by the Queue Manager.
                                                 */
#define  IX_CRYPTO_ACC_MAX_ACTIVE_SA_TUNNELS 1000  /**<
                                                    * Maximum active tunnels
                                                    * supported could be changed
                                                    * by the client based on the
                                                    * application's requirements.
                                                    * Number of active tunnels
                                                    * will not impact the
                                                    * performance but will have 
                                                    * an impact on the memory
                                                    * needed to keep the
                                                    * crypto context information
                                                    * will depend on it. Overall
                                                    * memory requirement depends
                                                    * on the number of tunnels.
                                                    */


/*
 * Constants value for DES algorithm in unit bytes
 */

#define IX_CRYPTO_ACC_DES_KEY_64     8
#define IX_CRYPTO_ACC_DES_BLOCK_64   8
#define IX_CRYPTO_ACC_DES_IV_64      8
#define IX_CRYPTO_ACC_3DES_KEY_192   24


/*
 * Constants value for AES algorithm in unit bytes
 */

#define IX_CRYPTO_ACC_AES_KEY_128    16
#define IX_CRYPTO_ACC_AES_KEY_192    24
#define IX_CRYPTO_ACC_AES_KEY_256    32
#define IX_CRYPTO_ACC_AES_BLOCK_128  16
#define IX_CRYPTO_ACC_AES_CBC_IV_128 16
#define IX_CRYPTO_ACC_AES_CTR_IV_64  8



/*
 * Constants value for SHA1 algorithm in unit bytes
 */

#define IX_CRYPTO_ACC_SHA1_KEY_160    20
#define IX_CRYPTO_ACC_SHA1_DIGEST_160 20


/*
 * Constants value for MD5 algorithm in unit bytes
 */

#define IX_CRYPTO_ACC_MD5_KEY_128     16
#define IX_CRYPTO_ACC_MD5_DIGEST_128  16


 
/*
 * Typedefs whose scope is limited to this file.
 */

/**
 * @enum IxCryptoAccOperation
 *
 * @brief Cryptographic Operation Definitions  
 *
 * @note Only one of these operation is supported.
 */ 
typedef enum 
{
    IX_CRYPTO_ACC_OP_ENCRYPT = 0, 
    IX_CRYPTO_ACC_OP_DECRYPT, 
    IX_CRYPTO_ACC_OP_AUTH_CALC,
    IX_CRYPTO_ACC_OP_AUTH_CHECK,
    IX_CRYPTO_ACC_OP_ENCRYPT_AUTH, 
    IX_CRYPTO_ACC_OP_AUTH_DECRYPT,
    IX_CRYPTO_ACC_OP_TYPE_OF_OPERATION
} IxCryptoAccOperation;



/**
 * @enum IxCryptoAccCipherAlgo
 *
 * @brief Cipher Algorithm Definitions  
 *
 * @note 3DES and AES will be supported if not violating import/export rules.
 */
typedef enum
{
    IX_CRYPTO_ACC_CIPHER_NULL = 0,    
    IX_CRYPTO_ACC_CIPHER_DES,
    IX_CRYPTO_ACC_CIPHER_3DES,
    IX_CRYPTO_ACC_CIPHER_AES,
    IX_CRYPTO_ACC_CIPHER_ALGO_TYPE
} IxCryptoAccCipherAlgo;



/**
 * @enum IxCryptoAccCipherMode
 *
 * @brief Cipher Mode Definitions  
 *
 * @note CFB and OFB are not supported.
 */
typedef enum
{
    IX_CRYPTO_ACC_MODE_NULL = 0,
    IX_CRYPTO_ACC_MODE_ECB,
    IX_CRYPTO_ACC_MODE_CBC,
    IX_CRYPTO_ACC_MODE_CTR,
    IX_CRYPTO_ACC_MODE_TYPE
} IxCryptoAccCipherMode;



/**
 * @enum IxCryptoAccAuthAlgo
 *
 * @brief Authentication Algorithm Definitions  
 *
 * @note Only two authentication algorithms are supported, SHA1 and MD5.
 */
typedef enum
{
    IX_CRYPTO_ACC_AUTH_NULL = 0,
    IX_CRYPTO_ACC_AUTH_SHA1,
    IX_CRYPTO_ACC_AUTH_MD5,
    IX_CRYPTO_ACC_AUTH_TYPE
} IxCryptoAccAuthAlgo;



/**
 * @enum IxCryptoAccStatus
 *
 * @brief Invalid Status Definitions  
 *
 * @note These status will be used by the APIs to return to the client.
 */
typedef enum
{
    IX_CRYPTO_ACC_STATUS_SUCCESS = IX_SUCCESS,
    IX_CRYPTO_ACC_STATUS_FAIL = IX_FAIL,
    IX_CRYPTO_ACC_STATUS_WAIT,
    IX_CRYPTO_ACC_STATUS_RETRY,
    IX_CRYPTO_ACC_STATUS_QUEUE_FULL,
    IX_CRYPTO_ACC_STATUS_OPERATION_NOT_SUPPORTED,
    IX_CRYPTO_ACC_STATUS_CIPHER_ALGO_NOT_SUPPORTED,
    IX_CRYPTO_ACC_STATUS_CIPHER_MODE_NOT_SUPPORTED,
    IX_CRYPTO_ACC_STATUS_CIPHER_INVALID_KEY_LEN,
    IX_CRYPTO_ACC_STATUS_CIPHER_INVALID_IV_LEN,
    IX_CRYPTO_ACC_STATUS_CIPHER_INVALID_BLOCK_LEN,
    IX_CRYPTO_ACC_STATUS_AUTH_ALGO_NOT_SUPPORTED,
    IX_CRYPTO_ACC_STATUS_AUTH_INVALID_DIGEST_LEN,
    IX_CRYPTO_ACC_STATUS_AUTH_INVALID_KEY_LEN,
    IX_CRYPTO_ACC_STATUS_AUTH_FAIL,
    IX_CRYPTO_ACC_STATUS_CRYPTO_CTX_NOT_VALID,
    IX_CRYPTO_ACC_STATUS_EXCEED_MAX_TUNNELS 
} IxCryptoAccStatus;



/**
 * @brief   Structure storing cipher configuration parameters required to
 *          perform security functionality.
 *
 * Structure is used to specify the cipher context required for using the 
 * @ref IxCryptoAcc interface.
 */ 
typedef struct
{
    IxCryptoAccCipherAlgo cipherAlgo; /**< Cipher Algorithm - DES, 3DES, 
                                       * AES
                                       */
    IxCryptoAccCipherMode cipherMode; /**< Cipher mode - ECB, CBC, CTR */
    UINT32 cipherKeyLen;  /**< Cipher key length in bytes */
    union 
    {
        UINT8 cipherKey[IX_CRYPTO_ACC_MAX_CIPHER_KEY_LENGTH]; /**< default
	                                                       * key array
                                                               */ 
        UINT8 desKey[IX_CRYPTO_ACC_DES_KEY_64];         /**< DES key */
        UINT8 tripleDesKey[IX_CRYPTO_ACC_3DES_KEY_192]; /**< 3DES key */
        UINT8 aesKey128[IX_CRYPTO_ACC_AES_KEY_128];     /**< AES-128 key */
        UINT8 aesKey192[IX_CRYPTO_ACC_AES_KEY_192];     /**< AES-192 key */
        UINT8 aesKey256[IX_CRYPTO_ACC_AES_KEY_256];     /**< AES-256 key */
    } key;                          /**< Cipher key */
    UINT32 cipherBlockLen;          /**< 
                                     * Input plaintext/ciphertext blocks
                                     * size in bytes.
                                    */
    UINT32 cipherInitialVectorLen;  /**< Length of IV in bytes */
    UINT32 securityParameterIndex;  /**< 
                                     * SPI is used in generating counter 
                                     * block for AES CTR mode.
                                     */       
} IxCryptoAccCipherCtx;



/**
 * @brief   Structure storing authentication configuration parameters required 
 *          to perform security functionality.
 *
 * Structure is used to specify the authentication required for using the 
 * @ref IxCryptoAcc interface.
 */
typedef struct
{
    IxCryptoAccAuthAlgo authAlgo; /**< authentication algorithm -
                                   * MD5, SHA1
                                   */
    UINT32 authDigestLen; /**< Digest length in bytes */
    UINT32 authKeyLen;    /**< Hash key length in bytes */
    union
    {
        UINT8 authKey[IX_CRYPTO_ACC_MAX_AUTH_KEY_LENGTH]; /**< default hash
	                                                   * key array
                                                           */
        UINT8 sha1Key[IX_CRYPTO_ACC_SHA1_KEY_160]; /**< SHA1 key */
        UINT8 md5Key[IX_CRYPTO_ACC_MD5_KEY_128];   /**< MD5 key */
    } key; /**< Hash key */      
} IxCryptoAccAuthCtx;   



/**
 * @brief   Structure storing configuration parameters required to
 *          perform security functionality.
 *
 * Structure is used to specify the context required for using the 
 * @ref IxCryptoAcc interface.
 *
 */ 
typedef struct
{
    IxCryptoAccOperation operation;     /**< Types of operation */
    IxCryptoAccCipherCtx cipherCtx;
    IxCryptoAccAuthCtx authCtx;
    BOOL useDifferentSrcAndDestMbufs;   /**< If TRUE, data read from srcMbuf, 
                                         *  results written to destMbuf, if 
                                         *  FALSE data read from srcMbuf, 
                                         *  written back to srcMbuf. 
                                         *  Default is FALSE. Note that only
                                         *  the crypted/authenticated data that 
                                         *  is copied, not the entire source 
                                         *  mbuf.
                                         */ 
} IxCryptoAccCtx;



/**
 *
 * @ingroup IxCryptoAcc
 *
 * @brief  Cryptographic Context registration completed callback notification.
 *
 * This function is called to notify a client that the Crypto Context has been 
 * registered. This function will return the status through the associated 
 * CryptoCtxId once the parameters are calculated by Network Processor Engine
 * is (NPE) and stored in the struct, Cryptographic Context Database (CCD)  
 * which reference by the CryptoCtxId. 
 * The CryptoCtxId is valid until ixCryptoAccCtxUnregister is invoked.
 * 
 * If the callback function return the IX_CRYPTO_ACC_STATUS_WAIT status, it 
 * indicates that registration is not completed yet, but the mbuf pointer 
 * will need to be freed by client and the client need to  wait for the next
 * completion indication.Registration completed successfully only if status 
 * IX_CRYPTO_ACC_STATUS_SUCCESS is received.
 *
 * @param   cryptoCtxId [in] This CryptoCtxId is provided when the 
 *          callback was registered for a particular Security Association (SA)  
 *          via ixCryptoAccAuthCryptPerform. CryptoCtxId points to a  
 *          struct (CCD) consists of cryptographic paramaters database  
 *          required by the Network Processor Engine (NPE).
 * @param   *pMbuf [in] Pointer to the unused mbuf (to be freed by 
 *          client). The client should free any mbuf that is not NULL.
 * @param   status [in] Status reporting to the client
 *          via IxCryptoAccStatus.
 *          
 * @note @li   IX_CRYPTO_ACC_STATUS_SUCCESS - registration successful.
 * @note @li   IX_CRYPTO_ACC_STATUS_FAIL - registration fail.
 * @note @li   IX_CRYPTO_ACC_STATUS_WAIT - registration not completed yet, 
 *             wait for the next completion indication. NPE is busy in 
 *             calculating the initial variables needed for the registration.
 *
 */
typedef void (*IxCryptoAccRegisterCompleteCallback)(
    UINT32 cryptoCtxId,
    IX_MBUF *pMbuf, 
    IxCryptoAccStatus status);

 
 
 /**
 *
 * @ingroup IxCryptoAcc
 *
 * @brief  Security process completed callback notification.
 *
 * This function is called to notify a client that the cryptographic transaction
 * has been completed. The CryptoCtxId is returned to the client to indicate
 * which process that has been completed. The CryptoCtxId is created via 
 * ixCryptoAccCtxRegister, and this ID is unique to a particular IPSec 
 * tunnel.
 *
 * @param   cryptoCtxId [in] The CryptoCtxId is provided when the   
 *          callback was registered via ixCryptoAccCtxRegister for a   
 *          particular Security Association (SA). Then the CryptoCtxId will    
 *          be passed ixCryptoAccAuthCryptPerform. CryptoCtxId points to  
 *          a struct into (CCD) consists of cryptographic paramaters database  
 *          required by the Network Processor Engine (NPE). It is returned to  
 *          the client to indicate that the process for the particular  
 *          CryptoCtxId is completed.
 * @param   *pSrcMbuf [in] Pointer to the source data mbuf   
 *          descriptor, whichis also processed data mbuf descriptor if 
 *          UseDifferentSrcAndDestMbufs is FALSE.
 * @param   *pDestMbuf [in] Only used if UseDifferentSrcAndDestMbufs  
 *          is TRUE. Pointer to the processed data mbuf descriptor .
 * @param   status [in] Status reporting to the client 
 *          via IxCryptoAccStatus.
 *
 * @note  @li  IX_CRYPTO_ACC_STATUS_SUCCESS - Operation completed successfully.
 * @note  @li  IX_CRYPTO_ACC_STATUS_FAIL - Operation fail.
 * @note  @li  IX_CRYPTO_ACC_STATUS_AUTH_FAIL - Authentication is unsuccessful.
 *             Note that when the authentication fails, the content of the 
 *             destination mbuf (which is the same as the source mbuf if a 
 *             normal in-line operation is performed) will be indeterminate.
 *       
 *
 */
typedef void (*IxCryptoAccPerformCompleteCallback)( 
    UINT32 cryptoCtxId,
    IX_MBUF *pSrcMbuf, 
    IX_MBUF *pDestMbuf,
    IxCryptoAccStatus status); 



/**
 *
 * @ingroup IxCryptoAcc
 *
 * @fn ixCryptoAccInit (void)
 *
 * @brief Initialise the Security Access component
 *
 * This API will initialize the Security Access component internals, such as 
 * Queue Manager, allocate memory and initialize all the variables. This API
 * must be called before any other API call on this component can be made. 
 * This API is called once only.
 *
 * @return @li  IX_CRYPTO_ACC_STATUS_SUCCESS - successfully initialize the 
 *              component.
 * @return @li  IX_CRYPTO_ACC_STATUS_FAIL - Initialisation failed for some 
 *              unspecified internal reason.
 *              
 * @li   Reentrant    : no
 * @li   ISR Callable : no
 *
 */
PUBLIC IxCryptoAccStatus
ixCryptoAccInit (void);



/**
 *
 * @ingroup IxCryptoAcc
 *
 * @fn ixCryptoAccCtxRegister (
    IxCryptoAccCtx *pAccCtx,
    IX_MBUF *pMbufPrimaryChainVar,
    IX_MBUF *pMbufSecondaryChainVar,
    IxCryptoAccRegisterCompleteCallback registerCallbackFn,
    IxCryptoAccPerformCompleteCallback performCallbackFn,
    UINT32 *pCryptoCtxId)
 *
 * @brief Cryptographic Context ID, CryptoCtxId Allocation for the Client
 *
 * This function is for the client to pass all the required information 
 * (eg. key, cipher algorithm, etc) for the hardware accelerator context. 
 * All the information will be converted into a cryptographic parameters 
 * structure to be shared with NPE. The structure will be associated with 
 * CryptoCtxId. (CryptoCtxId need to be passed back to client thru callback). 
 * Besides, two empty mbuf are required to be passed in through interface for 
 * the use of   access component and NPE to compute the primary and secondary 
 * initial chaining variables. pMbufPrimaryChainVar and pMbufSecondaryChainVar 
 * mbuf pointers must be NULL if authentication or combined service is not 
 * selected. 
 *          
 * @param   *pAccCtx [in] is a pointer to the Security Association 
 *          Database (SAD) which will assigned a CryptoCtxId.  
 *          The information inside SAD will be used for the creation of
 *          Cryptographic Context Database (CCD).
 * @param   *pMbufPrimaryChainVar [in] a pointer to an empty mbuf  
 *          (must not be chained) for the use of access component to compute 
 *          primary chaining variables for SHA1/MD5. This mbuf structure is  
 *          allocated by the client (mininum size of the cluster required is 
 *          64 bytes), only if SHA1/MD5 is selected, and the mbuf pointer must  
 *          be NULL if SHA1/MD5 is not selected. After the NPE complete the 
 *          computation, the mbuf is returned separately through the client's
 *          registered callback.
 * @param   *pMbufSecondaryChainVar [in] a pointer to an empty mbuf
 *          (must not be chained) for the use of access component to compute 
 *          secondary chaining variables for SHA1/MD5. This mbuf structure is  
 *          allocated by the client (mininum size of the cluster required is 
 *          64 bytes), only if SHA1/MD5 is selected, and the mbuf pointer must  
 *          be NULL if SHA1/MD5 is not selected. After the NPE complete the 
 *          computation, the mbuf is returned separately through the client's
 *          registered callback.
 * @param   registerCallbackFn [in] callback function pointer to return  
 *          CryptoCtxId the client when the registration is completed. This 
 *          cannot be NULL.
 * @param   performCallbackFn [in] callback function pointer to return the 
 *          processed buffer to the client with respect to the unique CryptoCtxId. 
 *          This cannot be NULL.
 * @param   *pCryptoCtxId: [inout] Crypto Context ID returned by access 
 *          component for the crypto context registered. 
 *
 * @return @li  IX_CRYPTO_ACC_STATUS_SUCCESS - Registration parameters valid
 * @return @li  IX_CRYPTO_ACC_STATUS_FAIL - Registration failed for some 
 *              unspecified internal reasons.
 * @return @li  IX_CRYPTO_ACC_STATUS_OPERATION_NOT_SUPPORTED - Invalid 
 *              operation requested by the client.
 * @return @li  IX_CRYPTO_ACC_STATUS_CIPHER_ALGO_NOT_SUPPORTED - Invalid 
 *              cipher algorithm requested by the client.
 * @return @li  IX_CRYPTO_ACC_STATUS_CIPHER_MODE_NOT_SUPPORTED - Invalid 
 *              cipher mode requested by the client.
 * @return @li  IX_CRYPTO_ACC_STATUS_CIPHER_INVALID_KEY_LEN - Invalid cipher 
 *              key length passed in by the client.
 * @return @li  IX_CRYPTO_ACC_STATUS_CIPHER_INVALID_BLOCK_LEN - Invalid cipher
 *              block length passed in by the client
 * @return @li  IX_CRYPTO_ACC_STATUS_CIPHER_INVALID_IV_LEN - Invalid IV 
 *              length passed in by the client.
 * @return @li  IX_CRYPTO_ACC_STATUS_AUTH_ALGO_NOT_SUPPORTED - Invalid 
 *              authentication requested by the client.
 * @return @li  IX_CRYPTO_ACC_STATUS_AUTH_INVALID_DIGEST_LEN - Invalid 
 *              authentication digest length.
 * @return @li  IX_CRYPTO_ACC_STATUS_AUTH_INVALID_KEY_LEN - Invalid 
 *              authentication key length.
 * @return @li  IX_CRYPTO_ACC_STATUS_EXCEED_MAX_TUNNELS - Exceed maximum 
 *              tunnels permitted.
 * @return @li  IX_CRYPTO_ACC_STATUS_QUEUE_FULL - Queue full status returned 
 *              and the registration request will be rejected.
 *              
 * @li   Reentrant    : no
 * @li   ISR Callable : no
 *
 */ 
PUBLIC IxCryptoAccStatus 
ixCryptoAccCtxRegister (
    IxCryptoAccCtx *pAccCtx,
    IX_MBUF *pMbufPrimaryChainVar,
    IX_MBUF *pMbufSecondaryChainVar,
    IxCryptoAccRegisterCompleteCallback registerCallbackFn,
    IxCryptoAccPerformCompleteCallback performCallbackFn,
    UINT32 *pCryptoCtxId);



/**
 * 
 * @ingroup IxCryptoAcc
 *
 * @fn ixCryptoAccCtxUnregister (UINT32 cryptoCtxId)
 *
 * @brief Unregister the CryptoCtxId from Cryptographic Context Database (CCD).
 *
 * This function is for freeing the particular CryptoCtxId from the Cryptographic 
 * Context Database.
 *
 * @param    cryptoCtxId [in] is the CCD pointer that is required  
 *           to be freed, which will be supplied by the client. The 
 *           CryptoCtxId must be a valid Id.
 *
 * @return @li  IX_CRYPTO_ACC_STATUS_SUCCESS - successfully unregister the 
 *              CryptoCtxId.
 * @return @li  IX_CRYPTO_ACC_STATUS_FAIL - Unregistration failed for some 
 *              unspecified internal reasons, e.g. uninitialized.
 * @return @li  IX_CRYPTO_ACC_STATUS_CRYPTO_CTX_NOT_VALID - invalid CryptoCtxId.
 * @return @li  IX_CRYPTO_ACC_STATUS_RETRY - retry the unregister operation as 
 *              there are still some ixCryptoAccAuthCryptPerform requests 
 *              associated with the CryptoCtxId pending on the NPE queue. 
 *              
 * @li   Reentrant    : no
 * @li   ISR Callable : no
 *
 */
PUBLIC IxCryptoAccStatus
ixCryptoAccCtxUnregister (UINT32 cryptoCtxId);



/**
 *
 * @ingroup IxCryptoAcc
 * 
 * @fn ixCryptoAccAuthCryptPerform (
    UINT32 cryptoCtxId,
    IX_MBUF *pSrcMbuf,
    IX_MBUF *pDestMbuf,
    UINT16 authStartOffset,
    UINT16 authDataLen,    
    UINT16 cryptStartOffset,   
    UINT16 cryptDataLen,
    UINT16 icvOffset,
    UINT8  *pIV)
 * 
 * @brief Perform Authentication and Decryption/Encryption functionality
 *
 * This function is called for authentication and decryption/encryption 
 * functionality.
 * Note that the restriction with respect to  AuthStartOffset, AuthStartLen, 
 * CryptStartOffset and CryptStartLen is that the crypted data must be a subset
 * of the authenticated data. The boundary relationship of 
 * [(authStartOffset+authDataLen) >= (cryptDataLen+cryptStartOffset) >=
 * (cryptStartOffset >= authStartOffset]
 * MUST BE SATISFIED.
 * There should not be any chained mbuf boundary with an ICV field.
 *
 * @param   cryptoCtxId [in] is the CCD pointer to be supplied by  
 *          the client. This CryptoCtxId is allocated via  
 *          ixCryptoAccCtxRegister. The CryptoCtxId must be a valid Id.
 * @param   *pSrcMbuf [in] is a pointer to the data to be processed.  
 *          This mbuf structure is allocated by the client. The processed data  
 *          will be stored in the location if UseDifferentSrcAndDestMbufs is 
 *          FALSE. The same pointer is then returned to the client via callback
 *          function.
 * @param   *pDestMbuf [in] only used if UseDifferentSrcAndDestMbufs  
 *          is TRUE. This is a pointer the buiffer that the reult is to be  
 *          written to. This mbuf structure is allocated by the client. The  
 *          processed data will be stored in the location. The same pointer  
 *          is then returned to the client via callback function   
 *          IxCryptoAccPerformCompleteCallback registered during 
 *          ixCryptoAccCtxRegister. The length of mbuf MUST be big enough to 
 *          hold the result of operation. The result of operation COULD NOT 
 *          span into two or more different mbufs, thus the mbuf supplied must 
 *          be at least the length of expected result. 
 * @param   authStartOffset [in] supplied by the client to indicate  
 *          the start of the payload to be authenticated.
 * @param   authDataLen [in] supplied by the client to indicate  
 *          the length of the payload to be authenticated in Bytes. The
 *	    maximum data length must not exceed 65471
 * @param   icvOffset [in] supplied by the client to indicate  
 *          the start of the ICV (Initial Chaining Variable) used for the
 *          authentication. This ICV field should not cross to multiple mbufs,
 *	    in a chained mbuf.
 * @param   cryptStartOffset [in] supplied by the client to indicate 
 *          the start of the payload to be decryption/encryption.
 * @param   cryptDataLen [in] supplied by the client to indicate the  
 *          length of the payload to be decryption/encryption in Bytes. The 
 *	    must be in the multiple of cipher block length.
 * @param   *pIV [in] Initialization Vector  supplied by the client 
 *          to be used for the decryption/encryption process.
 *          
 * @return @li  IX_CRYPTO_ACC_STATUS_SUCCESS - Operation requested is 
 *              successfully enqueued to hardware accelerator for processing.
 * @return @li  IX_CRYPTO_ACC_STATUS_FAIL - Cryptographic process  failed for 
 *              some unspecified internal reasons.
 * @return @li  IX_CRYPTO_ACC_STATUS_QUEUE_FULL - Cryptographic queue is full.
 * @return @li  IX_CRYPTO_ACC_STATUS_CIPHER_INVALID_BLOCK_LEN - Invalid plaintext
 *              / ciphertext block size passed in by the client.
 * @return @li  IX_CRYPTO_ACC_STATUS_CRYPTO_CTX_NOT_VALID - Crypto context is 
 *              not registered.
 *              
 * @li   Reentrant    : no
 * @li   ISR Callable : no
 *
 */ 
PUBLIC IxCryptoAccStatus
ixCryptoAccAuthCryptPerform (
    UINT32 cryptoCtxId,
    IX_MBUF *pSrcMbuf,
    IX_MBUF *pDestMbuf,
    UINT16 authStartOffset,
    UINT16 authDataLen,    
    UINT16 cryptStartOffset,   
    UINT16 cryptDataLen,
    UINT16 icvOffset,
    UINT8  *pIV);


                 
/**
 *
 * @ingroup IxCryptoAcc
 *
 * @fn ixCryptoAccShow (void)
 *
 * @brief API for printing statistic and status.
 *
 * This function is called by the client to print statistics, such as number 
 * of packets returned with operation fail, number of packets encrypted/
 * decrypted/authenticated while the function also print the current status of 
 * the queue, whether the queue is empty or full or current queue length.
 *          
 * @li   Reentrant    : no
 * @li   ISR Callable : no
 *
 */
PUBLIC void
ixCryptoAccShow (void);



/**
 *
 * @ingroup IxCryptoAcc
 *
 * @fn ixCryptoAccShowWithId (UINT32 cryptoCtxId)
 *
 * @brief API for printing statistic and status.
 *
 * This function prints all the statistics provided by ixCryptoAccShow. In
 * addition the function will print out the contents of the Crypto Context
 * corresponding to the CryptoCtxId supplied through the API
 *
 * @param   cryptoCtxId [in] Crypto Context ID which has been registered
 *          
 * @li   Reentrant    : no
 * @li   ISR Callable : no
 *
 */
PUBLIC void
ixCryptoAccShowWithId (UINT32 cryptoCtxId); 



/**
 *
 * @ingroup IxCryptoAcc
 *
 * @fn ixCryptoAccCryptoServiceStop (void)
 * 
 * @brief API to stop the security hardware accelerator services.
 *
 * This function is called to stop THE security hardware accelerator services. 
 * All the requests pending in queue will be completed before the security 
 * hardware accelerator halt. Any new requests issued after this call will be 
 * rejected. All the Crypto Context will be unregistered in this function call. 
 *
 *
 * @return @li  IX_CRYPTO_ACC_STATUS_SUCCESS - Operation requested is successful 
 *              with all the pending requests completed and CCD cleared.
 * @return @li  IX_CRYPTO_ACC_STATUS_FAIL - Cryptographic stop request failed 
 *              for some unspecified internal reasons.
 *                         
 * @li   Reentrant    : no
 * @li   ISR Callable : no
 *
 */ 
PUBLIC IxCryptoAccStatus
ixCryptoAccCryptoServiceStop (void);    
             
             
#endif /* ndef IXCRYPTOACC_H */
/**
 *@}
 */
