/** 
 * @file IxUART.h
 *
 * @date 12-OCT-01
 *
 * @brief Public header for the Intel IXP4xx internal UART, generic driver.
 * 
 * Design Notes:
 * This driver allows you to perform the following functions: 
 *	Device Initialization,
 *	send/receive characters.
 *
 * Perform Uart IOCTL for the following:
 *	Set/Get the current baud rate,
 *	set parity,
 *	set the number of Stop bits,
 *	set the character Length (5,6,7,8),
 *	enable/disable Hardware flow control.
 *
 * Only Polled mode is supported for now.
 *    
 * -- Intel Copyright Notice --
 *
 * @par
 * INTEL CONFIDENTIAL
 *
 * @par
 * Copyright 2002 Intel Corporation All Rights Reserved.
 *
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.  The Material
 * contains trade secrets and proprietary and confidential information of
 * Intel or its suppliers and licensors.  The Material is protected by
 * worldwide copyright and trade secret laws and treaty provisions. No
 * part of the Material may be used, copied, reproduced, modified,
 * published, uploaded, posted, transmitted, distributed, or disclosed in
 * any way without Intel's prior express written permission.
 *
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel or otherwise.  Any license under
 * such intellectual property rights must be express and approved by
 * Intel in writing.
 *
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * -- End Intel Copyright Notice --
*/

/**
 * @defgroup IxUARTAccAPI IXP4xx UART Access (IxUARTAcc) API
 *
 * @brief IXP4xx UARTAcc Driver Public API
 * 
 * @{
 */


/* Defaults */

/**
 * @def IX_UART_DEF_OPTS
 *
 * @brief The default hardware options to set the UART to -
 * no flow control, 8 bit word, 1 stop bit, no parity 
 */
// Jack ++ 2003/06/03
#define IX_UART_DEF_OPTS (CLOCAL | CS8)
//#define IX_UART_DEF_OPTS (CLOCAL | CS7)

/**
 * @def IX_UART_DEF_XMIT
 *
 * @brief The default UART FIFO size - must be no bigger than 64
 */
#define IX_UART_DEF_XMIT 64

/**
 * @def IX_UART_DEF_BAUD
 *
 * @brief The default UART baud rate - 9600
 */
// Jack ++ 2003/06/03
//#define IX_UART_DEF_BAUD 9600
#define IX_UART_DEF_BAUD 115200

/**
 * @def IX_UART_MIN_BAUD
 *
 * @brief The minimum UART baud rate - 9600
 */
#define IX_UART_MIN_BAUD 9600

/**
 * @def IX_UART_MAX_BAUD
 *
 * @brief The maximum UART baud rate - 926100
 */
#define IX_UART_MAX_BAUD 926100

/**
 * @def IX_UART_XTAL
 *
 * @brief The UART clock speed
 */
#define IX_UART_XTAL 14745600

/* IOCTL commands */
#define IX_BAUD_SET	0	/**< set the baud rate */
#define IX_BAUD_GET	1	/**< get the baud rate */
#define IX_MODE_SET	2	/**< set the UART mode of operation */
#define IX_MODE_GET	3	/**< get the current UART mode of operation */
#define IX_OPTS_SET	4	/**< set the UART device options */
#define IX_OPTS_GET	5	/**< get the UART device options */
#define IX_STATS_GET	6	/**< get the UART statistics */

/* POSIX style ioctl arguments */
#define CLOCAL		0x1	/**< software flow control */
#define CREAD		0x2	/**< enable interrupt receiver */
#define CSIZE		0xc	/**< characters size */
#define CS5		0x0	/**< 5 bits */
#define CS6		0x4	/**< 6 bits */
#define CS7		0x8	/**< 7 bits */
#define CS8		0xc	/**< 8 bits */
#define STOPB		0x20	/**< send two stop bits (else one) */
#define PARENB		0x40	/**< parity detection enabled (else disabled) */
#define PARODD		0x80	/**< odd parity  (else even) */

/**
 * @enum ixUARTMode
 * @brief The mode to set to UART to.
 */
typedef enum _ixUARTMode
{
    INTERRUPT=0,	/**< Interrupt mode */
    POLLED,		/**< Polled mode */
    LOOPBACK		/**< Loopback mode */
} ixUARTMode;

/**
 * @struct ixUARTStats
 * @brief Statistics for the UART.
 */
typedef struct _ixUARTStats
{
    UINT32  rxCount;
    UINT32  txCount;	
    UINT32  overrunErr;
    UINT32  parityErr;
    UINT32  framingErr;
    UINT32  breakErr;
} ixUARTStats;

/**
 * @struct ixUARTDev
 * @brief Device descriptor for the UART.
 */
typedef struct _ixUARTDev
{
    UINT8 *addr;	/**< device base address */
    ixUARTMode mode;	/**< interrupt, polled or loopback */
    int baudRate;	/**< baud rate */
    int freq;		/**< UART clock frequency */     
    int options;	/**< hardware options */
    int fifoSize;	/**< FIFO xmit size */

    ixUARTStats   stats;	/**< device statistics */
} ixUARTDev;

/**
 * @ingroup IxUARTAccAPI
 *
 * @fn IX_STATUS ixUARTInit(ixUARTDev* pUART)
 * 
 * @param pUART	- pointer to UART structure describing our device.
 * 
 * @brief Initialise the UART. This puts the chip in a quiescent state.
 * 
 * @return IX_SUCCESS - UART device successfully initialised.
 * @return IX_FAIL - Critical error, device not initialised.
 ***************************************************************************/
PUBLIC IX_STATUS ixUARTInit(ixUARTDev* pUART);

/**
 * @ingroup IxUARTAccAPI
 *
 * @fn IX_STATUS ixUARTPollOutput(ixUARTDev* pUART, int outChar)
 *
 * @param pUART - pointer to UART structure describing our device.
 * @param outChar - character to transmit.
 *
 * @brief Transmit a character in polled mode.
 *
 * @return IX_SUCCESS - character was successfully transmitted.
 * @return IX_FAIL - output buffer is full (try again).
 ***************************************************************************/
PUBLIC IX_STATUS ixUARTPollOutput(ixUARTDev* pUART, int outChar);

/**
 * @ingroup IxUARTAccAPI
 *
 * @fn IX_STATUS ixUARTPollInput(ixUARTDev* pUART, char *inChar)
 *
 * @param pUART	- pointer to UART structure describing our device.
 * @param *inChar - character read from the device.
 *
 * @brief Receive a character in polled mode.
 *
 * @return IX_SUCCESS - character was successfully read.
 * @return IX_FAIL - input buffer empty (try again).
 ***************************************************************************/
PUBLIC IX_STATUS ixUARTPollInput(ixUARTDev* pUART, char *inChar);

/**
 * @ingroup IxUARTAccAPI
 *
 * @fn IX_STATUS ixUARTIoctl(ixUARTDev* pUART, int cmd, void* arg)
 *
 * @param pUART	- pointer to UART structure describing our device.
 * @param cmd - an ioctl request code.
 * @param arg - optional argument used to set the device mode,
 *	 baud rate, and hardware options.
 *
 * @brief Perform I/O control routines on the device.
 *
 * @return IX_SUCCESS - requested feature was set/read successfully.
 * @return IX_FAIL - error setting/reading the requested feature.
 ***************************************************************************/    
PUBLIC IX_STATUS ixUARTIoctl(ixUARTDev* pUART, int cmd, void* arg);

/**
 * @} defgroup IxUARTAcc
 */
