/** 
 * @file:	 IxUARTAccCodelet.h
 *
 * @date	 16-OCT-01
 *
 * @brief This file contains the interface prototypes for Intel IXP4xx UART codelet.
 * 
 * -- Intel Copyright Notice --
 *
 * @par
 * INTEL CONFIDENTIAL
 *
 * @par
 * Copyright 2002 Intel Corporation All Rights Reserved.
 *
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.  The Material
 * contains trade secrets and proprietary and confidential information of
 * Intel or its suppliers and licensors.  The Material is protected by
 * worldwide copyright and trade secret laws and treaty provisions. No
 * part of the Material may be used, copied, reproduced, modified,
 * published, uploaded, posted, transmitted, distributed, or disclosed in
 * any way without Intel's prior express written permission.
 *
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel or otherwise.  Any license under
 * such intellectual property rights must be express and approved by
 * Intel in writing.
 *
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * -- End Intel Copyright Notice --
*
 */

/**
 * @defgroup Codelets IXP4xx Codelets
 *
 * @brief IXP4xx demo codelets
 *
 * @{
 */

/**
 * @ingroup Codelets
 *
 * @defgroup IxUARTAccCodelet IXP4xx UART Access Codelet (IxUARTAccCodelet) API
 *
 * @brief The interface prototypes for Intel IXP4xx UART codelet.
 *
 * <P>
 * The following scenarios are supported: 
 * <UL>
 * <LI> Device initialisation, setting up of baud rate, data size etc.
 * <LI> Character reception.
 * <LI> Character transmission.
 * <LI> S/W loopback function.
 * </UL>
 * You must have a PC with serial port supporting the baud rates here and
 * terminal software to view the data as it is passed to/from the UART.
 *    
 * @{
 */

#ifndef IXUARTACCCODELET_H
#define IXUARTACCCODELET_H

#define TRANSLATE_OUT_CHARS	/**< Map '\n' to '\r\n' on output */
#define TRANSLATE_IN_CHARS	/**< Map '\r' to '\n' on input */	

typedef enum _ixUARTAccPort
{
    IX_UART_PORT1 = 1,
    IX_UART_PORT2 = 2
} ixUARTAccPort;


/* Prototypes */

/**
 * @fn IX_STATUS ixUARTAccCodeletDevConfig(ixUARTAccPort uartNo)
 * 
 * @brief Configure a UART device.
 *
 * @param uartNo - Port number of UART to configure
 *
 * @return IX_SUCCESS - UART device succesfully initialised
 * @return IX_FAIL - error initialising UART device
 */

PUBLIC IX_STATUS ixUARTAccCodeletDevConfig(ixUARTAccPort uartNo);


/**
 * @fn ixUARTAccCodeletSwLoopback(ixUARTAccPort uartNo)
 * 
 * @brief Echo characters as they are received. Stops when a ^Z is received.
 *
 * @param pUART	- Port number of the UART we want to test.
 *
 * @return IX_SUCCESS - SW Loopback successfully started on selected device
 * @return IX_FAIL - error starting SW Loopback
 */

PUBLIC IX_STATUS ixUARTAccCodeletSwLoopback(ixUARTAccPort uartNo);


/**
 * @fn ixUARTAccCodeletWrite(ixUARTAccPort uartNo, char* buf)
 *
 * @brief Write to the serial device. This is a blocking call.
 * 
 * @param uartNo - Port number of UART to write to.
 * @param buf	- buffer containing characters to transmit.
 *
 * Write to the serial device. This is a blocking call.
 *
 * @return int	- number of characters actually written or IX_FAIL if
 * an unknown UART is selected. 
 */

PUBLIC int ixUARTAccCodeletWrite(ixUARTAccPort uartNo, char* buf);


/**
 * @fn ixUARTAccCodeletRead(ixUARTAccPort uartNo, char* buf, unsigned int count)
 *
 * @brief  Read from the serial device. This is a blocking call. The buffer 
   returned is NULL terminated.
 * 
 * @param uartNo - Port number of UART to read from.
 * @param buf	- buffer to place characters in.
 * @param count	- number of characters to read.
 *
 * @return int	- number of characters actually read or IX_FAIL is an unknown
 * UART is selected.
 */

PUBLIC int ixUARTAccCodeletRead(ixUARTAccPort uartNo, char* buf, unsigned int count);


/**
 * @fn ixUARTAccCodeletShow(ixUARTAccPort uartNo)
 * 
 * @brief Display statistics for the specified UART.
 *
 * @param uartNo - Port number of UART to display information for.
 *
 * @return IX_SUCCESS - Successfully display UART Statistics
 * @return IX_FAIL - error displaying UART statistics
 */

PUBLIC IX_STATUS ixUARTAccCodeletShow(ixUARTAccPort uartNo);


/**
 * @fn ixUARTAccCodeletBaudDemo(ixUARTAccPort uartNo)
 *
 * @brief Output a test string on the device at supported baud rates.
 *
 * @param uartNo - Port number of UART to test.
 *
 * @return IX_SUCCESS - Successfully running baud demo
 * @return  IX_FAIL - error running baud demo.
 */

PUBLIC IX_STATUS ixUARTAccCodeletBaudDemo(ixUARTAccPort uartNo);

#endif /* IXUARTACCCODELET_H */

/**@} defgroup IxUARTAccCodelet*/

/**@} defgroup Codelets*/
