/**
 * @file usbdeviceparam.h
 *
 * @author Intel Corporation
 * @date 30-OCT-2001
 *
 * @brief This file containes the USB device parameters used by the USB driver
 *
 * Device parameters:
 * - base address, number of endpoints, FIFO and packet sizes are UDC-dependent
 * - DMA channel buffer sizes can be customized and depend on the DMA controller
 *
 * -- Intel Copyright Notice --
 *
 * @par
 * INTEL CONFIDENTIAL
 *
 * @par
 * Copyright 2001-2002 Intel Corporation All Rights Reserved.
 *
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.  The Material
 * contains trade secrets and proprietary and confidential information of
 * Intel or its suppliers and licensors.  The Material is protected by
 * worldwide copyright and trade secret laws and treaty provisions. No
 * part of the Material may be used, copied, reproduced, modified,
 * published, uploaded, posted, transmitted, distributed, or disclosed in
 * any way without Intel's prior express written permission.
 *
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel or otherwise.  Any license under
 * such intellectual property rights must be express and approved by
 * Intel in writing.
 *
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * -- End Intel Copyright Notice --
 */

/**
 * @addtogroup DeviceParameters
 *
 * @brief Device parameters for the IXP425 USB Controller
 *
 * @{
 */

#ifndef usbdeviceparam_H

#ifndef __doxygen_HIDE

#define usbdeviceparam_H

#endif /* __doxygen_HIDE */

#include "ixp425.h"
#include "ixp425IntrCtl.h"

/** Base I/O address */
#define UDC_REGISTERS_BASE IXP425_USB_BASE

/** IRQ */
#define UDC_IRQ IXP425_INT_LVL_USB

/** Number of endpoints */
#define NUM_ENDPOINTS 16

/** SETUP packet size */
#define SETUP_PACKET_SIZE 8

/** CONTROL endpoint FIFO depth */
#define CONTROL_FIFO_SIZE 16

/** CONTROL endpoint packet size */
#define CONTROL_PACKET_SIZE 16

/** INTERRUPT endpoint FIFO depth */
#define INTERRUPT_FIFO_SIZE 8

/** INTERRUPT endpoint packet size */
#define INTERRUPT_PACKET_SIZE 8

/** BULK endpoint FIFO depth */
#define BULK_FIFO_SIZE 64

/** BULK endpoint packet size */
#define BULK_PACKET_SIZE 64

#ifdef IX_USB_DMA

/** BULK DMA channel buffer size */
#define BULK_DMA_SIZE 512

#endif /* IX_USB_DMA */

/** ISOCHRONOUS endpoint FIFO depth */
#define ISOCHRONOUS_FIFO_SIZE 256

/** ISOCHRONOUS endpoint packet size */
#define ISOCHRONOUS_PACKET_SIZE 256

#ifdef IX_USB_DMA

/** ISOCHRONOUS DMA channel buffer size */
#define ISOCHRONOUS_DMA_SIZE 256

#endif /* IX_USB_DMA */

/** Endpoint descriptions */
static int EPDescriptorTable[] =
{
    /* EP_0 */  USB_CONTROL     | USB_IN_OUT,
    /* EP_1 */  USB_BULK        | USB_IN,
    /* EP_2 */  USB_BULK        | USB_OUT,
    /* EP_3 */  USB_ISOCHRONOUS | USB_IN,
    /* EP_4 */  USB_ISOCHRONOUS | USB_OUT,
    /* EP_5 */  USB_INTERRUPT   | USB_IN,
    /* EP_6 */  USB_BULK        | USB_IN,
    /* EP_7 */  USB_BULK        | USB_OUT,
    /* EP_8 */  USB_ISOCHRONOUS | USB_IN,
    /* EP_9 */  USB_ISOCHRONOUS | USB_OUT,
    /* EP_10 */ USB_INTERRUPT   | USB_IN,
    /* EP_11 */ USB_BULK        | USB_IN,
    /* EP_12 */ USB_BULK        | USB_OUT,
    /* EP_13 */ USB_ISOCHRONOUS | USB_IN,
    /* EP_14 */ USB_ISOCHRONOUS | USB_OUT,
    /* EP_15 */ USB_INTERRUPT   | USB_IN
};

/**  Endpoint interrupt sources on UISR0 and UISR1 */
static int UDCEPInterrupt[] = 
{
/* UISR0 */
    /* EP_0 */  (0x1 << 0),
    /* EP_1 */  (0x1 << 1),
    /* EP_2 */  (0x1 << 2),
    /* EP_3 */  (0x1 << 3),
    /* EP_4 */  (0x1 << 4),
    /* EP_5 */  (0x1 << 5),
    /* EP_6 */  (0x1 << 6),
    /* EP_7 */  (0x1 << 7),
/* UISR1 */
    /* EP_8 */  (0x1 << 0),
    /* EP_9 */  (0x1 << 1),
    /* EP_10 */ (0x1 << 2),
    /* EP_11 */ (0x1 << 3),
    /* EP_12 */ (0x1 << 4),
    /* EP_13 */ (0x1 << 5),
    /* EP_14 */ (0x1 << 6),
    /* EP_15 */ (0x1 << 7)
};

#endif /* usbdeviceparam_H */

/**
 * @} addtogroup DeviceParameters
 */
