/**
 * @file   IxFpathAccCodeletMain.c
 * @date   May 29 2002
 * @brief  codelet main connection management.
 *
 * -- Intel Copyright Notice --
 * 
 * @par
 * Copyright 2002-2003 Intel Corporation All Rights Reserved.
 * 
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.
 * 
 * @par
 * The Material is protected by worldwide copyright and trade secret laws
 * and treaty provisions. No part of the Material may be used, copied,
 * reproduced, modified, published, uploaded, posted, transmitted,
 * distributed, or disclosed in any way except in accordance with the
 * applicable license agreement .
 * 
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel, except in accordance with the
 * applicable license agreement.
 * 
 * @par
 * Unless otherwise agreed by Intel in writing, you may not remove or
 * alter this notice or any other notice embedded in Materials by Intel
 * or Intel's suppliers or licensors in any way.
 * 
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * 
 * @par
 * -- End Intel Copyright Notice --
 */
#include <stdio.h>

#include "IxTypes.h"
#include "IxFpathAcc.h"
#include "IxOsServices.h"
#include "IxOsCacheMMU.h"
#include "IxAtmdAcc.h"
#include "IxAtmTypes.h"
#include "IxEthAcc.h"
#include "ix_ossl.h"


#include "IxFpathAccCodelet.h"
#include "IxFpathAccCodeletDefines_p.h"
#include "IxAssert.h"

/** 
 * Typedefs for this file.
 */

/* AAl5 Trailer, note endianness - network byte order is assumed */
typedef struct {
    UINT8 uui;          
    UINT8 cpi;          
    UINT16 length;      
    UINT32 crc;
} Aal5Trailer;


/** Forward Declaration */
void
ixFpCodeletConnectionDatabaseInit (void);

/*
 * Variables
 */

/** Connection Database - indexed by VcNumber */
IxFpCodeletFastPathConnection
    ixFpCodeletFastPathConnection[MAX_NUM_VCS];

/** Number of times the fast path buffer pool went empty */
unsigned int
    ixFpCodeletUnderRunCallbackCount = 0;

/** Flag to enable/disable trace */
BOOL ixFpCodeletTraceEnabled = FALSE;

/** Is the main codelet initialisation complete */
static BOOL
    fpathAccCodeletInitComplete = FALSE;

/* static prototypes */
PRIVATE void ixFpCodeletBufferPoolUnderRunCallback( IxFpathAccClientId clientId );
PRIVATE void ixFpCodeletAtmRxCallback(IxAtmLogicalPort port,
					IxAtmdAccUserId userId,
    					IxAtmdAccPduStatus status,
					IxAtmdAccClpStatus clp,
					IX_MBUF * buf);
PRIVATE void ixFpCodeletVcShow( VcNumber vc );

/* An arbitary number used to identify the application on the
 * fast path underrun callback API
 */
#define CLIENT_ID 99


/** 
 * @brief Buffer Pool Under-run handler
 *
 * This function illustrates an ixFpathAcc buffer pool under-run
 * callback.
 *
 * This callback is an abnormal event that occurs when the fast path 
 * buffer pool has been exhausted and inidcates that NPE-B is not
 * processing fast path PDUs as fast as NPE-A is accepting them.
 * 
 * After Initialization and initial buffer supply, the fast path buffer 
 * pool is replenished transparently to the user ( normally by NPE-B ).
 * Unlike other access layer components where under-run callbacks
 * may be used to drive a dynamic replenlish process, in fast path
 * this event is informative only.
 *
 * In this example code the event is simply logged and counted.
 */
PRIVATE void
ixFpCodeletBufferPoolUnderRunCallback( IxFpathAccClientId clientId )
{
    static unsigned int logCntr = 0;
    
    if ( logCntr++%1000 == 0 ) 
    {
        IXFPATH_CODELET_TRACE(( traceBuf, 
                    "IxFpCodelet: BufferPool UnderRun - clientId %d, Callback Cnt %i",
                    clientId,
                    logCntr ));
    }
    
    ixFpCodeletUnderRunCallbackCount++;
}

/**
 * @brief Initialise the codelet
 *
 * This function illustrates initialization of the ixFpathAcc
 * component.
 * 
 * ixFpathAcc should be initialized after the initialization of 
 * the subcomponents that are managed by ixFpCodeletUtilsInit().
 */
IX_STATUS
ixFpCodeletInit( void )
{
    IX_STATUS status = IX_FAIL;
    IX_MBUF *buf = NULL;
    int maxEntries;
    int count;

    if ( fpathAccCodeletInitComplete ) 
    {
        return IX_SUCCESS;
    }

    /* Format the connection database */
    ixFpCodeletConnectionDatabaseInit();

    /* Inialise supporting components: buffer pool, template utils, 
     * Atm, Eth, Npe's, and related subcomponents. 
     * This must be done before fpathAcc is initialized.
     */
    if(( ixFpCodeletBufManMbufPoolInit () != IX_SUCCESS ) ||
       ( ixFpCodeletTemplatesInit () != IX_SUCCESS ) ||
        /* Now initialise all other s/w components needed for this codelet.*/
       ( ixFpCodeletUtilsInit () != IX_SUCCESS ))
    {
        IXFPATH_CODELET_ERROR("Failed to initialize subcomponents");
        return IX_FAIL;
    }


    /* Now initialise the fpathAcc component */
    if( ixFpathAccInit() != IX_SUCCESS )
    {
        IXFPATH_CODELET_ERROR("ixFpathAccInit() Failed");
        return IX_FAIL;
    }


    /* Before starting to use the fast path the fast path
     * buffer pool must be populated, this should be done
     * after the component initialization
     */

    /* To find out the maximum number of mbufs that the pool
     * can hold, we can supply a NULL mbuf. The returned
     * value is the space remaining in the pool. After
     * initialization this will be pool size.
     */
    if( ixFpathAccBufferSupply (NULL, &maxEntries) != IX_SUCCESS )
    {
        IXFPATH_CODELET_ERROR("ixFpathAccBufferSupply() query Failed");
        return IX_FAIL;
    }

    /* Set a callback to be called if the buffer pool becomes
     * exhausted, this is an error
     */
    if( ixFpathAccBufferPoolUnderRunCallbackSet (
            ixFpCodeletBufferPoolUnderRunCallback,
            CLIENT_ID ) != IX_SUCCESS )
    {
        IXFPATH_CODELET_ERROR("ixFpathAccBufferPoolUnderRunCallbackSet() Failed");
        return IX_FAIL;
    }

    /* For the purposes of this codelet we
     * will completely fill the fast path pool with mbufs.
     * Note: in a real application the number of entries may
     * take into account the expected fast path traffic rate
     * and could be configured to less than the maximum.
     * Note: the API allows the user to supply, in any 
     * combination, a sequence of individual mbufs or mbuf chains. 
     * This example illustrates a sequence of individual mbufs.
     */
    count = maxEntries;
    while( count > 0 )
    { 
        buf = ixFpCodeletBufManMbufGet (FASTPATH_MBUF_SIZE);
        if( buf == NULL )
        {
            break;
        }

        if( ixFpathAccBufferSupply (buf, &count) != IX_SUCCESS )
        {
            break;
        }
         
    }

    if((count > 0 ) || (buf == NULL))
    {
        IXFPATH_CODELET_ERROR("ixFpathAccBufferSupply() Failed");
        return status;
    }


    /* At any time it is possible to query the fast path pool
     * to determine the size and level to which it has been
     * populated. We do this here to verify that the pool has
     * indeed been filled.
     */
    status = ixFpathAccBufferPoolInfoGet (&maxEntries, &count);
    if((status != IX_SUCCESS) || (count != maxEntries ))
    {
        IXFPATH_CODELET_ERROR("ixFpathAccBufferPoolInfoGet() Failed");
        return status;
    }

    fpathAccCodeletInitComplete = TRUE;

    return IX_SUCCESS;
}

/** 
 * @brief Handle received ATM PDUs
 *
 * This function implements an AAL5 receive function that can be
 * registered with AtmdAcc.
 *
 * This implementation serves only to show ixFpathAcc APIs
 * For a more complete illustration of ATM receive handling please refer 
 * to the aal5Codelet, and ethAal5App codelets
 *
 * Assumptions:-
 *
 *      The AAL5 pdu including trailer is completely contained
 *      within one mbuf, i.e. no mbuf chaining is supported.
 *
 * @param userId - configured to be the VcNumber (connection No.)at connect().
 * @param status - pdu status as returned by IxAtmdAcc.
 * @param clp    - cell loss priority status.
 * @param buf    - pointer to an mbuf containing a single Aal5 Pdu only.
 */
PRIVATE void
ixFpCodeletAtmRxCallback(
    IxAtmLogicalPort port,
    IxAtmdAccUserId userId,
    IxAtmdAccPduStatus status,
    IxAtmdAccClpStatus clp,
    IX_MBUF * buf)
{
    Aal5Trailer *trailer;
    IxFpCodeletFastPathConnection *connection;
    unsigned int sduLen;

    if((userId >= MAX_NUM_VCS) ||
       (buf == NULL ))
    {
        IXFPATH_CODELET_ERROR("Bad parameters");
        return;
    }

    if( ixFpCodeletTraceEnabled )
    {
        ixOsServLog ( 
            LOG_MESSAGE,
            "Atm Pdu received "
            "On UserId = %8X "
            "mbuf = %8X "
            "status = %d "
            "clp = %d ",
            (int) userId,
            (int) buf,
            (int) status,
            (int) clp,
            0,0 ); 
    }

    IX_ACC_DATA_CACHE_INVALIDATE(IX_MBUF_MDATA(buf), IX_MBUF_MLEN(buf));

    connection = ixFpCodeletFastPathConnectionGet ( userId );

    /* count stats */
    connection->aal5Stats.totalPduCount++;
 
    /* Check the buffer status
     */
    if( status == IX_ATMDACC_AAL5_VALID )
    { 
        /* keep statistics */
        connection->aal5Stats.validPduCount++;

        /* Strip off the Aal5 trailer, we get the aal5 sdu
         * size from the trailer, and adjust the mbuf length 
         * to this value.
         */
    
        /* Locate the trailer */
        trailer = (Aal5Trailer *) (IX_MBUF_MDATA(buf) + 
                                   IX_MBUF_MLEN(buf) - 
                                   sizeof(Aal5Trailer));
    
        /* extract the trailer */
        sduLen =  trailer->length;
    
        /* Adjust the mbuf length to this value */
        IX_MBUF_MLEN(buf)    = sduLen;
        IX_MBUF_PKT_LEN(buf) = sduLen;
    
        /* Now we pass the PDU up to the application layer
         * Note in this example code, the application layer will
         * be ixFpCodeletAtmRx_EthernetBridge(),
         * or ixFpathAccCodeleAtmRxRoutedIP()
         */
        (*connection->atmRxHandler) (
                    userId,
                    buf);
    }
    else
    {
        /* In this FpathAcc codelet we simply discard 
         * invalid PDUs.
         */
        connection->aal5Stats.errorCount++;
        ixFpCodeletBufManMbufFree (buf);
    }
}

/**
 * @brief Fast path VC create
 *
 * This function illustrates the creation of a VC.
 *
 * Note that function ixFpCodeletFastPathVcDelete() illustrates
 * the correct sequence of steps to be observed when disconnecting 
 * a fast path enabled VC.
 * 
 * @param vc - a codelet connection number.
 */
IX_STATUS
ixFpCodeletFastPathVcCreate( VcNumber vc )
{
    IxFpCodeletFastPathConnection *connection  = 
        &ixFpCodeletFastPathConnection[vc];

    /* Establish an ATM connection, specify the Utopia port, vpi, vci,
     * the ATM receive queue, register the Atm receive callback, and
     * the utility replenish callback
     */
    if( ixFpCodeletUtilsAtmRxVcRegisterConnect ( 
            connection->port,
            connection->vpi,
            connection->vci,
            IX_ATM_RX_A,
            ixFpCodeletAtmRxCallback,
            connection->vc,
            &connection->npeVcId ) != IX_SUCCESS )
    {
       IXFPATH_CODELET_ERROR("failed to connect on VC");
       return IX_FAIL;
    }

    connection->inUse = TRUE;
    return IX_SUCCESS;
} 


/** 
 * @brief Delete a fast path connection
 *
 * This function illustrates the correct sequence for disconnecting
 * a fast path VC.
 *
 * The following sequence MUST always be observed:-
 *
 *      1) The fast path is shut down
 *      2) The ATM VC is disconnected.
 * 
 * In this example function the order is guaranteed by the sequential
 * nature of the code.
 *
 * If the ATM connections and fast path configuration are to be conrolled 
 * in different OS threads, then a synchonization mechanism must be used 
 * to ensure this order.
 *
 * @param vc - a codelet connection number.
 */
IX_STATUS
ixFpCodeletFastPathVcDelete( VcNumber vc )
{
    IxFpCodeletFastPathConnection *connection  = 
        &ixFpCodeletFastPathConnection[vc];

    /* Shut down Fast Path first */
    if( ixFpathAccTemplatePairClear (connection->npeVcId) != IX_SUCCESS ) 
    {
        IXFPATH_CODELET_ERROR("Fast Path failed to shutdown");
        return IX_FAIL;
    }

    /* Disconnect the ATM VC afterwards */
    if( ixFpCodeletUtilsAtmRxDisconnect (connection->vc) != IX_SUCCESS )
    {
        IXFPATH_CODELET_ERROR("ATM VC failed to disconnect");
        return IX_FAIL;
    }
    
    connection->port  = 0;
    connection->vpi   = 0;
    connection->vci   = 0;
    connection->inUse = FALSE;

    return IX_SUCCESS;
}

/** 
 * @brief Initialize the codelet connection data base
 */
void
ixFpCodeletConnectionDatabaseInit( void )
{
    int vc;
    int flow;

 
    memset(ixFpCodeletFastPathConnection, 0, sizeof(ixFpCodeletFastPathConnection));

    for( vc=0; vc<MAX_NUM_VCS; vc++)
    /* reset all connections */
    {
       ixFpCodeletFastPathConnection[vc].vc = vc;                
       ixFpCodeletFastPathConnection[vc].encapsulationType = TYPE_NONE; 
       ixFpCodeletFastPathConnection[vc].npeVcId = IX_ATM_MAX_NUM_VC+1;
       for( flow=0; flow<MAX_FLOWS_PER_VC; flow++)
       {
           ixFpCodeletFastPathConnection[vc].activeIpFlow[flow].aged = TRUE;
       }
    }
}

/** 
 * @brief Get the connection data for a VC
 *
 * @param vc - the connection id of interest.
 */
IxFpCodeletFastPathConnection *
ixFpCodeletFastPathConnectionGet( VcNumber vc )
{
    return &ixFpCodeletFastPathConnection[ vc ];
}

/** 
 * @brief Show information about codelet configuration
 *
 * This function displays the configuration and statistics for a VC.
 * It also show the use of the ixFpathAccStatisticsGet API function
 *
 * @param vc - the connection id of interest.
 */
PRIVATE void
ixFpCodeletVcShow( VcNumber vc )
{
    BOOL fpathStatsAvailable;
    IxFpCodeletFastPathConnection *connection;
    int fpathPktCnt;
    int i;

    connection = ixFpCodeletFastPathConnectionGet (vc);

    if( connection->inUse )
    {
        printf("\n\n=== Fast Path Codelet Statisics ==============================================");
        printf("\n%2.2s %4s %3s %3s %7s %20s %15s %15s ", 
                 "VC", 
                 "Port", 
                 "Vpi",
                 "Vci",
                 "NpeVcId",  
                 "Encapsulation",
                 "Slow Path Pkts", 
                 "Fast Path Pkts");
    
        fpathStatsAvailable = FALSE;
        fpathPktCnt = 0;
    
        /* get fast path statistics for this VC */
        if ( connection->inUse ) 
        {
            if ( ixFpathAccStatisticsGet( connection->npeVcId, &fpathPktCnt)
                 == IX_SUCCESS )
            {
                fpathStatsAvailable = TRUE;
            }
        }
        printf("\n%2.2i %4i %3i %3i %7i %20s %15.15i ",
            vc,
            connection->port,
            connection->vpi,
            connection->vci,
            connection->npeVcId,
            ixFpCodeletTemplatesEncapuslationName(connection->encapsulationType),
            connection->aal5Stats.totalPduCount );
            
        fpathStatsAvailable ? 
            printf("%15.15i", fpathPktCnt ) : printf(" NOT AVAILABLE ");
    

        printf("\n=== AAL5 statistics ==========================================================");
        printf("\ntotal PDUs      Valid PDUs      Error PDUs");
        printf("\n%15.15i %15.15i %15.15i",
               connection->aal5Stats.totalPduCount,
               connection->aal5Stats.validPduCount,
               connection->aal5Stats.errorCount );

        
        printf("\n--- Encapsulation specific statistics ----------------------------------------");
        switch ( connection->encapsulationType )
        {
        case TYPE_ROUTEDIPVCMUX:
            printf("\nEthTx Errors    Flow Set Errors Tmpl Set Errors MAC lookup Errors");
            printf("\n%15.15i %15.15i %15.15i %15.15i",
                   connection->routedIpStats.ethernetTxErrorCount, 
                   connection->routedIpStats.flowSetErrorCount ,
                   connection->routedIpStats.templateSetErrorCount,
                   connection->routedIpStats.macLookupErrorCount );
            printf("\n--- Learning IP Flows --------------------------------------------------------");
            printf("\n      IP             MAC        Count      In Use" );
        
            for ( i=0; i<NUM_LEARNING_FLOWS; i++) 
            {
                printf("\n%3d.%3d.%3d.%3d  %2.2X:%2.2X:%2.2X:%2.2X:%2.2X:%2.2X %15.15i %4s",
                (connection->learningIpFlow[i].ipAddress >> 24) & 0xff,
                (connection->learningIpFlow[i].ipAddress >> 16) & 0xff,
                (connection->learningIpFlow[i].ipAddress >> 8) & 0xff,
                (connection->learningIpFlow[i].ipAddress >> 0) & 0xff,
                (connection->learningIpFlow[i].macAddress[0]),
                (connection->learningIpFlow[i].macAddress[1]),
                (connection->learningIpFlow[i].macAddress[2]),
                (connection->learningIpFlow[i].macAddress[3]),
                (connection->learningIpFlow[i].macAddress[4]),
                (connection->learningIpFlow[i].macAddress[5]),
                (connection->learningIpFlow[i].counter),
                (connection->learningIpFlow[i].inUse? "Yes":"No"));
            }
        
            printf("\n--- Active Fast Path IP Flows ------------------------------------------------");
            printf("\n      IP             MAC        Count      In Use" );
        
            for ( i=0; i<MAX_FLOWS_PER_VC; i++) 
            {
                printf("\n%3d.%3d.%3d.%3d  %2.2X:%2.2X:%2.2X:%2.2X:%2.2X:%2.2X %15.15i %4s",
                (connection->activeIpFlow[i].ipAddress >> 24) & 0xff,
                (connection->activeIpFlow[i].ipAddress >> 16) & 0xff,
                (connection->activeIpFlow[i].ipAddress >> 8) & 0xff,
                (connection->activeIpFlow[i].ipAddress >> 0) & 0xff,
                (connection->activeIpFlow[i].macAddress[0]),
                (connection->activeIpFlow[i].macAddress[1]),
                (connection->activeIpFlow[i].macAddress[2]),
                (connection->activeIpFlow[i].macAddress[3]),
                (connection->activeIpFlow[i].macAddress[4]),
                (connection->activeIpFlow[i].macAddress[5]),
                (connection->activeIpFlow[i].age),
                (connection->activeIpFlow[i].aged? "Yes":"No"));
            }
            break;

        case TYPE_1483BVCMUX:
            printf("\nLength Rejects  Other Rejects   New WAN MAC");
            printf("\n%15.15i %15.15i %15.15i",
                   connection->etherBridgeStats.lengthExceptionCount,
                   connection->etherBridgeStats.otherExceptionCount,
                   connection->etherBridgeStats.newWanMacCount );
        break;
        default:
            IXFPATH_CODELET_ERROR("Unsupported encapsulation type");
        break;        
        }
    }
    printf("\n");
}

/** 
 * @brief Show information about codelet configuration for all VCs
 */
void
ixFpCodeletShow(void)
{
    unsigned int vc;
    for ( vc=0; vc<MAX_NUM_VCS; vc++ ) 
    {
      ixFpCodeletVcShow( vc );
    }
}


PUBLIC void
ixFpAccCodeletMain( IxFpathCodeletType type )
{
UINT32 wait_count;

wait_count = IX_FPATHACC_CODELET_RUN_IN_SEC;    

switch (type)
{
	case ETHERNET_BRIDGE:
      {
	 /*Run the Fastpath Ethernet Bridge Codelet*/ 	    
         ixFpCodeletEthBStart();
           
	     /* Interval period to run the codelet before
              calling the ixFpCodeletEthBStop function
	     */
	     while (wait_count--)	
           {
		 ixOsServTaskSleep(IX_FPATHACC_SLEEP_IN_S);
           } 
	 
	  /* Show fastpath statistic before closing the link */ 
          ixFpCodeletShow();
	 
          /*Stop the Fastpath Ethernet Bridge Codelet*/ 	    
          ixFpCodeletEthBStop();
          break;
      }

	 case ROUTED_IP:
       {
	     /*Run the Fastpath Routed IP Codelet*/ 	    
           ixFpCodeletRoutedIpStart();
           
	     /* Interval period to run the codelet before
              calling the ixFpCodeletRoutedIpStop function
	     */
	     while (wait_count--)	
           {
		 ixOsServTaskSleep(IX_FPATHACC_SLEEP_IN_S);
           } 
	  
	  /* Show the Fastpath Statistic before closing the link */  
          ixFpCodeletShow();
	  
          /*Stop the Fastpath Routed IP Codelet*/ 	    
          ixFpCodeletRoutedIpStop();
          break;
       }

	 default:
	     IX_ASSERT(0);
          break;
  }

	 
}	









