/**
 * @file  IxFpathAccCodeletRoutedDemo.c
 * @date  June 24 2002
 * @brief This file contains the routed VC fast path demo.  
 *
 * This code provides a codelet entry point to demonstrate how to setup 
 * a fast path stream from ATM to Ethernet to support routing of 
 * IP packets. 
 *
 * -- Intel Copyright Notice --
 * 
 * @par
 * Copyright 2002-2003 Intel Corporation All Rights Reserved.
 * 
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.
 * 
 * @par
 * The Material is protected by worldwide copyright and trade secret laws
 * and treaty provisions. No part of the Material may be used, copied,
 * reproduced, modified, published, uploaded, posted, transmitted,
 * distributed, or disclosed in any way except in accordance with the
 * applicable license agreement .
 * 
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel, except in accordance with the
 * applicable license agreement.
 * 
 * @par
 * Unless otherwise agreed by Intel in writing, you may not remove or
 * alter this notice or any other notice embedded in Materials by Intel
 * or Intel's suppliers or licensors in any way.
 * 
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * 
 * @par
 * -- End Intel Copyright Notice --
 *
 * @sa IxFpathAccCodelet.h
 * @sa IxFpathAccCodeletDefines_p.h
 */

#include "IxTypes.h"
#include "IxAssert.h"
#include "IxAtmTypes.h"
#include "IxOsServices.h"
#include "IxFpathAcc.h"
#include "IxEthAcc.h"
#include "ix_ossl.h"
#include "IxOsCacheMMU.h"
#include "IxFpathAccCodeletMain_p.h"
#include "IxFpathAccCodeletBufMan_p.h"

/*
 * Local Variables
 */
ix_ossl_thread_t ixFpCodeletIpFlowAgeingTaskId;

/*
 * Forward Declarations
 */
PRIVATE IX_STATUS
ixFpCodeletRoutedIpVcCreate( 
    VcNumber vc, 
    IxAtmLogicalPort port,
    unsigned int vpi,
    unsigned int vci );

PRIVATE void  
ixFpCodeletAtmRxRoutedIP(
    IxAtmdAccUserId userId,
    IX_MBUF * buf);

PRIVATE IX_STATUS
ixFpCodeletDestMacAddressGet( UINT32 destIPaddress,  
                                    UINT8 *destMacAddress );

PRIVATE ix_error
ixFpCodeletIpFlowAgeingTask(
    void* arg,
    void **ptrRetObj);

PRIVATE void
ixFpCodeletIpFlowLearn( 
    IxAtmdAccUserId userId,
    UINT32 currentIpAddress,
    UINT8 *macAddress );


/** Interval of activitiy of an IP flow after which the flow is aged.*/
static unsigned int ixFpCodeletFlowAgeingInterval = 
    (IP_FLOW_AGEING_INTERVAL_MS / IP_FLOW_ACCOUNTING_INTERVAL_MS );

/** The threshold number of times a flow is seen before it is learned.*/
static unsigned int ixFpCodeletFlowLearningThreshold = 
    IP_FLOW_LEARNING_PDU_COUNT_THRESHOLD;

/** The period during which the IP flows are evaluated for learning */
static unsigned int ixFpCodeletFlowAccountingInterval = 
    IP_FLOW_ACCOUNTING_INTERVAL_MS;


/** The src mac address to be used for generated Ethernet headers */
static UINT8 sourceMacAddress[IX_IEEE803_MAC_ADDRESS_SIZE] = ETH1_MAC_ADDRESS;
static BOOL taskEnabled = TRUE;
static BOOL etherTypeSet = FALSE;


/** 
 * @brief Routed IP Demo Start
 *
 * This function illustrates the creation of a fast path VC 
 * for a Routed IP application.
 * See ixFpCodelet.h.
 */
PUBLIC void
ixFpCodeletRoutedIpDemoStart( void )
{
    /* First initialize the codelet */
    if( ixFpCodeletInit () != IX_SUCCESS )
    {
        IXFPATH_CODELET_ERROR("Failed to initialize codelet");
        return;
    }

    /* 
     *  Set the Ethernet type to be used for all Ethernet encapsulated IP packets
     *  forwarded on the Fast Path. This only needs to be done once per Ethernet
     *  port. 
     *
     *  NOTE - The only valid TherType value is 0x0800 for now since the NPE fast
     *         path classifier and modifiers engines only support IPv4.
     */
    if (!etherTypeSet) 
    {
        ixFpathAccEtherTypeSet( IX_FPATHACC_ETHERNET_PORT0, 0x0800);
        etherTypeSet = TRUE;
    }

    /* Create an ATM VC and configure it for fast path */ 
    if( ixFpCodeletRoutedIpVcCreate ( 
            VC_1, 
            FPATH_ROUTED_TEST_UTOPIA_PORT,
            FPATH_ROUTED_TEST_VPI,
            FPATH_ROUTED_TEST_VCI ) != IX_SUCCESS )
    {
        IXFPATH_CODELET_ERROR("Failed to establish fast path");
        return;
    }
   
    taskEnabled = TRUE;

    /* Create a task to perform the IP flow Learning/Ageing
     * function
     */
    if( ix_ossl_thread_create (
            ixFpCodeletIpFlowAgeingTask,
            NULL,
            &ixFpCodeletIpFlowAgeingTaskId) != IX_OSSL_ERROR_SUCCESS)
    {
        IXFPATH_CODELET_ERROR("Failed to create flow learning task");
    }
}

/** 
 * @brief Routed IP Demo Stop
 *
 * This function illustrates the correct steps for deletion of a fast path VC 
 * for a Routed IP application. See IxFpathAccCodelet.h.
 */
PUBLIC void
ixFpCodeletRoutedIpDemoStop( void )
{
    VcNumber vc;
    IxFpCodeletFastPathConnection *connection;
    for( vc=0; vc<MAX_NUM_VCS; vc++ )
    {
        connection = ixFpCodeletFastPathConnectionGet (vc);
        
        if( connection->inUse &&
            (connection->encapsulationType == TYPE_ROUTEDIPVCMUX ))
        {
            IXFPATH_CODELET_TRACE((traceBuf,"Shutting Down VC %d\n",vc));
            if( ixFpCodeletFastPathVcDelete(vc) == IX_SUCCESS ) 
            {
                IXFPATH_CODELET_TRACE((traceBuf,"Done\n"));
            }
            else
            {
                IXFPATH_CODELET_ERROR("ixFpCodeletFastPathVcDelete - Failed");
            }
        }    
    }

    /* Kill the task */
    taskEnabled = FALSE;
}


/** 
 * @brief Routed IP VC Create
 *
 * This function illustrates the creation of a fast path VC 
 * for a Routed IP application.
 *
 * @param vc   - demo connection number.
 * @param port - The ATM PHY logical port number.
 * @param vpi  - The virtual path index for the ATM VC
 * @param vci  - The virtual channel index for the ATM VC
 */
PRIVATE IX_STATUS
ixFpCodeletRoutedIpVcCreate( 
    VcNumber vc, 
    IxAtmLogicalPort port,
    unsigned int vpi,
    unsigned int vci )
{
    IxFpCodeletFastPathConnection *connection;

    connection = ixFpCodeletFastPathConnectionGet(vc);

    /* Configure the codelet for the routed IP demo application
     * set the encapsulation type and the rx handler
     */
    connection->encapsulationType = TYPE_ROUTEDIPVCMUX;
    connection->atmRxHandler      = ixFpCodeletAtmRxRoutedIP;
    connection->port  = port;
    connection->vpi   = vpi;
    connection->vci   = vci;
    connection->vc    = vc;

    /* Create an ATM VC and configure it for fast path */ 
    if( ixFpCodeletFastPathVcCreate ( vc ) != IX_SUCCESS )
    {
        IXFPATH_CODELET_ERROR("Failed to establish fast path");
        return IX_FAIL;
    }

    /* Create a base template pair */
    ixFpCodeletTemplatesBaseTemplatePairCreate (
        connection->vc,
        connection->encapsulationType );

    /* The templates are not specified to the NPEs until a IP flow is learned */
    return IX_SUCCESS;
}

/** 
 * @brief Handler for received ATM PDUs for the Routed IP application
 *
 * This function illustrates an ATM receive handler for an
 * routed IP application.
 *
 * In the routed IP application packets may flow on both the fast path, 
 * and the "so called" slow path simultaniously. In this application
 * the higher layer protocols take care of packet ordering considerations.
 *
 * Packets received on the Atm interface must simply be re-transmitted 
 * on the Ethernet interface. Before doing this (depending on the
 * application) it may be necessary to remove an RFC 1483 header.
 * For all applications an ethernet header must be pre-appended.
 * 
 * Note this demonstration code supports only a a routed IP
 * VC mux application.
 *
 * @param userId - Expected to be the VcNumber of the demo connection.
 * @param mbuf   - pointer to an mbuf containing an Aal5 SDU.
 */
PRIVATE void  
ixFpCodeletAtmRxRoutedIP(
    IxAtmdAccUserId userId,
    IX_MBUF * buf)
{
    UINT32 destIpAddress;
    IxFpCodeletFastPathConnection *connection;

    connection = ixFpCodeletFastPathConnectionGet ((VcNumber) userId);

    /* depending on the application the header is modified */
    IX_ASSERT(connection->encapsulationType == TYPE_ROUTEDIPVCMUX );

    /* Note if your mbufs are allocated from non cached memory
     * then this macro should be defined to do nothing 
     */
    IX_ACC_DATA_CACHE_INVALIDATE(IX_MBUF_MDATA(buf), IX_MBUF_MLEN(buf) );

    /* get dest IP address */
    destIpAddress = *( UINT32 *) &IX_MBUF_MDATA(buf)[IP_DEST_ADDR_OFFSET];

    /* Add a dummy ethernet header , space for this is
     * guaranteed by the define MBUF_HEADROOM 
     * see IxFpathAccCodeletDefines_p.h 
     */
    IX_MBUF_MDATA(buf)       -= SIZEOF_IEEE802_3_HDR;
    IX_MBUF_MLEN(buf)        += SIZEOF_IEEE802_3_HDR; 
    IX_MBUF_PKT_LEN(buf)     += SIZEOF_IEEE802_3_HDR;

    /* set the dest Mac Address, we pass the IP and a pointer
     * to the PDU field this function returns the MAC directly 
     * into the PDU 
     */
    if( ixFpCodeletDestMacAddressGet ( 
            destIpAddress,
            (UINT8 *)( IX_MBUF_MDATA(buf) + IEEE802_3_DEST_MAC_ADDR_OFFSET ))
       != IX_SUCCESS )
    {
        IXFPATH_CODELET_ERROR("could not get dest MAC address");
        connection->routedIpStats.macLookupErrorCount++;
        ixFpCodeletBufManMbufFree (buf);
        return;
    }

    /* set the source Mac Address */
    memcpy (
       ( IX_MBUF_MDATA(buf) + IX_IEEE803_MAC_ADDRESS_SIZE ),
       sourceMacAddress,
       IX_IEEE803_MAC_ADDRESS_SIZE);

    /* If this userId has been enabled for fast path
     * then learn/age this flow (has to be done before
     * sending the frame since the mbufs may be modified
     * by virtual to physical address changes
     */
     if( connection->inUse == TRUE )
     {
        ixFpCodeletIpFlowLearn( 
            userId, 
            destIpAddress,
            (UINT8 *)( IX_MBUF_MDATA(buf) + IEEE802_3_DEST_MAC_ADDR_OFFSET ));
     }

    /* flus the data to memory, the NPE will be able to read it */
    IX_ACC_DATA_CACHE_FLUSH(IX_MBUF_MDATA(buf), SIZEOF_IEEE802_3_HDR);

    /* Transmit the PDU on the ethernet port */
    if (ixEthAccPortTxFrameSubmit ( 
            IX_ETH_PORT_1,
            buf, 
            IX_ETH_ACC_TX_DEFAULT_PRIORITY) != IX_SUCCESS )
    {
        IXFPATH_CODELET_ERROR("failed to send pdu on eth1");
        connection->routedIpStats.ethernetTxErrorCount++;
        ixFpCodeletBufManMbufFree (buf);
    }

}


/** 
 * @brief Set the IP address and MAC address for a specified flow
 *
 * This function demonstrates how to use the ixFpathAcc API
 * to update the classifier/modifier templates for a specific flow.
 *
 * @param connection - Demo codelet connection structure.
 * @param flowNum    - 1 of 4 possible IP flows for the VC.
 * @param ipAddress  - the destination IP address that the classifier flow
 *                     will be provisioned with so that it can be matched.
 * @param macAddr    - the corresponding destination MAC address for the destination Ip
 *                     addr. This will be populated at the correct point in the
 *                     modifier template for the specified flow.
 */
PRIVATE void
ixFpCodeletIpFlowUpdate( 
    IxFpCodeletFastPathConnection *connection,
    int flowNum,
    UINT32 ipAddress,
    UINT8 *macAddress )
{
    IxFpCodeletTemplate classifier;
    IxFpCodeletTemplate modifier;   

    /* Set the flow in the template */
    if( ixFpCodeletTemplatesIpFlowSet (
            connection->vc,
            flowNum,
            ipAddress,
            macAddress )
            != IX_SUCCESS )
    {
        IXFPATH_CODELET_ERROR("Unable to set IP flow in template");
        connection->routedIpStats.flowSetErrorCount++;
        return;
    }

    /* Download the updated templates*/
    ixFpCodeletTemplatesTemplatePairGet (
        connection->vc, 
        &classifier, 
        &modifier);

    /* Flush the cache because this data
     * will be read by the Npes
     */
    IX_ACC_DATA_CACHE_FLUSH(classifier.data,
                            classifier.len);
    IX_ACC_DATA_CACHE_FLUSH(modifier.data,
                            modifier.len);

    if( ixFpathAccTemplatePairSet (
            connection->npeVcId,
            classifier.data,
            classifier.len,
            modifier.data,
            modifier.len ) != IX_SUCCESS )
    {
        IXFPATH_CODELET_ERROR("failed to set template pair");
        connection->routedIpStats.templateSetErrorCount++;
    }
}

/** 
 * @brief Age IP flows and check for new flows
 *
 * This function illustrates a very simple IP flow ageing/learning
 * task. It primary purpose is to demonstrate how to use the
 * ixFpathAcc component to dynamically update templates and enable fast path
 * for specific IP flows. 
 *
 * See also ixFpathAccCodeletIpFlowLearn()
 *
 * Four active IP flows may use the fast path at any time. Each
 * flow is associated with a specific dest IP address.
 *
 * To be a candidate for fast path a flow must have received at least 
 * IP_FLOW_LEARNING_THRESHOLD pdus, within the IP_FLOW_ACCOUNTING_INTERVAL.
 *
 * The function ixFpathAccCodeletIpFlowLearn() (called from the ATM receive
 * handler) maintains an alive counter for each learning flow, each time a pdu 
 * with a matching dest IP address is received the counter is incremented.
 *
 * If a learning flows exceeds the learnig threshold, then the first available
 * fast path flow ( if any ) is assigned and the classifier/modifier 
 * are updated and downloaded. 
 * 
 * Once active on the fast path the flows are automatically aged, and
 * after the IP_FLOW_AGEING_INTERVAL has elapsed are marked as available
 * and may either be re-learned or superceded.
 *
 * @param arg       - not used.
 * @param ptrRegObj - not used.
 */
PRIVATE ix_error
ixFpCodeletIpFlowAgeingTask(
    void* arg,
    void **ptrRetObj)
{
    int active;
    int learning;
    IxFpCodeletFastPathConnection *connection;
    ActiveIpFlow   *activeFlow;
    LearningIpFlow *learningFlow;
    VcNumber vc;
    
    ix_ossl_thread_set_priority ( 
        ixFpCodeletIpFlowAgeingTaskId, 
        IX_OSSL_THREAD_PRI_LOW);

    while(taskEnabled)
    {
        for( vc = 0; vc < MAX_NUM_VCS; vc++ )
        {
            connection = 
                ixFpCodeletFastPathConnectionGet ((VcNumber) vc);

            /* Age all active Flows */ 
            for( active=0; active<MAX_FLOWS_PER_VC; active++ )
            {
                activeFlow = &connection->activeIpFlow[active];
                if( !activeFlow->aged )
                {
                    activeFlow->age++;
                    if( activeFlow->age > ixFpCodeletFlowAgeingInterval )
                    {
                        /* By marking this flow as aged we
                         * make it available to other IP addresses.
                         * Until it is claimed however the existing traffic
                         * will continue to flow on the fast path, this does
                         * no harm
                         */
                        IXFPATH_CODELET_TRACE(( traceBuf,
                            "%3d.%3d.%3d.%3d - Aged on Flow %d\n",
                            (activeFlow->ipAddress >> 24) & 0xff,
                            (activeFlow->ipAddress >> 16) & 0xff,
                            (activeFlow->ipAddress >> 8) & 0xff, 
                            (activeFlow->ipAddress >> 0) & 0xff,
                            active )); 
                        activeFlow->aged = TRUE;
                        activeFlow->ipAddress = 0;
                    } 
                }
            } /* end for ( active.... */

            /* Check learning Flows and handle new flow requests */
            for( learning=0; learning<NUM_LEARNING_FLOWS; learning++ )
            {
                learningFlow = &connection->learningIpFlow[learning];
                if ( learningFlow->inUse )
                {
                    if ( learningFlow->counter > 
                         ( ixFpCodeletFlowLearningThreshold / 
                           NUM_LEARNING_FLOWS ))
                    {
                        /* This flow is sustaining a PDU rate
                         * above the learning threshold
                         * so we search for an aged 
                         * active flow to supercede.
                         */
                        for( active=0; active<MAX_FLOWS_PER_VC; active++)
                        {
                            activeFlow = &connection->activeIpFlow[active];
                            if( activeFlow->aged )
                            {    
                                  activeFlow->ipAddress = learningFlow->ipAddress;
                                  memcpy (activeFlow->macAddress,
                                          learningFlow->macAddress,
                                          IX_IEEE803_MAC_ADDRESS_SIZE);
                                 /* Set the flow in the template */
                                 ixFpCodeletIpFlowUpdate (
                                     connection,
                                     active,
                                     activeFlow->ipAddress,
                                     activeFlow->macAddress );

                                 IXFPATH_CODELET_TRACE(( traceBuf,
                                     "%3d.%3d.%3d.%3d - Fast Path on Flow %d\n",
                                     (activeFlow->ipAddress >> 24) & 0xff,
                                     (activeFlow->ipAddress >> 16) & 0xff,
                                     (activeFlow->ipAddress >> 8) & 0xff, 
                                     (activeFlow->ipAddress >> 0) & 0xff,
                                     active ));

                                 /* make this learning slot available again */
                                 learningFlow->inUse = FALSE;
                                 learningFlow->ipAddress = 0;
                                 /* make the active flow in use */
                                 activeFlow->age = 0;
                                 activeFlow->aged = FALSE;
                                 break;
                             }
                        }
                    }
                    else
                    {
                        /* This flow failed to make the grade
                         * and has not been learned make this learning slot
                         * available to other IP addresses for learning.
                         */
                        IXFPATH_CODELET_TRACE((traceBuf, 
                            "%3d.%3d.%3d.%3d - Below Learning threshold\n",
                            (learningFlow->ipAddress >> 24) & 0xff,
                            (learningFlow->ipAddress >> 16) & 0xff,
                            (learningFlow->ipAddress >> 8) & 0xff, 
                            (learningFlow->ipAddress >> 0) & 0xff )); 
                        learningFlow->inUse = FALSE;
                        learningFlow->ipAddress = 0;
                    }
                }
                else
                {
                    /* This learning flow is available so check
                     * if we have received a new flow request for
                     * this VC
                     */
                    if( connection->ipFlowRequest )
                    {
                        /* get the new flow details */
                        learningFlow->ipAddress  = connection->ipFlowRequestData.ipAddress;
                        memcpy (learningFlow->macAddress,
                                connection->ipFlowRequestData.macAddress,
                                IX_IEEE803_MAC_ADDRESS_SIZE);

                        IXFPATH_CODELET_TRACE(( traceBuf,
                            "%3d.%3d.%3d.%3d - Started Learning\n",
                            (learningFlow->ipAddress >> 24) & 0xff,
                            (learningFlow->ipAddress >> 16) & 0xff,
                            (learningFlow->ipAddress >> 8) & 0xff, 
                            (learningFlow->ipAddress >> 0) & 0xff )); 

                        /* mark the flow as active */
                        learningFlow->inUse = TRUE;
                        /* clear the flow request */
                        connection->ipFlowRequest = FALSE;
                     }
                } 
                /* reset the counter */
                learningFlow->counter = 0;
                /* This delay ensures that each VC is processed once every
                 * IP accounting interval, and that the update of each flow
                 * on a VC is staggered.
                 */
                ixOsServTaskSleep(ixFpCodeletFlowAccountingInterval/NUM_LEARNING_FLOWS);
            }
        }
    }
    return IX_SUCCESS;
}

/** 
 * @brief Routed IP demonstration - Learn new IP flows
 *
 * This function illustrates the rx pdu handling part of a very simple 
 * IP flow ageing/learning mechanism. 
 *
 * @sa ixFpathAccCodeletIpFlowAgeingTask()
 *
 * Per VC Four active IP flows may use the fast path at any time. Each
 * flow is associated with a specific dest IP address. 
 *
 * To be a candidate for fast path a flow must have received at least 
 * IP_FLOW_LEARNING_THRESHOLD pdus, within the IP_FLOW_ACCOUNTING_INTERVAL.
 *
 * This algorthim allows up to NUM_LEARNING_FLOWS flows per VC to be 
 * learned at any time. 
 *
 * An alive counter is maintained for each learning 
 * flow, each time a pdu with a matching dest IP address is received 
 * the counter is incremented.
 *
 * If a pdu with a non matching dest IP address is received then 
 * the pdu IP address and mac address are saved and a new flow request
 * signalled to the flow ageing learning task by means of a simple flag.
 *
 * Further pdus with non mathcing dest IP addresses are ignored until
 * the request has been processed by the ageing/learning task and the
 * flag has been cleared.
 *
 * <B> IMPORTANT: </B> 
 * @note that the Ageing/Learning task is made responsible 
 * for updating the classifier and modifier templates. This transfer of 
 * control is necessary because these relatively expensive operations 
 * must not be performed in the rx data path context.
 *
 * @param userId - connection number.
 * @param currentIpAddress - the destination IP address of the current
 *                           IP packet.
 * @param macAddress - the corresponding destination MAC Addr for 
 *                           the current destination IP address. 
 */
PRIVATE void
ixFpCodeletIpFlowLearn( 
    IxAtmdAccUserId userId,
    UINT32 currentIpAddress,
    UINT8 *macAddress )
{
    int flowNum;
    ActiveIpFlow *activeFlow;
    LearningIpFlow *learningFlow;
    IxFpCodeletFastPathConnection *connection;

    connection = ixFpCodeletFastPathConnectionGet ((VcNumber) userId);

    /* check if PDU is on an active flow */
    for( flowNum=0; flowNum<MAX_FLOWS_PER_VC; flowNum++ )
    {
        activeFlow = &connection->activeIpFlow[flowNum];

        if( activeFlow->ipAddress == currentIpAddress )
        {
            /* When a newly learned flow transitions to the fast path, 
             * we may see some residual PDUs on the slow path
             * we simply ignore them
             * Note: in a real applictaion this could be policed by
             * a counter to detect a flow that did not correctly
             * transition to the fast path.
             */ 
            return;
        }
    }

    /* check if PDU is on a learning flow */
    for( flowNum=0; flowNum<NUM_LEARNING_FLOWS; flowNum++ )
    {
        learningFlow = &connection->learningIpFlow[flowNum];

        if( learningFlow->ipAddress == currentIpAddress )
        {
            learningFlow->counter++;
            return;
        }
    }

    /* Was not active or learning,  so this is a candidate for
     * being a new flow. Signal a new flow to be learned.
     */
    if( !connection->ipFlowRequest )
    {
        connection->ipFlowRequestData.ipAddress = currentIpAddress;
        memcpy( connection->ipFlowRequestData.macAddress,
                macAddress,
                IX_IEEE803_MAC_ADDRESS_SIZE);
        connection->ipFlowRequest = TRUE;
    }
}

/** 
 * @brief Routed IP demonstration - Get the dest MAC address 
 *
 * This routed IP fast path demonstration does not include ARP 
 * functionality. In this simple demonstration the IP address 
 * is NOT checked, and constant destination MAC address is returned
 * ( defined in IxFpathCodeletDefines_p.h )
 * @param destIpAddress - the destination IP address of the current
 *                        IP packet.
 * @param destMacAddress - the corresponding destination MAC Addr for 
 *                        the current destination IP address. 
 *
 */  
PRIVATE IX_STATUS
ixFpCodeletDestMacAddressGet( UINT32 destIPaddress,  
                                    UINT8 *destMacAddress )
{
    UINT8 targetMac[] = FPATH_ROUTEDIP_TARGET_MAC;
    
    memcpy( destMacAddress,
            targetMac,
            IX_IEEE803_MAC_ADDRESS_SIZE );

    return IX_SUCCESS;
}


