/**
 * @file   IxFpathAccCodeletTemplates.c
 * @date   May 29 2002
 * @brief  This file contains the implementation of the fpathAcc codelet templates
 *              
 * <tt>INTEL CONFIDENTIAL
 * Copyright (c) 2000-2002 Intel Corporation, All Rights Reserved.</tt>
 *
 * -- Intel Copyright Notice --
 * 
 * @par
 * Copyright 2000-2003 Intel Corporation All Rights Reserved.
 * 
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.
 * 
 * @par
 * The Material is protected by worldwide copyright and trade secret laws
 * and treaty provisions. No part of the Material may be used, copied,
 * reproduced, modified, published, uploaded, posted, transmitted,
 * distributed, or disclosed in any way except in accordance with the
 * applicable license agreement .
 * 
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel, except in accordance with the
 * applicable license agreement.
 * 
 * @par
 * Unless otherwise agreed by Intel in writing, you may not remove or
 * alter this notice or any other notice embedded in Materials by Intel
 * or Intel's suppliers or licensors in any way.
 * 
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * 
 * @par
 * -- End Intel Copyright Notice --
 */

/**
 * <P>
 * <B> Classifier Construction; </B>
 * <P>
 * The classifiers are made up of fixed fields at the beginning of each 
 * template followed by a combination of classifier instructions which form 
 * the remainder of the template. 
 *
 * The value of the fixed fields and the opcode combination vary based on 
 * the protocol and encapsulation of the packets to be sent on the fast path.
 *
 * Essentailly a classifier defines a set of tests to be performed on a 
 * an incoming aal5 pdu received on a specific VC. If all tests pass, the
 * packet is sent directly to the Ethernet NPE for modification into an
 * ethernet frame.
 *
 * The first four bytes of will be used to indicate AAL5 Length and 
 * UU comparing values. AAL5 max length is stored at offset 0 of the template 
 * as 16-bit little-endian value, and UU comparing value is stored at offset 2.
 *
 * After that, a number of checking instructions.
 
 * This file will introduce two basic types of classifier; one for bridged 
 * Ethernet packets and one for Routed IP packets.
 *
 * Bridged and routed templates have two significant difference.Firstly Ip 
 * templates have a notion of packets flows (4) and secondly Ip templates 
 * are specified to the NPES dynamically( based on the packet content e.g. 
 * destination IP addresses). 

 * Dynamic template specification is normally associated with IP flow learning 
 * and aging where slow path packet inspection occurs and result in dynamic
 * changes to templates.
 *
 * Bridged templates have no concept of flows and are usually specified 
 * to the NPEs once per VC setup.
 *
 * <P>
 * <B> Global Classifier Conditions </B>
 * <P>
 * For any VC, independent of whether an AAL5 PDU passes or not the fast 
 * path qualification test, all of the following global conditions must be 
 * met (and therefore are hard coded in Classifier) before the PDU can be 
 * sent to FpathMdf:
 * <UL> 
 *    <LI> 1.The AAL5 PDU passed AAL CRC32 check.
 *    <LI> 2.The entire AAL5 PDU is contained in single mbuf.
 * </UL> 
 * If any one of the above conditions is not met, the PDU is sent to the slow
 * path as usual. If all the above conditions are met, the PDU is sent to the 
 * modifier engine in NPE-B with additional information in the descriptor
 *
 * <B> Bridged VC Classifier Template </B>
 * <P>
 * <UL> Basic Constuct
 *   <LI> MAX Frame Size Field (2 bytes): Maximum frame size of the incoming
 *                                        aal5 PDU.
 *   <LI> UUI Inidication (1 byte)      : The Aal5 UUI to match. 
 *   <LI> Unused (1 byte)
 *   <LI> WAN Mac Learning Instruction; 
 *   <LI> Protocol Encapsulation Match Instruction; 
 *   <LI> End Instruction; 
 * </UL>
 *
 * <P>
 * <B> Routed VC Classifier Template </B>
 * <P>
 * <UL> Basic Constuct
 *   <LI> MAX Frame Size Field (2 bytes): Maximum frame size of the incoming
 *                                        aal5 PDU.
 *   <LI> UUI Inidication (1 byte)      : The Aal5 UUI to match. 
 *   <LI> Unused (1 byte)
 *   <LI> Protocol Encapsulation Match Instruction; 
 *   <LI> Match Instruction to check the IP Version  ; 
 *   <LI> Upper Limit Instruction       ; to check the IP packet length doesn't 
 *                                        exceed a specified size. 
 *   <LI> Low Limit Instruction         ; to check the Ip Packet Time to  
 *                                        live is at least 1.
 *   <LI> Checksum Instruction          ; to ensure the IP header is ok 
 *   <LI> One out of four instruction   ; to match one of four destination IP 
 *                                        addresses 
 *   <LI> End Instruction; 
 *
 * </UL> 
 */

#include "IxTypes.h"
#include "IxAssert.h"
#include "IxFpathAcc.h"
#include "IxOsServices.h"
#include "IxOsCacheMMU.h"
#include "IxAtmdAcc.h"
#include "IxAtmTypes.h"
#include "IxEthAcc.h"

#include "IxFpathAccCodeletDefines_p.h"

/** Max Encapsulated Frame Sizes based on standards. These are the maximum 
    size of encapsulated frames, include 1483 header, contained within
    downstream aal PDUs that will be allowed on the fast path.

    These values are used in the fast path templates.
    */
#define MAXSIZEOF_1483_LLC_IEEE802_3_FRAME \
            (MAXSIZEOF_IEEE802_3_FRAME + \
             SIZEOF_1483_LLC_IEEE802_3_HDR)

#define MAXSIZEOF_1483_VCMUX_IEEE802_3_FRAME \
            (MAXSIZEOF_IEEE802_3_FRAME + \
             SIZEOF_1483_VCMUX_IEEE802_3_HDR)

#define MAXSIZEOF_1483_LLC_IP_FRAME \
            (MAXSIZEOF_IEEE802_3_FRAME + \
             SIZEOF_1483_LLC_ROUTED_IP_HDR)

#define MAXSIZEOF_1483_VCMUX_IP_FRAME \
            (MAXSIZEOF_IEEE802_3_FRAME + \
             SIZEOF_1483_VCMUX_ROUTED_IP_HDR)


/** 
 * There is a fixed offset for packet inspection of 16 bytes for
 * classifier opcodes. This is the offset at which the first byte
 * of a received PDU begins.
 */
#define CLS_FIXED_OFFSET 0x10

/**
 * Bridged LLC classifier template.
 *
 * Payload Format RFC1483 encapsulated Bridged Ethernet/802.3 PDU.
 *
 * <verbatim>
   +-------------------------------+ _
   |       LLC  0xAA-AA-03         |  }
   |-------------------------------|  }
   |        OUI 0x00-80-C2         |  }
   |-------------------------------|  }  10 Bytes
   |    PID 0x00-01 or 0x00-07     |  }
   |-------------------------------|  }
   |         PAD 0x00-00           |  }
   |-------------------------------| -  
   | MAC destination addr (6 bytes)|  }
   |-------------------------------|  }
   | MAC source address   (6 bytes)|  }  12 Bytes
   |-------------------------------|  }
   | Type/Length          (2 bytes)|  }
   |-------------------------------| _}
   .                               .
   .   (remainder of MAC frame)    .
   .       (46 -> 1500 Bytes)      .
   .                               .
   +-------------------------------+ -
   |  LAN FCS (if PID is 0x00-01)  |  }   2 Bytes
   +-------------------------------+ ------------------
                                       1524 Max Total
 * <endverbatim>
 * 
 * Above is the payload format for an Ethernet frame transported 
 * in an RFC1483 LLC encapsulation.
 * 
 */
static  
UINT8 classifier_BridgedLLC[] =
{
    /** FRAME SIZE - AAL5 Maximum length.  */    

    /* 
     * The classifier template will ensure that the total length of
     * the Aal5 payload is not greater than the MTU for ethernet
     * frames plus the Ethernet frame overhead plus the encapsulation
     * overhead. In this case this is 1500 + 14 + 10.
     */
    (MAXSIZEOF_1483_LLC_IEEE802_3_FRAME & 0xFF) , /* Low byte of max size */
    (MAXSIZEOF_1483_LLC_IEEE802_3_FRAME >> 8) ,   /* Hi byte of max size  */

    0x00, /** AAL UU indication. ***/
    0x00, /*** Unused, to align. ***/

    /* 
     * The first opcode in this template is the to cause NPE-A to inspect the 
     * src mac address of the incoming ethernet frame and report it to 
     * IxFpathAcc.
     */
    /** WAN MAC learning instruction. */    
    IX_FPATHACC_CLS_OPCODE_WAN_MAC_LEARN , /** OPCODE */
    CLS_FIXED_OFFSET + 
    SIZEOF_1483_LLC_IEEE802_3_HDR, /** OFFSET - Destination mac addr in the PDU */

    /* 
     * The next opcode in this template ensures that the RFC1483 header is
     * correct.
     */
    /** Match PLOA bytes instruction */
    IX_FPATHACC_CLS_OPCODE_MATCH_BYTES ,       /** OPCODE */
    CLS_FIXED_OFFSET,              /** OFFSET - Start of PDU */
    SIZEOF_1483_LLC_IEEE802_3_HDR, /** SIZE  RFC 1483 LLC Encapsulation Header */
    0xaa, /* DATA   */  /* +-------------------------------+ */
    0xaa,               /* |       LLC  0xAA-AA-03         | */
    0x03,               /* +-------------------------------+ */
    0x00,               /* |        OUI 0x00-80-C2         | */
    0x80,               /* +-------------------------------+ */
    0xc2,               /* |    PID 0x00-01 or 0x00-07     | */
    0x00,               /* +-------------------------------+ */
    0x07,               /* |         PAD 0x00-00           | */
    0x00,               /* +-------------------------------+ */
    0x00,

    /* If all of the classifier succeeds this far then the paket will go to
       the modifier. */

    /* End Template */
    IX_FPATHACC_CLS_OPCODE_END  
};

/**
 * Bridged VcMux Classifier Template
 *
 *  Payload Format for Bridged Ethernet/802.3 PDUs
 *  <verbatim>
   +-------------------------------+ -
   |       VCMUX  0x00-00          |  }   2 bytes
   |-------------------------------| -  
   | MAC destination addr (6 bytes)|  }
   |-------------------------------|  }
   | MAC source address   (6 bytes)|  }  12 Bytes
   |-------------------------------|  }
   | Type/Length          (2 bytes)|  }
   |-------------------------------| _}
   .                               .
   .   (remainder of MAC frame)    .
   .       (46 -> 1500 Bytes)      .
   .                               .
   +-------------------------------+ -
   |  LAN FCS (if PID is 0x00-01)  |  }   2 Bytes
   +-------------------------------+ ------------------
                                       1516 Bytes Total
 *  <endverbatim>
 */
static  
UINT8 classifier_BridgedVcMux[] =
{
    /*** FRAME SIZE - AAL5 Maximum length. */    
    (MAXSIZEOF_1483_VCMUX_IEEE802_3_FRAME & 0xFF), 
    (MAXSIZEOF_1483_VCMUX_IEEE802_3_FRAME >> 8), 

    0x00, /* AAL UU indication. */
    0x00, /* Unused, to align. */

    /** WAN MAC learning instruction. */    
    IX_FPATHACC_CLS_OPCODE_WAN_MAC_LEARN,
    CLS_FIXED_OFFSET + 
    SIZEOF_1483_VCMUX_IEEE802_3_HDR+
    SIZEOF_IEEE802_3_DEST_MAC_ADDR,/** OFFSET - Destination mac addr in the PDU */

    /** Match PLOA bytes instruction */
    IX_FPATHACC_CLS_OPCODE_MATCH_BYTES ,       /** OPCODE */
    CLS_FIXED_OFFSET,              /** OFFSET - Start of PDU */
    SIZEOF_1483_VCMUX_IEEE802_3_HDR, /** SIZE   */
    0x00,  /* DATA   */  /* +-------------------------------+*/
    0x00,                /* |       VCMUX  0x00-00          |*/
                         /* +-------------------------------+*/
    /* End Template */
    IX_FPATHACC_CLS_OPCODE_END  
};

/** Indices to offset into the classifier_IprIPFlow array.*/
#define IPFLOW_CLS_PAD_INDEX (0x00)
#define IPFLOW_CLS_1OF4_MATCH_CMD_INDEX \
           (IPFLOW_CLS_PAD_INDEX+1)
#define IPFLOW_CLS_1OF4_MATCH_CMD_OFFSET_INDEX \
           (IPFLOW_CLS_1OF4_MATCH_CMD_INDEX+1)
#define IPFLOW_CLS_FIRST_FLOW_INDEX \
           (IPFLOW_CLS_1OF4_MATCH_CMD_OFFSET_INDEX+2)
/*******************************************************************/
/* Routed RFC-1483 LLC NPE Classifier template */
/*******************************************************************/
/*
    Payload Format for Routed LLC IP PDUs
   +-------------------------------+  -           
   |       LLC  0xAA-AA-03         |   }          
   +-------------------------------+   }           
   |        OUI 0x00-00-00         |   } 8 bytes  
   +-------------------------------+   }          
   |       EtherType 0x08-00       |   }          
   +-------------------------------+  -           
   |             .                 |
   |           IP PDU              |
   |     (up to 2^16 - 9 octets)   |
   |             .                 |
   +-------------------------------+
*/
UINT8 classifier_IprLlc[] = 
{
    /* 
     * In this case the Aal5 pdu contains an IP packet along with a 
     * a 1483 LLC header. Since the 1483 header will be removed the
     * the IP packet must be small enough to fit in the 1500 byte payload
     * of an Ehternet Frame if it is to go on the fast path. This retriction
     * applies because the fast path does not have IP packet SARing
     * capabilites, a slow path IP stack must perform that operation.
     *
     * Hence, the classifier template will ensure that the total length of
     * the Aal5 payload is not greater than the MTU for ethernet
     * frames plus the encapsulation overhead. 
     *
     * In this case this breaks down as;
     *     1500 Ethernet payload + 8 byte LLC header.
     */
    /** AAL5 SDU Maximum length. - Must */
    (MAXSIZEOF_1483_LLC_IP_FRAME &  0xFF), 
    (MAXSIZEOF_1483_LLC_IP_FRAME >> 0x08), 

    0x00 , /* AAL UU indication. */
    0x00, /* Unused, to align. */

    /* 
     * The first opcode in this template ensures that the RFC1483 header is
     * correct.
     */
    /** Match: PLOA header. */
    IX_FPATHACC_CLS_OPCODE_MATCH_BYTES,       /** OPCODE */                                 
    CLS_FIXED_OFFSET,             /** OFFSET - Start of PDU */
    SIZEOF_1483_LLC_ROUTED_IP_HDR,/** SIZE- RFC 1483 LLC Encapsulation Header */
    0xaa,
    0xaa,
    0x03,
    0x00,
    0x00,
    0x00,
    0x08,
    0x00,

    /** 
      * Match: IP version and header length. 
      *
      * The IP header length value should be 5 because we do not send
      * packets which contain the "Options" field on the fast path. This
      * is for simplicity. Hence, the IP HDR is expected to be 20 bytes 
      * long. 
      */
    IX_FPATHACC_CLS_OPCODE_MATCH_BYTES,
    CLS_FIXED_OFFSET+
    SIZEOF_1483_LLC_ROUTED_IP_HDR, /** OFFSET */
    SIZEOF_VER_PLUS_HL,            /** SIZE */
    IPV4_HDR_ID,                   /** DATA */

    /** 
      * UpLimit: Make sure that the IP packet length is small enough to
      * fit in an ethernet frame. 
      */
    IX_FPATHACC_CLS_OPCODE_UPPER_LIMIT,
    CLS_FIXED_OFFSET+
    SIZEOF_1483_LLC_ROUTED_IP_HDR+ /** RFC 1483 LLC Encapsulation Header */
    IP_TOTAL_LEN_OFFSET,           /** Offset to length field within IP header */
    SIZEOF_IP_TOTAL_LEN,           /** SIZE */
    (MAXSIZEOF_IEEE802_3_PAYLOAD  & 0xFF), 
    (MAXSIZEOF_IEEE802_3_PAYLOAD  >> 8),  

    /** 
      * LowLimit: Check the the TTL is a minimum of 1. i.e. ensure this packet
      * doesn't need to be destroyed. 
      */
    IX_FPATHACC_CLS_OPCODE_LOWER_LIMIT,
    CLS_FIXED_OFFSET+
    SIZEOF_1483_LLC_ROUTED_IP_HDR+ /** RFC 1483 LLC Encapsulation Header */
    IP_TTL_OFFSET, /** OFFSET */
    SIZEOF_IP_TTL, /** DATA */
 
    /** 
      * Checksum the header - this doubles as a check to ensure no options
      * are set in the header as the checksum is calculated over 20bytes
      * instead of 24bytes.  
      */
    IX_FPATHACC_CLS_OPCODE_CHECKSUM, /* OPCODE */
    CLS_FIXED_OFFSET+
    SIZEOF_1483_LLC_ROUTED_IP_HDR, /** Begin checksum at the start of IP packet */
    NUMBEROF_SHORTS_TO_CHECKSUM_IN_IP_HDR, 

    /* Pad out the base template to a 4 byte boundary */
    IX_FPATHACC_CLS_OPCODE_NOP_PAD1,
    IX_FPATHACC_CLS_DUMMY_BYTE
    /* Note Flow Identification Commands are appended here*/ 
};

/*******************************************************************/
/* Routed RFC-1483 VCMUX NPE Classifier template */
/*******************************************************************/
/*
    Payload Format for Routed VCMUX IP PDUs
   +-------------------------------+             
   |             .                 |
   |           IP PDU              |
   |     (up to 2^16 - 9 octets)   |
   |             .                 |
   +-------------------------------+
*/
UINT8 classifier_IprVcMux[] = 
{
    /* AAL5 Maximum length. */
    (MAXSIZEOF_1483_VCMUX_IP_FRAME &  0xFF), 
    (MAXSIZEOF_1483_VCMUX_IP_FRAME >> 0x08), 

    0x00 , /* AAL UU indication. */
    0x00,  /* Unused, to align. */

    /** Match: IP version and header length. */
    /** This*/
    IX_FPATHACC_CLS_OPCODE_MATCH_BYTES,
    CLS_FIXED_OFFSET,
    SIZEOF_VER_PLUS_HL,           /** SIZE */
    IPV4_HDR_ID,                   /** DATA */

    /** UpLimit: packet length. */
    IX_FPATHACC_CLS_OPCODE_UPPER_LIMIT,
    CLS_FIXED_OFFSET+
    IP_TOTAL_LEN_OFFSET,           /** Offset to length field within IP header */
    SIZEOF_IP_TOTAL_LEN,             /** SIZE */
    (MAXSIZEOF_IEEE802_3_PAYLOAD  >> 8),  
    (MAXSIZEOF_IEEE802_3_PAYLOAD  & 0xFF), 

    /* LowLimit: TTL */
    IX_FPATHACC_CLS_OPCODE_LOWER_LIMIT,
    CLS_FIXED_OFFSET+
    IP_TTL_OFFSET, /** OFFSET */
    SIZEOF_IP_TTL, /** DATA */

    /* Checksum */
    IX_FPATHACC_CLS_OPCODE_CHECKSUM, /* OPCODE */
    CLS_FIXED_OFFSET,
    /** The size of the checksum area is expected to be 20 bytes */
    NUMBEROF_SHORTS_TO_CHECKSUM_IN_IP_HDR,                 /**  */

    /* Pad out the base template to a 4 byte boundary */
    IX_FPATHACC_CLS_OPCODE_NOP

    /* NOTE - The flow identification trailer will be appended here. */ 
};

/**
 * @brief Base IP Flow classifier CMD template. (partial)
 *
 * This the part of IP classifier templates which match IP packets to
 * be sent on the fast path. There is a max of four packet flows allowed.
 *
 * This partial template is appended to base IP template.
 * @sa ixFpCodeletTemplatesBaseTemplateCreate
 */
static UINT8 classifier_IprIPFlow[] =
{
    /* Pad the IP address to fall on a 4 byte boundary */
    IX_FPATHACC_CLS_OPCODE_NOP,

    /* Normal IP Flow, 1-out-of-4-match: Destination IP addresses. */
    IX_FPATHACC_CLS_OPCODE_1_OUT_OF_4_MATCH,
    
    IP_DEST_ADDR_OFFSET + 
    CLS_FIXED_OFFSET,  /** Note: value is adjusted to account for LLC or VCMUX
                         * headers. */
        SIZEOF_IP_DEST_ADDR,
        0,0,0,0, /* Empty space will be populated by IP learning function !! */
        0,0,0,0,
        0,0,0,0,
        0,0,0,0, 

    /* End Template */
    IX_FPATHACC_CLS_OPCODE_END
};

/** Just a dummy classifier */
static  
UINT8 classifier_dummy[] = 
{
    IX_FPATHACC_CLS_OPCODE_NOP,
    IX_FPATHACC_CLS_OPCODE_NOP,
    IX_FPATHACC_CLS_OPCODE_NOP,
    IX_FPATHACC_CLS_OPCODE_NOP,
    IX_FPATHACC_CLS_OPCODE_END,
};

#define MAX_TEMPLATE_SIZE 255

/** 
 *  <P>
 *  <B> Modifier construction: </B>
 *  <P>
 *
 *  Modifier templates are arranged such that the MAC address per flow 
 *  are split into two parts. The first 4 bytes of each flow MAC address 
 *  appear sequentially in the template. Followed by the second 2 bytes of 
 *  each flow mac address. 
 *
 *  This is the basic modifier template upon which all modifier
 *  templates for all encapsulations are built. 
 */
static  
UINT8 modifier_blank[] =
{
    /* Protocols. !! */
    0x00,0x00, /* */
    
    /* mData/mLen change. !!! */
    0x00,0x00, /* The length by which the packet header is shortened */

    0x00,0x00,0x00,0x00, /* 12: Dst MAC[0], byte 0-3 !! */
    0x00,0x00,0x00,0x00, /* 16: Dst MAC[1], byte 0-3 !! */
    0x00,0x00,0x00,0x00, /* 20: Dst MAC[2], byte 0-3 !! */
    0x00,0x00,0x00,0x00, /* 24: Dst MAC[3], byte 0-3 !! */

    0x00,0x00, /* 28: Dst MAC[0], byte 4-5 !! */
    0x00,0x00, /* 30: Dst MAC[1], byte 4-5 !! */
    0x00,0x00, /* 32: Dst MAC[2], byte 4-5 !! */
    0x00,0x00, /* 34: Dst MAC[3], byte 4-5 !! */

    0x00,0x00,0x00,0x00, /* 36: Local dst IP[0] !! */
    0x00,0x00,0x00,0x00, /* 40: Local dst IP[1] !! */
    0x00,0x00,0x00,0x00, /* 44: Local dst IP[2] !! */
    0x00,0x00,0x00,0x00, /* 48: Local dst IP[3] !! */

    0x00,0x00, /* 52: Local dst port[0] !! */
    0x00,0x00, /* 54: Local dst port[1] !! */
    0x00,0x00, /* 56: Local dst port[2] !! */
    0x00,0x00  /* 58: Local dst port[3] !! */
};

/* 
 * Definitions to sizes and offsets to members fileds within modifier template 
 */
#define SIZEOF_MDF_DST_MAC_FIRST_PART     4
#define SIZEOF_MDF_DST_MAC_SECOND_PART    2

#define OFFSET_MDF_DST_MAC_BYTE0       0x04
#define OFFSET_MDF_DST_MAC_BYTE4 \
            (OFFSET_MDF_DST_MAC_BYTE0)\
            +(SIZEOF_MDF_DST_MAC_FIRST_PART*MAX_FLOWS_PER_VC)
#define OFFSET_MDF_DST_IP \
            (OFFSET_MDF_DST_MAC_BYTE4)\
            +(SIZEOF_MDF_DST_MAC_SECOND_PART*MAX_FLOWS_PER_VC)
           
#define OFFSET_MDF_DST_PORT \
             (OFFSET_MDF_DST_IP)\
             +(SIZEOF_IP_DEST_ADDR*MAX_FLOWS_PER_VC) 

/** 
 * Valid protocol values for the first field of the modifier templates 
 */
#define PROTOCOL_BDG  0x00
#define PROTOCOL_IPR  0x01
#define PROTOCOL_NAT  0x02 /* Not support in this codelet */
#define PROTOCOL_NAPT 0x04 /* Not support in this codelet */

/** struture to hold the VC information */
typedef struct
{ 
    unsigned port;
    unsigned vpi;
    unsigned vci;

    EncapsulationType type;

    UINT8  *templateCls;   /* SPoints to the SDRAM copy of the classifier for */
                           /* this VC.*/
    UINT32 templateClsSize;/* Calssifier Size */
    UINT32 templateClsIpHdrOffset; /* IP flows only */

    UINT8  *templateMdf;   /* SPoints to the SDRAM copy of the modifier for */ 
    UINT32 templateMdfSize;
}VcInfo;

/** Holds the */
VcInfo vcInfo[MAX_NUM_VCS];

/** 
 * This type defines a classifier/modifier template pair
 */
typedef struct
{
    IxFpCodeletTemplate classifier; /* classifier template */
    IxFpCodeletTemplate modifier;   /* modifier template */
    unsigned int    clsIpHdrOffset; /* Offset from the start of a packet to the */
                                    /* IP addr. The offset varies based on the  */
                                    /* protocol encapsulation.*/

    UINT16 mdfProtocol;        /* The protocl for the template pair.*/
    UINT16 mdfHeaderAdjustVal; /* This value is placed in the modifier template */
                               /* and cooresponds number of bytes that the      */
                               /* modifier removes from the start of a fast path*/
                               /* packet. Effectively the amount of header info */
                               /* to remove.*/
} IxFpathTemplatePair;

/** 
 * Sets up the base template for each encapsulation type.
 */
IxFpathTemplatePair baseTemplateInfo[NUM_ENCAP_TYPES] =
{
    /* TYPE_NONE */
    {  
        {classifier_dummy, sizeof(classifier_dummy)}, /* Dummy */
        {modifier_blank, sizeof(modifier_blank)},
        0,  /* Offset to Ip header */
        PROTOCOL_BDG,
        0,
    }, /* Dummy */

    /* TYPE_1483BLLC */
    {
        {classifier_BridgedLLC,   sizeof(classifier_BridgedLLC)},
        {modifier_blank, sizeof(modifier_blank)},
        0,  /* Offset to Ip header */
        PROTOCOL_BDG,
        SIZEOF_1483_LLC_IEEE802_3_HDR
    },

    /* TYPE_1483BVCMUX */
    {
        {classifier_BridgedVcMux, sizeof(classifier_BridgedVcMux)},
        {modifier_blank, sizeof(modifier_blank)},
        0,  /* Offset to Ip header */
        PROTOCOL_BDG,
        SIZEOF_1483_VCMUX_IEEE802_3_HDR
    },

    /* TYPE_ROUTEDIPLLC */
    {
        {classifier_IprLlc,       sizeof(classifier_IprLlc)},
        {modifier_blank,          sizeof(modifier_blank)},
        SIZEOF_1483_LLC_ROUTED_IP_HDR,
        PROTOCOL_IPR,
        SIZEOF_1483_LLC_ROUTED_IP_HDR
    },

    /* TYPE_ROUTEDIPVCMUX*/
    {
        {classifier_IprVcMux,     sizeof(classifier_IprVcMux)},
        {modifier_blank,          sizeof(modifier_blank)},
        0,
        PROTOCOL_IPR,
        0,
    },
};

/**
 * Array containing the text string that describes each encapsulation type.
 */
char encapsulation[NUM_ENCAP_TYPES][20] =
{
    "TYPE_NONE\0",
    "TYPE_1483BLLC\0",
    "TYPE_1483BVCMUX\0",
    "TYPE_ROUTEDIPLLC\0",
    "TYPE_ROUTEDIPVCMUX\0",
};

/**
 * @brief Return the text string that describes the encapsulation type
 */
char*
ixFpCodeletTemplatesEncapuslationName( EncapsulationType t )
{
    if( t < NUM_ENCAP_TYPES )
    {
        return &encapsulation[t][0];
    }
    return NULL;
}

/**
 * @brief Initialise the template utility subcomponent.
 */
IX_STATUS
ixFpCodeletTemplatesInit(void)
{
    unsigned int vcCntr = 0;

    /* Zero the struct */
    memset(&vcInfo, 0x00, sizeof(vcInfo));

    for (vcCntr = 0 ; vcCntr < MAX_NUM_VCS ; vcCntr++)
    {
        vcInfo[vcCntr].templateCls = IX_ACC_DRV_DMA_MALLOC(MAX_TEMPLATE_SIZE);
        vcInfo[vcCntr].templateMdf = IX_ACC_DRV_DMA_MALLOC(MAX_TEMPLATE_SIZE);
    }
    return IX_SUCCESS;
}

/**
 * Offsets to fields inside the modifier template.
 */
#define MDF_PROTOCOL_FIELD_OFFSET   0x01
#define MDF_HEADER_LEN_FIELD_OFFSET 0x03

/** 
 * @brief Creates a base template pair for a VC 
 *
 * This function will create a base classifier and modifier template for
 * a vc based on the encapsulation. The resultant templates are stored
 * in SDRAM,they are not sent to the NPEs.
 *
 * For IP encapsulations the IP flow learning partial classifier is appended 
 * to the base template to form a complete classifier template. Initially
 * all 4 IP flows are inactive i.e.have IP addresses of 0.0.0.0.
 *
 * @param vc   - Codelet connection number. 
 * @param type - The VC encapsulation type.
 */
void 
ixFpCodeletTemplatesBaseTemplatePairCreate( 
    VcNumber vc, 
    EncapsulationType type )
{
    vcInfo[vc].type = type;

    /* 
     * CLASSIFIER CONSTRUCTION 
     */

    /* Copy the base classifier template for the vc from the base
       template array */
    memcpy(
       vcInfo[vc].templateCls, 
       baseTemplateInfo[type].classifier.data, 
       baseTemplateInfo[type].classifier.len);

    /* Store the classifier size */
    vcInfo[vc].templateClsSize = baseTemplateInfo[type].classifier.len;    

    /* Need to add the IP flow partial template to the end of the base
       template to form a complete template.*/
    switch (type) 
    {
        case TYPE_ROUTEDIPLLC:
        case TYPE_ROUTEDIPVCMUX:
            /* Append the IP Flow CMD*/ 
            memcpy(
                &(vcInfo[vc].templateCls[vcInfo[vc].templateClsSize]), 
                classifier_IprIPFlow, 
                sizeof(classifier_IprIPFlow));

	    /* Set offset parameter in the 1-out-of-4-match opcode for
               the IP Flows in the new classifier */			
            vcInfo[vc].templateCls[vcInfo[vc].templateClsSize+
                                   IPFLOW_CLS_1OF4_MATCH_CMD_OFFSET_INDEX]
                += baseTemplateInfo[type].clsIpHdrOffset;

	    /* Update the classifier size */			
            vcInfo[vc].templateClsSize += sizeof(classifier_IprIPFlow);
            break;

        case TYPE_NONE: 
        case TYPE_1483BLLC:
        case TYPE_1483BVCMUX:
            break;
        default:
            IX_ASSERT(0);
            break;
    }

    /** 
     * MODIFIER CONSTRUCTION 
     */

    /* Copy the base modifier template for the vc */
    memcpy(
       vcInfo[vc].templateMdf, 
       baseTemplateInfo[type].modifier.data, 
       baseTemplateInfo[type].modifier.len);

    vcInfo[vc].templateMdfSize = baseTemplateInfo[type].modifier.len;    

    /* Protocol field */
    vcInfo[vc].templateMdf[MDF_PROTOCOL_FIELD_OFFSET] = 
        baseTemplateInfo[type].mdfProtocol;

    vcInfo[vc].templateMdf[MDF_HEADER_LEN_FIELD_OFFSET] = 
        baseTemplateInfo[type].mdfHeaderAdjustVal;

    /** 
     * Some templates need extra commands appended so....
     * Get the classifier template (NAT/NAPT/IP Flow) to concatenate with the base
     * template. This function also adjusts the offsets accordingly
     */

}

/**
 * Defines to facilitate setting of WAN mac learning cmd in classifier
 * template
 * @sa ixFpathAccCodeletTemplatesWanMacLearningSet
 */
#define CLS_WAN_MAC_LEARN_OPCODE_INDEX_1       0x04
#define CLS_WAN_MAC_LEARN_OPCODE_INDEX_2_VCMUX 0x0b
#define CLS_WAN_MAC_LEARN_OPCODE_INDEX_2_LLC   0x13
#define CLS_WAN_MAC_LEARN_OFFSET_INDEX         0x06
#define IEEE802_3_LLC_SRC_HOST_FIELD_OFFSET    0x20
#define IEEE802_3_VCMUX_SRC_HOST_FIELD_OFFSET  0x18

/** 
 * @brief Fill in the WAN mac learning opcode in the classifier
 *
 * The Wan Mac learning opcode can appear at any point in the 
 * classifier template. This codelet allows it to be placed at the
 * beginning or end of the template. 
 *
 * @note If placed at the start of the classifier then all Ethernet 
 * packets of valid length have their WAN mac addresses learned even if
 * they eventually go the exception path. Id placed as the last opcode
 * in the classifier then only packets which qualify for the fast path 
 * will have their Src MAC addresses learned.
 *
 *
 */
IX_STATUS
ixFpCodeletTemplatesWanMacLearningSet(
    VcNumber vc, 
    BOOL enabled,
    BOOL firstInstuction)
{
    unsigned char firstOpcode; 
    unsigned char secondOpcode;

    /* N.B. This function just replaces some opcodes in the base template
            for a bridged encapsulation. */
    
    /* Configure opcode */
    if ( enabled ) 
    {
        firstOpcode = 
            (firstInstuction?IX_FPATHACC_CLS_OPCODE_WAN_MAC_LEARN:IX_FPATHACC_CLS_OPCODE_NOP_PAD1);

        secondOpcode = 
            (firstInstuction?IX_FPATHACC_CLS_OPCODE_NOP_PAD1:IX_FPATHACC_CLS_OPCODE_WAN_MAC_LEARN);
    }
    else
    {
        /* No MAC learning */
        firstOpcode  = IX_FPATHACC_CLS_OPCODE_NOP_PAD1;
        secondOpcode = IX_FPATHACC_CLS_OPCODE_NOP_PAD1;
    }

    switch ( vcInfo[vc].type ) 
    {
         case TYPE_1483BVCMUX:
             vcInfo[vc].templateCls[CLS_WAN_MAC_LEARN_OPCODE_INDEX_1]
                = firstOpcode;
             vcInfo[vc].templateCls[CLS_WAN_MAC_LEARN_OPCODE_INDEX_2_VCMUX]
                = secondOpcode;
         break;        

         case TYPE_1483BLLC:
             vcInfo[vc].templateCls[CLS_WAN_MAC_LEARN_OPCODE_INDEX_1]
                = firstOpcode;
             vcInfo[vc].templateCls[CLS_WAN_MAC_LEARN_OPCODE_INDEX_2_LLC]
                = secondOpcode;
         break;        

         default:
             return IX_FAIL;
         break;    
    }

    return IX_SUCCESS;
}

/** 
 * @brief Configures a template IP flow 
 *
 * @param vc         - Codelet connection number. 
 * @param flow       - specifies 1 of 4 flow positions within the temapltes.
 * @param destIpAddr - destination IP address to be filled into the classifier
 * @param destMac    - destination Mac address for the destIpAddr to be placed 
 *                     in the modifier template
 *
 * This function fill in the specified destination IP address into the correct
 * position in the classfier template based on the specified flow. 
 *
 * Correspondingly the destination mac address to be used by the modifier engine
 * when building the ethernet frame is filled in at the correct flow position.
 *
 * @note This only fills out the SDRAM copy of the templates.
 */
IX_STATUS
ixFpCodeletTemplatesIpFlowSet(
    VcNumber vc,         /* 0 - 7*/
    FlowId   flow,       /* 0 - 3 */
    UINT32   destIpAddr, /* Classifier matches this addr */
    UINT8   *destMac )   /* Modifier inserts this in Ethernet header */
{
    unsigned int mdfDestMacOffset    = 0;
    EncapsulationType type           = vcInfo[vc].type;

    unsigned int clsFlowOffset
        = baseTemplateInfo[type].classifier.len +
          IPFLOW_CLS_FIRST_FLOW_INDEX +
          (flow*4);
    
    /* Setup the classifier flow information (dest Ip Addr)*/
    vcInfo[vc].templateCls[clsFlowOffset+3] = (destIpAddr >> 0)  & 0xFF;
    vcInfo[vc].templateCls[clsFlowOffset+2] = (destIpAddr >> 8)  & 0xFF;
    vcInfo[vc].templateCls[clsFlowOffset+1] = (destIpAddr >> 16) & 0xFF;
    vcInfo[vc].templateCls[clsFlowOffset]   = (destIpAddr >> 24) & 0xFF;

    /* Setup the modifier flow information */
    mdfDestMacOffset = OFFSET_MDF_DST_MAC_BYTE0 +
        (SIZEOF_MDF_DST_MAC_FIRST_PART*flow);

    vcInfo[vc].templateMdf[mdfDestMacOffset++] = destMac[5];
    vcInfo[vc].templateMdf[mdfDestMacOffset++] = destMac[4];
    vcInfo[vc].templateMdf[mdfDestMacOffset++] = destMac[3];
    vcInfo[vc].templateMdf[mdfDestMacOffset++] = destMac[2];
    
    mdfDestMacOffset = OFFSET_MDF_DST_MAC_BYTE4 +
        (SIZEOF_MDF_DST_MAC_SECOND_PART*flow);
    vcInfo[vc].templateMdf[mdfDestMacOffset++] = destMac[1];
    vcInfo[vc].templateMdf[mdfDestMacOffset]   = destMac[0];

    return IX_SUCCESS;
}

/** 
 * @brief Return the template pair that has been built for the current Vc
 *
 * @param vc         - Codelet connection number. 
 * @param classifier - Pointer to a classifier.
 * @param modifier   - Pointer to a modifier.
 */
IX_STATUS
ixFpCodeletTemplatesTemplatePairGet(
    VcNumber vc,
    IxFpCodeletTemplate *classifier,
    IxFpCodeletTemplate *modifier )
{
    if ((vc < 0) || (vc > VC_7) ||
        ( classifier == NULL ) ||
        ( modifier == NULL ))
        return IX_FAIL;
    
    classifier->data = vcInfo[vc].templateCls;
    classifier->len  = vcInfo[vc].templateClsSize;
    modifier->data   = vcInfo[vc].templateMdf;
    modifier->len    = vcInfo[vc].templateMdfSize;
    return IX_SUCCESS;
}
    
/** 
 * @brief Print the modifier template for a VC
 *
 * @param vc - Codelet connection number. 
 */
IX_STATUS
ixFpCodeletTemplatesMdfPrint(VcNumber vc)
{
    unsigned int cntr = 0;

    if ( (vc < 0) || (vc > VC_7))
        return IX_FAIL;
    
    printf("\n Modifier for Vc %i; Size = %i; Addr 0x%x ", 
        vc, 
        vcInfo[vc].templateMdfSize,
        (unsigned int) &vcInfo[vc].templateMdf[0]);
    
    while(cntr < vcInfo[vc].templateMdfSize)
    {
        printf("\n  Byte 0x%2.2x: 0x%2.2x, 0x%2.2x, 0x%2.2x, 0x%2.2x, ", 
            cntr,
            vcInfo[vc].templateMdf[cntr],
            vcInfo[vc].templateMdf[cntr+1], 
            vcInfo[vc].templateMdf[cntr+2], 
            vcInfo[vc].templateMdf[cntr+3]);
	cntr+=4; 
             
    } 
    printf("\n Modifier End " );

    return IX_SUCCESS;
}

/** 
 * @brief Print the classifier template for a VC
 *
 * @param vc - Codelet connection number. 
 */
IX_STATUS
ixFpCodeletTemplatesClsPrint(VcNumber vc)
{
    unsigned int cntr = 0;

    if ( (vc < 0) || (vc > VC_7))
        return IX_FAIL;
    
    printf("\n Classifier for Vc %i Size = %i Addr 0x%x",
        vc, 
        vcInfo[vc].templateClsSize,
        (unsigned int) &vcInfo[vc].templateCls[0]);

    while(cntr < vcInfo[vc].templateClsSize)
    {
        printf("\n  Byte 0x%2.2x: 0x%2.2x, 0x%2.2x, 0x%2.2x, 0x%2.2x, ", 
            cntr,
            vcInfo[vc].templateCls[cntr],
            vcInfo[vc].templateCls[cntr+1], 
            vcInfo[vc].templateCls[cntr+2], 
            vcInfo[vc].templateCls[cntr+3]);
	cntr+=4; 
    } 
    printf("\n Classifier End " );

    return IX_SUCCESS;
}



