/**
 * @file  IxFpathAccCodeletTemplates_p.h
 * @date  May 29 2002
 * @brief This file contains code to generate classifier/modifier
 *        templates for the IxFpathAcc codelet.
 *
 * Design Notes:
 *
 * -- Intel Copyright Notice --
 * 
 * @par
 * Copyright 2002-2003 Intel Corporation All Rights Reserved.
 * 
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.
 * 
 * @par
 * The Material is protected by worldwide copyright and trade secret laws
 * and treaty provisions. No part of the Material may be used, copied,
 * reproduced, modified, published, uploaded, posted, transmitted,
 * distributed, or disclosed in any way except in accordance with the
 * applicable license agreement .
 * 
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel, except in accordance with the
 * applicable license agreement.
 * 
 * @par
 * Unless otherwise agreed by Intel in writing, you may not remove or
 * alter this notice or any other notice embedded in Materials by Intel
 * or Intel's suppliers or licensors in any way.
 * 
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * 
 * @par
 * -- End Intel Copyright Notice --
 */

#ifndef IX_FPATHACC_CODELET_TEMPLATES_P_H
#define IX_FPATHACC_CODELET_TEMPLATES_P_H

#include "IxTypes.h"
#include "IxAtmTypes.h"

/**
 *  defines
 */

/** Header defines c.f. Rfc1483 Multiprotocol Encapsulation
    over ATM Adaptation Layer 5 */
#define SIZEOF_1483_LLC_IEEE802_3_HDR   (10)
#define SIZEOF_1483_VCMUX_IEEE802_3_HDR (2)
#define SIZEOF_1483_LLC_ROUTED_IP_HDR   (8)
#define SIZEOF_1483_VCMUX_ROUTED_IP_HDR (0)

/** Ethernet Frame Size Defines */
#define SIZEOF_IEEE802_3_DEST_MAC_ADDR  (6)
#define SIZEOF_IEEE802_3_SRC_MAC_ADDR   (6)
#define SIZEOF_IEEE802_3_TYPE_FIELD     (2)
#define MAXSIZEOF_IEEE802_3_PAYLOAD     (1500)
#define SIZEOF_IEEE802_3_LAN_FCS        (0x02)

#define SIZEOF_IEEE802_3_HDR \
            (SIZEOF_IEEE802_3_DEST_MAC_ADDR + \
             SIZEOF_IEEE802_3_SRC_MAC_ADDR + \
             SIZEOF_IEEE802_3_TYPE_FIELD)

#define SIZEOF_IEEE802_3_OVERHEAD \
            (SIZEOF_IEEE802_3_HDR + \
             SIZEOF_IEEE802_3_LAN_FCS)

#define MAXSIZEOF_IEEE802_3_FRAME \
            (SIZEOF_IEEE802_3_OVERHEAD + \
             MAXSIZEOF_IEEE802_3_PAYLOAD)

#define IEEE802_3_DEST_MAC_ADDR_OFFSET 0

/** Offset from start of IP Packet */
#define SIZEOF_VER_PLUS_HL               (0x01)
#define SIZEOF_TYPE_OF_SERVICE           (0x01)
#define SIZEOF_IP_TOTAL_LEN              (0x02)
#define SIZEOF_IP_ID                     (0x02)
#define SIZEOF_IP_FLAGS_PLUS_FRAG_OFFSET (0x02)
#define SIZEOF_IP_TTL                    (0x01)
#define SIZEOF_IP_PROTOCOL               (0x01)
#define SIZEOF_IP_HDR_CHECKSUM           (0x02)
#define SIZEOF_IP_SRC_ADDR               (0x04)
#define SIZEOF_IP_DEST_ADDR              (0x04)

/** @note the IP header we use is always 20 bytes for
    simplicity. */
#define SIZEOF_IP_HDR_WITHOUT_OPTIONS    (0x14)
#define NUMBEROF_SHORTS_TO_CHECKSUM_IN_IP_HDR (0x0a)

/** Offsets into the IP HDR */
#define IP_DEST_ADDR_OFFSET (0x10)
#define IP_TOTAL_LEN_OFFSET (0x02)
#define IP_TTL_OFFSET       (0x08)


/** First byte of a typical IP packet */
#define IPV4_HDR_ID          (0x45)

/**
 * Supported channel encapsulation type.
 */
typedef enum
{
    TYPE_NONE = 0,          /* Raw VC - always slow path*/
    TYPE_1483BLLC,
    TYPE_1483BVCMUX,
    TYPE_ROUTEDIPLLC,
    TYPE_ROUTEDIPVCMUX,
    NUM_ENCAP_TYPES	/* N.B. Add any new encapsulation types before
		         * this line */
} EncapsulationType;

/** Allow a maximum of 8 VCs */
typedef enum
{
    VC_0 = 0,
    VC_1,
    VC_2,
    VC_3,
    VC_4,
    VC_5,
    VC_6,
    VC_7,
    MAX_NUM_VCS
} VcNumber;

/** Four IP flows per VC */
typedef enum
{
    IP_FLOW0 = 0,
    IP_FLOW1,
    IP_FLOW2,
    IP_FLOW3,
    MAX_FLOWS_PER_VC
}FlowId;

/**
 * This type describes a classifier or modifier template
 */
typedef struct
{
    UINT8 *data;        /**< pointer to the template data */
    unsigned int len;   /**< the length of the template data */
} IxFpCodeletTemplate;

/**
 * @brief Initialise this sub-component. Call this before using any other functions.
 */
IX_STATUS
ixFpCodeletTemplatesInit(void);

/**
 * @brief Create the basic template for the VC based on its encapsulation type
 */
void
ixFpCodeletTemplatesBaseTemplatePairCreate(
    VcNumber vc,
    EncapsulationType type );

/* ------------------------------------------------------------------------
 * @brief Set/Clears the Wan learn command in the xscale copy of a classifier
 *         template.
 * Set firstinstruction to FALSE if learning opcode to be last in the
 * template.
 */
IX_STATUS
ixFpCodeletTemplatesWanMacLearningSet(
    VcNumber vc,
    BOOL enabled,
    BOOL firstInstuction );

/**
 * @brief Set the src or destination ip address for an Ip flow on
 *         an Ip encapsulated VC.
 */
IX_STATUS
ixFpCodeletTemplatesIpFlowSet(
    VcNumber vc,        /* 0 - 7*/
    FlowId   flow,      /* 0 - 3 */
    UINT32 destIpAddr,  /* Classifier matches this addr */
    UINT8  *destMac );  /* Modifier inserts this in Ethernet header */

/**
 * @brief Return the template pair that has been built for the current Vc
 */
IX_STATUS
ixFpCodeletTemplatesTemplatePairGet(
    VcNumber vc,
    IxFpCodeletTemplate *classifier,
    IxFpCodeletTemplate *modifier );

/**
 * @brief Print out the classifier template for this VC
 */
IX_STATUS
ixFpCodeletTemplatesClsPrint(VcNumber vc);

/**
 * @brief Print out the modifier template for this VC
 */
IX_STATUS
ixFpCodeletTemplatesMdfPrint(VcNumber vc);

/**
 * @brief Get the encapsulation name.
 */
char *
ixFpCodeletTemplatesEncapuslationName( EncapsulationType t );

#endif /* IX_FPATHACC_CODELET_TEMPLATES_P_H */

