/**
 * @file IxCryptoAccCodelet.c
 *
 * @date October-31-2002
 *
 * @brief This file contains the implementation of the Crypto Access Codelet.
 *
 * Descriptions of the functions used in this codelet is contained in
 * IxCryptoAccCodelet_p.h
 *
 *
 * -- Intel Copyright Notice --
 * 
 * @par
 * Copyright 2002-2003 Intel Corporation All Rights Reserved.
 * 
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.
 * 
 * @par
 * The Material is protected by worldwide copyright and trade secret laws
 * and treaty provisions. No part of the Material may be used, copied,
 * reproduced, modified, published, uploaded, posted, transmitted,
 * distributed, or disclosed in any way except in accordance with the
 * applicable license agreement .
 * 
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel, except in accordance with the
 * applicable license agreement.
 * 
 * @par
 * Unless otherwise agreed by Intel in writing, you may not remove or
 * alter this notice or any other notice embedded in Materials by Intel
 * or Intel's suppliers or licensors in any way.
 * 
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * 
 * @par
 * -- End Intel Copyright Notice --
 */

/*
 * System include files.
 */

#include <stdio.h>
#include <stdlib.h>

#ifdef __vxworks
#include <end.h>    /* END drivers */
#include <endLib.h> /* END drivers */
#endif /* def __vxworks */


#include <linux/if_crypto.h>
/*
 * User include files.
 */
#include "IxQMgr.h"
#include "IxNpeDl.h"
#include "IxOsServices.h"
#include "ixp425.h"
#include "IxOsCacheMMU.h"
#include "ix_ossl.h"
#include "IxCryptoAccCodelet.h"
#include "IxCryptoAccCodelet_p.h"

/** Bits in byte */
#define BITS_IN_BYTE    8

/** QMgr dispatcher function pointer */
PRIVATE IxQMgrDispatcherFuncPtr ixCryptoAccCodeletDispatcherFunc;

/** QMgr dispatcher active flag */
PRIVATE volatile BOOL ixCryptoAccCodeletDispatcherPollStop = FALSE;

/** Private functions prototype */
PRIVATE void ixCryptoAccCodeletRegisterCB (
            UINT32 cryptoCtxId,
            IX_MBUF *pMbuf, 
            IxCryptoAccStatus status);

PRIVATE void ixCryptoAccCodeletPerformCB (
            UINT32 cryptoCtxId,
            IX_MBUF *pSrcMbuf, 
            IX_MBUF *pDestMbuf,
            IxCryptoAccStatus status);


/** Mbuf Pool */
PRIVATE IX_MBUF *ixCryptoAccCodeletCryptoBufPool;
PRIVATE IX_MBUF *ixCryptoAccCodeletCryptoFreeBufQHead = NULL;
PRIVATE IX_MBUF *ixCryptoAccCodeletCryptoFreeBufQTail = NULL;
PRIVATE INT8 *ixCryptoAccCodeletCryptoBufData;
PRIVATE UINT32 ixCryptoAccCodeletCryptoFreeBufQSize = 0;

/** Registration results  */
BOOL forwardRegisterCBCalled = FALSE;
BOOL reverseRegisterCBCalled = FALSE;

/** Callback function */
IxCryptoAccRegisterCompleteCallback registerCB;
IxCryptoAccPerformCompleteCallback performCB;

/** Length of MBuf payload (in bytes) */
PRIVATE UINT32 packetLength;

/** Performance log variables declaration 
 * Notes:
 * Throughput of the operation performed in codelets is calculated once every 
 * 1000 packets. The result is kept in the performanceLog array and is 
 * printed out after all the crypto perform service requests completed. Number
 * of requests is determined by the PERFORMANCE_WINDOW_SIZE. 1 unit in the 
 * window size represent 1000 requests (packets)
 */
PRIVATE UINT32 performanceLog [PERFORMANCE_WINDOW_SIZE];
PRIVATE UINT32 performanceNumEntries = 0;
PRIVATE UINT32 totalPacket = 0;
PRIVATE UINT32 timeStamp = 0;
PRIVATE UINT32 rateRatio;

/** Guard for once-only init code, when initialised this variable is set to 
 * TRUE 
 */
/* PRIVATE BOOL */ ixCryptoAccCodeletInitialised = FALSE;

/** Flag used when shutting down services, when TRUE, the function will 
 * shutdown 
 */
PRIVATE BOOL trafficFlowHalt = FALSE;

/** Initialised both CtxId to an invalid number as 0 (zero) is a valid ID.
 * The ID ranges from 0 to (IX_CRYPTO_ACC_MAX_ACTIVE_SA_TUNNELS - 1), hence
 * IX_CRYPTO_ACC_MAX_ACTIVE_SA_TUNNELS is invalid.
 */
PRIVATE UINT32 forwardCtxId = IX_CRYPTO_ACC_MAX_ACTIVE_SA_TUNNELS;
PRIVATE UINT32 reverseCtxId = IX_CRYPTO_ACC_MAX_ACTIVE_SA_TUNNELS;

/** Forward and reverse context declaration */
PRIVATE IxCryptoAccCtx forwardCtx, reverseCtx; 

/** Random IV value and IV should remain constant on both
 * directions; encrypt and decrypt.
 */
PRIVATE UINT8 IV[IX_CRYPTO_ACC_MAX_CIPHER_IV_LENGTH] 
/* no swap 
                = {0,2,4,6,
                   1,3,5,7,
                   3,4,1,2,
                   6,7,5,1}; 
*/
/* 4 bytes swap */
                = {6,4,2,0,
                   7,5,3,1,
                   2,1,4,3,
                   1,5,7,6}; 

/* 8 bytes swap  
		= {
                   7,5,3,1,
		   6,4,2,0,
                   1,5,7,6,
                   2,1,4,3
		   }; 
*/
/** Random cipher key */
PRIVATE UINT8 cipherKey[IX_CRYPTO_ACC_MAX_CIPHER_KEY_LENGTH] 
/* 4 bytes swaped */
        = {
	   0x04,0x03,0x02,0x01,
           0x08,0x07,0x06,0x05,
           0x0C,0x0B,0x0A,0x09,
           0x10,0x0F,0x0E,0x0D,
           0x14,0x13,0x12,0x11,
           0x18,0x17,0x16,0x15,
           0x1C,0x1B,0x1A,0x19,
           0x20,0x1F,0x1E,0x1D
	   };

/* never swaped 
        = {0x01,0x02,0x03,0x04,
           0x05,0x06,0x07,0x08,
           0x09,0x0A,0x0B,0x0C,
           0x0D,0x0E,0x0F,0x10,
           0x11,0x12,0x13,0x14,
           0x15,0x16,0x17,0x18,
           0x19,0x1A,0x1B,0x1C,
           0x1D,0x1E,0x1F,0x20};
*/
/** SHA1 key from test case 1 in RFC 2202 */
PRIVATE UINT8 authKey[IX_CRYPTO_ACC_MAX_AUTH_KEY_LENGTH]
        = {0x0B,0x0B,0x0B,0x0B,
           0x0B,0x0B,0x0B,0x0B,
           0x0B,0x0B,0x0B,0x0B,
           0x0B,0x0B,0x0B,0x0B,
           0x0B,0x0B,0x0B,0x0B};


/**
 * @fn ixCryptoAccCodeletBufIsQEmpty (void)
 *
 * @brief This inline function is used to check whether the buffer
 *        queue is empty.
 *
 */
INLINE BOOL
ixCryptoAccCodeletBufIsQEmpty (void)
{
    BOOL status = FALSE;
    if (0 == ixCryptoAccCodeletCryptoFreeBufQSize)
    { 
        status = TRUE;
    }
    return status;
}


/**
 * @fn ixCryptoAccCodeletBufIsQFull (void)
 *
 * @brief This inline function is used to check whether the buffer
 *        queue is full.
 *
 */
INLINE BOOL
ixCryptoAccCodeletBufIsQFull(void)
{
    BOOL status = FALSE;
    if (IX_CRYPTOACC_CODELET_CRYPTO_MBUF_POOL_SIZE 
        == ixCryptoAccCodeletCryptoFreeBufQSize)
    { 
        status = TRUE;
    }
    return status;
}


/**
 * @fn ixCryptoAccCodeletCryptoMbufToQTailAdd (IX_MBUF *pMbufToAdd)
 *
 * @brief This function adds a mbuf to the tail of the queue.
 *
 */
void ixCryptoAccCodeletCryptoMbufToQTailAdd (IX_MBUF *pMbufToAdd)
{
    IX_MBUF *pMbuf;
    
    if (!ixCryptoAccCodeletBufIsQFull())
    {
        pMbuf = ixCryptoAccCodeletCryptoFreeBufQTail;
        if (NULL != ixCryptoAccCodeletCryptoFreeBufQTail) 
        {
            /* Add pMbufToAdd in the last mbuf in the chain */
            IX_MBUF_NEXT_PKT_IN_CHAIN_PTR(pMbuf) = pMbufToAdd;    
        } /* End of if ixCryptoAccCodeletCryptoFreeBufQTail */
    
        /* Update the ixCryptoAccCodeletFreeBufQ with the last mbuf */
        ixCryptoAccCodeletCryptoFreeBufQTail = pMbufToAdd;
        IX_MBUF_NEXT_PKT_IN_CHAIN_PTR(pMbufToAdd) = NULL;
        ixCryptoAccCodeletCryptoFreeBufQSize++;
    } 
    else /* Buffer queue is full */
    {
        printf ("\nBuffer Queue is full\n");
    } /* End of if-else ixCryptoAccCodeletBufIsQFull() */
} /* End of ixCryptoAccCodeletCryptoMbufToQTailAdd () function */


/**
 * @fn ixCryptoAccCodeletCryptoMbufFromQHeadRemove (IX_MBUF **pMbufToRem)
 *
 * @brief This function removes a mbuf from the head of the queue.
 */
void ixCryptoAccCodeletCryptoMbufFromQHeadRemove(IX_MBUF **pMbufToRem)
{
    if (!ixCryptoAccCodeletBufIsQEmpty())
    {
        *pMbufToRem = ixCryptoAccCodeletCryptoFreeBufQHead;
        ixCryptoAccCodeletCryptoFreeBufQHead 
            = IX_MBUF_NEXT_PKT_IN_CHAIN_PTR((*pMbufToRem));
        ixCryptoAccCodeletCryptoFreeBufQSize--;
    } 
    else
    {
        *pMbufToRem = NULL;
        printf ("\nBuffer Queue is empty\n");   
    } /* end of if-else ixCryptoAccCodeletBufIsQEmpty */
} /* End of ixCryptoAccCodeletCryptoMbufFromQHeadRemove () function */


/**
 * @fn IX_STATUS ixCryptoAccCodeletCryptoMemPoolInit (UINT32 pktLen)
 *
 * @brief Initialise Crypto Access Codelet MBUF pool
 */
IX_STATUS ixCryptoAccCodeletCryptoMemPoolInit (UINT32 pktLen)
{
    UINT32 bufNo;
    IX_MBUF *pMbuf;
    UINT32 mBlkSize = IX_XSCALE_CACHE_LINE_SIZE * ((sizeof (IX_MBUF) + 
                        (IX_XSCALE_CACHE_LINE_SIZE - 1)) /
                        IX_XSCALE_CACHE_LINE_SIZE);    
    UINT32 pktSize = pktLen + IX_CRYPTOACC_CODELET_AUTH_DIGEST_LEN;
    UINT32 mDataSize = IX_XSCALE_CACHE_LINE_SIZE * ((pktSize + 
                        (IX_XSCALE_CACHE_LINE_SIZE - 1)) /
                        IX_XSCALE_CACHE_LINE_SIZE);

    ixCryptoAccCodeletCryptoBufPool = NULL;
    ixCryptoAccCodeletCryptoBufData = NULL;
    
    if (NULL == (ixCryptoAccCodeletCryptoBufPool 
                    = IX_ACC_DRV_DMA_MALLOC (
                        IX_CRYPTOACC_CODELET_CRYPTO_MBUF_POOL_SIZE
                        * mBlkSize)))
    {
        printf ("Error allocating mBuf pool for Crypto Codelet!\n");
        return IX_FAIL;
    }

    if (NULL == (ixCryptoAccCodeletCryptoBufData 
                    = IX_ACC_DRV_DMA_MALLOC(
                        IX_CRYPTOACC_CODELET_CRYPTO_MBUF_POOL_SIZE *
                        mDataSize)))
    {
        IX_CRYPTO_ACC_CODELET_DRV_DMA_FREE(
            ixCryptoAccCodeletCryptoBufPool,
            IX_CRYPTOACC_CODELET_CRYPTO_MBUF_POOL_SIZE * mBlkSize);

        printf ("Error allocating mBuf data pool for Crypto Codelet!\n");
        return IX_FAIL;
    }

    pMbuf = ixCryptoAccCodeletCryptoBufPool;
    /* Initilise mbuf pool head pointer */
    ixCryptoAccCodeletCryptoFreeBufQHead  = pMbuf;
    
    /* Format our mBufs */
    for (bufNo = 0;
        bufNo < IX_CRYPTOACC_CODELET_CRYPTO_MBUF_POOL_SIZE;
        bufNo++)
    {
        /* Assign memory block to mData pointer in mbuf */ 
        IX_MBUF_MDATA (pMbuf) 
            = &ixCryptoAccCodeletCryptoBufData[bufNo * mDataSize];
    
        /* Initialise mbuf length */
        IX_MBUF_MLEN(pMbuf) = mDataSize;
        
        /* No chains */
        IX_MBUF_NEXT_BUFFER_IN_PKT_PTR(pMbuf)= NULL;
    
        /* Add newly formated buffer to our free queue */
        ixCryptoAccCodeletCryptoMbufToQTailAdd (pMbuf);
        pMbuf = (IX_MBUF *)((UINT8 *) pMbuf + mBlkSize);
    }

    return IX_SUCCESS;
} /* End of ixCryptoAccCodeletCryptoMemPoolInit () function */


/**
 * @fn IX_STATUS ixCryptoAccCodeletCryptoMemPoolFree (UINT32 pktLen)
 *
 * @brief Free all resources in Crypto Access Codelet MBUF pool
 *
 */
void ixCryptoAccCodeletCryptoMemPoolFree (UINT32 pktLen)
{
    UINT32 mBlkSize = IX_XSCALE_CACHE_LINE_SIZE * ((sizeof (IX_MBUF) + 
                        (IX_XSCALE_CACHE_LINE_SIZE - 1)) /
                        IX_XSCALE_CACHE_LINE_SIZE);
    UINT32 pktSize = pktLen + IX_CRYPTOACC_CODELET_AUTH_DIGEST_LEN;
    UINT32 mDataSize = IX_XSCALE_CACHE_LINE_SIZE * ((pktSize + 
                        (IX_XSCALE_CACHE_LINE_SIZE - 1)) /
                        IX_XSCALE_CACHE_LINE_SIZE);          
                        
    IX_CRYPTO_ACC_CODELET_DRV_DMA_FREE(
        ixCryptoAccCodeletCryptoBufPool,
        IX_CRYPTOACC_CODELET_CRYPTO_MBUF_POOL_SIZE * mBlkSize);

    IX_CRYPTO_ACC_CODELET_DRV_DMA_FREE(
        ixCryptoAccCodeletCryptoBufData,
        IX_CRYPTOACC_CODELET_CRYPTO_MBUF_POOL_SIZE * mDataSize);
    
    ixCryptoAccCodeletCryptoFreeBufQHead = NULL;
    ixCryptoAccCodeletCryptoFreeBufQTail = NULL;
    ixCryptoAccCodeletCryptoFreeBufQSize = 0;
} /* End of ixCryptoAccCodeletMemPoolFree () function */


/**
 * @fn void ixCryptoAccCodeletDispatcherPoll (void)
 *
 * @brief This function polls the queues when poll mode set in the 
 *        function ixCryptoAccCodeletDispatcherStart. 
 */
void ixCryptoAccCodeletDispatcherPoll (void)
{
    /* Set dispatcher stop flag to FALSE */
    ixCryptoAccCodeletDispatcherPollStop = FALSE;
    
    while (1)
    {
        /* Exit thread if stop flag is set to TRUE */
        if (ixCryptoAccCodeletDispatcherPollStop)
        {
            break;  /* Exit the thread */
        }

#ifndef __vxworks
        /* In linux platform, the QMgr dispatcher thread needs to be put into 
         * sleep mode for 1 milisecond after each loop, in order to allow  
         * others thread to run.
         */
/*
        ixOsServTaskSleep (1);
*/
#endif /* ndef __vxworks */
{
	int kk,tt=0;
	for(kk=0;kk<10;kk++) tt++;
}
        /* The crypto NPE uses queues 29 and 30. */
        (*ixCryptoAccCodeletDispatcherFunc) (IX_QMGR_QUELOW_GROUP);
    }
    
} /* End of function ixCryptoAccCodeletDispatcherPoll() */


/**
 * @fn ixCryptoAccCodeletDispatcherStart (BOOL useInterrupt)
 *
 * @brief  This function starts the Queue manager dispatcher.
 *
 * @return IX_STATUS 
 */
IX_STATUS ixCryptoAccCodeletDispatcherStart (BOOL useInterrupt)
{
    UINT32 dispatchtid;

    /* Get QMgr dispatcher function pointer */
    ixQMgrDispatcherLoopGet (&ixCryptoAccCodeletDispatcherFunc);
    
    if (useInterrupt)   /* Interrupt mode */
    {
        printf ("\nStarting Interrupt mode for cryptoAcc!\n");
        
        /* Hook the QM QLOW dispatcher to the interrupt controller. 
         * The crypto NPE uses queues 29 and 30.
         */
        if (IX_SUCCESS != ixOsServIntBind (IXP425_INT_LVL_QM1,
                              (IxVoidFnVoidPtr)ixCryptoAccCodeletDispatcherFunc,
                              (void *)IX_QMGR_QUELOW_GROUP))
        {
            printf ("Failed to bind to QM1 interrupt\n");
            return IX_FAIL;
        }
    }
    else  /* Poll mode */
    {
        printf ("\nStarting Polling mode for cryptoAcc!\n");
        if (IX_OSSL_ERROR_SUCCESS != ix_ossl_thread_create (
                                        (ix_ossl_thread_entry_point_t)
                                        ixCryptoAccCodeletDispatcherPoll,
                                        NULL,
                                        &dispatchtid))
        {
            printf ("Error spawning dispatch task\n");
            return IX_FAIL;
        }

        ix_ossl_thread_set_priority (
            dispatchtid, 
            IX_CRYPTOACC_CODELET_QMR_DISPATCHER_PRIORITY);
    }

    return IX_SUCCESS;
} /* End of ixCryptoAccCodeletDispatcherStart () function */


/**
 * @fn ixCryptoAccCodeletRegisterCtx(
 *         UINT32 opType,
 *         IxCryptoAccCtx *forwardCtx,
 *         IxCryptoAccCtx *reverseCtx,
 *         UINT32 *forwardCtxId,
 *         UINT32 *reverseCtxId)
 *
 * @brief  This function registers to the Crypto Access Component for
 *         three different operation which demonstrates the capability
 *         of the security hardware accelerator.
 *         The types of operation are:
 *          1 - Encryption and Decryption
 *          2 - Authentication: Calculation and Check
 *          3 - Encryption/Decryption with authentication
 */
IX_STATUS ixCryptoAccCodeletRegisterCtx (
            UINT32 opType,
            IxCryptoAccCtx *pForwardCtx,
            IxCryptoAccCtx *pReverseCtx,
            UINT32 *pForwardCtxId,
            UINT32 *pReverseCtxId)
{
    IxCryptoAccStatus status;
    IX_MBUF *pPrimaryMbuf = NULL;
    IX_MBUF *pSecondaryMbuf = NULL;
    UINT16 i = 0;
    
    /* Set up forward and reverse context and then register it */
    switch (opType)
    {
        case IX_CRYPTOACC_CODELET_ENCRYPT:
            pForwardCtx->operation = IX_CRYPTO_ACC_OP_ENCRYPT;
            pReverseCtx->operation = IX_CRYPTO_ACC_OP_DECRYPT;
            break;
        
        case IX_CRYPTOACC_CODELET_AUTH:
            pForwardCtx->operation = IX_CRYPTO_ACC_OP_AUTH_CALC;
            pReverseCtx->operation = IX_CRYPTO_ACC_OP_AUTH_CHECK;
            break;
            
        case IX_CRYPTOACC_CODELET_ENCRYPT_AUTH:
            pForwardCtx->operation = IX_CRYPTO_ACC_OP_ENCRYPT_AUTH;
            pReverseCtx->operation = IX_CRYPTO_ACC_OP_AUTH_DECRYPT;
            break;
            
        default:
            printf ("\nInvalid Operation Type, %d\n", opType);
            return IX_FAIL;
    } /* End of switch (opType) */

    /* Primary Mbuf and Secondary Mbuf is required for authentication 
     * and combined service for forward context
     */
    if ((IX_CRYPTOACC_CODELET_AUTH == opType) || 
        (IX_CRYPTOACC_CODELET_ENCRYPT_AUTH == opType))
    {
        ixCryptoAccCodeletCryptoMbufFromQHeadRemove (&pPrimaryMbuf);
        ixCryptoAccCodeletCryptoMbufFromQHeadRemove (&pSecondaryMbuf);

        /* Check if the mbufs are NULL */
        if ((NULL == pPrimaryMbuf) || (NULL == pSecondaryMbuf))
        {
            printf ("\nInvalid mbufs!\n");
            return IX_FAIL;
        }
    }
  
    /* Register the forward context */
    status = ixCryptoAccCtxRegister (
                 pForwardCtx,
                 pPrimaryMbuf,
                 pSecondaryMbuf,
                 registerCB,
                 performCB,
                 pForwardCtxId);
                 
    if (IX_CRYPTO_ACC_STATUS_SUCCESS != status)
    {
        printf ("Forward registration Failed. Status = %d\n", status);
        return IX_FAIL;
    }

    i = 0;

    /* Wait for the forward register callback to return */
    while (!forwardRegisterCBCalled)
    {
	int kk,tt=0;
	for(kk=0;kk<10;kk++) tt++;
/*
        ixOsServTaskSleep (IX_CRYPTOACC_CODELET_REGISTER_WAIT_TIME);
*/        
        /* Wait for IX_CRYPTOACC_CODELET_MAX_TIME_WAIT times before return 
         * fail status 
         */
        if (IX_CRYPTOACC_CODELET_MAX_TIME_WAIT < i)
        {
            printf ("\nRegistration incomplete, timeout for forward callback\n");
            return IX_FAIL;
        }
        i++;
    } /* end of while (!forwardRegisterCBCalled) */

    /* Primary Mbuf and Secondary Mbuf are required for authentication 
     * and combined service for reverse context
     */
    if ((IX_CRYPTOACC_CODELET_AUTH == opType) || 
    (IX_CRYPTOACC_CODELET_ENCRYPT_AUTH == opType))
    {
        ixCryptoAccCodeletCryptoMbufFromQHeadRemove (&pPrimaryMbuf);
        ixCryptoAccCodeletCryptoMbufFromQHeadRemove (&pSecondaryMbuf);
    
        /* Check if the mbufs are NULL */
        if ((NULL == pPrimaryMbuf) || (NULL == pSecondaryMbuf))
        {
            printf ("\nInvalid mbufs!\n");
            return IX_FAIL;
        }
    }
    
    /* Register the reverse context */
    status = ixCryptoAccCtxRegister (
                 pReverseCtx,
                 pPrimaryMbuf,
                 pSecondaryMbuf,
                 registerCB,
                 performCB,
                 pReverseCtxId);
                 
    if (IX_CRYPTO_ACC_STATUS_SUCCESS != status)
    {
        printf ("Reverse registration Failed. Status = %d\n", status);
        return IX_FAIL;
    }

    i = 0;

    /* Wait for the reverse register callback to return */
    while (!reverseRegisterCBCalled)
    {
	int kk,tt=0;
	for(kk=0;kk<10;kk++) tt++;
/*
        ixOsServTaskSleep (IX_CRYPTOACC_CODELET_REGISTER_WAIT_TIME);
*/        
        if (IX_CRYPTOACC_CODELET_MAX_TIME_WAIT < i)
        {
            printf ("\nRegistration incomplete, timeout for reverse callback\n");
            return IX_FAIL;
        }
        i++;
    } /* end of while (!reverseRegisterCBCalled) */
    
    return IX_SUCCESS;
} /* End of ixCryptoAccCodeletRegisterCtx () function */


/**
 * @fn ixCryptoAccCodeletUnregisterCtx (void)
 *
 * @brief  This function unregisters all the contexts registered.
 */
void ixCryptoAccCodeletUnregisterCtx (void)
{
    if (IX_CRYPTO_ACC_STATUS_SUCCESS != 
        ixCryptoAccCtxUnregister (forwardCtxId))
    {
        printf ("\nFailed to unregister forward context %d\n",forwardCtxId);
    }

    if (IX_CRYPTO_ACC_STATUS_SUCCESS != 
        ixCryptoAccCtxUnregister (reverseCtxId))
    {
        printf ("\nFailed to unregister reverse context %d\n",reverseCtxId);
    }

    forwardRegisterCBCalled = FALSE;
    reverseRegisterCBCalled = FALSE;
    
    /* Initialised to an invalid number, zero is a valid ID */
    forwardCtxId = IX_CRYPTO_ACC_MAX_ACTIVE_SA_TUNNELS;
    reverseCtxId = IX_CRYPTO_ACC_MAX_ACTIVE_SA_TUNNELS;
}

/**
 * @fn  ixCryptoAccCodeletInit (void)
 * @brief This function initialises the neccessary components.
 *
 */
IX_STATUS ixCryptoAccJackInit (void)
{
    if (ixCryptoAccCodeletInitialised) 
    {
        printf ("Crypto codelet already initialised\n");
        return IX_SUCCESS;
    }

    /* Initialise stop flag to TRUE */
/*
    ixCryptoAccCodeletDispatcherPollStop = TRUE;
*/
#if 0
    /* Initialise NPEs */
    if (IX_SUCCESS != ixNpeDlNpeInitAndStart (IX_CRYPTOACC_CODELET_NPEC_IMAGE_ID))
    {
        printf ("Error initialising NPE!\n");
        return IX_FAIL;
    }
#endif
    /* Initialise cryptoAcc component */
    if (IX_CRYPTO_ACC_STATUS_SUCCESS != ixCryptoAccInit ())
    {
        printf ("Error initialising Crypto access component!\n");
        return IX_FAIL;
    }

    /* Initialize crypto context, both the forward and reverse 
     * context are the same. The only difference is the operation which
     * will be initialize in ixCryptoAccCodeletRegisterCtx () 
     */    
    forwardCtx.cipherCtx.cipherAlgo 
        = reverseCtx.cipherCtx.cipherAlgo 
        = IX_CRYPTOACC_CODELET_CIPHER_ALGO;
        
    forwardCtx.cipherCtx.cipherMode 
        = reverseCtx.cipherCtx.cipherMode 
        = IX_CRYPTOACC_CODELET_CIPHER_MODE;
        
    forwardCtx.cipherCtx.cipherKeyLen 
        = reverseCtx.cipherCtx.cipherKeyLen 
        = IX_CRYPTOACC_CODELET_CIPHER_KEY_LEN;
        
    forwardCtx.cipherCtx.cipherBlockLen 
        = reverseCtx.cipherCtx.cipherBlockLen 
        = IX_CRYPTOACC_CODELET_CIPHER_BLOCK_LEN;
        
    forwardCtx.cipherCtx.cipherInitialVectorLen 
        = reverseCtx.cipherCtx.cipherInitialVectorLen 
        = IX_CRYPTOACC_CODELET_CIPHER_IV_LEN;
        
    forwardCtx.cipherCtx.securityParameterIndex 
        = reverseCtx.cipherCtx.securityParameterIndex 
        = 0; /* a random number */
        
    memcpy (forwardCtx.cipherCtx.key.cipherKey,
        cipherKey,
        IX_CRYPTOACC_CODELET_CIPHER_KEY_LEN);

    {
	unsigned long *pp = (unsigned long *)(forwardCtx.cipherCtx.key.cipherKey);
	int ii;

	for(ii= 0; ii < IX_CRYPTOACC_CODELET_CIPHER_KEY_LEN / 4; ii++)
		printk("\ncipher key %8x",*(pp+ii));
	printk("\n");
    }

    memcpy (reverseCtx.cipherCtx.key.cipherKey,
        cipherKey,
        IX_CRYPTOACC_CODELET_CIPHER_KEY_LEN);
    
    forwardCtx.authCtx.authAlgo 
        = reverseCtx.authCtx.authAlgo 
        = IX_CRYPTOACC_CODELET_AUTH_ALGO;
        
    forwardCtx.authCtx.authDigestLen 
        = reverseCtx.authCtx.authDigestLen 
        = IX_CRYPTOACC_CODELET_AUTH_DIGEST_LEN;
        
    forwardCtx.authCtx.authKeyLen 
        = reverseCtx.authCtx.authKeyLen 
        = IX_CRYPTOACC_CODELET_AUTH_KEY_LEN;
        
    /* Using test case 1 in rfc 2202 */
    memcpy (forwardCtx.authCtx.key.authKey,
        authKey,
        IX_CRYPTOACC_CODELET_AUTH_KEY_LEN);
    memcpy (reverseCtx.authCtx.key.authKey,
        authKey,
        IX_CRYPTOACC_CODELET_AUTH_KEY_LEN);
    
    forwardCtx.useDifferentSrcAndDestMbufs 
        = reverseCtx.useDifferentSrcAndDestMbufs 
        = FALSE;
#if 0
    /* Start the Queue Manager dispatcher */   
    if ( IX_SUCCESS != 
     ixCryptoAccCodeletDispatcherStart (IX_CRYPTOACC_CODELET_QMGR_DISPATCHER_MODE))
    {
        printf ("Error starting queue manager dispatcher!\n");
        return IX_FAIL;
    }
#endif    
    /* set flag to TRUE to indicate codelet has been initialised successfully */
    ixCryptoAccCodeletInitialised = TRUE;
    printf ("cryptoAcc Codelet Initialiasation complete!\n\n");

    return IX_SUCCESS;
} /* End of ixCryptoAccCodeletInit () function */


/** 
 * @fn ixCryptoAccCodeletRegisterCB(
 *         UINT32 cryptoCtxId,
 *         IX_MBUF *pMbuf, 
 *         IxCryptoAccStatus status)
 *
 * @brief  Cryptographic Context registration callback function to be called
 *         in QMgr Dispatcher context when the registration completed.
 *
 * @param "UINT32 [in] cryptoCtxId" - Crypto Context ID
 * @param "IX_MBUF* [in] pMbuf" - mbuf pointer
 * @param "IxCryptoAccStatus [in] status" - status of callback
 *
 * @return None
 */
PRIVATE void ixCryptoAccCodeletRegisterCB (
        UINT32 cryptoCtxId,
        IX_MBUF *pMbuf, 
        IxCryptoAccStatus status)
{
    BOOL *pFlagToUse;

    /* Determine callback flag to be used from the context crypto registered. 
     * If forward context, forwardRegisterCBCalled is used; otherwise 
     * reverseRegisterCBCalled is used.  
     */
    if (cryptoCtxId == forwardCtxId)
    {
        pFlagToUse = &forwardRegisterCBCalled;
    }
    else /* (cryptoCtxId != forwardCtxId) */
    {   
        if (cryptoCtxId == reverseCtxId)
        {
            pFlagToUse = &reverseRegisterCBCalled;
        }
        else
        {
            ixOsServLog(LOG_ERROR, "\nInvalid ContextId %d\n", 
                cryptoCtxId, 0, 0, 0, 0, 0);
            return;
        } /* end of if-else (cryptoCtxId == reverseCtxId) */
    } /* end of if-else (cryptoCtxId == forwardCtxId) */

    /* We are expecting two callbacks for authentication and combined service 
     * The pMbuf could be null for encrypt/decrypt ONLY operation.
     */
    if (NULL != pMbuf)
    {
        /* Must return the buffer to the pool irrespective of the status */
        ixCryptoAccCodeletCryptoMbufToQTailAdd (pMbuf);
    }

    if (IX_CRYPTO_ACC_STATUS_SUCCESS == status)
    {
        *pFlagToUse = TRUE;
        ixOsServLog(LOG_MESSAGE, 
            "RegistrationCB is successful. Context %d\n", 
            cryptoCtxId, 
            0, 0, 0, 0, 0);
    } /* End of if (IX_CRYPTO_ACC_STATUS_SUCCESS == status) */
    else 
    {
        if (IX_CRYPTO_ACC_STATUS_WAIT == status)
        {
            ixOsServLog(LOG_MESSAGE, 
                "Register not completed yet; Context %d\n", 
                cryptoCtxId, 
                0, 0, 0, 0, 0);
        }
        else
        {
            ixOsServLog(LOG_WARNING, "Registration failed; Context %d\n", 
                cryptoCtxId, 0, 0, 0, 0, 0);
        }
    } /* End of if else (IX_CRYPTO_ACC_STATUS_SUCCESS == status) */
} /* End of ixCryptoAccCodeletRegisterCB () function */

/* Jack ++ */
__inline__ UINT32 ixOsServWordSwap(UINT32 wordIn)
{
    /* Storage for the swapped word */
    UINT32 wordOut;

    /* wordIn = A, B, C, D */
    __asm__ (" eor r1, %1, %1, ror #16;"  /* R1 =      A^C, B^D, C^A, D^B */
             " bic r1, r1, #0x00ff0000;"  /* R1 =      A^C, 0  , C^A, D^B */
             " mov %0, %1, ror #8;"       /* wordOut = D,   A,   B,   C   */
             " eor %0, %0, r1, lsr #8;"   /* wordOut = D,   C,   B,   A   */
             : "=r" (wordOut) : "r" (wordIn) : "r1");

    return wordOut;
}


/**
 * @fn ixCryptoAccCodeletPerformCB (
 *        UINT32 cryptoCtxId,
 *        IX_MBUF *pSrcMbuf, 
 *        IX_MBUF *pDestMbuf,
 *        IxCryptoAccStatus status)
 *
 * @brief  Crypto services request callback fuction to be called when the 
 *         crypto service s request completed. Performance rate is calculated 
 *         in this function.
 *
 * @param "UINT32 [in] cryptoCtxId" - Crypto Context ID
 * @param "IX_MBUF* [in] pSrcMbuf" - source mbuf pointer
 * @param "IX_MBUF* [in] pDestMbuf" - destination mbuf pointer
 * @param "IxCryptoAccStatus [in] status" - status of callback
 *
 * @return None
 */
PRIVATE void ixCryptoAccCodeletPerformCB (
        UINT32 cryptoCtxId,
        IX_MBUF *pSrcMbuf, 
        IX_MBUF *pDestMbuf,
        IxCryptoAccStatus status)
{
    IxCryptoAccStatus retStatus;
    UINT32 timeNow;     /**< Current timestamp */
    UINT32 timeDiff;    /**< Time taken to process 1000 packets in XSCALE TICK 
                         * unit
                         */ 

    switch (status)
    {
        case IX_CRYPTO_ACC_STATUS_SUCCESS:
            if (IX_CRYPTOACC_CODELET_BATCH_LEN <= totalPacket)
            {
                timeNow = ixOsServTimestampGet();
    {
	unsigned long *pp = (unsigned long *)(&pSrcMbuf->mBlkHdr.mData[0]);
	int ii;

	for(ii= 0; ii < packetLength / 4; ii++)
		printk("after data %8x\n",*(pp+ii));
    }
            
                /* Check if the timer wrap over, get the time taken and  
                 * divided it by 66 to get microseconds as XScale tick is 66MHz 
                 */
                if (timeNow < timeStamp)
                {
                    timeDiff = ((0xffffffff - timeStamp + timeNow + 1)/
                        IX_CRYPTOACC_CODELET_XSCALE_TICK);
                }
                else
                {
                    timeDiff = ((timeNow - timeStamp) /
                        IX_CRYPTOACC_CODELET_XSCALE_TICK);
                } /* end of if-else (timeNow < timeStamp) */
                
                /* Store performance rate for every 1000 packets */
                if (PERFORMANCE_WINDOW_SIZE > performanceNumEntries)
                {
                    performanceLog[performanceNumEntries] = rateRatio / timeDiff;
                    performanceNumEntries++;
                }
        
                /* Reset totalPacket for another measurement */
                totalPacket = 0;
        
                /* Start the timeStamp */
                timeStamp = timeNow;
            }
            else
            {
                totalPacket++;
            } /* End of if-else totalPacket */
            break;
      
        case IX_CRYPTO_ACC_STATUS_AUTH_FAIL:
            ixOsServLog(LOG_WARNING, "Authentication Failed\n",
                    0, 0, 0, 0, 0, 0);
            break;
    
        case IX_CRYPTO_ACC_STATUS_FAIL:
            ixOsServLog(LOG_WARNING, "Operation Failed\n", 
                    0, 0, 0, 0, 0, 0);
            break;
    
        default:
            ixOsServLog(LOG_ERROR, "Error in status message, %d\n", 
                status, 0, 0, 0, 0, 0);
    } /* End of switch (status) */

    /* Sample size more than user-specified size, halt the traffic */    
    if (PERFORMANCE_WINDOW_SIZE <= performanceNumEntries)
    {
        trafficFlowHalt = TRUE;
    }

    /* If we are shutting down, return mbuf to the pool and return */
    if (trafficFlowHalt)
    {
        /* Return the buffer(s) to the pool */
        if (NULL != pSrcMbuf)
        {
            ixCryptoAccCodeletCryptoMbufToQTailAdd (pSrcMbuf);
        }
        if (NULL != pDestMbuf)
        {
            ixCryptoAccCodeletCryptoMbufToQTailAdd (pDestMbuf);
        }
        return;
    }

    /* Switch Contexts ID, based on the context ID to determine what operation 
     * should be done on the packet. If cryptoCtxId is a forwardCtxId, the 
     * packet is encrypted / authenticated, send the packet to NPE again for
     * decryption / auth_check, and vise-versa.
     */
    if (cryptoCtxId == forwardCtxId)
    {
        /* The perform will keep on trying to write to the queue and will wait
         * if the queue is full
         */
        do
        {
            retStatus = ixCryptoAccAuthCryptPerform (
                            reverseCtxId,
                            pSrcMbuf,
                            pDestMbuf,
                            0,
                            packetLength,
                            0,
                            packetLength,
                            packetLength,
                            IV);
                                
        } while (IX_CRYPTO_ACC_STATUS_QUEUE_FULL == retStatus);
        /* End of do-while loop */
    } 
    else    /* cryptoCtxId == reverseCtxId */
    {
        do
        {
            retStatus = ixCryptoAccAuthCryptPerform (
                            forwardCtxId,
                            pSrcMbuf,
                            pDestMbuf,
                            0,
                            packetLength,
                            0,
                            packetLength,
                            packetLength,
                            IV);    
                 
        } while (IX_CRYPTO_ACC_STATUS_QUEUE_FULL == retStatus);
        /* End of while loop */
    } /* End of if else (cryptoCtxId == forwardCtxId) */
} /* End of ixCryptoAccCodeletPerformCB () function */


/**
 * @fn ixCryptoAccCodeletMain (
 *          IxCryptAccCodeletOperation operationType,
 *          UINT32 packetLen)
 *
 * @brief  This is the entry point function to the codelet to choose the 
 *         operation for the codelet and packet length to be used. 
 *         This is the main function of the codelet where crypto contexts
 *         registration and crypto perform services are done. Based on the  
 *         selected operation and selected packet length, packets are sent to 
 *         cryptoAcc for processing. 
 */
static int check_f = 0;

IX_STATUS ixCryptoAccCodeletMain ( struct crypto_request *req 
	/*     IxCryptAccCodeletOperation operationType,
               UINT32 packetLen */)
{
    IxCryptAccCodeletOperation operationType = 1;
    UINT32 packetLen = 64;

    IxCryptoAccStatus status;
    IX_MBUF *pMbuf;
    UINT32 pktNo;
    UINT32 i;
    UINT32 j;
    char data;


  if(!check_f) check_f = 1;
  else return 0;


    if ((IX_CRYPTOACC_CODELET_ENCRYPT > operationType) || 
        (IX_CRYPTOACC_CODELET_ENCRYPT_AUTH < operationType)) 
    {
        printf ("\nInvalid operation type selected!\n\n");
        ixCryptoAccCodeletUsageShow ();
        return IX_FAIL;
    }
    
    if ((IX_CRYPTOACC_CODELET_MAX_CRYPTO_PKT_SIZE >= packetLen) 
        && (IX_CRYPTOACC_CODELET_MIN_CRYPTO_PKT_SIZE <= packetLen))
    {
        /* IX_CRYPTOACC_CODELET_CIPHER_ALGO is defined at USER SETTING
         * in IxCryptoAccCodelet.h
         */
        if (IX_CRYPTO_ACC_CIPHER_AES == IX_CRYPTOACC_CODELET_CIPHER_ALGO)
        {
            /* Check if packet length is multiple of cipher block length */
            if (0 != (packetLen %  IX_CRYPTO_ACC_AES_BLOCK_128))
            {
                printf ("\nPacket length is not mulitple of AES cipher "
                        "block length - 16 bytes.\n\n");
                ixCryptoAccCodeletUsageShow ();
                return IX_FAIL;
            }
        }
        else /* DES or 3DES cipher algorithm */
        {
            /* Check if packet length is multiple of cipher block length */
            if (0 != (packetLen %  IX_CRYPTO_ACC_DES_BLOCK_64))
            {
                printf ("\nPacket length is not mulitple of DES cipher block "
                        "length - 8 bytes.\n\n");
                ixCryptoAccCodeletUsageShow ();
                return IX_FAIL;
            }
        }
    }
    else /* invalid packet length */
    {
        printf ("\nInvalid packet length selected!\n\n");
        ixCryptoAccCodeletUsageShow ();
        return IX_FAIL;    
    }
    
    /* Initialisation */
    if (IX_SUCCESS !=  ixCryptoAccJackInit ())
    {
        printf ("Initialisation Failed.\n");
        return IX_FAIL;
    }

    registerCB = ixCryptoAccCodeletRegisterCB;
    performCB = ixCryptoAccCodeletPerformCB;
    
    printf ("************************************************************\n");
    printf ("\tSecurity Hardware Accelerator Codelet\n");
    printf ("************************************************************\n\n");

    /* Store packet length into global variable for the use in perform CB */
    packetLength = packetLen;

    /* Alloc Mbufs pool */
    if (IX_SUCCESS 
        != ixCryptoAccCodeletCryptoMemPoolInit (packetLength))
    {
        printf ("Mbuf pool init failed\n");
        return IX_FAIL;
    }

    /* Calculate rate ratio used in performance calculation based 
     * on buffer size chosen */
    if (IX_CRYPTOACC_CODELET_ENCRYPT != operationType)
    {
        /* Include digest length into rateRatio as authentication
         * operation will insert digest into payload based on the length
         * of IX_CRYPTOACC_CODELET_AUTH_DIGEST_LEN
         */
        rateRatio = ((packetLength +
                      IX_CRYPTOACC_CODELET_AUTH_DIGEST_LEN)
                      * BITS_IN_BYTE * IX_CRYPTOACC_CODELET_BATCH_LEN);
    }
    else
    {
        rateRatio = (packetLength * BITS_IN_BYTE * IX_CRYPTOACC_CODELET_BATCH_LEN);
    }

    /* Registration */
    if (IX_SUCCESS != ixCryptoAccCodeletRegisterCtx(
                           operationType,
                           &forwardCtx,
                           &reverseCtx,
                           &forwardCtxId,
                           &reverseCtxId))
    {
        printf ("Registration Failed.\n");
        return IX_FAIL;
    }
/*
    for(pktNo = 0; 
        pktNo < IX_CRYPTOACC_CODELET_CRYPTO_MBUF_POOL_SIZE; 
        pktNo++)
*/
    {
        /* Get an mbuf */
        ixCryptoAccCodeletCryptoMbufFromQHeadRemove (&pMbuf);

        /* Check if the mbuf is NULL */
        if (NULL == pMbuf)
        {
            printf ("\nInvalid mbuf!\n");
            return IX_FAIL;
        }
        
        /* copy data into mBuf from the ASCII table 0 (0x30) to z (0x7a)*/
/*
	for(j=0;j<8;j++) pMbuf->mBlkHdr.mData[0*8+j] = (char)0x37 -j ;
	for(j=0;j<8;j++) pMbuf->mBlkHdr.mData[1*8+j] = (char)0x3f -j ;
	for(j=0;j<8;j++) pMbuf->mBlkHdr.mData[2*8+j] = (char)0x47 -j ;
	for(j=0;j<8;j++) pMbuf->mBlkHdr.mData[3*8+j] = (char)0x4f -j ;
	for(j=0;j<8;j++) pMbuf->mBlkHdr.mData[4*8+j] = (char)0x57 -j ;
	for(j=0;j<8;j++) pMbuf->mBlkHdr.mData[5*8+j] = (char)0x5f -j ;
	for(j=0;j<8;j++) pMbuf->mBlkHdr.mData[6*8+j] = (char)0x67 -j ;
	for(j=0;j<8;j++) pMbuf->mBlkHdr.mData[7*8+j] = (char)0x6f -j ;
*/

        data = (char) 0x30;
        for (j = 0; j < packetLength; j++)
        {
            pMbuf->mBlkHdr.mData[j] = data++;
            
            if ((char) 0x7b == data)
            {
                data =(char) 0x30;
            }
        }

#if 1
                { /* swap by 4 bytes */
                        UINT32 *P;
                        int P_len = pMbuf->m_len;
                        P = (UINT32 *)(pMbuf->m_data);

                        while(P_len > 0) {
                                *P = (UINT32)ixOsServWordSwap((UINT32)(*P));
                                P_len -= sizeof(UINT32);
                                P ++;
                        }
                }
#endif
/*
    {
	unsigned long *pp = (unsigned long *)(&pMbuf->mBlkHdr.mData[0]);
	int ii;

	for(ii= 0; ii < packetLength / 4; ii++)
		printk("\nbefore data %8x",*(pp+ii));
	printk("\n");
    }
*/
    {
	unsigned long *pp = (unsigned long *)(IV);
	int ii;

	for(ii= 0; ii < 2; ii++)
		printk("\n IV %8x",*(pp+ii));
	printk("\n");
    }


{
        unsigned long *ppp = (unsigned long *)(&cipherKey[0]);
        int iii;

        for(iii=0;iii < 6 ; iii ++)
                printk("3des key %8x\n",*(ppp+iii));
}
        
        /* This while loop will always call perform until it is  not 
         * successful but will continue to retry perform if the status 
         * is queue_full. Start with forwardCtxId and the reverseCtxId
         * will be used in the perform callback.
         */
#if 0
        do
        {
            if (0 == pktNo)
            {
                /* Start the timeStamp before Perform */
                timeStamp = ixOsServTimestampGet();
            }
#endif
            status = ixCryptoAccAuthCryptPerform (
                        forwardCtxId,
                        pMbuf,
                        NULL,
                        0,
                        packetLength,    
                        0, 
                        packetLength,
                        packetLength,
                        IV);
#if 0            
        } while (IX_CRYPTO_ACC_STATUS_QUEUE_FULL == status);
        /* End of do-while loop */
#endif        
            printf ("Placed %dth Packet in to flow\n", pktNo + 1);
    } /* End of for (i) loop */

/* Jack ++ */
/* return IX_FAIL; */
    
    printf ("\nWaiting for operation done [<10 min for the biggest packet]\n");

    /* Wait crypto services done. The task is put into sleep state while 
     * waiting for all the callbacks returned
     */        
#if 0
    do
    {
        ixOsServTaskSleep(IX_CRYPTOACC_CODELET_PERFORM_WAIT_TIME);            
    } while (!trafficFlowHalt);
#endif
    
    /* Wait for all buffers being returned in callbacks */
    while (!ixCryptoAccCodeletBufIsQFull())
    {
	int kk,tt=0;
	for(kk=0;kk<10;kk++) tt++;
/*
        ixOsServTaskSleep(IX_CRYPTOACC_CODELET_PERFORM_WAIT_TIME);
        printf (".");
	printk("\n");
*/
    }
/*    
    printf ("\nOperation DONE!\n");
*/
    /* Unregister contexts */
    ixCryptoAccCodeletUnregisterCtx();

    /* Free resources */
    ixCryptoAccCodeletCryptoMemPoolFree(packetLength);

    /* Dump performance data */
/*
    for (i = 0; i < PERFORMANCE_WINDOW_SIZE; i++)
    {
        printf ("[%2d] The rate is %d Mbits/sec\n", i, performanceLog[i]);
    }
*/    
    /* Reset performance entries counter */
    performanceNumEntries = 0;
    
    /* Reset stop flag */
    trafficFlowHalt = FALSE;
        
    if (ixCryptoAccCodeletInitialised) 
    {
    	ixCryptoAccCodeletDispatcherStop (
    	    IX_CRYPTOACC_CODELET_QMGR_DISPATCHER_MODE);
    	ixCryptoAccCodeletInitialised = FALSE;
    }

    return IX_SUCCESS;

} /* end of function ixCryptoAccCodeletMain () */


/**
 * @fn ixCryptoAccCodeletDispatcherStop (BOOL useInterrupt)
 *
 * @brief Stop QMgr dispatcher thread if QMgr dispatcher runs in poll mode 
 *        or unbind QMgr dispatcher from interrupt if it runs in interrupt mode.
 *
 */
IX_STATUS ixCryptoAccCodeletDispatcherStop (BOOL useInterrupt)
{
    if (useInterrupt)	/* Interrupt mode */
    {
    	/* 
    	 * Unhook the QM QLOW dispatcher from the interrupt controller. 
    	 */
    	if (IX_SUCCESS != ixOsServIntUnbind (IXP425_INT_LVL_QM1))
    	{
    	    printf ("Failed to unbind to QM1 interrupt\n");
    	    return IX_FAIL;
    	}
    }
    else /* poll mode */
    {
        if (!ixCryptoAccCodeletDispatcherPollStop)
        {
            /* Set stop flag to TRUE to stop the thread */
    	    ixCryptoAccCodeletDispatcherPollStop = TRUE;
        }
    } /* end of if (useInterrupt) */
    return (IX_SUCCESS);
} /* end of ixCryptoAccCodeletDispatcherStop () */


/**
 * @fn ixCryptoAccCodeletUsageShow (void)
 *
 * @brief Display user guide for Linux platform or VxWorks platform
 */
void ixCryptoAccCodeletUsageShow (void)
{
    printf ("Crypto Access Component Codelet User Guide\n");
    printf ("==========================================\n\n");
    
#ifdef __vxworks    
    printf ("  >ixCryptoAccCodeletMain (operationType, packetLen)\n\n");
#else
    printf ("  >insmod ixp400_codelets_cryptoAcc.o operationType=<operationType> "
            "packetLen=<packetLen>\n\n");
#endif
    printf (
        "  Where operationType:\n"
        "    1 = To encrypt and decrypt packets using selected cipher algorithm.\n"
        "    2 = To authenticate packets using selected authentication algorithm.\n"
        "    3 = To encrypt/authenticate and authenticate/decrypt packets using\n"
        "        selected cipher algorithm and selected authentication algorithm.\n\n"
        "  Where packetLen: \n"
        "    Packet length ranges from 64 bytes to 65456 bytes, if cipher\n"
        "    algorithm is DES/3DES, packet length must be multiple of 8-byte\n"
        "    (cipher block length); while AES algorithm must have packet length\n"
        "    that is multiple of 16-byte in size.\n\n\n");            

} /* end of ixCryptoAccCodeletUsageShow () */

/* End of IxCryptoAccCodelet.c */
