/**
 * @file IxCryptoAccCodelet.h
 *
 * @date October-31-2002
 *
 * @brief This is the header file for the Crypto Access Component Codelet.
 *
 * -- Intel Copyright Notice --
 * 
 * @par
 * Copyright 2002-2003 Intel Corporation All Rights Reserved.
 * 
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.
 * 
 * @par
 * The Material is protected by worldwide copyright and trade secret laws
 * and treaty provisions. No part of the Material may be used, copied,
 * reproduced, modified, published, uploaded, posted, transmitted,
 * distributed, or disclosed in any way except in accordance with the
 * applicable license agreement .
 * 
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel, except in accordance with the
 * applicable license agreement.
 * 
 * @par
 * Unless otherwise agreed by Intel in writing, you may not remove or
 * alter this notice or any other notice embedded in Materials by Intel
 * or Intel's suppliers or licensors in any way.
 * 
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * 
 * @par
 * -- End Intel Copyright Notice --
 */

/**
 * @defgroup Codelets IXP425 Codelets
 *
 * @brief IXP425 codelets
 *
 * @{
 */

/**
 * @ingroup Codelets
 *
 * @defgroup IxCryptoAccCodelet IXP425 Crypto Access Codelet 
 *           (IxCryptoAccCodelet) API
 *
 * @brief IXP425 Crypto Access Codelet API
 *
 * The codelet demonstrates how the Security Hardware Accelerator API 
 * can be used for cryptographic purposes.
 *
 * Three different cryptographic operations are demonstrated : 
 * <PRE> 
 *    1.   Encryption and decryption. Default is DES-CBC.
 *
 *    2.   Authentication calculation and check. Default is SHA1.
 *
 *    3.   Combined service of encryption/authentication calculation and
 *         authentication check/decryption. Default is DES-CBC mode with SHA1.
 * </PRE>
 *
 * Here is the list of functionalities of the codelet and they are in the 
 * order in which the codelet must follow in order to utilise the Security 
 * Hardware Accelerator, i.e. Initialise the related components, followed 
 * by crypto context registration, send in crypto services request using 
 * the registered crypto context, unregister crypto context, etc.
 * <PRE>
 *       - The codelet demonstrates how the Network Processor Engine (NPE) 
 * is initialised, how the NPE image is downloaded to the NPE, and 
 * and how the NPE is started.
 *
 *       - The codelet demonstrates how the Queue Manager is initialised to
 * polling or interrupt mode.
 *
 *       - The codelet shows how to set up the callback functions and 
 * how to register crypto contexts with cryptoAcc access component.
 *
 *       - The codelet demonstrates how the registration should be done for
 * the three different operations using the register API.
 *
 *       - The codelet demonstrates how the perform API can be used after
 * successful registration.
 *
 *       - The codelet demonstrates the use of the unregister API in the
 * event of re-starting the above mentioned operation.
 *
 *       - The codelet measures the performance of the Security Hardware
 * Accelerator for each of the three operations stated above.
 * </PRE>
 *
 * For both Linux and VxWorks, the type of cipher algorithm and authentication
 * algorithm can be set at USER SETTING portion. The key length, block length   
 * and IV length must be set according to the algorithm standard. Please refer 
 * to IxCryptoAcc.h for permitted setting for each algorithm. Any changes to 
 * the setting below will require the codes to be recompiled.
 *
 * <B>Default Cryptographic Setting for Performance Benchmarking</B>
 * <PRE>
 *    - Cipher Algorithm         : DES
 *    - Cipher Mode of Operation : CBC
 *    - Authentication Algorithm : SHA1 
 * </PRE>
 *
 * <B>Default Performance Sampling Size</B>
 * <PRE>
 *    - Packets per Sampling     : 1,000
 *    - Number of Samples        : 20
 *    - Total Packets            : 20,000
 * </PRE>
 *
 * <B>Default mbuf Pool Setting</B>
 * <PRE>
 *    - Number of mbufs in pool     : 20
 *    - Min of mbufs needed in pool : 2 (for crypto context registration 
 *                                    purpose)
 * </PRE>
 *
 * User may choose to run different operation with different packet length. 
 * The performance rate of selected operation with selected packet length will
 * be displayed after throughput rate has been captured for a predefined 
 * performance sampling size.
 *
 * 
 * <B>Notes for Performance Benchmarking:</B>
 * <PRE>
 *       - Packets are generated in XScale with random payload. No explicit
 * verification of payload will be carried out in the codelet.
 *
 *       - Encrypted packets are sent to NPE again for decryption and same 
 * applies to authentication operation. Packet payload is verified implicitly 
 * through this feedback system. 
 *
 *       - Throughput rate is captured from the point where packet is sent to 
 * NPE for processing to the point where NPE completes the encryption and/or
 * authentication operation and notifies XScale through callback.
 *
 *       - No external packet generator or benchmarking devices involved in this
 * codelet. Throughput rate is captured using XScale timestamp.
 *
 *       - Time taken for codelet to complete the operation depends on packet 
 * length selected. Larger packet will take more time to complete. 
 * </PRE>
 *
 * <B> VxWorks User Guide </B><BR>
 * ixCryptoAccCodeletMain() function is used as the entry point of execution for 
 * cryptoAcc Codelet. <BR>
 * It allows user to enter selection for different operations described below 
 * with different packet length. 
 *
 * <PRE>
 *  <I> Usage :
 *      >ixCryptoAccCodeletMain (operationType, packetLen)
 *      Where operationType:
 *           1 = To encrypt and decrypt packets using selected cipher algorithm.
 *           2 = To authenticate packets using selected authentication 
 *               algorithm.
 *           3 = To encrypt/authenticate and authenticate/decrypt packets using
 *               selected cipher algorithm and selected authentication 
 *               algorithm.
 *      Where packetLen: 
 *           Packet length ranges from 64 bytes to 65456 bytes, if cipher 
 *           algorithm is DES/3DES, packet length must be multiple of 8-byte
 *           (cipher block length); while AES algorithm must have packet length
 *           that is multiple of 16-byte in size.
 * </I>
 * </PRE>
 *
 * <B> Linux User Guide </B><BR>
 * The idea of using the ixCryptoAccCodeletMain() as the entry point of 
 * execution for cryptoAcc codelet. <BR>
 * The selected operation will be executed when user issues 'insmod' at command 
 * prompt.
 *
 * <PRE>
 * <I>  Usage :
 *      >insmod ixp400_codelets_cryptoAcc.o operationType=<x> packetLen=<y>
 *      Where x:
 *           1 = To encrypt and decrypt packets using selected cipher algorithm.
 *           2 = To authenticate packets using selected authentication 
 *               algorithm.
 *           3 = To encrypt/authenticate and authenticate/decrypt packets using
 *               selected cipher algorithm and selected authentication 
 *               algorithm.
 *      Where y: 
 *           Packet length ranges from 64 bytes to 65456 bytes, if cipher 
 *           algorithm is DES/3DES, packet length must be multiple of 8-byte
 *           (cipher block length); while AES algorithm must have packet length
 *           that is multiple of 16-byte in size.
 * </I>
 * </PRE>
 *
 * @{
 */
#ifndef IxCryptoAccCodelet_H
#define IxCryptoAccCodelet_H

#include "IxTypes.h"
#include "IxOsBuffMgt.h"
#include "IxNpeDl.h"
#include "IxOsCacheMMU.h"
#include "IxCryptoAcc.h"


#define INLINE __inline__


/**************************************************************************
 *              USER SETTING - NPE BUILD ID
 **************************************************************************/

/**
 * @ingroup IxCryptoAccCodelet
 *
 * @def IX_CRYPTOACC_CODELET_NPEC_IMAGE_ID
 * 
 * @brief The build ID for NPE C image, default is IX_NPEDL_NPEIMAGE_NPEC_CRYPTO_AES_ETH
 *
 * @note Please refer to IxNpeDl.h for NPE image definition
 *
 */
#define IX_CRYPTOACC_CODELET_NPEC_IMAGE_ID IX_NPEDL_NPEIMAGE_NPEC_CRYPTO_AES_ETH



/**************************************************************************
 *       USER SETTING - PERFORMANCE BENCHMARKING SETTING
 **************************************************************************/

/**
 * @ingroup IxCryptoAccCodelet
 *
 * @def   IX_CRYPTOACC_CODELET_CIPHER_ALGO
 *
 * @brief Cipher algorithm
 *
 */
#define IX_CRYPTOACC_CODELET_CIPHER_ALGO        (IX_CRYPTO_ACC_CIPHER_DES)

/**
 * @ingroup IxCryptoAccCodelet
 *
 * @def   IX_CRYPTOACC_CODELET_CIPHER_MODE
 *
 * @brief Mode of operation for the cipher algorithm
 *
 */
#define IX_CRYPTOACC_CODELET_CIPHER_MODE        (IX_CRYPTO_ACC_MODE_CBC)

/**
 * @ingroup IxCryptoAccCodelet
 *
 * @def   IX_CRYPTOACC_CODELET_CIPHER_KEY_LEN
 *
 * @brief Permitted key length according to the cipher algorithm
 *
 */
#define IX_CRYPTOACC_CODELET_CIPHER_KEY_LEN     (IX_CRYPTO_ACC_DES_KEY_64)

/**
 * @ingroup IxCryptoAccCodelet
 *
 * @def   IX_CRYPTOACC_CODELET_CIPHER_BLOCK_LEN
 *
 * @brief Cipher block length for the cipher algorithm
 *
 */
#define IX_CRYPTOACC_CODELET_CIPHER_BLOCK_LEN   (IX_CRYPTO_ACC_DES_BLOCK_64)

/**
 * @ingroup IxCryptoAccCodelet
 *
 * @def   IX_CRYPTOACC_CODELET_CIPHER_IV_LEN
 *
 * @brief Initialisation vector length for the selected mode of operation and 
 *        cipher algorithm
 *
 */
#define IX_CRYPTOACC_CODELET_CIPHER_IV_LEN      (IX_CRYPTO_ACC_DES_IV_64)

/**
 * @ingroup IxCryptoAccCodelet
 *
 * @def   IX_CRYPTOACC_CODELET_AUTH_ALGO
 *
 * @brief Authentication algorithm
 *
 */
#define IX_CRYPTOACC_CODELET_AUTH_ALGO          (IX_CRYPTO_ACC_AUTH_SHA1)

/**
 * @ingroup IxCryptoAccCodelet
 *
 * @def   IX_CRYPTOACC_CODELET_AUTH_KEY_LEN
 *
 * @brief Key length permitted according to the authentication algorithm 
 *        selected above
 *
 */
#define IX_CRYPTOACC_CODELET_AUTH_KEY_LEN       (IX_CRYPTO_ACC_SHA1_KEY_160)

/**
 * @ingroup IxCryptoAccCodelet
 *
 * @def   IX_CRYPTOACC_CODELET_AUTH_DIGEST_LEN
 *
 * @brief Message digest length for the authentication algorithm selecetd above
 *
 */
#define IX_CRYPTOACC_CODELET_AUTH_DIGEST_LEN    (IX_CRYPTO_ACC_SHA1_DIGEST_160)
 
/**
 * @ingroup IxCryptoAccCodelet
 *
 * @def PERFORMANCE_WINDOW_SIZE
 *
 * @brief Number of crypto packets need to be sent in for performance 
 *        benchmarking (sampling size), 1 unit represents 1000 of packets 
 *
 */
#define PERFORMANCE_WINDOW_SIZE                        20

/**
 * @ingroup IxCryptoAccCodelet
 *
 * @def IX_CRYPTOACC_CODELET_CRYPTO_MBUF_POOL_SIZE
 *
 * @brief Size of MBuf pool, ie number of buffers to circulate 
 *
 */

#define IX_CRYPTOACC_CODELET_CRYPTO_MBUF_POOL_SIZE    3


/**************************************************************************
 *                           SYSTEM SETTING 
 **************************************************************************/

/**
 * @ingroup IxCryptoAccCodelet
 *
 * @enum  IxCryptAccCodeletOperation
 *
 * @brief The codelet mode of operation
 *
 */
typedef enum
{
    IX_CRYPTOACC_CODELET_ENCRYPT = 1,  /**< Encryption and decryption */
    IX_CRYPTOACC_CODELET_AUTH,         /**< Authentication calculation
                                        * and check
                                        */ 
    IX_CRYPTOACC_CODELET_ENCRYPT_AUTH  /**< Combined service of encryption/
                                        * authentication calculation and
                                        * authentication check/decryption.
                                        */
} IxCryptAccCodeletOperation;

/**
 * @ingroup IxCryptoAccCodelet
 *  
 * @def   IX_CRYPTOACC_CODELET_BATCH_LEN
 *
 * @brief Number of packets per sampling for performance benchmarking 
 *
 */
#define IX_CRYPTOACC_CODELET_BATCH_LEN              1000

/**
 * @ingroup IxCryptoAccCodelet
 * 
 * @def   IX_CRYPTOACC_CODELET_QMR_DISPATCHER_PRIORITY
 *
 * @brief Recommended priority of queue manager dispatch loop 
 *
 */
#define IX_CRYPTOACC_CODELET_QMR_DISPATCHER_PRIORITY    150 

/**
 * @ingroup IxCryptoAccCodelet
 *
 * @def   IX_CRYPTOACC_CODELET_XSCALE_TICK
 *
 * @brief The XScale tick is 66 MHz
 *
 */
#define IX_CRYPTOACC_CODELET_XSCALE_TICK            66

/**
 * @ingroup IxCryptoAccCodelet
 * 
 * @def   IX_CRYPTOACC_CODELET_REGISTER_WAIT_TIME
 *
 * @brief The codelet register wait time (in ms) in a for loop 
 *
 */
#define IX_CRYPTOACC_CODELET_REGISTER_WAIT_TIME     500 

/**
 * @ingroup IxCryptoAccCodelet
 *
 * @def   IX_CRYPTOACC_CODELET_PERFORM_WAIT_TIME
 * 
 * @brief The codelet perform wait time (in ms) in a for loop 
 *
 */
#define IX_CRYPTOACC_CODELET_PERFORM_WAIT_TIME      500

/**
 * @ingroup IxCryptoAccCodelet
 * 
 * @def   IX_CRYPTOACC_CODELET_MAX_TIME_WAIT
 *
 * @brief Maximum time (in ms) to wait before exiting the program  
 *
 */
#define IX_CRYPTOACC_CODELET_MAX_TIME_WAIT          200 

/**
 * @ingroup IxCryptoAccCodelet
 *
 * @def   IX_CRYPTOACC_CODELET_MIN_CRYPTO_PKT_SIZE
 *
 * @brief The minimum packet size 
 *
 */
#define IX_CRYPTOACC_CODELET_MIN_CRYPTO_PKT_SIZE    64

/**
 * @ingroup IxCryptoAccCodelet
 *
 * @def   IX_CRYPTOACC_CODELET_MAX_CRYPTO_PKT_SIZE
 *
 * @brief The maximum packet size 
 *
 */
#define IX_CRYPTOACC_CODELET_MAX_CRYPTO_PKT_SIZE    65456

/**
 * @ingroup IxCryptoAccCodelet
 *
 * @def   IX_CRYPTOACC_CODELET_NPEB_BUILD_ID
 *
 * @brief NPE B Build ID 
 *
 */
#define IX_CRYPTOACC_CODELET_NPEB_BUILD_ID          1

/**
 * @ingroup IxCryptoAccCodelet
 *
 * @def IX_CRYPTOACC_CODELET_QMGR_DISPATCHER_MODE
 * 
 * @brief QMgr Dispatcher Mode, whether in interrupt (TRUE) or poll (FALSE) mode
 *
 * Note: <BR>
 * 1. QMgr dispatcher can be run in both poll mode and interrupt mode in 
 *    vxWorks platform. When QMgr dispatcher is running in interrupt mode, the 
 *    display of packet rate for Ethernet software loopback on the windshell
 *    is very slow if the crypto packet length chosen is very small. This is 
 *    due to the time taken by NPE to service crypto packet is very fast and 
 *    QMgr callbacks with higher priority are triggered frequently and blocked 
 *    the task of displaying.
 *
 * 2. In linux platform, QMgr Dispatcher is NOT advisible to run in 
 *    poll mode. This is due to the task scheduling in poll mode is not fair 
 *    enough between multiple threads that cause the performance rate to drop 
 *    tremendously. Thus the performance rate displayed in Linux platform  
 *    using poll mode is not accurate. Performance rate is more accurate when 
 *    QMgr dispatcher is running in interrupt mode.
 */
#ifdef __linux
/* use interrupts for performances */
#define IX_CRYPTOACC_CODELET_QMGR_DISPATCHER_MODE          (TRUE)
#else
/* use polled mode for performances */
#define IX_CRYPTOACC_CODELET_QMGR_DISPATCHER_MODE          (FALSE)
#endif


/**************************************************************************
 *                           Function Prototype 
 **************************************************************************/
        
/**
 * @ingroup IxCryptoAccCodelet
 *
 * @fn ixCryptoAccCodeletMain (
           IxCryptAccCodeletOperation operationType,
           UINT32 packetLen)
 *
 * @brief  This is the entry point function to the codelet to choose the 
 *         operation for the codelet and packet length to be used. 
 *         This is the main function of the codelet where crypto contexts
 *         registration and crypto perform services are done. Based on the  
 *         selected operation and selected packet length, packets are sent to 
 *         cryptoAcc for processing. 
 *
 * @param "IxCryptAccCodeletOperation [in] operationType" - operation for 
 *                                  codelet.
 *                                  <LI> 1: Encrypt/decrypt operation
 *                                  <LI> 2: Authentication operation
 *                                  <LI> 3: Encrypt/decrypt and authentication
 *                                          operation
 * @param "UINT32 [in] packetLen" - packet length
 *
 * @return  IX_STATUS
 *          @li IX_SUCCESS - codelet runs successfully
 *          @li IX_FAIL    - codelet fails
 */
IX_STATUS ixCryptoAccCodeletMain (
		struct crypto_request *req 
/*               IxCryptAccCodeletOperation operationType,
               UINT32 packetLen
*/
);


#endif /* IxCryptoAccCodelet_H */

/** @} defgroup IxCryptoAccCodelet*/

/** @} defgroup Codelet*/
