/**
 * @file IxCryptoAccCryptoPerform.c
 *
 * @date 03-October-2002
 *
 * @brief Module to perform crypto services via NPE.
 *
 *
 * Design Notes:
 *
 * -- Intel Copyright Notice --
 * 
 * @par
 * Copyright 2002-2003 Intel Corporation All Rights Reserved.
 * 
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.
 * 
 * @par
 * The Material is protected by worldwide copyright and trade secret laws
 * and treaty provisions. No part of the Material may be used, copied,
 * reproduced, modified, published, uploaded, posted, transmitted,
 * distributed, or disclosed in any way except in accordance with the
 * applicable license agreement .
 * 
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel, except in accordance with the
 * applicable license agreement.
 * 
 * @par
 * Unless otherwise agreed by Intel in writing, you may not remove or
 * alter this notice or any other notice embedded in Materials by Intel
 * or Intel's suppliers or licensors in any way.
 * 
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * 
 * @par
 * -- End Intel Copyright Notice --
 */

/**
 * Put the system defined include files required.
 */

#include <string.h>

/**
 * Put the user defined include files required.
 */

#include "IxTypes.h"
#include "IxOsServices.h"
#include "IxOsServices.h"
#include "IxOsCacheMMU.h"
#include "IxCryptoAcc.h"
#include "IxCryptoAcc_p.h"
#include "IxCryptoAccCryptoPerform_p.h"
#include "IxCryptoAccQAccess_p.h"
#include "IxCryptoAccDescMgmt_p.h"
#include "IxCryptoAccCCDMgmt_p.h"
#include "IxCryptoAccUtilities_p.h"

/**
 * Variable declarations global to this file only.  Externs are followed by
 * static variables.
 */


#define IX_CRYPTO_ACC_AES_CTR_IV_LENGTH_8  8  /**< Length of flag in AES IV
 					       */
#define IX_CRYPTO_ACC_AES_CTR_IV_OFFSET  4  /**< IV offset, 4th byte from flag
 						 */
#define IX_CRYPTO_ACC_AES_CTR_IV_COUNTER_OFFSET  12  /**< Counter offset (last
 						      * 4 bytes
						      */

/* extern IxCryptoAccCryptoCtx ixCryptoCtx[IX_CRYPTO_ACC_MAX_ACTIVE_SA_TUNNELS];
 * is declared in IxCryptoAccCCDMgmt_p.h
 */


/**
 *
 * @fn      ixCryptoPerformAesCtrBlkCreate
 * @brief   To create initial counter block for the use of AES CTR encryption
 *          mode operation.
 * 
 * @note    Please refer to IxCryptoAcc.h for the version of internet draft
 *          for AES CTR implementation
 *
 */

PUBLIC void 
ixCryptoPerformAesCtrBlkCreate (
    IxCryptoQDescriptor *pQDescriptor,
    UINT8 *pIV,
    UINT32 securityParameterIndex)
{
    /* 
     * SPI is contained in the 1,2,3 bytes of the 4 word IV with flags
     * occupying the 0 byte. Flags are always set to zero. The shift of
     * 16 and 8 is to get the right order into bytes 1, 2, and 3 of the
     * IV.
     */
    pQDescriptor->npeQDesc.IV[0] = 0; /* Set flag to zero */
    pQDescriptor->npeQDesc.IV[1] = (securityParameterIndex&0xFF0000)>>16;
    pQDescriptor->npeQDesc.IV[2] = (securityParameterIndex&0x00FF00)>>8;
    pQDescriptor->npeQDesc.IV[3] = (securityParameterIndex&0x0000FF);
    
    /* Set IV in second and third words */
    memcpy (&(pQDescriptor->npeQDesc.IV[IX_CRYPTO_ACC_AES_CTR_IV_OFFSET]),pIV, 
	    IX_CRYPTO_ACC_AES_CTR_IV_LENGTH_8);

    /* Set last IV word to zero */
    pQDescriptor->npeQDesc.IV[IX_CRYPTO_ACC_AES_CTR_IV_COUNTER_OFFSET] = 0;
    pQDescriptor->npeQDesc.IV[IX_CRYPTO_ACC_AES_CTR_IV_COUNTER_OFFSET+1] = 0;
    pQDescriptor->npeQDesc.IV[IX_CRYPTO_ACC_AES_CTR_IV_COUNTER_OFFSET+2] = 0;
    pQDescriptor->npeQDesc.IV[IX_CRYPTO_ACC_AES_CTR_IV_COUNTER_OFFSET+3] = 0;

} /* End of ixCryptoPerformAesCtrBlkCreate function */



/**
 *
 * @fn      ixCryptoPerformBufferOffsetBoundaryCheck
 * @brief   To check the boundary condition of the buffer offset provided 
 *          by the client.
 *
 */

#ifndef NDEBUG
IxCryptoAccStatus 
ixCryptoPerformBufferOffsetBoundaryCheck (
    UINT32 npeOperation,
    IX_MBUF *pSrcMbuf,
    IX_MBUF *pDestMbuf,
    UINT16 authStartOffset,
    UINT16 authDataLen,
    UINT16 cryptStartOffset,
    UINT16 cryptDataLen)
{
    IX_MBUF *pLengthMbufPtr;
    int length;
    int cipherLen = 0;
    int authLen = 0;
    BOOL opAuth = FALSE;  /* True if authenticating */
    BOOL opCrypt = FALSE; /* True if doing cryption */

    /* Mask the IxCryptoNpeOperationMode HMac bit to obtain the current status 
     * of Authentication */
    opAuth  = (IX_CRYPTO_NPE_OP_HMAC_MODE_IS_ENABLED == 
		(npeOperation & IX_CRYPTO_NPE_OP_HMAC_DISABLE_MASK));
    /* Mask the IxCryptoNpeOperationMode Crypt bit to obtain the current status 
     * of Cryption */
    opCrypt = (IX_CRYPTO_NPE_OP_CRYPT_MODE_IS_DISABLED != 
	       (npeOperation & IX_CRYPTO_NPE_OP_CRYPT_ENABLE_MASK));

    /* Get total length of chained mbuf - only the source mbuf is chained. The
     * destination mbuf will never be chained
     */
    pLengthMbufPtr = pSrcMbuf;
    length = IX_MBUF_MLEN (pSrcMbuf);
    cipherLen = (int)(cryptStartOffset+cryptDataLen);
    authLen = (int)(authStartOffset+authDataLen);

    while (NULL != IX_MBUF_NEXT_BUFFER_IN_PKT_PTR (pLengthMbufPtr))
    {
	pLengthMbufPtr = IX_MBUF_NEXT_BUFFER_IN_PKT_PTR (pLengthMbufPtr);
	length += IX_MBUF_MLEN (pLengthMbufPtr);
    }   

    /* Check at least one is set, do not support no auth and no crypt */
    if (opAuth||opCrypt) 
    {
	if (opCrypt)
	{
	    /* Check if Mbuf is longer than cryptStartOffset+cryptDataLen */
	    if (length < cipherLen)
	    {
		/* Log error message in debugging mode */
		IX_CRYPTO_ACC_LOG(LOG_MESSAGE, "Source Mbuf is short = %d\n \
             			      cryptStartOffset+cryptDataLen is %d\n\n",
				  length,cipherLen,0,0,0,0);	
		return IX_CRYPTO_ACC_STATUS_FAIL;
	    }
	} /* End of Crypt mode check */
    
	if (opAuth)
	{
	    /* Check if the AuthDataLen is greater than 65471 or 0xFFBF  
	     * refer to NPE FS */
	    if (IX_CRYPTO_MAX_AUTH_LENGTH < (UINT32)(authDataLen))
	    {
		/* Log error message in debugging mode */
		IX_CRYPTO_ACC_LOG (LOG_MESSAGE, "Authentication data length is \
                 %d, greater than IX_CRYPTO_MAX_AUTH_LENGTH of %d",
				   authDataLen,IX_CRYPTO_MAX_AUTH_LENGTH,0,0,0,0);
		return IX_CRYPTO_ACC_STATUS_FAIL;
	    } /* End of Max auth length check */

	    /* Check if Mbuf is longer than authStartOffset+authDataLen */
	    if (length <authLen)
	    {
		/* Log error message in debugging mode */
		IX_CRYPTO_ACC_LOG(LOG_MESSAGE, "Source Mbuf is short = %d\n \
             			      authStartOffset+authDataLen is %d\n\n",
				  length,authLen,0,0,0,0);	
		return IX_CRYPTO_ACC_STATUS_FAIL;
	    }
	} /* End of OpAuth Only Check */

	/* If we are doing both auth and crypt, check relative positions 
	 * ((authStartOffset+authDataLen)>=(cryptDataLen+cryptStartOffset))&&
	 * ((cryptDataLen+cryptStartOffset)>=cryptStartOffset)&&
	 * (cryptStartOffset>=authStartOffset).
	 * The simplified version of the relative position check above is used
	 * in the actual codes below.
	 */
	if (opAuth&&opCrypt)
	{
	    if (((authStartOffset+authDataLen)>=(cryptDataLen+cryptStartOffset))&&
		(cryptStartOffset>=authStartOffset))
	    {
		return IX_CRYPTO_ACC_STATUS_SUCCESS;
	    }
	    else /* End of if (relative check) */
	    {
		return IX_CRYPTO_ACC_STATUS_FAIL;
	    }
	} /* End of if opAuth&&opCrypt */

	/* Check for Destination Mbuf if operationMode is non-inplace */
	if (IX_CRYPTO_NPE_OP_TRANSFER_MODE_IS_IN_PLACE != 
	    (npeOperation & IX_CRYPTO_NPE_OP_TRANSFER_MODE_MASK))
	{
	    if (opAuth)
	    {
		/* MLEN is used instead of PKT_LEN to ensure all the data will
		 * fit into one mbuf.
		 */
		if (IX_MBUF_MLEN (pDestMbuf) < authLen)
		{
		    /* Log error message in debugging mode */
		    IX_CRYPTO_ACC_LOG(LOG_MESSAGE, "Dest Mbuf is short = %d\n \
	         		  CryptStartOffset+CryptDataLen is %d\n",
				      IX_MBUF_MLEN (pDestMbuf),authLen,0,0,0,0);
		    return IX_CRYPTO_ACC_STATUS_FAIL;
		} /* End of DestMbuf length check for authentication*/
	    } /* End of authentication mode check */
	    if (opCrypt)
	    {
		if (IX_MBUF_MLEN (pDestMbuf) < cipherLen)
		{
		    /* Log error message in debugging mode */
		    IX_CRYPTO_ACC_LOG(LOG_MESSAGE, "Dest Mbuf is short = %d\n \
	         		      CryptStartOffset+CryptDataLen is %d\n",
				      IX_MBUF_MLEN (pDestMbuf),cipherLen,0,0,0,0);
		    return IX_CRYPTO_ACC_STATUS_FAIL;
		} /* End of DestMbuf length check for authentication*/
	    } /* End of Crypt mode check */
	} /* End of DestMbuf check */
    } /* End of if neither opAuth nor opCrypt */
    else /* Report error when neither opAuth nor opCrypt */
    {
	/* Log error message in debugging mode */
	IX_CRYPTO_ACC_LOG (LOG_ERROR, "Neither authentication nor \
					encryption is enabled",0,0,0,0,0,0);
	return IX_CRYPTO_ACC_STATUS_FAIL;
    } /* End of OpAuth || OpCrypt Check */
    /* Otherwise nothing more to do */
    return IX_CRYPTO_ACC_STATUS_SUCCESS;

} /* End of ixCryptoPerformBufferOffsetBoundaryCheck */
#endif /* ifndef NDEBUG */
/**
 * @fn      ixCryptoPerformCipherBlkLengthCheck
 * @brief   To check the data length against the pre-determined cipher block 
 *          length (8 or 16 bytes) stated in the Security Hardware Accelerator
 *          Function Specification.
 *
 */

#ifndef NDEBUG
IxCryptoAccStatus ixCryptoPerformCipherBlkLengthCheck (
    UINT32 npeOperation,
    UINT16 cipherBlkLen,
    UINT16 cryptStartOffset,
    UINT16 cryptDataLen)
{
    /*
     * Need to check:
     * 1. DataLen not 0 and cipherBlkLen not 0 (to protect divide by 0)
     * 2. DataLen is multiple of Cipher Block Length.
     */
    if ((0 != cryptDataLen) && (0 != cipherBlkLen))
    {
	/* Check whether cipherBlkLen is multiple of cryptDataLen */
	if (0 != (cryptDataLen%cipherBlkLen))
	{	
	    IX_CRYPTO_ACC_LOG (LOG_MESSAGE, "BlkLen not multiple of DataLen!\n \
					cryptDataLen = %d\n cipherBlkLen= %d\n",
			       cryptDataLen,cipherBlkLen,0,0,0,0);
	    return IX_CRYPTO_ACC_STATUS_CIPHER_INVALID_BLOCK_LEN;
	} /* End of Valid Block length check */
    } /* End of (NULL != cryptDataLen)&&(NULL != cipherBlkLen) */
    else /* Return error is either one of DataLen or BlkLen is zero */
    {
	/* Log error message in debugging mode */
	IX_CRYPTO_ACC_LOG (LOG_MESSAGE, "Invalid block length!\n \
					cryptDataLen = %d\n cipherBlkLen= %d\n",
			   cryptDataLen,cipherBlkLen,0,0,0,0);
	return IX_CRYPTO_ACC_STATUS_CIPHER_INVALID_BLOCK_LEN;
    }
		
    return IX_CRYPTO_ACC_STATUS_SUCCESS;
} /* End of ixCryptoPerformCipherBlkLengthCheck */
#endif /* ifndef NDEBUG */
