/**
 * @file IxEthDataPlane.c
 *
 * @author Intel Corporation
 * @date 12-Feb-2002
 *
 * @brief This file contains the implementation of the IXPxxx 
 * Ethernet Access Data plane component
 *
 * Design Notes:
 *
 * -- Intel Copyright Notice --
 * 
 * @par
 * Copyright 2002-2003 Intel Corporation All Rights Reserved.
 * 
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.
 * 
 * @par
 * The Material is protected by worldwide copyright and trade secret laws
 * and treaty provisions. No part of the Material may be used, copied,
 * reproduced, modified, published, uploaded, posted, transmitted,
 * distributed, or disclosed in any way except in accordance with the
 * applicable license agreement .
 * 
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel, except in accordance with the
 * applicable license agreement.
 * 
 * @par
 * Unless otherwise agreed by Intel in writing, you may not remove or
 * alter this notice or any other notice embedded in Materials by Intel
 * or Intel's suppliers or licensors in any way.
 * 
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * 
 * @par
 * -- End Intel Copyright Notice --
 */

#include "IxAssert.h"
#include "IxNpeMh.h"
#include "IxEthAcc.h"
#include "IxOsServices.h"
#include "IxOsServicesMemAccess.h"
#include "IxEthAccFpathDep.h"
#include "IxEthDBPortDefs.h"
#include "IxFeatureCtrl.h"
#include "IxEthAcc_p.h"
#include "IxEthAccQueueAssign_p.h"

#include <stdio.h>

/**
 * private functions prototype
 */

/* Remove INLINE keyword to resolve cannot inline warning (Linux). SCR1421 */
PRIVATE UINT32 ixEthAccMbufQPrepare(IX_MBUF *mbuf);

/* Remove INLINE keyword to resolve cannot inline warning (Linux). SCR1421 */
PRIVATE IxEthAccStatus
ixEthAccTxSwQHighestPriorityGet(IxEthAccPortId portId,
				IxEthAccTxPriority *priorityPtr);

/* Remove INLINE keyword to resolve cannot inline warning (Linux). SCR1421 */
PRIVATE IxEthAccStatus
ixEthAccTxFromSwQ(IxEthAccPortId portId,
		  IxEthAccTxPriority priority);

/* Remove INLINE keyword to resolve cannot inline warning (Linux). SCR1421 */
PRIVATE IxEthAccStatus
ixEthAccRxFreeFromSwQ(IxEthAccPortId portId);

/* Remove INLINE keyword to resolve cannot inline warning (Linux). SCR1421 */
PRIVATE IX_MBUF *ixEthAccMbufFromTxQ(UINT32 qbuf);

/* Remove INLINE keyword to resolve cannot inline warning (Linux). SCR1421 */
PRIVATE IX_STATUS ixEthAccQmgrLockTxWrite(IxEthAccPortId portId, UINT32 qBuffer);

PRIVATE INLINE IX_MBUF *ixEthAccMbufFromRxQ(UINT32 qbuf);
PRIVATE INLINE IX_STATUS ixEthAccQmgrLockRxWrite(IxEthAccPortId portId, UINT32 qBuffer);
PRIVATE INLINE IX_STATUS ixEthAccQmgrTxWrite(IxEthAccPortId portId, UINT32 qBuffer, UINT32 priority);

/**
 * @addtogroup IxEthAccPri
 *@{
 */

#define TX_INC(port,field) \
        ixEthAccPortData[port].ixEthAccTxData.stats.field++
#define RX_INC(port,field) \
        ixEthAccPortData[port].ixEthAccRxData.stats.field++
#define TX_STATS_INC(port,field) \
        IX_ETH_ACC_STATS_INC(ixEthAccPortData[port].ixEthAccTxData.stats.field)
#define RX_STATS_INC(port,field) \
        IX_ETH_ACC_STATS_INC(ixEthAccPortData[port].ixEthAccRxData.stats.field)

extern ixEthAccPortDataInfo   ixEthAccPortData[];

PRIVATE IxEthAccDataPlaneStats     ixEthAccDataStats;

#if IX_ETH_ACC_FPATH_AWARE
PRIVATE UINT32 txFramesCurrentlySubmitted[IX_ETH_ACC_NUMBER_OF_PORTS];
#endif

PRIVATE IxFastMutex txWriteMutex[IX_ETH_ACC_NUMBER_OF_PORTS];
PRIVATE IxFastMutex rxWriteMutex[IX_ETH_ACC_NUMBER_OF_PORTS];

PRIVATE UINT32 ixEthAccMbufQPrepare(IX_MBUF *mbuf)
{
    UINT32 qbuf;

    /* test for unchained mbufs */
    if (IX_MBUF_NEXT_BUFFER_IN_PKT_PTR(mbuf) == NULL)
    {
	/* unchained mbufs */
	IX_ETH_ACC_STATS_INC(ixEthAccDataStats.unchainedMBufs);

	/* virt2phys data pointer */
	IX_MBUF_MDATA(mbuf) = (char *)
	    IX_MMU_VIRTUAL_TO_PHYSICAL_TRANSLATION(IX_MBUF_MDATA(mbuf));
	
	/* endianess swap data pointer */
        IX_ETH_ACC_ENDIAN_UNCHAINED_MBUF_SWAP(mbuf);
	
	/* flush MBUF itself after all address conversions */
	IX_ACC_DATA_CACHE_FLUSH(mbuf, sizeof(IX_MBUF));
    }
    else
    {
	/* chained mbufs */
	IX_MBUF *ptr;
	IX_MBUF *next_ptr;
    
	IX_ETH_ACC_STATS_INC(ixEthAccDataStats.chainedMBufs);

	for (ptr = mbuf; ptr != NULL; ptr = next_ptr)
	{
	    /* we must save virtual next chain pointer */
	    next_ptr = IX_MBUF_NEXT_BUFFER_IN_PKT_PTR(ptr);
 
	    /* virt2phys data pointer */
	    IX_MBUF_NEXT_BUFFER_IN_PKT_PTR(ptr) = (IX_MBUF *)
		IX_MMU_VIRTUAL_TO_PHYSICAL_TRANSLATION(next_ptr);
	    IX_MBUF_MDATA(ptr) = (char *)
		IX_MMU_VIRTUAL_TO_PHYSICAL_TRANSLATION(IX_MBUF_MDATA(ptr));

	    /* endian processing */
	    IX_ETH_ACC_ENDIAN_MBUF_SWAP(ptr);

	    /* flush MBUF itself after all address conversions */
	    IX_ACC_DATA_CACHE_FLUSH(ptr, sizeof(IX_MBUF));
	}
    }
    
    /* virt2phys mbuf itself */
    qbuf = (UINT32)IX_MMU_VIRTUAL_TO_PHYSICAL_TRANSLATION(mbuf);

    /* Ensure the bits which are reserved to exchange information with
     * the NPE are cleared 
     *
     * If the mbuf address is not correctly aligned, or from an
     * incompatible memory range, there is no point to continue
     */
    IX_ENSURE(((qbuf & ~IX_ETHNPE_QM_Q_RXENET_ADDR_MASK) == 0), 
	      "Invalid address range");

    return qbuf;
}

PRIVATE IX_MBUF *ixEthAccMbufFromTxQ(UINT32 qbuf)
{
    IX_MBUF *mbuf;

    /* mask mbuf to cleare align and system used bits */
    qbuf &= IX_ETHNPE_QM_Q_RXENET_ADDR_MASK;

#if IX_ACC_DRAM_PHYS_OFFSET != 0
    /* restore the original address pointer (if PHYS_OFFSET is not 0) */
    qbuf |= IX_ACC_DRAM_PHYS_OFFSET & ~IX_ETHNPE_QM_Q_RXENET_ADDR_MASK;
#endif

    /* phys2virt mbuf itself */
    mbuf = (IX_MBUF *)IX_MMU_PHYSICAL_TO_VIRTUAL_TRANSLATION(qbuf);

    /* test for unchained mbufs */
    if (IX_MBUF_NEXT_BUFFER_IN_PKT_PTR(mbuf) == NULL)
    {
	/* unchained mbufs */
	/* endianess swap only the necessary fields */
        IX_ETH_ACC_ENDIAN_UNCHAINED_MBUF_SWAP(mbuf);
        
	/* phys2virt data pointer */
	IX_MBUF_MDATA(mbuf) = (char *)
	    IX_MMU_PHYSICAL_TO_VIRTUAL_TRANSLATION(IX_MBUF_MDATA(mbuf));
    }
    else
    {
	IX_MBUF *ptr;

	/* chained mbufs */
	for (ptr = mbuf; ptr != NULL; ptr = IX_MBUF_NEXT_BUFFER_IN_PKT_PTR(ptr))
	{
	    IX_ETH_ACC_ENDIAN_MBUF_SWAP(ptr);

	    /* phys2virt data pointer */
	    IX_MBUF_MDATA(ptr) = (char *)
		IX_MMU_PHYSICAL_TO_VIRTUAL_TRANSLATION(IX_MBUF_MDATA(ptr));
	    
	    IX_MBUF_NEXT_BUFFER_IN_PKT_PTR(ptr) = (IX_MBUF *)
	       IX_MMU_PHYSICAL_TO_VIRTUAL_TRANSLATION(
	       IX_MBUF_NEXT_BUFFER_IN_PKT_PTR(ptr));
	}
    }

    return mbuf;
}

PRIVATE INLINE IX_MBUF *ixEthAccMbufFromRxQ(UINT32 qbuf)
{
    IX_MBUF *mbuf;

    /* mask mbuf to cleare align and system used bits */
    qbuf &= IX_ETHNPE_QM_Q_RXENET_ADDR_MASK;

#if IX_ACC_DRAM_PHYS_OFFSET != 0
    /* restore the original address pointer (if PHYS_OFFSET is not 0) */
    qbuf |= IX_ACC_DRAM_PHYS_OFFSET & ~IX_ETHNPE_QM_Q_RXENET_ADDR_MASK;
#endif

    /* phys2virt mbuf itself */
    mbuf = (IX_MBUF *)IX_MMU_PHYSICAL_TO_VIRTUAL_TRANSLATION(qbuf);

    IX_ACC_DATA_CACHE_INVALIDATE(mbuf, sizeof(IX_MBUF));

    /* test for unchained mbufs */
    if (IX_MBUF_NEXT_BUFFER_IN_PKT_PTR(mbuf))
    {
	ixEthAccDataStats.unexpectedError++;
	IX_ETH_ACC_FATAL_LOG("IXETHACC: BUFFER CHAIN RECEIVED FROM THE ETH NPE!: buf=%p, next=%p Fixing...\n  ", 
			     (UINT32)mbuf , 
			     (UINT32)IX_MBUF_NEXT_BUFFER_IN_PKT_PTR(mbuf), 
			     0, 0, 0, 0);
	
	/* Fix it, so we can continue */
	IX_MBUF_NEXT_BUFFER_IN_PKT_PTR(mbuf) = NULL;
    }
    
    /* endianess swap only the necessary fields */
    IX_ETH_ACC_ENDIAN_UNCHAINED_MBUF_SWAP(mbuf);
    
    /* phys2virt data pointer */
    IX_MBUF_MDATA(mbuf) = (char *)
	IX_MMU_PHYSICAL_TO_VIRTUAL_TRANSLATION(IX_MBUF_MDATA(mbuf));

    return mbuf;
}

PRIVATE IX_STATUS
ixEthAccQmgrLockTxWrite(IxEthAccPortId portId, UINT32 qBuffer)
{
    IX_STATUS qStatus;
    if (ixOsServFastMutexTryLock(&txWriteMutex[portId]) == IX_SUCCESS)
    {
	qStatus = ixQMgrQWrite(
	       IX_ETH_ACC_PORT_TO_TX_Q_ID(portId),
	       &qBuffer);
#ifndef NDEBUG  
	if (qStatus != IX_SUCCESS)
	{
	    TX_STATS_INC(portId, txOverflow);	    
	}
#endif
	ixOsServFastMutexUnlock(&txWriteMutex[portId]);
    }
    else
    {
	TX_STATS_INC(portId, txLock);
	qStatus = IX_QMGR_Q_OVERFLOW;
    }
    return qStatus;
}

PRIVATE INLINE IX_STATUS
ixEthAccQmgrLockRxWrite(IxEthAccPortId portId, UINT32 qBuffer)
{
    IX_STATUS qStatus;
    if (ixOsServFastMutexTryLock(&rxWriteMutex[portId]) == IX_SUCCESS)
    {
	qStatus = ixQMgrQWrite(
	       IX_ETH_ACC_PORT_TO_RX_FREE_Q_ID(portId),
	       &qBuffer);	    
#ifndef NDEBUG  
	if (qStatus != IX_SUCCESS)
	{
	    RX_STATS_INC(portId, rxFreeOverflow);	    
	}
#endif
	ixOsServFastMutexUnlock(&rxWriteMutex[portId]);
    }
    else
    {
	RX_STATS_INC(portId, rxFreeLock);
	qStatus = IX_QMGR_Q_OVERFLOW;
    }
    return qStatus;
}

PRIVATE INLINE IX_STATUS
ixEthAccQmgrTxWrite(IxEthAccPortId portId, UINT32 qBuffer, UINT32 priority)
{
    /* fill the priority field */
    qBuffer |= (priority << IX_ETHNPE_QM_Q_FIELD_PRIOR_R);

#if IX_ETH_ACC_FPATH_AWARE
    if(txFramesCurrentlySubmitted[portId] <
       IX_ETH_ACC_MAX_TX_FRAMES_TO_SUBMIT)
    {
	return ixEthAccQmgrLockTxWrite(portId, qBuffer);	    
    }
    else
    {
	return IX_QMGR_Q_OVERFLOW;
    }	
#else
    return ixEthAccQmgrLockTxWrite(portId, qBuffer);	    
#endif
}

/**
 * @fn PRIVATE
 *
 * @brief This function will discover the highest priority S/W Tx Q that
 *        has entries in it
 *
 * @param portId - (in) the id of the port whose S/W Tx queues are to be searched
 *        priorityPtr - (out) the priority of the highest priority occupied q will be written
 *                      here
 *
 * @return IX_ETH_ACC_SUCCESS if an occupied Q is found
 *         IX_ETH_ACC_FAIL if no Q has entries
 *
 * @internal
 */
PRIVATE IxEthAccStatus
ixEthAccTxSwQHighestPriorityGet(IxEthAccPortId portId,
				IxEthAccTxPriority *priorityPtr)
{
    if (ixEthAccPortData[portId].ixEthAccTxData.schDiscipline 
	== FIFO_NO_PRIORITY)
    {
	if(IX_ETH_ACC_DATAPLANE_IS_Q_EMPTY(ixEthAccPortData[portId].
	       ixEthAccTxData.txQ[IX_ETH_ACC_TX_DEFAULT_PRIORITY]))
	{
	    return IX_ETH_ACC_FAIL;
	}
	else
	{
	    *priorityPtr = IX_ETH_ACC_TX_DEFAULT_PRIORITY;
	    TX_STATS_INC(portId,txPriority[*priorityPtr]);
	    return IX_ETH_ACC_SUCCESS;
	}
    }
    else
    {
	IxEthAccTxPriority highestPriority = IX_ETH_ACC_TX_PRIORITY_7;       
	while(1)
	{
	    if(!IX_ETH_ACC_DATAPLANE_IS_Q_EMPTY(ixEthAccPortData[portId].
	       ixEthAccTxData.txQ[highestPriority]))
	    {

		*priorityPtr = highestPriority;
		TX_STATS_INC(portId,txPriority[highestPriority]);
		return IX_ETH_ACC_SUCCESS;

	    }
	    if (highestPriority == IX_ETH_ACC_TX_PRIORITY_0)
	    {
		return IX_ETH_ACC_FAIL;
	    }
	    highestPriority--;
	}
    }
}

/**
 * @fn PRIVATE
 *
 * @brief This function will take a buffer from a TX S/W Q and attempt
 *        to add it to the relevant TX H/W Q
 *
 * @param portId - the port whose TX queue is to be written to
 *        priority - identifies the queue from which the entry is to be read
 *
 * @return none
 *
 * @internal
 */
PRIVATE IxEthAccStatus
ixEthAccTxFromSwQ(IxEthAccPortId portId,
		  IxEthAccTxPriority priority)
{
    IX_MBUF        *mbuf;
    IX_STATUS	   qStatus;

    IX_ENSURE((UINT32)priority <= (UINT32)7, "Invalid priority");

    IX_ETH_ACC_DATAPLANE_REMOVE_MBUF_FROM_Q_HEAD(
	ixEthAccPortData[portId].ixEthAccTxData.txQ[priority], 
	mbuf);
    
    if (mbuf != NULL)
    {
	/*
	 * Add the Tx buffer to the H/W Tx Q  
	 */

	/* 
	 * We do not need to flush here as it is already done
	 * in TxFrameSubmit()
	 */
	qStatus = ixEthAccQmgrTxWrite(portId, 
	      IX_MMU_VIRTUAL_TO_PHYSICAL_TRANSLATION((UINT32)mbuf),
	      priority);
	
	if (qStatus == IX_SUCCESS)
	{
	    TX_STATS_INC(portId,txFromSwQOK);
	    IX_ETH_ACC_FPATH_DELTA_INC(txFramesCurrentlySubmitted[portId]);
	    return IX_SUCCESS;
	}
	else if (qStatus == IX_QMGR_Q_OVERFLOW)
	{
	    /*
	     * H/W Q overflow, need to save the buffer 
	     * back on the s/w Q.
	     * we must put it back on the head of the q to avoid 
	     * reordering packet tx
	     */
	    TX_STATS_INC(portId,txFromSwQDelayed);
	    IX_ETH_ACC_DATAPLANE_ADD_MBUF_TO_Q_HEAD(
		ixEthAccPortData[portId].ixEthAccTxData.txQ[priority], 
		mbuf);

	    /*enable Q notification*/
	    qStatus = ixQMgrNotificationEnable(
		IX_ETH_ACC_PORT_TO_TX_Q_ID(portId),
		IX_ETH_ACC_PORT_TO_TX_Q_SOURCE(portId));	 
        
            if (qStatus != IX_SUCCESS && qStatus != IX_QMGR_WARNING)
            {
		TX_INC(portId,txUnexpectedError);
		IX_ETH_ACC_FATAL_LOG(
	            "IXETHACC:ixEthAccPortTxFromSwQ:Error: %u\n", 
	            qStatus, 0, 0, 0, 0, 0);
            }
	}
	else 
	{
	    TX_INC(portId,txUnexpectedError);

	    /* recovery attempt */
	    IX_ETH_ACC_DATAPLANE_ADD_MBUF_TO_Q_HEAD(
		ixEthAccPortData[portId].ixEthAccTxData.txQ[priority], 
		mbuf);
	    
	    IX_ETH_ACC_FATAL_LOG(
		"IXETHACC:ixEthAccTxFromSwQ:Error: QM status 0x%08X\n", 
		qStatus, 0, 0, 0, 0, 0);	
	}
    }
    else
    {
	/* sw queue is empty */
    }
    return IX_ETH_ACC_FAIL;
}

/**
 * @fn PRIVATE
 *
 * @brief This function will take a buffer from a RXfree S/W Q and attempt
 *        to add it to the relevant RxFree H/W Q
 *
 * @param portId - the port whose RXFree queue is to be written to
 *
 * @return none
 *
 * @internal
 */
PRIVATE IxEthAccStatus
ixEthAccRxFreeFromSwQ(IxEthAccPortId portId)
{
    IX_MBUF        *mbuf;
    IX_STATUS	   qStatus = IX_SUCCESS;

    IX_ETH_ACC_DATAPLANE_REMOVE_MBUF_FROM_Q_HEAD(
	  ixEthAccPortData[portId].ixEthAccRxData.freeBufferList, 
	  mbuf);
    if (mbuf != NULL)
    {
	/* flush the mbuf header (modified in the sw Q) */
	IX_ACC_DATA_CACHE_FLUSH(mbuf, sizeof(IX_MBUF));
		
	/*
	 * Add The Rx Buffer to the H/W Free buffer Q if possible  
	 */
	qStatus = ixEthAccQmgrLockRxWrite(portId, 
	  IX_MMU_VIRTUAL_TO_PHYSICAL_TRANSLATION((UINT32)mbuf));

	if (qStatus == IX_SUCCESS)
	{
	    RX_STATS_INC(portId,rxFreeRepFromSwQOK);
	    /*
	     * Buffer added to h/w Q.
	     */
	    return IX_SUCCESS;
	}
	else if (qStatus == IX_QMGR_Q_OVERFLOW)
	{
	    /*
	     * H/W Q overflow, need to save the buffer back on the s/w Q.
	     */
	    RX_STATS_INC(portId,rxFreeRepFromSwQDelayed);

	    IX_ETH_ACC_DATAPLANE_ADD_MBUF_TO_Q_HEAD(
		   ixEthAccPortData[portId].ixEthAccRxData.freeBufferList, 
		   mbuf);
	}
	else 
	{
	    /* unexpected qmgr error */
	    RX_INC(portId,rxUnexpectedError);

	    IX_ETH_ACC_DATAPLANE_ADD_MBUF_TO_Q_HEAD(
		    ixEthAccPortData[portId].ixEthAccRxData.freeBufferList, 
		    mbuf);

	    IX_ETH_ACC_FATAL_LOG("IXETHACC:rxfreeFromSwQ:Error: QM status 0x%08X\n", 
				 qStatus, 0, 0, 0, 0, 0);
	}
    }
    else
    {
	/* sw queue is empty */
    }
    return IX_ETH_ACC_FAIL;
}


IX_ETH_ACC_PUBLIC
IxEthAccStatus ixEthAccInitDataPlane()
{	
    UINT32 portId;
    
    /* 
     * Initialize the service and register callback to other services.
     */

    IX_ETH_ACC_MEMSET(&ixEthAccDataStats, 0, sizeof(ixEthAccDataStats));

    for(portId=IX_ETH_PORT_1; portId<=IX_ETH_PORT_2; portId++)
    {
	ixOsServFastMutexInit(&txWriteMutex[portId]);
	ixOsServFastMutexInit(&rxWriteMutex[portId]);
    }

    return (IX_ETH_ACC_SUCCESS);
}


IX_ETH_ACC_PUBLIC
IxEthAccStatus ixEthAccPortTxDoneCallbackRegister(IxEthAccPortId portId, 
						  IxEthAccPortTxDoneCallback
						  txCallbackFn, 
						  UINT32 callbackTag)
{	
    if (!IX_ETH_ACC_IS_SERVICE_INITIALIZED())
    {
	return (IX_ETH_ACC_FAIL);
    }
    if (!IX_ETH_ACC_IS_PORT_VALID(portId))
    {
	return (IX_ETH_ACC_INVALID_PORT);
    }

    if (IX_ETH_ACC_SUCCESS != ixEthAccSingleEthNpeCheck(portId))
    {
        IX_ETH_ACC_WARNING_LOG("EthAcc: Unavailable Eth %d: Cannot register TxDone Callback.\n",(INT32)portId,0,0,0,0,0);
        return IX_ETH_ACC_SUCCESS ;
    }

    if (!IX_ETH_IS_PORT_INITIALIZED(portId))
    {
	return (IX_ETH_ACC_PORT_UNINITIALIZED);
    }
    if (txCallbackFn == 0)
	/* Check for null function pointer here. */
    {
	return (IX_ETH_ACC_INVALID_ARG);
    }
    ixEthAccPortData[portId].ixEthAccTxData.txBufferDoneCallbackFn = txCallbackFn;
    ixEthAccPortData[portId].ixEthAccTxData.txCallbackTag = callbackTag;
    return (IX_ETH_ACC_SUCCESS);
}


IX_ETH_ACC_PUBLIC
IxEthAccStatus ixEthAccPortRxCallbackRegister(IxEthAccPortId portId, 
					      IxEthAccPortRxCallback
					      rxCallbackFn, UINT32
					      callbackTag)
{	
    if (!IX_ETH_ACC_IS_SERVICE_INITIALIZED())
    {
	return (IX_ETH_ACC_FAIL);
    }
    if (!IX_ETH_ACC_IS_PORT_VALID(portId))
    {
	return (IX_ETH_ACC_INVALID_PORT);
    }

    if (IX_ETH_ACC_SUCCESS != ixEthAccSingleEthNpeCheck(portId))
    {
        IX_ETH_ACC_WARNING_LOG("EthAcc: Unavailable Eth %d: Cannot register Rx Callback.\n",(INT32)portId,0,0,0,0,0);
        return IX_ETH_ACC_SUCCESS ;
    }
 
    if (!IX_ETH_IS_PORT_INITIALIZED(portId))
    {
	return (IX_ETH_ACC_PORT_UNINITIALIZED);
    }
    
    /* Check for null function pointer here. */
    if (rxCallbackFn == NULL)
    {
	return (IX_ETH_ACC_INVALID_ARG);
    }
    ixEthAccPortData[portId].ixEthAccRxData.rxCallbackFn = rxCallbackFn;
    ixEthAccPortData[portId].ixEthAccRxData.rxCallbackTag = callbackTag;
    return (IX_ETH_ACC_SUCCESS);
}

IX_ETH_ACC_PUBLIC
IxEthAccStatus ixEthAccPortTxFrameSubmit(IxEthAccPortId portId, 
					 IX_MBUF *buffer, 
					 IxEthAccTxPriority priority)
{	
    IX_STATUS	qStatus = IX_SUCCESS;
    UINT32      qBuffer;
    IxEthAccTxPriority highestPriority;
#if !IX_ETH_ACC_FPATH_AWARE
    IxQMgrQStatus txQStatus;    
#endif

#ifndef NDEBUG 
    if (buffer == NULL)
    {
	return (IX_ETH_ACC_FAIL);
    }
    if (!IX_ETH_ACC_IS_SERVICE_INITIALIZED())
    {
	return (IX_ETH_ACC_FAIL);
    }
    if (!IX_ETH_ACC_IS_PORT_VALID(portId))
    {
	return (IX_ETH_ACC_INVALID_PORT);
    }

    if (IX_ETH_ACC_SUCCESS != ixEthAccSingleEthNpeCheck(portId))
    {
        IX_ETH_ACC_FATAL_LOG("EthAcc: Unavailable Eth %d: Cannot submit Tx Frame.\n",
			     (INT32)portId,0,0,0,0,0);
        return IX_ETH_ACC_PORT_UNINITIALIZED ;
    }

    if (!IX_ETH_IS_PORT_INITIALIZED(portId))
    {
	return (IX_ETH_ACC_PORT_UNINITIALIZED);
    }
    if ((UINT32)priority > (UINT32)IX_ETH_ACC_TX_PRIORITY_7)
    {
	return (IX_ETH_ACC_INVALID_ARG);
    }
#endif 

    /*
     * Need to Flush the MBUF and its contents (data) as it may be
     * read from the NPE. Convert virtual addresses to physical addresses also.
     */
    qBuffer = ixEthAccMbufQPrepare(buffer);

    /*
     * If no fifo priority ...
     */
    if (ixEthAccPortData[portId].ixEthAccTxData.schDiscipline == 
	FIFO_NO_PRIORITY)
    {
	/*
	 * Add The Tx Buffer to the H/W Tx Q if possible  
	 */
	qStatus = ixEthAccQmgrTxWrite(portId,
				      qBuffer,
				      IX_ETH_ACC_TX_DEFAULT_PRIORITY);

	if (qStatus == IX_SUCCESS)
	{
	    TX_STATS_INC(portId,txQOK);
	    IX_ETH_ACC_FPATH_DELTA_INC(txFramesCurrentlySubmitted[portId]);

	    /*  
	     * Happy day scenario : Buffer added to h/w Q.
	     */
	    return (IX_SUCCESS);
	}
	else if (qStatus == IX_QMGR_Q_OVERFLOW)
	{
	    /*
	     * We were unable to write the buffer to the 
	     * appropriate H/W Q,  Save it in the sw Q.
	     * (use the default priority queue regardless of 
	     * input parameter)
	     */
	    priority = IX_ETH_ACC_TX_DEFAULT_PRIORITY;
	}
	else 
	{
	    /* unexpected qmgr error */
	    TX_INC(portId,txUnexpectedError);
	    IX_ETH_ACC_FATAL_LOG(
		"IXETHACC:ixEthAccPortTxFrameSubmit:Error: qStatus = %u\n", 
		(UINT32)qStatus, 0, 0, 0, 0, 0);
	    return (IX_ETH_ACC_FAIL);
	}
    }
    else if (ixEthAccPortData[portId].ixEthAccTxData.schDiscipline == 
	     FIFO_PRIORITY)
    {
	
	/*
	 * For priority transmission, put the frame directly on the H/W queue
	 * if the H/W queue is empty, otherwise, put it in a S/W Q
	 */
#if !IX_ETH_ACC_FPATH_AWARE
	/*
	 * For priority transmission, if Fast Path is not enabled, then 
	 * we only submit a prioritised frame if the tx q is empty
	 * For Fast Path enabled systems, we attempt to submit regardless,
	 * and rely on the ixEthAccQmgrTxWrite function's buffer submission
	 * tracking to signal an overflow if all non fast-path slots
	 * are occupied
	 */
	/*non fast path case*/
	ixQMgrQStatusGet(IX_ETH_ACC_PORT_TO_TX_Q_ID(portId), &txQStatus);
	if((txQStatus & IX_QMGR_Q_STATUS_E_BIT_MASK) != 0)
	{
	    /*The tx queue is empty, check whether there are buffers on the s/w queues*/
	    if(ixEthAccTxSwQHighestPriorityGet(portId,  &highestPriority)
	       !=IX_ETH_ACC_FAIL)
	    {
		/*there are buffers on the s/w queues, submit them*/
		ixEthAccTxFromSwQ(portId, highestPriority);

		/* the queue was empty, 1 buffer is supplied
		 * but is likely to be transmitted now and the 
		 * hw queue is empty again, so submit more from
		 * the sw queues
		 */
		if(ixEthAccTxSwQHighestPriorityGet(portId,  &highestPriority)
		   !=IX_ETH_ACC_FAIL)
		{
		    ixEthAccTxFromSwQ(portId, highestPriority);
		    /*
		     * and force the buffer supplied to be placed 
		     * on a priority queue
		     */
		    qStatus = IX_QMGR_Q_OVERFLOW;
		}
		else
		{
		    /*there are no buffers in the s/w queues, submit directly*/
		    qStatus = ixEthAccQmgrTxWrite(portId, qBuffer, priority);
		}
	    }
	    else
	    {
		/*there are no buffers in the s/w queues, submit directly*/
		qStatus = ixEthAccQmgrTxWrite(portId, qBuffer, priority);
	    }
	}
	else
	{
	    qStatus = IX_QMGR_Q_OVERFLOW;
	}
#else
	/*Fast path case*/
	/*First check if there are any non-fastpath slots available*/
	if(txFramesCurrentlySubmitted[portId] <
	   IX_ETH_ACC_MAX_TX_FRAMES_TO_SUBMIT)
	{
	    /*we need to pull a buffer off the S/w queues if one exists*/
	    if(ixEthAccTxSwQHighestPriorityGet(portId,  &highestPriority)
	       !=IX_ETH_ACC_FAIL)
	    {
		/*there are buffers on the s/w queues, submit from them*/
		ixEthAccTxFromSwQ(portId, highestPriority);

		/* force the buffer supplied to be placed 
		 * on a priority queue
		 */
		qStatus = IX_QMGR_Q_OVERFLOW;
	    }
	    else
	    {
		/*there are no buffers in the s/w queues, submit directly*/
		qStatus = ixEthAccQmgrTxWrite(portId, qBuffer, priority);
	    }
	}
	else
	{
	    /* there are no non-fastpath slots available, 
	     * so put the buffer on a s/w queue
	     */
	    qStatus = IX_QMGR_Q_OVERFLOW;
	}
#endif	    
    }
    else 
    {
	TX_INC(portId,txUnexpectedError);
	IX_ETH_ACC_FATAL_LOG(
	    "IXETHACC:ixEthAccPortTxFrameSubmit:Error: wrong schedule discipline setup\n", 
	    0, 0, 0, 0, 0, 0);
	return (IX_ETH_ACC_FAIL);
    }
    
    if(qStatus == IX_SUCCESS ) 
    {
	IX_ETH_ACC_FPATH_DELTA_INC(txFramesCurrentlySubmitted[portId]);
	TX_STATS_INC(portId,txQOK);
	return IX_ETH_ACC_SUCCESS;
    }
    else if(qStatus == IX_QMGR_Q_OVERFLOW)
    {
	TX_STATS_INC(portId,txQDelayed);
	/*
	 * We were unable to write the buffer to the 
	 * appropriate H/W Q,  Save it in a s/w Q.
	 */
	IX_ETH_ACC_DATAPLANE_ADD_MBUF_TO_Q_TAIL(
		ixEthAccPortData[portId].
		ixEthAccTxData.txQ[priority], 
		buffer);
	    
	qStatus = ixQMgrNotificationEnable(
		IX_ETH_ACC_PORT_TO_TX_Q_ID(portId),
		IX_ETH_ACC_PORT_TO_TX_Q_SOURCE(portId));

        if (qStatus != IX_SUCCESS)
	{
	    if (qStatus == IX_QMGR_WARNING)
	    {
		/* notification is enabled for a queue 
		 * which is already empty (the condition is already met)
		 * and there will be no more queue event to drain the sw queue
		 */
		TX_STATS_INC(portId,txLateNotificationEnabled);
		
		/* pull a buffer from the sw queue */
		if(ixEthAccTxSwQHighestPriorityGet(portId,  &highestPriority)
		   !=IX_ETH_ACC_FAIL)
		{
		    /*there are buffers on the s/w queues, submit from them*/
		    ixEthAccTxFromSwQ(portId, highestPriority);
		}
	    }
	    else
	    {
		TX_INC(portId,txUnexpectedError);
		IX_ETH_ACC_FATAL_LOG(
		     "IXETHACC:ixEthAccPortTxFrameSubmit:Error: %u\n", 
		     qStatus, 0, 0, 0, 0, 0);
	    }
        }
    }
    else
    {
	TX_INC(portId,txUnexpectedError);
	IX_ETH_ACC_FATAL_LOG(
	     "IXETHACC:ixEthAccPortTxFrameSubmit:Error: %u\n", 
	     qStatus, 0, 0, 0, 0, 0);
	return (IX_ETH_ACC_FAIL);
    }

    return (IX_ETH_ACC_SUCCESS);
}


/**
 * @fn PRIVATE
 *
 * @brief
 *
 * @param
 *
 * @return none
 *
 * @internal
 */

IX_ETH_ACC_PUBLIC
IxEthAccStatus ixEthAccPortRxFreeReplenish(IxEthAccPortId portId, 
					   IX_MBUF *buffer)
{	
    IX_STATUS	qStatus = IX_SUCCESS;
    UINT32      qBuffer;
    
    /* 
     * Check buffer is valid.
     */
    
#ifndef NDEBUG 
    if (buffer == 0)
    {
	return (IX_ETH_ACC_FAIL);
    }
    if (!IX_ETH_ACC_IS_SERVICE_INITIALIZED())
    {
	return (IX_ETH_ACC_FAIL);
    }
    if (!IX_ETH_ACC_IS_PORT_VALID(portId))
    {
	return (IX_ETH_ACC_INVALID_PORT);
    }

    if (IX_ETH_ACC_SUCCESS != ixEthAccSingleEthNpeCheck(portId))
    {
        IX_ETH_ACC_FATAL_LOG("EthAcc: Unavailable Eth %d: Cannot replenish Rx Free Q.\n",(INT32)portId,0,0,0,0,0);
        return IX_ETH_ACC_PORT_UNINITIALIZED ;
    }

    if (!IX_ETH_IS_PORT_INITIALIZED(portId))
    {
	return (IX_ETH_ACC_PORT_UNINITIALIZED);
    }
#endif 
    
    /*
     * Need to Flush the MBUF only (not data) as it may be read from the NPE 
     */
    qBuffer = ixEthAccMbufQPrepare(buffer);

    /*
     * Add The Rx Buffer to the H/W Free buffer Q if possible  
     */
    qStatus = ixEthAccQmgrLockRxWrite(portId, qBuffer);

    if (qStatus == IX_SUCCESS)
    {
	RX_STATS_INC(portId,rxFreeRepOK);
	/* 
	 * Buffer added to h/w Q.
	 */
	return (IX_SUCCESS);
    }
    else if (qStatus == IX_QMGR_Q_OVERFLOW)
    {
	RX_STATS_INC(portId,rxFreeRepDelayed);
	/*
	 * We were unable to write the buffer to the approprate H/W Q, 
	 * Save it in a s/w Q.
	 */
	IX_ETH_ACC_DATAPLANE_ADD_MBUF_TO_Q_HEAD(
	    ixEthAccPortData[portId].ixEthAccRxData.freeBufferList, 
	    buffer);

	qStatus = ixQMgrNotificationEnable(
	    IX_ETH_ACC_PORT_TO_RX_FREE_Q_ID(portId),
	    IX_ETH_ACC_PORT_TO_RX_FREE_Q_SOURCE(portId));
        
        if (qStatus != IX_SUCCESS)
	{
	    if (qStatus == IX_QMGR_WARNING)
	    {
		/* notification is enabled for a queue 
		 * which is already empty (the condition is already met)
		 * and there will be no more queue event to drain the sw queue
		 * move an entry from the sw queue to the hw queue */
		RX_STATS_INC(portId,rxFreeLateNotificationEnabled);
		ixEthAccRxFreeFromSwQ(portId);
	    }
	    else
	    {
		RX_INC(portId,rxUnexpectedError);
		IX_ETH_ACC_FATAL_LOG(
		     "IXETHACC:ixEthAccRxPortFreeReplenish:Error: %u\n", 
		     qStatus, 0, 0, 0, 0, 0);
	    }
        }
    }
    else 
    {
	RX_INC(portId,rxUnexpectedError);
	IX_ETH_ACC_FATAL_LOG(
	    "IXETHACC:ixEthAccRxPortFreeReplenish:Error: qStatus = %u\n", 
	    (UINT32)qStatus, 0, 0, 0, 0, 0);
        return(IX_ETH_ACC_FAIL);
    }
    return (IX_ETH_ACC_SUCCESS);
}


IX_ETH_ACC_PUBLIC
IxEthAccStatus ixEthAccTxSchedulingDisciplineSet(IxEthAccPortId portId, 
						 IxEthAccTxSchedulerDiscipline
						 sched)
{	
    if (!IX_ETH_ACC_IS_SERVICE_INITIALIZED())
    {
	return (IX_ETH_ACC_FAIL);
    }
    if (!IX_ETH_ACC_IS_PORT_VALID(portId))
    {
	return (IX_ETH_ACC_INVALID_PORT);
    }

    if (IX_ETH_ACC_SUCCESS != ixEthAccSingleEthNpeCheck(portId))
    {
        IX_ETH_ACC_WARNING_LOG("EthAcc: Unavailable Eth %d: Cannot set Tx Scheduling Discipline.\n",(INT32)portId,0,0,0,0,0);
        return IX_ETH_ACC_SUCCESS ;
    }

    if (!IX_ETH_IS_PORT_INITIALIZED(portId))
    {
	return (IX_ETH_ACC_PORT_UNINITIALIZED);
    }
    ixEthAccPortData[portId].ixEthAccTxData.schDiscipline = sched;
    return (IX_ETH_ACC_SUCCESS);
}
/**
 * @fn PRIVATE ixEthRxFrameQMCallback 
 *
 * @brief receive callback for Frame receive Q from NPE
 *
 * @param @ref IxQMgrCallback
 *
 * @return none
 *
 * @internal
 *
 * @warning If NDEBUG is set
 * any portID returned via the received callback is invalid if
 * greater than 5. If your system defines more than 6 ports (0 to 5) in
 * IxEthDBPortDefs.h you MUST check this value before using it
 * in conjunction with the Ethernet Learning/Filtering Database.
 * If greater than 5 it should be set to IX_ETH_DB_NUMBER_OF_PORTS,
 * which guarantees the value is outside the number of ports in use.
 * This problem is the result of the NPE Rx descriptor format
 * limitation, which can report valid ports only between 0 and 5, 
 * and will use 6 for "reserved" and "7" for "not found", while
 * the XScale Ethernet Database can use these numbers for legitimate
 * ports.
 */
void ixEthRxFrameQMCallback(IxQMgrQId qId, IxQMgrCallbackId callbackId)
{	
    IX_MBUF    *mbufPtr;
    UINT32     portId;
    UINT32     npeId;
    BOOL       additionalReadsRequired = FALSE;
    UINT32     rxQReadStatus;
    UINT32     qEntry;
    UINT32     *qEntryPtr;
static UINT32 rxQEntry[IX_ETH_ACC_MAX_RX_FRAME_CONSUME_PER_CALLBACK+1];

    /* 
     * Indication that there are Rx frames in the QM Rx Frame Q.
     */
    IX_ETH_ACC_STATS_INC(ixEthAccDataStats.rxCallbackCounter);
    do
    {
	qEntryPtr = rxQEntry;
	rxQReadStatus = ixQMgrQBurstRead(IX_ETH_ACC_RX_FRAME_ETH_Q, 
		 IX_ETH_ACC_MAX_RX_FRAME_CONSUME_PER_CALLBACK, 
		 qEntryPtr);

	if (rxQReadStatus == IX_QMGR_Q_UNDERFLOW)
	{
	    /*We've read all entries from the q*/
	    additionalReadsRequired = FALSE;
	}
	else if(rxQReadStatus == IX_SUCCESS)
	{
	    /*We've not read all entries from the q*/
	    additionalReadsRequired = TRUE;
	}
	else
	{
	    ixEthAccDataStats.unexpectedError++;
	    /*major error*/
	    IX_ETH_ACC_FATAL_LOG(
		"IXETHACC:ixEthRxFrameQMCallback:Error: %u\n", 
		(UINT32)rxQReadStatus, 0, 0, 0, 0, 0);
	    return;
	}

	while(*qEntryPtr != 0)
	{
	    qEntry = *qEntryPtr++;
	    /* 
	     * Pick out mbuf from message.
	     */
	    mbufPtr = ixEthAccMbufFromRxQ(qEntry);
	    
#ifndef NDEBUG
	    if (mbufPtr == (IX_MBUF *) NULL)
	    {
		ixEthAccDataStats.unexpectedError++;
		IX_ETH_ACC_FATAL_LOG(
		    "IXETHACC:ixEthRxQMCallback:Error: Null Mbuf Ptr\n", 
		    0, 0, 0, 0, 0, 0);
		return;
	    }
#endif

	    /*
	     * Get Port and Npe ID from message.
	     */	    
	    portId = ((IX_ETHNPE_QM_Q_RXENET_PORTID_MASK & 
		       qEntry) >> 
		      IX_ETHNPE_QM_Q_FIELD_PORTID_R);
	    
	    npeId = ((IX_ETHNPE_QM_Q_RXENET_NPEID_MASK & 
		      qEntry) >> 
		     IX_ETHNPE_QM_Q_FIELD_NPEID_R);
   
	    /* 
	     * Call user level callback.
	     */
	    RX_STATS_INC(npeId,rxFrameClientCallback);
	    ixEthAccPortData[npeId].ixEthAccRxData.rxCallbackFn(
		ixEthAccPortData[npeId].ixEthAccRxData.rxCallbackTag, 
		mbufPtr, 
		portId);
	}
    } while (additionalReadsRequired);

    return;
}
/**
 * @fn PRIVATE
 *
 * @brief
 *
 * @param
 *
 * @return none
 *
 * @internal
 */
void ixEthRxFreeQMCallback(IxQMgrQId qId, IxQMgrCallbackId callbackId)
{	
    IxEthAccPortId	portId = (IxEthAccPortId) callbackId;
    int		        lockVal;
    UINT32		maxQWritesToPerform = IX_ETH_ACC_MAX_RX_FREE_BUFFERS_LOAD;
    IX_STATUS	        qStatus = IX_SUCCESS;

    /*
     * We have reached a low threshold on one of the Rx Free Qs
     */
    
    /*note that due to the fact that we are working off an Empty threshold, this callback
      need only write a single entry to the Rx Free queue in order to re-arm the notification
    */
    
    RX_STATS_INC(portId,rxFreeLowCallback);
    
    /* 
     * Get buffers from approprite S/W Rx freeBufferList Q.
     */
    
#ifndef NDEBUG 
    if (!IX_ETH_ACC_IS_PORT_VALID(portId))
    {
	IX_ETH_ACC_FATAL_LOG(
	    "IXETHACC:ixEthRxFreeQMCallback:Error: Invalid Port 0x%08X\n", 
	    portId, 0, 0, 0, 0, 0);
	return;
    }
#endif
    IX_ETH_ACC_DATA_PLANE_LOCK(lockVal);
    if (IX_ETH_ACC_DATAPLANE_IS_Q_EMPTY(ixEthAccPortData[portId].
					ixEthAccRxData.freeBufferList))
    {
	/*  
	 * Turn off Q callback notification for Q in Question.
	 */
	qStatus = ixQMgrNotificationDisable(
	    IX_ETH_ACC_PORT_TO_RX_FREE_Q_ID(portId));
	

	IX_ETH_ACC_DATA_PLANE_UNLOCK(lockVal);

	if (qStatus != IX_SUCCESS)
	{
	    RX_INC(portId,rxUnexpectedError);
	    IX_ETH_ACC_FATAL_LOG(
		"IXETHACC:ixEthRxFreeQMCallback:Error: QM status 0x%08X\n", 
		qStatus, 0, 0, 0, 0, 0);
	    return;
	}
    }
    else 
    {
	IX_ETH_ACC_DATA_PLANE_UNLOCK(lockVal);
	/*
	 * Load the H/W Q with buffers from the s/w Q.
	 */
	
	do
	{
	    /* 
	     * Consume Q entries. - Note Q contains Physical addresss, 
	     * and have already been flushed to memory,
	     * And endianess converted if required.
	     */
	    if (ixEthAccRxFreeFromSwQ(portId) != IX_SUCCESS)
	    {
		/*
		 * No more entries in s/w Q.
		 * Turn off Q callback indication
		 */
		
		IX_ETH_ACC_DATA_PLANE_LOCK(lockVal);
		if (IX_ETH_ACC_DATAPLANE_IS_Q_EMPTY(ixEthAccPortData[portId].
		    ixEthAccRxData.freeBufferList))
		{
		    qStatus = ixQMgrNotificationDisable(
			IX_ETH_ACC_PORT_TO_RX_FREE_Q_ID(portId));
		}
		IX_ETH_ACC_DATA_PLANE_UNLOCK(lockVal);
		break;
	    }
	}
	while (--maxQWritesToPerform);
    }
}
/**
 * @fn PRIVATE
 *
 * @brief
 *
 * @param
 *
 * @return none
 *
 * @internal
 */
void ixEthTxFrameQMCallback(IxQMgrQId qId, IxQMgrCallbackId callbackId)
{
    IxEthAccPortId portId = (IxEthAccPortId) callbackId;
    int		   lockVal;
    UINT32	   maxQWritesToPerform = IX_ETH_ACC_MAX_TX_FRAME_TX_CONSUME_PER_CALLBACK;
    IX_STATUS	   qStatus = IX_SUCCESS;
    IxEthAccTxPriority highestPriority;

    
    /*
     * We have reached a low threshold on the Tx Q, and are being asked to 
     * supply a buffer for transmission from our S/W TX queues
     */
    TX_STATS_INC(portId,txLowThreshCallback);

    /* 
     * Get buffers from approprite Q.
     */
    
#ifndef NDEBUG
    if (!IX_ETH_ACC_IS_PORT_VALID(portId))
    {
	IX_ETH_ACC_FATAL_LOG(
	    "IXETHACC:ixEthTxFrameQMCallback:Error: Invalid Port 0x%08X\n", 
	    portId, 0, 0, 0, 0, 0);
	return;
    }
#endif

    do
    {
	/* 
	 * Consume Q entries. - Note Q contains Physical addresss, 
	 * and have already been flushed to memory,
	 * And endianess converted if required.
	 */
	
	IX_ETH_ACC_DATA_PLANE_LOCK(lockVal);
	
	if(ixEthAccTxSwQHighestPriorityGet(portId, &highestPriority) ==
	   IX_ETH_ACC_FAIL)
	{
	    /*
	     * No more entries in s/w Q.
	     * Turn off Q callback indication
	     */
	    qStatus = ixQMgrNotificationDisable(
		IX_ETH_ACC_PORT_TO_TX_Q_ID(portId));

	    IX_ETH_ACC_DATA_PLANE_UNLOCK(lockVal);

	    if (qStatus != IX_SUCCESS)
	    {
		ixEthAccDataStats.unexpectedError++;
		IX_ETH_ACC_FATAL_LOG(
		    "IXETHACC:ixEthTxFrameQMCallback:Error: QM status 0x%08X\n", 
		    qStatus, 0, 0, 0, 0, 0);
	    }
	    
	    return;	    
	}
	else
	{
	    IX_ETH_ACC_DATA_PLANE_UNLOCK(lockVal);
	    if (ixEthAccTxFromSwQ(portId,highestPriority)!=IX_SUCCESS)
	    {
		RX_INC(portId,rxUnexpectedError);
		return;
	    }
	}	
    }  	
    while (--maxQWritesToPerform);
}

/**
 * @fn PRIVATE
 *
 * @brief
 *
 * @param
 *
 * @return none
 *
 * @internal
 */

void ixEthTxFrameDoneQMCallback(IxQMgrQId qId, IxQMgrCallbackId callbackId)
{	
    IX_MBUF    *mbufPtr;
    UINT32     portId;
    BOOL       additionalReadsRequired = FALSE;
#if IX_ETH_ACC_SUBMIT_FROM_TX_DONE	    
    UINT32     highestPriority;
#endif
    UINT32 qEntry;
    UINT32 *qEntryPtr;
    UINT32 txDoneQReadStatus;
    static UINT32 txDoneQEntry[IX_ETH_ACC_MAX_TX_FRAME_DONE_CONSUME_PER_CALLBACK + 1];

    /* 
     * Indication that Tx frames have been transmitted from the NPE.
     */
    
    IX_ETH_ACC_STATS_INC(ixEthAccDataStats.txDoneCallbackCounter);
    
    do{
	qEntryPtr = txDoneQEntry;
	txDoneQReadStatus = ixQMgrQBurstRead(IX_ETH_ACC_TX_FRAME_DONE_ETH_Q, 
	     IX_ETH_ACC_MAX_TX_FRAME_DONE_CONSUME_PER_CALLBACK,
	     qEntryPtr);
	
	if (txDoneQReadStatus == IX_QMGR_Q_UNDERFLOW)
	{
	    additionalReadsRequired = FALSE;
	}
	else if(txDoneQReadStatus == IX_SUCCESS)
	{
	    additionalReadsRequired = TRUE;
	}
	else
	{
	    /*major error*/
	    ixEthAccDataStats.unexpectedError++;
	    IX_ETH_ACC_FATAL_LOG(
		"IXETHACC:ixEthTxFrameDoneQMCallback:Error: %u\n", 
		(UINT32)txDoneQReadStatus, 0, 0, 0, 0, 0);
	    return;
	}

	while(*qEntryPtr != 0)
	{	    
	    qEntry = *qEntryPtr++;

	    /* 
	     * Pick out mbuf from message.
	     * Convert Ptr to Virtual. 
	     */
	    mbufPtr = ixEthAccMbufFromTxQ(qEntry);
#ifndef NDEBUG
	    if (mbufPtr == (IX_MBUF *) NULL)
	    {
		ixEthAccDataStats.unexpectedError++;
		IX_ETH_ACC_FATAL_LOG(
		    "IXETHACC:ixEthTxFrameDoneQMCallback:Error: Null Mbuf Ptr\n", 
		    0, 0, 0, 0, 0, 0);
		return;
	    }
#endif
	    
	    /*
	     * Get NPE id from message, then convert to portId.
	     */
	    
	    portId = ((IX_ETHNPE_QM_Q_TXENETDONE_NPEID_MASK &
		       qEntry) >> 
		      IX_ETHNPE_QM_Q_FIELD_NPEID_R);
	    portId = IX_ETH_ACC_QM_NPE_ID_TO_PORT_ID(portId);

#ifndef NDEBUG
	    /* Prudent to at least check the port is within range */
	    if (portId >= IX_ETH_ACC_NUMBER_OF_PORTS)
	    {
		ixEthAccDataStats.unexpectedError++;
		IX_ETH_ACC_FATAL_LOG(
		    "IXETHACC:ixEthTxFrameDoneQMCallback: Illegal port: %u\n", 
		    (UINT32)portId, 0, 0, 0, 0, 0);
		return;
	    }
#endif

	    IX_ETH_ACC_FPATH_DELTA_DEC(txFramesCurrentlySubmitted[portId]);
	    TX_STATS_INC(portId,txDoneClientCallback);

	    /* 
	     * Call user level callback.
	     */
	    ixEthAccPortData[portId].ixEthAccTxData.txBufferDoneCallbackFn(
		ixEthAccPortData[portId].ixEthAccTxData.txCallbackTag, 
		mbufPtr);

	    /* @todo Enabling this causes a performance drop of approximately 50%
	     */

#if IX_ETH_ACC_SUBMIT_FROM_TX_DONE	    
	    /*We should be able to pull a buffer off the S/W Tx Qs
	      and transmit it now*/
	    if(txFramesCurrentlySubmitted[portId] <
	       IX_ETH_ACC_MAX_TX_FRAMES_TO_SUBMIT)
	    {
		TX_STATS_INC(portId,txFromTxDoneAttempts);
		
		if(ixEthAccTxSwQHighestPriorityGet(portId, 
						   &highestPriority)
		   !=IX_ETH_ACC_FAIL)
		{
		    ixEthAccTxFromSwQ(portId, highestPriority);
		}
	    }
#endif
	}
    } while( additionalReadsRequired );

    return;
}

void ixEthAccRecoverRxFreeQBuffers(IxEthAccPortId portId )
{
    IX_STATUS   qStatus;
    static UINT32 rxFreeQEntry[IX_QMGR_Q_SIZE128 + 1];
    UINT32 *qEntryPtr;
    UINT32     qEntry;
    IX_MBUF    *buffer;

    if (IX_ETH_ACC_SUCCESS != ixEthAccSingleEthNpeCheck(portId))
    {
        IX_ETH_ACC_WARNING_LOG("EthAcc: Unavailable Eth %d: We should not have any RxFree Q Buffers.\n",(INT32)portId,0,0,0,0,0);
        return ; 
    }
    
    /* disable rx free Q notifications */
    ixQMgrNotificationDisable(IX_ETH_ACC_PORT_TO_RX_FREE_Q_ID(portId));

    /*
     * Drain Rx Replenish Q.
     */
    qEntryPtr = rxFreeQEntry;
    qStatus = ixQMgrQBurstRead(
	 IX_ETH_ACC_PORT_TO_RX_FREE_Q_ID(portId), 
	 IX_QMGR_Q_SIZE128,
	 qEntryPtr);
    /*
     * Recover Entries in Rx Replenish Q.
     */
    while (*qEntryPtr != 0)
    {
	qEntry = *qEntryPtr++;

	buffer = ixEthAccMbufFromRxQ(qEntry);

	RX_STATS_INC(portId,rxDuringDisable);
	ixEthAccPortData[portId].ixEthAccRxData.rxCallbackFn(
	     ixEthAccPortData[portId].ixEthAccRxData.rxCallbackTag, 
	     buffer,
	     IX_ETH_DB_NUMBER_OF_PORTS /* port not found */);
    }
}

void  
ixEthAccRecoverTxSubmittedQBuffers(IxEthAccPortId portId)
{
    IX_STATUS   qStatus = IX_SUCCESS;
#if IX_ETH_ACC_FPATH_AWARE 
    IX_STATUS   qWrStatus = IX_SUCCESS;
#endif
    static UINT32 txQEntry[IX_QMGR_Q_SIZE128 + 1];
    UINT32 *qEntryPtr;
    UINT32     qEntry;
    IX_MBUF    *buffer;

    if (IX_ETH_ACC_SUCCESS != ixEthAccSingleEthNpeCheck(portId))
    {
        IX_ETH_ACC_WARNING_LOG("EthAcc: Unavailable Eth %d: We should not have any Tx submitted Q Buffers.\n",(INT32)portId,0,0,0,0,0);
        return; 
    }

    /* disable tx Q notifications */
    ixQMgrNotificationDisable(IX_ETH_ACC_PORT_TO_TX_Q_ID(portId));

    /*
     * Drain tx Q.
     */
    qEntryPtr = txQEntry;
    qStatus = ixQMgrQBurstRead(
       IX_ETH_ACC_PORT_TO_TX_Q_ID(portId),
       IX_QMGR_Q_SIZE128,
       qEntryPtr);

    /*
     * Recover Entries in Rx Replenish Q.
     */
    while (*qEntryPtr != 0)
    {
	qEntry = *qEntryPtr++;

#if IX_ETH_ACC_FPATH_AWARE 

	/*
	 * Check if QEntry is a fast path descripor 
	 */	   
	if ( qEntry & IX_ETHNPE_QM_Q_TXENET_FPBIT_MASK )
	{
	    /*
	     * This is a fast path buffer, return to the same Q.
	     * (@todo : check if it is appropriate to return to the
	     * queue when the NPE is supposed to not transmit
	     * It should go to the fast path free queue.
	     */
	    qWrStatus = ixQMgrQWrite( IX_ETH_ACC_PORT_TO_TX_Q_ID(portId), &qEntry);	    
	    if ( qWrStatus != IX_SUCCESS )
	    {
		TX_INC(portId,txUnexpectedError);
		IX_ETH_ACC_FATAL_LOG(
		     "IXETHACC:ixEthAccRecoverTxSubmittedQBuffers:Error: qWrStatus = %u\n", 
		     (UINT32)qWrStatus, 0, 0, 0, 0, 0);
	    }
	}
	else 
#endif
	{
	    buffer = ixEthAccMbufFromTxQ(qEntry);
	    
	    TX_STATS_INC(portId,txDoneDuringDisable);
	    IX_ETH_ACC_FPATH_DELTA_DEC(txFramesCurrentlySubmitted[portId]);
	    
	    ixEthAccPortData[portId].ixEthAccTxData.txBufferDoneCallbackFn(
		   ixEthAccPortData[portId].ixEthAccTxData.txCallbackTag, 
		   buffer);
	}
    }
}

void  
ixEthAccRecoverSwQBuffers(IxEthAccPortId portId)
{
    IxEthAccTxPriority priority;
    IX_MBUF *mbuf;

    if (IX_ETH_ACC_SUCCESS != ixEthAccSingleEthNpeCheck(portId))
    {
        IX_ETH_ACC_WARNING_LOG("EthAcc: Unavailable Eth %d: We should not have Sw Q Buffers.\n",(INT32)portId,0,0,0,0,0);
        return; 
    }

    /* tx queues */
    for (priority = IX_ETH_ACC_TX_PRIORITY_0 ; 
	 priority <= IX_ETH_ACC_TX_PRIORITY_7 ;
	 priority++)
    {
        do
        {
	    /* get sw q entry */
            IX_ETH_ACC_DATAPLANE_REMOVE_MBUF_FROM_Q_HEAD(
		 ixEthAccPortData[portId].ixEthAccTxData.txQ[priority], mbuf);

            if (mbuf != NULL)
            {
		TX_STATS_INC(portId,txPriority[priority]);

		/* convert to mbuf header fields to virtual address */
		mbuf = ixEthAccMbufFromTxQ((UINT32)IX_MMU_VIRTUAL_TO_PHYSICAL_TRANSLATION(mbuf));

                TX_STATS_INC(portId,txDoneSwQDuringDisable);
		
	        ixEthAccPortData[portId].ixEthAccTxData.txBufferDoneCallbackFn(
                    ixEthAccPortData[portId].ixEthAccTxData.txCallbackTag, 
                    mbuf);
            }
        } 
        while (mbuf != NULL);
    }

    /* rx queue */
    do
    {
	/* get sw q entry */
        IX_ETH_ACC_DATAPLANE_REMOVE_MBUF_FROM_Q_HEAD(ixEthAccPortData[portId].
	     ixEthAccRxData.freeBufferList, mbuf);

        if (mbuf != NULL)
        {
	    /* convert to mbuf header fields to virtual address */
	    mbuf = ixEthAccMbufFromRxQ((UINT32)IX_MMU_VIRTUAL_TO_PHYSICAL_TRANSLATION(mbuf));

	    RX_STATS_INC(portId,rxSwQDuringDisable);

            ixEthAccPortData[portId].ixEthAccRxData.rxCallbackFn(
                ixEthAccPortData[portId].ixEthAccRxData.rxCallbackTag, 
                mbuf,
                IX_ETH_DB_NUMBER_OF_PORTS /* port not found */);

        } 
    }
    while (mbuf != NULL);
}


IX_ETH_ACC_PUBLIC
void ixEthAccDataPlaneShow(void)
{
    UINT32 numTx0Entries;
    UINT32 numTx1Entries;
    UINT32 numTxDoneEntries;
    UINT32 numRxEntries;
    UINT32 numRxFree0Entries;
    UINT32 numRxFree1Entries;
    UINT32 portId;
#ifndef NDEBUG
    UINT32 priority;
    UINT32 numBuffersInRx=0;
    UINT32 numBuffersInTx=0;
    UINT32 numBuffersInSwQ=0;
    UINT32 totalBuffers=0;
    UINT32 rxFreeCallbackCounter = 0;
    UINT32 txCallbackCounter = 0;
#endif
    UINT32 key;

    /* snapshot of stats */
    ixEthAccTxDataStats tx[IX_ETH_ACC_NUMBER_OF_PORTS];
    ixEthAccRxDataStats rx[IX_ETH_ACC_NUMBER_OF_PORTS];
    IxEthAccDataPlaneStats stats;

    if (!IX_ETH_ACC_IS_SERVICE_INITIALIZED())
    {
	return;
    }

    /* get a reliable snapshot */
    key = ixOsServIntLock();

    numTx0Entries = 0;
    ixQMgrQNumEntriesGet(IX_ETH_ACC_TX_FRAME_ENET0_Q, &numTx0Entries);
    numTx1Entries = 0;
    ixQMgrQNumEntriesGet(IX_ETH_ACC_TX_FRAME_ENET1_Q, &numTx1Entries);
    numTxDoneEntries = 0;
    ixQMgrQNumEntriesGet( IX_ETH_ACC_TX_FRAME_DONE_ETH_Q, &numTxDoneEntries);
    numRxEntries = 0;
    ixQMgrQNumEntriesGet(IX_ETH_ACC_RX_FRAME_ETH_Q, &numRxEntries);
    numRxFree0Entries = 0;
    ixQMgrQNumEntriesGet(IX_ETH_ACC_RX_FREE_BUFF_ENET0_Q, &numRxFree0Entries);
    numRxFree1Entries = 0;
    ixQMgrQNumEntriesGet(IX_ETH_ACC_RX_FREE_BUFF_ENET1_Q, &numRxFree1Entries);

    for(portId=IX_ETH_PORT_1; portId<=IX_ETH_PORT_2; portId++)
    {
	memcpy(&tx[portId], 
	       &ixEthAccPortData[portId].ixEthAccTxData.stats,
	       sizeof(tx[portId]));
	memcpy(&rx[portId], 
	       &ixEthAccPortData[portId].ixEthAccRxData.stats,
	       sizeof(rx[portId]));
    }
    memcpy(&stats, &ixEthAccDataStats, sizeof(stats));
    
    ixOsServIntUnlock(key);

#ifdef NDEBUG
    printf("Detailed statistics collection not supported in this load\n");
#endif

    /* print snapshot */
    for(portId=IX_ETH_PORT_1; portId<=IX_ETH_PORT_2; portId++)
    {
       if ((ixFeatureCtrlProductIdRead() & IX_FEATURE_CTRL_SILICON_STEPPING_MASK) == 
            IX_FEATURE_CTRL_SILICON_TYPE_B0) 
       {  
            if ((IX_ETH_PORT_1 == portId) && 
                (ixFeatureCtrlComponentCheck(IX_FEATURECTRL_ETH0) ==  
                 IX_FEATURE_CTRL_COMPONENT_DISABLED)) 
            {							      
               continue ;
            } 
            if ((IX_ETH_PORT_2 == portId) && 
                (ixFeatureCtrlComponentCheck(IX_FEATURECTRL_ETH1) == 
                 IX_FEATURE_CTRL_COMPONENT_DISABLED)) 
            { 
	       continue ;
            } 
        }        

	printf("PORT %u --------------------------------\n",
	       portId);
#ifndef NDEBUG 
	printf("Tx Done Frames                : %u\n", 
	       tx[portId].txDoneClientCallback + 
	       tx[portId].txDoneSwQDuringDisable + 
	       tx[portId].txDoneDuringDisable);
	printf("Tx Frames                     : %u\n", 
	       tx[portId].txQOK + tx[portId].txQDelayed);
	printf("Tx H/W Q Added OK             : %u\n", 
	       tx[portId].txQOK);
	printf("Tx H/W Q Delayed              : %u\n", 
	       tx[portId].txQDelayed);
	printf("Tx From S/W Q Added OK        : %u\n", 
	       tx[portId].txFromSwQOK);
	printf("Tx From S/W Q Delayed         : %u\n", 
	       tx[portId].txFromSwQDelayed);
	printf("Tx Overflow                   : %u\n", 
	       tx[portId].txOverflow);
	printf("Tx Mutual Lock                : %u\n", 
	       tx[portId].txLock);
	printf("Tx Late Ntf Enabled           : %u\n", 
	       tx[portId].txLateNotificationEnabled);
	printf("Tx Low Thresh CB              : %u\n", 
	       tx[portId].txLowThreshCallback);
	printf("Tx From Tx Done Attempts      : %u\n", 
	       tx[portId].txFromTxDoneAttempts);
	printf("Tx Done from H/W Q (Disable)  : %u\n", 
	       tx[portId].txDoneDuringDisable);
	printf("Tx Done from S/W Q (Disable)  : %u\n", 
	       tx[portId].txDoneSwQDuringDisable);
	for (priority = IX_ETH_ACC_TX_PRIORITY_0;
	     priority <= IX_ETH_ACC_TX_PRIORITY_7;
	     priority++)
	{
	    if (tx[portId].txPriority[priority])
	    {
		printf("Tx Priority %u                 : %u\n", 
		       priority,
		       tx[portId].txPriority[priority]);
	    }
	}
#endif
	printf("Tx unexpected errors          : %u (should be 0)\n", 
	       tx[portId].txUnexpectedError);	
	
#ifndef NDEBUG
	printf("Rx Frames                     : %u\n", 
	       rx[portId].rxFrameClientCallback + 
	       rx[portId].rxSwQDuringDisable+ 
	       rx[portId].rxDuringDisable);
	printf("Rx Free Replenish             : %u\n", 
	       rx[portId].rxFreeRepOK + rx[portId].rxFreeRepDelayed);
	printf("Rx Free H/W Q Added OK        : %u\n", 
	       rx[portId].rxFreeRepOK);
	printf("Rx Free H/W Q Delayed         : %u\n", 
	       rx[portId].rxFreeRepDelayed);
	printf("Rx Free From S/W Q Added OK   : %u\n", 
	       rx[portId].rxFreeRepFromSwQOK);
	printf("Rx Free From S/W Q Delayed    : %u\n", 
	       rx[portId].rxFreeRepFromSwQDelayed);
	printf("Rx Free Overflow              : %u\n", 
	       rx[portId].rxFreeOverflow);
	printf("Rx Free Mutual Lock           : %u\n", 
	       rx[portId].rxFreeLock);
	printf("Rx Free Late Ntf Enabled      : %u\n", 
	       rx[portId].rxFreeLateNotificationEnabled);
	printf("Rx Free Low CB                : %u\n",
	       rx[portId].rxFreeLowCallback);
	printf("Rx From H/W Q (Disable)       : %u\n", 
	       rx[portId].rxDuringDisable);
	printf("Rx From S/W Q (Disable)       : %u\n", 
	       rx[portId].rxSwQDuringDisable);
#endif
	printf("Rx unexpected errors          : %u (should be 0)\n", 
	       rx[portId].rxUnexpectedError);	
	
#ifndef NDEBUG
	numBuffersInTx = tx[portId].txQOK + 
	    tx[portId].txQDelayed -
	    tx[portId].txDoneClientCallback - 
	    tx[portId].txDoneSwQDuringDisable - 
	    tx[portId].txDoneDuringDisable;

	printf("# Tx Buffers currently for transmission : %u\n", 
	       numBuffersInTx);
	
	numBuffersInRx = rx[portId].rxFreeRepOK + 
	    rx[portId].rxFreeRepDelayed - 
	    rx[portId].rxFrameClientCallback - 
	    rx[portId].rxSwQDuringDisable - 
	    rx[portId].rxDuringDisable;

	printf("# Rx Buffers currently for reception    : %u\n", 
	       numBuffersInRx);	

	totalBuffers += numBuffersInRx + numBuffersInTx;
#endif
    }
    
    printf("---------------------------------------\n");

#ifndef NDEBUG
    printf("\n");
    printf("Mbufs :\n");   
    printf("Unchained mbufs               : %u\n", 
	   stats.unchainedMBufs);
    printf("Chained bufs                  : %u\n", 
	   stats.chainedMBufs);

    printf("\n");
    printf("Software queue usage :\n");   
    printf("Buffers added to S/W Q        : %u\n", 
	   stats.addToSwQ);
    printf("Buffers removed from S/W Q    : %u\n", 
	   stats.removeFromSwQ);

    printf("\n");
    printf("Hardware queues callbacks :\n");   

    for(portId=IX_ETH_PORT_1; portId<=IX_ETH_PORT_2; portId++)
    {
	rxFreeCallbackCounter += rx[portId].rxFreeLowCallback;
	txCallbackCounter += tx[portId].txLowThreshCallback;
    }
    printf("Tx Done QM Callback invoked   : %u\n", 
	   stats.txDoneCallbackCounter);
    printf("Tx QM Callback invoked        : %u\n", 
	   txCallbackCounter);
    printf("Rx QM Callback invoked        : %u\n", 
	   stats.rxCallbackCounter);
    printf("Rx Free QM Callback invoked   : %u\n", 
	   rxFreeCallbackCounter);
#endif
    printf("Unexpected errors in CB       : %u (should be 0)\n", 
	   stats.unexpectedError);
    printf("\n");

    printf("Hardware queues levels :\n");   
    printf("Transmit Port 1 Q             : %u \n",numTx0Entries);
    printf("Transmit Port 2 Q             : %u \n",numTx1Entries);
    printf("Transmit Done Q               : %u \n",numTxDoneEntries);
    printf("Receive Q                     : %u \n",numRxEntries);
    printf("Receive Free Port 1 Q         : %u \n",numRxFree0Entries);
    printf("Receive Free Port 2 Q         : %u \n",numRxFree1Entries);

#ifndef NDEBUG
    printf("\n");
    printf("# Total Buffers accounted for : %u\n", 
	   totalBuffers);

    numBuffersInSwQ = ixEthAccDataStats.addToSwQ - 
	ixEthAccDataStats.removeFromSwQ;

    printf("    Buffers in S/W Qs         : %u\n", 
	   numBuffersInSwQ);
    printf("    Buffers in H/W Qs or NPEs : %u\n", 
	   totalBuffers - numBuffersInSwQ);
#endif

#if IX_ETH_ACC_FPATH_AWARE
    printf("\n");
    printf("Currently Running In FAST PATH AWARE MODE\n");
    printf("Tx Frames Currently Submitted port 0 : %u\n",
	   txFramesCurrentlySubmitted[0]);
    printf("Tx Frames Currently Submitted port 1 : %u\n",
	   txFramesCurrentlySubmitted[1]);
#else
    printf("\n");
    printf("Currently Running In NON - FAST PATH AWARE MODE\n");
#endif
}





