/***
 * @file IxFpathConfigControl.c
 *
 * @author Intel Corporation
 * @date 15 April 2002
 *
 * @brief Fast Path Config Control
 *
 * responsible for controlling:-
 *  @li setting Ethernet Type, 
 *  @li setting Wan Mac Address callback, 
 *  @li setting/clearing of  classifier/modifier templates,
 *  @li retrieval of Fpath statistics.
 *
 * Design Notes:
 *    
 *
 * -- Intel Copyright Notice --
 * 
 * @par
 * Copyright 2002-2003 Intel Corporation All Rights Reserved.
 * 
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.
 * 
 * @par
 * The Material is protected by worldwide copyright and trade secret laws
 * and treaty provisions. No part of the Material may be used, copied,
 * reproduced, modified, published, uploaded, posted, transmitted,
 * distributed, or disclosed in any way except in accordance with the
 * applicable license agreement .
 * 
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel, except in accordance with the
 * applicable license agreement.
 * 
 * @par
 * Unless otherwise agreed by Intel in writing, you may not remove or
 * alter this notice or any other notice embedded in Materials by Intel
 * or Intel's suppliers or licensors in any way.
 * 
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * 
 * @par
 * -- End Intel Copyright Notice --
 */

/*
 * Sytem defined include files
 */
#include <stdio.h>


/*
 * Put the user defined include files required.
 */
#include "IxTypes.h"
#include "IxAssert.h"
#include "IxOsServices.h"
#include "IxFpathDefines_p.h"
#include "IxFpathConfigControl_p.h"
#include "IxFpathNpeInterface_p.h"

/*
 *#defines
 */
#define IX_FPATHACC_DUMMY_VC_ID IX_ATM_MAX_NUM_VC
#define IX_FPATHACC_MAX_NUM_FPINDEX 8
#define IX_FPATHACC_INVALID_FPINDEX IX_FPATHACC_MAX_NUM_FPINDEX+1


/* 
 * Local variables go here
 */

static BOOL ixFpathCfgCtrlInitDone = FALSE;


/* This type defines a descriptor that describes everything
 * ixFpathAcc needs to know about a VC
 */
typedef struct
{
   IxFpathIndex fpathIndex;
   IxFpathAccNewWanMacAddrCallback clientWanMacCallback;
   IxFpathAccClientId wanMacClientId;
} IxFpathCfgCtrlVcDesc;


/* A table of Vc descriptors that can be indexed by
 * npeVcId - this one greater than the number of VCs
 * to accommodate a dummy entry.
 */
static IxFpathCfgCtrlVcDesc
ixFpathCfgCtrlNpeVcDesc[ IX_ATM_MAX_NUM_VC+1 ];

/* A table of npeVcIds that can be indexed by fastpathIndex.
*/
static IxAtmNpeRxVcId
ixFpathCfgCtrlNpeRxVcId[ IX_FPATHACC_MAX_NUM_FPINDEX ];

/* statistics counters
*/
static unsigned int ixFpathAccWanMacCallbackCount = 0;
static unsigned int ixFpathAccDummyWanMacCallbackCount = 0;
static unsigned int ixFpathAccInvalidWanMacCallbackCount = 0;
static unsigned int ixFpathAccDisabledWanMacCallbackCount = 0;

/* static prototypes
*/
PRIVATE void ixFpathNpeIfWanMacMsgCallback(IxFpathIndex fpIndex, unsigned char *macAddr);
PRIVATE BOOL ixFpathFreeFpIndexGet(IxFpathIndex *fpIndex);

/* ------------------------------------------------------
 * The default client Wan Mac Address message handler
 */
static void
ixFpathAccDummyWanMacAddrCallback (
    IxFpathAccClientId clientId,
    unsigned char *macAddress)
{
    if(ixFpathAccDummyWanMacCallbackCount == 0)
    {
    ixOsServLog( LOG_ERROR,
                 "WanMacAddress learned on an Active VC with no Wan Mac CB set, %i ",
                 clientId,
                 0,0,0,0,0);
    }
    ixFpathAccDummyWanMacCallbackCount++;
}



/* ------------------------------------------------------
 * The local Wan Mac Address message handler that we register
 * maps fpathIndex from the Npe, to clientId, and invokes
 * the registered client Wan Mac Address handler
 */
PRIVATE void
ixFpathNpeIfWanMacMsgCallback(IxFpathIndex fpIndex, unsigned char *macAddr)
{
    IxAtmNpeRxVcId npeVcId;

    IX_ASSERT( macAddr != NULL ); 

    /* sanity check the fpIndex */
    if ( ( fpIndex >= 0 ) && 
         ( fpIndex < IX_FPATHACC_MAX_NUM_FPINDEX ) )
    {
         if ( ixFpathCfgCtrlNpeRxVcId[fpIndex] 
              < IX_FPATHACC_DUMMY_VC_ID)
         {
             /* We guarantee the npeVcId is in range */
             npeVcId = ixFpathCfgCtrlNpeRxVcId[fpIndex];
            
             /* This is safe - there is always a dummy callback registered */
             ixFpathCfgCtrlNpeVcDesc[npeVcId].clientWanMacCallback(
             ixFpathCfgCtrlNpeVcDesc[npeVcId].wanMacClientId,
                 macAddr);
             ixFpathAccWanMacCallbackCount++;

             return;
         }
         else
         {
             if( ixFpathAccDisabledWanMacCallbackCount == 0 )
             {
                ixOsServLog( LOG_ERROR, 
                             "WanMacAddress learned on an disabled FpIndex, %i",
                             fpIndex,
                             0,0,0,0,0);
             }
             ixFpathAccDisabledWanMacCallbackCount++;
         }
    }
    else
    {
        if( ixFpathAccInvalidWanMacCallbackCount == 0 )
        {
            ixOsServLog( LOG_ERROR, 
                        "WanMacAddress learned on an invalid FpIndex, %i",
                        fpIndex,
                        0,0,0,0,0);
        }
        ixFpathAccInvalidWanMacCallbackCount++;
    }
}

/* ------------------------------------------------------
 * Initialise 
 */
IX_STATUS
ixFpathCfgCtrlInit( void )
{
    IxAtmNpeRxVcId npeVcId;
    IxFpathIndex  fpathIndex;
    IxFpathCfgCtrlVcDesc *fpVcDesc;
    IX_STATUS returnStatus = IX_FAIL;

    if ( ixFpathCfgCtrlInitDone ) 
    {
        return IX_FAIL;
    }

    /* initialise the fast path Vc descriptor table */
    /* Note the extra entry at the end of table is dummy for unsed fpIndexTable entries */
    for( npeVcId = 0; npeVcId <= IX_ATM_MAX_NUM_VC; npeVcId++ )
    {
        fpVcDesc = &ixFpathCfgCtrlNpeVcDesc[ npeVcId ];

        fpVcDesc->fpathIndex = IX_FPATHACC_INVALID_FPINDEX;
        fpVcDesc->clientWanMacCallback = ixFpathAccDummyWanMacAddrCallback;
        fpVcDesc->wanMacClientId = 0 ;
    }

    /* initialize the fast path index table */
    for( fpathIndex = 0; fpathIndex < IX_FPATHACC_MAX_NUM_FPINDEX;  fpathIndex++)
    {
        ixFpathCfgCtrlNpeRxVcId[fpathIndex] = IX_FPATHACC_DUMMY_VC_ID;  /* Always point to dummy */
    }

    /* Initialise the Npe Interface */
    if (ixFpathNpeIfInit() == IX_SUCCESS)
    {
        returnStatus = ixFpathNpeIfWanMacCallbackRegister(ixFpathNpeIfWanMacMsgCallback);
    }

    if ( returnStatus == IX_SUCCESS ) 
    {
        ixFpathCfgCtrlInitDone = TRUE;
    }

    return returnStatus;

}


/* ----------------------------------------------------
 * Search for and return a free fpIndex
 */
PRIVATE BOOL
ixFpathFreeFpIndexGet(IxFpathIndex *fpIndex)
{
    IxFpathIndex fpId;

    for( fpId = 0; 
         ( ( fpId < IX_FPATHACC_MAX_NUM_FPINDEX ) && 
           ( ixFpathCfgCtrlNpeRxVcId[fpId] < IX_FPATHACC_DUMMY_VC_ID ) ); 
         fpId++ ) 
    {
        /* we dont do anything, we are only looking for the index
         * at which the loop exits
         */
    }
    *fpIndex = fpId;
    return( fpId != IX_FPATHACC_MAX_NUM_FPINDEX );
}



/* ------------------------------------------------------
 * set a classifier/modifier template pair
 */
IX_STATUS
ixFpathCfgCtrlTemplatePairSet(IxAtmNpeRxVcId npeVcId, 
                              unsigned char *classifierTemplate, 
                              int classifierTemplateLength, 
                              unsigned char *modifierTemplate, 
                              int modifierTemplateLength )
{
    IxFpathIndex fpIndex;

    IX_ENSURE( ixFpathCfgCtrlInitDone, "\nixFpathCfgCtrl used before initialised\n");

    /* Get the fpIndex for this VC */
    fpIndex = ixFpathCfgCtrlNpeVcDesc[npeVcId].fpathIndex;

    if ( fpIndex == IX_FPATHACC_INVALID_FPINDEX )
    {
        /* check  that a free fpIndex exists */
        if ( !ixFpathFreeFpIndexGet( &fpIndex ) )
        {
            return IX_FAIL;
        } 
    }
    else
    {
        /* Sanity check.
         * The VC table says this is the fpIndex associated with the VC,
         * so lets check that the fpIndex table agrees
         */
        IX_ENSURE((ixFpathCfgCtrlNpeRxVcId[fpIndex] == npeVcId),"ixFpathAcc Inconsistent npeVcid/FpIndex");

        /* active fpath indices must be disabled before downloading */
        if( ixFpathNpeIfDisableSend(IX_NPEMH_NPEID_NPEA, fpIndex, npeVcId)
                != IX_SUCCESS )
        {
            return IX_FAIL;;
        }
    }
    
    /* Always disable the VC and fpIndex */
    ixFpathCfgCtrlNpeVcDesc[npeVcId].fpathIndex 
        = IX_FPATHACC_INVALID_FPINDEX;

    ixFpathCfgCtrlNpeRxVcId[fpIndex] = IX_FPATHACC_DUMMY_VC_ID;

    /* Download the modifier, classifier, and enable */
    if ( ixFpathNpeIfModifierDownload(
             IX_NPEMH_NPEID_NPEB, 
             fpIndex, 
             modifierTemplate,
             modifierTemplateLength ) == IX_SUCCESS ) 
    {
        if ( ixFpathNpeIfClassifierDownload( 
                 IX_NPEMH_NPEID_NPEA, 
                 fpIndex, 
                 classifierTemplate,
                 classifierTemplateLength) == IX_SUCCESS ) 
        {
            if ( ixFpathNpeIfEnableSend(
                     IX_NPEMH_NPEID_NPEA, 
                     fpIndex, 
                     npeVcId) == IX_SUCCESS ) 
            {
                ixFpathCfgCtrlNpeVcDesc[npeVcId].fpathIndex = fpIndex;
                ixFpathCfgCtrlNpeRxVcId[fpIndex] = npeVcId;
                return IX_SUCCESS;
            }
        }
    }
    return IX_FAIL;
}

/* ------------------------------------------------------
 * clear a classifier/modifier template pair
 */
IX_STATUS
ixFpathCfgCtrlTemplatePairClear(IxAtmNpeRxVcId npeVcId)
{
    IxFpathIndex fpIndex;
    IX_STATUS returnStatus = IX_SUCCESS;

    IX_ENSURE( ixFpathCfgCtrlInitDone, "\nixFpathCfgCtrl used before initialised\n");

    /* Get the fpIndex for this VC */
    fpIndex = ixFpathCfgCtrlNpeVcDesc[npeVcId].fpathIndex;

    if ( fpIndex == IX_FPATHACC_INVALID_FPINDEX )
    {
        /* This VC is already disabled */
        returnStatus = IX_FAIL;
    }
    else
    {
        /* Sanity check.
         * The VC table says this is the fpIndex associated with the VC,
         * so lets check that the fpIndex table agrees
         */
        IX_ASSERT( ixFpathCfgCtrlNpeRxVcId[fpIndex] == npeVcId );
   
        /* active fpath indices must be disable before downloading */
        returnStatus 
            = ixFpathNpeIfDisableSend(IX_NPEMH_NPEID_NPEA, fpIndex, npeVcId);

        /* N.B. This should only be done when the template is disabled. */
        if (returnStatus == IX_SUCCESS)
        {
            returnStatus = ixFpathNpeIfStatisticsClear(
                               IX_NPEMH_NPEID_NPEA, 
                               fpIndex);
        }
        
        /* Always disable the VC and fpIndex */
        ixFpathCfgCtrlNpeVcDesc[npeVcId].fpathIndex 
            = IX_FPATHACC_INVALID_FPINDEX;

        ixFpathCfgCtrlNpeRxVcId[fpIndex] 
            = IX_FPATHACC_DUMMY_VC_ID;
    }

    return returnStatus;
}
    
/* ------------------------------------------------------
 * set the ethernet type for the specified port
 */
IX_STATUS
ixFpathCfgCtrlEtherTypeSet(IxFpathEthernetPort port, 
                           UINT16 etherType )
{
    IxNpeMhNpeId npeId;

    IX_ENSURE( ixFpathCfgCtrlInitDone, "ixFpathCfgCtrl used before initialised");

    /* convert the port to an Npe Id */
    switch( port )
    {
    case IX_FPATHACC_ETHERNET_PORT0:
        npeId = IX_NPEMH_NPEID_NPEB;
        break;
    default:
        /* bad port number */
        return IX_FAIL;
        break;
    }

    /* send the ether type to the Npe */
    return ixFpathNpeIfEtherTypeSend( npeId, etherType );
}

/* ------------------------------------------------------
 * set the Wan Mac Address callback
 */
IX_STATUS
ixFpathCfgCtrlWanMacAddrCallbackSet( 
    IxFpathAccNewWanMacAddrCallback macAddrCallback, 
    IxAtmNpeRxVcId npeVcId, 
    IxFpathAccClientId clientId)
{
    IX_ENSURE( ixFpathCfgCtrlInitDone, "\nixFpathCfgCtrl used before initialised\n");

    /* Sanity check, before writing to the table check that the 
     * Vcid is in a valid range, and not enabled for fast path. 
     */
    IX_ENSURE((npeVcId < IX_FPATHACC_DUMMY_VC_ID), 
              "\nixFpathCfgCtrlWanMacAddrCallbackSet Bad npeVcid\n");

    if( ixFpathCfgCtrlNpeVcDesc[npeVcId].fpathIndex == IX_FPATHACC_INVALID_FPINDEX )
    {
        /* if the user is trying to clear the callback ? */
        if( macAddrCallback == NULL )
        {
            /* restore the dummy callback */
            ixFpathCfgCtrlNpeVcDesc[npeVcId].clientWanMacCallback 
                = ixFpathAccDummyWanMacAddrCallback;
            ixFpathCfgCtrlNpeVcDesc[npeVcId].wanMacClientId 
                = 0;
        }
        else
        {
            /* store the new callback */
            ixFpathCfgCtrlNpeVcDesc[npeVcId].clientWanMacCallback 
                = macAddrCallback;
            ixFpathCfgCtrlNpeVcDesc[npeVcId].wanMacClientId 
                = clientId;
        }
        return IX_SUCCESS;
    }
    return IX_FAIL;
}

/* ------------------------------------------------------
 * get the fast path statistics
 */
IX_STATUS ixFpathCfgCtrlStatisticsGet(
    IxAtmNpeRxVcId npeVcId,
    int *packetCnt)
{
    IxFpathIndex fpIndex;
    IX_STATUS returnStatus = IX_SUCCESS;

    IX_ENSURE( ixFpathCfgCtrlInitDone, "\nixFpathCfgCtrl used before initialised\n");

    /* Get the fpIndex for this VC */
    fpIndex = ixFpathCfgCtrlNpeVcDesc[npeVcId].fpathIndex;

    if ( fpIndex == IX_FPATHACC_INVALID_FPINDEX )
    {
        /* This VC is disabled */
        returnStatus = IX_FAIL;
    }
    else
    {
        /* Sanity check.
         * The VC table says this is the fpIndex associated with the VC,
         * so lets check that the fpIndex table agrees
         */
        IX_ASSERT( ixFpathCfgCtrlNpeRxVcId[fpIndex] == npeVcId );

        /* active fpath indices must be disabled before downloading */
        returnStatus 
            = ixFpathNpeIfStatisticsGet(IX_NPEMH_NPEID_NPEA, fpIndex, packetCnt);

    }

    return returnStatus;
}



/* ----------------------------------------------------
 * Show
 */
void
ixFpathCfgCtrlShow( void )
{
    int i;

    if(!ixFpathCfgCtrlInitDone)
    {
        printf("\nNot Initialized");    
    }
    else
    {
        printf("\n---------------------------");
        printf("\nWAN MAC callback statistics");
        printf("\n---------------------------");
        printf("\nWan Mac Callback Count %u",
                ixFpathAccWanMacCallbackCount);
        printf("\nDummy Wan Mac Callback Count %u",
                ixFpathAccDummyWanMacCallbackCount);  
        printf("\nInvalid Wan Mac Callback Count %u",
                ixFpathAccInvalidWanMacCallbackCount);    
        printf("\nDisabled Wan Mac Callback Count %u",
                ixFpathAccDisabledWanMacCallbackCount);    

        printf("\n--------");
        printf("\nVC Table");
        printf("\n--------");
        printf("\nVC\tfpIndex\tClientId");
        for( i=0; i<IX_ATM_MAX_NUM_VC+1; i++ )
        {

            if( ixFpathCfgCtrlNpeVcDesc[i].fpathIndex < IX_FPATHACC_MAX_NUM_FPINDEX )
            {
                if( ixFpathCfgCtrlNpeRxVcId[ixFpathCfgCtrlNpeVcDesc[i].fpathIndex] != (IxAtmNpeRxVcId) i)
                {
                    printf("\nWarning Inconsistent fpIndex in the following entry"); 
                }
            }
            printf("\n%i\t%i\t%i",
                   i,
                   ixFpathCfgCtrlNpeVcDesc[i].fpathIndex,
                   ixFpathCfgCtrlNpeVcDesc[i].wanMacClientId );
        }

        printf("\n-----------");
        printf("\nIndex Table");
        printf("\n-----------");
        printf("\nfpIndex\tNpeVcId");
        for( i=0; i<IX_FPATHACC_MAX_NUM_FPINDEX; i++)
        {
            printf("\n%i\t%i",
                   i,
                   ixFpathCfgCtrlNpeRxVcId[i] );

        }
        ixFpathNpeIfShow();
    }
}



