/**
 * @file IxCryptoAccCodelet_p.h
 *
 * @date Jun-8-2003
 *
 * @brief This is private header file for the Crypto Access Component Codelet
 *
 * -- Intel Copyright Notice --
 * 
 * @par
 * Copyright 2002-2003 Intel Corporation All Rights Reserved.
 * 
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.
 * 
 * @par
 * The Material is protected by worldwide copyright and trade secret laws
 * and treaty provisions. No part of the Material may be used, copied,
 * reproduced, modified, published, uploaded, posted, transmitted,
 * distributed, or disclosed in any way except in accordance with the
 * applicable license agreement .
 * 
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel, except in accordance with the
 * applicable license agreement.
 * 
 * @par
 * Unless otherwise agreed by Intel in writing, you may not remove or
 * alter this notice or any other notice embedded in Materials by Intel
 * or Intel's suppliers or licensors in any way.
 * 
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * 
 * @par
 * -- End Intel Copyright Notice --
 */


#ifndef IxCryptoAccCodelet_p_H
#define IxCryptoAccCodelet_p_H

#include "IxTypes.h"
#include "IxOsBuffMgt.h"
#include "IxNpeDl.h"
#include "IxOsCacheMMU.h"
#include "IxCryptoAcc.h"

/** Global variables */
extern BOOL forwardRegisterCBCalled;    /**< callback flag */
extern BOOL reverseRegisterCBCalled;    /**< callback flag */
extern IxCryptoAccRegisterCompleteCallback registerCB; /**< register callback 
                                                        * function
                                                        */
extern IxCryptoAccPerformCompleteCallback performCB;   /**< perform callback 
                                                        * function
                                                        */


/**************************************************************************
 *                           Function Prototype 
 **************************************************************************/

/**
 * @fn  ixCryptoAccCodeletInit (void)
 * @brief This function initialises the neccessary components.
 *
 * @note The initialisation sequence is as follows:
 * <UL>
 *  <LI> Initialise Network Processor Engine (NPE)
 *  <LI> Initialise Queue manager
 *  <LI> Start NPE
 *  <LI> Initialise Crypto Access component
 *  <LI> Start Queue manager dispatcher. 
 * </UL>  
 *
 * @param None
 *
 * @return @li IX_SUCCESS - Successful initialisation.
 *         @li IX_FAIL - Initialisation failed.
 */
PRIVATE IX_STATUS ixCryptoAccCodeletInit (void);

/**
 * @def IX_CRYPTO_ACC_CODELET_DRV_DMA_FREE
 *
 * @brief To free the memory allocated through IX_CRYPTO_ACC_DRV_DMA_MALLOC
 *        function.
 *
 * @note Example of usage: IX_CRYPTO_ACC_CODELET_DRV_DMA_FREE(pData, size)
 *
 * @param "UINT32 [in] *pData" - memory block pointer to be freed.
 * @param "UINT32 [in] size" - size of memory block to be freed.
 *
 * @return None
 *
 */
#define IX_CRYPTO_ACC_CODELET_DRV_DMA_FREE(pData, size) \
            IX_ACC_DRV_DMA_FREE(pData,size)


/**
 * @fn ixCryptoAccCodeletBufIsQEmpty (void)
 *
 * @brief This inline function is used to check whether the buffer
 *        queue is empty.
 *
 * @param None
 *
 * @return BOOL
 *             - TRUE - the queue is full.
 *             - FALSE - the queue is not full.
 */
INLINE BOOL 
ixCryptoAccCodeletBufIsQEmpty (void);


/**
 * @fn ixCryptoAccCodeletBufIsQFull(void)
 *
 * @brief This inline function is used to check whether the buffer
 *        queue is full.
 *
 * @param None
 *
 * @return BOOL
 *             - TRUE - the queue is full.
 *             - FALSE - the queue is not full.
 */
INLINE BOOL 
ixCryptoAccCodeletBufIsQFull (void);


/**
 * @fn ixCryptoAccCodeletCryptoMbufToQTailAdd (IX_MBUF *pMbufToAdd)
 *
 * @brief This function adds a mbuf to the tail of the queue.  
 *
 * @param "IX_MBUF* [in] pMbufToAdd" - Pointer of the mbuf to be added  to the
 *                    tail of the queue.
 *
 * @return None.
 */
void ixCryptoAccCodeletCryptoMbufToQTailAdd (IX_MBUF *pMbufToAdd);


/**
 * @fn ixCryptoAccCodeletCryptoMbufFromQHeadRemove(IX_MBUF **pMbufToRem)
 *
 * @brief This function removes a mbuf from the head of the queue.
 *
 * @param "IX_MBUF** [in] pMbufToRem" - Pointer to a mbuf pointer to be
 *                     removed from the head of the queue.
 *
 * @return None.
 */
void ixCryptoAccCodeletCryptoMbufFromQHeadRemove (IX_MBUF **pMbufToRem);


/**
 * @fn ixCryptoAccCodeletCryptoMemPoolInit(UINT32 pktLen)
 *
 * @brief This function initialises mbuf pool for the Crypto
 *        Access Codelet.
 *
 * @param "UINT32 [in] pktLen" - packet length size for mbuf
 *
 * @return @li IX_SUCCESS - Successful mbuf pool initialisation.
 *         @li IX_FAIL - Mbuf pool initialisation failed.
 */
IX_STATUS ixCryptoAccCodeletCryptoMemPoolInit (UINT32 pktLen);


/**
 * @fn IX_STATUS ixCryptoAccCodeletCryptoMemPoolFree (UINT32 pktLen)
 *
 * @brief Free all resources in Crypto Access Codelet MBUF pool
 *
 * @param "UINT32 [in] pktLen" - packet length size for mbuf
 *
 * @return None
 */
void ixCryptoAccCodeletCryptoMemPoolFree (UINT32 pktLen);
 
 
/**
 * @fn ixCryptoAccCodeletNpeInit(IxNpeDlNpeId npeId, IxNpeDlBuildId buildId)
 *
 * @brief This function is to download the latest crypto NPE image into NPE.
 *
 * @param "IxNpeDlNpeId [in] npeId" - The ID of the NPE which the microcode 
 *                       downloaded to.
 * @param "IxNpeDlBuildId [in] buildId" - The build ID for image which indicate 
 *                       functionality
 *
 * @return @li IX_SUCCESS - Successful NPE initialisation.
 *         @li IX_FAIL - NPE initialisation failed.
 */
IX_STATUS ixCryptoAccCodeletNpeInit (
                        IxNpeDlNpeId npeId,
                        IxNpeDlBuildId buildId);


/**
 * @fn ixCryptoAccCodeletDispatcherPoll(void)
 *
 * @brief This function polls the queues when poll mode set in the 
 *        function ixCryptoAccCodeletDispatcherStart.
 *
 * @param None
 * 
 * @return None
 */
void ixCryptoAccCodeletDispatcherPoll (void);


/**
 * @fn ixCryptoAccCodeletDispatcherStart(BOOL useInterrupt)
 *
 * @brief  This function starts the Queue manager dispatcher.
 *
 * @param "BOOL [in] useInterrupt" - start in interrupt or poll mode. TRUE is 
 *                                   interrupt mode while FALSE is poll mode.
 * 
 * @return  @li IX_SUCCESS - Dispatcher successfully started.
 *          @li IX_FAIL - Error starting dispatcher.
 */
IX_STATUS ixCryptoAccCodeletDispatcherStart (BOOL useInterrupt);


/**
 * @fn ixCryptoAccCodeletRegisterCtx(
          UINT32 opType,
          IxCryptoAccCtx *pForwardCtx,
          IxCryptoAccCtx *pReverseCtx,
          UINT32 *pForwardCtxId,
          UINT32 *pReverseCtxId)
 *
 * @brief  This function registers to the Crypto Access Component for
 *         three different operation which demonstrates the capability
 *         of the security hardware accelerator.
 *         The types of operation are:
 *         1 - Encryption and Decryption
 *         2 - Authentication: Calculation and Check
 *         3 - Encryption/Decryption with authentication
 *
 * @param "UINT32 [in] opType" - Type of operation to be registered. This
 *                 parameter will only take 3 different unsigned
 *                 integer as shown above in the brief, e.g. 1 for
 *                 encryption and decryption only and so on.
 * @param "IxCryptoAccCtx* [in] pForwardCtx" - pointer to Crypto Context for
 *                                            forward direction, e.g encryption,
 *                                            auth_calc.
 * @param "IxCryptoAccCtx* [in] pReverseCtx" - pointer to Crypto Context for 
 *                                            reverse direction, e.g decryption,
 *                                            auth_check.
 * @param "UINT32* [in] pForwardCtxId" - pointer to Crypto Context ID
 * @param "UINT32* [in] pReverseCtxId" - pointer to Crypto Context ID
 * 
 * @return @li IX_SUCCESS - Register successfully
 *         @li IX_FAIL - Error in registering
 */
IX_STATUS ixCryptoAccCodeletRegisterCtx (
            UINT32 opType,
            IxCryptoAccCtx *pForwardCtx,
            IxCryptoAccCtx *pReverseCtx,
            UINT32 *pForwardCtxId,
            UINT32 *pReverseCtxId);


/**
 * @fn ixCryptoAccCodeletUnregisterCtx (void)
 *
 * @brief  This function unregisters all the contexts registered.
 *
 * @param None
 * 
 * @return None
 */
void ixCryptoAccCodeletUnregisterCtx (void);


/**
 * @fn ixCryptoAccCodeletDispatcherStop (BOOL useInterrupt)
 *
 * @brief Stop QMgr dispatcher thread if QMgr dispatcher runs in poll mode 
 *        or unbind QMgr dispatcher from interrupt if it runs in interrupt mode.
 *
 * @param "BOOL [in] useInterrupt" - start in interrupt or poll mode. TRUE is 
 *                                   interrupt mode while FALSE is poll mode.
 * 
 * @return  @li IX_SUCCESS - Dispatcher successfully stopped.
 *          @li IX_FAIL - Error stopping dispatcher.
 */
IX_STATUS ixCryptoAccCodeletDispatcherStop (BOOL useInterrupt);


/**
 * @fn ixCryptoAccCodeletUsageShow (void)
 *
 * @brief Display user guide for Linux platform or VxWorks platform
 *
 * @param None
 *
 * @return None
 */
void ixCryptoAccCodeletUsageShow (void);

#endif /* IxCryptoAccCodelet_p_H */

