/**
 * @file  IxFpathAccCodeletBridgeDemo.c
 * @date  May 29 2002
 * @brief Bridge VC Fast Path Demonstration
 *
 * This code provides a codelet entry point to demonstrate how to setup 
 * a fast path stream from ATM to Ethernet to support bridging of 
 * Ethernet packets using RFC1483 VC Mux.. 
 * 
 * -- Intel Copyright Notice --
 * 
 * @par
 * Copyright 2002-2003 Intel Corporation All Rights Reserved.
 * 
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.
 * 
 * @par
 * The Material is protected by worldwide copyright and trade secret laws
 * and treaty provisions. No part of the Material may be used, copied,
 * reproduced, modified, published, uploaded, posted, transmitted,
 * distributed, or disclosed in any way except in accordance with the
 * applicable license agreement .
 * 
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel, except in accordance with the
 * applicable license agreement.
 * 
 * @par
 * Unless otherwise agreed by Intel in writing, you may not remove or
 * alter this notice or any other notice embedded in Materials by Intel
 * or Intel's suppliers or licensors in any way.
 * 
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * 
 * @par
 * -- End Intel Copyright Notice --
 */


#include <stdio.h>
#include <time.h>

#include "IxTypes.h"
#include "IxAssert.h"
#include "IxAtmTypes.h"
#include "IxOsCacheMMU.h"
#include "IxEthAcc.h"
#include "ix_ossl.h"
#include "IxFpathAcc.h"
#include "IxFpathAccCodeletMain_p.h"
#include "IxFpathAccCodeletBufMan_p.h"

/*
 * Macros
 */
#define WAN_MAC_ADDR_LEARNING_DB_SIZE \
    (WAN_MAC_LEARNING_DB_DISPLAY_NUM_ENTRIES * 2)

#define WAN_ADDR_DISPLAY_DELAY_MS 100

/*
 * Types
 */
typedef struct 
{
    VcNumber vc;
    UINT8 macAddress[IX_IEEE803_MAC_ADDRESS_SIZE];
    unsigned int timestamp;
} WanMacLearningDbEntry;

typedef struct 
{
  unsigned int index;      
  unsigned int wanMacCount;
  WanMacLearningDbEntry entry[WAN_MAC_ADDR_LEARNING_DB_SIZE];
} IxFpCodeletWanMacLearningDb;


/*
 * Local Data
 */
static IxFpCodeletWanMacLearningDb ixFpCodeletWanMacLearningDb;
static IxFpCodeletWanMacLearningDb ixFpCodeletWanMacLearningDbSnapShot;
static BOOL ixFpCodeletWanMacCbTraceEnabled = FALSE;
static ix_ossl_thread_t ixFpCodeletWanMacDisplayingTaskId;
static BOOL taskEnabled = TRUE;



/*
 * Forward Declarations
 */
PRIVATE IX_STATUS
ixFpCodeletEthBVcCreate(
    VcNumber vc, 
    IxAtmLogicalPort port,
    unsigned int vpi,
    unsigned int vci );

PRIVATE void
ixFpCodeletNewWanMacAddressHandler(
    IxFpathAccClientId userId,
    unsigned char *macAddress);

void  
ixFpCodeletAtmRxEthernetBridge(
    IxAtmdAccUserId userId,
    IX_MBUF * buf);

ix_error
ixFpCodeletWanMacDisplayingTask(
    void* arg,
    void **ptrRetObj);

void 
ixFpCodeletWanMacLearningDbInit( void );


/* ------------------------------------------------------------------------
 * Ethernet Bridge Demo
 *
 * This function illustrates the creation of a fast path VC 
 * for an Ethernet Bridge application.
 */
PUBLIC void
ixFpCodeletEthBDemoStart( void )
{
    /* First initialize the codelet and other supporting s/w components */
    if( ixFpCodeletInit () != IX_SUCCESS )
    {
        IXFPATH_CODELET_ERROR("Failed to initialize codelet");
        return;
    }

    ixFpCodeletWanMacLearningDbInit ();


    /* Create an ATM VC and configure it for fast path */ 
    if( ixFpCodeletEthBVcCreate ( 
            VC_0,
            FPATH_BRIDGED_TEST_UTOPIA_PORT,
            FPATH_BRIDGED_TEST_VPI,
            FPATH_BRIDGED_TEST_VCI ) != IX_SUCCESS )
    {
        IXFPATH_CODELET_ERROR("Failed to establish fast path");
        return;
    }

    taskEnabled = TRUE;
    /* Create a task to Display the MAC address buffer
     */
    if( ix_ossl_thread_create (
            ixFpCodeletWanMacDisplayingTask,
            NULL,
            &ixFpCodeletWanMacDisplayingTaskId) != IX_OSSL_ERROR_SUCCESS)
    {
        IXFPATH_CODELET_ERROR("Failed to create Wan Mac Display Task");
    }
}

/** 
 * @brief Ethernet Bridge Demo Stop
 *
 * This function illustrates the correct steps for deletion of a fast path VC 
 * for the Ethernet Bridge application. See IxFpCodelet.h.
 */
PUBLIC void
ixFpCodeletEthBDemoStop( void )
{
    VcNumber vc;
    IxFpCodeletFastPathConnection *connection;
    for( vc=0; vc<MAX_NUM_VCS; vc++ )
    {
        connection = ixFpCodeletFastPathConnectionGet (vc);
        
        if( connection->inUse &&
            (connection->encapsulationType == TYPE_1483BVCMUX ))
        {
            IXFPATH_CODELET_TRACE((traceBuf,"Shutting Down VC %d\n",vc));
            if( ixFpCodeletFastPathVcDelete(vc) == IX_SUCCESS ) 
            {
                IXFPATH_CODELET_TRACE((traceBuf,"Done\n"));
            }
            else
            {
                IXFPATH_CODELET_ERROR("ixFpCodeletFastPathVcDelete - Failed");
            }
        }    
    }

    /* Kill the task */
    taskEnabled = FALSE;
}

/**
 * @brief Ethernet Bridge VC Create
 *
 * This function illustrates the creation of a fast path VC 
 * for an Ethernet Bridge application.
 *
 * @param vc   - demo connection number.
 * @param port - The ATM PHY logical port number.
 * @param vpi  - The virtual path index for the ATM VC
 * @param vci  - The virtual channel index for the ATM VC
 */
PRIVATE IX_STATUS
ixFpCodeletEthBVcCreate(
    VcNumber vc, 
    IxAtmLogicalPort port,
    unsigned int vpi,
    unsigned int vci )
{
    IxFpCodeletFastPathConnection *connection;
    IxFpCodeletTemplate classifier;
    IxFpCodeletTemplate modifier;   

    /* we will establish this demo appliaction on VC_0 */
    connection = ixFpCodeletFastPathConnectionGet  (vc);

    /* Configure the codelet for the the ethernet bridge demo application
     * set the encapsulation type and the rx handler
     */
    connection->encapsulationType = TYPE_1483BVCMUX;
    connection->atmRxHandler      = ixFpCodeletAtmRxEthernetBridge;
    connection->port  = port;
    connection->vpi   = vpi;
    connection->vci   = vci;
    connection->vc    = vc;

    /* Create an ATM VC and configure it for fast path */ 
    if( ixFpCodeletFastPathVcCreate(vc) != IX_SUCCESS )
    {
        IXFPATH_CODELET_ERROR("Failed to establish fast path");
        return IX_FAIL;
    }

    /* Create a base template pair */
    ixFpCodeletTemplatesBaseTemplatePairCreate (
        connection->vc,
        connection->encapsulationType );

    /* Wan Mac learning can optionaly be setup 
     * refer to ixFpCodeletDefines_p.h
     */
    if( WAN_MAC_LEARNING_ENABLED )
    {
        /* Configure Wan Mac Learning
         * This involves registering a Wan Mac callback handler,
         * and enabling Wan Mac Learning in the classifier template 
         */
     
        /* register the callback */
        if( ixFpathAccWanMacAddrCallbackSet ( 
                ixFpCodeletNewWanMacAddressHandler,
                connection->npeVcId,
                (IxFpathAccClientId) connection->vc) != IX_SUCCESS )
        {
           IXFPATH_CODELET_ERROR("failed to register Wan Mac callback");
           return IX_FAIL;
        }
     
        /* enable Wan Mac learning in the template */
        ixFpCodeletTemplatesWanMacLearningSet (
            connection->vc, 
            TRUE,   /* enable Wan Mac learning */
            TRUE);  /* place the command at the start of the classifier 
                     * i.e. the WAN MAC address will be learned before
                     * any other classification is done
                     */
     
    }

    /* Retrieve the new templates */
    ixFpCodeletTemplatesTemplatePairGet (
        connection->vc, 
        &classifier, 
        &modifier);

    /* Flush the cache because this data
     * will be read by the Npes
     */
    IX_ACC_DATA_CACHE_FLUSH(modifier.data,
                            modifier.len);

    /* Download the templates to Npes */
    if( ixFpathAccTemplatePairSet (
            /* N.B. The npeVcId was obtained when making the VC connection
             * on IxAtmdAcc. 
             */
            connection->npeVcId,
            classifier.data,
            classifier.len,
            modifier.data,
            modifier.len ) != IX_SUCCESS )
    {
        IXFPATH_CODELET_ERROR("failed to set template pair");
        return IX_FAIL;
    }

    return IX_SUCCESS;
}

/** 
 * @brief Handle received ATM PDUs for the Ethernet Bridge Application
 *
 * This function illustrates an ATM receive handler for an
 * Ethernet Bridge Application.
 *
 * In the Ethernet bridge application IEEE 802.3 packet ordering must be
 * maintained. To meet this requirenment the user must configure the 
 * system such that under normal operating conditions all PDUs are 
 * accepted for fast path.
 *
 * Only exception packets are expected to be handled by this function. 
 * For the purposes of this demonstration code these are simply, counted and discarded.
 *
 * @param userId - Expected to be the VcNumber of the demo connection.
 * @param mbuf   - pointer to an mbuf containing an Aal5 SDU.
 */
PRIVATE void  
ixFpCodeletAtmRxEthernetBridge(
    IxAtmdAccUserId userId,
    IX_MBUF * buf)
{
    IxFpCodeletFastPathConnection *connection;

    connection = ixFpCodeletFastPathConnectionGet ((VcNumber) userId);
    /* There are three legitimate reasons to receive an exception packet.
     * 1) the PDU failed classification because the payload content 
     *    did not satisfy the classifier template rules.
     * 2) the PDU was longer than maximum Fpath PDU size.
     * 3) the ethernet tx queue was full.
     *
     * By testing the payload length we can distinguish length rejects
     * for statistics purposes
     */
     IX_ASSERT( connection->encapsulationType == TYPE_1483BVCMUX );

     if( IX_MBUF_PKT_LEN(buf) >  FASTPATH_MBUF_SIZE )
     {
        /* the pdu was too long for fast path */
        connection->etherBridgeStats.lengthExceptionCount++;
     }
     else
     {
        /* some other kind of reject */
        connection->etherBridgeStats.otherExceptionCount++;
     }

     /* return the buffer to the free pool */
     ixFpCodeletBufManMbufFree (buf);
}



/** 
 * @brief Handle a new WAN MAC address
 *
 * This function illustrates a WAN MAC callback handler and may be 
 * registered with IxFpathAcc to receive notification of new 
 * WAN source MAC addresses.
 *
 * Each time a new source MAC address is detected, and at least every
 * 500ms in the presence of a constant source MAC address, the IxFpathAcc
 * component will invoke this function.
 *
 * Typically the user will wish to "learn" these new addresses. 
 * This IxFpathAcc demonstration code simply keeps a copy of the last 
 * 16 new MAC addresses.
 */
PRIVATE void
ixFpCodeletNewWanMacAddressHandler(
    IxFpathAccClientId userId,
    unsigned char *macAddress)
{
    unsigned int index = ixFpCodeletWanMacLearningDb.index;

    if( ixFpCodeletWanMacCbTraceEnabled )
    {
        IXFPATH_CODELET_TRACE(( traceBuf,
            "IxFpCodelet"
            "WanMacCb Received "
            "On userId = %8X "
            "macAddr = %2.2X:%2.2X:%2.2X:%2.2X:%2.2X:%2.2X\n",
            userId,
            macAddress[0],
            macAddress[1],
            macAddress[2],
            macAddress[3],
            macAddress[4],
            macAddress[5])); 
    }

    /* save the userId */
    ixFpCodeletWanMacLearningDb.entry[index].vc = userId;
    /* Get timestamp */
    ixFpCodeletWanMacLearningDb.entry[index].timestamp =
        ixOsServTimestampGet();
    /* save the MAC */
    memcpy (&ixFpCodeletWanMacLearningDb.entry[index].macAddress, 
            macAddress,
            IX_IEEE803_MAC_ADDRESS_SIZE);
    /* count this entry */
    ixFpCodeletWanMacLearningDb.wanMacCount++;
    /* advance the index and wrap */
    ixFpCodeletWanMacLearningDb.index++;
    ixFpCodeletWanMacLearningDb.index %= WAN_MAC_ADDR_LEARNING_DB_SIZE;
}

/** 
 * @brief Initialize the WAN MAC learning database
 */
PRIVATE void 
ixFpCodeletWanMacLearningDbInit( void )
{
    ixFpCodeletWanMacLearningDb.index = 0;
    ixFpCodeletWanMacLearningDb.wanMacCount = 0;
}

/** 
 * @brief Display the Wan MAC learning database.
 *
 * Display the last WAN_MAC_LEARNING_DB_DISPLAY_NUM_ENTRIES in 
 * the WAN MAC learning database
 */
PRIVATE void 
ixFpCodeletWanMacLearningDbDisplay( void )
{
    unsigned int timestamp;
    unsigned int index;
    unsigned int count = WAN_MAC_LEARNING_DB_DISPLAY_NUM_ENTRIES;
    int lockKey;

    /* It is not possible to display the new WAN MAC addresses
     * as fast as they are being notified, therefore we take a 
     * snapshot of the buffer and display this at our leisure.
     * Note Int locks around this snapshot ensure the consistency
     * of the copy.
     */

    lockKey = ixOsServIntLock();    

    memcpy( &ixFpCodeletWanMacLearningDbSnapShot,
            &ixFpCodeletWanMacLearningDb,
            sizeof( IxFpCodeletWanMacLearningDb ));

    ixOsServIntUnlock(lockKey);

    /* decide how may entries will be displayed
     * and get the starting index. If there are less
     * than WAN_MAC_LEARNING_DB_DISPLAY_NUM_ENTRIES to display
     * then we just display how ever many there are. 
     */
    if ( ixFpCodeletWanMacLearningDbSnapShot.wanMacCount < 
         WAN_MAC_LEARNING_DB_DISPLAY_NUM_ENTRIES )
    {
        count = ixFpCodeletWanMacLearningDbSnapShot.wanMacCount;
    }     

    index = ixFpCodeletWanMacLearningDbSnapShot.index;

    /* Display a banner */
    IXFPATH_CODELET_TRACE((traceBuf,"--- WAN MAC Learning history ----------------------------\n"));
    IXFPATH_CODELET_TRACE((traceBuf,"Time(us) VC MAC\n"));
                                                 
    while (count--) 
    {   
        /* process the entries in reverse chronological
         * order, starting one behind the current index
         * because it has not yet been written
         */
        index--;
        index %= WAN_MAC_ADDR_LEARNING_DB_SIZE;

        /* convert timestamp to micro seconds */
        timestamp = 
            ( ixFpCodeletWanMacLearningDbSnapShot.entry[index].timestamp
              / IXP425_PERIPHERAL_BUS_CLOCK );

        /* Display the entry: timestamp, VC, and MAC address */
        IXFPATH_CODELET_TRACE((traceBuf,"%8.8u %2.2d %2.2X:%2.2X:%2.2X:%2.2X:%2.2X:%2.2X\n",
            timestamp,
            ixFpCodeletWanMacLearningDbSnapShot.entry[index].vc,
            ixFpCodeletWanMacLearningDbSnapShot.entry[index].macAddress[0],
            ixFpCodeletWanMacLearningDbSnapShot.entry[index].macAddress[1],
            ixFpCodeletWanMacLearningDbSnapShot.entry[index].macAddress[2],
            ixFpCodeletWanMacLearningDbSnapShot.entry[index].macAddress[3],
            ixFpCodeletWanMacLearningDbSnapShot.entry[index].macAddress[4],
            ixFpCodeletWanMacLearningDbSnapShot.entry[index].macAddress[5]));
    }
    IXFPATH_CODELET_TRACE((traceBuf,"\n"));
}

/***
 * @brief Periodic display of Wan MAC address database
 * This function implements a task to demonstrate the WAM MAC
 * address learning function. 
 *
 * The purpose of this task is to periodically display a snapshot of
 * the WAN MAC database
 */
PRIVATE ix_error
ixFpCodeletWanMacDisplayingTask(
    void* arg,
    void **ptrRetObj)
{
    unsigned int count;
    ix_ossl_thread_set_priority ( 
    ixFpCodeletWanMacDisplayingTaskId, 
    IX_OSSL_THREAD_PRI_LOW);

    while(taskEnabled)
    {
        ixFpCodeletWanMacLearningDbDisplay();
        count = ( WAN_MAC_LEARNING_DB_DISPLAY_INTERVAL_S * 
                    (1000/WAN_ADDR_DISPLAY_DELAY_MS) );

        while( count-- )
        {
            ixOsServTaskSleep(WAN_ADDR_DISPLAY_DELAY_MS);        
        }
    }
    return IX_SUCCESS;
}
