/**
 * @file   IxFpathAccCodeletUtils.c
 * @date   June 23 2002
 * @brief  This file contains Atm, Eth, and Npe related utility code 
 *         to be used by the fpathAcc codelet. 
 * 
 * -- Intel Copyright Notice --
 * 
 * @par
 * Copyright 2002-2003 Intel Corporation All Rights Reserved.
 * 
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.
 * 
 * @par
 * The Material is protected by worldwide copyright and trade secret laws
 * and treaty provisions. No part of the Material may be used, copied,
 * reproduced, modified, published, uploaded, posted, transmitted,
 * distributed, or disclosed in any way except in accordance with the
 * applicable license agreement .
 * 
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel, except in accordance with the
 * applicable license agreement.
 * 
 * @par
 * Unless otherwise agreed by Intel in writing, you may not remove or
 * alter this notice or any other notice embedded in Materials by Intel
 * or Intel's suppliers or licensors in any way.
 * 
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * 
 * @par
 * -- End Intel Copyright Notice --
 */
#include <taskLib.h>
#include <logLib.h>
#include <ioLib.h>

#ifdef __vxworks
#include <end.h>    /* END drivers */
#include <endLib.h> /* END drivers */
#endif

#include "ixp425.h"
#include "IxTypes.h"
#include "IxAssert.h"
#include "IxEthAcc.h"
#include "IxAtmTypes.h"
#include "IxAtmdAcc.h"
#include "IxFpathAcc.h"
#include "IxAtmm.h"
#include "IxNpeMh.h"
#include "IxQMgr.h"
#include "IxOsServices.h"
#include "IxNpeDl.h"
#include "IxOsCacheMMU.h"
#include "ix_ossl.h"
#include "IxNpeMicrocode.h"
#include "IxFpathAccCodeletDefines_p.h"

/** 
 * Types
 */
typedef struct
{
    BOOL               inUse;
    IxAtmConnId        connId;
    IxAtmSchedulerVcId vcId;
    IxAtmNpeRxVcId     npeVcId;
    IxAtmLogicalPort   port;
}IxFpCodeletUtilsAtmConnection;

/** 
 * Local variables
 */
static IxQMgrDispatcherFuncPtr dispatcherFunc ;

/** 
 * Atm Connection data base
 */
static IxFpCodeletUtilsAtmConnection 
    ixFpCodeletUtilsAtmRxConnection[IX_ATM_MAX_NUM_VC];

static IxFpCodeletUtilsAtmConnection
    ixFpCodeletUtilsAtmTxConnection[IX_ATM_MAX_NUM_VC];

/* Ethernet */
typedef struct
{
    BOOL traceEnabled;
    unsigned int replenishCount;
    unsigned int txDoneCallbackCount;
    unsigned int rxCallbackCount;
} IxFpCodeletUtilsEthStatistics;

static IxFpCodeletUtilsEthStatistics
ixFpCodeletUtilsEthStatistics = { FALSE,0,0,0 };


/* Atm */
typedef struct
{
    BOOL traceEnabled;
    unsigned int replenishCount;
    unsigned int txDoneCallbackCount;
} IxFpCodeletUtilsAtmStatistics;

static IxFpCodeletUtilsAtmStatistics
ixFpCodeletUtilsAtmStatistics = { FALSE,0,0 };

/* private prototypes */
PRIVATE IX_STATUS ixFpCodeletUtilsQMgrInit(void);
PRIVATE IX_STATUS ixFpCodeletUtilsNpeMhInit(void);
PRIVATE IX_STATUS ixFpCodeletUtilsNpeAInit(void);
PRIVATE IX_STATUS ixFpCodeletUtilsNpeBInit(void);
PRIVATE IX_STATUS ixFpCodeletUtilsNpeCInit(void);
PRIVATE IX_STATUS ixFpCodeletUtilsUtopiaInit(void);
PRIVATE IX_STATUS ixFpCodeletUtilsEthReplenish (IxAtmLogicalPort port);
PRIVATE void ixFpCodeletUtilsEthTxDoneCb (IxAtmLogicalPort port, IX_MBUF *buf);
PRIVATE void ixFpCodeletUtilsEthRxCb (IxAtmLogicalPort port, IX_MBUF *buf, IxEthAccPortId portId);
PRIVATE IX_STATUS ixFpCodeletUtilsEthMiiInit(BOOL speed, BOOL duplex, BOOL autoneg);
PRIVATE IX_STATUS ixFpCodeletUtilsEthAccInit(void);
PRIVATE IxFpCodeletUtilsAtmConnection *ixFpCodeletUtilsAtmRxConnectionGet(unsigned int id);
PRIVATE void ixFpCodeletUtilsConnectionDatabaseInit(void);

/**---
 * Local defines
 */

#define IX_ETH_ACC_MII_MAX_PHY_NUM  0x2

#define COMP_INIT(status, COMPONENT) do { \
    status = COMPONENT(); \
    if (status != IX_SUCCESS) \
    { \
	IXFPATH_CODELET_TRACE((traceBuf, "%s %s", #COMPONENT, "Init Failed\n"));\
	IXFPATH_CODELET_ERROR(("Init Failed\n"));\
    } \
    else \
    { \
	IXFPATH_CODELET_TRACE((traceBuf, "%s %s", #COMPONENT, "Init Succeeded\n"));\
    } \
} while(0)





/**---
 * Vars that define the environment
 */

/* UTOPIA modes */
static BOOL ixFpCodeletUtopiaLoopback   = FALSE;
static IxAtmmPhyMode fpIxFpCodeletPhyMode = IX_ATMM_MPHY_MODE;

/* qMgr mode */
static BOOL ixFpCodeletQMgrUseInterrupts = FALSE;


/**---
 * Function Implementations
 */


/**
 * QMgr dispatcher task.
 * Using this makes it possible to put debug break to callbacks
 */
static ix_ossl_thread_t tId;
static ix_error
ixFpCodeletUtilsQMgrDispatchTask ( void* arg, 
               void** ptrRetObj)
{
    ix_ossl_thread_set_priority( tId, IX_OSSL_THREAD_PRI_LOW);

    while (TRUE)
    {
	(*dispatcherFunc)(IX_QMGR_QUELOW_GROUP);
	(*dispatcherFunc)(IX_QMGR_QUEUPP_GROUP);

        /* Share the CPU and allow other threads to run */
	ixOsServTaskSleep(0);
    }
}


/**
 * Initialize QMgr
 * Default behaviour is to start qMgr with a dispatch task
 * as debuging callbacks is easier this way, since you
 * cant break in ISRs,
 * To use interrupts set the variable ixFpCodelet_qMgrUseInterrupts to true;
 */
PRIVATE IX_STATUS
ixFpCodeletUtilsQMgrInit(void)
{
    IX_STATUS status = IX_SUCCESS;

    /* Initialize QMGR */
    if( ixQMgrInit() != IX_SUCCESS )
    {
        return IX_FAIL;
    }

    ixQMgrDispatcherLoopGet(&dispatcherFunc);

    if( ixFpCodeletQMgrUseInterrupts )
    {
        /* Hook the QM QUELOW dispatcher to interrupt */
        status |= ixOsServIntBind (IXP425_INT_LVL_QM1,
                                   (IxVoidFnVoidPtr)dispatcherFunc,
                                   (void *) IX_QMGR_QUELOW_GROUP);
        /* Hook the QM QUEUPP dispatcher to interrupt */
        status |= ixOsServIntBind ( IXP425_INT_LVL_QM2,
                                   (IxVoidFnVoidPtr)dispatcherFunc,
                                   (void *) IX_QMGR_QUEUPP_GROUP);
        return status;
    }


    if ( ix_ossl_thread_create (
             ixFpCodeletUtilsQMgrDispatchTask,
             NULL,
             &tId) != IX_OSSL_ERROR_SUCCESS )
    {
        return IX_FAIL;
    } 

    return IX_SUCCESS;
}


/**
 * Initialise Npe Message Handler
 */
PRIVATE IX_STATUS
ixFpCodeletUtilsNpeMhInit(void)
{
    return ixNpeMhInitialize (IX_NPEMH_NPEINTERRUPTS_YES);
}


/**
 * Initialise NpeA
 */
PRIVATE IX_STATUS
ixFpCodeletUtilsNpeAInit(void)
{
    return ixNpeDlNpeInitAndStart (NPEA_IMAGEID);
}


/**
 * Initialise NpeB
 */
PRIVATE IX_STATUS
ixFpCodeletUtilsNpeBInit(void)
{
    return ixNpeDlNpeInitAndStart (NPEB_IMAGEID);
}

/**
 * Initialise NpeC
 */
PRIVATE IX_STATUS
ixFpCodeletUtilsNpeCInit(void)
{
    return ixNpeDlNpeInitAndStart (NPEC_IMAGEID);
}



/**
 * Initialise the utopia ports
 * The default behaviour ( ixFpCodeletUtopiaLoopback = FALSE ) is no
 * loopback, all ports are initialized.
 *
 * If ixFpCodeletUtopiaLoopback is set to TRUE a single port ( port 0 )
 * is initiallized in loopback mode.
 *
 */
PRIVATE IX_STATUS
ixFpCodeletUtilsUtopiaInit( void )
{
    IxAtmmPortCfg portCfgs[IX_UTOPIA_MAX_PORTS];
    IxAtmmUtopiaLoopbackMode loopbackMode;
    IxAtmLogicalPort port    = IX_UTOPIA_PORT_0;
    IxAtmLogicalPort maxPort;
    IX_STATUS status = IX_SUCCESS;

    /* Clear portCfgs */
    memset(portCfgs, 0, sizeof(portCfgs));

    /* setup portCfgs for one port */
    if (ixFpCodeletUtopiaLoopback)
    {
        /* loopback mode supports only 1 porti, and the PHY should be 0  */
        printf("\nUtopia port initialised in loopback Mode\n");
        maxPort                        = 1;
	loopbackMode                   = IX_ATMM_UTOPIA_LOOPBACK_ENABLED;
	portCfgs[port].UtopiaTxPhyAddr = IX_UTOPIA_PORT_0;
	portCfgs[port].UtopiaRxPhyAddr = IX_UTOPIA_PORT_0;
    }
    else
    {
/* Set this during compilation to true when require to operate with the
 * UTOPIA as clock master. Needed for example when connecting with Adtech 
 * AX4000 ATM test equipment.Note this Applies only to Validation boards.
 */
#ifdef VALIDATION_PLATFORM_USED
	ixFpCodeletUtopiaFpgaStimulusAsMasterSet();
#endif

        maxPort      = IX_UTOPIA_MAX_PORTS;
	loopbackMode = IX_ATMM_UTOPIA_LOOPBACK_DISABLED;
        
        for (port = IX_UTOPIA_PORT_0; port < maxPort; port++ ) 
        {
	    portCfgs[port].UtopiaTxPhyAddr = MPHY_UTOPIA_PHY_ADDR+port;
	    portCfgs[port].UtopiaRxPhyAddr = MPHY_UTOPIA_PHY_ADDR+port;
        }
    }

    if( ixAtmmUtopiaInit (
                 maxPort,
                 fpIxFpCodeletPhyMode,
                 portCfgs,
                 loopbackMode) != IX_SUCCESS )
    {
        return IX_FAIL;
    }

    for (port = IX_UTOPIA_PORT_0;
         (port < maxPort)&&(status == IX_SUCCESS); port++ ) 
    {
        
        status |= ixAtmmPortInitialize (
                      port, 
                      UTOPIA_PORT_RATE, 
                      UTOPIA_PORT_RATE);
        
        status |= ixAtmmPortEnable(port);
    }
    return status;
}

/**
 * Replenish the ethernet rx free queue
 */
PRIVATE IX_STATUS
ixFpCodeletUtilsEthReplenish (IxAtmLogicalPort port)
{
    IX_MBUF *buf            = NULL;
    IX_STATUS retval        = IX_SUCCESS;
    unsigned int numEntries = ETH_INITIAL_RX_REPLENISH_NUM;

    while ((retval == IX_SUCCESS) && (numEntries >0))
    {
        ixFpCodeletUtilsEthStatistics.replenishCount++;

        buf = ixFpCodeletBufManMbufGet (ETH_RX_MBUF_SIZE);

        if (buf != NULL)
        {
	    retval = ixEthAccPortRxFreeReplenish (port, buf);
            numEntries -= 1;
        }
        else
        {
            retval = IX_FAIL;
        }
    }
    return retval;
}

/**
 * Ethernet transmit done callback
 */
PRIVATE void
ixFpCodeletUtilsEthTxDoneCb (IxAtmLogicalPort port, IX_MBUF *buf)
{
    ixFpCodeletUtilsEthStatistics.txDoneCallbackCount++;

    if( ixFpCodeletUtilsEthStatistics.traceEnabled )
    {
        IXFPATH_CODELET_TRACE ((
            traceBuf,
            "ixFpCodelet ixFpCodeletEthTxDoneCb()"));
    }
    ixFpCodeletBufManMbufFree (buf);
}



/**
 * Ethernet receive callback
 */
PRIVATE void
ixFpCodeletUtilsEthRxCb (IxAtmLogicalPort port, IX_MBUF *buf, IxEthAccPortId portId)
{
    ixFpCodeletUtilsEthStatistics.rxCallbackCount++;

    if( ixFpCodeletUtilsEthStatistics.traceEnabled )
    {
        
        IXFPATH_CODELET_TRACE((
            traceBuf,
            "Received a Eth Frame mbuf = 0x%8.8X"
            "length %d,"
            "port %x,"
            "mac address %2.2X%2.2X%2.2X%2.2X%2.2X%2.2X\n",
            (int) buf,
            IX_MBUF_PKT_LEN(buf), 
            port,
            IX_MBUF_MDATA(buf)[0],
	    IX_MBUF_MDATA(buf)[1],
	    IX_MBUF_MDATA(buf)[2],
	    IX_MBUF_MDATA(buf)[3],
	    IX_MBUF_MDATA(buf)[4],
	    IX_MBUF_MDATA(buf)[5] ));
    }

    /* In this IxFpathAcc codelet we simply
     * discard received ethernet PDUs
     */
    ixFpCodeletBufManMbufFree (buf); 
}


/**
 *  Initialise the MII phy
 */
PRIVATE IX_STATUS
ixFpCodeletUtilsEthMiiInit(BOOL speed, BOOL duplex, BOOL autoneg)
{
    BOOL getLink, getSpeed, getDuplex,getAutoneg;
    UINT32 phyAddresses[IX_ETH_ACC_MII_MAX_PHY_NUM];
    BOOL phyPresent[IXP425_ETH_ACC_MII_MAX_ADDR];
    UINT32 i;
    UINT32 j;

    printf("Scanning for Ethernet PHYs\n");

    if(ixEthAccMiiPhyScan (phyPresent) == ERROR)
    {	
	printf("ixEthAccMiiPhyScan Failed\n");   
        return IX_FAIL;
    }   
    else
    {
	printf("phy init Succeeded\n");
	printf("PHYs discovered at addresses: ");
	j=0;
	for(i=0;i<IX_ETH_ACC_MII_MAX_PHY_NUM;i++)
	{
	    if(phyPresent[i] == TRUE)
	    {
		printf("%d ",i);
		phyAddresses[j]=i;
		j++;
	    }
	}
	printf("\n");
    }
    
    /*reset each phy*/
    for(i=0;i<IX_ETH_ACC_MII_MAX_PHY_NUM; i++)
    {
	printf("Resetting PHY %d address %d\n", i,phyAddresses[i]);
	ixEthAccMiiPhyReset (phyAddresses[i]);
	taskDelay (200);
    }

    /*Set each phy's config*/

    for(i=0;i<IX_ETH_ACC_MII_MAX_PHY_NUM; i++)
    {
        unsigned int retry = 20;
	printf("Configuring PHY %d address:%d\n", i,phyAddresses[i]);
	ixEthAccMiiPhyConfig (phyAddresses[i],
			      speed,
			      duplex, 
			      autoneg);
	do
        {
           ixOsServTaskSleep(100);  /* 100 milliseconds */
            ixEthAccMiiLinkStatus (
            phyAddresses[i],
            &getLink,
            &getDuplex,
            &getSpeed,
            &getAutoneg);
        }
        while (retry-- && getLink == FALSE);
	ixEthAccMiiShow (phyAddresses[i]);	
    }

    
    for(i=IX_ETH_PORT_1;i<=IX_ETH_PORT_2;i++)
    {
	ixEthAccMiiLinkStatus (
            phyAddresses[i], 
            &getLink, 
            &getDuplex, 
            &getSpeed, 
            &getAutoneg);

	if(getDuplex)
	{
	    ixEthAccPortDuplexModeSet (i, IX_ETH_ACC_FULL_DUPLEX);
	}
	else
	{
	    ixEthAccPortDuplexModeSet (i, IX_ETH_ACC_HALF_DUPLEX);
        }
    }
    
    printf("MII PHY configuration complete\n");
    return IX_SUCCESS;
}

/**
 * Initialise EthAcc and its envronment
 * Tx FCS can be disabled by setting the var ixFpCodelet_ethTxFrameFCSDisable 
 * to true.
 */


IxEthAccMacAddr macAddr1 = { ETH1_MAC_ADDRESS };
IxEthAccMacAddr macAddr2 = { ETH2_MAC_ADDRESS };



PRIVATE IX_STATUS
ixFpCodeletUtilsEthAccInit(void)
{
    /*Initialise Ethernet Access component*/
    if((ixEthAccInit ()!= IX_ETH_ACC_SUCCESS) ||
       (ixEthAccPortInit (IX_ETH_PORT_1) !=IX_ETH_ACC_SUCCESS)||
       (ixEthAccPortInit (IX_ETH_PORT_2) !=IX_ETH_ACC_SUCCESS)||

       (ixFpCodeletUtilsEthMiiInit (FALSE,FALSE,TRUE) != IX_SUCCESS)||

       /* initPorts();*/
       (ixEthAccPortUnicastMacAddressSet (IX_ETH_PORT_1, &macAddr1) !=IX_ETH_ACC_SUCCESS)||
       (ixEthAccPortUnicastMacAddressSet (IX_ETH_PORT_2, &macAddr2) !=IX_ETH_ACC_SUCCESS)||

       (ixEthAccPortTxDoneCallbackRegister (IX_ETH_PORT_1, ixFpCodeletUtilsEthTxDoneCb, IX_ETH_PORT_1) !=IX_ETH_ACC_SUCCESS)||
       (ixEthAccPortTxDoneCallbackRegister (IX_ETH_PORT_2, ixFpCodeletUtilsEthTxDoneCb, IX_ETH_PORT_2) !=IX_ETH_ACC_SUCCESS)||

       (ixEthAccTxSchedulingDisciplineSet (IX_ETH_PORT_1,FIFO_NO_PRIORITY)!=IX_ETH_ACC_SUCCESS)||
       (ixEthAccTxSchedulingDisciplineSet (IX_ETH_PORT_2,FIFO_NO_PRIORITY)!=IX_ETH_ACC_SUCCESS)||

       (ixEthAccPortRxCallbackRegister (IX_ETH_PORT_1, ixFpCodeletUtilsEthRxCb, IX_ETH_PORT_1 )!=IX_ETH_ACC_SUCCESS)||
       (ixEthAccPortRxCallbackRegister (IX_ETH_PORT_2, ixFpCodeletUtilsEthRxCb, IX_ETH_PORT_2 )!=IX_ETH_ACC_SUCCESS)||

       (ixFpCodeletUtilsEthReplenish (IX_ETH_PORT_1) != IX_SUCCESS)||
       (ixFpCodeletUtilsEthReplenish (IX_ETH_PORT_2) != IX_SUCCESS)||

       (ixEthAccPortEnable (IX_ETH_PORT_1) != IX_ETH_ACC_SUCCESS)||
       (ixEthAccPortEnable (IX_ETH_PORT_2) != IX_ETH_ACC_SUCCESS)
       )
    {
        return IX_FAIL;
    }
    return IX_SUCCESS;
}

/**
 * Return a pointer to the Rx connection data for
 * the specified connection number
 */
PRIVATE IxFpCodeletUtilsAtmConnection *
ixFpCodeletUtilsAtmRxConnectionGet( unsigned int id )
{
    return &ixFpCodeletUtilsAtmRxConnection[id];
}

/**
 * Replenish an Atm Rx Free queue
 */
void 
ixFpCodeletUtilsAtmReplenish (IxAtmdAccUserId userId)
{
    unsigned int numEntries;

    IX_MBUF *buf = NULL;
    IX_STATUS retval = IX_SUCCESS;
    IxAtmConnId connId = ixFpCodeletUtilsAtmRxConnectionGet (userId)->connId;

    retval = ixAtmdAccRxVcFreeEntriesQuery (connId, &numEntries);

    while ((retval == IX_SUCCESS) && (numEntries >0))
    {
        ixFpCodeletUtilsAtmStatistics.replenishCount++;

        buf = ixFpCodeletBufManMbufGet (ATM_RX_MBUF_SIZE);

        if (buf != NULL)
        {
            retval = ixAtmdAccRxVcFreeReplenish (connId, buf);
            numEntries -= 1;
        }
        else
        {
            retval = IX_FAIL;
        }
    }

    if (retval != IX_SUCCESS)
    {
         IXFPATH_CODELET_ERROR("Failed to Replenish Rx Free Q");
    }
}


/**
 * Atm Rx connect and register all callbacks
 */
IX_STATUS
ixFpCodeletUtilsAtmRxVcRegisterConnect(IxAtmLogicalPort port,
		    unsigned rxvpi,
		    unsigned rxvci,
		    IxAtmRxQueueId rxQueueId,
		    IxAtmdAccRxVcRxCallback rxCallback,
		    int userId,
                    IxAtmNpeRxVcId *npeVcId )
{
    IxAtmmVc rxVc;
    unsigned int rxFreeQueueSize;
    IxFpCodeletUtilsAtmConnection *connection; 

    IX_ASSERT((userId >= 0) && (userId < IX_ATM_MAX_NUM_VC));

    connection = ixFpCodeletUtilsAtmRxConnectionGet (userId);

    if (connection->inUse == TRUE)
    {
        IXFPATH_CODELET_ERROR("Rx VC already registered");
        return IX_FAIL;
    }

    connection->port = port;

    memset(&rxVc, 0, sizeof(rxVc));

    /* Setup Rx Vc descriptor */
    rxVc.vpi = rxvpi;
    rxVc.vci = rxvci;
    rxVc.direction = IX_ATMM_VC_DIRECTION_RX;
    rxVc.trafficDesc.atmService = IX_ATM_UBR;

    /* Setup rx VC */
    if( ixAtmmVcRegister (
            port,
            &rxVc,
	    &connection->vcId) != IX_SUCCESS)
    {
	IXFPATH_CODELET_ERROR("Failed to register Rx VC");
	return IX_FAIL;
    }

    /* Connect Rx to the VC */
    if( ixAtmdAccRxVcConnect (
            port,
	    rxvpi,
	    rxvci,
            IX_ATMDACC_AAL5,
	    rxQueueId,
            userId,
	    rxCallback,
	    IX_ATMDACC_DEFAULT_REPLENISH_COUNT,
	    &connection->connId,
            &connection->npeVcId ) != IX_SUCCESS)
    {
	IXFPATH_CODELET_ERROR("Failed to connect to Rx VC");
	return IX_FAIL;
    }

    /* return the NpeVcId to the client */
    *npeVcId = connection->npeVcId; 

    if( ixAtmdAccRxVcFreeEntriesQuery(
            connection->connId,
            &rxFreeQueueSize) != IX_SUCCESS)
    {
	IXFPATH_CODELET_ERROR("Failed to get nof entries in rx free queue");
	return IX_FAIL;
    }

    /* Replenish Rx buffers */
    ixFpCodeletUtilsAtmReplenish(userId);

    if( ixAtmdAccRxVcFreeLowCallbackRegister (
            connection->connId,
            0,
            ixFpCodeletUtilsAtmReplenish) != IX_SUCCESS)
    {
	IXFPATH_CODELET_ERROR("Failed to register RxFreeLowCalback");
	return IX_FAIL;
    }

    /* Enable receive */
    if ( ixAtmdAccRxVcEnable (connection->connId) != IX_SUCCESS)
    {
	IXFPATH_CODELET_ERROR("Failed to enable Rx VC");
	return IX_FAIL;
    }

    connection->inUse = TRUE;
    return IX_SUCCESS;
}


/** 
 * Atm Rx Disconnect
 */
#define ATM_DISCONNECT_ATTEMPTS 2
#define ATM_DISCONNECT_DELAY_MS 100
IX_STATUS
ixFpCodeletUtilsAtmRxDisconnect( IxAtmdAccUserId userId )
{
    IxAtmConnId connId;
    unsigned int disconnectAttempts;
    IX_STATUS status;

    connId = ixFpCodeletUtilsAtmRxConnectionGet( userId )->connId;

    status = IX_ATMDACC_RESOURCES_STILL_ALLOCATED;
    disconnectAttempts = ATM_DISCONNECT_ATTEMPTS;

    status = ixAtmdAccRxVcDisable (connId);
    if( status == IX_SUCCESS )
    {
        while( disconnectAttempts--)
        {
            status = ixAtmdAccRxVcTryDisconnect (connId);
            if( status ==  IX_SUCCESS )
            {
                break;
            }
            /* delay 16ms and try again */
            ixOsServTaskSleep(ATM_DISCONNECT_DELAY_MS);
        }
    }
    return status;
}


/**
*/
PRIVATE void
ixFpCodeletUtilsConnectionDatabaseInit( void )
{
    int connNum;

    memset( ixFpCodeletUtilsAtmRxConnection, 0, 
            sizeof(ixFpCodeletUtilsAtmRxConnection));

    memset( ixFpCodeletUtilsAtmTxConnection, 0, 
            sizeof(ixFpCodeletUtilsAtmTxConnection));

    for( connNum = 0; connNum < IX_ATM_MAX_NUM_VC; connNum++ )
    {
        ixFpCodeletUtilsAtmRxConnection[connNum].inUse   = FALSE;
        ixFpCodeletUtilsAtmTxConnection[connNum].inUse   = FALSE;
    }
}



/**
*/
IX_STATUS 
ixFpCodeletUtilsInit( void )
{
    IX_STATUS status = IX_SUCCESS;

    /* */
    logInit (ioTaskStdGet(0,1), 1);
    logFdSet (ioTaskStdGet(0,1));
    ixOsServLogLevelSet (LOG_ALL);
    taskPrioritySet(taskIdSelf(), 100);

#ifdef __vxworks
    /* When the ixe drivers are running, the codelets
    * cannot run.
    */
    if (endFindByName ("ixe", 0) != NULL)
    {
      printf("FAIL : Driver ixe0 detected\n");
      return IX_FAIL;
    }
    if (endFindByName ("ixe", 1) != NULL)
    {
      printf("FAIL : Driver ixe1 detected\n");
      return IX_FAIL;
    }
#endif

    ixFpCodeletUtilsConnectionDatabaseInit();

    COMP_INIT(status, ixFpCodeletUtilsQMgrInit);    
    COMP_INIT(status, ixFpCodeletUtilsNpeMhInit);
    COMP_INIT(status, ixAtmSchInit);
    COMP_INIT(status, ixAtmdAccInit);
    COMP_INIT(status, ixFpCodeletUtilsNpeAInit);
    COMP_INIT(status, ixFpCodeletUtilsNpeBInit);
    COMP_INIT(status, ixFpCodeletUtilsNpeCInit);
    COMP_INIT(status, ixAtmmInit);
    COMP_INIT(status, ixFpCodeletUtilsUtopiaInit);
    COMP_INIT(status, ixFpCodeletUtilsEthAccInit);


    IXFPATH_CODELET_TRACE(( traceBuf, "Initialization Phase Complete\n"));
    return status;
}




