/**
 * @file IxCryptoAcc.c
 *
 * @date October-03-2002
 *
 * @brief  Source file for Public API (Component Interface module)
 *
 *
 * Design Notes:
 *
 * -- Intel Copyright Notice --
 * 
 * @par
 * Copyright 2002-2003 Intel Corporation All Rights Reserved.
 * 
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.
 * 
 * @par
 * The Material is protected by worldwide copyright and trade secret laws
 * and treaty provisions. No part of the Material may be used, copied,
 * reproduced, modified, published, uploaded, posted, transmitted,
 * distributed, or disclosed in any way except in accordance with the
 * applicable license agreement .
 * 
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel, except in accordance with the
 * applicable license agreement.
 * 
 * @par
 * Unless otherwise agreed by Intel in writing, you may not remove or
 * alter this notice or any other notice embedded in Materials by Intel
 * or Intel's suppliers or licensors in any way.
 * 
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * 
 * @par
 * -- End Intel Copyright Notice --
 */


/*
* Put the user defined include files required.
*/
#include "IxOsServices.h"
#include "IxCryptoAcc.h"
#include "IxFeatureCtrl.h"
#include "IxCryptoAcc_p.h"
#include "IxCryptoAccCCDMgmt_p.h"
#include "IxCryptoAccDescMgmt_p.h"
#include "IxCryptoAccQAccess_p.h"
#include "IxCryptoAccCryptoRegister_p.h"
#include "IxCryptoAccCryptoPerform_p.h"
#include "IxCryptoAccUtilities_p.h"


#define IX_CRYPTO_ACC_DELAY_IN_MS   50    /**< Time taken in task delay */
 
#define IX_CRYPTO_ACC_TIMER_COUNT   10    /**< Timer counter to prevent
                                            * deadlock
                                            */

#define IX_CRYPTO_ACC_HASH_FIRST_PAD_BYTE  0x80 /**< Padding value to be  
                                                 * used in hash data block 
                                                 * padding (first byte of
                                                 * padded length only)
                                                 */

#define IX_CRYPTO_ACC_HASH_FIRST_PAD_BYTE_LEN 1 /**< Length of first pad 
                                                 * byte in byte
                                                 */
                                                 
#define IX_CRYPTO_ACC_HASH_TOTAL_BIT_HIGH_LEN 4 /**< Length of field for
                                                 * total bit (high 32-bit) 
                                                 * for hash data block
                                                 */
                                                 
#define IX_CRYPTO_ACC_HASH_TOTAL_BIT_LOW_LEN 4 /**< Length of field for
                                                 * total bit (low 32-bit) 
                                                 * for hash data block
                                                 */

#define IX_CRYPTO_ACC_HASH_TOTAL_BIT_LEN \
            (IX_CRYPTO_ACC_HASH_TOTAL_BIT_HIGH_LEN \
             + IX_CRYPTO_ACC_HASH_TOTAL_BIT_LOW_LEN) /**< Length of field for
                                                      * total bit of hash data
                                                      * block 
                                                      */                                                      

#define IX_CRYPTO_ACC_HASH_DATA_REMAINDER \
            (IX_CRYPTO_AUTHENTICATION_DATA_BLOCK_LENGTH \
             - IX_CRYPTO_ACC_HASH_FIRST_PAD_BYTE_LEN \
             - IX_CRYPTO_ACC_HASH_TOTAL_BIT_LEN)  /**< Remainder value for 
                                                   * hash data block, where
                                                   * zero padding is not 
                                                   * needed
                                                   */

/*
* Variable declarations global to this file only.  Externs are followed by
* static variables.
*/
static BOOL ixCryptoAccServiceInit = FALSE; /**< TRUE if component 
                                             * initialized 
                                             */
static BOOL ixCryptoAccServiceStop = FALSE; /**< TRUE if crypto service 
                                             * is halt 
                                             */


/**
 * @def IX_CRYPTO_IS_SERVICE_INITIALIZED
 * @brief Check if component has been initialized
 *
 * return none
 *
 */
#define IX_CRYPTO_IS_SERVICE_INITIALIZED() (TRUE == ixCryptoAccServiceInit)



/**
 * @def IX_CRYPTO_IS_SERVICE_STOPPED
 * @brief Check if component service has been stopped
 *
 * return none
 *
 */
#define IX_CRYPTO_IS_SERVICE_STOPPED() (TRUE == ixCryptoAccServiceStop)



/**
 * @fn ixCryptoAccInit
 * @brief Initialise the Security Access component
 *
 */
PUBLIC IxCryptoAccStatus
ixCryptoAccInit (void)
{    
    if (IX_FEATURE_CTRL_SILICON_TYPE_B0 == 
	(ixFeatureCtrlProductIdRead() & IX_FEATURE_CTRL_SILICON_STEPPING_MASK))
    {
        if (IX_FEATURE_CTRL_COMPONENT_DISABLED == 
	        ixFeatureCtrlComponentCheck(IX_FEATURECTRL_DES))
        {
	    ixOsServLog(LOG_WARNING,"Warning: the DES coprocessor you"
	        " specified does not exist\n",0,0,0,0,0,0);
        }

	if ( IX_FEATURE_CTRL_COMPONENT_DISABLED ==
	         ixFeatureCtrlComponentCheck(IX_FEATURECTRL_HASH))
        {
	    ixOsServLog(LOG_WARNING,"Warning: the HASH coprocessor you"
	        " specified does not exist\n",0,0,0,0,0,0);
        }

        if (IX_FEATURE_CTRL_COMPONENT_DISABLED ==
	        ixFeatureCtrlComponentCheck(IX_FEATURECTRL_AES))
        {
	    ixOsServLog(LOG_WARNING,"Warning: the AES coprocessor you"
                " specified does not exist\n",0,0,0,0,0,0);
        }
    }
     
    /* Check if the access component has been intialized, if not, 
     * initialize Descriptor Mgmt, CCD Mgmt, Statistics and QAccess 
     * modules.
     */
    if (!IX_CRYPTO_IS_SERVICE_INITIALIZED ()
        && !IX_CRYPTO_IS_SERVICE_STOPPED ())
    {       
        /* Initializing Descriptor Management Module */
        if (IX_CRYPTO_ACC_STATUS_SUCCESS != ixCryptoDescMgmtInit ())
        {
            /* Log error message in debugging mode */
            IX_CRYPTO_ACC_LOG (LOG_ERROR,
                "ixCryptoDescMgmtInit FAILED.\n", 
                0, 0, 0, 0, 0, 0);
             
            return IX_CRYPTO_ACC_STATUS_FAIL;
        } /* end of if (ixCryptoDescMgmtInit) */
            
        /* Initializing Crypto Context Database Management Module */
        if (IX_CRYPTO_ACC_STATUS_SUCCESS != ixCryptoCCDMgmtInit ())
        {
            /* Log error message in debugging mode */
            IX_CRYPTO_ACC_LOG (LOG_ERROR,
                "ixCryptoCCDMgmtInit FAILED.\n", 
                0, 0, 0, 0, 0, 0);
                
            return IX_CRYPTO_ACC_STATUS_FAIL;
        } /* end of if (ixCryptoCCDMgmtInit) */
            
        /* Initializing Queue Access Module */
        if (IX_CRYPTO_ACC_STATUS_SUCCESS != ixCryptoQAccessInit ())
        {
            /* Log error message in debugging mode */
            IX_CRYPTO_ACC_LOG (LOG_ERROR,
                "ixCryptoQAccessInit FAILED.\n", 
                0, 0, 0, 0, 0, 0);
                
            return IX_CRYPTO_ACC_STATUS_FAIL;
        } /* end of if (ixCryptoQAccessInit) */
        
        /* The initialisation is complete hence set the flag to TRUE */
        ixCryptoAccServiceInit = TRUE;
        
        /* Initialization successfully */
        return IX_CRYPTO_ACC_STATUS_SUCCESS;
    }
    else /* end of if (IX_CRYPTO_IS_SERVICE_INITIALIZED) &&
            (IX_CRYPTO_IS_SERVICE_STOPPED) */
    {
        /* Log error message in debugging mode */
        IX_CRYPTO_ACC_LOG (LOG_ERROR,
            "IxCrypto is initialized or stopped before this \
            function call.\n", 
            0, 0, 0, 0, 0, 0);

        return IX_CRYPTO_ACC_STATUS_FAIL;
    } /* end of if-else (IX_CRYPTO_IS_SERVICE_INITIALIZED) &&
     (IX_CRYPTO_IS_SERVICE_STOPPED) */       
} /* end of ixCryptoAccInit () function */

 
/**
 * @fn ixCryptoAccCtxRegister
 *
 * @brief Crypto context registration. Cryptographic Context ID (cryptoCtxId)
 *        for the registered crypto context obtained from this registration
 *        request will be used in ixCryptoAccAuthCryptPerform requests.
 *
 */
PUBLIC IxCryptoAccStatus 
ixCryptoAccCtxRegister (
    IxCryptoAccCtx *pAccCtx,
    IX_MBUF *pMbufPrimaryChainVar,
    IX_MBUF *pMbufSecondaryChainVar,
    IxCryptoAccRegisterCompleteCallback registerCallbackFn,
    IxCryptoAccPerformCompleteCallback performCallbackFn,
    UINT32 *pCryptoCtxId)
{
    IxCryptoAccStatus status = IX_CRYPTO_ACC_STATUS_SUCCESS;
    UINT32 npeCryptoParamIndex = 0;
    BOOL callbackFlag = FALSE;

    /* Check if the client context and CryptoCtxId pointer are NULL, 
     * return error if either one is NULL 
     */
    if ((NULL == pAccCtx) || (NULL == pCryptoCtxId))
    {
        /* Log error message in debugging mode */
        IX_CRYPTO_ACC_LOG (
            LOG_MESSAGE, 
            "Client's Ctx is %x\n pCryptoCtxId is %x\n", 
            (int) pAccCtx, 
            (int) pCryptoCtxId,
            0, 0, 0, 0);

        return IX_CRYPTO_ACC_STATUS_FAIL;
    } /* end of if (pAccCtx) || (pCryptoCtxId) */
    
    /* Check if the callback functions provided by client is NULL, return
     * error if either one is NULL
     */
    if ((NULL == registerCallbackFn) || (NULL == performCallbackFn))
    {
        /* Log error message in debugging mode */
        IX_CRYPTO_ACC_LOG (
            LOG_MESSAGE, 
            "Register Callback: %x\nPerform Callback: %x\n", 
            (int) registerCallbackFn,
            (int) performCallbackFn, 
            0, 0, 0, 0);

        return IX_CRYPTO_ACC_STATUS_FAIL;
    } /* end of if (registerCallbackFn) || (performCallbackFn) */

    /* Check if the access component has been intialized and the crypto 
     * service is running 
     */
    if (IX_CRYPTO_IS_SERVICE_INITIALIZED() 
        && !IX_CRYPTO_IS_SERVICE_STOPPED())
    {
        /* Get an empty Crypto Context from CCD */
        status = ixCryptoCCDMgmtCryptoCtxGet (pCryptoCtxId);
    
        if (IX_CRYPTO_ACC_STATUS_SUCCESS == status)
        {
            /* Store callback function pointers into Crypto Context */
            ixCryptoCtx[*pCryptoCtxId].registerCallbackFn 
                = registerCallbackFn;
            ixCryptoCtx[*pCryptoCtxId].performCallbackFn
                = performCallbackFn;    
        }
        else /* No Crypto Context is allocated */
        {
            return status;
        } /* end of if-else (status) */

        /* Switch case for Operation registered by client */
        switch (pAccCtx->operation)
        {
            /* Encryption operation, notes: the code for the case of 
             * IX_CRYPTO_ACC_OP_ENCRYPT fall through to the case 
             * IX_CRYPTO_ACC_OP_DECRYPT
             */
            case IX_CRYPTO_ACC_OP_ENCRYPT:
            
            /* Deceryption operation */
            case IX_CRYPTO_ACC_OP_DECRYPT:
                /* Register cipher algorithm */
                status = ixCryptoRegisterCipherAlgoRegister (
                             *pCryptoCtxId,
                             pAccCtx, 
                             &npeCryptoParamIndex,
                             IX_CRYPTO_OP_REGISTER);
                             
                /* Check if cipher context registration success */
                if (IX_CRYPTO_ACC_STATUS_SUCCESS == status)
                {
                    /* Assign NPE operation to CRYPT for the Crypto Context */
                    ixCryptoCtx[*pCryptoCtxId].npeOperationMode.npeOperation
                        = IX_CRYPTO_NPE_OP_CRYPT;
    
                    /* Set direction fo cryption operation, default is decrypt,
                     * OR the NpeOperation with the mask will set the crypt
                     * direction bit to 1 and turn it to be encrypt operation
                     */
                    if (IX_CRYPTO_ACC_OP_ENCRYPT == pAccCtx->operation)
                    {
                        ixCryptoCtx[*pCryptoCtxId].npeOperationMode.npeOperation
                            = (((ixCryptoCtx[*pCryptoCtxId].npeOperationMode.
                                npeOperation)
                                | IX_CRYPTO_NPE_OP_CRYPT_DIR_MASK) 
                                & IX_CRYPTO_NPE_OP_MASK);
                    } /* end of if (pAccCtx->operation) */
                    
                    /* Set callbackFlag to TRUE to trigger client's 
                     * callback for the case of DES/3DES encryption and 
                     * decryption operation only
                     */                 
                    if (IX_CRYPTO_ACC_CIPHER_AES 
                        != pAccCtx->cipherCtx.cipherAlgo)
                    {
                        /* Callback is called only if DES or 3DES, callback 
                         * for AES will be called in QMgr callback context
                         */
                        callbackFlag = TRUE;
                    } /* end of if (pAccCtx->cipherCtx.cipherAlgo) */
                } /* end of if (status) */
                break;
    
            /* Authentication Computation operation, notes: the code for the  
             * case of IX_CRYPTO_ACC_OP_AUTH_CALC fall through to the case 
             * IX_CRYPTO_ACC_OP_AUTH_CHECK 
             */
            case IX_CRYPTO_ACC_OP_AUTH_CALC:
            
            /* Authentication Verification operation */
            case IX_CRYPTO_ACC_OP_AUTH_CHECK: 
                /* Register authentication algorithm */
                status = ixCryptoRegisterAuthAlgoRegister (
                             *pCryptoCtxId,
                             pAccCtx, 
                             pMbufPrimaryChainVar,
                             pMbufSecondaryChainVar, 
                             &npeCryptoParamIndex,
                             IX_CRYPTO_OP_REGISTER);
    
                /* Check if authentication context registration
                 * successful 
                 */
                if (IX_CRYPTO_ACC_STATUS_SUCCESS == status)
                {
                    /* Assign NPE operation */
                    if (IX_CRYPTO_ACC_OP_AUTH_CALC == pAccCtx->operation)
                    {
                        ixCryptoCtx[*pCryptoCtxId].npeOperationMode.npeOperation
                            = IX_CRYPTO_NPE_OP_HMAC_GEN_ICV;
                    }
                    else /* IX_CRYPTO_ACC_OP_AUTH_CHECK */
                    {
                        ixCryptoCtx[*pCryptoCtxId].npeOperationMode.npeOperation
                            = IX_CRYPTO_NPE_OP_HMAC_VER_ICV;            
                    } /* end of if-else (pAccCtx->Operation) */
                } /* end of if (status) */
                break;
    
            /* Combined service, encryption followed by authentication,
             * notes: the code for the case of IX_CRYPTO_ACC_OP_ENCRYPT_AUTH
             * fall through to the case IX_CRYPTO_ACC_OP_AUTH_DECRYPT
             */
            case IX_CRYPTO_ACC_OP_ENCRYPT_AUTH: 
            
            /* Combined service, authentication followed by decryption */
            case IX_CRYPTO_ACC_OP_AUTH_DECRYPT:

                /* Register cipher algorithm */
                status = ixCryptoRegisterCipherAlgoRegister (
                             *pCryptoCtxId,
                             pAccCtx, 
                             &npeCryptoParamIndex,
                             IX_CRYPTO_OP_WAIT);

                /* Check if cipher context registration successful */
                if (IX_CRYPTO_ACC_STATUS_SUCCESS == status)
                {
                    /* Register authentication algorithm */
                    status = ixCryptoRegisterAuthAlgoRegister (
                                 *pCryptoCtxId,
                                 pAccCtx, 
                                 pMbufPrimaryChainVar,
                                 pMbufSecondaryChainVar, 
                                 &npeCryptoParamIndex,
                                 IX_CRYPTO_OP_REGISTER);

                    /* Check if authentication context registration 
                     * successful 
                     */
                    if (IX_CRYPTO_ACC_STATUS_SUCCESS == status)
                    {
                        /* Assign NPE operation */
                        if (IX_CRYPTO_ACC_OP_ENCRYPT_AUTH == pAccCtx->operation)
                        {
                            ixCryptoCtx[*pCryptoCtxId].npeOperationMode.
                                npeOperation 
                                    = IX_CRYPTO_NPE_OP_ENCRYPT_HMAC_GEN_ICV;
                        }
                        else /* IX_CRYPTO_ACC_OP_AUTH_DECRYPT */
                        {
                            ixCryptoCtx[*pCryptoCtxId].npeOperationMode.
                                npeOperation
                                    = IX_CRYPTO_NPE_OP_HMAC_VER_ICV_DECRYPT;
                        } /* end of if-else (pAccCtx->operation) */
                    } /* end of if (status) */
                } /* end of if (status) */
                break;
    
            /* all others case */
            default:
                /* operation not supported */
                status = IX_CRYPTO_ACC_STATUS_OPERATION_NOT_SUPPORTED;  

        } /* end of switch (pAccCtx->operation) */
                
        if (IX_CRYPTO_ACC_STATUS_SUCCESS == status) 
        {
            /* Set transfer mode, default is in-place transfer mode, OR
             * the NpeOperation with the mask will set the transfer mode
             * to 1 and turn it to non in-place operation
             */
            if (pAccCtx->useDifferentSrcAndDestMbufs)
            {
                ixCryptoCtx[*pCryptoCtxId].npeOperationMode.npeOperation
                    = (((ixCryptoCtx[*pCryptoCtxId].npeOperationMode.
                          npeOperation)
                          | IX_CRYPTO_NPE_OP_TRANSFER_MODE_MASK) 
                          & IX_CRYPTO_NPE_OP_MASK);
            } /* end of if (pAccCtx->useDifferentSrcAndDestMbufs) */
    
            /* Assign Init Length for NPE Crypto Parameters structure in 
             * Crypto Context 
             */
            ixCryptoCtx[*pCryptoCtxId].npeOperationMode.initLength 
                = npeCryptoParamIndex;

             /* Check if client's callback need to be called here */
             if (callbackFlag) /* Callback needed */
             {
                    /* Mark valid bit in Crypto Context to indicate the 
                     * Crypto Context has been registered successfully
                     */
                    ixCryptoCtx[*pCryptoCtxId].valid = TRUE;
                        
                    /* Call client's callback */
                    ixCryptoCtx[*pCryptoCtxId].registerCallbackFn(
                        *pCryptoCtxId,
                        NULL, 
                        status);                
            } /* end of if (callbackFlag) */                

            return status;
        }
        else /* Registration failed */
        {
            /* Release Crypto Context allocated previously */
            ixCryptoCCDMgmtCryptoCtxRelease (*pCryptoCtxId);
    
            return status;
        } /* end of if (status) */
    }
    else /* (IX_CRYPTO_IS_SERVICE_INITIALIZED) &&
          * (!IX_CRYPTO_IS_SERVICE_STOPPED)
          */
    {
        
        /* Log error message in debugging mode */
            IX_CRYPTO_ACC_LOG (LOG_ERROR, 
                "Component has not been initialized\n", 
                0, 0, 0, 0, 0, 0);
                
        return IX_CRYPTO_ACC_STATUS_FAIL;
    } /* end of if-else (IX_CRYPTO_IS_SERVICE_INITIALIZED) &&
       * (!IX_CRYPTO_IS_SERVICE_STOPPED) 
       */
} /* End of ixCryptoAccCtxRegister () function */



/**
 * @fn   ixCryptoAccAuthCryptPerform
 * @brief Perform Authentication and Decryption/Encryption functionalities
 *
 */

PUBLIC IxCryptoAccStatus
ixCryptoAccAuthCryptPerform (
    UINT32 cryptoCtxId,
    IX_MBUF *pSrcMbuf,
    IX_MBUF *pDestMbuf,
    UINT16 authStartOffset,
    UINT16 authDataLen,
    UINT16 cryptStartOffset,
    UINT16 cryptDataLen,
    UINT16 icvOffset,
    UINT8  *pIV)
{
    IxCryptoQDescriptor *pQDesc = NULL;
    UINT32 operationMode;
    UINT32 npeOperation;
    IxCryptoNpeOperationStatus operStatus = IX_CRYPTO_OP_PERFORM;
    UINT32 *pCipherCfg;
    IxCryptoAccStatus status;
    IX_MBUF *pMbuf = pSrcMbuf;
    INT32 newIcvOffset = icvOffset;
    BOOL useDiffBuf = FALSE;
    UINT8 *tempIcvAddr;
    UINT32 npeCryptCfg;
    UINT32 cipherMode;

/* Jack ++ 2003/11/13 */
    UINT8 *tosAddr;
    UINT8 *frag_offAddr;
    UINT8 *ttlAddr;
    UINT8 *checkAddr;
/* Jack -- */
       
    /* Check if the access component has been intialized and the crypto service
       is running */
    if (IX_CRYPTO_IS_SERVICE_INITIALIZED ()
        && !IX_CRYPTO_IS_SERVICE_STOPPED ())
    {
        /* Check the validity of CryptoCtxId */
        status = ixCryptoCCDMgmtCtxValidCheck (cryptoCtxId);
        if (IX_CRYPTO_ACC_STATUS_SUCCESS != status)
        {
            /* Log error message in debugging mode */
            IX_CRYPTO_ACC_LOG (LOG_MESSAGE,
                "cryptoCtxId %d is not valid or registered.\n", 
                cryptoCtxId, 0, 0, 0, 0, 0);
            return status;          
        }       
     
        /* Now we know CtxId valid, look up record for the NPE operationMode */
        operationMode = (ixCryptoCtx[cryptoCtxId].npeOperationMode.npeOperation
                            & MASK_8_BIT);
           
        /* Check for NULL MBufs, and the destination mbuf will only be checked 
         * if the transfer mode is non in place */
        if  ((NULL == pSrcMbuf)||
             ((IX_CRYPTO_NPE_OP_TRANSFER_MODE_IS_IN_PLACE != (operationMode 
             & IX_CRYPTO_NPE_OP_TRANSFER_MODE_MASK) &&
             (NULL==pDestMbuf))))
        {
            /* Log error message in debugging mode */
            IX_CRYPTO_ACC_LOG (
                LOG_MESSAGE,
                "Input Mbufs not valid. pSrcMbuf=0x%X\n \
                pDestMbuf=0x%X\n",
                (int) pSrcMbuf,
                (int) pDestMbuf, 
                0, 0, 0, 0);
            return IX_CRYPTO_ACC_STATUS_FAIL;       
        }
    
/* #ifndef NDEBUG     */
#if 0
        /* Perform boundary check. 
         * Note that the icvOffset is not checked against the mbuf length.
         * The actual length of the mbuf will not be known. Only the length
         * of the data stored in a mbuf is known.
         */
        status = ixCryptoPerformBufferOffsetBoundaryCheck (
                     operationMode,
                     pSrcMbuf,
                     pDestMbuf,
                     authStartOffset,
                     authDataLen,
                     cryptStartOffset,
                     cryptDataLen);
        if (IX_CRYPTO_ACC_STATUS_SUCCESS != status)
        {
            return status;
        } /* End of Boundary Check */       
#endif
 
        /* Get a free Q descriptor from Q descriptors pool. Return if 
         * unsuccessful */
        if (IX_CRYPTO_ACC_STATUS_SUCCESS != 
            ixCryptoDescMgmtQDescriptorGet (&pQDesc))
        {
 
            /* Error logging is done in the DescMgmt module */
            return IX_CRYPTO_ACC_STATUS_FAIL;
        } /* End of ixCryptoDescMgmtQDescriptorGet */

        /* Get NPE operation */
        npeOperation = operationMode & (~IX_CRYPTO_NPE_OP_TRANSFER_MODE_MASK);
 
        /* Get operation transfer mode */
        if (IX_CRYPTO_NPE_OP_TRANSFER_MODE_IS_IN_PLACE != (operationMode 
             & IX_CRYPTO_NPE_OP_TRANSFER_MODE_MASK))
        {
 
            /* set to TRUE if non in-place operation */
            useDiffBuf = TRUE;
	    IX_CRYPTO_DATA_CACHE_FLUSH (
		IX_MBUF_MDATA (pDestMbuf),
		IX_MBUF_MLEN (pDestMbuf));
        }
        
        /* Switch case based on NPE operation. Two tasks need to be 
         * executed before sending the request to NPE.
         * 1. Check if the ICV is in the middle of data. If the ICV
         *    is in the middle of data and the NPE operation is 
         *    IX_CRYPTO_NPE_OP_HMAC_VER_ICV or 
         *    IX_CRYPTO_NPE_OP_HMAC_VER_ICV_DECRYPT; copy the original 
         *    ICV value into queue descriptor and store the original ICV 
         *    offset into queue descriptor too. Pass in the new ICV address to 
         *    NPE with the ICV field in the source mbuf is cleared to zero.
         * 2. Check if the operation is non in-place operation, if it
         *    is non in-place operation, shift the mdata pointer in 
         *    destination mbuf back to authStartOffset/cryptStartOffset
         *    depending on the NPE operation. As NPE always start writing to 
         *    offset 0 instead of authStartOffset and cryptStartOffset. Thus 
         *    the mdata pointer need to be shifted in order to get the data 
         *    write to the correct offset
         */        
        switch (npeOperation)
        {
            /* HMAC verification only, or combined service with
             * HMAC verification operation. Notes : the code for 
             * HMAC verification fall thorugh to combined service
             * IX_CRYPTO_NPE_OP_HMAC_VER_ICV_DECRYPT
             */
            case IX_CRYPTO_NPE_OP_HMAC_VER_ICV:
            case IX_CRYPTO_NPE_OP_HMAC_VER_ICV_DECRYPT:
 
                if (useDiffBuf) /* If non in-place operation */
                {
                    /* Shift mbuf mdata pointer to right by authStartOffset */
                    IX_MBUF_MDATA (pDestMbuf) += authStartOffset;
                } /* end of if (useDiffBuf) */

                /* Store original ICV offset into queue descriptor */
                pQDesc->value.originalIcvOffset = icvOffset;

                /* Get ICV address */
                tempIcvAddr = (UINT8 *)
                                  ixCryptoUtilMbufOffsetToAddressConvert (
                                  pSrcMbuf,
                                  icvOffset,
                                  FALSE);
                                  
                /* Store original ICV value into queue descriptor */
                memcpy (
                    pQDesc->integrityCheckValue, 
                    tempIcvAddr, 
                    ixCryptoCtx[cryptoCtxId].digestLength );

                if ((icvOffset >= authStartOffset) &&
                    (icvOffset < (authStartOffset + authDataLen)))
                {       

		/* Jack ++ 2003/11/13 */
                /* Get TOS address */
                tosAddr = (UINT8 *)
                                  ixCryptoUtilMbufOffsetToAddressConvert (
                                  pSrcMbuf,
                                  1,
                                  FALSE);
                                  
                /* Store original TOS value into queue descriptor */
                memcpy (
                    &pQDesc->tos, 
                    tosAddr, 
                    sizeof(UINT8));

		memset (
			tosAddr, 
			0, 
			sizeof(UINT8) );

                /* Get FRAG_OFF address */
                frag_offAddr = (UINT8 *)
                                  ixCryptoUtilMbufOffsetToAddressConvert (
                                  pSrcMbuf,
                                  6,
                                  FALSE);
                                  
                /* Store original FRAG_OFF value into queue descriptor */
                memcpy (
                    &pQDesc->frag_off[0],
                    frag_offAddr, 
                    2 * sizeof(UINT8));

		memset (
			frag_offAddr, 
			0, 
			2 * sizeof(UINT8) );

                /* Get TTL address */
                ttlAddr = (UINT8 *)
                                  ixCryptoUtilMbufOffsetToAddressConvert (
                                  pSrcMbuf,
                                  8,
                                  FALSE);
                                  
                /* Store original TTL value into queue descriptor */
                memcpy (
                    &pQDesc->ttl, 
                    ttlAddr, 
                    sizeof(UINT8));

		memset (
			ttlAddr, 
			0, 
			sizeof(UINT8) );

                /* Get CHECK address */
                checkAddr = (UINT8 *)
                                  ixCryptoUtilMbufOffsetToAddressConvert (
                                  pSrcMbuf,
                                  10,
                                  FALSE);
                                  
                /* Store original CHECK value into queue descriptor */
                memcpy (
                    &pQDesc->check[0],
                    checkAddr, 
                    2 * sizeof(UINT8));

		memset (
			checkAddr, 
			0, 
			2 * sizeof(UINT8) );

                    memset (
                        tempIcvAddr, 
                        0, 
                        ixCryptoCtx[cryptoCtxId].digestLength );
                }
                
                /* Assign ICV value location to queue descriptor */    
                tempIcvAddr = pQDesc->integrityCheckValue;
                
                /* Convert ICV address to physical address and then 
                   swap to network byte order */
                pQDesc->npeQDesc.address.icvAddr
                    = (UINT32)IX_CRYPTO_VIRTUAL_TO_PHYSICAL_TRANSLATION (
                          (UINT32) tempIcvAddr);

                pQDesc->npeQDesc.address.icvAddr
                    = IX_CRYPTO_CONVERT_WORD_TO_NETWORK_ORDER 
  		      (pQDesc->npeQDesc.address.icvAddr);

                break; 
            
            /* HMAC generation only, or combined service with
             * HMAC generation operation. Notes : the code for 
             * HMAC generation fall thorugh to combined service
             * IX_CRYPTO_NPE_OP_ENCRYPT_HMAC_GEN_ICV
             */
            case IX_CRYPTO_NPE_OP_HMAC_GEN_ICV:
            case IX_CRYPTO_NPE_OP_ENCRYPT_HMAC_GEN_ICV:

                /* Store original ICV offset into queue descriptor */
                pQDesc->value.originalIcvOffset = icvOffset;
 
                if (useDiffBuf) /* If non in-place */
                {
                    pMbuf = pDestMbuf;
                    IX_MBUF_MDATA (pDestMbuf) += authStartOffset;
                    newIcvOffset -= authStartOffset;
                } /* end of if (useDiffBuf) */
                
                /* Get ICV address */
                tempIcvAddr = (UINT8 *)
                    ixCryptoUtilMbufOffsetToAddressConvert (
                        pMbuf,
                        icvOffset,
                        FALSE);

/////////////////
		/* Jack ++ 2003/11/13 */
                /* Get TOS address */
                if ((icvOffset >= authStartOffset) &&
                    (icvOffset < (authStartOffset + authDataLen)))
                {       
                    memset (
                        tempIcvAddr, 
                        0, 
                        ixCryptoCtx[cryptoCtxId].digestLength );

                tosAddr = (UINT8 *)
                                  ixCryptoUtilMbufOffsetToAddressConvert (
                                  pMbuf,
                                  1,
                                  FALSE);

                /* Store original TOS value into queue descriptor */
                memcpy (
                    &pQDesc->tos, 
                    tosAddr, 
                    sizeof(UINT8));
                                  
		memset (
			tosAddr, 
			0, 
			sizeof(UINT8) );

                /* Get FRAG_OFF address */
                frag_offAddr = (UINT8 *)
                                  ixCryptoUtilMbufOffsetToAddressConvert (
                                  pMbuf,
                                  6,
                                  FALSE);
                /* Store original FRAG_OFF value into queue descriptor */
                memcpy (
                    &pQDesc->frag_off[0], 
                    frag_offAddr, 
                    2 * sizeof(UINT8));
                                  
		memset (
			frag_offAddr, 
			0, 
			2 * sizeof(UINT8) );

                /* Get TTL address */
                ttlAddr = (UINT8 *)
                                  ixCryptoUtilMbufOffsetToAddressConvert (
                                  pMbuf,
                                  8,
                                  FALSE);

                /* Store original TTL value into queue descriptor */
                memcpy (
                    &pQDesc->ttl, 
                    ttlAddr, 
                    sizeof(UINT8));
                                  
		memset (
			ttlAddr, 
			0, 
			sizeof(UINT8) );

                /* Get CHECK address */
                checkAddr = (UINT8 *)
                                  ixCryptoUtilMbufOffsetToAddressConvert (
                                  pMbuf,
                                  10,
                                  FALSE);

                /* Store original CHECK value into queue descriptor */
                memcpy (
                    &pQDesc->check[0], 
                    checkAddr, 
                    2 * sizeof(UINT8));
                                  
		memset (
			checkAddr, 
			0, 
			2 * sizeof(UINT8) );

		}
////////////////

                /* Convert ICV address into physical address then
                   swap to network byte order  */
                pQDesc->npeQDesc.address.icvAddr
                    = (UINT32)IX_CRYPTO_VIRTUAL_TO_PHYSICAL_TRANSLATION (
                          tempIcvAddr);

                pQDesc->npeQDesc.address.icvAddr
		    = IX_CRYPTO_CONVERT_WORD_TO_NETWORK_ORDER 
                      (pQDesc->npeQDesc.address.icvAddr);    

                break; 
           
            /* Encrypt/Decrypt onlyoperation */
            default :
 
                if (useDiffBuf) /* If non in-place operation */
                {
                    
                    IX_MBUF_MDATA (pDestMbuf) += cryptStartOffset;
                } /* end of if (useDiffBuf) */
                break;
        } /* end of switch case (npeOperation) */ 
/* Jack ++ for debug */
#if 0
{
        int yi;
        UINT32 *p;
        p = (UINT32 *)(pMbuf->m_data);
        for(yi=0;yi < (pMbuf->m_len) /4;yi++)
                printk("in IxCryptoAcc.c before perf src %x\n",*(p+yi));
}
#endif
/* Jack -- */
        
        /* 
         * If we are doing crypto we need to set up IV. No IV needed for 
         * Authentication. The cipher block length check is only performed 
         * when the operationMode is crypt */   
        if(IX_CRYPTO_NPE_OP_CRYPT_MODE_IS_DISABLED != 
           (operationMode & IX_CRYPTO_NPE_OP_CRYPT_ENABLE_MASK )) 
        { 
 
            /* 
             * We are doing some Crypto: Now we can be sure that the 
             * NpeCryptCfgWord is at start of UINT8 NpeCryptoInfo array
             */ 
/* #ifndef NDEBUG */
#if 0
            /* Check for the correct cipher text and plain text length */          
            status = ixCryptoPerformCipherBlkLengthCheck (
                         operationMode,
                         ixCryptoCtx[cryptoCtxId].cipherBlockLength,
                         cryptStartOffset,
                         cryptDataLen);
                
            if (IX_CRYPTO_ACC_STATUS_SUCCESS != status)
            {

                /* Log error message in debugging mode */
                IX_CRYPTO_ACC_LOG (LOG_MESSAGE, 
		    "IX_CRYPTO_ACC_STATUS_CIPHER_INVALID_BLOCK_LEN.\n", 
                    0, 0, 0, 0, 0, 0);
                /* Need to release descriptor before returning. 
                 * Descriptor release will log a message when fail */
                ixCryptoDescMgmtQDescriptorRelease(pQDesc);
                return status;          
            } /* End of invalid block length */ 
#endif

	    /* Read NPE Crypto Cfg Word from CCD database. */
            pCipherCfg = (UINT32 *)ixCryptoCtx[cryptoCtxId].
		             pNpeCryptoParam->npeCryptoInfo;

	    /* Convert to Network Order (Big Endian) */
            npeCryptCfg = IX_CRYPTO_CONVERT_WORD_TO_NETWORK_ORDER (
		              *pCipherCfg);

	    /* Extract CipherMode from NPE Crypt Cfg Word */
	    cipherMode = ((npeCryptCfg >>
			   IX_CRYPTO_NPE_CRYPT_CFG_CIPHERMODE_LOC) &
			   IX_CRYPTO_NPE_CRYPT_CFG_CIPHERMODE_MASK);

            /* Check that IV needed, for either CBC or CTR  mode */
            if(IX_CRYPTO_NPE_CRYPT_CFG_ECB != cipherMode)
            {
 
                /* First check IV not NULL */
                if(NULL==pIV)
                {   
 
                    /* Log error message in debugging mode */
                    IX_CRYPTO_ACC_LOG (LOG_MESSAGE,
                               "Input IV not valid == NULL.\n", 
                               0, 0, 0, 0, 0, 0);
                    /* Descriptor release will log a message when fail */
                    ixCryptoDescMgmtQDescriptorRelease(pQDesc);
                    
                    return IX_CRYPTO_ACC_STATUS_FAIL;  
                } /* End of pIV Null check */
                
                /*
                 * Check if this is AES CTR. CTR Mode must be handled 
                 * differently
                 */
                if (IX_CRYPTO_NPE_CRYPT_CFG_CTR != cipherMode)  
                {
 
                    memcpy (pQDesc->npeQDesc.IV, pIV, 
                        ixCryptoCtx[cryptoCtxId].cipherIvLength);
                }
                else /* Create block for AES */
                {           
                    ixCryptoPerformAesCtrBlkCreate (
                        pQDesc, 
                        pIV, 
                        ixCryptoCtx[cryptoCtxId].
                        securityParameterIndex);
                } /* End of if-else CTR Check */    
            } /* End of if(ECB)  */      
        } /* End of if (Crypto) */
                  
        /* Construct Q descriptor message */   
        pQDesc->cryptoCtxId = cryptoCtxId;
        pQDesc->operStatus = operStatus;
    
        /* Set to NPE operation and initial values in Crypto Param structure 
         * As npeOperation and initLength are byte-length member fields,
         * conversion is not needed in data coherency endianness. 
         */
         pQDesc->npeQDesc.npeOperationMode.npeOperation
             = ixCryptoCtx[cryptoCtxId].npeOperationMode.npeOperation;

         pQDesc->npeQDesc.npeOperationMode.initLength
             = ixCryptoCtx[cryptoCtxId].npeOperationMode.initLength;
              
        /* Convert source mbuf into NPE format. Virtual address must be used when
         * flushing to external memory */
        pQDesc->npeQDesc.srcMbufAddr = (UINT32)
            (ixCryptoUtilMbufToNpeFormatConvert(pSrcMbuf));
          
        /* If operationMode is in place, DestMbuf will be empty and don't need to
         * be written to the pQDesc or flushed to memory.
         */
        if (IX_CRYPTO_NPE_OP_TRANSFER_MODE_IS_IN_PLACE != 
            (operationMode & IX_CRYPTO_NPE_OP_TRANSFER_MODE_MASK))
        {
 
            /* Convert destination mbuf into NPE format */
            pQDesc->npeQDesc.destMbufAddr = (UINT32)
                (ixCryptoUtilMbufToNpeFormatConvert (pDestMbuf));
        }
        
        /* Offset and length of data */
        pQDesc->npeQDesc.authStartOffset
            = IX_CRYPTO_CONVERT_SHORT_TO_NETWORK_ORDER(authStartOffset);
    
        pQDesc->npeQDesc.authLength
            = IX_CRYPTO_CONVERT_SHORT_TO_NETWORK_ORDER(authDataLen);

        pQDesc->npeQDesc.cryptStartOffset
            = IX_CRYPTO_CONVERT_SHORT_TO_NETWORK_ORDER(cryptStartOffset);

        pQDesc->npeQDesc.cryptLength
            = IX_CRYPTO_CONVERT_SHORT_TO_NETWORK_ORDER(cryptDataLen);
        
        /* Write Crypto Param Address and flush*/
        pQDesc->npeQDesc.cryptoCtxNpeAddr 
            = (UINT32)IX_CRYPTO_VIRTUAL_TO_PHYSICAL_TRANSLATION (
            ixCryptoCtx[cryptoCtxId].pNpeCryptoParam);

        pQDesc->npeQDesc.cryptoCtxNpeAddr 
	    = IX_CRYPTO_CONVERT_WORD_TO_NETWORK_ORDER 
              (pQDesc->npeQDesc.cryptoCtxNpeAddr);
        
        /* Flush Q descriptor message from cache */
        IX_CRYPTO_DATA_CACHE_FLUSH (pQDesc, sizeof (IxCryptoQDescriptor));
        
        /* translate pQDesc after flush */    
        pQDesc = (IxCryptoQDescriptor*) 
            IX_CRYPTO_VIRTUAL_TO_PHYSICAL_TRANSLATION (pQDesc);
    
        /* Enqueue request to hardware queue using QMgr */
        status = ixCryptoQAccessQueueWrite (IX_CRYPTO_ACC_CRYPTO_REQ_Q,
                     (UINT32 *) &pQDesc);               
    
        if (IX_CRYPTO_ACC_STATUS_SUCCESS == status)
        {
            ixCryptoCtx[cryptoCtxId].reqCount++;
        }
        else /* End of status check for Queue Write */ 
        {
            /* Log error message in debugging mode */
            IX_CRYPTO_ACC_LOG (LOG_MESSAGE,
                "ixCryptoQAccessQueueWrite failed.\n", 
                0, 0, 0, 0, 0, 0);

            /* Convert Q descriptor address back to virtual address 
             * before releasing it
             */
            pQDesc = (IxCryptoQDescriptor *) 
                         IX_CRYPTO_PHYSICAL_TO_VIRTUAL_TRANSLATION(
                         (UINT32) pQDesc);
                           
            /* Convert mbuf back to host format from NPE format */
            pSrcMbuf = ixCryptoUtilMbufFromNpeFormatConvert
                            ((IX_MBUF *) pQDesc->npeQDesc.srcMbufAddr);
            pDestMbuf = ixCryptoUtilMbufFromNpeFormatConvert
                            ((IX_MBUF *) pQDesc->npeQDesc.destMbufAddr);
            
            /* When the Queue write fails, pQDesc needs to be released */
            ixCryptoDescMgmtQDescriptorRelease(pQDesc);
     
        } /* End of if-else for Queue Write check */
        return status;
    }
    else /* end of if (IX_CRYPTO_IS_SERVICE_INITIALIZED) &&
        (IX_CRYPTO_IS_SERVICE_STOPPED) */
    {
        /* Log error message in debugging mode */
        IX_CRYPTO_ACC_LOG (LOG_ERROR,
                   "IxCrypto is not initialized or it is stopped.\n", 
                   0, 0, 0, 0, 0, 0);   
        return IX_CRYPTO_ACC_STATUS_FAIL;
    }/* end of if-else (IX_CRYPTO_IS_SERVICE_INITIALIZED) &&
    (IX_CRYPTO_IS_SERVICE_STOPPED) */  
} /* end of ixCryptoAccAuthCryptPerform () function */


/**
 * @fn ixCryptoAccCtxUnregister
 * @brief Unregister the crypto context from Cryptographic Context Database.
 *
 */
PUBLIC IxCryptoAccStatus
ixCryptoAccCtxUnregister (UINT32 cryptoCtxId)
{
    IxCryptoAccStatus status;

    /* Check if the access component has been intialized and the crypto 
     * service is running 
     */
    if (IX_CRYPTO_IS_SERVICE_INITIALIZED()
        && !IX_CRYPTO_IS_SERVICE_STOPPED())
    {
        /* Check if the Crypto Context ID provided points to a valid Crypto
         * Context 
         */
        status = ixCryptoCCDMgmtCtxValidCheck (cryptoCtxId);
        
        if (IX_CRYPTO_ACC_STATUS_SUCCESS == status) 
        {
            /* Check if all requests from this cryptoCtxId have been
             * completed 
             */
            if (IX_CRYPTO_CTX_IS_ALL_TASK_DONE (
                ixCryptoCtx[cryptoCtxId].reqCount,
                ixCryptoCtx[cryptoCtxId].reqDoneCount))
            {
                /* Unregister the Crypto Context */
                return ixCryptoCCDMgmtCryptoCtxRelease (cryptoCtxId);              
            }
            else /* IX_CRYPTO_CTX_IS_ALL_TASK_DONE == FALSE */
            {
                /* There are requests pending in the queue for this Crypto 
                 * Context, cancel the unregistration and send Retry 
                 * message back to the client 
                 */
                return IX_CRYPTO_ACC_STATUS_RETRY;
            } /* end of if-else (IX_CRYPTO_CTX_IS_ALL_TASK_DONE) */
        }
        else /* Crypto context does not exist */
        {
            /* Log error message in debugging mode */
            IX_CRYPTO_ACC_LOG (LOG_MESSAGE, 
                "Crypto Context %d does not exist.\n", 
                cryptoCtxId, 
                0, 0, 0, 0, 0);
                
            return status;
        } /*end of if (status) for (ixCryptoCCDMgmtCtxValidCheck) */
    }
    else /* (IX_CRYPTO_IS_SERVICE_INITIALIZED) &&
          * (!IX_CRYPTO_IS_SERVICE_STOPPED) 
          */
    {
        /* Log error message in debugging mode */
            IX_CRYPTO_ACC_LOG (LOG_ERROR, 
                "Component has not been initialized\n", 
                0, 0, 0, 0, 0, 0);
                
        return IX_CRYPTO_ACC_STATUS_FAIL;
    } /* end of if-else (IX_CRYPTO_IS_SERVICE_INITIALIZED) &&
       * (!IX_CRYPTO_IS_SERVICE_STOPPED) 
       */
} /* End of ixCryptoAccCtxUnregister () */



/**
 * @fn   ixCryptoAccCryptoServiceStop
 * @brief API to stop the security hardware accelerator services.
 *
 */
PUBLIC IxCryptoAccStatus
ixCryptoAccCryptoServiceStop (void)
{
    UINT32 timer = 0;
        
    /* Set the flag to TRUE to prevent new service request */
    ixCryptoAccServiceStop = TRUE;
    
    while (IX_CRYPTO_ACC_STATUS_SUCCESS 
        != ixCryptoDescMgmtAllQDescriptorInPoolCheck())
    {
        /* Put the task into sleep mode and wait for the NPE to complete
         * all the tasks in the queue 
         */
        ixOsServTaskSleep(IX_CRYPTO_ACC_DELAY_IN_MS);

        /* Timer to avoid infinite lock */
        timer++;
        if (IX_CRYPTO_ACC_TIMER_COUNT < timer )
        {
            /* Log error message in debugging mode */
            IX_CRYPTO_ACC_LOG (
            LOG_FATAL,
            "Stop function failed!\n", 
            0, 0, 0, 0, 0, 0);
            return IX_CRYPTO_ACC_STATUS_FAIL;
    
        } /* End of if timer loop */

    } /* end of while (ixCryptoDescMgmtAllQDescriptorInPoolCheck) */
    
    /* Release all Crypto Contexts and this will not fail */
    ixCryptoCCDMgmtCryptoCtxReleaseAll ();

    /* Release memeory pool allocated to NPE Crypto Param */
    ixCryptoCCDMgmtNpeCryptoParamPoolFree ();
    
    /* Release memory pool allocated to Desc pool */
    ixCryptoDescMgmtDescPoolFree ();
    
    return IX_CRYPTO_ACC_STATUS_SUCCESS;
} /* end of ixCryptoAccCryptoServiceStop () function */



/**
 * @fn   ixCryptoAccShow
 * @brief API for printing statistic and status.
 *
 */

PUBLIC void
ixCryptoAccShow (void)
{
    /* Show request queue and done queue status */
    ixCryptoQAccessQueueStatusShow (IX_CRYPTO_ACC_CRYPTO_REQ_Q); 
    ixCryptoQAccessQueueStatusShow (IX_CRYPTO_ACC_CRYPTO_DONE_Q);
       
    /* To show the number of descriptors have been used, and the pool 
     * usage
     */
    ixCryptoDescMgmtShow ();

    /* To show number of Crypto Contexts have been registered */
    ixCryptoCCDMgmtShow ();


} /* end of ixCryptoAccShow () function */


/**
 * @fn   ixCryptoAccShowWithId
 * @brief API for printing statistic and status for particular Context.
 *
 */

PUBLIC void
ixCryptoAccShowWithId (UINT32 cryptoCtxId)
{
    if (IX_CRYPTO_ACC_STATUS_SUCCESS == 
    ixCryptoCCDMgmtCtxValidCheck (cryptoCtxId))
    {
        /* Show request queue and done queue status, number of descriptors have 
         * been used, and the pool usage and show number of Crypto Contexts
         * that have been registered */
        ixCryptoAccShow ();

        /* To show contents of Crypto Context which has been registered */
        ixCryptoCCDMgmtCryptoCtxShow (cryptoCtxId);
    }
    else
    {
        /* Log error message in debugging mode */
        IX_CRYPTO_ACC_LOG (LOG_MESSAGE,
                   "Invalid cryptoCtxId.\n", 
                   0, 0, 0, 0, 0, 0);
    } /* End of if-else Ctx Check */
    
    
} /* end of ixCryptoAccShowWithId () */



/**
 * @fn ixCryptoAccHashKeyGenerate
 *
 * @brief This function is used to generate authentication key needed in HMAC
 *        authentication if the authentication key is greater than 64 bytes.
 *        New authentication key of L bytes size will be generated in this 
 *        function (L = 20 for SHA1, L = 16 for MD5). Please refer to RFC2104 
 *        for more details on the key size. The authentication key is padded 
 *        (extended) so that its length (in bits) is congruent to 448, 
 *        modulo 512 (please refer to RFC1321). 
 *
 * @note  
 * 1. This API will be called ONLY if the authentication key > 64 bytes. If
 *    key size <= 64 bytes, authentication key is used directly in Crypto
 *    context for registration, no further hashing is needed.
 * 2. This function must be called prior the Crypto Context registration if 
 *    the key size > 64. Result from this function will be used as 
 *    authentication key in the Crypto Context registration.
 * 3. It is client's responsibility to ensure the mbuf is big enough to
 *    hold the result (generated authentication key). No checking on mbuf 
 *    length will be done.
 *
 */
PUBLIC IxCryptoAccStatus
ixCryptoAccHashKeyGenerate (
    IxCryptoAccAuthAlgo hashAlgo,
    IX_MBUF *pMbufHashKey,
    IxCryptoAccHashKeyGenCompleteCallback hashKeyCallbackFn,
    UINT16 hashKeyStartOffset,
    UINT16 hashKeyLen,
    UINT16 hashKeyDestOffset,
    UINT32 *pHashKeyId)
{
    union
    {
        IxCryptoNpeHashCfgWord hashCfgWord;
        UINT32 npeHashCfgWord;
    } cfgWord;

    UINT32 npeHashConfigWord;
    INT8 *pHashKey;
    IxCryptoAccStatus status = IX_CRYPTO_ACC_STATUS_SUCCESS;
    IxCryptoQDescriptor *pQDesc = NULL;
    UINT32 initLength;
    UINT32 destAddr;
    UINT32 padLength;
    UINT32 lengthInBit;
    UINT32 i;
    UINT8 md5ChainVar[IX_CRYPTO_NPE_CRYPTO_PARAM_MD5_CHAIN_VAR_LEN] =
        MD5_CHAIN_VAR;
    UINT8 sha1ChainVar[IX_CRYPTO_NPE_CRYPTO_PARAM_SHA1_CHAIN_VAR_LEN] =
        SHA1_CHAIN_VAR;

    /* Check if the mbuf and hashKeyID pointer are NULL,
     * return error if either one is NULL
     */
    if ((NULL == pMbufHashKey) || (NULL == pHashKeyId))
    {
        /* Log error message in debugging mode */
        IX_CRYPTO_ACC_LOG (
            LOG_MESSAGE,
            "Mbuf is %x\n pHashKeyId is %x\n",
            (int) pMbufHashKey,
            (int) pHashKeyId,
            0, 0, 0, 0);

        return IX_CRYPTO_ACC_STATUS_FAIL;
    } /* end of if (pMbufHashKey) || (pHashKeyId) */

    /* Check if the mbuf provided is chained mbufs */
    if (NULL != IX_MBUF_NEXT_BUFFER_IN_PKT_PTR (pMbufHashKey))
    {
        /* chained mbuf, return  error */
        return IX_CRYPTO_ACC_STATUS_FAIL;
    } /* end of if (IX_MBUF_NEXT_BUFFER_IN_PKT_PTR (pMbufHashKey) */

    /* Check if the callback function provided by client is NULL, return
     * error if NULL
     */
    if (NULL == hashKeyCallbackFn)
    {
        /* Log error message in debugging mode */
        IX_CRYPTO_ACC_LOG (
            LOG_MESSAGE,
            "Hash key Callback: %x\n",
            (int) hashKeyCallbackFn,
            0,0, 0, 0, 0);

        return IX_CRYPTO_ACC_STATUS_FAIL;
    } /* end of if (hashKeyCallbackFn) */

    /* Check if the access component has been intialized and the crypto
     * service is running
     */
    if (IX_CRYPTO_IS_SERVICE_INITIALIZED()
        && !IX_CRYPTO_IS_SERVICE_STOPPED())
    {
        /* Check hash key length */
        if (IX_CRYPTO_ACC_MAX_AUTH_KEY_LENGTH >= hashKeyLen)
        {
            /* invalid key length */
            return IX_CRYPTO_ACC_STATUS_AUTH_INVALID_KEY_LEN;
        }

        /* get mData pointer of mbuf */
        pHashKey = IX_MBUF_MDATA (pMbufHashKey);

	/* Clear NPE hash config word to zero */
	cfgWord.npeHashCfgWord = 0;
        
        /* Hash algorithm switch case options
         * For each case of valid hash algorithm, hash config will be created.
         * Hash key will be padded to multiple of 512-bit according to the 
         * padding scheme in RFC1321 (MD5 hash algorithm) and RFC3174 (SHA1
         * hash algorithm).
         */
        switch (hashAlgo)
        {
            /* SHA1 algorithm */
            case IX_CRYPTO_ACC_AUTH_SHA1:

                /* Set authentication digest length to key length */
                cfgWord.hashCfgWord.digestLength
                    = (IX_CRYPTO_ACC_SHA1_KEY_160 / WORD_IN_BYTES);

                /* Set authentication algorithm */
                cfgWord.hashCfgWord.hashAlgo = IX_CRYPTO_NPE_HASH_CFG_SHA1;

                /* Set chaining variable length */
                cfgWord.hashCfgWord.CVLength
                    = (IX_CRYPTO_NPE_CRYPTO_PARAM_SHA1_CHAIN_VAR_LEN /
                        WORD_IN_BYTES);

                /* Init length for chaining variables computation is the length
                 * of hash config word plus length of primary chaining variable
                 */
                initLength = IX_CRYPTO_NPE_CRYPTO_PARAM_HASH_CFG_WORD_LEN
                                + IX_CRYPTO_NPE_CRYPTO_PARAM_SHA1_CHAIN_VAR_LEN;
                                
                /* Configure endian mode of NPE, by default is Big endian.
                 * No swapping is required
                 */
                cfgWord.hashCfgWord.chnWr = IX_CRYPTO_NPE_CFG_ENDIAN_NO_SWAP;
                cfgWord.hashCfgWord.chnRd = IX_CRYPTO_NPE_CFG_ENDIAN_NO_SWAP;
                cfgWord.hashCfgWord.hdWr = IX_CRYPTO_NPE_CFG_ENDIAN_NO_SWAP;
                cfgWord.hashCfgWord.hdRd = IX_CRYPTO_NPE_CFG_ENDIAN_NO_SWAP;
                
                /* Do padding, if the hash key is not congruent 448 bit 
                 * (56 bytes), modulo 512-bit (64 bytes). 
                 */
                if (IX_CRYPTO_ACC_HASH_DATA_REMAINDER ==
                    (hashKeyLen % IX_CRYPTO_AUTHENTICATION_DATA_BLOCK_LENGTH))
                {
                    /* length of original hash key in bits */
                    lengthInBit = hashKeyLen * BYTE_IN_BITS;
                    
                    /* pad hash key with value 0x80 */
                    memset (&pHashKey[hashKeyStartOffset + hashKeyLen], 
                        IX_CRYPTO_ACC_HASH_FIRST_PAD_BYTE, 
                        IX_CRYPTO_ACC_HASH_FIRST_PAD_BYTE_LEN);
                    
                    /* set total bit (high 32-bit) to zero, key size will
                     * not exceed 2^32
                     */
                    memset (
                        &pHashKey[hashKeyStartOffset + hashKeyLen
                        + IX_CRYPTO_ACC_HASH_FIRST_PAD_BYTE_LEN], 
                        0, 
                        IX_CRYPTO_ACC_HASH_TOTAL_BIT_HIGH_LEN);
                    
                    /* set total bit field (low 32-bit) */
                    for (i = 0; i < IX_CRYPTO_ACC_HASH_TOTAL_BIT_LOW_LEN; i++)
                    {
                        pHashKey[hashKeyStartOffset + hashKeyLen 
                            + IX_CRYPTO_ACC_HASH_FIRST_PAD_BYTE_LEN
                            + IX_CRYPTO_ACC_HASH_TOTAL_BIT_HIGH_LEN + i] 
                            = (lengthInBit >> ((MSB_BYTE - i) * BYTE_IN_BITS) 
			       & MASK_8_BIT);
                    }
                    
                    /* total hash key length after padding */
                    hashKeyLen = hashKeyLen 
                                 + IX_CRYPTO_ACC_HASH_TOTAL_BIT_LEN
                                 + IX_CRYPTO_ACC_HASH_FIRST_PAD_BYTE_LEN;
                }
                else
                {
                    /* length of original hash key in bits */
                    lengthInBit = hashKeyLen * BYTE_IN_BITS;
                    
                    /* calculate zero padding length (bytes) for original 
                     * hash key to make it multiple of 512-bit
                     */
                    padLength = IX_CRYPTO_AUTHENTICATION_DATA_BLOCK_LENGTH -
                                ((hashKeyLen + IX_CRYPTO_ACC_HASH_TOTAL_BIT_LEN
                                + IX_CRYPTO_ACC_HASH_FIRST_PAD_BYTE_LEN) % 
                                IX_CRYPTO_AUTHENTICATION_DATA_BLOCK_LENGTH);
                    
                    /* pad hash key with value 0x80 */
                    memset (&pHashKey[hashKeyStartOffset + hashKeyLen], 
                        IX_CRYPTO_ACC_HASH_FIRST_PAD_BYTE, 
                        IX_CRYPTO_ACC_HASH_FIRST_PAD_BYTE_LEN);
                        
                    /* pad original hash key with 0s at the back */
                    memset (
                        &pHashKey[hashKeyStartOffset + hashKeyLen 
                        + IX_CRYPTO_ACC_HASH_FIRST_PAD_BYTE_LEN], 
                        0, padLength);
                    
                    /* set total bit (high 32-bit) to zero, key size will
                     * not exceed 2^32
                     */
                    memset (&pHashKey[hashKeyStartOffset + hashKeyLen
                        + IX_CRYPTO_ACC_HASH_FIRST_PAD_BYTE_LEN
                        + padLength], 
                        0, 
                        IX_CRYPTO_ACC_HASH_TOTAL_BIT_HIGH_LEN);
                    
                    /* set total bit field (low 32-bit) */
                    for (i = 0; i < IX_CRYPTO_ACC_HASH_TOTAL_BIT_LOW_LEN; i++)
                    {
                        pHashKey[hashKeyStartOffset + hashKeyLen 
                            + IX_CRYPTO_ACC_HASH_FIRST_PAD_BYTE_LEN
                            + padLength 
                            + IX_CRYPTO_ACC_HASH_TOTAL_BIT_HIGH_LEN
                            + i] 
                            = (lengthInBit >> ((MSB_BYTE - i) * BYTE_IN_BITS) 
			       & MASK_8_BIT);
                    }
                    
                    /* total hash key length after padding */
                    hashKeyLen = hashKeyLen 
                                 + padLength
                                 + IX_CRYPTO_ACC_HASH_TOTAL_BIT_LEN
                                 + IX_CRYPTO_ACC_HASH_FIRST_PAD_BYTE_LEN;
                }
                
                break;

            /* MD5 algorithm */
            case IX_CRYPTO_ACC_AUTH_MD5:

                /* Set authentication digest length */
                cfgWord.hashCfgWord.digestLength
                    = (IX_CRYPTO_ACC_MD5_KEY_128 / WORD_IN_BYTES);

                /* Set authentication algorithm */
                cfgWord.hashCfgWord.hashAlgo = IX_CRYPTO_NPE_HASH_CFG_MD5;

                /* Set chaining variable length */
                cfgWord.hashCfgWord.CVLength
                    = (IX_CRYPTO_NPE_CRYPTO_PARAM_MD5_CHAIN_VAR_LEN /
                        WORD_IN_BYTES);

                /* Init length for chaining variables computation is the length
                 * of hash config word plus length of primary chaining variable
                 */
                initLength = IX_CRYPTO_NPE_CRYPTO_PARAM_HASH_CFG_WORD_LEN
                                + IX_CRYPTO_NPE_CRYPTO_PARAM_MD5_CHAIN_VAR_LEN;
                                
                /* Configure endian mode of NPE, MD5 hashing co-processor
                 * operate in Little Endian, byte swapping is required
                 */
                cfgWord.hashCfgWord.chnWr = IX_CRYPTO_NPE_CFG_BYTE_SWAP;
                cfgWord.hashCfgWord.chnRd = IX_CRYPTO_NPE_CFG_BYTE_SWAP;
                cfgWord.hashCfgWord.hdWr = IX_CRYPTO_NPE_CFG_BYTE_SWAP;
                cfgWord.hashCfgWord.hdRd = IX_CRYPTO_NPE_CFG_BYTE_SWAP;
                
                /* Do padding, if the hash key is not congruent 448 bit 
                 * (56 bytes), modulo 512-bit (64 bytes). 
                 */
                if (IX_CRYPTO_ACC_HASH_DATA_REMAINDER == 
                    (hashKeyLen % IX_CRYPTO_AUTHENTICATION_DATA_BLOCK_LENGTH))
                {
                    /* length of original hash key in bits */
                    lengthInBit = hashKeyLen * BYTE_IN_BITS;
                    
                    /* pad hash key with value 0x80 */
                    memset (&pHashKey[hashKeyStartOffset + hashKeyLen], 
                        IX_CRYPTO_ACC_HASH_FIRST_PAD_BYTE, 
                        IX_CRYPTO_ACC_HASH_FIRST_PAD_BYTE_LEN);
                        
                    /* set total bit field (low 32-bit) 
                     * Note :  MD5 operate in little endian, byte swap 
                     * needed while setting the total bit field
                     */
                    for (i = 0; i < IX_CRYPTO_ACC_HASH_TOTAL_BIT_LOW_LEN; i++)
                    {
                        pHashKey[hashKeyStartOffset + hashKeyLen + 
                            IX_CRYPTO_ACC_HASH_FIRST_PAD_BYTE_LEN +i] 
                                = (lengthInBit >> (i * BYTE_IN_BITS) 
				   & MASK_8_BIT);
                    }
                    
                    /* set total bit (high 32-bit) to zero, key size will
                     * not exceed 2^32
                     */
                    memset (
                        &pHashKey[hashKeyStartOffset + hashKeyLen
                        + IX_CRYPTO_ACC_HASH_FIRST_PAD_BYTE_LEN
                        + IX_CRYPTO_ACC_HASH_TOTAL_BIT_LOW_LEN], 
                        0, 
                        IX_CRYPTO_ACC_HASH_TOTAL_BIT_HIGH_LEN);
                    
                    /* total hash key length after padding */
                    hashKeyLen = hashKeyLen 
                                 + IX_CRYPTO_ACC_HASH_TOTAL_BIT_LEN
                                 + IX_CRYPTO_ACC_HASH_FIRST_PAD_BYTE_LEN;
                }
                else
                {
                    /* length of original hash key in bits */
                    lengthInBit = hashKeyLen * BYTE_IN_BITS;
                    
                    /* calculate zero padding length (bytes) for original 
                     * hash key to make it multiple of 512-bit
                     */
                    padLength = IX_CRYPTO_AUTHENTICATION_DATA_BLOCK_LENGTH -
                                ((hashKeyLen + IX_CRYPTO_ACC_HASH_TOTAL_BIT_LEN
                                + IX_CRYPTO_ACC_HASH_FIRST_PAD_BYTE_LEN) % 
                                IX_CRYPTO_AUTHENTICATION_DATA_BLOCK_LENGTH);
                    
                    /* pad hash key with value 0x80 */
                    memset (&pHashKey[hashKeyStartOffset + hashKeyLen], 
                        IX_CRYPTO_ACC_HASH_FIRST_PAD_BYTE, 
                        IX_CRYPTO_ACC_HASH_FIRST_PAD_BYTE_LEN);
                    
                    /* pad original hash key with 0s at the back */
                    memset (
                        &pHashKey[hashKeyStartOffset + hashKeyLen 
                        + IX_CRYPTO_ACC_HASH_FIRST_PAD_BYTE_LEN], 
                        0, 
                        padLength);
                    
                    /* set total bit field (low 32-bit) 
                     * Note :  MD5 operate in little endian, byte swap 
                     * needed while setting the total bit field
                     */
                    for (i = 0; i < IX_CRYPTO_ACC_HASH_TOTAL_BIT_LOW_LEN; i++)
                    {
                        pHashKey[hashKeyStartOffset + hashKeyLen + 
                            IX_CRYPTO_ACC_HASH_FIRST_PAD_BYTE_LEN 
                            + padLength +i] 
                                = (lengthInBit >> (i * BYTE_IN_BITS) 
				   & MASK_8_BIT);
                    }
                    
                    /* set total bit (high 32-bit) to zero, key size will
                     * not exceed 2^32
                     */
                    memset (
                        &pHashKey[hashKeyStartOffset + hashKeyLen
                        + IX_CRYPTO_ACC_HASH_FIRST_PAD_BYTE_LEN
                        + padLength 
                        + IX_CRYPTO_ACC_HASH_TOTAL_BIT_LOW_LEN],
                        0, 
                        IX_CRYPTO_ACC_HASH_TOTAL_BIT_HIGH_LEN);
                    
                    /* total hash key length after padding */
                    hashKeyLen = hashKeyLen 
                                 + padLength
                                 + IX_CRYPTO_ACC_HASH_TOTAL_BIT_LEN
                                 + IX_CRYPTO_ACC_HASH_FIRST_PAD_BYTE_LEN;
                }
                
                break;

            /* invalid authentication algorithm */
            default:
                return IX_CRYPTO_ACC_STATUS_AUTH_ALGO_NOT_SUPPORTED;

        } /* end of switch case (hashAlgo) */

        /* Get Crypto Param for hash key generation */
        status = ixCryptoCCDMgmtKeyCryptoParamGet (pHashKeyId);

        if (IX_CRYPTO_ACC_STATUS_SUCCESS == status)
        {
            /* Store callback function pointers into Crypto Context */
            ixKeyCryptoParam[*pHashKeyId].hashKeyCallbackFn
                = hashKeyCallbackFn;
        }
        else /* No Crypto Context is allocated */
        {
            return status;
        } /* end of if-else (status) */

	/* Generate NPE Hash Config word in NPE format (Big Endian) */
	ixCryptoUtilNpeHashCfgGenerate (
	    &npeHashConfigWord, 
	    &cfgWord.hashCfgWord);

        /* Copy Hash Config word into NPE Crypto Param structure */
        for ( i = 0;
            i < IX_CRYPTO_NPE_CRYPTO_PARAM_HASH_CFG_WORD_LEN;
            i++)
        {
            /* Shift Hash Config word into the array of UINT8,
             * most significant byte is shifted into the array with
             * smaller index.
             * Ex :
             *   -------------------
             *  | b0 | b1 | b2 | b3 |
             *   -------------------
             *  array[n] = b0
             *  array[n+1] = b1
             */
            ixKeyCryptoParam[*pHashKeyId].pNpeCryptoParam->npeCryptoInfo[i]
                = ((npeHashConfigWord >> ((MSB_BYTE - i) * BYTE_IN_BITS))
                  & MASK_8_BIT);
        } /* end of for (i) */

        /* Log NPE Crypt Config word in debugging mode */
        IX_CRYPTO_ACC_LOG (
            LOG_MESSAGE,
            "Hash Cfg word is %x\n",
            npeHashConfigWord,
            0, 0, 0, 0, 0);
        
        /* Initialize NPE crypto param structure with algorithm dependent
         * initial chaining variables value
         */
        if (IX_CRYPTO_ACC_AUTH_MD5 == hashAlgo)
        {
            /* Copy MD5 initial chaining variable value into NPE Crypto
             * param structure
             */
            memcpy (&(ixKeyCryptoParam[*pHashKeyId].pNpeCryptoParam->
                npeCryptoInfo[IX_CRYPTO_NPE_CRYPTO_PARAM_HASH_CFG_WORD_LEN]),
                md5ChainVar,
                IX_CRYPTO_NPE_CRYPTO_PARAM_MD5_CHAIN_VAR_LEN);
        }
        else /* IX_CRYPTO_ACC_AUTH_SHA1 */
        {
            /* Copy SHA1 initial chaining variable value into NPE Crypto
             * param structure
             */
            memcpy (&(ixKeyCryptoParam[*pHashKeyId].pNpeCryptoParam->
                npeCryptoInfo[IX_CRYPTO_NPE_CRYPTO_PARAM_HASH_CFG_WORD_LEN]),
                sha1ChainVar,
                IX_CRYPTO_NPE_CRYPTO_PARAM_SHA1_CHAIN_VAR_LEN);
        } /* end of if-else (hashAlgo) */

        /* Flush NPE Crypto Param structure from cache into SDRAM */
        IX_CRYPTO_DATA_CACHE_FLUSH (
            ixKeyCryptoParam[*pHashKeyId].pNpeCryptoParam,
            IX_CRYPTO_NPE_CRYPTO_PARAM_SIZE);
        
        /* Get a free Q descriptor from Q descriptors pool
         * Notes: Q descriptor is pushed back to the descriptor pool if the
         *        queue write operation failed. If the queue write operation
         *        pass, the Q descriptor will be processed by NPE and being
         *        release in QMgr dispatcher context through registered
         *        ixCryptoQAccessReqDoneQMgrCallback function
         */
        status =ixCryptoDescMgmtQDescriptorGet (&pQDesc);

        if (IX_CRYPTO_ACC_STATUS_SUCCESS == status)
        {
            /* Store hashKeyID & operStatus for the use in QMGr callback */
            pQDesc->value.keyId = *pHashKeyId;
            pQDesc->operStatus = IX_CRYPTO_OP_HASH_GEN_KEY;

            /* Get destination address */
            destAddr = ixCryptoUtilMbufOffsetToAddressConvert (
                           pMbufHashKey,
                           hashKeyDestOffset,
                           FALSE);

            /* Convert destination address to physical address
               and swap to network byte order */
            pQDesc->npeQDesc.address.icvAddr
                = (UINT32)IX_CRYPTO_VIRTUAL_TO_PHYSICAL_TRANSLATION (
                  destAddr);
 
            pQDesc->npeQDesc.address.icvAddr
	        = IX_CRYPTO_CONVERT_WORD_TO_NETWORK_ORDER (
		      pQDesc->npeQDesc.address.icvAddr);  
            
            /* Convert mbuf into NPE format */
            pQDesc->npeQDesc.srcMbufAddr
                = (UINT32) ixCryptoUtilMbufToNpeFormatConvert(pMbufHashKey);

            /* Set to chaining variables generation operation */
            pQDesc->npeQDesc.npeOperationMode.npeOperation
                = IX_CRYPTO_NPE_OP_HASH_GEN_ICV;

            /* Set length of initial values in Crypto Param structure */
            pQDesc->npeQDesc.npeOperationMode.initLength = initLength;

            /* Offset and length of data */
            pQDesc->npeQDesc.authStartOffset 
                 = IX_CRYPTO_CONVERT_SHORT_TO_NETWORK_ORDER (hashKeyStartOffset);

            pQDesc->npeQDesc.authLength 
                 = IX_CRYPTO_CONVERT_SHORT_TO_NETWORK_ORDER (hashKeyLen);

            /* Convert virtual address to physical address, NPE only takes
             * physical address
             */

            /* NPE Crypto Param structure start address */
            pQDesc->npeQDesc.cryptoCtxNpeAddr
                = IX_CRYPTO_VIRTUAL_TO_PHYSICAL_TRANSLATION
                  ((UINT32) ixKeyCryptoParam[*pHashKeyId].pNpeCryptoParam
                  ->npeCryptoInfo);

             pQDesc->npeQDesc.cryptoCtxNpeAddr 
 	        = IX_CRYPTO_CONVERT_WORD_TO_NETWORK_ORDER 
                  (pQDesc->npeQDesc.cryptoCtxNpeAddr);

            /* Flush Q descriptor message from cache */
            IX_CRYPTO_DATA_CACHE_FLUSH (pQDesc, sizeof (IxCryptoQDescriptor));

            /* Convert Q descriptor address to physical address */
            pQDesc = (IxCryptoQDescriptor *)
                         IX_CRYPTO_VIRTUAL_TO_PHYSICAL_TRANSLATION(
                         (UINT32) pQDesc);

            /* Enqueue request to hardware queue */
            status = ixCryptoQAccessQueueWrite (
                         IX_CRYPTO_ACC_CRYPTO_REQ_Q,
                         (UINT32 *) &pQDesc);

            /* Check if operation enqueued successfully */
            if (IX_CRYPTO_ACC_STATUS_SUCCESS != status)
            {
                /* Relaese NPE crypto param back to the pool */
                ixCryptoCCDMgmtKeyCryptoParamRelease (*pHashKeyId);

                /* Convert Q descriptor address back to virtual address
                 * before releasing it
                 */
                pQDesc = (IxCryptoQDescriptor *)
                         IX_CRYPTO_PHYSICAL_TO_VIRTUAL_TRANSLATION(
                         (UINT32) pQDesc);

                /* Convert mbuf back to host format from NPE format */
                pMbufHashKey = ixCryptoUtilMbufFromNpeFormatConvert
                                ((IX_MBUF *) pQDesc->npeQDesc.srcMbufAddr);

                /* Release Q descriptor back to the descriptor pool */
                ixCryptoDescMgmtQDescriptorRelease (pQDesc);
            }/* end of if-else (status) */
        } /* end of if (status) for (ixCryptoDescMgmtQDescriptorGet) */

        return status;
    }
    else /* (IX_CRYPTO_IS_SERVICE_INITIALIZED) &&
          * (!IX_CRYPTO_IS_SERVICE_STOPPED)
          */
    {

        /* Log error message in debugging mode */
            IX_CRYPTO_ACC_LOG (LOG_ERROR,
                "Component has not been initialized\n",
                0, 0, 0, 0, 0, 0);

        return IX_CRYPTO_ACC_STATUS_FAIL;
    } /* end of if-else (IX_CRYPTO_IS_SERVICE_INITIALIZED) &&
       * (!IX_CRYPTO_IS_SERVICE_STOPPED)
       */
} /* end of ixCryptoAccHashKeyGenerate () function */
