/**
 * @file IxCryptoAccCCDMgmt.c
 *
 * @date October-03-2002
 *
 * @brief  Source file for Crypto Context database(CCD) management 
 *         module
 *
 *
 * Design Notes:
 *
 * -- Intel Copyright Notice --
 * 
 * @par
 * Copyright 2002-2003 Intel Corporation All Rights Reserved.
 * 
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.
 * 
 * @par
 * The Material is protected by worldwide copyright and trade secret laws
 * and treaty provisions. No part of the Material may be used, copied,
 * reproduced, modified, published, uploaded, posted, transmitted,
 * distributed, or disclosed in any way except in accordance with the
 * applicable license agreement .
 * 
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel, except in accordance with the
 * applicable license agreement.
 * 
 * @par
 * Unless otherwise agreed by Intel in writing, you may not remove or
 * alter this notice or any other notice embedded in Materials by Intel
 * or Intel's suppliers or licensors in any way.
 * 
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * 
 * @par
 * -- End Intel Copyright Notice --
 */


/*
 * Put the system defined include files required.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/*
 * Put the user defined include files required.
 */
#include "IxCryptoAccCCDMgmt_p.h"
#include "IxCryptoAccUtilities_p.h"


/*
 * Global variables
 */
IxCryptoAccCryptoCtx ixCryptoCtx[IX_CRYPTO_ACC_MAX_ACTIVE_SA_TUNNELS];
                                   /**< Crypto Context database */
IxCryptoAccKeyCryptoParam ixKeyCryptoParam[IX_CRYPTO_ACC_KEY_CRYPTO_PARAM_NUM];
                         /**< Extra NPE Crypto Param for key generation */


/*
 * Variable declarations global to this file only.  Externs are followed by
 * static variables.
 */
static UINT32 cryptoCtxCount = 0;  /**< Index to Crypto Context List,
                                    * Crypto Context list head pointer
                                    */
static UINT32 cryptoCtxList[IX_CRYPTO_ACC_MAX_ACTIVE_SA_TUNNELS];
                                   /**< List of Crypto Context ID in CCD
                                    * this array is used as a FILO push-pop
                                    * list in crypto context get and release
                                    * funciton
                                    */
static UINT8 *pCryptoParamPool = NULL;  /**< NPE Crypto Param memory
                                         * pool pointer
                                         */
static UINT32 keyCryptoParamCount = 0;  /**< Index to crypto param List for,
                                         * key generation.
                                         * Key Crypto Param list head pointer
                                         */
static UINT32 keyCryptoParamList[IX_CRYPTO_ACC_KEY_CRYPTO_PARAM_NUM];
                                   /**< List of NPE crypto param structure,
                                    * this array is used as a FILO push-pop
                                    * list in crypto param get and release
                                    * funciton
                                    */
static INT32 keyCryptoParamLock; /**< Protect critical sections in this
                                  * module
                                  */
static BOOL initDone = FALSE;           /**< flag used to prevent multiple
                                         * initialization
                                         */

#ifndef NDEBUG
static UINT32 cryptoCtxDrainCount = 0;  /**< Number of times exceed
                                         * maximum tunnels allowed
                                         */
#endif /* ndef NDEBUG */


/**
 * @fn      ixCryptoCCDMgmtInit
 * @brief   Initialize Crypto Context database
 *
 */
IxCryptoAccStatus
ixCryptoCCDMgmtInit (void)
{
    UINT8 *pNpeCryptoParamPool = NULL;
    UINT32 i;

/* Jack ++ */
    if (1 /*!initDone*/)
    {
        /* Allocate memory to NPE Crypto Parameters Pool */
        pNpeCryptoParamPool = IX_CRYPTO_ACC_DRV_DMA_MALLOC(
                                  (IX_CRYPTO_ACC_MAX_ACTIVE_SA_TUNNELS +
                                  IX_CRYPTO_ACC_KEY_CRYPTO_PARAM_NUM)
                                  * IX_CRYPTO_NPE_CRYPTO_PARAM_SIZE);

        if (NULL == pNpeCryptoParamPool)
        {
            /* Log error message in debugging mode */
            IX_CRYPTO_ACC_LOG (
                LOG_ERROR,
                "Memory allocation for NPE Crypto Param Pool failed.\n",
                0, 0, 0, 0, 0, 0);

            /* Memory allocation failed */
            return IX_CRYPTO_ACC_STATUS_FAIL;
        } /* end of if (pNpeCryptoParamPool) */

        /* Store crypto param pool pointer, this pointer will be
         * used  in crypto param pool free function
         */
        pCryptoParamPool = pNpeCryptoParamPool;

        for (i = 0;
            i < IX_CRYPTO_ACC_MAX_ACTIVE_SA_TUNNELS;
            i++)
        {
            /* Assign Crypto Context ID to Crypto Context list */
            cryptoCtxList[i] = i;

            /* Initialize valid bit in Crypto Context to FALSE */
            ixCryptoCtx[i].valid = FALSE;

            /* Allocate a NPE Crypto Param structure from the NPE
             * Crypto Parameters Pool and assign NPE Crypto Parameters
             * structure to Crypto Context
             */
            ixCryptoCtx[i].pNpeCryptoParam =
                (IxCryptoNpeCryptoParam *) pNpeCryptoParamPool;

            /* Move the NPE Crypto Param Pool pointer to point to next
             * NPE Crypto Param structure */
            pNpeCryptoParamPool += IX_CRYPTO_NPE_CRYPTO_PARAM_SIZE;
        } /* end of for (i) */

        /* Set cryptoCtxList head pointer to last element in the list */
        cryptoCtxCount = IX_CRYPTO_ACC_MAX_ACTIVE_SA_TUNNELS;

        for (i = 0;
            i < IX_CRYPTO_ACC_KEY_CRYPTO_PARAM_NUM;
            i++)
        {
            /* Assign Crypto Context ID to Crypto Context list */
            keyCryptoParamList[i] = i;

            /* Allocate a NPE Crypto Param structure from the NPE
             * Crypto Parameters Pool and assign NPE Crypto Parameters
             * structure to Key Crypto Param
             */
            ixKeyCryptoParam[i].pNpeCryptoParam =
                (IxCryptoNpeCryptoParam *) pNpeCryptoParamPool;

            /* Move the NPE Crypto Param Pool pointer to point to next
             * NPE Crypto Param structure */
            pNpeCryptoParamPool += IX_CRYPTO_NPE_CRYPTO_PARAM_SIZE;
        } /* end of for (i) */

        /* Set keyCryptoParamList head pointer to last element in the list */
        keyCryptoParamCount = IX_CRYPTO_ACC_KEY_CRYPTO_PARAM_NUM;

        /* Set initDone flag to TRUE to indicate intialization has been
         * completed
         */
        initDone = TRUE;

        return IX_CRYPTO_ACC_STATUS_SUCCESS;
    }
    else /* CCD Mgmt module has been initialized */
    {
        return IX_CRYPTO_ACC_STATUS_FAIL;
    } /* end of if-else (initDone) */
} /* end of ixCryptoCCDMgmtInit ()function */



/**
 * @fn      ixCryptoCCDMgmtCryptoCtxGet
 * @brief   Get a Crypto Context ID from CCD pool. The CryptoCtxId will serve
 *          as index of the array to the CCD database.
 *
 */
IxCryptoAccStatus
ixCryptoCCDMgmtCryptoCtxGet (UINT32 *pCtxId)
{
    if (cryptoCtxCount > 0)  /* Unused Crypto Context available */
    {
        /* Get a Crypto Context ID from CCD */
        *pCtxId = cryptoCtxList[--cryptoCtxCount];

        /* Clear NPE Param structure */
        memset (
            ixCryptoCtx[*pCtxId].pNpeCryptoParam->npeCryptoInfo,
            0,
            IX_CRYPTO_MAX_CRYPTO_INFO_BYTES);

        /* Reset statistic count in Crypto Context */
        ixCryptoCtx[*pCtxId].reqCount = 0;
        ixCryptoCtx[*pCtxId].reqDoneCount = 0;

        /* Clear valid flag for reverse AES key */
        ixCryptoCtx[*pCtxId].validAndKeyId = 0;

        return IX_CRYPTO_ACC_STATUS_SUCCESS;
    }
    else /* cryptoCtxCount <= 0) */
    {
#ifndef NDEBUG
        cryptoCtxDrainCount++;
#endif /* ndef NDEBUG */

        /* Exceed Maximum active tunnels allowed */
        return IX_CRYPTO_ACC_STATUS_EXCEED_MAX_TUNNELS;
    } /* end of if-else(cryptoCtxCount) */
} /* end of ixCryptoCCDMgmtCryptoCtxGet () function */


/**
 * @fn      ixCryptoCCDMgmtCryptoCtxRelease
 * @brief   Release Crypto Context back to the CCD pool.
 *
 */
IxCryptoAccStatus
ixCryptoCCDMgmtCryptoCtxRelease (UINT32 cryptoCtxId)
{
    /* Mark valid bit in Crypto Context to FALSE */
    ixCryptoCtx[cryptoCtxId].valid = FALSE;

    if (cryptoCtxCount < IX_CRYPTO_ACC_MAX_ACTIVE_SA_TUNNELS)
    {
        /* Push the Crypto Context back to the CCD */
        cryptoCtxList[cryptoCtxCount++] = cryptoCtxId;

        return IX_CRYPTO_ACC_STATUS_SUCCESS;
    }
    else /* cryptoCtxCount >= IX_CRYPTO_ACC_MAX_ACTIVE_SA_TUNNELS) */
    {
        /* Log error message in debugging mode */
        IX_CRYPTO_ACC_LOG (
            LOG_ERROR,
            "Crypto Context %d release failed.\n",
            cryptoCtxId,
            0, 0, 0, 0, 0);

        /* CCD overflow */
        return  IX_CRYPTO_ACC_STATUS_FAIL;
    } /* end of if-else(cryptoCtxCount) */
} /* end of ixCryptoCCDMgmtCryptoCtxRelease () function */


/**
 * @fn      ixCryptoCCDMgmtKeyCryptoParamGet
 * @brief   Get a key ID from key crypto param pool. The keyId will serve
 *          as index of the array to the key crypto param database.
 *
 */
IxCryptoAccStatus
ixCryptoCCDMgmtKeyCryptoParamGet (UINT32 *pKeyId)
{
    IxCryptoAccStatus status;

    /* Lock keyCryptoParamList to protect critical section */
    IX_CRYPTO_ACC_LOCK (keyCryptoParamLock);

    if (keyCryptoParamCount > 0)  /* Unused Crypto Param available */
    {
        /* Get a key ID from key crypto param pool */
        *pKeyId = keyCryptoParamList[--keyCryptoParamCount];

        /* Clear NPE Param structure */
        memset (
            ixKeyCryptoParam[*pKeyId].pNpeCryptoParam->npeCryptoInfo,
            0,
            IX_CRYPTO_MAX_CRYPTO_INFO_BYTES);

        status = IX_CRYPTO_ACC_STATUS_SUCCESS;
    }
    else /* keyCryptoParamCount <= 0) */
    {
        /* Run out of Key Crypto Param, should not happen as the number
         * of key crypto param prepared is same as maximum og queue
         * entries
         */
        status = IX_CRYPTO_ACC_STATUS_FAIL;
    } /* end of if-else(keyCryptoParamCount) */

    /* Unlock given mutex */
    IX_CRYPTO_ACC_UNLOCK (keyCryptoParamLock);

    return status;
} /* end of ixCryptoCCDMgmtKeyCryptoParamGet () function */



/**
 * @fn      ixCryptoCCDMgmtKeyCryptoParamRelease
 * @brief   Release Crypto Param structure back to the key crypto param pool.
 *
 */
IxCryptoAccStatus
ixCryptoCCDMgmtKeyCryptoParamRelease (UINT32 keyId)
{
    IxCryptoAccStatus status;

    /* Lock keyCryptoParamList to protect critical section */
    IX_CRYPTO_ACC_LOCK (keyCryptoParamLock);

    if (keyCryptoParamCount < IX_CRYPTO_ACC_KEY_CRYPTO_PARAM_NUM)
    {
        /* Push the crypto param back to the pool */
        keyCryptoParamList[keyCryptoParamCount++] = keyId;

        status = IX_CRYPTO_ACC_STATUS_SUCCESS;
    }
    else /* keyCryptoParamCount >= IX_CRYPTO_ACC_KEY_CRYPTO_PARAM_NUM) */
    {
        /* Log error message in debugging mode */
        IX_CRYPTO_ACC_LOG (
            LOG_ERROR,
            "Key Crypto Param %d release failed.\n",
            keyId,
            0, 0, 0, 0, 0);

        /* Key Crypto Param overflow */
        status = IX_CRYPTO_ACC_STATUS_FAIL;
    } /* end of if-else(keyCryptoParamCount) */

    /* Unlock given mutex */
    IX_CRYPTO_ACC_UNLOCK (keyCryptoParamLock);

    return status;
} /* end of ixCryptoCCDMgmtKeyCryptoParamRelease () function */



/**
 * @fn      ixCryptoCCDMgmtCryptoCtxReleaseAll
 * @brief   Release ALL Crypto Context back to the CCD pool.
 *
 */
void
ixCryptoCCDMgmtCryptoCtxReleaseAll (void)
{
    UINT i;

    for (i = 0;
        i < IX_CRYPTO_ACC_MAX_ACTIVE_SA_TUNNELS;
        i++)
    {
        /* Mark valid bit in Crypto Context to FALSE */
        ixCryptoCtx[i].valid = FALSE;

        /* Assign Crypto Context ID to Crypto Context list */
        cryptoCtxList[i] = i;
    } /* end of for (cryptoCtxCount) */
    
    /* Reset Crypto Context List head pointer */
    cryptoCtxCount = IX_CRYPTO_ACC_MAX_ACTIVE_SA_TUNNELS;

} /* end of ixCryptoCCDMgmtCryptoCtxReleaseAll () function */



/**
 * @fn      ixCryptoCCDMgmtCtxValidCheck
 * @brief   Check whether the Crypto Contex is valid
 *
 */
IxCryptoAccStatus
ixCryptoCCDMgmtCtxValidCheck (UINT32 cryptoCtxId)
{
    if ((IX_CRYPTO_ACC_MAX_ACTIVE_SA_TUNNELS > cryptoCtxId)
        && ixCryptoCtx[cryptoCtxId].valid) /* Crypto Context valid */
    {
        return IX_CRYPTO_ACC_STATUS_SUCCESS;
    }
    else /* invalid Crypto Context */
    {
        return IX_CRYPTO_ACC_STATUS_CRYPTO_CTX_NOT_VALID;
    } /* end of if-else (ixCryptoCtx[cryptoCtxId].valid) */
} /* end of ixCryptoCCDMgmtCtxValidCheck () function */



/**
 * @fn      ixCryptoCCDMgmtCryptoCtxShow
 * @brief   Show the contents of Crypto Context which has been registered
 *
 */
void
ixCryptoCCDMgmtCryptoCtxShow (UINT32 cryptoCtxId)
{
    printf ("\n\nCrypto Context %5d \n", cryptoCtxId);
    printf ("---------------------\n");
    printf ("NPE operation ............. : %10x\n",
        ixCryptoCtx[cryptoCtxId].npeOperationMode.npeOperation);
    printf ("Init length ............... : %10x\n",
        ixCryptoCtx[cryptoCtxId].npeOperationMode.initLength);
    printf ("Requests Count ............ : %10u\n", 
        ixCryptoCtx[cryptoCtxId].reqCount);
    printf ("Requests Done Count ....... : %10u\n\n", 
        ixCryptoCtx[cryptoCtxId].reqDoneCount);

} /* end of ixCryptoCCDMgmtCryptoCtxShow () function */



/**
 * @fn      ixCryptoCCDMgmtShow
 * @brief   Show the numbers of Crypto Context have been registered.
 *
 */
void
ixCryptoCCDMgmtShow (void)
{
    /* CCD Management module display */
    printf ("\n\nCrypto Context Database  memory allocation\n");
    printf ("------------------------------------------\n");
    printf ("CCD size in bytes ........................ : %10u bytes\n",
       (UINT32) (IX_CRYPTO_CTX_SIZE * IX_CRYPTO_ACC_MAX_ACTIVE_SA_TUNNELS));
    printf ("Crypto Context size ...................... : %10u bytes\n", 
       (UINT32) IX_CRYPTO_CTX_SIZE);
    printf ("Maximum Crypto Context allowed ........... : %10u\n",
        IX_CRYPTO_ACC_MAX_ACTIVE_SA_TUNNELS);
    printf ("Number of Crypto Context registered ...... : %10u\n",
        (IX_CRYPTO_ACC_MAX_ACTIVE_SA_TUNNELS - cryptoCtxCount));

#ifndef NDEBUG
    printf ("No of times exceed maximum tunnels allowed : %10u\n",
        cryptoCtxDrainCount);
#endif /* ndef NDEBUG */
} /* end of ixCryptoCCDMgmtShow () function */



/**
 * @fn ixCryptoCCDMgmtNpeCryptoParamPoolFree
 *
 * @brief To free the memory allocated to NPE crypto param pool through 
 *        malloc function
 *
 */
void 
ixCryptoCCDMgmtNpeCryptoParamPoolFree (void)
{
    /* Check if the pool has been allocated */
    if(NULL != pCryptoParamPool)
    {
        /* free the memory allocated */
        IX_CRYPTO_ACC_DRV_DMA_FREE (pCryptoParamPool,
				    IX_CRYPTO_ACC_MAX_ACTIVE_SA_TUNNELS
				    * IX_CRYPTO_NPE_CRYPTO_PARAM_SIZE);
        
        /* Set flag to FALSE to indicate the CCD Crypto Param pool has 
         * been freed 
         */
        initDone = FALSE;
        
        printf ("CCD Crypto Param pool has been freed.\n");
    }/* end of if (pNpeCryptoParamPool) */ 
} /* end of ixCryptoCCDMgmtNpeCryptoParamPoolFree () function */
