/**
 * @file IxCryptoAccCryptoPerform_p.h
 *
 * @date 03-October-2002
 *
 * @brief Header file for the perform module.
 *
 *
 * Design Notes:
 *
 * -- Intel Copyright Notice --
 * 
 * @par
 * Copyright 2002-2003 Intel Corporation All Rights Reserved.
 * 
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.
 * 
 * @par
 * The Material is protected by worldwide copyright and trade secret laws
 * and treaty provisions. No part of the Material may be used, copied,
 * reproduced, modified, published, uploaded, posted, transmitted,
 * distributed, or disclosed in any way except in accordance with the
 * applicable license agreement .
 * 
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel, except in accordance with the
 * applicable license agreement.
 * 
 * @par
 * Unless otherwise agreed by Intel in writing, you may not remove or
 * alter this notice or any other notice embedded in Materials by Intel
 * or Intel's suppliers or licensors in any way.
 * 
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * 
 * @par
 * -- End Intel Copyright Notice --
 */


#ifndef IXCRYPTOACCCRYPTOPERFORM_P_H
#define IXCRYPTOACCCRYPTOPERFORM_P_H

/*
 * Os/System dependancies.
 */
#include <IxTypes.h>
#include <IxOsBuffMgt.h>

/*
 * Intermodule dependancies
 */

/**
 *
 * @fn ixCryptoPerformAesCtrBlkCreate
 *
 * @brief   To create initial counter block for the use of AES CTR encryption 
 *          mode operation.
 *
 * @note This function will create the initial counter block to be used in the 
 * 	 cryption operation by the NPE.
 *
 * @param   IxCryptoQDescriptor [in] *pQDescriptor - pointer to the queue 
 *          descriptor.
 * @param   UINT8 [in] *pIV - pointer to the initialization vector.
 * @param   UINT32 [in] SecurityParameterIndex - SPI for the AES.
 *
 * @return: none
 *
 * @li   Reentrant    : no
 * @li   ISR Callable : no
 *
 */

PUBLIC void 
ixCryptoPerformAesCtrBlkCreate (
    IxCryptoQDescriptor *pQDescriptor,
    UINT8 *pIV,
    UINT32 SecurityParameterIndex);

/**
 *
 * @fn ixCryptoPerformBufferOffsetBoundaryCheck
 *
 * @brief   To check the boundary condition of the buffer offset provided by
 *          the client.
 *
 * @note This function will ensure that the boundary conditions are satisfied:
 * 	 Auth offset <= Crypt offset <= Crypt offset + Crypt length <= Auth 
 *	 offset + Auth length.
 *       It also checks that offsets/lengths given fit inside mBufs passed.
 *	 The boundary check in this function will be activated under DEBUG mode
 *	 ONLY in order to optimize the performance in the data path.
 *
 * @param   UINT32 [in] npeOperation - NPE Operation in NPE operation mode of
 *	    struct IxCryptoNpeOperationMode..
 * @param   IX_MBUF [in] *pSrcMbuf - Source Mbuf pointer.
 * @param   IX_MBUF [in] *pDesMbuf - Destination Mbuf pointer.
 * @param   UINT16 [in] AuthStartOffset - Start address of the data to be 
 *          authenticated.
 * @param   UINT16 [in] AuthDataLen - The length of data to be authenticated.
 * @param   UINT16 [in] CryptStartOffset - Start address of the data to be 
 *          encrypted/decrypted.
 * @param   UINT16 [in] CryptDataLen - The length of data to be en/decrypted.
 *
 * @return @li  IX_CRYPTO_ACC_STATUS_SUCCESS - successfully checked the 
 *              valid buffer offset and length.
 * @return @li  IX_CRYPTO_ACC_STATUS_FAIL - one of the parameters is invalid.
 *
 * @li   Reentrant    : no
 * @li   ISR Callable : no
 */

IxCryptoAccStatus 
ixCryptoPerformBufferOffsetBoundaryCheck (
    UINT32 npeOperation,
    IX_MBUF *pSrcMbuf,
    IX_MBUF *pDestMbuf,
    UINT16 authStartOffset,
    UINT16 authDataLen,
    UINT16 cryptStartOffset,
    UINT16 cryptDataLen);

/**
 *
 * @fn ixCryptoPerformCipherBlkLengthCheck
 *
 * @brief   To check the data length against the pre-determined cipher block
 *          length (8 or 16 bytes) stated in the Security Hardware Accelerator 
 *          Function Spec.
 *  
 * @note This function will ensure the correct block size for symmetric 
 *	 en/decryption. The size for DES is 8 bytes while for AES' block size 
 *	 is of 16 bytes.
 *	 The block length check in this function will be activated under DEBUG
 *	 mode ONLY in order to optimize the performance in the data path.
 *
 * @param   UINT32 [in] npeOperation - NPE Operation in NPE operation mode of
 *	    struct IxCryptoNpeOperationMode.
 * @param   UINT16 [in] cipherBlkLen - the length of the cipher block.
 * @param   UINT16 [in] cryptStartOffset - the crypt start offset.
 * @param   UINT16 [in] cryptDataLen - the length of the data or plaintext..
 *
 * @return @li   * @param   IX_MBUF [in] *pSrcMbuf - Source Mbuf pointer.
 * @param   IX_MBUF [in] *pDesMbuf - Destination Mbuf pointer.
 *
 * @return @li  IX_CRYPTO_ACC_STATUS_SUCCESS - Successful block length check.
 * @return @li  IX_CRYPTO_ACC_STATUS_CIPHER_INVALID_BLOCK_LEN - Invalid block
 *              length for encryption.
 *
 * @li   Reentrant    : no
 * @li   ISR Callable : no
 *
 */

IxCryptoAccStatus 
ixCryptoPerformCipherBlkLengthCheck (
    UINT32 npeOperation,
    UINT16 cipherBlkLen,
    UINT16 cryptStartOffset,	
    UINT16 cryptDataLen);

#endif /* ifndef IXCRYPTOACCCRYPTOPERFORM_P_H */



