/**
 * @file IxCryptoAccQAccess.c
 *
 * @date October-03-2002
 *
 * @brief Source file for Queue Access module. This module is the 
 *        only interface to QMgr for IxCryptoAcc software component. This 
 *        module provides function to read from or write to the queue, to 
 *        show the queue status and also the notification callback for QMgr.
 *
 *
 * Design Notes:
 *
 * -- Intel Copyright Notice --
 * 
 * @par
 * Copyright 2002-2003 Intel Corporation All Rights Reserved.
 * 
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.
 * 
 * @par
 * The Material is protected by worldwide copyright and trade secret laws
 * and treaty provisions. No part of the Material may be used, copied,
 * reproduced, modified, published, uploaded, posted, transmitted,
 * distributed, or disclosed in any way except in accordance with the
 * applicable license agreement .
 * 
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel, except in accordance with the
 * applicable license agreement.
 * 
 * @par
 * Unless otherwise agreed by Intel in writing, you may not remove or
 * alter this notice or any other notice embedded in Materials by Intel
 * or Intel's suppliers or licensors in any way.
 * 
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * 
 * @par
 * -- End Intel Copyright Notice --
 */


/*
 * System defined include files required.
 */
#include <stdio.h>

/*
 * Put the user defined include files required.
 */
#include "IxCryptoAcc_p.h"
#include "IxCryptoAccQAccess_p.h"
#include "IxCryptoAccDescMgmt_p.h"
#include "IxCryptoAccCCDMgmt_p.h"
#include "IxCryptoAccUtilities_p.h"


#define IX_CRYPTO_ACC_QM_QUEUE_DISPATCH_PRIORITY IX_QMGR_Q_PRIORITY_0
                                        /**< Priority of queue in QMgr 
                                         * dispatcher (higest priority)
                                         */
#define IX_CRYPTO_ACC_CRYPTO_REQ_Q_SOURCE 0
                                        /**< Notification callback trigger 
                                         * source. It is a dummy definition, 
                                         * not used in the code as IxCryptoAcc
                                         * always write to the queue once 
                                         * there is a request from client.
                                         */
#define IX_CRYPTO_ACC_CRYPTO_DONE_Q_SOURCE IX_QMGR_Q_SOURCE_ID_NOT_E
                                        /**< Notification callback trigger
                                         * source
                                         */

#define IX_CRYPTO_NPE_OP_STATUS_MASK    0x80000000  /**< Q descriptor  
                                                     * authentication status 
                                                     * mask, MSB
                                                     */                                             
#define IX_CRYPTO_NPE_OP_AUTH_FAIL      0x80000000  /**< NPE Operation failed,
                                                     * authentication failed 
                                                     */                        

/*
* Variable declarations global to this file only.  Externs are followed by
* static variables.
*/
static BOOL initDone = FALSE;      /**< flag used to prevent multiple
                                    * initialization
                                    */
static IxCryptoAccStats cryptoStats;      /**< Statistics data structure */

                                    
/* Queue configuration info */                                    
static IxCryptoQCfgInfo queueInfo[IX_CRYPTO_ACC_NUM_OF_Q]=
{
  { 
    IX_CRYPTO_ACC_CRYPTO_REQ_Q,       /**< Queue ID */
    "Crypto Req Q",                   /**< Queue name */
    NULL,                             /**< Notification callback */
    IX_CRYPTO_ACC_CRYPTO_REQ_Q_CB_TAG, /**< Callback tag */ 
    IX_QMGR_Q_SIZE64,                 /**< Q Size*/
    IX_QMGR_Q_ENTRY_SIZE1,            /**< Queue Entry Sizes - all Q
                                       * entries are single word entries */
    FALSE,                            /**< Disable Q notification at 
                                       * startup 
                                       */
    IX_CRYPTO_ACC_CRYPTO_REQ_Q_SOURCE,/**< Q Condition to drive callback */
    IX_QMGR_Q_WM_LEVEL0,              /**< Q Low watermark - not used */
    IX_QMGR_Q_WM_LEVEL64,             /**< Q High watermark - not used */    
  },

  { 
    IX_CRYPTO_ACC_CRYPTO_DONE_Q,        /**< Queue ID */
    "Crypto Done Q",                    /**< Queue name */
    ixCryptoQAccessReqDoneQMgrCallback, /**< Notification callback */
    IX_CRYPTO_ACC_CRYPTO_DONE_Q_CB_TAG, /**< Callback tag */ 
    IX_QMGR_Q_SIZE64,                   /**< Q Size*/
    IX_QMGR_Q_ENTRY_SIZE1,              /**< Queue Entry Sizes - all Q
                                         * entries are single word entries */
    TRUE,                               /**< Enable Q notification at 
                                         * startup 
                                         */
    IX_CRYPTO_ACC_CRYPTO_DONE_Q_SOURCE, /**< Q Condition to drive callback */
    IX_QMGR_Q_WM_LEVEL0,                /**< Q Low watermark */
    IX_QMGR_Q_WM_LEVEL64,               /**< Q High watermark - not used */    
  }
};

/* From CCD management module */
extern IxCryptoAccStats cryptoStats;


/**
 * @fn ixCryptoQAccessInit
 * @brief Queue configuration and callback registration function
 *
 */
IxCryptoAccStatus 
ixCryptoQAccessInit (void)
{
    UINT32 i;
    
    if (initDone) /* If the module has been initialized */
    {
        return IX_CRYPTO_ACC_STATUS_FAIL;
    } /* end of if (initDone) */
    
    for (i = 0; i < IX_CRYPTO_ACC_NUM_OF_Q; i++)
    {
        /* Configure each Q */
        if (IX_SUCCESS != (ixQMgrQConfig (
                              queueInfo[i].qName,
                              queueInfo[i].qId,
                              queueInfo[i].qSize,
                              queueInfo[i].qWords)))
        {
            /* configuration failed */
            return IX_CRYPTO_ACC_STATUS_FAIL;
        } /* end of if (ixQMgrQConfig) */

        /* Set notification condition for Q */  
        if (queueInfo[i].qNotificationEnableAtStartup) 
        {
            /* Set Q watermarks for each Q */
            if (IX_SUCCESS != ixQMgrWatermarkSet (
                                  queueInfo[i].qId,
                                  queueInfo[i].AlmostEmptyThreshold,
                                  queueInfo[i].AlmostFullThreshold))
            {
                /* configuration failed */
                return IX_CRYPTO_ACC_STATUS_FAIL;
            } /* end of if (ixQMgrWatermarkSet) */
            
            /* Set dispatcher priority for the Q */
            if (IX_SUCCESS != ixQMgrDispatcherPrioritySet ( 
                                  queueInfo[i].qId, 
                                  IX_CRYPTO_ACC_QM_QUEUE_DISPATCH_PRIORITY))
            {
                /* configuration failed */
                return IX_CRYPTO_ACC_STATUS_FAIL;
            } /* end of if (ixQMgrDispatcherPrioritySet) */
       
            /* Register callbacks for each Q */ 
            if (IX_SUCCESS != ixQMgrNotificationCallbackSet (
                                  queueInfo[i].qId,
                                  queueInfo[i].qCallback,
                                  queueInfo[i].callbackTag))
            {
                /* callback registration failed */
                return IX_CRYPTO_ACC_STATUS_FAIL;
            } /* end of if (ixQMgrNotificationCallbackSet) */

            if (IX_SUCCESS != ixQMgrNotificationEnable (
                                   queueInfo[i].qId,
                                   queueInfo[i].qConditionSource))
            {
                /* configuration failed */
                return IX_CRYPTO_ACC_STATUS_FAIL;
            } /* end of if (ixQMgrNotificationEnable) */
        } /* end of if (queueInfo[i].qNotificationEnableAtStartup) */        
    } /* end of for (i) */
    
     /* Initialize all the member in the cryptoStats struct */
    cryptoStats.cryptoSuccessCounter = 0;
    cryptoStats.cryptoFailCounter = 0;
    cryptoStats.qOverflowCounter = 0;
    cryptoStats.qUnderflowCounter = 0;
    cryptoStats.qDescAddrInvalidCounter = 0;
    
    /* Set flag to indicate this module has been initializaed */
    initDone = TRUE;
    
    return IX_CRYPTO_ACC_STATUS_SUCCESS;

} /* end of ixCryptoQAccessInit () function */



/**
 * @fn ixCryptoQAccessReqDoneQMgrCallback
 * @brief Notification callback registered with QManager. Once the 
 *        threshold of the interrupt source is reached, the notification
 *        callback is triggered and is executed.
 *
 */
void 
ixCryptoQAccessReqDoneQMgrCallback (
    IxQMgrQId queueId, 
    IxQMgrCallbackId callbackId)
{
    UINT32 messageCount = 0;
    IxCryptoQDescriptor *pQDesc = NULL;
    IX_MBUF *pSrcMbuf;
    IX_MBUF *pDestMbuf;
    UINT32 operationResult;
    UINT32 cryptoCtxId;
    IX_STATUS qStatus;
    IxCryptoAccStatus status;
    UINT32 npeOperation;
    BOOL useDiffBuf = FALSE;
    UINT8 *tempIcvAddr;
    UINT32 keyId;
/* Jack ++ 2003/11/14 */
    UINT8 *tosAddr;
    UINT8 *frag_offAddr;
    UINT8 *ttlAddr;
    UINT8 *checkAddr;


    do
    {
        /* Read queue entry from hardware queue */
        qStatus = ixQMgrQRead (queueId, &operationResult);

        /* Check if the queue read operation pass */
        if (IX_SUCCESS == qStatus)
        {
            /* Count how many messages got this callback */
            messageCount++;
    
            /* Reset src mbuf and dest mbuf pointers to NULL */
            pSrcMbuf = NULL;
            pDestMbuf = NULL;
            
            pQDesc = (IxCryptoQDescriptor *)
            (operationResult & IX_CRYPTO_Q_DESC_ADDR_MASK);
    
            /* Convert Q descriptor pointer from physical address to virtual
             * address
             */
            pQDesc = (IxCryptoQDescriptor *) 
            IX_CRYPTO_PHYSICAL_TO_VIRTUAL_TRANSLATION(
                (UINT32) pQDesc);
                         
            if (NULL != pQDesc) /* if pQDesc not NULL */
            {       
                /* Invalidate Q descriptor message in cache */
                IX_ACC_DATA_CACHE_INVALIDATE (pQDesc, 
                                  sizeof (IxCryptoQDescriptor));
            
                cryptoCtxId = pQDesc->cryptoCtxId; /* Get Crypto Context ID */
        
                /* Check if the source mbuf is NULL */
                if (NULL != (IX_MBUF *) pQDesc->npeQDesc.srcMbufAddr)
                {

                    /* Convert mbuf header and its contents from NPE format
                     * to host format
                     */
                    pSrcMbuf = ixCryptoUtilMbufFromNpeFormatConvert
                                   ((IX_MBUF *) pQDesc->npeQDesc.srcMbufAddr);
                } /* end of if (pQDesc->NpeQDesc.SrcMbufAddr) */
            
                /* Get NPE operation from Q descriptor */
                npeOperation 
                    = ((pQDesc->npeQDesc.npeOperationMode.npeOperation
                           & (~IX_CRYPTO_NPE_OP_TRANSFER_MODE_MASK)) 
                           & MASK_8_BIT);
                
                if (IX_CRYPTO_NPE_OP_TRANSFER_MODE_IS_IN_PLACE 
                    != (pQDesc->npeQDesc.npeOperationMode.npeOperation
                    & IX_CRYPTO_NPE_OP_TRANSFER_MODE_MASK))
                {
                    /* set to TRUE if non in-place operation */
                    useDiffBuf = TRUE;
                }
            
                /* Switch case based on NPE operation. Two tasks need to be 
                 * executed in the callback context depending on the NPE 
                 * operation completed.
                 * 1. Check if the ICV is in the middle of data. If the ICV
                 *    is in the middle of data and the NPE operation is 
                 *    IX_CRYPTO_NPE_OP_HMAC_VER_ICV or 
                 *    IX_CRYPTO_NPE_OP_HMAC_VER_ICV_DECRYPT; copy the original 
                 *    ICV value from queue descriptor and write it back to 
                 *    source and/or destination mbuf (depends on transfer 
                 *    mode) based on the original ICV offset stored in queue 
                 *    descriptor.
                 * 2. Check if the operation is non in-place operation, if it
                 *    is non in-place operation, shift the mdata pointer in
                 *    destination mbuf back to its original offset. As the
                 *    mdata pointer is shifted according to authStartOffset or
                 *    cryptStartOffset to cover the constraint of NPE in 
                 *    writing the result to offset 0 in destination mbuf.
                 *    NPE always writes to offset 0 instead of authStartOffset
                 *    and cryptStartOffset
                 */                
/* Jack ++ for swap */
#if 0
                {
                        UINT32 *P;
                        int P_len = pSrcMbuf->m_len;
                        P = (UINT32 *)(pSrcMbuf->m_data);

                        while(P_len > 0) {
                                *P = (UINT32)ixOsServWordSwap((UINT32)(*P));
                                P_len -= sizeof(UINT32);
                                P ++;
                        }
                }
#endif
/* Jack -- */
                switch (npeOperation)
                {
                    /* HMAC verification only, or combined service with
                     * HMAC verification operation. Notes : the code for 
                     * HMAC verification fall thorugh to combined service
                     * IX_CRYPTO_NPE_OP_HMAC_VER_ICV_DECRYPT
                     */
                    case IX_CRYPTO_NPE_OP_HMAC_VER_ICV:
                    case IX_CRYPTO_NPE_OP_HMAC_VER_ICV_DECRYPT:
                        /* Check if the destination mbuf is NULL */ 
                        if (NULL != (IX_MBUF *) pQDesc->npeQDesc.destMbufAddr)
                        {
                            /* Convert mbuf header and its contents from NPE 
                             * format to host format
                             */
                            pDestMbuf 
                                = ixCryptoUtilDestMbufFromNpeFormatConvert 
                                    ((IX_MBUF *) pQDesc->npeQDesc.destMbufAddr,
                                    pQDesc->npeQDesc.authStartOffset);
                        } /* end of if (pQDesc->NpeQDesc.DestMbufAddr) */
                        
                        pQDesc->npeQDesc.authStartOffset
 			   = IX_CRYPTO_CONVERT_SHORT_TO_HOST_ORDER (
                             pQDesc->npeQDesc.authStartOffset);
  
                        pQDesc->npeQDesc.authLength 
                           = IX_CRYPTO_CONVERT_SHORT_TO_HOST_ORDER (
                             pQDesc->npeQDesc.authLength);

                        /* If ICV in the middle of data */
                        if ((pQDesc->value.originalIcvOffset >= 
                            pQDesc->npeQDesc.authStartOffset) &&
                            (pQDesc->value.originalIcvOffset <
                            (pQDesc->npeQDesc.authStartOffset + 
                            pQDesc->npeQDesc.authLength)))
                        {

                            /* Get ICV address from source mbuf */
                            tempIcvAddr = (UINT8 *)
                                ixCryptoUtilMbufOffsetToAddressConvert (
                                pSrcMbuf,
                                pQDesc->value.originalIcvOffset,
                                FALSE);
                                    
                            /* Restore ICV value into source mbuf */        
                            memcpy (
                                tempIcvAddr, 
                                pQDesc->integrityCheckValue,
                                ixCryptoCtx[cryptoCtxId].digestLength);
//////////////////
                            /* Get TOS address from source mbuf */
                            tosAddr = (UINT8 *)
                                ixCryptoUtilMbufOffsetToAddressConvert (
                                pSrcMbuf,
                                1,
                                FALSE);
                                    
                            /* Restore TOS value into source mbuf */        
                            memcpy (
                                tosAddr, 
                                &pQDesc->tos,
                                sizeof(UINT8));

                            /* Get FRAG_OFF address from source mbuf */
                            frag_offAddr = (UINT8 *)
                                ixCryptoUtilMbufOffsetToAddressConvert (
                                pSrcMbuf,
                                6,
                                FALSE);
                                    
                            /* Restore FRAG_OFF value into source mbuf */        
                            memcpy (
                                frag_offAddr, 
                                &pQDesc->frag_off[0],
                                2 * sizeof(UINT8));

                            /* Get TTL address from source mbuf */
                            ttlAddr = (UINT8 *)
                                ixCryptoUtilMbufOffsetToAddressConvert (
                                pSrcMbuf,
                                8,
                                FALSE);
                                    
                            /* Restore TTL value into source mbuf */        
                            memcpy (
                                ttlAddr, 
                                &pQDesc->ttl,
                                sizeof(UINT8));

                            /* Get CHECK address from source mbuf */
                            checkAddr = (UINT8 *)
                                ixCryptoUtilMbufOffsetToAddressConvert (
                                pSrcMbuf,
                                10,
                                FALSE);
                                    
                            /* Restore CHECK value into source mbuf */        
                            memcpy (
                                checkAddr, 
                                &pQDesc->check[0],
                                2 * sizeof(UINT8));

/////////////////
                        } /* end of if (pQDesc->value.originalIcvOffset) */

                        if (useDiffBuf) /* If non in-place operation */
                        {

                            /* Restore ICV value into destination mbuf */
                            memcpy (
                                (IX_MBUF_MDATA (pDestMbuf) + 
                                pQDesc->value.originalIcvOffset), 
                                pQDesc->integrityCheckValue,
                                ixCryptoCtx[cryptoCtxId].digestLength);
                        }                        
                        break;
                        
                    /* HMAC generation only, or combined service with
                     * HMAC generation operation. Notes : the code for 
                     * HMAC generation fall thorugh to combined service
                     * IX_CRYPTO_NPE_OP_ENCRYPT_HMAC_GEN_ICV
                     */
                    case IX_CRYPTO_NPE_OP_HMAC_GEN_ICV:
                    case IX_CRYPTO_NPE_OP_ENCRYPT_HMAC_GEN_ICV:
                        /* Check if the destination mbuf is NULL */ 
                        /* If ICV in the middle of data */
                        if ((pQDesc->value.originalIcvOffset >= 
                            pQDesc->npeQDesc.authStartOffset) &&
                            (pQDesc->value.originalIcvOffset <
                            (pQDesc->npeQDesc.authStartOffset + 
                            pQDesc->npeQDesc.authLength)))
                        {
//////////////////
                            /* Get TOS address from source mbuf */
                            tosAddr = (UINT8 *)
                                ixCryptoUtilMbufOffsetToAddressConvert (
                                pSrcMbuf,
                                1,
                                FALSE);
                                    
                            /* Restore TOS value into source mbuf */        
                            memcpy (
                                tosAddr, 
                                &pQDesc->tos,
                                sizeof(UINT8));

                            /* Get FRAG_OFF address from source mbuf */
                            frag_offAddr = (UINT8 *)
                                ixCryptoUtilMbufOffsetToAddressConvert (
                                pSrcMbuf,
                                6,
                                FALSE);
                                    
                            /* Restore FRAG_OFF value into source mbuf */        
                            memcpy (
                                frag_offAddr, 
                                &pQDesc->frag_off[0],
                                2 * sizeof(UINT8));

                            /* Get TTL address from source mbuf */
                            ttlAddr = (UINT8 *)
                                ixCryptoUtilMbufOffsetToAddressConvert (
                                pSrcMbuf,
                                8,
                                FALSE);
                                    
                            /* Restore TTL value into source mbuf */        
                            memcpy (
                                ttlAddr, 
                                &pQDesc->ttl,
                                sizeof(UINT8));

                            /* Get CHECK address from source mbuf */
                            checkAddr = (UINT8 *)
                                ixCryptoUtilMbufOffsetToAddressConvert (
                                pSrcMbuf,
                                10,
                                FALSE);
                                    
                            /* Restore CHECK value into source mbuf */        
                            memcpy (
                                checkAddr, 
                                &pQDesc->check[0],
                                2 * sizeof(UINT8));

/////////////////
                        } /* end of if (pQDesc->value.originalIcvOffset) */


                        if (NULL != (IX_MBUF *) pQDesc->npeQDesc.destMbufAddr)
                        {
                            /* Convert mbuf header and its contents from NPE 
                             * format to host format
                             */
                            pDestMbuf 
                                = ixCryptoUtilDestMbufFromNpeFormatConvert 
                                    ((IX_MBUF *) pQDesc->npeQDesc.destMbufAddr,
                                    pQDesc->npeQDesc.authStartOffset);
                        } /* end of if (pQDesc->NpeQDesc.DestMbufAddr) */




                        break;
                   
                    /* Encrypt/Decrypt only operation */
                    default :
                        /* Check if the destination mbuf is NULL */ 
                        if (NULL != (IX_MBUF *) pQDesc->npeQDesc.destMbufAddr)
                        {
                            /* Convert mbuf header and its contents from NPE 
                             * format to host format
                             */
                            pDestMbuf 
                                = ixCryptoUtilDestMbufFromNpeFormatConvert 
                                    ((IX_MBUF *) pQDesc->npeQDesc.destMbufAddr,
                                    pQDesc->npeQDesc.cryptStartOffset);
                        } /* end of if (pQDesc->NpeQDesc.DestMbufAddr) */
                        break;
                } /* end of switch case (npeOperation) */                    

                /* Check Authentication status */
/*
printk("operationResult %x IX_CRYPTO_NPE_OP_STATUS_MASK %x result %x\n",
operationResult,
IX_CRYPTO_NPE_OP_STATUS_MASK,
(operationResult & IX_CRYPTO_NPE_OP_STATUS_MASK)
);
*/
                if (IX_CRYPTO_NPE_OP_AUTH_FAIL 
                    != (operationResult & IX_CRYPTO_NPE_OP_STATUS_MASK)) /* pass */
                {
                
                    /* Increment counter for number of requests completed
                     * successfully without error 
                     */
                    cryptoStats.cryptoSuccessCounter++;

                    /* Check NPE operation status flag to determine which
                     * callback function to be called
                     */
                    switch (pQDesc->operStatus)
                    {
                        /* Crypto Perform operation */
                        case IX_CRYPTO_OP_PERFORM:
                            /* Crypto Perform operation completed successfully */
                            status =  IX_CRYPTO_ACC_STATUS_SUCCESS;

                            /* Crypto Perform operation, call the perform 
                             * complete callback function
                             */

                            ixCryptoCtx[cryptoCtxId].performCallbackFn(
                                cryptoCtxId,
                                pSrcMbuf, 
                                pDestMbuf,
                                status);
                                        
                            break;
                
                            /* Crypto Register operation */
                        case IX_CRYPTO_OP_REGISTER:

                            /* If reverse AES key generated, need to copy the
                             * AES key into the crypto param in ixCryptoCtx for
                             * crypto perform service
                             */
                            if (IX_CRYPTO_ACC_REV_AES_KEY_VALID ==
                                (ixCryptoCtx[cryptoCtxId].validAndKeyId &
                                IX_CRYPTO_ACC_REV_AES_KEY_VALID))
                            {
                                /* Invalidate crypto context to get updated primary
                                 * and secondary chaining variables
                                 */
                                IX_ACC_DATA_CACHE_INVALIDATE (
                                    ixCryptoCtx[cryptoCtxId].pNpeCryptoParam,
                                    IX_CRYPTO_NPE_CRYPTO_PARAM_SIZE);

                                keyId = ixCryptoCtx[cryptoCtxId].validAndKeyId & MASK_8_BIT;

                                /* Invalidate key crypto param to get reverse
                                 * AES key
                                 */
                                IX_ACC_DATA_CACHE_INVALIDATE (
                                    ixKeyCryptoParam[keyId].pNpeCryptoParam,
                                    IX_CRYPTO_NPE_CRYPTO_PARAM_SIZE);

                                /* Copy rever AES key from key Crypto Param into
                                 * ixCryptoCtx
                                 */
                                memcpy (
                                    &((ixCryptoCtx[cryptoCtxId].pNpeCryptoParam)
                                    ->npeCryptoInfo[IX_CRYPTO_NPE_CRYPTO_PARAM_CRYPTO_CFG_WORD_LEN]),
                                    &((ixKeyCryptoParam[keyId].pNpeCryptoParam)
                                    ->npeCryptoInfo[IX_CRYPTO_NPE_CRYPTO_PARAM_CRYPTO_CFG_WORD_LEN]),
                                    ixKeyCryptoParam[keyId].keyLength);

                                /* Flush NPE Crypto Param structure from cache
                                 * into SDRAM
                                 */
                                IX_CRYPTO_DATA_CACHE_FLUSH (
	                                ixCryptoCtx[cryptoCtxId].pNpeCryptoParam,
	                                IX_CRYPTO_NPE_CRYPTO_PARAM_SIZE);

                                 /* Release key crypto param */
                                 status = ixCryptoCCDMgmtKeyCryptoParamRelease (
                                            keyId);
                            }

                            /* Crypto Register operation completed successfully */
                            status =  IX_CRYPTO_ACC_STATUS_SUCCESS;

                            /* Mark valid bit in Crypto Context to indicate the
                             * Crypto Context has been registered successfully
                             */
                            ixCryptoCtx[cryptoCtxId].valid = TRUE;

                            /* Crypto Register operation, call the register
                             * complete callback function
                             */
                            ixCryptoCtx[cryptoCtxId].registerCallbackFn(
                                cryptoCtxId,
                                pSrcMbuf,
                                status);
                                    
                            break;
                                 
                            /* Crypto Register Operation */
                        case IX_CRYPTO_OP_WAIT:
                                 
                            /* Crypto Register operation, part of the operation 
                             * completed only, wait for next completion 
                             * indication, and release the buffer
                             */
                            status = IX_CRYPTO_ACC_STATUS_WAIT;
                
                            /* Crypto Register operation, call the register 
                             * complete callback function
                             */
                            ixCryptoCtx[cryptoCtxId].registerCallbackFn(
                                cryptoCtxId,
                                pSrcMbuf,
                                status);

                            break;

                        /* Hash key generation operation */
                        case IX_CRYPTO_OP_HASH_GEN_KEY:

                            /* get keyId from QDesc */
                            keyId = pQDesc->value.keyId;

                            /* Hash key generation operation completed 
                             * successfully 
                             */
                            status =  IX_CRYPTO_ACC_STATUS_SUCCESS;

                            /* Hash key generation operation, call the hash key
                             * generation complete callback function
                             */
                            ixKeyCryptoParam[keyId].hashKeyCallbackFn(
                                keyId,
                                pSrcMbuf,
                                status);

                            /* Release key crypto param */
                            status = ixCryptoCCDMgmtKeyCryptoParamRelease (
                                         keyId);

                            break;

                            /* Default error case, the code should not reach this
                             * default case in normal operation
                             */
                        default:
                            status = IX_CRYPTO_ACC_STATUS_FAIL;

                            /* Log error message in debugging mode */
                            IX_CRYPTO_ACC_LOG(
                                LOG_ERROR,
                                "Wrong NPE Operation status for CryptoCtx %d\n",
                                cryptoCtxId,
                                0, 0, 0, 0, 0);
                                    
                            break;
                                                    
                    } /* end of switch case (pQDesc->OperStatus) */
                }
                else /* Authentication operation failed */
                {
                    status = IX_CRYPTO_ACC_STATUS_AUTH_FAIL;
                        
                    /* Increment counter for number of requests completed 
                     * but operation failed 
                     */
                    cryptoStats.cryptoFailCounter++;
                        
                    /* Logging error message in debugging mode */
                    IX_CRYPTO_ACC_LOG(
                        LOG_MESSAGE, 
                        "Authentication operation failed.\n",
                        0, 0, 0, 0, 0, 0);
                            
                    /* Call client's perform complete callback function,
                     * the callback function is registered during the Crypto 
                     * Context registration process.
                     * Note: Perform compleet callback is always called for the 
                     * case with AuthStatus == Q_DESC_OPERATION_FAIL as only the
                     * crypto perform request with operation authentication will
                     * have this status.
                     */
                    ixCryptoCtx[cryptoCtxId].performCallbackFn(
                        cryptoCtxId,
                        pSrcMbuf, 
                        pDestMbuf,
                        status);
                                  
                } /* end of if-else (AuthStatus) */
                    
                /* Release Q descriptor back to descriptor pool */
                if (IX_CRYPTO_ACC_STATUS_SUCCESS != 
                    ixCryptoDescMgmtQDescriptorRelease (pQDesc))
                {
                    /* Log error message in debugging mode */
                    IX_CRYPTO_ACC_LOG(
                        LOG_ERROR, 
                        "Q descriptor release failed for Crypto Ctx %d.\n",
                        cryptoCtxId,
                        0, 0, 0, 0, 0);
                }
                        
                /* Increment counter in the Crypto Context to keep track the 
                 * number of requests have been completed, log the number even 
                 * if operation failed
                 * Note : Only Increment the counter for crypto register and
                 *        crypto perform request, as hash key generation will
                 *        not have a valid cryptoCtxId
                 */
                if (IX_CRYPTO_OP_HASH_GEN_KEY != pQDesc->operStatus)
                {
                    ixCryptoCtx[cryptoCtxId].reqDoneCount++;
                }
            }
            else /* if pQDesc is NULL */
            {
                /* Increment counter for number of Q descriptor address received
                 * is NULL
                 */
                cryptoStats.qDescAddrInvalidCounter++;
                    
                /* Logging error message in debugging mode */
                IX_CRYPTO_ACC_LOG(LOG_ERROR, 
                    "Q desc read from Crypto Done Q is NULL\n",
                    0, 0, 0, 0, 0, 0);
            } /* end of if-else (pQDesc) */
        }
        else /* Else for if (qStatus != IX_SUCCESS) */
        {
            if (IX_QMGR_Q_UNDERFLOW == qStatus )
            {
                /* 
                 * Underflow. Only error if this first time in loop
                 * Log with statistics
                 */
                if (0 == messageCount)
                {
                    cryptoStats.qUnderflowCounter++;
                }
            }
            else /* queue read failed */
            {
                /* Logging error message in debugging mode */
                IX_CRYPTO_ACC_LOG(LOG_ERROR, 
                    "Queue read failed with status %d.\n",
                    qStatus,
                    0, 0, 0, 0, 0);
            }
        } /* End of if (qStatus != IX_SUCCESS) */        
    } while (IX_SUCCESS == qStatus);

} /* end of ixCryptoQAccessReqDoneQMCallback () function */


/**
 * @fn      ixCryptoQAccessQueueWrite
 * @brief   Write queue entry to hardware queue through QMgr
 * @note    There is no corresponding Read function, as no statistic
 *          collection for reads (inderflow is common occurance using
 *          mechanism used in callback).
 *
 */
IxCryptoAccStatus
ixCryptoQAccessQueueWrite (
    IxQMgrQId queueId,
    UINT32 *pQEntry)
{
    if (IX_SUCCESS == ixQMgrQWrite (queueId, pQEntry))
    {
        return IX_CRYPTO_ACC_STATUS_SUCCESS;
    }
    else /* queue overflow */
    {
	/* IX_CRYPTO_ACC_STATUS_QUEUE_FULL, increase the counter */
        cryptoStats.qOverflowCounter++;
        return IX_CRYPTO_ACC_STATUS_QUEUE_FULL;
    } /* end of if-else (ixQMgrWrite) */
} /* end of ixCryptoQAccessQueueWrite () function */



/**
 * @fn      ixCryptoQAccessQueueStatusShow
 * @brief   Show queue status
 *
 */
IxCryptoAccStatus
ixCryptoQAccessQueueStatusShow (IxQMgrQId QueueId)
{
    if (IX_SUCCESS == ixQMgrQShow (QueueId))
    {
        printf ("\n\nNumber of successful crypto request          : %d \n\n", 
                cryptoStats.cryptoSuccessCounter);
                
        printf ("Number of unsuccessful crypto request        : %d \n\n", 
                cryptoStats.cryptoFailCounter);
        
        printf ("Number of times Crypto Req Q overflow        : %d \n\n", 
                cryptoStats.qOverflowCounter);
        
        printf ("Number of times Crypto Done Q underflow      : %d \n\n", 
                cryptoStats.qUnderflowCounter);
                
        printf ("Number of times Q descriptor received is NULL: %d \n\n", 
                cryptoStats.qDescAddrInvalidCounter);
        
        return IX_CRYPTO_ACC_STATUS_SUCCESS;
    }
    else /* Q not configured */
    {
        return IX_CRYPTO_ACC_STATUS_FAIL;
    } /* end of if-else (ixQMgrShow) */
} /* end of ixCryptoQAccessQueueStatusShow () function*/
