/**
 * @file IxCryptoAccUtilities_p.h
 *
 * @date October-03-2002
 *
 * @brief Header file for the utilities module.
 *
 *
 * Design Notes:
 *
 * -- Intel Copyright Notice --
 * 
 * @par
 * Copyright 2002-2003 Intel Corporation All Rights Reserved.
 * 
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.
 * 
 * @par
 * The Material is protected by worldwide copyright and trade secret laws
 * and treaty provisions. No part of the Material may be used, copied,
 * reproduced, modified, published, uploaded, posted, transmitted,
 * distributed, or disclosed in any way except in accordance with the
 * applicable license agreement .
 * 
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel, except in accordance with the
 * applicable license agreement.
 * 
 * @par
 * Unless otherwise agreed by Intel in writing, you may not remove or
 * alter this notice or any other notice embedded in Materials by Intel
 * or Intel's suppliers or licensors in any way.
 * 
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * 
 * @par
 * -- End Intel Copyright Notice --
 */


#ifndef IXCRYPTOACCUTILITIES_P_H
#define IXCRYPTOACCUTILITIES_P_H

/*
 * Os/System dependancies.
 */
#include "IxTypes.h"
#include "IxOsServices.h"
#include "IxOsCacheMMU.h"
#include "IxOsBuffMgt.h"
#include "IxOsServicesMemAccess.h"
#include "IxCryptoAcc_p.h"

#ifdef __vxworks

#include <vxWorks.h>
#include <cacheLib.h>

#endif /* def __vxworks */


/**
 * @def IX_CRYPTO_CONVERT_WORD_TO_NETWORK_ORDER
 *
 * @brief Convert word into network order (Big Endian).
 *
 * @param UINT32 [inout] data - word to be converted.
 *
 * @return: none
 *
 */ 
#define IX_CRYPTO_CONVERT_WORD_TO_NETWORK_ORDER(data) \
        IX_OSSERV_SWAP_NPE_SHARED_LONG(data)


/**
 * @def IX_CRYPTO_CONVERT_WORD_TO_HOST_ORDER
 *
 * @brief Convert word into host order (Little Endian or Big Endian). 
 *
 * @param UINT32 [inout] data - words to be converted.
 *
 * @return none
 *
 */
#define IX_CRYPTO_CONVERT_WORD_TO_HOST_ORDER(data) \
        IX_OSSERV_SWAP_NPE_SHARED_LONG(data)

/**
 * @def IX_CRYPTO_CONVERT_SHORT_TO_NETWORK_ORDER
 *
 * @brief Convert short word into network order (Big Endian).
 *
 * @param UINT16 [inout] data - short word to be converted.
 *
 * @return: none
 *
 */ 
#define IX_CRYPTO_CONVERT_SHORT_TO_NETWORK_ORDER(data) \
        IX_OSSERV_SWAP_NPE_SHARED_SHORT(data) 

/**
 * @def IX_CRYPTO_CONVERT_SHORT_TO_HOST_ORDER
 *
 * @brief Convert short word into host order (Little Endian or Big Endian). 
 *
 * @param UINT16 [inout] data - short word to be converted.
 *
 * @return none
 *
 */
#define IX_CRYPTO_CONVERT_SHORT_TO_HOST_ORDER(data) \
        IX_OSSERV_SWAP_NPE_SHARED_SHORT(data)

/**
 * @def IX_CRYPTO_VIRTUAL_TO_PHYSICAL_TRANSLATION
 *
 * @brief To convert logical/virtual address to physical address. It will
 *        return a virtual address for the provided  physical address.
 *
 * @note Example of usage x = IX_CRYPTO_VIRTUAL_TO_PHYSICAL_TRANSLATION(x)
 * 
 * @param INT32 [in] address - virtual address to be translated.
 *
 * @return virtual address for the provided  physical address.
 *
 */
#define IX_CRYPTO_VIRTUAL_TO_PHYSICAL_TRANSLATION(address)\
             IX_MMU_VIRTUAL_TO_PHYSICAL_TRANSLATION(address)


/**
 * @def IX_CRYPTO_PHYSICAL_TO_VIRTUAL_TRANSLATION
 *
 * @brief To convert physical address to logical/virtual address. It will
 *        return a physical address for the provided virtual.
 *
 * @note Example of usage: x = IX_CRYPTO_PHYSICAL_TO_VIRTUAL_TRANSLATION(x)
 *
 * @param UINT32 [in] address - physical address to be translated.
 *
 * @return physical address for the provided virtual address
 *
 */
#define IX_CRYPTO_PHYSICAL_TO_VIRTUAL_TRANSLATION(address)\
             IX_MMU_PHYSICAL_TO_VIRTUAL_TRANSLATION (address)


/**
 * @def IX_CRYPTO_ACC_DRV_DMA_MALLOC
 *
 * @brief To allocate memory for driver use, that will be shared between
 *        XSCale and NPE.
 *
 * @note Example of usage: x = IX_CRYPTO_ACC_DRV_DMA_MALLOC(x)
 *
 * @param UINT32 [in] size - memory size in bytes to be allocated.
 *
 * @return  void * Pointer to memory that can be used between XScale and NPE's.
 *
 */
#define IX_CRYPTO_ACC_DRV_DMA_MALLOC(size)\
             IX_ACC_DRV_DMA_MALLOC(size)
             

/**
 * @def IX_CRYPTO_ACC_DRV_DMA_FREE
 *
 * @brief To free the memory allocated through IX_CRYPTO_ACC_DRV_DMA_MALLOC
 *        function.
 *
 * @note Example of usage: IX_CRYPTO_ACC_DRV_DMA_FREE(pData, size)
 *
 * @param UINT8 [in] *pData - memory block pointer to be freed.
 * @param UINT32 [in] size -  number of bytes of memory allocated.
 *
 * @return None
 *
 */
#define IX_CRYPTO_ACC_DRV_DMA_FREE(pData, size)\
             IX_ACC_DRV_DMA_FREE(pData,size)

             
/**
 * @def IX_CRYPTO_DATA_CACHE_INVALIDATE
 *
 * @brief To invalidate a cache range.
 *
 * @param UINT32 [in] address - cache address to be invalidated.
 * @param UINT32 [in] size - the size in bytes of cache to be invalidated.
 *
 * @return none
 *
 */
#define IX_CRYPTO_DATA_CACHE_INVALIDATE(address, size)\
             IX_ACC_DATA_CACHE_INVALIDATE(address, size)


/**
 * @def IX_CRYPTO_DATA_CACHE_FLUSH
 *
 * @brief To flush a cache range to physical memory. 
 *
 * @param UINT32 [in] address - cache address to be flushed to the physical 
 *	  memory.
 * @param UINT32 [in] size - the size in bytes of cache to be flushed to the 
 *	  physical memory.
 *
 * @return none
 *
 */
#define IX_CRYPTO_DATA_CACHE_FLUSH(address, size)\
             IX_ACC_DATA_CACHE_FLUSH(address, size)



/**
 * @fn ixCryptoUtilMbufOffsetToAddressConvert
 *
 * @brief To convert offset to data block of mbuf into 32-bit address.
 *
 * @param IX_MBUF [in] *pMbuf - pointer to the mbuf.
 * @param INT32 [in] offset - offset of the mbuf.
 * @param BOOL [in] useDiffBuf - TRUE if non in-place operation
 *                               FALSE if in-place operation 
 *
 * @return UINT32 address of data pointer in mbuf.
 *
 * Reentrant    : no
 * ISR Callable : no
 *
 */
UINT32 
ixCryptoUtilMbufOffsetToAddressConvert (
    IX_MBUF *pMbuf,
    INT32 offset,
    BOOL useDiffBuf);
                                        

/**
 * @fn ixCryptoUtilMbufToNpeFormatConvert
 *
 * @brief To convert chained mbuf to NPE format, including endianess
 *        conversion (to Big Endian) and virtual address to physical address
 *        conversion.  And then flush the data to physical memory.
 *
 * @param IX_MBUF [in] *pMbuf - pointer to the mbuf data block.
 *
 * @return IX_MBUF* - the converted mbuf is returned.
 *
 * Reentrant    : no
 * ISR Callable : no
 *
 */
IX_MBUF*
ixCryptoUtilMbufToNpeFormatConvert (IX_MBUF *pMbuf);


/**
 * @fn ixCryptoUtilMbufFromNpeFormatConvert
 *
 * @brief To convert chained mbuf from NPE format, including endianess
 *        conversion to host endianess and physical address to virtual 
 *        address conversion. And invalidate the cache.
 *
 * @param IX_MBUF [in] *pMbuf - pointer to the mbuf data block.
 *
 * @return IX_MBUF* - the converted mbuf is returned.
 *
 * Reentrant    : no
 * ISR Callable : no
 *
 */
IX_MBUF*
ixCryptoUtilMbufFromNpeFormatConvert (IX_MBUF *pMbuf);


/**
 * @fn ixCryptoUtilDestMbufFromNpeFormatConvert
 *
 * @brief To convert destination mbuf from NPE format, including endianess
 *        conversion to host endianess and physical address to virtual 
 *        address conversion. Shift the mData pointer back to its original 
 *        location according to the shift offset and invalidate the cache. 
 *
 * @param IX_MBUF [in] *pMbuf - pointer to the mbuf data block.
 * @param UINT16 [in] mDataShiftOffset - shift offset value for mData pointer
 *
 * @return IX_MBUF* - the converted mbuf is returned.
 *
 * Reentrant    : no
 * ISR Callable : no
 *
 */
IX_MBUF*
ixCryptoUtilDestMbufFromNpeFormatConvert (
    IX_MBUF *pMbuf,
    UINT16 mDataShiftOffset);


/**
 * @fn ixCryptoUtilNpeCryptCfgGenerate
 *
 * @brief To generate Npe Crypt Cfg word that will be written to NPE.  
 *
 * @param IxCryptoNpeCryptCfgWord [in] *pNpeCryptCfg - pointer to the 
 *        IxCryptoNpeCryptCfgWord struct.
 * @param UINT32 [out] *pNpeCryptCfgWord - NPE Crypt Cfg Word
 *
 * Reentrant    : no
 * ISR Callable : no
 *
 */
void
ixCryptoUtilNpeCryptCfgGenerate (
    UINT32 *pNpeCryptCfgWord, 
    IxCryptoNpeCryptCfgWord *pNpeCryptCfg);


/**
 * @fn ixCryptoUtilNpeHashCfgGenerate
 *
 * @brief To generate Npe Hash Cfg word that will be written to NPE.  
 *
 * @param IxCryptoNpeHashCfgWord [in] *pNpeHashCfg - pointer to the 
 *        IxCryptoNpeHashCfgWord struct.
 * @param UINT32 [out] *pNpeHashCfgWord - NPE Hash Cfg Word
 *
 * Reentrant    : no
 * ISR Callable : no
 *
 */
void
ixCryptoUtilNpeHashCfgGenerate (
    UINT32 *pNpeHashCfgWord, 
    IxCryptoNpeHashCfgWord *pNpeHashCfg);


/**
 * @def IX_CRYPTO_ACC_LOCK
 *
 * @brief Locks out IRQs
 *
 * @return INT32 lockVal - lock key used by IX_CRYPTO_ACC_UNLOCK to unlock 
 *         IRQs later
 *
 */
#define IX_CRYPTO_ACC_LOCK(lockVal) lockVal = ixOsServIntLock();


/**
 * @def IX_CRYPTO_ACC_UNLOCK
 *
 * @brief Unlock IRQs
 *
 * @param INT32 [in] lockVal - lock key previously obtained with 
 *        IX_CRYPTO_ACC_LOCK
 *
 * @return None
 *
 */
#define IX_CRYPTO_ACC_UNLOCK(lockVal) ixOsServIntUnlock(lockVal);


/**
 * @def IX_CRYPTO_ACC_LOG
 *
 * @brief logs a formatted message
 *
 * @param int [in] level - trace level.
 * @param char [in] format - format string, similar to printf().
 * @param int [in] a - first argument to display.
 * @param int [in] b - second argument to display.
 * @param int [in] c - third argument to display.
 * @param int [in] d - fourth argument to display.
 * @param int [in] e - fifth argument to display.
 * @param int [in] f - sixth argument to display.
 *
 * @return none
 *
 */
/* #ifndef NDEBUG */
/* Jack ++ */
#if 0
#define IX_CRYPTO_ACC_LOG(level, str, a, b, c, d, e, f)\
             (ixOsServLog (level, str, a, b, c, d, e, f))
#else
#define IX_CRYPTO_ACC_LOG(level, str, a, b, c, d, e, f) /* do nothing */
#endif /* def NDEBUG */


#endif /* ndef IXCRYPTOACCUTILITIES_P_H */
