/**
 * @file   IxCryptoAcc.h
 *
 * @date   2-OCT-2002
 *
 * @brief   Header file for the IXP425 Security component (IxCryptoAcc)
 *
 * -- Intel Copyright Notice --
 * 
 * @par
 * Copyright 2002-2003 Intel Corporation All Rights Reserved.
 * 
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.
 * 
 * @par
 * The Material is protected by worldwide copyright and trade secret laws
 * and treaty provisions. No part of the Material may be used, copied,
 * reproduced, modified, published, uploaded, posted, transmitted,
 * distributed, or disclosed in any way except in accordance with the
 * applicable license agreement .
 * 
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel, except in accordance with the
 * applicable license agreement.
 * 
 * @par
 * Unless otherwise agreed by Intel in writing, you may not remove or
 * alter this notice or any other notice embedded in Materials by Intel
 * or Intel's suppliers or licensors in any way.
 * 
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * 
 * @par
 * -- End Intel Copyright Notice --
 */


/**
 * @defgroup IxCryptoAcc IXP425 Security (IxCryptoAcc) API
 *
 * @brief IXP425 Security component Public API
 *
 * @{
 */

#ifndef IXCRYPTOACC_H
#define IXCRYPTOACC_H

/*
 * Put the user defined include files required
 */

#include "IxOsBuffMgt.h"
#include "IxTypes.h"

/* Jack ++ 2003/09/26 */
#include <linux/interrupt.h>
#include <linux/spinlock.h> 

/*
 * #defines and macros used in this file.
 */

#define IX_CRYPTO_ACC_MAX_CIPHER_KEY_LENGTH 32  /**<
                                                 * Max length (byte) of cipher
                                                 * key for DES (64 bit),
                                                 * 3DES (192 bit), AES (128, 192
                                                 * & 256 bit).
                                                 */
#define IX_CRYPTO_ACC_MAX_CIPHER_IV_LENGTH 16   /**<
                                                 * Max IV length in byte
                                                 */
#define IX_CRYPTO_ACC_MAX_AUTH_KEY_LENGTH 64    /**<
                                                 * Max length (byte) of
                                                 * authentication key for
                                                 * SHA1 and MD5.
                                                 * Key size > 64 bytes need to 
                                                 * be hashed to produce 
                                                 * shorter key by calling API 
                                                 * ixCryptoAccHashKeyGenerate().
                                                 * (SHA1 : 20 bytes, MD5 : 16 
                                                 * bytes by default according 
                                                 * to RFC2104.       
                                                 * If key size <= 64 bytes,
                                                 * authentication key can be 
                                                 * used for registration 
                                                 * directly.
                                                 */
#define IX_CRYPTO_ACC_MAX_AUTH_IV_LENGTH 20     /**<
                                                 * Max length (byte) of initial
                                                 * chaining variable for
                                                 * SHA1 (160 bit) and
                                                 * MD5 (128 bit).
                                                 */
#define  IX_CRYPTO_ACC_MAX_QUEUE_DEPTH 64       /**<
                                                 * Max queue depth supported
                                                 * by the Queue Manager.
                                                 */
#define  IX_CRYPTO_ACC_MAX_ACTIVE_SA_TUNNELS 1000  /**<
                                                    * Maximum active tunnels
                                                    * supported could be changed
                                                    * by the client based on the
                                                    * application's requirements.
                                                    * Number of active tunnels
                                                    * will not impact the
                                                    * performance but will have
                                                    * an impact on the memory
                                                    * needed to keep the
                                                    * crypto context information
                                                    * will depend on it. Overall
                                                    * memory requirement depends
                                                    * on the number of tunnels.
                                                    */


/*
 * Constants value for DES algorithm in unit bytes
 */

#define IX_CRYPTO_ACC_DES_KEY_64     8  /**<
                                         * DES key length in bytes
                                         */
#define IX_CRYPTO_ACC_DES_BLOCK_64   8  /**<
                                         * DES cipher block length in bytes
                                         */
#define IX_CRYPTO_ACC_DES_IV_64      8  /**<
                                         * DES initialization vector length 
                                         * in bytes
                                         */
#define IX_CRYPTO_ACC_3DES_KEY_192   24 /**<
                                         * 3DES key length in bytes
                                         */


/*
 * Constants value for AES algorithm in unit bytes
 */

#define IX_CRYPTO_ACC_AES_KEY_128    16 /**<
                                         * AES-128 key length in bytes
                                         */
#define IX_CRYPTO_ACC_AES_KEY_192    24 /**<
                                         * AES-192 key length in bytes
                                         */
#define IX_CRYPTO_ACC_AES_KEY_256    32 /**<
                                         * AES-256 key length in bytes
                                         */
#define IX_CRYPTO_ACC_AES_BLOCK_128  16 /**<
                                         * AES cipher block length in bytes
                                         */
#define IX_CRYPTO_ACC_AES_CBC_IV_128 16 /**<
                                         * AES initialization vector length 
                                         * in bytes for CBC mode
                                         */
#define IX_CRYPTO_ACC_AES_CTR_IV_64  8  /**<
                                         * AES initialization vector length 
                                         * in bytes for CTR mode
                                         */



/*
 * Constants value for SHA1 algorithm in unit bytes
 */

#define IX_CRYPTO_ACC_SHA1_KEY_160    20    /**<
                                             * SHA1 key length in bytes
                                             */
#define IX_CRYPTO_ACC_SHA1_DIGEST_160 20    /**< 
                                             * SHA1 message digest length
                                             * in bytes
                                             */


/*
 * Constants value for MD5 algorithm in unit bytes
 */

#define IX_CRYPTO_ACC_MD5_KEY_128     16    /**<
                                             * MD5 key length in bytes
                                             */
#define IX_CRYPTO_ACC_MD5_DIGEST_128  16    /**< 
                                             * MD5 message digest length
                                             * in bytes
                                             */


 
/*
 * Typedefs whose scope is limited to this file.
 */

/**
 * @enum IxCryptoAccOperation
 *
 * @brief Cryptographic Operation Definitions  
 *
 */ 
typedef enum 
{
    IX_CRYPTO_ACC_OP_ENCRYPT = 0,       /**< Encrypt operation */
    IX_CRYPTO_ACC_OP_DECRYPT,           /**< Decrypt operation */
    IX_CRYPTO_ACC_OP_AUTH_CALC,         /**< 
                                         * Authentication calculation 
                                         * operation
                                         */
    IX_CRYPTO_ACC_OP_AUTH_CHECK,        /**<
                                         * Authentication verification 
                                         * operation
                                         */
    IX_CRYPTO_ACC_OP_ENCRYPT_AUTH,      /**<
                                         * Encryption followed by 
                                         * authentication calculation 
                                         * operation
                                         */
    IX_CRYPTO_ACC_OP_AUTH_DECRYPT,      /**<
                                         * Authentication verification 
                                         * followed by decryption
                                         */
    IX_CRYPTO_ACC_OP_TYPE_OF_OPERATION  /**< 
                                         * Maximum value for types of
                                         * operation
                                         */
} IxCryptoAccOperation;



/**
 * @enum IxCryptoAccCipherAlgo
 *
 * @brief Cipher Algorithm Definitions  
 *
 * @note 3DES and AES will be supported if not violating import/export rules.
 */
typedef enum
{
    IX_CRYPTO_ACC_CIPHER_NULL = 0,    /**< NULL encryption */
    IX_CRYPTO_ACC_CIPHER_DES,         /**< DES algorithm */
    IX_CRYPTO_ACC_CIPHER_3DES,        /**< Triple DES algorithm */
    IX_CRYPTO_ACC_CIPHER_AES,         /**< AES algorithm */
    IX_CRYPTO_ACC_CIPHER_ALGO_TYPE    /**< 
                                       * Maximum value for types of cipher 
                                       * algorithm
                                       */
} IxCryptoAccCipherAlgo;



/**
 * @enum IxCryptoAccCipherMode
 *
 * @brief Cipher Mode Definitions  
 *
 * @note CFB and OFB are not supported.
 */
typedef enum
{
    IX_CRYPTO_ACC_MODE_NULL = 0,    /**< NULL cipher mode */
    IX_CRYPTO_ACC_MODE_ECB,         /**< ECB mode of operation */
    IX_CRYPTO_ACC_MODE_CBC,         /**< CBC mode of operation */
    IX_CRYPTO_ACC_MODE_CTR,         /**< 
                                     * CTR mode of operation, only 
                                     * applicable to AES
                                     */
    IX_CRYPTO_ACC_MODE_TYPE         /**< 
                                     * Maximum value for types of 
                                     * operation 
                                     */
} IxCryptoAccCipherMode;



/**
 * @enum IxCryptoAccAuthAlgo
 *
 * @brief Authentication Algorithm Definitions  
 *
 * @note Only two authentication algorithms are supported, SHA1 and MD5.
 */
typedef enum
{
    IX_CRYPTO_ACC_AUTH_NULL = 0,    /**< NULL authentication */
    IX_CRYPTO_ACC_AUTH_SHA1,        /**< SHA1 algorithm */
    IX_CRYPTO_ACC_AUTH_MD5,         /**< MD5 algorithm */
    IX_CRYPTO_ACC_AUTH_TYPE         /**< 
                                     * Maximum value for types of 
                                     * authentication algorithm
                                     */
} IxCryptoAccAuthAlgo;



/**
 * @enum IxCryptoAccStatus
 *
 * @brief Status Definitions  
 *
 * @note These status will be used by the APIs to return to the client.
 */
typedef enum
{
    IX_CRYPTO_ACC_STATUS_SUCCESS = IX_SUCCESS,      /**< Success status */
    IX_CRYPTO_ACC_STATUS_FAIL = IX_FAIL,            /**< Fail status */
    IX_CRYPTO_ACC_STATUS_WAIT,                      /**< Wait status */
    IX_CRYPTO_ACC_STATUS_RETRY,                     /**< Retry status */
    IX_CRYPTO_ACC_STATUS_QUEUE_FULL,                /**< Queue full */
    IX_CRYPTO_ACC_STATUS_OPERATION_NOT_SUPPORTED,   /**< Invalid operation */
    IX_CRYPTO_ACC_STATUS_CIPHER_ALGO_NOT_SUPPORTED, /**<
                                                     * Invalid cipher algorithm
                                                     */
    IX_CRYPTO_ACC_STATUS_CIPHER_MODE_NOT_SUPPORTED, /**< 
                                                     * Invalid cipher mode of
                                                     * operation
                                                     */
    IX_CRYPTO_ACC_STATUS_CIPHER_INVALID_KEY_LEN,    /**<
                                                     * Invalid cipher key
                                                     *length
                                                     */
    IX_CRYPTO_ACC_STATUS_CIPHER_INVALID_IV_LEN,     /**< 
                                                     * Invalid IV length
                                                     */
    IX_CRYPTO_ACC_STATUS_CIPHER_INVALID_BLOCK_LEN,  /**<
                                                     * Invalid cipher block
                                                     * length
                                                     */
    IX_CRYPTO_ACC_STATUS_AUTH_ALGO_NOT_SUPPORTED,   /**< 
                                                     * Invalid authentication 
                                                     * algorithm
                                                     */
    IX_CRYPTO_ACC_STATUS_AUTH_INVALID_DIGEST_LEN,   /**<
                                                     * Invalid message digest
                                                     * length
                                                     */
    IX_CRYPTO_ACC_STATUS_AUTH_INVALID_KEY_LEN,      /**< 
                                                     * Invalid authentication 
                                                     * key length 
                                                     */
    IX_CRYPTO_ACC_STATUS_AUTH_FAIL,                 /**<  
                                                     * Authentication 
                                                     * verification failed 
                                                     */
    IX_CRYPTO_ACC_STATUS_CRYPTO_CTX_NOT_VALID,      /**< 
                                                     * Invalid crypto 
                                                     * context ID
                                                     */
    IX_CRYPTO_ACC_STATUS_EXCEED_MAX_TUNNELS         /**< 
                                                     * Exceed maximum number 
                                                     * of crypto contexts 
                                                     * allocation
                                                     */
} IxCryptoAccStatus;



/**
 * @brief   Structure storing cipher configuration parameters required to
 *          perform security functionality.
 *
 * Structure is used to specify the cipher context required for using the 
 * @ref IxCryptoAcc interface.
 */ 
typedef struct
{
    IxCryptoAccCipherAlgo cipherAlgo; /**< Cipher Algorithm - DES, 3DES, 
                                       * AES
                                       */
    IxCryptoAccCipherMode cipherMode; /**< Cipher mode - ECB, CBC, CTR */
    UINT32 cipherKeyLen;  /**< Cipher key length in bytes */
    union 
    {
        UINT8 cipherKey[IX_CRYPTO_ACC_MAX_CIPHER_KEY_LENGTH]; /**< default
                                                               * key array
                                                               */ 
        UINT8 desKey[IX_CRYPTO_ACC_DES_KEY_64];         /**< DES key */
        UINT8 tripleDesKey[IX_CRYPTO_ACC_3DES_KEY_192]; /**< 3DES key */
        UINT8 aesKey128[IX_CRYPTO_ACC_AES_KEY_128];     /**< AES-128 key */
        UINT8 aesKey192[IX_CRYPTO_ACC_AES_KEY_192];     /**< AES-192 key */
        UINT8 aesKey256[IX_CRYPTO_ACC_AES_KEY_256];     /**< AES-256 key */
    } key;                          /**< Cipher key */
    UINT32 cipherBlockLen;          /**< 
                                     * Cipher block length in bytes.
                                     * (DES/3DES - 8 bytes, AES - 16 bytes)
                                     */
    UINT32 cipherInitialVectorLen;  /**< Length of IV in bytes */
    UINT32 securityParameterIndex;  /**< 
                                     * SPI is used in generating counter 
                                     * block for AES CTR mode.
                                     */       
} IxCryptoAccCipherCtx;



/**
 * @brief   Structure storing authentication configuration parameters required 
 *          to perform security functionality.
 *
 * Structure is used to specify the authentication context required for using  
 * the @ref IxCryptoAcc interface.
 */
typedef struct
{
    IxCryptoAccAuthAlgo authAlgo; /**< authentication algorithm -
                                   * MD5, SHA1
                                   */
    UINT32 authDigestLen; /**< Digest length in bytes */
    UINT32 authKeyLen;    /**< Hash key length in bytes */
    union   
    {
        UINT8 authKey[IX_CRYPTO_ACC_MAX_AUTH_KEY_LENGTH]; /**< default hash
                                                           * key array
                                                           */
        UINT8 sha1Key[IX_CRYPTO_ACC_MAX_AUTH_KEY_LENGTH]; /**< SHA1 key */
        UINT8 md5Key[IX_CRYPTO_ACC_MAX_AUTH_KEY_LENGTH];  /**< MD5 key */
    } key; /**< Hash key */      
} IxCryptoAccAuthCtx;   



/**
 * @brief   Structure storing configuration parameters required to
 *          perform security functionality.
 *
 * Structure is used to specify the crypto context (hardware accelerator
 * context) required for using the @ref IxCryptoAcc interface.
 *
 */ 

typedef struct
{
    IxCryptoAccOperation operation;     /**< Types of operation */
    IxCryptoAccCipherCtx cipherCtx;     /**< Cipher context */
    IxCryptoAccAuthCtx authCtx;         /**< Authentication context */
    BOOL useDifferentSrcAndDestMbufs;   /**< 
                                         *  If TRUE, data is read from srcMbuf, 
                                         *  result is written to destMbuf
                                         *  (non in-place operation). 
                                         *  If FALSE data is read from srcMbuf, 
                                         *  and written back to srcMbuf 
                                         *  (in-place operation). 
                                         *  Default is FALSE. Note that only
                                         *  the crypted/authenticated data that 
                                         *  is copied, not the entire source 
                                         *  mbuf.
                                         */ 
/* Jack ++ */
	spinlock_t crypto_lock;
	UINT8 ivec[IX_CRYPTO_ACC_MAX_CIPHER_IV_LENGTH];
        UINT16 authStartOffset;
        UINT16 authDataLen;
        UINT16 cryptStartOffset;
        UINT16 cryptDataLen;
        UINT16 icvOffset;
} IxCryptoAccCtx;



/**
 *
 * @ingroup IxCryptoAcc
 *
 * @brief  Cryptographic Context registration complete callback notification.
 *
 * This function is called to notify a client that the Crypto Context has 
 * been registered. This function returns status through the associated 
 * cryptoCtxId once the initial values needed are calculated by Network 
 * Processor Engine (NPE) and stored in Cryptographic Context Database (CCD).  
 *
 * The CryptoCtxId is valid until ixCryptoAccCtxUnregister is invoked.
 * 
 * If the callback function returns the IX_CRYPTO_ACC_STATUS_WAIT status, it 
 * indicates that registration is not complete yet, but the mbuf pointer 
 * needs to be freed by client. Client needs to wait for the next
 * completion indication. Registration complete successfully only if status 
 * IX_CRYPTO_ACC_STATUS_SUCCESS is received.
 *
 * @param   cryptoCtxId [in] This crypto context ID is provided when client 
 *          sends in request via ixCryptoAccCtxRegister API to register crypto   
 *          context. cryptoCtxId points to a struct consists of cryptographic 
 *          paramaters required by the Network Processor Engine (NPE) in CCD 
 *          database.
 * @param   *pMbuf [in] Pointer to the mbuf (to be freed by 
 *          client). The client should free any mbuf that is not NULL.
 * @param   status [in] Status (IxCryptoAccStatus) reporting to client.
 *          
 * @note @li   IX_CRYPTO_ACC_STATUS_SUCCESS - registration is successful.
 * @note @li   IX_CRYPTO_ACC_STATUS_FAIL - registration failed.
 * @note @li   IX_CRYPTO_ACC_STATUS_WAIT - registration is not complete yet, 
 *             wait for next completion indication. NPE is busy in 
 *             calculating the initial variables needed for the registration.
 *
 */
typedef void (*IxCryptoAccRegisterCompleteCallback)(
    UINT32 cryptoCtxId,
    IX_MBUF *pMbuf, 
    IxCryptoAccStatus status);


/**
 *
 * @ingroup IxCryptoAcc
 *
 * @brief  Hash key generation complete callback notification.
 *
 * This function is called to notify a client that the hash key has been 
 * generated. This function will return the status through the associated 
 * hashKeyId  once the key is calculated by Network Processor Engine (NPE).
 * 
 * hashKeyId becomes invalid once the notification callback is called.
 *
 * @param   hashKeyId [in] This hashKeyId is provided when client sends in
 *          request to hash the key. 
 * @param   *pMbufHashKey [in] Pointer to the mbuf that contains original key
 *          and generated hash key. Client will need to copy the generated 
 *          hash key from the mbuf into crypto context and used it as hash 
 *          key for crypto registration request. 
 *          This mbuf will be freed by client. 
 * @param   status [in] Status (IxCryptoAccStatus) reporting to client.
 *          
 * @note @li   IX_CRYPTO_ACC_STATUS_SUCCESS - hash key generation is successful.
 * @note @li   IX_CRYPTO_ACC_STATUS_FAIL - hash key generation failed.
 *
 */
typedef void (*IxCryptoAccHashKeyGenCompleteCallback)(
    UINT32 hashKeyId,
    IX_MBUF *pMbufHashKey, 
    IxCryptoAccStatus status);
 
 
 /**
 *
 * @ingroup IxCryptoAcc
 *
 * @brief  Harware accelerator service request complete callback notification.
 *
 * This function is called to notify a client that the cryptographic transaction
 * has been completed. The cryptoCtxId and status of completed operation are 
 * returned to the client through this callback function to indicate operation
 * which crypto context has been completed. The CryptoCtxId is obtained via 
 * ixCryptoAccCtxRegister, and this ID is unique to a particular IPSec 
 * tunnel.
 *
 * @param   cryptoCtxId [in] This crypto context ID is provided when client 
 *          sends in request via ixCryptoAccCtxRegister API to register crypto   
 *          context. cryptoCtxId points to a struct consists of cryptographic 
 *          paramaters required by the Network Processor Engine (NPE) in CCD 
 *          database.
 * @param   *pSrcMbuf [in] Pointer to the source mbuf which contains the data    
 *          to be processed. It is also the output mbuf which contains the  
 *          processed data if UseDifferentSrcAndDestMbufs is FALSE.
 * @param   *pDestMbuf [in] Only used if UseDifferentSrcAndDestMbufs  
 *          is TRUE. Pointer to the output mbuf which contains processed data.
 * @param   status [in] Status reporting to the client 
 *          via IxCryptoAccStatus.
 *
 * @note  @li  IX_CRYPTO_ACC_STATUS_SUCCESS - Operation is completed successfully.
 * @note  @li  IX_CRYPTO_ACC_STATUS_FAIL - Operation failed.
 * @note  @li  IX_CRYPTO_ACC_STATUS_AUTH_FAIL - Authentication is unsuccessful.
 *             Note that when authentication fails, the content of the 
 *             destination mbuf (which is the same as the source mbuf if a 
 *             normal in-place operation is performed) will be undetermined.
 *
 */
typedef void (*IxCryptoAccPerformCompleteCallback)( 
    UINT32 cryptoCtxId,
    IX_MBUF *pSrcMbuf, 
    IX_MBUF *pDestMbuf,
    IxCryptoAccStatus status); 



/**
 *
 * @ingroup IxCryptoAcc
 *
 * @fn IxCryptoAccStatus ixCryptoAccInit (void)
 *
 * @brief Initialise the Security Access component
 *
 * This API will initialize all the internal resources of IxCryptoAcc access 
 * component. It will  check whether the specified AES or DES coprocessor is 
 * present and issue a warning if not. This API must be called before any other 
 * API call on this component can be made. This API is called once only.
 *
 * @param None
 *
 * @return @li  IX_CRYPTO_ACC_STATUS_SUCCESS - successfully initialize the 
 *              component; a warning is issued if coprocessor is not present.
 * @return @li  IX_CRYPTO_ACC_STATUS_FAIL - Initialization failed for some 
 *              unspecified internal reasons.
 *              
 * @li   Reentrant    : no
 * @li   ISR Callable : no
 *
 */
PUBLIC IxCryptoAccStatus
ixCryptoAccInit (void);



/**
 *
 * @ingroup IxCryptoAcc
 *
 * @fn IxCryptoAccStatus ixCryptoAccCtxRegister (
    IxCryptoAccCtx *pAccCtx,
    IX_MBUF *pMbufPrimaryChainVar,
    IX_MBUF *pMbufSecondaryChainVar,
    IxCryptoAccRegisterCompleteCallback registerCallbackFn,
    IxCryptoAccPerformCompleteCallback performCallbackFn,
    UINT32 *pCryptoCtxId)
 *
 * @brief Crypto context registration. Cryptographic Context ID (cryptoCtxId)
 *        for the registered crypto context obtained from this registration
 *        request will be used in ixCryptoAccAuthCryptPerform requests.  
 *
 * This function is used to register all the required information 
 * (eg. key, cipher algorithm, etc) for hardware accelerator services with 
 * IxCryptoAcc component. Those information will be stored in CCD database.  
 * All the information will be converted into a cryptographic parameters 
 * structure to be shared with NPE. The structure will be associated with 
 * a unique crypto context ID (cryptoCtxId). cryptoCtxId is passed back to 
 * client for future reference in callback function and also for hardware 
 * accelerator service requests. 
 * Besides, two empty mbuf are required to be passed in through interface for 
 * the use of access component and NPE to compute the primary and secondary 
 * initial chaining variables. pMbufPrimaryChainVar and pMbufSecondaryChainVar 
 * mbuf pointers must be NULL if authentication or combined service is not 
 * selected. 
 *          
 * @note
 * 2 scenarios of crypto context registration depending on authentication key 
 * size (in bytes) :
 * 1. If (L <= key size <= 64), then call this API (ixCryptoAccCtxRegister)
 *    directly. L = 16 for MD5 and L = 20 for SHA1. 
 * 2. If (key size > 64), then authentication key needs to be hashed to become
 *    shorter key first by calling another API ixCryptoAccHashKeyGenerate ().
 *    Please follow steps below for this case
 *    - Call ixCryptoAccHashKeyGenerate () to hash the authentication key
 *    - Wait for the callback from ixCryptoAccHashKeyGenerate (). Copy 
 *      generated authentication key from mbuf into IxCryptoAccCtx as 
 *      authentication key.
 *    - Call this API (ixCryptoAccCtxRegister)to register the context.
 *    
 * @param   *pAccCtx [in] is a pointer to hardware accelerator context. 
 *          Information required in hardware accelerator context, such as key 
 *          and algorithm configuration can be extracted from Security 
 *          Association Database (SAD). 
 * @param   *pMbufPrimaryChainVar [in] a pointer to an empty mbuf  
 *          (must not be chained) for the use of access component to compute 
 *          primary chaining variables for SHA1/MD5. This mbuf structure is  
 *          allocated by the client (mininum size of the cluster required is 
 *          64 bytes), only if SHA1/MD5 is selected, and the mbuf pointer must  
 *          be NULL if SHA1/MD5 is not selected. After the NPE complete the 
 *          computation, the mbuf is returned separately through the client's
 *          registered callback.
 * @param   *pMbufSecondaryChainVar [in] a pointer to an empty mbuf
 *          (must not be chained) for the use of access component to compute 
 *          secondary chaining variables for SHA1/MD5. This mbuf structure is  
 *          allocated by the client (mininum size of the cluster required is 
 *          64 bytes), only if SHA1/MD5 is selected, and the mbuf pointer must  
 *          be NULL if SHA1/MD5 is not selected. After the NPE complete the 
 *          computation, the mbuf is returned separately through the client's
 *          registered callback.
 * @param   registerCallbackFn [in] callback function pointer to return  
 *          crypto context registration status to client when the registration 
 *          is complete. This cannot be NULL.  
 * @param   performCallbackFn [in] callback function pointer to return the 
 *          processed buffer to the client with respect to the unique CryptoCtxId. 
 *          This cannot be NULL.
 * @param   *pCryptoCtxId: [inout] Crypto Context ID returned by access 
 *          component for the crypto context registered. 
 *
 * @return @li  IX_CRYPTO_ACC_STATUS_SUCCESS - Registration parameters are valid
 * @return @li  IX_CRYPTO_ACC_STATUS_FAIL - Registration failed for some 
 *              unspecified internal reasons.
 * @return @li  IX_CRYPTO_ACC_STATUS_OPERATION_NOT_SUPPORTED - Invalid 
 *              operation requested by the client.
 * @return @li  IX_CRYPTO_ACC_STATUS_CIPHER_ALGO_NOT_SUPPORTED - Invalid 
 *              cipher algorithm requested by the client.
 * @return @li  IX_CRYPTO_ACC_STATUS_CIPHER_MODE_NOT_SUPPORTED - Invalid 
 *              cipher mode requested by the client.
 * @return @li  IX_CRYPTO_ACC_STATUS_CIPHER_INVALID_KEY_LEN - Invalid cipher 
 *              key length passed in by the client.
 * @return @li  IX_CRYPTO_ACC_STATUS_CIPHER_INVALID_BLOCK_LEN - Invalid cipher
 *              block length passed in by the client
 * @return @li  IX_CRYPTO_ACC_STATUS_CIPHER_INVALID_IV_LEN - Invalid IV 
 *              length passed in by the client.
 * @return @li  IX_CRYPTO_ACC_STATUS_AUTH_ALGO_NOT_SUPPORTED - Invalid 
 *              authentication algorithm requested by the client.
 * @return @li  IX_CRYPTO_ACC_STATUS_AUTH_INVALID_DIGEST_LEN - Invalid 
 *              authentication digest length.
 * @return @li  IX_CRYPTO_ACC_STATUS_AUTH_INVALID_KEY_LEN - Invalid 
 *              authentication key length.
 * @return @li  IX_CRYPTO_ACC_STATUS_EXCEED_MAX_TUNNELS - Exceed maximum 
 *              tunnels permitted.
 * @return @li  IX_CRYPTO_ACC_STATUS_QUEUE_FULL - Queue full status returned 
 *              and the registration request will be rejected.
 *              
 * @li   Reentrant    : no
 * @li   ISR Callable : no
 *
 */ 
PUBLIC IxCryptoAccStatus 
ixCryptoAccCtxRegister (
    IxCryptoAccCtx *pAccCtx,
    IX_MBUF *pMbufPrimaryChainVar,
    IX_MBUF *pMbufSecondaryChainVar,
    IxCryptoAccRegisterCompleteCallback registerCallbackFn,
    IxCryptoAccPerformCompleteCallback performCallbackFn,
    UINT32 *pCryptoCtxId);



/**
 * 
 * @ingroup IxCryptoAcc
 *
 * @fn IxCryptoAccStatus ixCryptoAccCtxUnregister (UINT32 cryptoCtxId)
 *
 * @brief Unregister the crypto context from Cryptographic Context Database.
 *
 * This function is for freeing the particular crypto context (reference 
 *      through CryptoCtxId) from the Cryptographic Context Database.
 *
 * @param    cryptoCtxId [in] is pointer to crypto context that is required  
 *           to be freed, which will be supplied by the client. The 
 *           CryptoCtxId must be a valid Id.
 *
 * @return @li  IX_CRYPTO_ACC_STATUS_SUCCESS - successfully unregister the 
 *              crypto context.
 * @return @li  IX_CRYPTO_ACC_STATUS_FAIL - Unregistration failed for some 
 *              unspecified internal reasons, e.g. uninitialized.
 * @return @li  IX_CRYPTO_ACC_STATUS_CRYPTO_CTX_NOT_VALID - invalid crypto 
 *              context.
 * @return @li  IX_CRYPTO_ACC_STATUS_RETRY - retry the unregister operation as 
 *              there are still some ixCryptoAccAuthCryptPerform requests 
 *              associated with the cryptoCtxId pending on the NPE queue. 
 *              
 * @li   Reentrant    : no
 * @li   ISR Callable : no
 *
 */
PUBLIC IxCryptoAccStatus
ixCryptoAccCtxUnregister (UINT32 cryptoCtxId);



/**
 *
 * @ingroup IxCryptoAcc
 * 
 * @fn IxCryptoAccStatus ixCryptoAccAuthCryptPerform (
    UINT32 cryptoCtxId,
    IX_MBUF *pSrcMbuf,
    IX_MBUF *pDestMbuf,
    UINT16 authStartOffset,
    UINT16 authDataLen,    
    UINT16 cryptStartOffset,   
    UINT16 cryptDataLen,
    UINT16 icvOffset,
    UINT8  *pIV)
 * 
 * @brief Perform Authentication and Decryption/Encryption functionalities
 *
 * This function is called for authentication and decryption/encryption 
 * functionalities service request.
 * Note that the restriction with respect to authStartOffset, authStartLen, 
 * cryptStartOffset and cryptStartLen is that the crypted data must be a subset
 * of the authenticated data. The boundary relationship of 
 * [(authStartOffset + authDataLen) >= (cryptDataLen + cryptStartOffset) >=
 * (cryptStartOffset >= authStartOffset]
 * MUST BE SATISFIED.
 * There should not be any chained mbuf boundary within an ICV field.
 *
 * @param   cryptoCtxId [in] is the crypto context pointer to be supplied by  
 *          the client. This cryptoCtxId is obtained via  
 *          ixCryptoAccCtxRegister. The cryptoCtxId must be a valid Id.
 * @param   *pSrcMbuf [in] is a pointer to mbuf which contains data to be   
 *          processed. This mbuf structure is allocated by client. Result of 
 *          this request will be stored in the same mbuf and overwritten 
 *          the original data if UseDifferentSrcAndDestMbufs flag in 
 *          IxCryptoAccCtx is set to FALSE (in-place operation). Otherwise, if
 *          UseDifferentSrcAndDestMbufs flag is set to TRUE, the result will be
 *          written into destination mbuf (non in-place operation) and the 
 *          original data in this mbuf will remain unchanged.
 *          The same pointer is then returned to the client via registered 
 *          IxCryptoAccPerformCompleteCallback callback function.
 * @param   *pDestMbuf [in] only used if UseDifferentSrcAndDestMbufs  
 *          is TRUE. This is the buffer where the result is written to. 
 *          This mbuf structure is allocated by client. The length of mbuf 
 *          MUST be big enough to hold the result of operation. The result of  
 *          operation COULD NOT span into two or more different mbufs, thus 
 *          the mbuf supplied must be at least the length of expected result. 
 *          The same pointer is then returned to the client via registered 
 *          IxCryptoAccPerformCompleteCallback callback function.
 * @param   authStartOffset [in] supplied by the client to indicate  
 *          the start of the payload to be authenticated.
 * @param   authDataLen [in] supplied by the client to indicate  
 *          the length of the payload to be authenticated in Bytes. The
 *          maximum data length must not exceed 65471 bytes.
 * @param   cryptStartOffset [in] supplied by the client to indicate 
 *          the start of the payload to be decrypted/encrypted.
 * @param   cryptDataLen [in] supplied by the client to indicate the  
 *          length of the payload to be decrypted/encrypted in Bytes. The 
 *          payload to be decrypted/encrypted must have the length that is 
 *          multiple of cipher block length in size.
 * @param   icvOffset [in] supplied by the client to indicate  
 *          the start of the ICV (Integrity Check Value) used for the
 *          authentication. This ICV field should not be splitted across 
 *          multiple mbufs in a chained mbuf.
 * @param   *pIV [in] Initialization Vector  supplied by the client 
 *          to be used for the decryption/encryption process.
 *          
 * @return @li  IX_CRYPTO_ACC_STATUS_SUCCESS - Operation requested is 
 *              successfully enqueued to hardware accelerator for processing.
 * @return @li  IX_CRYPTO_ACC_STATUS_FAIL - Cryptographic process  failed for 
 *              some unspecified internal reasons.
 * @return @li  IX_CRYPTO_ACC_STATUS_QUEUE_FULL - Cryptographic queue is full.
 * @return @li  IX_CRYPTO_ACC_STATUS_CIPHER_INVALID_BLOCK_LEN - Invalid plaintext
 *              / ciphertext block size passed in by the client.
 * @return @li  IX_CRYPTO_ACC_STATUS_CRYPTO_CTX_NOT_VALID - Crypto context is 
 *              not registered.
 *
 * @note 
 * -  Client shall handle IP mutable fields.
 * -  Client shall pad the IP datagram to be a mutltiple of cipher block size, 
 *    using ESP trailer for encryption (RFC2406, explicit padding)
 * -  NPE shall pad the IP datagram to be a multiple of hashing block size, 
 *    specified by the authentication algorithm (RFC2402, implicit padding)
 * -  For authentication generation operation, client needs to clear the field
 *    which hold the authentication data (ICV) to zeroes. While for the 
 *    authentication verification operation, client needs to supply the 
 *    authentication data at the ICV field for NPE to verify the ICV value.
 *    If ICV is embedded in the payload, client DOES NOT need to move 
 *    the ICV to the front / the back of the payload and clear the original
 *    ICV field. cryptoAcc access component will handle this.
 * -  AES CTR couter block implementation is based on internet draft, 
 *    draft-ietf-ipsec-ciph-aes-ctr-00.txt (July 2002), which can be found at 
 *    IETF website, www.ietf.org
 *          
 *              
 * @li   Reentrant    : no
 * @li   ISR Callable : yes
 *
 */ 
PUBLIC IxCryptoAccStatus
ixCryptoAccAuthCryptPerform (
    UINT32 cryptoCtxId,
    IX_MBUF *pSrcMbuf,
    IX_MBUF *pDestMbuf,
    UINT16 authStartOffset,
    UINT16 authDataLen,    
    UINT16 cryptStartOffset,   
    UINT16 cryptDataLen,
    UINT16 icvOffset,
    UINT8  *pIV);


                 
/**
 *
 * @ingroup IxCryptoAcc
 *
 * @fn void ixCryptoAccShow (void)
 *
 * @brief API for printing statistics and status.
 *
 * This function is called by the client to print statistics, such as number 
 * of packets returned with operation fail, number of packets encrypted/
 * decrypted/authenticated while the function also print the current status of 
 * the queue, whether the queue is empty or full or current queue length.
 *
 * @param None
 *
 * @return None
 *          
 * @li   Reentrant    : no
 * @li   ISR Callable : no
 *
 */
PUBLIC void
ixCryptoAccShow (void);



/**
 *
 * @ingroup IxCryptoAcc
 *
 * @fn void ixCryptoAccShowWithId (UINT32 cryptoCtxId)
 *
 * @brief API for printing statistic and status.
 *
 * This function prints all the statistics provided by ixCryptoAccShow. In
 * addition the function will print out the contents of the Crypto Context
 * corresponding to the CryptoCtxId supplied through the API
 *
 * @param   cryptoCtxId [in] Crypto Context ID which has been registered
 *
 * @return  None
 *          
 * @li   Reentrant    : no
 * @li   ISR Callable : no
 *
 */
PUBLIC void
ixCryptoAccShowWithId (UINT32 cryptoCtxId); 



/**
 *
 * @ingroup IxCryptoAcc
 *
 * @fn IxCryptoAccStatus ixCryptoAccCryptoServiceStop (void)
 * 
 * @brief API to stop the security hardware accelerator services.
 *
 * This function is called to stop the security hardware accelerator services. 
 * All the requests pending in queues will be completed before the security 
 * hardware accelerator halt. Any new requests issued after this call will be 
 * rejected. All the Crypto Contexts will be unregistered in this function call. 
 *
 * @param None
 *
 * @return @li  IX_CRYPTO_ACC_STATUS_SUCCESS - Operation requested is successful 
 *              with all the pending requests are completed and CCD is cleared.
 * @return @li  IX_CRYPTO_ACC_STATUS_FAIL - Cryptographic stop request failed 
 *              for some unspecified internal reasons.
 *
 * @li   Reentrant    : no
 * @li   ISR Callable : no
 *
 */
PUBLIC IxCryptoAccStatus
ixCryptoAccCryptoServiceStop (void);


/**
 *
 * @ingroup IxCryptoAcc
 *
 * @fn IxCryptoAccStatus ixCryptoAccHashKeyGenerate (
    IxCryptoAccAuthAlgo hashAlgo,
    IX_MBUF *pMbufHashKey,
    IxCryptoAccHashKeyGenCompleteCallback hashKeyCallbackFn,
    UINT16 hashKeyStartOffset,
    UINT16 hashKeyLen,
    UINT16 hashKeyDestOffset,
    UINT32 *pHashKeyId)
 *
 * @brief This function is used to generate authentication key needed in HMAC
 *        authentication if the authentication key is greater than 64 bytes.
 *        New authentication key of L bytes size will be generated in this 
 *        function (L = 20 for SHA1, L = 16 for MD5). Please refer to RFC2104 
 *        for more details on the key size. The authentication key is padded 
 *        (extended) so that its length (in bits) is congruent to 448, 
 *        modulo 512 (please refer to RFC1321). 
 *
 * @note  
 * 1. This API will be called ONLY if the authentication key > 64 bytes. If
 *    key size is >=L and <= 64 bytes, authentication key is used directly in 
 *    Crypto context for registration, no further hashing is needed.
 * 2. This function must be called prior the Crypto Context registration if 
 *    the key size > 64. Result from this function will be used as 
 *    authentication key in the Crypto Context registration.
 * 3. It is client's responsibility to ensure the mbuf is big enough to
 *    hold the result (generated authentication key). No checking on mbuf 
 *    length will be done.
 *
 * @param   hashAlgo [in] is hashing algorithm to be used in generating the
 *          hash key. hashAlgo used to generate the key must be the same
 *          algorithm that will be used in crypto register and crypto perform
 *          services.
 * @param   *pMbufHashKey [in] a pointer to an mbuf that contains the 
 *          authentication key to be hashed. This mbuf structure is
 *          allocated by the client and the mbuf pointer must not be NULL. 
 *          After the NPE complete the computation, the mbuf is returned
 *          separately through the client's registered callback. The 
 *          authentication key generated will be stored in this mbuf, pointed  
 *          by hashKeyDestOffset. This must NOT be a chained mbuf.
 * @param   hashKeyCallbackFn [in] callback function pointer to be called
 *          when the hash key generation operation is completed. This 
 *          cannot be NULL.
 * @param   hashKeyStartOffset [in] offset to the mbuf mdata which contain
 *          the original hash key.
 * @param   hashKeyLen [in] key size
 * @param   hashKeyDestOffset [in] offset to the mbuf mdata to store the 
 *          generated result (authentication key)
 * @param   *pHashKeyId: [inout] Hash Key ID returned by access component
 *          to identify hash key generation. 
 *
 * @return @li  IX_CRYPTO_ACC_STATUS_SUCCESS - Hash key generation parameters
 *              are valid
 * @return @li  IX_CRYPTO_ACC_STATUS_FAIL - Hash key generation failed.
 * @return @li  IX_CRYPTO_ACC_STATUS_AUTH_ALGO_NOT_SUPPORTED - Invalid 
 *              hash algorithm  requested by the client.
 * @return @li  IX_CRYPTO_ACC_STATUS_AUTH_INVALID_KEY_LEN - Invalid 
 *              hash key length (key size <= 64 bytes for MD5/SHA1). If this
 *              error status is received, it means that the authentication key 
 *              (<= 64 bytes) can be used directly for crypto register function
 *              (ixCryptoAccCtxRegister), the key does not need to be hashed 
 *              to become shorter key.
 * @return @li  IX_CRYPTO_ACC_STATUS_QUEUE_FULL - Queue full status returned 
 *              and the authentication key generation request will be rejected.
 *
 * @li   Reentrant    : no
 * @li   ISR Callable : no
 *
 */ 
PUBLIC IxCryptoAccStatus
ixCryptoAccHashKeyGenerate (
    IxCryptoAccAuthAlgo hashAlgo,
    IX_MBUF *pMbufHashKey,
    IxCryptoAccHashKeyGenCompleteCallback hashKeyCallbackFn,
    UINT16 hashKeyStartOffset,
    UINT16 hashKeyLen,
    UINT16 hashKeyDestOffset,
    UINT32 *pHashKeyId);

             
#endif /* ndef IXCRYPTOACC_H */
/**
 *@}
 */
