/**
 * @file IxFpathAcc.h
 *
 * @date 09 Jan 2002
 *
 * @brief This file contains the public API for the IXP425 Fast Path Access
 * component.
 *
 * @par OverView
 * 
 * The IXP425 provides a mechanism to support higher levels of packet throughput
 * for Atm to Ethernet packets. This mechanism is termed the fast path.
 *
 * The IxFpathAcc component provides access to the fast path services provided
 * by NPEs. Currently fast path functionality is supported only by NPE-A (Atm)
 * and NPE-B (Ethernet) together, in the WAN to LAN direction.
 *
 * The fast path service operates by classify incoming packets achording to
 * a set of classification instructions (classification template) to determine if
 * they qualify to be sent on the fast path. Qualifying packets are sent directly
 * to NPE-B where they are modified achording to a set modification instructions
 * (modification template) before being transmitted on the Ehternet port.
 *
 * The IxFpathAcc components allows upto 8 classification/modification pairs to be
 * configured on the NPEs at a time. One classification pair per ATM VC.\
 *
 * @par Standard Operations
 *
 * For incoming Aal5 packets, the fast path service will check the aal5 CRC and
 * any packets which fail CRC are automatically sent the slow path.
 *
 * Aal5 packets which qualify for the fast path will have the all5 trailer stripped
 * automatically.
 *
 * If the Ethernet service (NPE Service) is overloaded i.e. transmit Q is full
 * packets are automaitically diverted on to the slow path.
 *
 * @par Application support and MAC Address service.
 * 
 * The fastpath is specifically designed to support applications such as Ethernet
 * half bridging and home gateway routing. Hence in support of Ehternet bridging
 * to multiple VCs environment the fastpath provides access to src Mac addresses
 * of incoming ATM packets per PVC. The fast path does not learn any MAC addresses
 * but makes them available to the Xscale to be learned as required. IxFpathAcc
 * provides a callback mechanism which is invoked everytime a packet with a new
 * MAC address is observed on a VC. One MAC address callback can registered with
 * the service per VC.
 *
 * @par Fast Path Buffers
 * 
 * The Fast Path service requires a buffers to be used to store and convey
 * packets on the fast path. A single buffer pool is used for this purpose and must
 * be populated by supplying buffers to IxFpathAcc before any packets can be conveyed
 * on the fast path.
 *
 * Unlike certain other access components buffers are not assigned per VC. A buffer
 * can therefore contain a packet from any incoming VC which has a classification
 * template pair assigned.
 *
 * Hence all fast path buffers must be of same size. The fast path service will not chain
 * buffers together to accommodate packets which would exceed a single buffers capacity.
 * Any such packet are automatically disqualified from the fast path and sent on the
 * slow path.
 *
 * When supplying buffers to the fast path pool the size of first buffer supplied
 * will become the size which will be accepted for all subsequent buffers.
 *
 * @par Enabling/Modifying/Disabling the Fast Path for a VC
 * 
 * Once a template pair has been accepted by IxFpathAcc for a VC the fast path is enabled
 * automatically on that VC. The next packet to arrive will be subjected to the fast path
 * function.
 *
 * Packets, however, will not be transported on the fast path if IxAtmd has not enabled
 * the VC, or if IxEthAcc has not enabled the Ethernet port.
 *
 * Modification of a template pair requires fast path processing for the particular VC
 * to be disabled for a brief period while the new template is installed. Packets will
 * continue to flow via the slow path.
 *
 * The service can be disabled by clearing the template pair for that VC.
 *
 * The exact content of the classfication/modification templates is outside the scope of
 * this component.
 *
 * @par Component Support
 * 
 * The IxFpathAcc supplements the services provided by IxAtmdAcc and IxEthAcc and as such
 * cannot perform its service without support from these components.

 * @par Dependencies
 * 
 * The fast path requires IxQmgr and IxNpeMh to be configured before
 * it is initialized
 *
 * @par Routing(IPv4) Support
 * 
 * The current software supports forwarding of IPv4 packets only on the fast 
 * path. Future software releases may provide support for IPv6 hence the 
 * Ethernet Type field is available for configuration on the API.
 *
 * -- Intel Copyright Notice --
 * 
 * @par
 * Copyright 2002-2003 Intel Corporation All Rights Reserved.
 * 
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.
 * 
 * @par
 * The Material is protected by worldwide copyright and trade secret laws
 * and treaty provisions. No part of the Material may be used, copied,
 * reproduced, modified, published, uploaded, posted, transmitted,
 * distributed, or disclosed in any way except in accordance with the
 * applicable license agreement .
 * 
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel, except in accordance with the
 * applicable license agreement.
 * 
 * @par
 * Unless otherwise agreed by Intel in writing, you may not remove or
 * alter this notice or any other notice embedded in Materials by Intel
 * or Intel's suppliers or licensors in any way.
 * 
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * 
 * @par
 * -- End Intel Copyright Notice --
*/

/**
 * @defgroup IxFpathAccAPI IXP425 Fast Path Access (IxFpathAcc) API
 *
 * @brief The public API for the IXP425 IxFpathAcc Component.
 *
 * @{
 */

#ifndef IXFPATHACC_H
#define IXFPATHACC_H

#include "IxTypes.h"
#include "IxOsBuffMgt.h"
#include "IxAtmTypes.h"
/**
 * The Ethernet port identifiers.
 */
typedef enum
{
    IX_FPATHACC_ETHERNET_PORT0 = 0, /**< Ethernet port 0 - attached to NPE-B >*/
    IX_FPATHACC_ETHERNET_MAX_PORTS/**< Maximum value of ethernet port range */
} IxFpathEthernetPort;


/**
 * @brief Client Id type
 */
typedef unsigned int IxFpathAccClientId;


/**
 *
 * @ingroup IxFpathAccAPI
 *
 * @def IX_FPATH_POOL_FULL
 *
 * @brief Buffer pool is full
 *
 * This constant is a return value used to tell the user that the IxFpathAcc
 * buffer pool is full.
 *
 */
#define IX_FPATH_POOL_FULL 2



/**
*
* @ingroup IxFpathAccAPI
*
* @brief  New WAN MAC address notification.
*
* This function is called to notify a client that the fast path has identified
* a change in src MAC address of the incoming packet stream on a VC.
*
* This function is only called if WanMacLearning instruction is enabled in the
* classifier template set for the VC.
*
* This function will be called every time a new WAN MAC address is learned,
* and is also guaranteed to be called every 500ms when the WAN MAC address is
* constant.
*
* @note - If the WAN MAC learning instruction is the first instruction in a
* classifier template all packets on the VC will be examined for MAC address
* changes. If it is the last instruction then onlt packets qualifying
* for the fast path will be examined.
*
* @param clientId  ID provided at registration of this callback.
* @param macAddress A pointer to the latest observed Mac Address for the VC.
*
* @return @li None
*
*/
typedef void (*IxFpathAccNewWanMacAddrCallback)(
                  IxFpathAccClientId clientId,
                  unsigned char *macAddress);

/**
*
* @ingroup IxFpathAccAPI
*
* @brief  The fast path buffer pool has suffered an under-run event.
*
* This function is called to notify a client that the fast path NPEs have
* requested more buffers from the pool than were placed in the pool at
* initialisation. The possible result is that packets are sent on the
* slow path instead of the fast path.
*
* This function will called exactly once every time the pool goes from empty
* to under-run.
*
* @param clientId ID provided at registration of this callback.
*
* @return @li None
*
* @sa ixFpathAccBufferPoolUnderRunCallbackSet
* @sa ixFpathAccBufferSupply
*
*/
typedef void (*IxFpathAccBufPoolUnderRunCallback)(IxFpathAccClientId clientId);

/**
*
* @ingroup IxFpathAccAPI
*
* @fn ixFpathAccInit(void)
*
* @brief Initialise the Fpath Access component
*
* This function will initialise the Fpath Access component internals and
* must be called before any other API call on this component can be made.
*
* @note DO NOT CALL THIS FUNCTION IF IX_ETH_ACC_FPATH_AWARE IS NOT SET  see
*       IxEthAcc.h.
*
* @return @li IX_SUCCESS - successfully initialised the component
* @return @li IX_FAIL Initialisation failed for some unspecified internal
* reason.
*/
PUBLIC IX_STATUS
ixFpathAccInit(void);

/**
*
* @ingroup IxFpathAccAPI
*
* @fn ixFpathAccEtherTypeSet(IxFpathEthernetPort port, UINT16 etherType)
*
* @brief Set the Ethernet type for frames on the specified port
*
* This function sets the Ethernet type value to be used in fastpath frames
* on the specified Ethernet port where the modifier engine must build
* Ehternet frame header from scratch.
*
* Hence, this function must be called before any fastpath templates are
* supplied to the service for VCs which build Ehternet frame headers, however
* IxFpathAcc will not prevent the Ether Type being set at any time, even while
* templates are set.

* Furthermore the current fast path implementation only supports forwarding of
* IPv4 packets which requires a value of 0x0800 for the Ehternet Type. Other
* values can be supplied but the modifier engine operation only supports IPv4
* right now.
*
* @sa ixFpathAccTemplatePairSet
*
* @param port which Ethernet port to configure.
* @param EtherType the Ethernet type.
*
* @return @li IX_SUCCESS - successful configured the ether type for the
* specified port.
* @return @li IX_FAIL Invalid port or the NPE failed to accept the request.
*
*/
PUBLIC IX_STATUS
ixFpathAccEtherTypeSet(IxFpathEthernetPort port, UINT16 etherType);

/**
*
* @ingroup IxFpathAccAPI
*
* @fn ixFpathAccBufferSupply(IX_MBUF *mBuf, int *unpopulatedEntryCnt)
*
* @brief Populate the Fast Path buffer pool
*
* This function supplys a single mbuf or chain of mbufs to the fast path
* service. An mbuf chain is broken into single mbufs and placed in the pool.
*
* All mbufs must be the same size and will be used to transport packets
* between the NPEs. The buffer pool is common so any buffer can be used for
* any fast path Vc.
*
* This function should be called before any templates are supplied to the service
* otherwise packets will be sent on the slow path. Usually this pool should be
* filled completely.
*
* If the number of elements in a chain of buffers supplied to this call exceeds
* the capacity of the pool the call will fail.
*
* @note  Ensure that the buffer size is sufficent for the largest packet to
* be sent on the fast path for any VC. Otherwise the packet will be sent on the
* slow path.
*
* @param mbuf single buffer or chain of buffers to be added to the pool. This
* parameter can be null. The first buffer supplied will set the acceptable size
* for all subsequent buffers.
*
* @param unpopulatedEntryCnt points to a variable which will filled a
* value representing the space in entries remaining unfilled in the
* pool.
*
* @sa ixFpathAccBufferPoolUnderRunCallbackSet
* @sa ixFpathAccBufferPoolInfoGet
*
* @return @li IX_SUCCESS - successful add a buffer chain to the pool.
* @return @li IX_FPATH_POOL_FULL - the pool has reach its capacity.
* @return @li IX_FAIL - Invalid parameter or internal error.
*
*/
PUBLIC IX_STATUS
ixFpathAccBufferSupply(IX_MBUF *mBuf, int *unpopulatedEntryCnt);

/**
*
* @ingroup IxFpathAccAPI
*
* @fn ixFpathAccBufferPoolInfoGet(int *maxEntries, int *populatedEntryCnt)
*
* @brief Get the fast path buffer pool info.
*
* This function gets the value for the maximum number of buffers which the
* fast path buffer pool can accommodate, it also retrieves the number of
* fast path buffers which have been supplied to the pool so far.
*
* This function will fail if a template is set for any VC.
*
* @param maxEntries points to a variable which will be filled with the
* maximum number of buffers which the fast path free pool can contain.
*
* @param populatedEntryCnt points to a variable which will filled a
* value representing number of fast path buffers curently supplied to
* the fast path pool.
*
* @sa ixFpathAccBufferPoolUnderRunCallbackSet
* @sa ixFpathAccBufferSupply
*
* @return @li IX_SUCCESS - successful add a buffer chain to the pool.
* @return @li IX_FAIL - Invalid parameter, a temaplte is set for a VC or
*                       internal error.
*
*/
PUBLIC IX_STATUS
ixFpathAccBufferPoolInfoGet(int *maxEntries, int *populatedEntryCnt);

/**
*
* @ingroup IxFpathAccAPI
*
* @fn ixFpathAccBufferPoolUnderRunCallbackSet(
    IxFpathAccBufPoolUnderRunCallback callback,
    IxFpathAccClientId clientId)
*
* @brief Set the buffer pool under-run notification callback
*
* Set the buffer under-run notification function. If one has already been set then
* it is superceed. Hence only one notofication callback is supported.
*
* @note The callback will be executed in the context of a Qmgr dispatch notification.
* @note The under-run callback can only be set when there is no active template
*
* @sa IxFpathAccBufPoolUnderRunCallback
*
* @param callback function pointer to be stored. This can be NULL.
* @param clientId Id supplied as a parameter to the callback funciton when invoked.
* this value is ignored if the callback is NULL.
*
* @return @li IX_SUCCESS notication callback has been set.
* @return @li IX_FAIL IxFpathAcc not yet initialised or some internal error has occurred.
*
*/
PUBLIC IX_STATUS
ixFpathAccBufferPoolUnderRunCallbackSet(
    IxFpathAccBufPoolUnderRunCallback callback,
    IxFpathAccClientId clientId);

/**
*
* @ingroup IxFpathAccAPI
*
* @fn ixFpathAccTemplatePairSet(
    IxAtmNpeRxVcId npeVcId,
    unsigned  char *classifierTemplate,
    int classifierTemplateLength,
    unsigned  char *modifierTemplate,
    int modifierTemplateLength)
*
* @brief Set the classifier/modifier template for a specific VC.
*
* This function provisions the NPEs with the supplied classifier/modifier
* templates. The desired VC is indicated by the npeVcId and must have been
* previously obtained from IxAtmdAcc.
*
* Packets will begin to flow on the fast path once the template pair has
* been installed.
*
* The fast path requires a valid npeVcId at all times so if a VC configured
* via IxAtmdAcc is removed or changed such that it's npeVcId is different
* then the npeVcId must be refreshed by calling this function.
* Otherwise the fast path will not be enabled on the VC. In general the
* fast path service for a VC before it is disabled on IxAtmdAcc.
*
* The integrity of the templates supplied to the service are not validated.
* Hence the it is the callers responsibility to ensure that the template contains
* a valid set of template instructions.
*
* If this function is called twice or more in succession for a particular VC,
* the template values of the most recent call are provisioned on the NPEs. Hence
* this function can be used to periodically update the template pair with having
* to call ixFpathAccTemplatePairClear.
*
* @note
* No fast path processing occurs on a VC while template are being set. A VC which has
* templates already set and hence has fast path enabled will have fast path processing
* disabled while the template update is executed.
*
* @note
* Template data MUST be in main memory when calling this API. The NPE engines
* copy the data from main memory only and not from cache. Failure to ensure
* this could cause the NPEs to be provisioned with erroneous template contents
* and cause unpredictable behavior.
*
* @sa ixFpathAccTemplatePairClear
*
* @param npeVcId	Identifier for a the VC obtained from IxAtmdAcc
*
* @param *classifierTemplate Pointer to byte array containing a combination of
* classfier instructions. This parameter cannot be NULL. The template must be
* constructed in network byte order ( i.e. big endian ). The template is
* internally copied by the fpath component.This template applies to a single VC and will be applied to all packets
* incoming on the VC once this function returns.
*
* @param classifierTemplateLength The length in bytes of the classifier array. This value
* cannot be 0.
*
* @param *modifierTemplate Pointer to a byte array containing a combination of modifier
* instructions. The template is internally copied by the fpath component
* The pointer cannot be NULL.The template must be constructed in network
* byte order ( i.e. big endian ).  This combination of modifier instructions will be
* applied to every packet which is passed by the classifier.
*
* @param modifierTemplateLength The length in bytes of the modifier array. This value
* cannot be 0.
*
* @return @li IX_SUCCESS Successfully provision the fast path with the supplied templates
* @return @li IX_FAIL Invalid parameter or some unspecified internal failure has occurred.
*
*/
PUBLIC IX_STATUS
ixFpathAccTemplatePairSet(
    IxAtmNpeRxVcId npeVcId,
    unsigned  char *classifierTemplate,
    int classifierTemplateLength,
    unsigned  char *modifierTemplate,
    int modifierTemplateLength);

/**
*
* @ingroup IxFpathAccAPI
*
* @fn ixFpathAccTemplatePairClear(IxAtmNpeRxVcId npeVcId)
*
* @brief Disable a template pair for a VC
*
* This function disables fast path processing for a particular VC by clearing the
* previously provisioned template pair for that VC. It also has the side effect
* of clearing the statistics for the npeVcId.
*
* @sa ixFpathAccTemplatePairSet
*
* @param npeVcId The identifier for a VC supplied previously in a call to ixFpathAccTemplatePairSet.
*
* @return @li IX_SUCCESS Successfully completed the clear operation
* @return @li IX_FAIL invalid parameter or some unspecified internal error has
* occurred. The state of the Npe(s) is indeterminate.
*
*/
PUBLIC IX_STATUS
ixFpathAccTemplatePairClear(IxAtmNpeRxVcId npeVcId);


/**
*
* @ingroup IxFpathAccAPI
*
* @fn ixFpathAccWanMacAddrCallbackSet( IxFpathAccNewWanMacAddrCallback macAddrCallback,
                                IxAtmNpeRxVcId npeVcId,
                                IxFpathAccClientId clientId)
*
* @brief Set the Mac Address callback for a VC.
*
* This function registers a callback to be called every time a new mac address is
* observed on the specified VC.
*
* All Mac address transitions may not be notified in the case of high traffic loads.
*
* A WAN MAC address detect instruction must be presented in the classifier template
* for MAC address observation service to be performed on a packet. Furthermore if the
* instruction appears as the first instruction in the template, MAC addresses for all
* packets on the VC will be observed, regardless of whether the packet ultimately goes
* fast path or slow path. On the other hand if it appears as the last instruction in
* the template only MAC addresses of packets which qualify for the past path will be
* observed.
*
* Mac address observation is disabled if a template pair is cleared or when a template
* pair is being update via the ixFpathAccTemplatePairSet call.
*
* Only one callback can be set per VC. Successive calls to this function will replace
* existing callbacks with newly supplied callback.
*
* @note the callback function is execute in the context of the IxNpeMh message
* notification.
*
* @sa ixFpathAccTemplatePairSet
*
* @note - The WAN MAC address notification must be set before a template is
* set for a particular VC. Once a template is set this function will fail for
* that VC.
*
* @param macAddrCallback new Mac address callback. This parameter can be set to NULL
* to clear a previously submitted callback.
*
* @param npeVcId VC indentifier matching the npeVcId supplied in the call to
* ixFpathAccTemplatePairSet.
*
* @param clientId identifier to be used as a parameter to the callback.
*
* @return @li IX_SUCCESS successfully stored the callback function.
* @return @li IX_FAIL invlaid parameter or some unspecified internal error
* has occurred.  The state of the Npe(s) is indeterminate.
*
*/
PUBLIC IX_STATUS
ixFpathAccWanMacAddrCallbackSet( IxFpathAccNewWanMacAddrCallback macAddrCallback,
                                IxAtmNpeRxVcId npeVcId,
                                IxFpathAccClientId clientId);

/**
*
* @ingroup IxFpathAccAPI
*
* @fn ixFpathAccStatisticsGet(
    IxAtmNpeRxVcId npeVcId,
    int *packetCnt)
*
* @brief  Retrieve the fast path statistics for a VC .
*
* The fast path service keeps a count of all packets which are transferred on the
* fast path.
*
* This value is cumulative for the same VC id. If the npeVcId changes the statistics
* count is cleared.
*
*
* @param npeVcId VC identifier as provided to ixFpathAccTemplatePairSet.
* @param packetCnt current packet count for the specified VC.
*
* @return @li IX_SUCCESS successfully retrieved the stats count.
* @return @li IX_FAIL invalid parameter or some unspecified internal error has
* occurred.
*/
PUBLIC IX_STATUS
ixFpathAccStatisticsGet(
    IxAtmNpeRxVcId npeVcId,
    int *packetCnt);

/**
*
* @ingroup IxFpathAccAPI
*
* @fn ixFpathAccShow(void)
*
* @brief  Display some component information for debug purposes
*
* Show some internal operation information relating to the fast path service.
*
* At a minimum the following will be show.
* @li the active fast path VCs
* @li the number of times a template has been set for particular VC.
* @li the last 5 MAC addresses for a particular VC
* @li a count of the number of mac address callbacks invoked for a VC.
* @li the number of under-run events which has occurred for the fast path buffer pool.
* @li the current level of the buffer pool.
* @li the number of fast path buffers returned via IxAtmdAcc.
* @li the physical address of the templates currently active.
*
* @param None
* @return @li None
*/
PUBLIC IX_STATUS
ixFpathAccShow(void);

/**
 * @ingroup IxFpathAccAPI
 *
 * @defgroup IxFpathAccAPIClassOpCode IXP425 Fast Path Classifier Opcode Definitions
 *
 * @brief The definitions for the IXP425 IxFpathAcc Classifier Opcodes.
 *
 * Opcodes used in classifier templates - NPE-A interprets a sequences of
 * these opcodes known as a template.
 *
 * @par
 * Classifier Instruction Format
 *
 * @par 
 * @li @b OpName (Opcode 1 byte):   
 *           Classifier operation
 *
 * @par 
 * @li @b Offset (Nb 1 byte):
 *           Position of the first byte to check, unsigned char, 16 is the
 *           first byte of the current cell; positions 0-15 indicate bytes
 *           in the previous cell and are therefore only valid after a
 *           Save16Bytes operation.
 *
 * @par 
 * @li @b OperandCount (Nd 1 byte):
 *           Number of bytes of PDU data needed to perform the checking.
 *
 * @par 
 * @li @b Parameters (Np bytes depending on Opcode):
 *           Some opcodes require opcode specific parameters. The number
 *           of which depends on the opcode and the OperandCount.
 *
 * @verbatim
 * OpName (value)    Offset  OperandCount Parameters 
 *
 * End(0)              N/A     0    value No more checking is needed
 * UpperLimit(1)       2,4     Nd   Nd byte data as an unsigned integer must be
 *                                  <= Nd byte of Pt as an unsigned integer.
 * LowerLimit(2)       N/A     1    One byte data as an unsigned integer must be
 *                                  >= One byte of Pt as an unsigned integer.
 * Match(3)            1,2,3,       Nd  Nd byte data must be == Nd byte of Pt.
 *                     4,8,10
 * Checksum(4)         10      0    Nd many short (i.e. 2-byte) data 16-bit
 *                                  checksum must be == 0xFFFF.
 * 1-out-of-4-match(5) 2,4,12       Nd byte data must be == one of the 4 Nd -byte
 *                     16,48        data in Pt.
 * A-or-B match(6)     N/A     2    One byte data must be either one of the two bytes
 *                                  in Pt.
 * MacWanLearn(7)      N/A     0    Inspect and report WAN MAC src addresses.
 * Conditional (8)     Special 1    If one byte match, skip next SpecialNd many bytes
 * Branch              Nd           of TempCls. Else go to the next test.
 * Save16Bytes(9)      N/A     0    D0 must be 48 for this instruction to work as
 *                                  intended. This instruction copies the last 16 bytes
 *                                  of the current cell's payload to a VC-specific
 *                                  temporary buffer and pauses the execution of the
 *                                  classifier until the next AAL5 cell is received on
 *                                  this VC. When the next cell is received, the
 *                                  previous cell's data is automatically prepended
 *                                  to the current cell and the classifier is restarted
 *                                  on the operation following Save16Bytes in the
 *                                  template. This instruction allows the classifier
 *                                  to cross cell boundaries subject to the
 *                                  restriction of a 16-byte overlap.
 * NOPpad3(10)         N/A     0    This instruction is used only to align the Pt
 *                                  field of a subsequent 1-out-of-4-match instruction.
 *                                  When the classifier encounters this instruction it
 *                                  will advance the TempCls pointer by three, i.e.
 *                                  past this NOPpad3 instruction and the following
 *                                  two bytes.
 * NOPpad2(11)         N/A     0    This instruction is used only to align the Pt
 *                                  field of a subsequent 1-out-of-4-match instruction.
 *                                  When the classifier encounters this instruction
 *                                  it will advance the  TempCls pointer by two,
 *                                  i.e. past this NOPpad2 instruction and the
 *                                  following byte.
 * NOPpad1(12)         N/A     0    This instruction is used only to align the Pt
 *                                  field of a subsequent 1-out-of-4-match instruction.
 *                                  When the classifier encounters this instruction
 *                                  it will advance the Classifier pointer by one,
 *                                  i.e. past this NOPpad1 instruction.
 * reserved(13)        N/A     N/A  THIS IS AN ILLEGAL OPCODE. It is used internally
 *                                  by the NPE to efficiently implement the NOP
 *                                  instructions above.
 * @endverbatim
 * 
 * @{
 */
#define IX_FPATHACC_CLS_OPCODE_END                0x00 /**< End */
#define IX_FPATHACC_CLS_OPCODE_UPPER_LIMIT        0x01 /**< UpperLimit */
#define IX_FPATHACC_CLS_OPCODE_LOWER_LIMIT        0x02 /**< LowerLimit */
#define IX_FPATHACC_CLS_OPCODE_MATCH_BYTES        0x03 /**< Match */
#define IX_FPATHACC_CLS_OPCODE_CHECKSUM           0x04 /**< Checksum */
#define IX_FPATHACC_CLS_OPCODE_1_OUT_OF_4_MATCH   0x05 /**< 1 out of 4 match */
#define IX_FPATHACC_CLS_OPCODE_A_OR_B_MATCH       0x06 /**< A or B match */
#define IX_FPATHACC_CLS_OPCODE_WAN_MAC_LEARN      0x07 /**< Wan MAC Learn */
#define IX_FPATHACC_CLS_OPCODE_CONDITIONAL_BRANCH 0x08 /**< Conditional branch */
#define IX_FPATHACC_CLS_OPCODE_COPY16             0x09 /**< Save 16 Bytes */
#define IX_FPATHACC_CLS_OPCODE_NOP_PAD3           0x0A /**< Padding */
#define IX_FPATHACC_CLS_OPCODE_NOP_PAD2           0x0B /**< Padding */
#define IX_FPATHACC_CLS_OPCODE_NOP_PAD1           0x0C /**< Padding */
#define IX_FPATHACC_CLS_OPCODE_NOP                0x0C /**< Padding */
#define IX_FPATHACC_CLS_DUMMY_BYTE                0xFF /**< Reserved */

/**
 * @} defgroup IxFpathAccAPIClassOpCode
 */

/**
 * @ingroup IxFpathAccAPI
 *
 * @defgroup IxFpathAccAPIModOpCode IXP425 Fast Path Modifier Template
 *
 * @brief The definitions for the IXP425 IxFpathAcc Modifier.
 *
 * Structure used in modifier templates - NPE-B interprets a sequences of
 * these known as a template.
 *
 * @verbatim

//==========================================================================
// Modification template  structures/constants
//==========================================================================
// TempMdf structure format:
//
//
//   0         2         4         6         8        10        12
//   +----+----+----+----+----+----+----+----+----+----+----+----+
//   | Opcode  |  mData  |  Dst MAC addr 0   |  Dst MAC addr 1   |
//   |         |  offset |    (bytes 0:3)    |    (bytes 0:3)    |
//   +----+----+----+----+----+----+----+----+----+----+----+----+
//  12        14        16        18        20        22        24
//   +----+----+----+----+----+----+----+----+----+----+----+----+
//   |  Dst MAC addr 2   |  Dst MAC addr 3   | DstMAC0 | DstMAC1 |
//   |    (bytes 0:3)    |    (bytes 0:3)    |  (4:5)  |  (4:5)  |
//   +----+----+----+----+----+----+----+----+----+----+----+----+
//  24        26        28        30        32        34        36
//   +----+----+----+----+----+----+----+----+----+----+----+----+
//   | DstMAC2 | DstMAC3 |   Dst IP addr 0   |   Dst IP addr 1   |
//   |  (4:5)  |  (4:5)  |                   |                   |
//   +----+----+----+----+----+----+----+----+----+----+----+----+
//  36        38        40        42        44        46        48
//   +----+----+----+----+----+----+----+----+----+----+----+----+
//   |   Dst IP addr 2   |   Dst IP addr 3   | DstPort | DstPort |
//   |                   |                   |    0    |    1    |
//   +----+----+----+----+----+----+----+----+----+----+----+----+
//  48        50        52        54        56        58        60
//   +----+----+----+----+----+----+----+----+----+----+----+----+
//   | DstPort | DstPort |   IPCS  |   IPCS  |   IPCS  |   IPCS  |
//   |    2    |    3    |    0    |    1    |    2    |    3    |
//   +----+----+----+----+----+----+----+----+----+----+----+----+
//  60        62        64        66        68
//   +----+----+----+----+----+----+----+----+
//   |  PortCS |  PortCS | PortCS  | PortCS  |
//   |    0    |    1    |    2    |    3    |
//   +----+----+----+----+----+----+----+----|
//
// Note that the Source MAC address and Ethernet type (include in the
// pre-pended Ethernet header when IP protocol processing is done) are
// the same for all VCs, so these fields are stored in universally
// accessable global variables.  Also, in the case of static NAT, the
// single local Destination IP and Port are stored at index of the
// respective arrays.
//
//==========================================================================
 * @endverbatim
 * 
 * @{
 */

/**
 * @} defgroup IxFpathAccAPIModOpCode
 */


#endif /* IXFPATH_H */
/**
 * @} defgroup IxFpathAccAPI
 */


