#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/partitions.h>
#include <linux/ioport.h>
#include <asm/io.h>

#define WINDOW_ADDR 	0x50000000
#define BUSWIDTH 	2
#define WINDOW_SIZE	0x01000000

#if defined(__LITTLE_ENDIAN)
#define B1(h) ((h) & 0xFF)
#define B2(h) (((h) >> 8) & 0xFF)
#elif defined(__BIG_ENDIAN)
#define B1(h) (((h) >> 8) & 0xFF)
#define B2(h) ((h) & 0xFF)
#elif
#error Byte order is not defined -- include "asm/byteorder.h"!
#endif

static void *memcpy16(void *dest, const void *src, int n)
{
    int i;
    u8 *d = (u8 *)dest;
    u16 *s = (u16 *)src;
    u16 data;
    
    for (i = 0; i < n / 2; i++) 
    {
	data = s[i];
	d[i*2] = B1(data);
	d[i*2+1] = B2(data);
    }

    if (n & 1)
	d[n-1] = B1(s[i]); 

    return dest;
}

static __u16 ixp425_read16(struct map_info *map, unsigned long ofs)
{
    return *(__u16 *)(map->map_priv_1 + ofs);
}

static void ixp425_copy_from(struct map_info *map, void *to,
    unsigned long from, ssize_t len)
{
#ifdef CONFIG_IXP425_JFFS2_WORKAROUND
    memcpy16(to, (void *)(map->map_priv_1 + from), len);
#else
    memcpy(to, (void *)(map->map_priv_1 + from), len);
#endif
}

static void ixp425_write16(struct map_info *map, __u16 d, unsigned long adr)
{
    *(__u16 *)(map->map_priv_1 + adr) = d;
}

static struct map_info ixp425_map = {
    name: 	"ixp425 Flash",
    buswidth: 	BUSWIDTH,
    read16:	ixp425_read16,
    copy_from:	ixp425_copy_from,
    write16:	ixp425_write16,
};

#ifdef CONFIG_MTD_REDBOOT_PARTS
static struct mtd_partition *parsed_parts;
#endif

#ifndef CONFIG_DIST_LSP

static struct mtd_partition ixp425_partitions[] = {
    {
	name:	"openrg",
	offset:	0x00040000,
	size:	0x00F80000,
    },
};

#else

static struct mtd_partition ixp425_partitions[] = {
    {
	name:	"image",
	offset:	0x00040000,
	size:	0x00400000, /* 4M for linux kernel + cramfs + initrd image */
    },
    {
	name:	"user",
	offset:	0x00440000,
	size:	0x00B80000, /* Rest of flash space minus redboot configuration
	                     * sectors (-256k) at the end of flash
			     */
    },
};

#endif

#define NB_OF(x)  (sizeof(x)/sizeof(x[0]))

static struct mtd_info *ixp425_mtd;
static struct resource *mtd_resource;

#ifdef DUMP_FLASH
static void hex_dump(__u32 *dp, int size)
{
    for (; size > 0; size--, dp++)
    {
	if (!((long)dp % 16))
	    printk("\n%p: ", dp);
	printk(" %08x", *dp);
    }
    printk("\n");
}
#endif

#ifdef MODULE
static void __exit ixp425_uninit(void)
#else
static void __init ixp425_uninit(void)
#endif
{
    if (ixp425_mtd)
    {
	del_mtd_partitions(ixp425_mtd);
	map_destroy(ixp425_mtd);
    }
    if (ixp425_map.map_priv_1)
	iounmap((void *)ixp425_map.map_priv_1);
    if (mtd_resource)
	release_mem_region(WINDOW_ADDR, WINDOW_SIZE);
    
#ifdef CONFIG_MTD_REDBOOT_PARTS
    if (parsed_parts)
	kfree(parsed_parts);
#endif

    /* Disable flash write */
    *(__u32 *)IXP425_EXP_CS0_REG_VIRT &= ~IXP425_FLASH_WRITABLE;
}

extern int mtd_boot_partition;
extern u32 kernel_image_start;
static int __init ixp425_init(void)
{
    int res, npart;

    /* Enable flash write */
    *(__u32 *)IXP425_EXP_CS0_REG_VIRT |= IXP425_FLASH_WRITABLE;

    /* Request IO region, doesn't perform any IO mapping
     * only checks for conflicts and registers fo future checking
     */
    ixp425_map.map_priv_1 = 0;
    mtd_resource = 
	request_mem_region(WINDOW_ADDR, WINDOW_SIZE, "ixp425 Flash");
    if (!mtd_resource)
    {
	printk(KERN_ERR
	    "ixp425 flash: Could not request mem region. "
	    "Probably memory region conflict.\n");
	res = -ENOMEM;
	goto Error;
    }

    /* Perform Flash IO region map into virtual address space */
    ixp425_map.map_priv_1 =
	(unsigned long)ioremap(WINDOW_ADDR, WINDOW_SIZE);
    if (!ixp425_map.map_priv_1)
    {
	printk("ixp425 Flash: Failed to map IO region. (ioremap)\n");
	res = -EIO;
	goto Error;
    }
    ixp425_map.size = WINDOW_SIZE;

#ifdef DUMP_FLASH
    hex_dump((void *)ixp425_map.map_priv_1, ixp425_map.size);
#endif

    /* Probe for the CFI complaint chip
     * suposed to be 28F128J3A
     */
    ixp425_mtd = do_map_probe("cfi_probe", &ixp425_map);
    if (!ixp425_mtd)
    {
	res = -ENXIO;
	goto Error;
    }
    ixp425_mtd->module = THIS_MODULE;
   
    /* Initialize flash partiotions 
     * Note: Redeboot partition info table can be parsed by MTD, and used
     *       instead of hard-coded partions. TBD
     */

#ifdef CONFIG_MTD_REDBOOT_PARTS
    /* Try to parse RedBoot partitions */
    npart = parse_redboot_partitions(ixp425_mtd, &parsed_parts);
    if (npart > 0)
    {
	/* found "npart" RedBoot partitions */
	
	res = add_mtd_partitions(ixp425_mtd, parsed_parts, npart);

	if (!res)
	{
	    /* locate the image partition */
	    int i;
	    int koffset = kernel_image_start - IXP425_EXP_BUS_BASE2_PHYS;
		
	    for (i = 0; i < npart; i++)
	    {
		if ((parsed_parts[i].offset <= koffset)
		    && (koffset < parsed_parts[i].offset
		    + parsed_parts[i].size))
		{
		    mtd_boot_partition = i;

		    printk("Found boot mtd partition %d: %s.\n", i, parsed_parts[i].name);
		    break;
		}
	    }
	}
    }
    else   
	res = -EIO;

    if (res)
#endif
    {
	printk("Using predefined MTD partitions.\n");
	/* RedBoot partitions not found - use hardcoded partition table */
	res = add_mtd_partitions(ixp425_mtd, ixp425_partitions,
	    NB_OF(ixp425_partitions));
    }

    if (res)
	goto Error;

    return res;
Error:
    ixp425_uninit();
    return res;
}

module_init(ixp425_init);
module_exit(ixp425_uninit);

MODULE_DESCRIPTION("MTD map driver for ixp425 evaluation board");

