/*
 *	Handle firewalling
 *	Linux ethernet bridge
 *
 *	Authors:
 *	Lennert Buytenhek               <buytenh@gnu.org>
 *	Bart de Schuymer		<bart.de.schuymer@pandora.be>
 *
 *	$Id: br_passthrough.c,v 1.2 2001/12/17 04:04:22 steve Exp $
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/ip.h>
#include <linux/netdevice.h>
#include <linux/skbuff.h>
#include <linux/if_ether.h>
#include <linux/netfilter_bridge.h>
#include <linux/netfilter_ipv4.h>
#include <net/ip.h>
#include <asm/uaccess.h>
#include <asm/checksum.h>
#include "../br_private.h"


//#define dprintk printk
#define dprintk(args...)


static inline void __maybe_fixup_src_address(struct sk_buff *skb)
{
#ifdef HAVE_SKBUFF_ORIGSRCDSTADDR
	if (skb->origsrcaddr != skb->nh.iph->saddr &&
	    inet_addr_type(skb->nh.iph->saddr) == RTN_LOCAL) {
		memcpy(skb->mac.ethernet->h_source,
			skb->dev->br_port->br->dev.dev_addr,
			ETH_ALEN);
	}
#endif
}



/* PF_BRIDGE/PRE_ROUTING *********************************************/
static struct net_device __fake_net_device = {
	hard_header_len:	ETH_HLEN
};

static struct dst_entry __fake_dst_entry = {
	__refcnt:		ATOMIC_INIT(1),
	dev:			&__fake_net_device,
	pmtu:			1500
};


static int br_nf_pre_routing_finish(struct sk_buff *skb)
{
	struct net_device *dev = skb->dev;
	struct iphdr *iph = skb->nh.iph;

	dprintk(KERN_NOTICE "br_nf_pre_routing_finish.. ");

	if (skb->origdstaddr != iph->daddr) {
		if (ip_route_input(skb, iph->daddr, iph->saddr, iph->tos, dev)) {
			struct rtable *rt;

/* 2004/2/25 Steve ++ */
//			if (ip_route_output(&rt, iph->daddr, 0, iph->tos, 0)) {
			if (ip_route_output(&rt, iph->daddr, 0, iph->tos, 0, 0)) {
/* Steve -- */
				dprintk("must drop\n");
				kfree_skb(skb);
				return 0;
			}

			dprintk("divertaway ");
			skb->dst = (struct dst_entry *)rt;
			memset(skb->mac.ethernet->h_dest, 0x22, ETH_ALEN);
		} else {
			dprintk("local ");
			memcpy(skb->mac.ethernet->h_dest, dev->dev_addr, ETH_ALEN);
		}
	} else {
		skb->dst = &__fake_dst_entry;
		dst_hold(skb->dst);
	}

	dprintk("\n");

	skb->dev = skb->physindev;
	br_handle_frame_finish(skb);

	return 0;
}

/* (PRE) okfn == br_handle_frame_finish */
static unsigned int br_nf_pre_routing(unsigned int hook, struct sk_buff **pskb, const struct net_device *in, const struct net_device *out, int (*okfn)(struct sk_buff *))
{
	struct iphdr *iph;
	__u32 len;
	struct sk_buff *skb, *arpkb = *pskb;

	/* 2001/12/10 Steve ++ 
	 * Change ethernet header, let this packet jump to forward chain.
	 */
	if (arpkb->mac.ethernet->h_proto == __constant_htons(ETH_P_ARP)) {
		struct arphdr *arp = arpkb->nh.arph;
		unsigned char *sha = (unsigned char *)arp + sizeof(struct arphdr);
		struct net_device *br_dev = &(in->br_port->br->dev);
		u32 sip, tip;
/* 2003/7/2 Steve ++ 
 * Proxy arp 
 */
		unsigned char *arp_ptr = sha;
/* Steve -- */

 		if (arp->ar_hrd != __constant_htons(ARPHRD_ETHER) ||
		    arp->ar_hln != br_dev->addr_len ||
		    arp->ar_pro != __constant_htons(ETH_P_IP) ||
		    arp->ar_pln != 4)
			goto noarp;

		sha += br_dev->addr_len;
		memcpy(&sip, sha, 4);
		sha += (4 + br_dev->addr_len);
		memcpy(&tip, sha, 4);

/* 2003/7/2 Steve ++ */
                if (arp->ar_op == __constant_htons(ARPOP_REQUEST) &&
                    inet_addr_type(tip) != RTN_LOCAL &&
                    inet_addr_type(sip) != RTN_LOCAL &&
                    pneigh_lookup(&arp_tbl, &tip, in, 0)) {
                        arp_send(ARPOP_REPLY,
                                 ETH_P_ARP,
                                 sip,
                                 br_dev,
                                 tip,
                                 arp_ptr,
                                 br_dev->dev_addr,
                                 arp_ptr);

                        goto out;
                }
/* Steve -- */

		if (arp->ar_op == __constant_htons(ARPOP_REQUEST) ||
		    arp->ar_op == __constant_htons(ARPOP_REPLY)) {
			if (inet_addr_type(tip) != RTN_LOCAL &&
			    inet_addr_type(sip) != RTN_LOCAL) {
				memcpy(arpkb->mac.ethernet->h_source, br_dev->dev_addr, ETH_ALEN);
				memset(arpkb->mac.ethernet->h_dest, 0xff, ETH_ALEN);
			}
		}
	}
	/* Steve -- */
noarp:

	if ((*pskb)->mac.ethernet->h_proto != __constant_htons(ETH_P_IP))
		return NF_ACCEPT;

	dprintk(KERN_NOTICE "br_nf_pre_routing.. ");
	if ((skb = skb_share_check(*pskb, GFP_ATOMIC)) == NULL)
		goto out;

	if (!pskb_may_pull(skb, sizeof(struct iphdr)))
		goto inhdr_error;

	iph = skb->nh.iph;
	if (iph->ihl < 5 || iph->version != 4)
		goto inhdr_error;

	if (!pskb_may_pull(skb, 4*iph->ihl))
		goto inhdr_error;

	if (ip_fast_csum((__u8 *)iph, iph->ihl) != 0)
		goto inhdr_error;

	len = ntohs(iph->tot_len);
	if (skb->len < len || len < 4*iph->ihl)
		goto inhdr_error;

	if (skb->len > len) {
		__pskb_trim(skb, len);
		if (skb->ip_summed == CHECKSUM_HW)
			skb->ip_summed = CHECKSUM_NONE;
	}

	skb->physindev = skb->dev;
	skb->dev = &(skb->dev->br_port->br->dev);
	skb->origdstaddr = iph->daddr;
	dprintk("calling br_nf_pre_routing_finish\n");
	NF_HOOK(PF_INET, NF_IP_PRE_ROUTING, skb, skb->dev, NULL,
		br_nf_pre_routing_finish);

	return NF_STOLEN;

inhdr_error:
//	IP_INC_STATS_BH(IpInHdrErrors);
out:
	dprintk("error out\n");
	return NF_DROP;
}

unsigned int
dummy_nf_local(unsigned int hook, struct sk_buff **pskb)
{
        struct sk_buff *skb = *pskb;

        dprintk("into dummy_nf_local: 0x4e75736f6674");
        if (skb->pkt_type != PACKET_OTHERHOST)
                return NF_STOLEN;
}


/* PF_BRIDGE/LOCAL_IN ************************************************/
static unsigned int br_nf_local_in(unsigned int hook, struct sk_buff **pskb, const struct net_device *in, const struct net_device *out, int (*okfn)(struct sk_buff *))
{
	struct sk_buff *skb = *pskb;
	struct net_device *br_dev = &(in->br_port->br->dev);

	/* 2001/12/10 Steve ++ 
	 * If arp request or reply is not ours, drop it.
	 */
	if (skb->mac.ethernet->h_proto == __constant_htons(ETH_P_ARP)) {
		struct arphdr *arp = skb->nh.arph;
		unsigned char *sha = (unsigned char *)arp + sizeof(struct arphdr);
		u32 sip, tip;

		if (arp->ar_hrd != __constant_htons(ARPHRD_ETHER) ||
		    arp->ar_hln != br_dev->addr_len ||
		    arp->ar_pro != __constant_htons(ETH_P_IP) ||
		    arp->ar_pln != 4)
			goto noarp;

		sha += br_dev->addr_len;
		memcpy(&sip, sha, 4);
		sha += (4 + br_dev->addr_len);
		memcpy(&tip, sha, 4);

		if (skb->nh.arph->ar_op == __constant_htons(ARPOP_REQUEST) ||
		    skb->nh.arph->ar_op == __constant_htons(ARPOP_REPLY)) {
			if (inet_addr_type(tip) != RTN_LOCAL &&
			    inet_addr_type(sip) != RTN_LOCAL)
				return NF_DROP;
		}
	}
	/* Steve -- */

noarp:
	if (skb->mac.ethernet->h_proto != __constant_htons(ETH_P_IP))
		return NF_ACCEPT;

	dprintk(KERN_NOTICE "br_nf_local_in.. ");
	if (skb->pkt_type != PACKET_OTHERHOST) {
		dprintk("ip_rcv_finish\n");
		IP_INC_STATS_BH(IpInReceives);
		if (skb->dst == &__fake_dst_entry)
			skb->dst = NULL;
		ip_rcv_finish(skb);
		return NF_STOLEN;
	}

	dprintk("drop\n");
	return NF_DROP;
}


/* PF_BRIDGE/FORWARD *************************************************/
static int br_nf_forward_finish(struct sk_buff *skb)
{
	dprintk(KERN_NOTICE "br_nf_forward_finish..\n");
//	__maybe_fixup_src_address(skb);
	br_forward_finish(skb);

	return 0;
}

static int br_nf_forward_postrouting(struct sk_buff *skb)
{
	dprintk(KERN_NOTICE "br_nf_forward_postrouting..\n");
	skb->origsrcaddr = skb->nh.iph->saddr;

	/* @@@ why doesn't br, br work here? */
	NF_HOOK(PF_INET, NF_IP_POST_ROUTING, skb, skb->physindev, skb->dev,
		br_nf_forward_finish);

	return 0;
}

static inline int output_maybe_reroute(struct sk_buff *skb)
{
	dprintk(KERN_NOTICE "output_maybe_reroute\n");

	/* @@@ this changes source MAC */
	return skb->dst->output(skb);
}

/* (PRE) okfn == br_forward_finish */
static unsigned int br_nf_forward(unsigned int hook, struct sk_buff **pskb, const struct net_device *in, const struct net_device *out, int (*okfn)(struct sk_buff *))
{
//	struct net_device *br;
	struct sk_buff *skb = *pskb;
	int (*outfn)(struct sk_buff *);
/* 2004/1/14 Steve ++ */
        struct iphdr *iph;
/* Steve -- */

	/* 2001/12/10 Steve ++ 
	 * Change ethernet and arp header.
	 */
	if (skb->mac.ethernet->h_proto == __constant_htons(ETH_P_ARP)) {
		struct arphdr *arp = skb->nh.arph;
		struct net_device *br_dev = &(in->br_port->br->dev);
		unsigned char *sha = (unsigned char *)arp + sizeof(struct arphdr);
		u32 sip, tip;

		if (arp->ar_hrd != __constant_htons(ARPHRD_ETHER) ||
		    arp->ar_hln != br_dev->addr_len ||
		    arp->ar_pro != __constant_htons(ETH_P_IP) ||
		    arp->ar_pln != 4)
			goto noarp;

		if (skb->nh.arph->ar_op == __constant_htons(ARPOP_REQUEST) ||
		    skb->nh.arph->ar_op == __constant_htons(ARPOP_REPLY)) {
			unsigned char *arpdata = sha;

			arpdata += br_dev->addr_len;
			memcpy(&sip, arpdata, 4);
			arpdata += (4 + br_dev->addr_len);
			memcpy(&tip, arpdata, 4);

			if (inet_addr_type(tip) != RTN_LOCAL &&
			    inet_addr_type(sip) != RTN_LOCAL) {
#if 0
				memcpy(skb->mac.ethernet->h_source, br_dev->dev_addr, ETH_ALEN);
				memset(skb->mac.ethernet->h_dest, 0xff, ETH_ALEN);
#endif
				memcpy(sha, br_dev->dev_addr, br_dev->addr_len);
				sha += (ETH_ALEN + 4);
				memset(sha, 0, br_dev->addr_len);
			} else
				return NF_DROP;
		}
	}
	/* Steve -- */

noarp:
	if (skb->mac.ethernet->h_proto != __constant_htons(ETH_P_IP))
		return NF_ACCEPT;

/* 2004/1/14 Steve ++ */
        iph = skb->nh.iph;

        if (in && inet_addr_type(iph->daddr) == RTN_BROADCAST &&
            !memcmp(skb->mac.ethernet->h_dest, in->broadcast, in->addr_len))
                return NF_ACCEPT;
/* Steve -- */

	dprintk(KERN_NOTICE "br_nf_forward.. ");

	skb->physoutdev = skb->dev;

	outfn = br_nf_forward_postrouting;
	if (skb->origdstaddr != skb->nh.iph->daddr)
		outfn = output_maybe_reroute;

	dprintk("call with outfn==%s\n", outfn==output_maybe_reroute?"omr":"bnfp");

	/* @@@ br = &(skb->dev->br_port->br->dev) */
	/* @@@ why doesn't br, br work here? */
	NF_HOOK(PF_INET, NF_IP_FORWARD, skb, skb->physindev, skb->dev, outfn);

	return NF_STOLEN;
}


/* PF_BRIDGE/LOCAL_OUT ***********************************************/
static int br_nf_local_out_finish(struct sk_buff *skb)
{
	dprintk(KERN_NOTICE "br_nf_local_out_finish.. ");

	if (skb->nh.iph->frag_off) {
		struct hh_cache *hh = skb->dst->hh;

		if (hh == NULL) {
			dprintk(KERN_NOTICE "drop\n");
			kfree_skb(skb);
			return 0;
		}

		read_lock_bh(&hh->hh_lock);
		memcpy(skb->data - 16, hh->hh_data, 16);
		read_unlock_bh(&hh->hh_lock);
	}

	dprintk("\n");
//	__maybe_fixup_src_address(skb);
	br_forward_finish(skb);

	return 0;
}

static int br_nf_local_out_postrouting(struct sk_buff *skb)
{
	dprintk(KERN_NOTICE "br_nf_local_out_postrouting..\n");
	skb->origsrcaddr = skb->nh.iph->saddr;

	/* because we don't want to filter on physoutdev in nat/POSTROUTING */
/* 2003/4/16 Steve ++ */
#if 0
	skb->physoutdev = NULL;
#endif
/* Steve -- */
	NF_HOOK_THRESH(PF_INET, NF_IP_POST_ROUTING, skb, skb->realindev,
			&(skb->dev->br_port->br->dev),
			br_nf_local_out_finish,
			INT_MIN+1);

	return 0;
}

/* (PRE) okfn == br_forward_finish */
static unsigned int br_nf_local_out(unsigned int hook, struct sk_buff **pskb, const struct net_device *in, const struct net_device *out, int (*okfn)(struct sk_buff *))
{
	struct sk_buff *skb = *pskb;
	int hookno;

	if (skb->mac.ethernet->h_proto != __constant_htons(ETH_P_IP))
		return NF_ACCEPT;

	dprintk(KERN_NOTICE "br_nf_local_out.. ");

	skb->physoutdev = skb->dev;

	hookno = NF_IP_LOCAL_OUT;
	if (skb->dst != NULL && skb->dst->input == ip_forward)
		hookno = NF_IP_FORWARD;

	dprintk("%s\n", hookno==NF_IP_LOCAL_OUT?"LO":"FW");

	NF_HOOK_THRESH(PF_INET, hookno, skb, skb->realindev,
			&(skb->dev->br_port->br->dev),
			br_nf_local_out_postrouting,
			NF_IP_PRI_BRIDGE_SABOTAGE+1);

	return NF_STOLEN;
}


/* IPv4/SABOTAGE *****************************************************/
static unsigned int ipv4_sabotage(unsigned int hook, struct sk_buff **pskb, const struct net_device *in, const struct net_device *out, int (*okfn)(struct sk_buff *))
{
	if (out->hard_start_xmit == br_dev_xmit) {
		struct sk_buff *skb = *pskb;

		if (hook == NF_IP_FORWARD)
			skb->realindev = (struct net_device *)in;

		dprintk(KERN_NOTICE "ipv4_sabotage %i %p\n", hook, skb);
		okfn(skb);
		return NF_STOLEN;
	}

	return NF_ACCEPT;
}


static struct nf_hook_ops br_nf_ops[] = {
	{ { NULL, NULL }, br_nf_pre_routing, PF_BRIDGE, NF_BR_PRE_ROUTING, 0 },
	{ { NULL, NULL }, br_nf_local_in, PF_BRIDGE, NF_BR_LOCAL_IN, 0 },
	{ { NULL, NULL }, br_nf_forward, PF_BRIDGE, NF_BR_FORWARD, 0 },
	{ { NULL, NULL }, br_nf_local_out, PF_BRIDGE, NF_BR_LOCAL_OUT, INT_MIN },
	{ { NULL, NULL }, ipv4_sabotage, PF_INET, NF_IP_FORWARD, NF_IP_PRI_BRIDGE_SABOTAGE },
	{ { NULL, NULL }, ipv4_sabotage, PF_INET, NF_IP_LOCAL_OUT, NF_IP_PRI_BRIDGE_SABOTAGE },
	{ { NULL, NULL }, ipv4_sabotage, PF_INET, NF_IP_POST_ROUTING, INT_MIN },
};

#define NUMHOOKS (sizeof(br_nf_ops)/sizeof(br_nf_ops[0]))


static int __init init(void)
{
	int i;

	for (i=0;i<NUMHOOKS;i++) {
		int ret;

		if ((ret = nf_register_hook(&br_nf_ops[i])) >= 0)
			continue;

		while (i--)
			nf_unregister_hook(&br_nf_ops[i]);

		return ret;
	}

	printk(KERN_NOTICE "Bridge firewalling registered\n");

	return 0;
}

static void __exit fini(void)
{
	int i;

	for (i=NUMHOOKS-1;i>=0;i--)
		nf_unregister_hook(&br_nf_ops[i]);
}

module_init(init);
module_exit(fini);
