/*
 *  ebtables
 *
 *	Authors:
 *	Bart De Schuymer		<bart.de.schuymer@pandora.be>
 *
 *	ebtables.c,v 1.0prerelease 2001/08/01
 *
 *  This code is stongly inspired on the iptables code which is
 *  Copyright (C) 1999 Paul `Rusty' Russell & Michael J. Neuling
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 */

/* Deal with CONFIG_MODVERSIONS */
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif

/* used for print_string */
#include <linux/sched.h>    /* For current */
#include <linux/tty.h>      /* For the tty declarations */

#include <linux/kernel.h>   /* We're doing kernel work */
#include <linux/module.h>
#include <linux/vmalloc.h>
#include <linux/skbuff.h>
#include <linux/if_ether.h>
#include <linux/netfilter_bridge.h>
#include <linux/ebtables.h>
#include <linux/socket.h> /* PF_BRIDGE */
#include <linux/spinlock.h> /* rwlock_t */
#include <asm/errno.h>
#include <asm/uaccess.h> /* copy_[to,from]_user */
#include <linux/smp.h> /* multiprocessors */
#include <linux/ip.h> /* log ip source && destination */

#define BUGPRINT(format, args...) printk("kernel msg: ebtables bug: please report to author: "format, ## args)
/*#define BUGPRINT(format, args...)*/
#define MEMPRINT(format, args...) printk("kernel msg: ebtables : out of memory: "format, ## args)
/*#define MEMPRINT(format, args...)*/

// this array makes it easy to know the number of chains in a table
int numc[EPF_NUMTABLES] = {EPF_NUMCHAINS0, EPF_NUMCHAINS1};

static char *table[EPF_NUMTABLES];/* the tables, consisting of the chains per hook */
static struct epf_entries **pointers[EPF_NUMTABLES];// these point to the beginning of the chains

/* counter variables */
// per table, per cpu, per chain we have counters
static struct epf_counter **counters[EPF_NUMTABLES];
static unsigned short allowc[EPF_NUMTABLES];// counters are on or off?

// information that isn't always supplied for every rule (e.g. with -j SNAT --to-source)
static char *additional_infos[EPF_NUMTABLES];
static int additional_size[EPF_NUMTABLES];

/* table lock */
static rwlock_t epf_lock;
/* log lock */
static spinlock_t epf_log_lock = SPIN_LOCK_UNLOCKED;

static unsigned int in_kernel = 0;// This contains a bitmask of the options loaded into the kernel

// I've had a good look at ipt_LOG.c for this, a very good look ;-)
static void epf_log(struct sk_buff *skb, struct epf_log_info *loginfo,
	const struct net_device *in, const struct net_device *out){

	char level_string[4] = "< >";
	level_string[1] = '0' + (loginfo->loglevel % 8);

	spin_lock_bh(&epf_log_lock);
	printk(level_string);
	printk("%s IN=%s OUT=%s ",
	       loginfo->prefix,
	       in ? in->name : "",
	       out ? out->name : "");// max length: 29 + 10 + 2 * 16

	if (skb->dev->hard_header_len) {
		int i;
		unsigned char *p = (skb->mac.ethernet)->h_source;
		printk("MAC source = ");
		for (i = 0; i < ETH_ALEN; i++,p++)
			printk("%02x%c", *p,
			       i == ETH_ALEN - 1
			       ? ' ':':');// length: 31
		printk("MAC dest = ");
		p = (skb->mac.ethernet)->h_dest;
		for (i = 0; i < ETH_ALEN; i++,p++)
			printk("%02x%c", *p,
			       i == ETH_ALEN - 1
			       ? ' ':':');// length: 29
	}
	printk("proto = 0x%04x", ntohs(((*skb).mac.ethernet)->h_proto));// length: 14
	if ((loginfo->bitmask & EPF_LOG_IP) && skb->mac.ethernet->h_proto == htons(ETH_P_IP)){
		// log ip information
		struct iphdr *iph = skb->nh.iph;
		printk(" IP SRC=%u.%u.%u.%u IP DST=%u.%u.%u.%u,", NIPQUAD(iph->saddr), NIPQUAD(iph->daddr));// max length: 46
		printk(" IP tos=0x%02X, IP proto=%d", iph->tos, iph->protocol);// max length: 26
	}
	printk("\n");
	spin_unlock_bh(&epf_log_lock);
}

static int epf_dev_check(char *entry, const struct net_device *device){
	if (*entry == '\0') return 0;
	if (!device) return 1;
	return strncmp(entry, device->name, IFNAMSIZ);
}	

/* called under read_lock_bh */
static void increase_counter(unsigned int chain, unsigned char table, int pos){
	int i, where = 0;
	for (i = 0; i < chain; i++)
		where += pointers[table][i]->n;
	where += pos;
	(counters[table][cpu_number_map(smp_processor_id())][where].pcnt)++;
}

// This used to be a small function :-)
static unsigned int br_hook (unsigned int chain, struct sk_buff **pskb,
			const struct net_device *in,
			const struct net_device *out,
			int (*okfn)(struct sk_buff *), unsigned char table_nr)
{
	int i, nentries;
	struct epf_entry *point;

	read_lock_bh(&epf_lock);
	nentries = (pointers[table_nr][chain])->n;
	point = (struct epf_entry *)(pointers[table_nr][chain]->data);
	#define FWINV(bool,invflg) ((bool) ^ !!(point->invflags & invflg))
	#define FWINV2(bool,invflg) ((bool) ^ !!(infostuff->invflags & invflg))
	for (i = 0; i < nentries; i++){
		// first we compare the protocol field
		if ( ( point->bitmask & EPF_NOPROTO ||
		   FWINV(point->ethproto == ((**pskb).mac.ethernet)->h_proto, EPF_IPROTO)
		   || FWINV(ntohs(((**pskb).mac.ethernet)->h_proto) < 0x0800 && (point->bitmask & EPF_802_3), EPF_IPROTO) )
		// next the devices
		   && FWINV(!epf_dev_check((char *)(point->in), in), EPF_IIN)
		   && FWINV(!epf_dev_check((char *)(point->out), out), EPF_IOUT) ){
			unsigned int addinfo_offset = 0;

			// next we compare the mac's
			if ( (point->bitmask & EPF_SOURCEMAC) &&
			      FWINV(!!memcmp(point->sourcemac, ((**pskb).mac.ethernet)->h_source, ETH_ALEN), EPF_ISOURCE) )
				goto letscontinue;

			if ( (point->bitmask & EPF_DESTMAC) &&
			      FWINV(!!memcmp(point->destmac, ((**pskb).mac.ethernet)->h_dest, ETH_ALEN), EPF_IDEST) )
				goto letscontinue;

			// next we compare other possible info

#ifdef CONFIG_BRIDGE_EBT_IPF
			// obviously this option can only exist along with this filter rule: -p 0x0800
			// the userspace program and do_replace() have made sure this is the fact
			if (point->bitmask & EPF_IP){
				struct epf_ip_info *infostuff = (struct epf_ip_info *) (point->additional_info + addinfo_offset);

				addinfo_offset += sizeof(struct epf_ip_info);
				if (infostuff->bitmask & EPF_IP_TOS && FWINV2(infostuff->tos != ((**pskb).nh.iph)->tos, EPF_IP_TOS))
					goto letscontinue;
				if (infostuff->bitmask & EPF_IP_PROTO && FWINV2(infostuff->protocol != ((**pskb).nh.iph)->protocol, EPF_IP_PROTO))
					goto letscontinue;
				if (infostuff->bitmask & EPF_IP_SOURCE &&
				    FWINV2((((**pskb).nh.iph)->saddr & infostuff->smsk) != infostuff->saddr, EPF_IP_SOURCE))
					goto letscontinue;
				if ((infostuff->bitmask & EPF_IP_DEST) &&
				    FWINV2((((**pskb).nh.iph)->daddr & infostuff->dmsk) != infostuff->daddr, EPF_IP_SOURCE))
					goto letscontinue;
			}
#endif
			// ok, the rule matches
			if (allowc[table_nr] == EPF_COUNTERS)
			increase_counter(chain, table_nr, i);

			if (point->target == EPF_ACCEPT){
				read_unlock_bh(&epf_lock);
				return NF_ACCEPT;
			}
			if (point->target == EPF_DROP){
				read_unlock_bh(&epf_lock);
				return NF_DROP;
			}
			// the rule probably just wants to see the counters
			// this target shouldnt happen much so we put it behind
			if (point->target == EPF_CONTINUE)
				goto letscontinue;
#ifdef CONFIG_BRIDGE_EBT_NAT
			// userspace makes sure no NAT happens in FORWARD or INPUT
			// even if it would happen there, this won't harm the kernel in any way
			if (point->target == EPF_SNAT){
				memcpy(((**pskb).mac.ethernet)->h_source, point->additional_info + addinfo_offset, ETH_ALEN * sizeof(unsigned char));
				read_unlock_bh(&epf_lock);
				return NF_ACCEPT;
			}
			if (point->target == EPF_DNAT){
				memcpy(((**pskb).mac.ethernet)->h_dest, point->additional_info + addinfo_offset, ETH_ALEN * sizeof(unsigned char));
				read_unlock_bh(&epf_lock);
				return NF_ACCEPT;
			}
#endif
#ifdef CONFIG_BRIDGE_EBT_LOG
			epf_log(*pskb, (struct epf_log_info *)(point->additional_info + addinfo_offset), in, out);
#endif
		}
letscontinue:
		point++;
	}

	if ( (pointers[table_nr][chain])->policy == EPF_ACCEPT ){
		read_unlock_bh(&epf_lock);
		return NF_ACCEPT;
	}
	read_unlock_bh(&epf_lock);
	return NF_DROP;
}

#ifdef CONFIG_BRIDGE_EBT_FILTER
static unsigned int br_filter (unsigned int hook, struct sk_buff **pskb,
			const struct net_device *in,
			const struct net_device *out,
			int (*okfn)(struct sk_buff *)){
	// LOCAL_IN = 0
	// FORWARD = 1
	// LOCAL_OUT = 2
	return br_hook(hook -1, pskb, in, out, okfn, EPF_FILTERTABLE);
}
#endif

#ifdef CONFIG_BRIDGE_EBT_NAT
static unsigned int br_nat (unsigned int hook, struct sk_buff **pskb,
			const struct net_device *in,
			const struct net_device *out,
			int (*okfn)(struct sk_buff *)){
	// PRE_ROUTING = 0
	// LOCAL_OUT = 1
	// POST_ROUTING = 2
	if (hook == NF_BR_LOCAL_OUT)
		return br_hook(1, pskb, in, out, okfn, EPF_NATTABLE);
	else {
		if (hook == NF_BR_POST_ROUTING)
			return br_hook(2, pskb, in, out, okfn, EPF_NATTABLE);
		else
			return br_hook(0, pskb, in, out, okfn, EPF_NATTABLE);
	}
}
#endif

static int switch_nocounters(unsigned char table_nr){
	int i;

	if (counters[table_nr]){
		for (i = 0; i < smp_num_cpus; i++)
			vfree(counters[table_nr][i]);
		vfree(counters[table_nr]);
		counters[table_nr] = NULL;
	}
	allowc[table_nr] = EPF_NOCOUNTERS;
	return 0;
}

static int switch_counters(unsigned char table_nr){
	int i, j, countersize = 0;

	for (i = 0; i < numc[table_nr]; i++)
		countersize += pointers[table_nr][i]->n;
	if (!countersize){
		allowc[table_nr] = EPF_COUNTERS;// the counters are turned on, but there are no rules
		return 0;
	}
	if ( !(counters[table_nr] = (struct epf_counter **)vmalloc(smp_num_cpus * sizeof(struct epf_counter *))) ){
		MEMPRINT("switch_c && nomemory\n");
		return -ENOMEM;
	}
	countersize *= sizeof(struct epf_counter);
	for (i = 0; i < smp_num_cpus; i++){
		if ( !(counters[table_nr][i] = (struct epf_counter *)vmalloc(countersize)) )
			goto undo_sw;
		memset(counters[table_nr][i], 0, countersize);
	}
	allowc[table_nr] = EPF_COUNTERS;
	return 0;
undo_sw:
	MEMPRINT("switch_c && nomemory2\n");
	for (j = 0; j < i; j++)
		vfree(counters[table_nr][j]);
	vfree(counters[table_nr]);
	counters[table_nr] = NULL;
	return -ENOMEM;
}

/* big one huh? */
static int do_replace(void *user, unsigned int len)
{
	int ret;
	unsigned char table_nr = EPF_NUMTABLES;
	struct epf_info tmp;
	struct epf_counter *counterstmp, **newcounters;
	char *new_, *point, *new_addinfos = NULL;
	unsigned int test = sizeof(tmp),
		i, j, countersize = 0, oldcz, add_info_size = 0;

	if (len < sizeof(tmp)) return -EINVAL;
 	if (copy_from_user(&tmp, user, sizeof(tmp)) != 0)
		return -EFAULT;
	if ( (table_nr = tmp.table) >= EPF_NUMTABLES)
		return -EINVAL;

	if ((table_nr == EPF_NATTABLE && !(in_kernel & EPF_IN_KERNEL_NAT)) ||
	    (table_nr == EPF_FILTERTABLE && !(in_kernel & EPF_IN_KERNEL_FILTER)))
		return -EINVAL;

	test += numc[table_nr] * sizeof(struct epf_entries);

	if (tmp.allow_counters != allowc[table_nr])
		return -EINVAL;
	for (i = 0;i < numc[table_nr]; i++){
		test += tmp.nentries[i] * sizeof(struct epf_entry);
		countersize += tmp.nentries[i];/* the number of entries (needed to allocate counter space) */
	}
	countersize *= sizeof(struct epf_counter);

	if (test != len){
		BUGPRINT("do_replace && !samelength\n");
		return -EINVAL;
	}
	/* allocate memory for the new table */
	if ( !(new_ = (char *)vmalloc(len - sizeof(tmp))) )
		return -ENOMEM;

 	if (copy_from_user(new_, user + sizeof(tmp), len - sizeof(tmp))){
		ret = -EFAULT;
		goto free_new;
	}

	/* test it for bad input */
	point = new_;
	for (i = 0;i < numc[table_nr]; i++){
		if ( ((struct epf_entries*)point)->policy != EPF_ACCEPT && ((struct epf_entries*)point)->policy != EPF_DROP ){
			ret = -EINVAL;
			goto free_new;
		}
		if ( ((struct epf_entries*)point)->n != tmp.nentries[i] ){
			ret = -EINVAL;
			goto free_new;
		}
		point += sizeof(struct epf_entries);
		for (j = 0; j < tmp.nentries[i]; j++){
			struct epf_entry *hlp = (struct epf_entry *)point;

			// illegal target
			if (hlp->target >= NUMTARGETS){
				ret = -EINVAL;
				goto free_new;
			}

			// requested functionality not in kernel
			if ( (hlp->target == EPF_LOG && !(in_kernel & EPF_IN_KERNEL_LOG)) ||
			     ( (hlp->target == EPF_SNAT || hlp->target == EPF_DNAT) && !(in_kernel & EPF_IN_KERNEL_NAT)) ||
			     (hlp->bitmask & EPF_IP && !(in_kernel & EPF_IN_KERNEL_IPF)) ){
				ret = -EINVAL;
				goto free_new;
			}
			/* We need to know how much space to allocate for all the additional info.
			 * We'll allocate one hunk of memory
			 */
			if (hlp->target == EPF_LOG || hlp->target == EPF_SNAT || hlp->target == EPF_DNAT || hlp->bitmask & EPF_IP){
				if (!(hlp->additional_info)){
					ret = -EINVAL;
					goto free_new;
				}
				if (hlp->target == EPF_LOG)
					add_info_size += sizeof(struct epf_log_info);
				if (hlp->target == EPF_SNAT || hlp->target == EPF_DNAT)
					add_info_size += sizeof(struct epf_nat_info);
				if (hlp->bitmask & EPF_IP)
					add_info_size += sizeof(struct epf_ip_info);
			}
			else /* we are really tough on the userspace program */
				if (hlp->additional_info){
					ret = -EINVAL;
					goto free_new;
				}

			// obviously, we can only check ip related stuff when the protocol number is that for ip
			if (hlp->bitmask & EPF_IP && hlp->ethproto != htons(0x0800)){
				ret = -EINVAL;
				goto free_new;
			}
			point += sizeof(struct epf_entry);
		}
	}

	/* We first deal with the additional info. If there's an error
	 * we can deal with it before getting any locks and stuff.
	 * The LOG target uses this facility.
	 */
	if (add_info_size){
		char *addinfo_pointer;
		if ( !(new_addinfos = (char *)vmalloc(add_info_size)) ){
			ret = -ENOMEM;
			goto free_new;
		}
		/* we copy the info from userspace and set up the pointers */
		addinfo_pointer = new_addinfos;
		point = new_;
		for (i = 0; i < numc[table_nr]; i++){
			point += sizeof(struct epf_entries);
			for (j = 0; j < tmp.nentries[i]; j++){
				struct epf_entry *hlp = (struct epf_entry *)point;
				unsigned int addinfo_offset = 0;
				char *userspace_p = hlp->additional_info;
				// the order in which info is added to the same epf_entry must be the
				// same as the order used in br_hook().
				// the target info has to be the last in the order, because the target info
				// needs to be valuated after the other info. An other order would cause
				// unnecessary dificulty in br_hook().
				if (hlp->bitmask & EPF_IP){
				 	if (copy_from_user(addinfo_pointer, userspace_p + addinfo_offset, sizeof(struct epf_ip_info))){
						ret = -EFAULT;
						goto free_additional;
					}
					if (!addinfo_offset)
						hlp->additional_info = addinfo_pointer;
					addinfo_pointer += sizeof(struct epf_ip_info);
					addinfo_offset += sizeof(struct epf_ip_info);
				}
				if (hlp->target == EPF_LOG){
				 	if (copy_from_user(addinfo_pointer, userspace_p + addinfo_offset, sizeof(struct epf_log_info))){
						ret = -EFAULT;
						goto free_additional;
					}
					if (!addinfo_offset)
						hlp->additional_info = addinfo_pointer;
					addinfo_pointer += sizeof(struct epf_log_info);
					addinfo_offset += sizeof(struct epf_log_info);
				}
				if (hlp->target == EPF_SNAT || hlp->target == EPF_DNAT){
					if (copy_from_user(addinfo_pointer, userspace_p + addinfo_offset, sizeof(struct epf_nat_info))){
						ret = -EFAULT;
						goto free_additional;
					}
					if (!addinfo_offset)
						hlp->additional_info = addinfo_pointer;
					addinfo_pointer += sizeof(struct epf_nat_info);
					addinfo_offset += sizeof(struct epf_nat_info);
				}
				point += sizeof(struct epf_entry);
			}
		}
	}

	write_lock_bh(&epf_lock);

		if (allowc[table_nr] == EPF_COUNTERS){
			for (i = 0, oldcz = 0; i < numc[table_nr]; i++)
				oldcz += pointers[table_nr][i]->n;
			/* copy our counters to userspace
			 * userspace must make sure it has allocated enough memory
			 */
			if (oldcz){
				if ( !(counterstmp = (struct epf_counter *)vmalloc(oldcz * sizeof(struct epf_counter))) ){
					ret = -ENOMEM;
					goto free_new_unlock;
				}
				memcpy(counterstmp, counters[table_nr][0], oldcz * sizeof(struct epf_counter));
				for (i = 1; i < smp_num_cpus; i++)
					for (j = 0; j < oldcz; j++)
						counterstmp[j].pcnt += counters[table_nr][i][j].pcnt;
				if (copy_to_user(user, counterstmp, oldcz * sizeof(struct epf_counter))){
					vfree(counterstmp);
					ret = -EFAULT;
					goto free_new_unlock;
				}
				vfree(counterstmp);
			}
			if (countersize){
				if ( !(newcounters = (struct epf_counter **)vmalloc(smp_num_cpus * sizeof(struct epf_counter *))) ){
					ret = -ENOMEM;
					goto free_new_unlock;
				}
				/* make new counters (initialized to 0) */
				for (i = 0; i < smp_num_cpus; i++){
					if ( !(newcounters[i] = (struct epf_counter *)vmalloc(countersize)) ){
						ret = -ENOMEM;
						goto undo;
					}
					memset(newcounters[i], 0, countersize);
				}
				/* delete old counters */
				if (counters[table_nr]){
					for (i = 0; i < smp_num_cpus; i++)
						vfree(counters[table_nr][i]);
					vfree(counters[table_nr]);
				}
				counters[table_nr] = newcounters;
			}
			else{
				/* just delete old counters, no longer needed */
				if (counters[table_nr]){
					for (i = 0; i < smp_num_cpus; i++)
						vfree(counters[table_nr][i]);
					vfree(counters[table_nr]);
					counters[table_nr] = NULL;
				}
			}
		}

		/* make our new table */
		vfree(table[table_nr]);
		table[table_nr] = new_;
		/* set up the additional info */
		if (additional_infos[table_nr])
			vfree(additional_infos[table_nr]);
		additional_infos[table_nr] = new_addinfos;
		additional_size[table_nr] = add_info_size;
		/* update pointers[] */
		point = table[table_nr];
		for (i = 0; i < numc[table_nr]; i++){
			pointers[table_nr][i] = (struct epf_entries*)point;
			point += tmp.nentries[i] * sizeof(struct epf_entry) + sizeof(struct epf_entries);
		}

	write_unlock_bh(&epf_lock);
	return 0;

undo:
	for (j = 0; j < i; j++)
		vfree(newcounters[j]);
	vfree(newcounters);
	printk("qdsf\n");
free_new_unlock:
	write_unlock_bh(&epf_lock);
	printk("qdsf\n");
free_additional:
	if (new_addinfos)
		vfree(new_addinfos);
	printk("qdsf\n");
free_new:
	vfree(new_);
	return ret;
}

/* userspace just supplied us with counters */
static int update_counters(void *user, unsigned int len){
	int i, number = 0;
	struct epf_counter *point, *point2, *tmp;
	unsigned char table_nr;

	if (len < sizeof(unsigned char))
		return -EINVAL;
	if (copy_from_user(&table_nr, user, sizeof(unsigned char)))
		return -EFAULT;
	if ((table_nr == EPF_NATTABLE && !(in_kernel & EPF_IN_KERNEL_NAT)) ||
	    (table_nr == EPF_FILTERTABLE && !(in_kernel & EPF_IN_KERNEL_FILTER)) || table_nr >= EPF_NUMTABLES)
		return -EINVAL;
	for (i = 0; i < numc[table_nr]; i++)
		number += pointers[table_nr][i]->n;
	if (len != number * sizeof(struct epf_counter) + sizeof(unsigned char)){
		BUGPRINT("update_counters && !samelength\n");
		return -EINVAL;
	}
	if ( !(tmp = (struct epf_counter *)vmalloc(number *sizeof(struct epf_counter))) ){
		MEMPRINT("updata_counters && nomemory\n");
		return -ENOMEM;
	}
	if ( copy_from_user(tmp, user + sizeof(unsigned char), number * sizeof(struct epf_counter)) ){
		BUGPRINT("updata_counters && !cfu\n");
		vfree(tmp);
		return -EFAULT;
	}
	/* we add to the first cpu's counters */
	point = counters[table_nr][0];
	point2 = tmp;
	for (i = 0; i < number; i++){
		point->pcnt += point2->pcnt;
		point++;
		point2++;
	}
	vfree(tmp);
	return 0;
}

static int copy_entries_to_user(void *user, int *len, unsigned char table_nr){
	int test = numc[table_nr] * sizeof(struct epf_entries), test2 = 0, i, j, number = 0;
	struct epf_counter *countershlp = NULL; /* NULL keeps gcc happy */

	/* check size */
	for (i = 0; i < numc[table_nr]; i++)
		number += pointers[table_nr][i]->n;
	test += number * sizeof(struct epf_entry);
	if (allowc[table_nr] == EPF_COUNTERS)
		test2 = number * sizeof(struct epf_counter);
	if ( (test + test2) != *len){
		BUGPRINT("cetu && !samelength\n");
		return -EINVAL;
	}
	if (allowc[table_nr] == EPF_COUNTERS && counters[table_nr]){

		if (!test2){// this should really never happen
			BUGPRINT("cetu && !test2\n");
			return -EFAULT;
		}
		if ( !(countershlp = (struct epf_counter *)vmalloc(test2)) ){
			MEMPRINT("cetu && nomemory\n");
			return -ENOMEM;
		}
		memcpy(countershlp, counters[table_nr][0], test2);
		for (i = 1; i < smp_num_cpus; i++)
			for (j = 0; j < number; j++)
				countershlp[j].pcnt += counters[table_nr][i][j].pcnt;

		if (copy_to_user(user + test, countershlp, test2)){
			BUGPRINT("cetu && !ctu\n");
			vfree(countershlp);
			return -EFAULT;
		}
		vfree(countershlp);
	}
	if (copy_to_user(user, table[table_nr], test)){
		BUGPRINT("cetu && !ctu2\n");
		return -EFAULT;
	}
	return 0;
}

static int
do_epf_set_ctl(struct sock *sk,	int cmd, void *user, unsigned int len)
{
	int ret;
	unsigned char table_nr;
	unsigned short allc = 10;// makes gcc happy

	switch(cmd) {
	case EPF_SO_SET_ENTRIES:
		ret = do_replace(user, len);
		break;
	case EPF_SO_SET_COUNTERS:
		write_lock_bh(&epf_lock);
		ret = update_counters(user, len);
		write_unlock_bh(&epf_lock);
		break;
	case EPF_SO_SET_ALLOWC:
		if (len != sizeof(unsigned short) + sizeof(unsigned char)){
			ret = -EINVAL;
			break;
		}
	 	if ( copy_from_user(&table_nr, user, sizeof(unsigned char)) ){
			ret = -EFAULT;
	 		break;
		}
		if ((table_nr == EPF_NATTABLE && !(in_kernel & EPF_IN_KERNEL_NAT)) ||
		    (table_nr == EPF_FILTERTABLE && !(in_kernel & EPF_IN_KERNEL_FILTER)) ||
		     table_nr >= EPF_NUMTABLES){
			ret = -EINVAL;
			break;
		}
	 	if ( copy_from_user(&allc, user + sizeof(unsigned char), sizeof(unsigned short)) ){
			ret = -EFAULT;
	 		break;
		}
		if (allc != EPF_COUNTERS && allc != EPF_NOCOUNTERS){
			ret = -EINVAL;
			break;
		}
		write_lock_bh(&epf_lock);
		if (allc == allowc[table_nr]){
			ret = 0;
			write_unlock_bh(&epf_lock);
			break;
		}
		if (allowc[table_nr] == EPF_COUNTERS)
			ret = switch_nocounters(table_nr);
		else
			ret = switch_counters(table_nr);
		write_unlock_bh(&epf_lock);
		break;
	default:
		ret = -EINVAL;
  }
	return ret;
}

static int
do_epf_get_ctl(struct sock *sk,	int cmd, void *user, int *len)
{
	struct epf_info help;
	int i, ret;
	unsigned char table_nr;


	switch(cmd) {
	case EPF_SO_GET_IN_KERNEL:
		if (*len != sizeof(unsigned int))
			return -EINVAL;
 		if (copy_to_user(user, &in_kernel, sizeof(unsigned int)))
			return -EFAULT;
		return 0;
	}		
	if (*len < sizeof(unsigned char))
		return -EINVAL;
	if (copy_from_user(&table_nr, user, sizeof(unsigned char)) != 0)
		return -EFAULT;
	if ((table_nr == EPF_NATTABLE && !(in_kernel & EPF_IN_KERNEL_NAT)) ||
	    (table_nr == EPF_FILTERTABLE && !(in_kernel & EPF_IN_KERNEL_FILTER)))
		return -EINVAL;

	switch(cmd) {
	case EPF_SO_GET_INFO:
		if (table_nr >= EPF_NUMTABLES){
			ret = -EINVAL;
			break;
		}
		if (*len != sizeof(struct epf_info)){
			ret = -EINVAL;
			break;
		}
		read_lock_bh(&epf_lock);
			for (i = 0; i < numc[table_nr]; i++)
				help.nentries[i] = pointers[table_nr][i]->n;
			help.allow_counters = allowc[table_nr];
			if (copy_to_user(user, &help, *len) != 0){
				read_unlock_bh(&epf_lock);
				ret = -EFAULT;
				break;
			}
		read_unlock_bh(&epf_lock);
		ret = 0;
		break;

	case EPF_SO_GET_ENTRIES:
		if (table_nr >= EPF_NUMTABLES){
			ret = -EINVAL;
			break;
		}
		write_lock_bh(&epf_lock);
		ret = copy_entries_to_user(user, len, table_nr);
		write_unlock_bh(&epf_lock);
		break;			

	case EPF_SO_GET_ADDITIONAL:
		if (table_nr >= EPF_NUMTABLES){
			ret = -EINVAL;
			break;
		}
		read_lock_bh(&epf_lock);
		if (*len != additional_size[table_nr]){
			read_unlock_bh(&epf_lock);
			ret = -EINVAL;
			break;
		}
		if (copy_to_user(user, additional_infos[table_nr], *len) != 0){
			read_unlock_bh(&epf_lock);
			ret = -EFAULT;
			break;
		}
		read_unlock_bh(&epf_lock);
		ret = 0;
		break;

	default:
		ret = -EINVAL;
	}

	return ret;
}

static struct nf_sockopt_ops epf_sockopts
= { { NULL, NULL }, PF_INET, EPF_BASE_CTL, EPF_SO_SET_MAX+1, do_epf_set_ctl,
    EPF_BASE_CTL, EPF_SO_GET_MAX+1, do_epf_get_ctl, 0, NULL  };


#ifdef CONFIG_BRIDGE_EBT_FILTER
static struct nf_hook_ops epf_br_ops_filter[6] = {
	{ { NULL, NULL }, br_filter, PF_BRIDGE, NF_BR_LOCAL_IN, -200},
	{ { NULL, NULL }, br_filter, PF_BRIDGE, NF_BR_FORWARD, -200},
	{ { NULL, NULL }, br_filter, PF_BRIDGE, NF_BR_LOCAL_OUT, -200},
};
#endif
#ifdef CONFIG_BRIDGE_EBT_NAT
static struct nf_hook_ops epf_br_ops_nat[6] = {
	{ { NULL, NULL }, br_nat, PF_BRIDGE, NF_BR_LOCAL_OUT, -300},
	{ { NULL, NULL }, br_nat, PF_BRIDGE, NF_BR_POST_ROUTING, -300},
	{ { NULL, NULL }, br_nat, PF_BRIDGE, NF_BR_PRE_ROUTING, -300}
};
#endif

/* Copyright (C) 1998 by Ori Pomerantz */
/* Print the string to the appropriate tty, the one
 * the current task uses */
void print_string(char *str)
{
  struct tty_struct *my_tty;

  /* The tty for the current task */
  my_tty = current->tty;

  /* If my_tty is NULL, it means that the current task
   * has no tty you can print to (this is possible, for
   * example, if it's a daemon). In this case, there's
   * nothing we can do. */
  if (my_tty != NULL) {

    /* my_tty->driver is a struct which holds the tty's
     * functions, one of which (write) is used to
     * write strings to the tty. It can be used to take
     * a string either from the user's memory segment
     * or the kernel's memory segment.
     *
     * The function's first parameter is the tty to
     * write to, because the  same function would
     * normally be used for all tty's of a certain type.
     * The second parameter controls whether the
     * function receives a string from kernel memory
     * (false, 0) or from user memory (true, non zero).
     * The third parameter is a pointer to a string,
     * and the fourth parameter is the length of
     * the string.
     */
    (*(my_tty->driver).write)(
        my_tty, /* The tty itself */
        0, /* We don't take the string from user space */
        str, /* String */
        strlen(str));  /* Length */

    /* ttys were originally hardware devices, which
     * (usually) adhered strictly to the ASCII standard.
     * According to ASCII, to move to a new line you
     * need two characters, a carriage return and a
     * line feed. In Unix, on the other hand, the
     * ASCII line feed is used for both purposes - so
     * we can't just use \n, because it wouldn't have
     * a carriage return and the next line will
     * start at the column right
     *                          after the line feed.
     *
     * BTW, this is the reason why the text file
     *  is different between Unix and Windows.
     * In CP/M and its derivatives, such as MS-DOS and
     * Windows, the ASCII standard was strictly
     * adhered to, and therefore a new line requires
     * both a line feed and a carriage return.
     */
    (*(my_tty->driver).write)(
      my_tty,
      0,
      "\015\012",
      2);
  }
}

static int __init init(void)
{
	int ret, i, j;

#ifdef CONFIG_BRIDGE_EBT_FILTER
	for (i = 0; i < 3; i++)
		if ((ret = nf_register_hook(&epf_br_ops_filter[i])) < 0)
			goto clean11;
#endif
#ifdef CONFIG_BRIDGE_EBT_NAT
	for (i = 0; i < 3; i++)
		if ((ret = nf_register_hook(&epf_br_ops_nat[i])) < 0)
			goto clean1;
#endif

	/* Register setsockopt */
	if ((ret = nf_register_sockopt(&epf_sockopts)) < 0)
		goto clean2;

#ifdef CONFIG_BRIDGE_EBT_FILTER
	in_kernel |= EPF_IN_KERNEL_FILTER;
#endif
#ifdef CONFIG_BRIDGE_EBT_NAT
	in_kernel |= EPF_IN_KERNEL_NAT;
#endif
#ifdef CONFIG_BRIDGE_EBT_LOG
	in_kernel |= EPF_IN_KERNEL_LOG;
#endif
#ifdef CONFIG_BRIDGE_EBT_IPF
	in_kernel |= EPF_IN_KERNEL_IPF;
#endif

	for (j = 0; j < EPF_NUMTABLES; j++){
		if ((j == EPF_FILTERTABLE && (in_kernel & EPF_IN_KERNEL_FILTER)) ||
		    (j == EPF_NATTABLE && (in_kernel & EPF_IN_KERNEL_NAT))){
			allowc[j] = EPF_NOCOUNTERS;
			counters[j] = NULL;
			additional_infos[j] = NULL;
			additional_size[j] = 0;
			if ( !(pointers[j] = vmalloc(sizeof(struct epf_entries *) * numc[j])) )
				goto free_pointers;
		}
	}

	for (j = 0; j < EPF_NUMTABLES; j++){
		if ((j == EPF_FILTERTABLE && (in_kernel & EPF_IN_KERNEL_FILTER)) ||
		    (j == EPF_NATTABLE && (in_kernel & EPF_IN_KERNEL_NAT))){
			if ( (table[j] = vmalloc(sizeof(struct epf_entries) * numc[j])) ){
 		   memcpy(table[j], &start, numc[j] * sizeof(struct epf_entries));
				for ( i = 0; i < numc[j]; i++)
					pointers[j][i] = (struct epf_entries *)(table[j] + i * sizeof(struct epf_entries));
			}
			else
				goto free_table;
		}
	}
	rwlock_init(&epf_lock);
  print_string("Bridge ethernet protocol filter registered");

	return ret;
free_table:
	for (i = 0; i < j; i++)
		if ((i == EPF_FILTERTABLE && (in_kernel & EPF_IN_KERNEL_FILTER)) ||
		    (i == EPF_NATTABLE && (in_kernel & EPF_IN_KERNEL_NAT)))
			vfree(table[i]);
	j = EPF_NUMTABLES;
free_pointers:
	for (i = 0; i < j; i++)
		if ((i == EPF_FILTERTABLE && (in_kernel & EPF_IN_KERNEL_FILTER)) ||
		    (i == EPF_NATTABLE && (in_kernel & EPF_IN_KERNEL_NAT)))
			vfree(pointers[i]);
	nf_unregister_sockopt(&epf_sockopts);
clean2:
#ifdef CONFIG_BRIDGE_EBT_NAT
	i = 3;
clean1:		
	for (j = 0; j < i; j++)
		nf_unregister_hook(&epf_br_ops_nat[j]);
#endif
#ifdef CONFIG_BRIDGE_EBT_FILTER
	i = 3;
clean11:
	for (j = 0; j < i; j++)
		nf_unregister_hook(&epf_br_ops_filter[j]);
#endif
	ret = -EFAULT;
  print_string("Bridge ethernet protocol filter couldn't be registered");
	return ret;
}

static void __exit fini(void){
	int i, j;

	write_lock_bh(&epf_lock);// we want total control
	for (i = 0; i < EPF_NUMTABLES; i++){
		if ((i == EPF_FILTERTABLE && (in_kernel & EPF_IN_KERNEL_FILTER)) ||
		    (i == EPF_NATTABLE && (in_kernel & EPF_IN_KERNEL_NAT))){
			if (additional_infos[i])
				vfree(additional_infos[i]);
			if (pointers[i])
				vfree(pointers[i]);
			if (table[i])
				vfree(table[i]);
			if (counters[i]){
				for (j = 0; j < smp_num_cpus; j++)
					vfree(counters[i][j]);
				vfree(counters[i]);
			}
		}
	}

#ifdef CONFIG_BRIDGE_EBT_NAT
	for (i = 0; i < 3; i++)
		nf_unregister_hook(&epf_br_ops_nat[i]);
#endif
#ifdef CONFIG_BRIDGE_EBT_FILTER
	for (i = 0; i < 3; i++)
		nf_unregister_hook(&epf_br_ops_filter[i]);
#endif
	nf_unregister_sockopt(&epf_sockopts);
	write_unlock_bh(&epf_lock);
  print_string("Bridge ethernet protocol filter unregistered");
}

module_init(init);
module_exit(fini);
