#include <linux/module.h>
#include <linux/netfilter.h>
#include <linux/ip.h>
#include <net/checksum.h>
#include <net/udp.h>

#include <linux/netfilter_ipv4/lockhelp.h>
#include <linux/netfilter_ipv4/ip_conntrack_helper.h>

DECLARE_LOCK(ip_ipsec_lock);
struct module *ip_conntrack_ipsec = THIS_MODULE;

#define	IPSEC_PORT 500
#define	IPSEC_NAT_T_PORT 4500

#if 0
#define DEBUGP printk
#else
#define DEBUGP(format, args...)
#endif

static int ipsec_help_msg(const struct iphdr *iph, size_t len,
		          struct ip_conntrack *ct,
		         enum ip_conntrack_info ctinfo)
{
	int dir = CTINFO2DIR(ctinfo);
	struct ip_conntrack_expect expect, *exp = &expect;

	if (dir != IP_CT_DIR_ORIGINAL)
                return NF_ACCEPT;

	LOCK_BH(&ip_ipsec_lock);
	exp->tuple = ((struct ip_conntrack_tuple)
		{ { ct->tuplehash[dir].tuple.src.ip,
		    { 0 } },
		  { ct->tuplehash[dir].tuple.dst.ip,
		    { 0 },
		    IPPROTO_UDP }});
	exp->mask = ((struct ip_conntrack_tuple)
		{ { 0xFFFFFFFF, { 0 } },
		  { 0xFFFFFFFF, { 0 }, 0xFFFF }});

	exp->expectfn = NULL;

	DEBUGP("ip_ct_ipsec_help_msg: client %u.%u.%u.%u:%u -> %u.%u.%u.%u:%u!\n",
	       NIPQUAD(exp->tuple.src.ip), ntohs(exp->tuple.src.u.tcp.port),
	       NIPQUAD(exp->tuple.dst.ip), ntohs(exp->tuple.dst.u.tcp.port));

	ip_conntrack_expect_related(ct, exp);
	UNLOCK_BH(&ip_ipsec_lock);
		    
	return NF_ACCEPT;
}

static int help(const struct iphdr *iph,
		 size_t len,
		 struct ip_conntrack *ct,
		 enum ip_conntrack_info ctinfo)
{
	struct udphdr *udph = (void *)iph + iph->ihl * 4;
	size_t udplen;

	DEBUGP("ip_ct_ipsec_help: help entered\n");

	if (ctinfo != IP_CT_ESTABLISHED
	    && ctinfo != IP_CT_ESTABLISHED + IP_CT_IS_REPLY) {
		DEBUGP("ip_ct_ipsec_help: Conntrackinfo = %u\n", ctinfo);
		return NF_ACCEPT;
	}

	/* Not whole UDP header? */
	udplen = len - iph->ihl * 4;
	if (udplen < sizeof(struct udphdr)) {
		DEBUGP("ip_ct_ipsec_help: too short for udph, udplen = %u\n", (unsigned)udplen);
		return NF_ACCEPT;
	}

	/* Checksum invalid?  Ignore. */
	/* FIXME: Source route IP option packets --RR */
	if (csum_tcpudp_magic(iph->saddr, iph->daddr, udplen, IPPROTO_UDP,
			      csum_partial((char *)udph, udplen, 0))) {
		DEBUGP("ip_ct_ipsec_help: bad csum: %p %u %u.%u.%u.%u %u.%u.%u.%u\n",
		       udph, udplen, NIPQUAD(iph->saddr),
		       NIPQUAD(iph->daddr));
		return NF_ACCEPT;
	}

	ipsec_help_msg(iph, len, ct, ctinfo);
	return NF_ACCEPT;
}

static struct ip_conntrack_helper ipsec_helpers =
					      { { NULL, NULL },
					        { { 0, { __constant_htons(IPSEC_PORT) } },
					          { 0, { 0 }, IPPROTO_UDP } },
					        { { 0, { 0xFFFF } },
					          { 0, { 0 }, 0xFFFF } },
					        8,
					        help };

static int __init init(void)
{
	ip_conntrack_helper_register(&ipsec_helpers);
	return 0;
}

static void __exit fini(void)
{
	ip_conntrack_helper_unregister(&ipsec_helpers);
}

module_init(init);
module_exit(fini);
