/* 
 * talk extension for IP connection tracking. 
 * Jozsef Kadlecsik <kadlec@blackhole.kfki.hu>
 *
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License
 *      as published by the Free Software Foundation; either version
 *      2 of the License, or (at your option) any later version.
 **
 *     Module load syntax:
 *     insmod ip_conntrack_talk.o otalk=[0|1] talk=[0|1]
 *
 *     The default is otalk=1 talk=1
 *     i.e. support both old talk and talk protocols
 *
 **
 *
 *		ASCII art on talk protocols
 *	
 *	
 *	caller server		    callee server
 *		|     \	          /
 *		|	\       /
 *		|	  \   /
 *		|	    /  
 *	 	|	  /   \
 *	      2 |     1 /       \ 3
 *	caller client  ----------- callee client
 *	               		 4
 *
 *	1. caller client <-> callee server: LOOK_UP, then ANNOUNCE invitation 
 *    ( 2. caller client <-> caller server: LEAVE_INVITE to server )
 *	3. callee client <-> caller server: LOOK_UP invitation
 *	4. callee client <-> caller client: talk data channel
 *
 * [1]: M. Hunter, talk: a historical protocol for interactive communication
 *      draft-hunter-talk-00.txt
 * [2]: D.B. Chapman, E.D. Zwicky: Building Internet Firewalls (O'Reilly)	
 */
#include <linux/module.h>
#include <linux/netfilter.h>
#include <linux/ip.h>
#include <net/checksum.h>
#include <net/udp.h>

#include <linux/netfilter_ipv4/lockhelp.h>
#include <linux/netfilter_ipv4/ip_conntrack_helper.h>
#include <linux/netfilter_ipv4/ip_conntrack_talk.h>

/* Default both old talk and talk are supported */
static int otalk = 1;
static int talk = 1;
MODULE_AUTHOR("Jozsef Kadlecsik <kadlec@blackhole.kfki.hu>");
MODULE_DESCRIPTION("talk connection tracking module");
#ifdef MODULE_PARM
MODULE_PARM(otalk, "i");
MODULE_PARM_DESC(otalk, "support old talk protocol");
MODULE_PARM(talk, "i");
MODULE_PARM_DESC(talk, "support talk protocol");
#endif

DECLARE_LOCK(ip_talk_lock);
struct module *ip_conntrack_talk = THIS_MODULE;

#if 0
#define DEBUGP printk
#else
#define DEBUGP(format, args...)
#endif

static int talk_help_msg(const struct iphdr *iph, size_t len,
		         struct ip_conntrack *ct,
		         enum ip_conntrack_info ctinfo,
		         int version,
		         u_char type,
		         struct talk_addr *addr,
		         struct talk_addr *ctl_addr)
{
	int dir = CTINFO2DIR(ctinfo);
	struct ip_conntrack_expect expect, *exp = &expect;
	struct ip_ct_talk_expect *exp_talk_info = &exp->help.exp_talk_info;

	DEBUGP("ip_ct_talk_help_msg: %u.%u.%u.%u:%u, ctl %u.%u.%u.%u:%u, type %d\n",
		NIPQUAD(addr->ta_addr), ntohs(addr->ta_port),
		NIPQUAD(ctl_addr->ta_addr), ntohs(ctl_addr->ta_port), type);

	if (type != ANNOUNCE) 
		return NF_ACCEPT;

	DEBUGP("ip_ct_talk_help_msg: match, %u.%u.%u.%u:%u, ctl %u.%u.%u.%u:%u!\n",
		NIPQUAD(addr->ta_addr), ntohs(addr->ta_port),
		NIPQUAD(ctl_addr->ta_addr), ntohs(ctl_addr->ta_port));

	/* update the talk info */
	LOCK_BH(&ip_talk_lock);
	exp_talk_info->port = htons(version ? TALK_PORT : OTALK_PORT);

	/* expect callee client -> caller server message */
	exp->tuple = ((struct ip_conntrack_tuple)
		{ { ct->tuplehash[!dir].tuple.src.ip,
		    { 0 } },
		  { ctl_addr->ta_addr,
		    { htons(version ? TALK_PORT : OTALK_PORT) },
		    IPPROTO_UDP }});
	exp->mask = ((struct ip_conntrack_tuple)
		{ { 0xFFFFFFFF, { 0 } },
		  { 0xFFFFFFFF, { 0xFFFF }, 0xFFFF }});

	exp->expectfn = NULL;

	DEBUGP("ip_ct_talk_help_msg: callee client %u.%u.%u.%u:%u -> caller daemon %u.%u.%u.%u:%u!\n",
	       NIPQUAD(exp->tuple.src.ip), ntohs(exp->tuple.src.u.tcp.port),
	       NIPQUAD(exp->tuple.dst.ip), ntohs(exp->tuple.dst.u.tcp.port));

	/* Ignore failure; should only happen with NAT */
	ip_conntrack_expect_related(ct, &expect);
	UNLOCK_BH(&ip_talk_lock);
		    
	return NF_ACCEPT;
}

static int talk_help_response(const struct iphdr *iph, size_t len,
		              struct ip_conntrack *ct,
		              enum ip_conntrack_info ctinfo,
		              int version,
		              u_char type,
		              u_char answer,
		              struct talk_addr *addr)
{
	int dir = CTINFO2DIR(ctinfo);
	struct ip_conntrack_expect expect, *exp = &expect;
	struct ip_ct_talk_expect *exp_talk_info = &exp->help.exp_talk_info;

	DEBUGP("ip_ct_talk_help_response: %u.%u.%u.%u:%u, type %d answer %d\n",
		NIPQUAD(addr->ta_addr), ntohs(addr->ta_port),
		type, answer);

	if (!(type == LOOK_UP 
	      && answer == SUCCESS))
		return NF_ACCEPT;

	DEBUGP("ip_ct_talk_help_response: match, %u.%u.%u.%u:%u!\n",
		NIPQUAD(addr->ta_addr), ntohs(addr->ta_port));

	/* update the talk info */
	LOCK_BH(&ip_talk_lock);
	exp_talk_info->port = addr->ta_port;

	/* expect callee client -> caller client connection */
	exp->tuple = ((struct ip_conntrack_tuple)
		{ { ct->tuplehash[!dir].tuple.src.ip,
		    { 0 } },
		  { addr->ta_addr,
		    { addr->ta_port },
		    IPPROTO_TCP }});
	exp->mask = ((struct ip_conntrack_tuple)
		{ { 0xFFFFFFFF, { 0 } },
		  { 0xFFFFFFFF, { 0xFFFF }, 0xFFFF }});

	exp->expectfn = NULL;

	DEBUGP("ip_ct_talk_help_response: callee client %u.%u.%u.%u:%u -> caller client %u.%u.%u.%u:%u!\n",
	       NIPQUAD(exp->tuple.src.ip), ntohs(exp->tuple.src.u.tcp.port),
	       NIPQUAD(exp->tuple.dst.ip), ntohs(exp->tuple.dst.u.tcp.port));

	/* Ignore failure; should only happen with NAT */
	ip_conntrack_expect_related(ct, &expect);
	UNLOCK_BH(&ip_talk_lock);
		    
	return NF_ACCEPT;
}

/* FIXME: This should be in userspace.  Later. */
static int talk_help(const struct iphdr *iph, size_t len,
		     struct ip_conntrack *ct,
		     enum ip_conntrack_info ctinfo,
		     int version)
{
	struct udphdr *udph = (void *)iph + iph->ihl * 4;
	const char *data = (const char *)udph + sizeof(struct udphdr);
	size_t udplen;

	DEBUGP("ip_ct_talk_help: help entered\n");

	/* Until there's been traffic both ways, don't look in packets. */
	if (ctinfo != IP_CT_ESTABLISHED
	    && ctinfo != IP_CT_ESTABLISHED + IP_CT_IS_REPLY) {
		DEBUGP("ip_ct_talk_help: Conntrackinfo = %u\n", ctinfo);
		return NF_ACCEPT;
	}

	/* Not whole UDP header? */
	udplen = len - iph->ihl * 4;
	if (udplen < sizeof(struct udphdr)) {
		DEBUGP("ip_ct_talk_help: too short for udph, udplen = %u\n", (unsigned)udplen);
		return NF_ACCEPT;
	}

	/* Checksum invalid?  Ignore. */
	/* FIXME: Source route IP option packets --RR */
	if (csum_tcpudp_magic(iph->saddr, iph->daddr, udplen, IPPROTO_UDP,
			      csum_partial((char *)udph, udplen, 0))) {
		DEBUGP("ip_ct_talk_help: bad csum: %p %u %u.%u.%u.%u %u.%u.%u.%u\n",
		       udph, udplen, NIPQUAD(iph->saddr),
		       NIPQUAD(iph->daddr));
		return NF_ACCEPT;
	}
	
	DEBUGP("ip_ct_talk_help: %u.%u.%u.%u:%u->%u.%u.%u.%u:%u\n",
		NIPQUAD(iph->saddr), ntohs(udph->source), NIPQUAD(iph->daddr), ntohs(udph->dest));

	if (udplen == sizeof(struct udphdr) 
		      + (version ? sizeof(struct talk_msg) : sizeof(struct otalk_msg)))
		return talk_help_msg(iph, len, ct, ctinfo, version,
					  TALK_MSG(data, version, type), 
					  &(TALK_MSG(data, version, addr)),
					  &(TALK_MSG(data, version, ctl_addr)));
	else if (udplen == sizeof(struct udphdr) 
			   + (version ? sizeof(struct talk_response) : sizeof(struct otalk_response)))
		return talk_help_response(iph, len, ct, ctinfo, version,
					  TALK_RESP(data, version, type), 
					  TALK_RESP(data, version, answer),
					  &(TALK_RESP(data, version, addr)));
	else {	
		DEBUGP("ip_ct_talk_help: not talk message/response, datalen = %u\n", 
		       (unsigned)udplen - sizeof(struct udphdr));
		return NF_ACCEPT;
	}
}

static int ohelp(const struct iphdr *iph,
		 size_t len,
		 struct ip_conntrack *ct,
		 enum ip_conntrack_info ctinfo)
{
	return talk_help(iph, len, ct, ctinfo, OTALK_VERSION);
}

static int help(const struct iphdr *iph,
		size_t len,
		struct ip_conntrack *ct,
		enum ip_conntrack_info ctinfo)
{
	return talk_help(iph, len, ct, ctinfo, TALK_VERSION);
}

static struct ip_conntrack_helper talk_helpers[2] = { 
					      { { NULL, NULL },
					        { { 0, { __constant_htons(OTALK_PORT) } },
					          { 0, { 0 }, IPPROTO_UDP } },
					        { { 0, { 0xFFFF } },
					          { 0, { 0 }, 0xFFFF } },
					        1,
					        ohelp },
                                              { { NULL, NULL },
					        { { 0, { __constant_htons(TALK_PORT) } },
					          { 0, { 0 }, IPPROTO_UDP } },
					        { { 0, { 0xFFFF } },
					          { 0, { 0 }, 0xFFFF } },
					        1,
					        help }
					    };

static int __init init(void)
{
	if (otalk > 0)
		ip_conntrack_helper_register(&talk_helpers[0]);
	if (talk > 0)
		ip_conntrack_helper_register(&talk_helpers[1]);
		
	return 0;
}

static void __exit fini(void)
{
	if (otalk > 0)
		ip_conntrack_helper_unregister(&talk_helpers[0]);
	if (talk > 0)
		ip_conntrack_helper_unregister(&talk_helpers[1]);
}

module_init(init);
module_exit(fini);
