/*
 * ip_nat_pptp.c	- Version 1.2
 *
 * NAT support for PPTP (Point to Point Tunneling Protocol).
 * PPTP is a a protocol for creating virtual private networks.
 * It is a specification defined by Microsoft and some vendors
 * working with Microsoft.  PPTP is built on top of a modified
 * version of the Internet Generic Routing Encapsulation Protocol.
 * GRE is defined in RFC 1701 and RFC 1702.  Documentation of
 * PPTP can be found in RFC 2637
 *
 * (C) 2000-2003 by Harald Welte <laforge@gnumonks.org>
 *
 * Development of this code funded by Astaro AG (http://www.astaro.com/)
 *
 * TODO: - Support for multiple calls within one session
 * 	   (needs netfilter newnat code)
 * 	 - NAT to a unique tuple, not to TCP source port
 * 	   (needs netfilter tuple reservation)
 * 
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/ip.h>
#include <linux/tcp.h>
#include <net/tcp.h>
#include <linux/netfilter_ipv4/ip_nat.h>
#include <linux/netfilter_ipv4/ip_nat_rule.h>
#include <linux/netfilter_ipv4/ip_nat_helper.h>
#include <linux/netfilter_ipv4/ip_conntrack_helper.h>
#include <linux/netfilter_ipv4/ip_conntrack_pptp.h>

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Harald Welte <laforge@gnumonks.org>");
MODULE_DESCRIPTION("Netfilter NAT helper module for PPTP");


#if 0
#include "ip_conntrack_pptp_priv.h"
#define DEBUGP printk
#else
#define DEBUGP(format, args...)
#endif

static unsigned int
pptp_nat_expected(struct sk_buff **pskb,
		  unsigned int hooknum,
		  struct ip_conntrack *ct,
		  struct ip_nat_info *info)
{
	struct ip_conntrack *master = master_ct(ct);
	struct ip_nat_multi_range mr;
	struct ip_ct_pptp_master *ct_pptp_info;
	u_int32_t newip;
	int ret = 0;
	struct iphdr *iph = (*pskb)->nh.iph;
	struct gre_hdr_pptp *grehdr = (void *)iph + iph->ihl*4;
/* 2003/10/16 Steve ++ */
	struct ip_ct_pptp_expect *exp_info;
/* Steve -- */

	IP_NF_ASSERT(info);
	IP_NF_ASSERT(master);
	IP_NF_ASSERT(!(info->initialized & (1 << HOOK2MANIP(hooknum))));

	DEBUGP("we have a connection!\n");

	LOCK_BH(&ip_pptp_lock);

	ct_pptp_info = &master->help.ct_pptp_info;
/* 2003/10/16 Steve ++ */
	exp_info = &ct->master->help.exp_pptp_info;
/* Steve -- */
	DEBUGP("server id: %x client id: %x gre version: %x protocol: %x callid: %x\n",
		htons(ct_pptp_info->pac_call_id),
		htons(ct_pptp_info->pns_call_id),
		grehdr->version,
		grehdr->protocol,
		grehdr->call_id);

	if (HOOK2MANIP(hooknum) == IP_NAT_MANIP_SRC) {
/* 2003/10/16 Steve ++ */
		if (exp_info->dir == IP_CT_DIR_ORIGINAL)
			newip = master->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.ip;
		else
			newip = master->tuplehash[IP_CT_DIR_REPLY].tuple.dst.ip;
#if 0
		if (grehdr->call_id == htons(ct_pptp_info->pns_call_id))
			newip = master->tuplehash[IP_CT_DIR_REPLY].tuple.src.ip;
		else
			newip = master->tuplehash[IP_CT_DIR_REPLY].tuple.dst.ip;
#endif
/* Steve -- */
		mr.rangesize = 1;
		mr.range[0].flags = IP_NAT_RANGE_MAP_IPS;
		mr.range[0].min_ip = mr.range[0].max_ip = newip;

		DEBUGP("change ip to %u.%u.%u.%u\n", 
			NIPQUAD(newip));
		ret = ip_nat_setup_info(ct, &mr, hooknum);
	} else {
/* 2003/10/16 Steve ++ */
		if (exp_info->dir == IP_CT_DIR_ORIGINAL)
			newip = master->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.ip;
		else
			newip = master->tuplehash[IP_CT_DIR_REPLY].tuple.src.ip;
#if 0
		if (grehdr->call_id == htons(ct_pptp_info->pns_call_id))
			newip = master->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.ip;
		else
			newip = master->tuplehash[IP_CT_DIR_REPLY].tuple.src.ip;
#endif
		
/* Steve -- */
		mr.rangesize = 1;
		mr.range[0].flags = IP_NAT_RANGE_MAP_IPS;
		mr.range[0].min_ip = mr.range[0].max_ip = newip;

		DEBUGP("change ip to %u.%u.%u.%u\n", 
			NIPQUAD(newip));
		ret = ip_nat_setup_info(ct, &mr, hooknum);
	}
	UNLOCK_BH(&ip_pptp_lock);
	return ret;

}

static unsigned int tcp_help(struct ip_conntrack *ct,
                             struct ip_conntrack_expect *exp,
                             struct ip_nat_info *info,
                             enum ip_conntrack_info ctinfo,
                             unsigned int hooknum, struct sk_buff **pskb)
{
	return NF_ACCEPT;
}

/* nat helper struct for control connection */
static struct ip_nat_helper pptp_tcp_helper = { 
	{ NULL, NULL },
	{ { 0, { __constant_htons(PPTP_CONTROL_PORT) } },
	  { 0, { 0 }, IPPROTO_TCP } },
	{ { 0, { 0xFFFF } },
	  { 0, { 0 }, 0xFFFF } },
	tcp_help, pptp_nat_expected , NULL, "PPTP" };

			  
static int __init init(void)
{
	DEBUGP("init_module\n" );

        if (ip_nat_helper_register(&pptp_tcp_helper))
		return -EIO;

        return 0;
}

static void __exit fini(void)
{
	DEBUGP("cleanup_module\n" );
        ip_nat_helper_unregister(&pptp_tcp_helper);
}

module_init(init);
module_exit(fini);
