/*
 * This is a module which is used for directly resending a received packet to
 * an interface specified by the user.
 * This module is particularly useful when the destination IP address is an
 * address owned by the router itself. Packets with such addresses are locally
 * delivered by the kernel and cannot be forwarded using the standard routing 
 * mechanisms.

 * Based on code from: ipt_MIRROR.c and various sources.
 *
 * Copyright (C) 2002 Cedric de Launois <delaunois@info.ucl.ac.be>
 *
 * This software is distributed under GNU GPL v2, 1991
 */

#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/ip.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ipt_ROUTE.h>
#include <linux/netdevice.h>
#include <linux/route.h>
#include <net/ip.h>
#include <net/icmp.h>
/* 2003/1/10 Steve ++ */
#include <net/tcp.h>
#include <net/udp.h>
/* Steve -- */

#if 0
#define DEBUGP printk
#else
#define DEBUGP(format, args...)
#endif


/* PRE : skb is the packet and ifindex is the interface index through which 
 *       the packet should be rerouted
 * POST: if success, skb->dev is the output device with index 'ifindex'
 *       to which the packet must be sent and skb->dst is not NULL
 * RETURN: -1 if an error occured
 *          1 if the packet has been succesfully rerouted to the device with
 *            index 'ifindex', using the kernel routing table
 *          0 if the kernel routing table didn't yield the requested 
 *            interface
 */
static int route(struct sk_buff *skb, unsigned int ifindex)
{
	int err;
	struct rtable *rt;
        struct iphdr *iph = skb->nh.iph;
	struct rt_key key = { 
		dst:iph->daddr,
		src:0,
		oif:ifindex, 
		tos:RT_TOS(iph->tos) 
	};

	/* Trying to route the packet with the standard routing table. */
	if ((err = ip_route_output_key(&rt, &key))) {
		DEBUGP("ipt_ROUTE: couldn't route pkt (err: %i)",err);
		return -1;
	}

	/* Check if the interface we are leaving by is the same as the
	 * one we requested.
	 */
	if (rt->u.dst.dev->ifindex == ifindex) {
		/* Drop old route. */
		dst_release(skb->dst);
		skb->dst = &rt->u.dst;
		skb->dev = skb->dst->dev;
		return 1;
	}

	/* The interface selected by the routing table is not the one
	 * specified by the user. This may happen because the dst address
	 * is one of our own addresses. This module makes it possible to
	 * reroute such packets by forcing the device to use.
	 */
	return 0;
}


/* Stolen from ip_finish_output2
 * PRE : skb->dev is set to the device we are leaving by
 *       skb->dst is not NULL
 * POST: the packet is sent with the link layer header pushed
 */
static inline int ip_direct_send(struct sk_buff *skb)
{
	struct dst_entry *dst = skb->dst;
	struct hh_cache *hh;

	hh = dst->hh;
	if (hh) {
		read_lock_bh(&hh->hh_lock);
		memcpy(skb->data - 16, hh->hh_data, 16);
		read_unlock_bh(&hh->hh_lock);
		skb_push(skb, hh->hh_len);
		return hh->hh_output(skb);
	} else if (dst->neighbour)
		return dst->neighbour->output(skb);

	if (net_ratelimit())
		DEBUGP(KERN_DEBUG "ipt_ROUTE: no hdr & no neighbour cache!\n");
	kfree_skb(skb);
	return -EINVAL;
}


/* PRE : skb->dev is set to the device we are leaving by
 * POST: - the packet is directly sent to the skb->dev device, without 
 *         pushing the link layer header.
 *       - skb has been freed !!
 */
static inline int dev_direct_send(struct sk_buff *skb)
{
	return dev_queue_xmit(skb);
}


static unsigned int ipt_route_target(struct sk_buff **pskb,
				      unsigned int hooknum,
				      const struct net_device *in,
				      const struct net_device *out,
				      const void *targinfo,
				      void *userinfo)
{
	const struct ipt_route_target_info *route_info = targinfo;
	struct net_device *dev_out = NULL;
	struct sk_buff *skb = *pskb;
/* 2003/1/10 Steve ++ */
	struct sk_buff *nskb, *oskb;
/* Steve -- */
	unsigned int index = 0;

/* 2003/1/10 Steve ++ */
	nskb = skb_copy(skb, GFP_ATOMIC);
	if (!nskb)
		return IPT_CONTINUE;

	oskb = skb;
	skb = nskb;
#if 0
	/* raw socket (tcpdump) may have clone of incoming
	 * skb: don't disturb it
	 */
	if (skb_cloned(skb) && !skb->sk) {
		struct sk_buff *nskb = skb_copy(skb, GFP_ATOMIC);
		if (!nskb)
			return NF_DROP;
		kfree_skb(skb);
		skb = nskb;
	}
#endif
/* Steve -- */

	/* If we are not at FORWARD hook
	 * the TTL isn't decreased by the IP stack
	 */
	if (hooknum != NF_IP_FORWARD) {
		struct iphdr *iph = skb->nh.iph;

		if (iph->ttl <= 1) {
			struct rtable *rt;

/* 2003/11/6 Steve ++ */
#if 0
			if (ip_route_output(&rt, iph->saddr, iph->daddr,
					    RT_TOS(iph->tos) | RTO_CONN,
					    0)) {
#endif
			if (ip_route_output(&rt, iph->saddr, iph->daddr,
					    RT_TOS(iph->tos) | RTO_CONN,
					    0, 0)) {
/* Steve -- */
/* 2003/1/10 Steve ++ */
#if 0
				return NF_DROP;
#endif
				kfree_skb(skb);
				skb = oskb;
				return IPT_CONTINUE;
/* Steve -- */
			}

			if (skb->dev == rt->u.dst.dev) {
				/* Drop old route. */
				dst_release(skb->dst);
				skb->dst = &rt->u.dst;

				/* this will traverse normal stack, and 
				 * thus call conntrack on the icmp packet */
				icmp_send(skb, ICMP_TIME_EXCEEDED, 
					  ICMP_EXC_TTL, 0);
			}
/* 2003/1/10 Steve ++ */
#if 0
			return NF_DROP;
#endif
			kfree_skb(skb);
			skb = oskb;
			return IPT_CONTINUE;
/* Steve -- */
		}

		ip_decrease_ttl(iph);
	}

	/* This packet will not be the same as before: clear nf fields */
	nf_conntrack_put(skb->nfct);
	skb->nfct = NULL;
	skb->nfcache = 0;
#ifdef CONFIG_NETFILTER_DEBUG
	skb->nf_debug = 0;
#endif

	/* Getting the actual interface index */
	if ((index = route_info->if_index) == 0) {
		/* The user specified a ifname, not a ifindex */
		if ((dev_out = dev_get_by_name(route_info->if_name)))
			index = dev_out->ifindex;
/* 2003/1/10 Steve ++ */
		else {
			kfree_skb(skb);
			skb = oskb;
			return IPT_CONTINUE;
		}
#if 0
		else return NF_DROP;
#endif
/* Steve -- */
	}

	/* Trying the standard way */
	if (route(skb, index)==1) {
/* 2002/1/10 Steve ++ */
		struct iphdr *iph = skb->nh.iph;
/* Steve -- */
		if (dev_out) dev_put(dev_out);
/* 2002/1/10 Steve ++ */
		if (route_info->ip != 0) {
			iph->check = ip_nat_cheat_check(~iph->daddr, route_info->ip,
        	                                        iph->check);

			//printk("route_info->ip = %x\n", route_info->ip);
			iph->daddr = route_info->ip;
			//printk("route_info->ip = %x\n", route_info->ip);
		} else {
			iph->check = ip_nat_cheat_check(~iph->daddr, 0xffffffff,
        	                                        iph->check);
			iph->daddr = 0xffffffff;
		}

		if (iph->protocol == IPPROTO_UDP) {
			struct udphdr *udph;
			u_int32_t udplen;

			udph = (void *)iph + iph->ihl*4;
			udplen = skb->len - iph->ihl*4;

			skb->csum = csum_partial((char *)udph + sizeof(struct udphdr),
                                             udplen - sizeof(struct udphdr),
                                             0);


			udph->check = 0;
			udph->check = csum_tcpudp_magic(iph->saddr, iph->daddr,
							udplen, IPPROTO_UDP,
							csum_partial((char *)udph, sizeof(struct udphdr),
							(skb)->csum));
		} else if (iph->protocol == IPPROTO_TCP) {
			struct tcphdr *tcph;
			u_int32_t tcplen;

			tcph = (void *)iph + iph->ihl*4;
			tcplen = skb->len - iph->ihl*4;

			skb->csum = csum_partial((char *)tcph + tcph->doff*4,
						 tcplen - tcph->doff*4, 0);

			tcph->check = 0;
			tcph->check = tcp_v4_check(tcph, tcplen, iph->saddr, iph->daddr,
						   csum_partial((char *)tcph, tcph->doff*4,
						   skb->csum));
		}
#if 0
		if (ip_direct_send(skb) == -EINVAL) return NF_DROP;
		return NF_STOLEN;
#endif
		if (ip_direct_send(skb) == -EINVAL) {
			skb = oskb;
			return IPT_CONTINUE;
		}

		skb = oskb;
		return IPT_CONTINUE;
/* Steve -- */
	}

	/* Failed to correctly route. Trying the hard way */
	if (!dev_out) dev_out = dev_get_by_index(index);

	if (dev_out) {

		/* Drop old route. */
		dst_release(skb->dst);
		skb->dst = NULL;

		/* Send the packet. This will also free skb ! 
		 * Do not go through the POST_ROUTING hook because 
		 * skb->dst is not set and because it will probably
		 * get confused by the destination IP address.
		 */
		skb->dev = dev_out;
		dev_direct_send(skb);
		dev_put(dev_out);

/* 2003/1/10 Steve ++ */
		skb = oskb;
		return IPT_CONTINUE;
#if 0
		return NF_STOLEN;
#endif
/* Steve -- */
	}

	DEBUGP("ipt_ROUTE: Couldn't route the packet!\n");
/* 2003/1/10 Steve ++ */
	kfree_skb(skb);
	skb = oskb;
	return IPT_CONTINUE;
#if 0
	return NF_DROP;
#endif
/* Steve -- */
}


static int ipt_route_checkentry(const char *tablename,
				 const struct ipt_entry *e,
				 void *targinfo,
				 unsigned int targinfosize,
				 unsigned int hook_mask)
{
	if (hook_mask & ~(1 << NF_IP_PRE_ROUTING)) {
		printk("ipt_ROUTE: bad hook\n");
		return 0;
	}

	if (targinfosize != IPT_ALIGN(sizeof(struct ipt_route_target_info))) {
		DEBUGP(KERN_WARNING "ipt_ROUTE: targinfosize %u != %Zu\n",
		       targinfosize,
		       IPT_ALIGN(sizeof(struct ipt_route_target_info)));
		return 0;
	}

	return 1;
}


static struct ipt_target ipt_route_reg
= { { NULL, NULL }, "ROUTE", ipt_route_target, ipt_route_checkentry, NULL,
    THIS_MODULE };


static int __init init(void)
{
	if (ipt_register_target(&ipt_route_reg))
		return -EINVAL;

	return 0;
}


static void __exit fini(void)
{
	ipt_unregister_target(&ipt_route_reg);
}

module_init(init);
module_exit(fini);
MODULE_LICENSE("GPL");
