#include <linux/module.h>
#include <linux/proc_fs.h>
#include <linux/spinlock.h>
#include <linux/interrupt.h>
#include <asm/uaccess.h>
#include <linux/ctype.h>
#include <linux/vmalloc.h>
#include <linux/ip.h>
#include <linux/if_ether.h>

#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ipt_blacklist.h>

struct blacklist_list {
	char mac_addr[6];
	struct blacklist_list *next;
};

struct blacklist_table {
	int count;
	struct blacklist_list *list;
};

static struct blacklist_table blacklist;
static spinlock_t list_lock = SPIN_LOCK_UNLOCKED;
static unsigned int maxtry = 0;

static void
add_blacklist(char *mac_addr)
{
	struct blacklist_list *list = blacklist.list;
	struct blacklist_list *prev, *next;

	prev = list;
	while (list) {
		if (!memcmp(list->mac_addr, mac_addr, 6))
			return;
		prev = list;
		list = list->next;
	}

	next = kmalloc(sizeof(struct blacklist_list), GFP_KERNEL);
	next->next = NULL;
	memcpy(next->mac_addr, mac_addr, 6);
	if (prev)
		prev->next = next;
	else
		blacklist.list = next;
	blacklist.count++;
}

static void
del_blacklist(char *mac_addr)
{
	struct blacklist_list *list = blacklist.list;
	struct blacklist_list *prev, *next;

	prev = list;
	while (list) {
		if (!memcmp(list->mac_addr, mac_addr, 6)) {
			next = list->next;
			if (blacklist.list == list) {
				prev = list->next;
				blacklist.list = list->next;
			} else
				prev->next = next;
			kfree(list);
			list = next;
			blacklist.count--;
			return;
		}

		prev = list;
		list = list->next;
	}
}

static int 
set_blacklist(struct file *file,
	      const char *input,
	      unsigned long size,
	      void *data)
{
	int i = 0;
	unsigned long j;
	char buffer[1024];
	int len = 0, options = 0;
	char *c, *cp, *prev;
	char mac_addr[6];

	if (size > 1024) len = 1024; else len = size;
	if (copy_from_user(buffer, input, len))
		return -EFAULT;
	buffer[len-1] = '\0';
	cp = buffer;
	if (*cp == '+')
		options = 1;
	else if (*cp == '-')
		options = 0;
	else {
		maxtry = simple_strtoul(cp, NULL, 10);
		return len;
	}

	cp++;
	prev = cp;
	while (cp) {
		if (*cp == ':') {
			*cp = '\0';
			j = simple_strtoul(prev, &c, 16);
			mac_addr[i] = (unsigned char)(j & 0xff);
			prev = prev+3;
			i++;
                }
		if (i == 5)
			break;
		cp++;
	}
	j = simple_strtoul(prev, &c, 16);
	mac_addr[i] = (unsigned char)(j & 0xff);


	if (options)
		add_blacklist(mac_addr);
	else
		del_blacklist(mac_addr);
	return size;
}

static int 
get_blacklist(char *buffer,
	       char **start,
	       off_t offset,
	       int length)
{
	int len = 0, last_len = 0;
	off_t pos = 0;
	off_t begin = 0;
	struct blacklist_list *list = blacklist.list;
	struct blacklist_list *prev, *next;

	spin_lock_bh(&list_lock);
	len += sprintf(buffer+len, "%u\n", maxtry);
	prev = list;
	while (list) {
		last_len = len;
		len += sprintf(buffer+len, "%2x:%2x:%2x:%2x:%2x:%2x\n", 
				list->mac_addr[0],
				list->mac_addr[1],
				list->mac_addr[2],
				list->mac_addr[3],
				list->mac_addr[4],
				list->mac_addr[5]
				);

		pos = begin + len;
		if(pos < offset) { len = 0; begin = pos; }
		if(pos > offset + length) { len = last_len; break; }

		prev = list;
		list = list->next;
	}

	*start = buffer + (offset - begin);
	len -= (offset - begin);
	if(len > length) len = length;
	spin_unlock_bh(&list_lock);
        return len;
}

static int 
match(const struct sk_buff *skb,
      const struct net_device *in,
      const struct net_device *out,
      const void *matchinfo,
      int offset,
      const void *hdr,
      u_int16_t datalen,
      int *hotdrop)
{
	const struct ipt_blacklist_info *info = matchinfo;
	struct ethhdr *eth = (skb)->mac.ethernet;
	struct blacklist_list *list = blacklist.list;

	while (list) {
		if (!memcpy(eth->h_source, list->mac_addr, 6))
			return 1;

		list = list->next;
	}

	return 0;
}

static int
checkentry(const char *tablename,
           const struct ipt_ip *ip,
           void *matchinfo,
           unsigned int matchsize,
           unsigned int hook_mask)
{
        if (matchsize != IPT_ALIGN(sizeof(struct ipt_blacklist_info)))
                return 0;

        return 1;
}

static struct ipt_match blacklist_match =
{ { NULL, NULL }, "blacklist", &match, &checkentry, NULL, THIS_MODULE };

static int
__init init(void)
{
	struct proc_dir_entry *proc;

	proc = proc_net_create("ipt_blacklist", 0, get_blacklist);
	if (!proc)
		return -ENOMEM;
	proc->owner = THIS_MODULE;
	proc->write_proc = set_blacklist;

	blacklist.count = 0;
	blacklist.list = NULL;

        if (ipt_register_match(&blacklist_match))
                return -EINVAL;

        return 0;
}


static void
__exit fini(void)
{
	remove_proc_entry("ipt_blacklist", proc_net);
	ipt_unregister_match(&blacklist_match);
}

module_init(init);
module_exit(fini);
