#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/netfilter_ipv4/ip_conntrack.h>
#include <linux/proc_fs.h>
#include <linux/ctype.h>
#include <net/ip.h>
#include <net/tcp.h>

#include "regexp/regexp.c"

#include <linux/netfilter_ipv4/ipt_layer7.h>
#include <linux/netfilter_ipv4/ip_tables.h>

MODULE_AUTHOR("Matthew Strait <quadong@users.sf.net>, Ethan Sommer <sommere@users.sf.net>");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("iptables application layer match module");

#if defined(CONFIG_IP_NF_MATCH_LAYER7_DEBUG)
  #define DPRINTK(format,args...) printk(format,##args)
#else
  #define DPRINTK(format,args...)
#endif

/* 2004/5/28 Justin++ */
//#define MAXDATALEN 2048
#define MAXDATALEN 800
/* Justin-- */

static struct pattern_cache {
	char * regex_string;
	regexp * pattern;
	struct pattern_cache * next;
} *first_pattern_cache = NULL;

static spinlock_t l7_lock = SPIN_LOCK_UNLOCKED;

static regexp 
*compile_and_cache(char *regex_string, char *protocol) 
{
        struct pattern_cache *node               = first_pattern_cache;
        struct pattern_cache *last_pattern_cache = first_pattern_cache;
        unsigned int len;

        while (node != NULL) {
                if (!strcmp(node->regex_string, regex_string)) 
                        return node->pattern;

                last_pattern_cache = node; /* points at the last non-NULL node */
                node = node->next;
        }

        /* If we reach the end of the list, then we have not yet cached
           the pattern for this regex. Let's do that now. */
        node                = kmalloc(sizeof(struct pattern_cache), GFP_ATOMIC);
        node->regex_string  = kmalloc(strlen(regex_string) + 1,     GFP_ATOMIC);
        node->pattern       = kmalloc(sizeof(struct regexp),        GFP_ATOMIC);
        node->next          = NULL;

        if(first_pattern_cache == NULL) /* list is empty */
                first_pattern_cache = node; /* make node the beginning */
        else
                last_pattern_cache->next = node; /* attach node to the end */

        /* copy the string and compile the regex */
        len = strlen(regex_string);
        if ( (node->pattern = regcomp(regex_string, &len)) == NULL) {
                printk("<3>layer7: ERROR COMPILING REGEX \"%s\" (%s)\n", 
			regex_string, protocol);
                /* since the pattern is NULL, it is already cached -- as NULL*/
        }

        strcpy(node->regex_string, regex_string);
        return node->pattern;
}

#if CONFIG_IP_NF_MATCH_LAYER7_DEBUG
/* For printing.  Converts an unfriendly string into a friendly one by replacing 
unprintables with periods and all whitespace with " ". */
static char * friendly_print(unsigned char * s)
{
	char * f = kmalloc(strlen(s) + 1, GFP_ATOMIC);
	int i;
	for(i = 0; i < strlen(s); i++){
		if(isprint(s[i]) && s[i] < 128)	f[i] = s[i];
		else if(isspace(s[i]))		f[i] = ' ';
		else 				f[i] = '.';
	}
	f[i] = '\0';
	return f;
}
#endif

static int 
can_handle(const struct sk_buff *skb)
{
	if (!skb->nh.iph) /* not IP */
		return 0;
	if(skb->nh.iph->protocol != IPPROTO_TCP &&
	   skb->nh.iph->protocol != IPPROTO_UDP &&
	   skb->nh.iph->protocol != IPPROTO_ICMP)
		return 0;

	return 1;
}

/* Returns offset the into the skb->data that the application data starts */
static int app_data_offset(const struct sk_buff *skb)
{
	/* In case we are ported somewhere (ebtables?) where skb->nh.iph 
	isn't set, this can be gotten from 4*(skb->data[0] & 0x0f) as well. */
        int ip_hl = 4*skb->nh.iph->ihl;

        if( skb->nh.iph->protocol == IPPROTO_TCP ) {
                /* 12 == offset into TCP header for the header length field. 
		Can't get this with skb->h.th->doff because the tcphdr 
		struct doesn't get set when routing (this is confirmed to be 
		true in Netfilter as well as QoS.) */
                int tcp_hl = 4*(skb->data[ip_hl + 12] >> 4);

                return ip_hl + tcp_hl;
        } else if( skb->nh.iph->protocol == IPPROTO_UDP  ) {
                return ip_hl + 8; /* UDP header is always 8 bytes */
        } else if( skb->nh.iph->protocol == IPPROTO_ICMP ) {
                return ip_hl + 8; /* ICMP header is 8 bytes */
        } else {
                printk(KERN_ERR "layer7: tried to handle unknown protocol!\n");
                return ip_hl + 8; /* something reasonable */
        }
}

/* add the new app data to the conntrack.  Return number of bytes added. */
static int add_data(char *layer7data, 
			char * app_data, int appdatalen)
{
	int length = 0, i;

	/* Convert the packet into a valid string by stripping nulls. Also make
	everything lower case (our regex lib doesn't have a case insensitivity
	option). Add it to the end of the current data. */
	for(i = 0; i < MAXDATALEN-1 && i < appdatalen; i++) {
		if(app_data[i] != '\0') {
			layer7data[length] = tolower(app_data[i]);
			length++;
		}
	}

	layer7data[length] = '\0';
	return length;
}

/* Returns true on match and false otherwise.  */
static int match(struct sk_buff *skb, 
		 const struct net_device *in,
                 const struct net_device *out,
		 const void *matchinfo,
                 int offset,
		 int *hotdrop)
{
	struct ipt_layer7_info *info = (struct ipt_layer7_info *)matchinfo;
	enum ip_conntrack_info master_ctinfo, ctinfo;
	struct ip_conntrack *master_conntrack, *conntrack;
	unsigned char *app_data;  
	unsigned int pattern_result, appdatalen;
	regexp *comppattern;
	static spinlock_t list_lock = SPIN_LOCK_UNLOCKED;
/* 2004/10/26 Steve ++ */
	char layer7data[MAXDATALEN];
/* Steve -- */

	spin_lock(&list_lock);
	if (!can_handle(skb)) {
		DPRINTK("layer7: This is some protocol I can't handle\n");
		spin_unlock(&list_lock);
		return info->invert;
	}

	comppattern = compile_and_cache(info->pattern, info->protocol);

	app_data = skb->data + app_data_offset(skb);
	appdatalen = skb->tail - app_data;

        if(!(conntrack = ip_conntrack_get((struct sk_buff *)skb, &ctinfo)) ||
           !(master_conntrack = ip_conntrack_get((struct sk_buff *)skb, &master_ctinfo))) {
                printk(KERN_ERR "layer7: error getting conntrack, giving up.\n");
		spin_unlock(&list_lock);
		return info->invert;
        }
	
        /* Try to get a master conntrack (and its master etc) for FTP, etc. */
        while (master_ct(master_conntrack) != NULL)
                master_conntrack = master_ct(master_conntrack);

	layer7data[0] = '\0';
//	if(!skb->cb[0])	{
	if (1) {
		int newbytes = add_data(layer7data, app_data, appdatalen);

		if(newbytes == 0) { /* didn't add any data */
			skb->cb[0] = 1;
			/* Didn't match before, not going to match now */
			spin_unlock(&list_lock);
			return info->invert;
		}
	}

	/* If the regexp failed to compile, don't bother running it */
	if(comppattern && regexec(comppattern, layer7data)) {
		DPRINTK("layer7: regexec positive: %s!\n", info->protocol);
		pattern_result = 1;
	} else pattern_result = 0;

	/* mark the packet seen */
	skb->cb[0] = 1;

	spin_unlock(&list_lock);
    	return (pattern_result ^ info->invert);
}

static int checkentry(const char *tablename, const struct ipt_ip *ip,
           void *matchinfo, unsigned int matchsize, unsigned int hook_mask)
{
       if (matchsize != IPT_ALIGN(sizeof(struct ipt_layer7_info))) 
               return 0;
       return 1;
}

static struct ipt_match layer7_match
= { { NULL, NULL }, "layer7", &match, &checkentry, NULL, THIS_MODULE };

static int __init init(void)
{
	return ipt_register_match(&layer7_match);
}

static void __exit fini(void)
{
	ipt_unregister_match(&layer7_match);
}

module_init(init);
module_exit(fini);
