/* 
 * netfilter module to enforce network quotas
 *
 * Sam Johnston <samj@samj.net>
 */
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/spinlock.h>
#include <linux/interrupt.h>

#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ipt_quota.h>
/* 2003/4/17 Steve ++ */
#include <linux/netfilter_ipv4/ip_conntrack.h>
#include <linux/timer.h>
#include <linux/if_ether.h>

static struct timer_list quota_timer;
static struct tm currenttime;
/* Steve -- */

MODULE_LICENSE("GPL");

static spinlock_t quota_lock = SPIN_LOCK_UNLOCKED;

static int
match(const struct sk_buff *skb,
      const struct net_device *in,
      const struct net_device *out,
      const void *matchinfo,
      int offset, const void *hdr, u_int16_t datalen, int *hotdrop)
{
        struct ipt_quota_info *q = (struct ipt_quota_info *) matchinfo;
	struct ip_conntrack *ct;
	enum ip_conntrack_info ctinfo;

        spin_lock_bh(&quota_lock);
	ct = ip_conntrack_get((struct sk_buff *)skb, &ctinfo);

	if ((q->per_quota > 0) && ct) {
		ct->total_bytes += (skb->len + ETH_HLEN);
		if (ct->total_bytes > q->per_quota) {
			spin_unlock_bh(&quota_lock);
			return 1;
		}
	}

	if (q->quota > 0 && ct) {
		if (q->year != currenttime.tm_year ||
		    q->mon != currenttime.tm_mon ||
		    q->day != currenttime.tm_mday) {
			q->quota_usage = 0;
			q->year = currenttime.tm_year;
			q->mon = currenttime.tm_mon;
			q->day = currenttime.tm_mday;
		}
		q->quota_usage += (skb->len + ETH_HLEN);
		if (q->quota_usage > q->quota) {
			spin_unlock_bh(&quota_lock);
			return 1;
		}
	}

        spin_unlock_bh(&quota_lock);
        return 0;
}

static int
checkentry(const char *tablename,
           const struct ipt_ip *ip,
           void *matchinfo, unsigned int matchsize, unsigned int hook_mask)
{
        struct ipt_quota_info *q = (struct ipt_quota_info *) matchinfo;
        /* TODO: spinlocks? sanity checks? */
        if (matchsize != IPT_ALIGN(sizeof (struct ipt_quota_info)))
                return 0;

	q->year = currenttime.tm_year;
	q->mon = currenttime.tm_mon;
	q->day = currenttime.tm_mday;
        return 1;
}

static void
update_quota_time(unsigned long timeout)
{
	struct timeval currenttvsec;

	spin_lock_bh(&quota_lock);
	do_gettimeofday(&currenttvsec);
	localtime(&currenttvsec.tv_sec, &currenttime);
	quota_timer.expires = jiffies + (3 * HZ);
	add_timer(&quota_timer);
        spin_unlock_bh(&quota_lock);
}

static struct ipt_match quota_match
    = { {NULL, NULL}, "quota", &match, &checkentry, NULL, THIS_MODULE };

static int __init
init(void)
{
	init_timer(&quota_timer);
	quota_timer.function = update_quota_time;
	quota_timer.expires = jiffies + (3 * HZ);
	add_timer(&quota_timer);
        return ipt_register_match(&quota_match);
}

static void __exit
fini(void)
{
        ipt_unregister_match(&quota_match);
}

module_init(init);
module_exit(fini);

