/* Kernel module to check if the source address has been seen recently. */
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/proc_fs.h>
#include <linux/spinlock.h>
#include <linux/interrupt.h>
#include <asm/uaccess.h>
#include <linux/ctype.h>
#include <linux/ip.h>

#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ipt_recent.h>

/* Defaults, these can be overridden on the module command-line. */
/* 2002/3/18 Steve ++ */
#define IP_LIST_TOT		50
#define	IP_LIST_HASH_LOG	6
#define IP_LIST_ARRAY_SIZE	(1 << IP_LIST_HASH_LOG)
/* Steve -- */

/* 2002/3/15 Steve ++ */
static u_int32_t timeout = 60 * 10;
/* Steve -- */
/* 2005/5/4 Steve ++ */
static u_int32_t duration = 0;
/* Steve -- */

/* Structure of our list of recently seen addresses. */
struct recent_ip_list {
  u_int32_t saddr;
  u_int32_t last_seen;
/* 2005/5/4 Steve ++ */
  u_int32_t first_seen;
/* Steve -- */
  u_int32_t login;
/* 2003/12/1 Steve ++ */
  char user[IPT_RECENT_NAME_LEN];
/* Steve -- */
};

/* Structure of our linked list of tables of recent lists. */
struct recent_ip_tables {
  char name[IPT_RECENT_NAME_LEN];
  int count;
  struct recent_ip_list *table[IP_LIST_ARRAY_SIZE];
  struct recent_ip_tables *next;
  spinlock_t list_lock;
  struct proc_dir_entry *status_proc;
};

/* Our current list of addresses we have recently seen.
 * Only added to on a --set, and only updated on --set || --update 
 */
static struct recent_ip_tables *r_tables;

/* We protect r_list with this spinlock so two processors are not modifying
 * the list at the same time. 
 */
static spinlock_t recent_lock = SPIN_LOCK_UNLOCKED;

/* Our /proc/net/ipt_recent entry */
static struct proc_dir_entry *proc_net_ipt_recent;

/* Function declaration for later. */
static int
match(const struct sk_buff *skb,
      const struct net_device *in,
      const struct net_device *out,
      const void *matchinfo,
      int offset,
      const void *hdr,
      u_int16_t datalen,
      int *hotdrop);

/* This is the function which produces the output for our /proc output
 * interface which lists each IP address, the last seen time and the 
 * other recent times the address was seen.
 */

/* 2003/12/3 Steve ++ */
static inline unsigned int
hash_func(unsigned long addr)
{
        int hash;
	unsigned long value;

        value = addr;
        hash = 0;
        do {
                hash ^= value;
        } while ((value >>= IP_LIST_HASH_LOG));

        return hash & (IP_LIST_ARRAY_SIZE - 1);
#if 0
	unsigned int i = (addr % IP_LIST_ARRAY_SIZE);
	return i;
#endif
}
/* Steve -- */
 
static int ip_recent_get_info(char *buffer, char **start, off_t offset, int length, int *eof, void *data)
{
	int len = 0, count, last_len = 0;
	off_t pos = 0;
	off_t begin = 0;
	struct recent_ip_tables *curr_table;
/* 2003/12/3 Steve ++ */
	int i = 0;
/* Steve -- */
/* 2004/3/7 Steve ++ */
	unsigned int diff = 0;
/* Steve -- */

	curr_table = (struct recent_ip_tables*) data;

	spin_lock_bh(&curr_table->list_lock);
/* 2003/12/3 steve ++ */
	for (i = 0; i < IP_LIST_ARRAY_SIZE; i++) {
		for(count = 0; count < IP_LIST_TOT; count++) {
			if(!curr_table->table[i][count].saddr) continue;
/* 2005/5/4 Steve ++ */
#if 0
			if ((curr_table->table[i][count].last_seen+(timeout*HZ)) < jiffies) {
#endif
			if ((curr_table->table[i][count].last_seen+(timeout*HZ)) < jiffies ||
			    (duration && (curr_table->table[i][count].first_seen + (duration*HZ)) < jiffies)) {
				curr_table->table[i][count].saddr = 0;
				curr_table->table[i][count].last_seen = 0;
				curr_table->table[i][count].login = 0;
/* 2005/5/4 Steve ++ */
				curr_table->table[i][count].first_seen = 0;
/* Steve -- */
				continue;
			}
			last_len = len;
			len += sprintf(buffer+len,"%u.%u.%u.%u ",NIPQUAD(curr_table->table[i][count].saddr));
/* 2004/3/7 Steve ++ */
			if (curr_table->table[i][count].last_seen >= jiffies)
				diff = 0;
			else
				diff = jiffies - curr_table->table[i][count].last_seen;
			len += sprintf(buffer+len,"%u ", diff);
//			len += sprintf(buffer+len,"%u ",curr_table->table[i][count].last_seen);
/* Steve -- */
/* 2003/12/1 Steve ++ */
			len += sprintf(buffer+len,"%u ", curr_table->table[i][count].login);
			len += sprintf(buffer+len,"%s ", curr_table->table[i][count].user);
/* Steve -- */
/* 2005/5/4 Steve ++ */
			len += sprintf(buffer+len,"%u ", curr_table->table[i][count].first_seen);
			len += sprintf(buffer+len,"%u ", duration);
/* Steve -- */
			len += sprintf(buffer+len,"\n");
			pos = begin + len;
			if(pos < offset) { len = 0; begin = pos; }
			if(pos > offset + length) { len = last_len; break; }
		}
	}

	*start = buffer + (offset - begin);
	len -= (offset - begin);
	if(len > length) len = length;

	spin_unlock_bh(&curr_table->list_lock);
	return len;
}

/* ip_recent_ctrl provides an interface for users to modify the table
 * directly.  This allows adding entries, removing entries, and
 * flushing the entire table.
 * This is done by opening up the appropriate table for writing and
 * sending one of:
 * xx.xx.xx.xx   -- Add entry to table with current time
 * +xx.xx.xx.xx  -- Add entry to table with current time
 * -xx.xx.xx.xx  -- Remove entry from table
 * -0.0.0.0      -- Flush table, remove all entries
 */

static int ip_recent_ctrl(struct file *file, const char *input, unsigned long size, void *data)
{
	static const u_int32_t max[4] = { 0xffffffff, 0xffffff, 0xffff, 0xff };
	u_int32_t val;
	int base;
	char c, *cp;
	union iaddr {
		uint8_t bytes[4];
		uint32_t word;
	} res;
	uint8_t *pp = res.bytes;
	int digit;

	char buffer[1024], user[IPT_RECENT_NAME_LEN];
	int len, check_set = 0, count;
	u_int32_t saddr = 0;
	struct sk_buff skb;
	struct ipt_recent_info info;
	struct recent_ip_tables *curr_table;
/* 2003/12/1 Steve ++ */
	int i = 0;
	unsigned int hashnum;
/* Steve -- */
/* 2004/3/7 Steve ++ */
	char last_seen[IPT_RECENT_NAME_LEN], login[IPT_RECENT_NAME_LEN];
/* Steve -- */

	curr_table = (struct recent_ip_tables*) data;

	if(size > 1024) len = 1024; else len = size;

	if(copy_from_user(buffer,input,len)) return -EFAULT;

	buffer[len-1] = '\0';

	cp = buffer;
/* 2002/3/18 Steve ++ */
	if(buffer[0] >= '0' && buffer[0] <= '9') check_set = IPT_RECENT_SECONDS;
/* Steve -- */
	if(buffer[0] == '+') { check_set = IPT_RECENT_SET; cp++; }
	if(buffer[0] == '-') { check_set = IPT_RECENT_REMOVE; cp++; }
/* 2005/5/4 Steve ++ */
	if (buffer[0] == '*') { check_set = IPT_RECENT_SET_DURATION; cp++; }
/* Steve -- */
	if(!check_set) return len;

/* 2005/5/4 Steve ++ */
	if (check_set == IPT_RECENT_SET_DURATION) {
		duration = simple_strtoul(cp, NULL, 10) * 3600;
		return len;
	}
/* Steve -- */

/* 2002/3/18 Steve ++ */
	if (check_set == IPT_RECENT_SECONDS) {
		timeout = simple_strtoul(cp, NULL, 10);
		return len;
	}
/* Steve -- */
	/* Get saddr (effectively inet_aton()) */
	/* Shamelessly stolen from libc, a function in the kernel for doing
	 * this would, of course, be greatly preferred, but our options appear
	 * to be rather limited, so we will just do it ourselves here.
	 */
	res.word = 0;

	c = *cp;
	for(;;) {
		if(!isdigit(c)) return len;
		val = 0; base = 10; digit = 0;
		if(c == '0') {
			c = *++cp;
			if(c == 'x' || c == 'X') base = 16, c = *++cp;
			else { base = 8; digit = 1; }
		}
		for(;;) {
			if(isascii(c) && isdigit(c)) {
				if(base == 8 && (c == '8' || c == '0')) return len;
				val = (val * base) + (c - '0');
				c = *++cp;
				digit = 1;
			} else if(base == 16 && isascii(c) && isxdigit(c)) {
				val = (val << 4) | (c + 10 - (islower(c) ? 'a' : 'A'));
				c = *++cp;
				digit = 1;
			} else break;
		}
		if(c == '.') {
			if(pp > res.bytes + 2 || val > 0xff) return len;
			*pp++ = val;
			c = *++cp;
		} else break;
	}
	if(c != '\0' && (!isascii(c) || !isspace(c))) return len;
	if(!digit) return len;

	if(val > max[pp - res.bytes]) return len;
	saddr = res.word | htonl(val);
/* 2003/12/3 Steve ++ */
	hashnum = hash_func(saddr);
/* Steve -- */

/* 2004/3/7 Steve ++ */
	while (1) {
		if (!*cp || !isspace(*cp))
			break;

		cp++;
		continue;
	}
/* Steve -- */

/* 2003/12/1 Steve ++ */
	memset(user, '\0', sizeof(user));
	for (i = 0; i < (IPT_RECENT_NAME_LEN - 1);) {
		if (!*cp || isspace(*cp))
			break;
#if 0
		if (isspace(*cp)) {
			cp++;
			continue;
		}
#endif
		user[i] = *cp;
		i++; cp++;
	}
/* Steve -- */
/* 2004/3/7 Steve ++ */
	while (1) {
		if (!*cp || !isspace(*cp))
			break;

		cp++;
		continue;
	}
	memset(last_seen, '\0', sizeof(last_seen));
	for (i = 0; i < (IPT_RECENT_NAME_LEN - 1);) {
		if (!*cp || isspace(*cp))
			break;

		last_seen[i] = *cp;
		i++; cp++;
	}

	while (1) {
		if (!*cp || !isspace(*cp))
			break;

		cp++;
		continue;
	}
	memset(login, '\0', sizeof(login));
	for (i = 0; i < (IPT_RECENT_NAME_LEN - 1);) {
		if (!*cp || isspace(*cp))
			break;

		login[i] = *cp;
		i++; cp++;
	}
/* Steve -- */


/* 2003/12/1 Steve ++ */	
//	if(!saddr && check_set == IPT_RECENT_SET) return len;
	if((!saddr || !user[0]) && check_set == IPT_RECENT_SET) return len;
/* Steve -- */

	/* Check if we are asked to flush the entire table */
	if(!saddr && check_set == IPT_RECENT_REMOVE) {
		spin_lock_bh(&curr_table->list_lock);
		for(count = 0; count < IP_LIST_TOT; count++) {
			curr_table->table[hashnum][count].last_seen = 0;
			curr_table->table[hashnum][count].saddr = 0;
/* 2003/12/1 Steve ++ */
			curr_table->table[hashnum][count].login = 0;
			memset(curr_table->table[hashnum][count].user, '\0', IPT_RECENT_NAME_LEN);
/* Steve -- */
/* 2005/5/4 Steve ++ */
			curr_table->table[hashnum][count].first_seen = 0;
/* Steve -- */
		}
		spin_unlock_bh(&curr_table->list_lock);
		return len;
	}

	/* Set up and just call match */
	info.check_set = check_set;
	info.invert = 0;
/* 2002/3/18 Steve ++ */
	strncpy(info.name,curr_table->name, IPT_RECENT_NAME_LEN);
/* Steve -- */
/* 2003/12/1 Steve ++ */
	strcpy(info.user, user);
/* Steve -- */
/* 2004/3/7 Steve ++ */
	if (login[0])
		info.login = simple_strtoul(login, NULL, 10);
	else
		info.login = 0;
	if (last_seen[0])
		info.last_seen = simple_strtoul(last_seen, NULL, 10);
	else
		info.last_seen = 0;
/* Steve -- */
	skb.nh.iph = kmalloc(sizeof(struct iphdr),GFP_KERNEL);
	if(!skb.nh.iph) { return -ENOMEM; }

	skb.nh.iph->saddr = saddr;
	match(&skb,NULL,NULL,&info,0,NULL,sizeof(info),NULL);
	kfree(skb.nh.iph);

	return len;
}

/* 'match' is our primary function, called by the kernel whenever a rule is
 * hit with our module as an option to it.
 * What this function does depends on what was specifically asked of it by
 * the user:
 * --set -- Add or update last seen time of the source address of the packet
 *   -- matchinfo->check_set == IPT_RECENT_SET
 * --rcheck -- Just check if the source address is in the list
 *   -- matchinfo->check_set == IPT_RECENT_CHECK
 * --update -- If the source address is in the list, update last_seen
 *   -- matchinfo->check_set == IPT_RECENT_UPDATE
 * --remove -- If the source address is in the list, remove it
 *   -- matchinfo->check_set == IPT_RECENT_REMOVE
 * --seconds -- Option to --rcheck/--update, only match if last_seen within seconds
 *   -- matchinfo->seconds
 * --hitcount -- Option to --rcheck/--update, only match if seen hitcount times
 *   -- matchinfo->hit_count
 * --seconds and --hitcount can be combined
 */
static int
match(const struct sk_buff *skb,
      const struct net_device *in,
      const struct net_device *out,
      const void *matchinfo,
      int offset,
      const void *hdr,
      u_int16_t datalen,
      int *hotdrop)
{
	int count, oldest = 0, ans;
	unsigned long now = jiffies;
	unsigned long oldest_time;
	const struct ipt_recent_info *info = matchinfo;
	u_int32_t saddr = skb->nh.iph->saddr;
	struct recent_ip_tables *curr_table;
	struct recent_ip_tables *last_table;
	struct recent_ip_list *r_list;
/* 2003/12/1 Steve ++ */
	char *user = info->user;
	unsigned int hashnum;
	struct timeval currenttvsec;
/* Steve -- */
/* 2004/3/7 Steve ++ */
	unsigned int last_seen = info->last_seen;
	unsigned int login = info->login;
/* Steve -- */

	/* Default is false ^ info->invert */
	ans = info->invert;

	/* Find the right table */
	spin_lock_bh(&recent_lock);
	curr_table = r_tables;
/* 2002/3/18 Steve ++ */
	while( (last_table = curr_table) && strncmp(info->name,curr_table->name, IPT_RECENT_NAME_LEN) && (curr_table = curr_table->next) ); 
/* Steve -- */
	spin_unlock_bh(&recent_lock);

	/* Table with this name not found, match impossible */
	if(!curr_table) { return ans; }

	/* Make sure no one is changing the list while we work with it */
	spin_lock_bh(&curr_table->list_lock);

	hashnum = hash_func(saddr);
	r_list = curr_table->table[hashnum];

	for(count = 0, oldest_time = r_list[0].last_seen; count < IP_LIST_TOT; count++) {
		/* Keep track of empty spaces which can be filled up later */
		if(!r_list[count].saddr) { oldest = count; continue; }
		/* Keep track of the oldest entry if we have been asked to add this entry
		 * to the table if it does not exist. */
		if(info->check_set & IPT_RECENT_SET && r_list[count].last_seen < oldest_time) {
			oldest_time = r_list[count].last_seen;
			oldest = count;
		}

		if(r_list[count].saddr == saddr) {
			if((!user[0] || (strlen(user) == strlen(r_list[count].user) && 
			    !memcmp(user, r_list[count].user, strlen(user))))) { 
				if(info->check_set & IPT_RECENT_CHECK || info->check_set & IPT_RECENT_UPDATE) {
				/* 2002/3/15 Steve ++ */
					if((time_before (now,r_list[count].last_seen+timeout*HZ)) &&
					   (!duration || (r_list[count].first_seen+(HZ*duration)) > now))
						ans = !info->invert; 
					else 
						ans = info->invert;
				/* Steve -- */
				}

				/* If and only if we have been asked to SET, or to UPDATE (on match) do we add the
				 * current timestamp to the last_seen. */
				if((info->check_set & IPT_RECENT_SET && (ans = !info->invert)) || (info->check_set & IPT_RECENT_UPDATE && ans)) {
					r_list[count].last_seen = now;
				}

				/* If we have been asked to remove the entry from the list, just set it to 0 */
/* 2002/3/15 Steve ++ */
				if(info->check_set & IPT_RECENT_REMOVE ||
				  (info->check_set & IPT_RECENT_UPDATE && !ans)) {
/* Steve -- */
					r_list[count].last_seen = 0;
					r_list[count].saddr = 0;
					r_list[count].login = 0;
/* 2005/5/4 Steve ++ */
					r_list[count].first_seen = 0;
/* Steve -- */
					ans = !info->invert;
				}
			} else {
/* 2003/12/1 Steve ++ */
				if (info->check_set & IPT_RECENT_SET) {
					do_gettimeofday(&currenttvsec);
					strcpy(r_list[count].user, user);
					if (!last_seen)
						r_list[count].last_seen = now;
					else {
						if (now > last_seen)
							r_list[count].last_seen = now - last_seen;
						else
							r_list[count].last_seen = 0;
					}
					if (!login)
						r_list[count].login = currenttvsec.tv_sec;
					else
						r_list[count].login = login;
/* 2005/5/4 Steve ++ */
					r_list[count].first_seen = now;
/* Steve -- */
				}
/* Steve -- */
			}
			spin_unlock_bh(&curr_table->list_lock);
			return ans;
		}
/* 2002/3/18 Steve ++ */
		else {
/* 2005/5/4 Steve ++ */
#if 0
			if(!time_before(now,r_list[count].last_seen+timeout*HZ)) {
#endif
			if ((!time_before(now,r_list[count].last_seen+timeout*HZ)) ||
			   (duration && (r_list[count].first_seen+(duration*HZ)) < now)) {
/* Steve -- */
				r_list[count].last_seen = 0;
				r_list[count].saddr = 0;
				r_list[count].login = 0;
/* 2005/5/4 Steve ++ */
				r_list[count].first_seen = 0;
/* Steve -- */
			}
		}
/* Steve -- */
	}

	/* If the address is not currently in the list and we have been asked to SET
	 * then we need to add it to the list.  We do this by replacing the oldest
	 * entry in the list which is determined when we step through the list. */
	if(info->check_set & IPT_RECENT_SET) {
		do_gettimeofday(&currenttvsec);
		r_list[oldest].saddr = saddr;
		if (!last_seen)
			r_list[oldest].last_seen = now;
		else {
			if (now > last_seen)
				r_list[oldest].last_seen = now - last_seen;
			else
				r_list[oldest].last_seen = 0;
		}
/* 2003/12/1 Steve ++ */
		if (!login)
			r_list[oldest].login = currenttvsec.tv_sec;
		else
			r_list[oldest].login = login;
		strcpy(r_list[oldest].user, user);
/* Steve -- */

/* 2005/5/4 Steve ++ */		
		r_list[oldest].first_seen = now;
/* Steve -- */
		ans = !info->invert;
	}

	spin_unlock_bh(&curr_table->list_lock);
	return ans;
}

/* This function is to verify that the rule given during the userspace iptables
 * command is correct.
 * If the command is valid then we check if the table name referred to by the
 * rule exists, if not it is created.
 */
static int
checkentry(const char *tablename,
           const struct ipt_ip *ip,
           void *matchinfo,
           unsigned int matchsize,
           unsigned int hook_mask)
{
	int flag = 0;
	const struct ipt_recent_info *info = matchinfo;
	struct recent_ip_tables *curr_table, *last_table;
/* 2003/12/3 Steve ++ */
	int i = 0;
/* Steve -- */

	if (matchsize != IPT_ALIGN(sizeof(struct ipt_recent_info))) return 0;

	/* seconds and hit_count only valid for CHECK/UPDATE */
	/* 2002/3/15 Steve ++ */
	if (info->check_set & IPT_RECENT_SET) {
		flag++;
	}
	if (info->check_set & IPT_RECENT_REMOVE) {
		flag++;
	}
	if(info->check_set & IPT_RECENT_CHECK) {
		flag++;
	}
	if(info->check_set & IPT_RECENT_UPDATE) {
		flag++;
	}
	/* Steve -- */
/* 2005/5/4 Steve ++ */
	if (info->check_set & IPT_RECENT_SET_DURATION) {
		flag++;
	}
/* Steve -- */

	/* One and only one of these should ever be set */
	if(flag != 1) return 0;

	/* Name must be set to something */
	if(!info->name || !info->name[0]) return 0;

	/* Things look good, create a list for this if it does not exist */
	/* Lock the linked list while we play with it */
//	spin_lock_bh(&recent_lock);

	/* Look for an entry with this name already created */
	/* Finds the end of the list and the entry before the end if current name does not exist */
	curr_table = r_tables;
/* 2002/3/18 Steve ++ */
	while( (last_table = curr_table) && strncmp(info->name,curr_table->name,IPT_RECENT_NAME_LEN) && (curr_table = curr_table->next) ); 
/* Steve -- */
	/* If a table already exists just increment the count on that table and return */
	if(curr_table) { curr_table->count++; return 1; }

	/* Table with this name not found */
	/* Allocate memory for new linked list item */
	curr_table = kmalloc(sizeof(struct recent_ip_tables),GFP_KERNEL);
	if(curr_table == NULL) { return -ENOMEM; }

	curr_table->list_lock = SPIN_LOCK_UNLOCKED;
	curr_table->next = NULL;
	curr_table->count = 1;
/* 2002/3/18 Steve ++ */
	strncpy(curr_table->name,info->name,IPT_RECENT_NAME_LEN);
/* Steve -- */

	/* Allocate memory for this table and the list of packets in each entry. */
/* 2003/12/3 Steve ++ */
/* 2002/3/18 Steve ++ */
        curr_table->table[0] = (struct recent_ip_list *)__get_free_pages(GFP_KERNEL, 5);
/* Steve -- */
        if(curr_table->table[0] == NULL) { return -ENOMEM; }
	for (i = 0; i < IP_LIST_ARRAY_SIZE; i++) {
		curr_table->table[i] = (struct recent_ip_list *)((unsigned long)curr_table->table[0] + (unsigned long)(sizeof(struct recent_ip_list) * IP_LIST_TOT * i));
		memset(curr_table->table[i], 0, sizeof(struct recent_ip_list)*IP_LIST_TOT);
	}
/* Steve -- */

	/* Create our proc 'status' entry. */
	curr_table->status_proc = create_proc_entry(curr_table->name, 0644, proc_net_ipt_recent);
	if (curr_table->status_proc) curr_table->status_proc->owner = THIS_MODULE;
	else {
        /* 2002/3/18 Steve ++ */
                free_pages((unsigned long)curr_table->table[0], 5);
        /* Steve -- */
		kfree(curr_table);
		return -ENOMEM;
	}
	curr_table->status_proc->read_proc = ip_recent_get_info;
	curr_table->status_proc->write_proc = ip_recent_ctrl;
	curr_table->status_proc->data = curr_table;

	/* Put the new table in place */
	if(!last_table) r_tables = curr_table; else last_table->next = curr_table;

//	spin_unlock_bh(&recent_lock);

	return 1;
}

/* This function is called in the event that a rule matching this module is
 * removed.
 * When this happens we need to check if there are no other rules matching
 * the table given.  If that is the case then we remove the table and clean
 * up its memory.
 */
static void
destroy(void *matchinfo, unsigned int matchsize)
{
	const struct ipt_recent_info *info = matchinfo;
	struct recent_ip_tables *curr_table, *last_table;
	if(matchsize != IPT_ALIGN(sizeof(struct ipt_recent_info))) return;

	/* Lock the linked list while we play with it */
//	spin_lock_bh(&recent_lock);

	/* Look for an entry with this name already created */
	/* Finds the end of the list and the entry before the end if current name does not exist */
	last_table = NULL;
	curr_table = r_tables;
	if(!curr_table) { return; }
/* 2002/3/18 Steve ++ */
	while( strncmp(info->name,curr_table->name,IPT_RECENT_NAME_LEN) && (last_table = curr_table) && (curr_table = curr_table->next) );
/* Steve -- */
	/* If a table does not exist then do nothing and return */
	if(!curr_table) { return; }

	curr_table->count--;
	/* If count is still non-zero then there are still rules referenceing it so we do nothing */
	if(curr_table->count) { return; }

	/* Count must be zero so we remove this table from the list */
	if(last_table) last_table->next = curr_table->next; else r_tables = curr_table->next;

	/* lock to make sure any late-runners still using this after we removed it from
	 * the list finish up then remove everything */
//	spin_lock_bh(&curr_table->list_lock);
	remove_proc_entry(curr_table->name,proc_net_ipt_recent);
        /* 2002/3/18 Steve ++ */
        free_pages((unsigned long)curr_table->table[0], 5);
        /* Steve ++ */
//	spin_unlock_bh(&curr_table->list_lock);
	kfree(curr_table);

//	spin_unlock_bh(&recent_lock);

	return;
}

/* This is the structure we pass to ipt_register to register our
 * module with iptables.
 */
static struct ipt_match recent_match
= { { NULL, NULL }, "recent", &match, &checkentry, &destroy, THIS_MODULE };

/* Kernel module initialization. */
static int __init init(void)
{
	proc_net_ipt_recent = proc_mkdir("ipt_recent",proc_net);
	if(!proc_net_ipt_recent) return -ENOMEM;

	return ipt_register_match(&recent_match);
}

/* Kernel module destruction. */
static void __exit fini(void)
{
	ipt_unregister_match(&recent_match);

	remove_proc_entry("ipt_recent",proc_net);
}

/* Register our module with the kernel. */
module_init(init);
module_exit(fini);
