/* Kernel module to control the rate
 *
 * Jrme de Vivie   <devivie@info.enserb.u-bordeaux.fr>
 * Herv Eychenne   <eychenne@info.enserb.u-bordeaux.fr>
 *
 * 2 September 1999: Changed from the target RATE to the match
 *                   `limit', removed logging.  Did I mention that
 *                   Alexey is a fucking genius?
 *                   Rusty Russell (rusty@rustcorp.com.au).  */
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/spinlock.h>
#include <linux/interrupt.h>

#include <linux/ip.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ipt_siplimit.h>

/* The algorithm used is the Simple Token Bucket Filter (TBF)
 * see net/sched/sch_tbf.c in the linux source tree
 */

static spinlock_t sip_limit_lock = SPIN_LOCK_UNLOCKED;

/* Rusty: This is my (non-mathematically-inclined) understanding of
   this algorithm.  The `average rate' in jiffies becomes your initial
   amount of credit `credit' and the most credit you can ever have
   `credit_cap'.  The `peak rate' becomes the cost of passing the
   test, `cost'.

   `prev' tracks the last packet hit: you gain one credit per jiffy.
   If you get credit balance more than this, the extra credit is
   discarded.  Every time the match passes, you lose `cost' credits;
   if you don't have that many, the test fails.

   See Alexey's formal explanation in net/sched/sch_tbf.c.

   To avoid underflow, we multiply by 128 (ie. you get 128 credits per
   jiffy).  Hence a cost of 2^32-1, means one pass per 32768 seconds
   at 1024HZ (or one every 9 hours).  A cost of 1 means 12800 passes
   per second at 100HZ.  */

#define SIP_CREDITS_PER_JIFFY 128


/* 2003/4/30 Steve ++ */
static inline int hashfunc(struct in_addr addr)
{
        unsigned int value;
        int hash;

        value = addr.s_addr;
        hash = 0;
        do {
                hash ^= value;
        } while ((value >>= SIP_HASH_LOG));

        return hash & (SIP_HASH_SIZE - 1);
}
/* Steve -- */

static int
ipt_siplimit_match(const struct sk_buff *skb,
		const struct net_device *in,
		const struct net_device *out,
		const void *matchinfo,
		int offset,
		const void *hdr,
		u_int16_t datalen,
		int *hotdrop)
{
	struct ipt_siprateinfo *r;
	unsigned long now = jiffies;
/* 2003/4/30 Steve ++ */
	struct in_addr addr;
	struct sip_host *curr, *last, **head;
	int hash, count;
	struct iphdr *ip_hdr;
	struct dropsip *drop_curr, *drop_last, **drop_head;
	int drop_hash, drop_count;
/* Steve -- */

	spin_lock_bh(&sip_limit_lock);

	r = ((struct ipt_siprateinfo *)matchinfo)->master;
	ip_hdr = skb->nh.iph;
	addr.s_addr = ip_hdr->saddr;
	if (!addr.s_addr) {
		spin_unlock_bh(&sip_limit_lock);
		return 1;
	}

	drop_count = 0;
	drop_last = NULL;
	if ((drop_curr = *(drop_head = &(r->drop->hash[drop_hash = hashfunc(addr)]))))
		do {
			if (drop_curr->src_addr.s_addr == addr.s_addr &&
			    drop_curr->next_time > now) {
				drop_curr->next_time = now + (r->timeout * HZ);
				spin_unlock_bh(&sip_limit_lock);
				return 0;
			}

			drop_count++;
			if (drop_curr->next) drop_last = drop_curr;
		} while ((drop_curr = drop_curr->next));

	count = 0;
	last = NULL;
	if ((curr = *(head = &(r->state->hash[hash = hashfunc(addr)]))))
		do {
			if (curr->src_addr.s_addr == addr.s_addr) break;

			count++;
			if (curr->next) last = curr;
		} while ((curr = curr->next));

	if (curr) {
		curr->credit += (now - xchg(&curr->prev, now)) * SIP_CREDITS_PER_JIFFY;
		if (curr->credit > curr->credit_cap)
			curr->credit = curr->credit_cap;

		if (curr->credit >= curr->cost) {
			curr->credit -= curr->cost;
			spin_unlock_bh(&sip_limit_lock);
			return 1;
		}

		goto nomatch;
	}

	if (count >= SIP_HASH_MAX && last) last->next = NULL;

	if (r->state->list[r->state->index].src_addr.s_addr)
		head = &r->state->hash[hashfunc(r->state->list[r->state->index].src_addr)];
	else
		head = &last;

	last = NULL;
	if ((curr = *head))
	do {
		if (curr == &r->state->list[r->state->index]) break;
		last = curr;
	} while ((curr = curr->next));

	if (curr) {
		if (last)
			last->next = last->next->next;
		else if (*head)
			*head = (*head)->next;
	}

	/* Get our list entry */
	curr = &r->state->list[r->state->index++];
	if (r->state->index >= SIP_LIST_SIZE) r->state->index = 0;

	/* Link it into the hash table */
	head = &r->state->hash[hash];
	curr->next = *head;
	*head = curr;
	curr->src_addr = addr;
	curr->prev = now;
	curr->credit_cap = r->credit_cap;
	curr->credit = r->credit_cap;
	curr->cost = r->cost;

	spin_unlock_bh(&sip_limit_lock);
	return 1;

nomatch:
	if (drop_count >= SIP_HASH_MAX && drop_last) drop_last->next = NULL;

	if (r->drop->list[r->drop->index].src_addr.s_addr)
		drop_head = &r->drop->hash[hashfunc(r->drop->list[r->drop->index].src_addr)];
	else
		drop_head = &drop_last;

	drop_last = NULL;
	if ((drop_curr = *drop_head))
	do {
		if (drop_curr == &r->drop->list[r->drop->index]) break;
		drop_last = drop_curr;
	} while ((drop_curr = drop_curr->next));

	if (drop_curr) {
		if (drop_last)
			drop_last->next = drop_last->next->next;
		else if (*drop_head)
			*drop_head = (*drop_head)->next;
	}

	/* Get our list entry */
	drop_curr = &r->drop->list[r->drop->index++];
	if (r->drop->index >= SIP_LIST_SIZE) r->drop->index = 0;

	/* Link it into the hash table */
	drop_head = &r->drop->hash[drop_hash];
	drop_curr->next = *drop_head;
	*drop_head = drop_curr;
	drop_curr->src_addr = addr;
	if (r->timeout > 0)
		drop_curr->next_time = now + (r->timeout * HZ);
	else
		drop_curr->next_time = 0;

	spin_unlock_bh(&sip_limit_lock);
	return 0;
}

/* Precision saver. */
static u_int32_t
user2credits(u_int32_t user)
{
	/* If multiplying would overflow... */
	if (user > 0xFFFFFFFF / (HZ*SIP_CREDITS_PER_JIFFY))
		/* Divide first. */
		return (user / SIP_IPT_LIMIT_SCALE) * HZ * SIP_CREDITS_PER_JIFFY;

	return (user * HZ * SIP_CREDITS_PER_JIFFY) / SIP_IPT_LIMIT_SCALE;
}

static int
ipt_siplimit_checkentry(const char *tablename,
		     const struct ipt_ip *ip,
		     void *matchinfo,
		     unsigned int matchsize,
		     unsigned int hook_mask)
{
	struct ipt_siprateinfo *r = matchinfo;

	if (matchsize != IPT_ALIGN(sizeof(struct ipt_siprateinfo)))
		return 0;

	/* Check for overflow. */
	if (r->burst == 0
	    || user2credits(r->avg * r->burst) < user2credits(r->avg)) {
		printk("Call rusty: overflow in ipt_siplimit: %u/%u\n",
		       r->avg, r->burst);
		return 0;
	}

	/* User avg in seconds * IPT_LIMIT_SCALE: convert to jiffies *
	   128. */
	r->prev = jiffies;
	r->credit = user2credits(r->avg * r->burst);	 /* Credits full. */
	r->credit_cap = user2credits(r->avg * r->burst); /* Credits full. */
	r->cost = user2credits(r->avg);

	r->state = (struct limitstate *)kmalloc(sizeof(struct limitstate), GFP_KERNEL);
	memset(r->state, '\0', sizeof(struct limitstate));
	r->drop = (struct droplist *)kmalloc(sizeof(struct droplist), GFP_KERNEL);
	memset(r->drop, '\0', sizeof(struct droplist));
	/* For SMP, we only want to use one set of counters. */
	r->master = r;

	return 1;
}

static void
ipt_siplimit_destroy(void *matchinfo, unsigned int matchsize)
{
        struct ipt_siprateinfo *r = matchinfo;

        if (matchsize != IPT_ALIGN(sizeof(struct ipt_siprateinfo)))
                return;

	spin_lock_bh(&sip_limit_lock);
	if (r->state)
		kfree(r->state);
	if (r->drop)
		kfree(r->drop);
	spin_unlock_bh(&sip_limit_lock);
	return;
}

static struct ipt_match ipt_siplimit_reg
= { { NULL, NULL }, "siplimit", ipt_siplimit_match, ipt_siplimit_checkentry, ipt_siplimit_destroy, THIS_MODULE };

static int __init init(void)
{
	if (ipt_register_match(&ipt_siplimit_reg))
		return -EINVAL;
	return 0;
}

static void __exit fini(void)
{
	ipt_unregister_match(&ipt_siplimit_reg);
}

module_init(init);
module_exit(fini);
