#include <linux/module.h>
#include <linux/netfilter.h>
#include <linux/ip.h>
#include <net/checksum.h>
#include <net/udp.h>
#include <net/tcp.h>
#include <linux/skbuff.h>
#include <linux/spinlock.h>

#include <linux/netfilter_ipv4/ipt_skype.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ip_conntrack.h>

#define	SKYPE_DNS_PORT	53
#define	SKYPE_HTTP_PORT 80
#define SKYPE_HTTPS_PORT 443

#define	SKYPE_DOMAIN_SIZE 16
static unsigned char skypeDomain[] =
{
	0x05, 0x73, 0x6e, 0x6f, 0x64, 0x65,
	0x05, 0x73, 0x6b, 0x79, 0x70, 0x65,
	0x03, 0x6e, 0x65, 0x74
};

#define SKYPE_HTTP_SIZE 5
#define	SKYPE_HTTP_PATTERN "HTTP/"

#define SKYPE_HTTPS_SIZE 26
static unsigned char skypehttps[] =
{
	0x80, 0x46, 0x01, 0x03, 0x01, 0x00, 0x2d, 0x00, 
	0x00, 0x00, 0x10, 0x00, 0x00, 0x05, 0x00, 0x00,
	0x04, 0x00, 0x00, 0x0a, 0x00, 0x00, 0x09, 0x00,
	0x00, 0x64
};

static struct skypedroplist *skypedrop;
static struct skyperetry *skyperetry;

static spinlock_t skype_lock = SPIN_LOCK_UNLOCKED;

static inline int 
hashfunc(struct in_addr addr)
{
	unsigned int value;
	int hash;

	value = addr.s_addr;
	hash = 0;
	do {
		hash ^= value;
	} while ((value >>= SKYPE_SIP_HASH_LOG));

	return hash & (SKYPE_SIP_HASH_SIZE - 1);
}

static int
find_skype_domain(const void *data, unsigned int datalen)
{
	char *c = (char *)data;
	int i = 0;

	if (datalen >= SKYPE_DOMAIN_SIZE) {
		for (i = 0; i < (datalen - SKYPE_DOMAIN_SIZE); i++) {
			if (!memcmp(c, skypeDomain, SKYPE_DOMAIN_SIZE))
				return 1;

			c++;
		}
	}

	return 0;
}

static int
find_skype_pattern(const struct sk_buff *skb,
		   struct skypeinfo *skypeinfo)
{
	struct iphdr *iph = skb->nh.iph;
	struct udphdr *udph = (void *)iph + iph->ihl * 4;
	unsigned int len = skb->len - iph->ihl * 4 - sizeof(struct udphdr);
	const void *data = (void *)((const char *)udph + sizeof(struct udphdr));
	u32	pkt_daddr;
	struct in_addr addr;
	struct skyperetrylist *curr, *last, **head;
        int hash = 0, count = 0;

	if (len != 18 && len != 23)
		return 0;

	if (len == 23) {
		memcpy(&pkt_daddr, (data+8), 4);
		if (pkt_daddr == iph->daddr)
			return 1;
	} else if (len == 18) {
		addr.s_addr = iph->saddr;

	        if ((curr = *(head = &(skyperetry->hash[hash = hashfunc(addr)]))))
			do {
				if (curr->skypeinfo.saddr.s_addr == skypeinfo->saddr.s_addr) {
					if (curr->skypeinfo.daddr.s_addr != skypeinfo->daddr.s_addr &&
					    curr->skypeinfo.port == skypeinfo->port &&
					   (curr->skypeinfo.match + (600 * HZ)) > jiffies) {
						curr->skypeinfo.match = jiffies;
						return 1;
					} 
				}

				count++;
				if (curr->next) last = curr;
			} while ((curr = curr->next));
	}

	if (count >= SKYPE_SIP_HASH_MAX && last) last->next = NULL;

	if (skyperetry->list[skyperetry->index].skypeinfo.saddr.s_addr)
		head = &skyperetry->hash[hashfunc(skyperetry->list[skyperetry->index].skypeinfo.saddr)];
	else
		head = &last;

	last = NULL;

	if ((curr = *head))
		do {
			if (curr == &skyperetry->list[skyperetry->index]) break;
				last = curr;
		} while ((curr = curr->next));

	if (curr) {
		if (last)
			last->next = last->next->next;
		else if (*head)
			*head = (*head)->next;
	}

	curr = &skyperetry->list[skyperetry->index++];

	if (skyperetry->index >= SKYPE_SIP_LIST_SIZE)
		skyperetry->index = 0;

	head = &skyperetry->hash[hash];
	curr->next = *head;
	*head = curr;

	curr->skypeinfo.saddr.s_addr = skypeinfo->saddr.s_addr;
	curr->skypeinfo.daddr.s_addr = skypeinfo->daddr.s_addr;
	curr->skypeinfo.port = skypeinfo->port;
	curr->skypeinfo.match = jiffies;

	return 0;
}

static int
get_tcp_state(struct ip_conntrack *ct,
	      struct tcphdr *tcph,
	      unsigned int len,
	      enum ip_conntrack_info ctinfo)
{
	if (tcph->syn) {
		if (ctinfo >= IP_CT_IS_REPLY &&
		    ct->skype.state == 0) {
			ct->skype.seq = tcph->ack_seq;
			ct->skype.ack = tcph->seq;
			ct->skype.state = 1;
		} else if (ctinfo == IP_CT_NEW) {
			ct->skype.seq = tcph->seq;
			ct->skype.ack = tcph->ack_seq;
			ct->skype.state = 0;
		}

		return 0;
	}

	if (ct->skype.seq == tcph->seq &&
	    htonl(ntohl(ct->skype.ack) + 1) == tcph->ack_seq &&
	    ctinfo < IP_CT_IS_REPLY &&
	    ct->skype.state == 1 &&
	    len == 0) {
		ct->skype.seq = tcph->seq;
		ct->skype.ack = tcph->ack_seq;
		ct->skype.state = 2;
	}

	return 0;
}

static int
check_skype_https(struct ip_conntrack *ct,
		  struct tcphdr *tcph,
		  unsigned int datalen,
		  enum ip_conntrack_info ctinfo,
		  const void *data)
{
	char *c = (char *)data;

	if (ctinfo < IP_CT_IS_REPLY &&
	    ct->skype.seq == tcph->seq &&
	    ct->skype.ack == tcph->ack_seq &&
	    datalen == 72) {
		if (!memcmp(c, skypehttps, SKYPE_HTTPS_SIZE)) {
			ct->skype.state = 4;
			return 1;
		}
	}

	return 0;
}

static int
check_skype_http(struct ip_conntrack *ct,
		 struct tcphdr *tcph,
		 unsigned int datalen,
		 enum ip_conntrack_info ctinfo,
		 const void *data)
{
	int i = 0, httpmatch = 0;
	char *c = (char *)data;

	if (ctinfo < IP_CT_IS_REPLY &&
	    ct->skype.seq == tcph->seq &&
	    ct->skype.ack == tcph->ack_seq &&
	    datalen == 16) {
		httpmatch = 1;
		for (i = 0; i < (datalen - SKYPE_HTTP_SIZE); i++) {
			if (!memcmp(c, SKYPE_HTTP_PATTERN, SKYPE_HTTP_SIZE))
				return 0;

			c++;
		}
	}

/* 2004/9/21 Steve ++
 * Detect http proxy
 */
        if (ctinfo < IP_CT_IS_REPLY &&
            datalen == 72) {
                if (!memcmp(c, skypehttps, SKYPE_HTTPS_SIZE)) {
                        httpmatch = 1;
                }
        }
/* Steve -- */

	if (httpmatch) {
		ct->skype.state = 4;
		return 1;
	}

	return 0;
}

static int
check_skype_nonstandard(struct ip_conntrack *ct,
			struct tcphdr *tcph,
			unsigned int len,
			enum ip_conntrack_info ctinfo)
{
	if (ctinfo < IP_CT_IS_REPLY &&
	    ct->skype.seq == tcph->seq &&
	    ct->skype.ack == tcph->ack_seq &&
	    ct->skype.state == 2 &&
	    len == 14) {
		ct->skype.seq = htonl(ntohl(tcph->seq) + len);
		ct->skype.ack = tcph->ack_seq;
		ct->skype.state = 3;
		return 0;
	}

	if (ctinfo >= IP_CT_IS_REPLY &&
	    ct->skype.state == 3 &&
	    ct->skype.seq == tcph->ack_seq &&
	    ct->skype.ack == tcph->seq &&
	    len == 14) {
		ct->skype.state = 4;
		return 1;
	}

	return 0;
}

static int
check_tcp_skype(struct ip_conntrack *ct,
		struct tcphdr *tcph,
		const void *data,
		unsigned int len,
		enum ip_conntrack_info ctinfo)
{
	int state = 0;

	if (ct->skype.state == 4)
		return 1;

	get_tcp_state(ct, tcph, len, ctinfo);
	if (ct->skype.state < 2)
		return 0;

	state = check_skype_nonstandard(ct, tcph, len, ctinfo);
	if (state)
		return 1;

	if (tcph->dest == htons(SKYPE_HTTP_PORT)) {
		state = check_skype_http(ct, tcph, len, ctinfo, data);
		if (state)
			return 1;
	}

	if (tcph->dest == htons(SKYPE_HTTPS_PORT)) {
		state = check_skype_https(ct, tcph, len, ctinfo, data);
		if (state)
			return 1;
	}

	return 0;
}

static int
match(const struct sk_buff *skb,
      const struct net_device *in,
      const struct net_device *out,
      const void *matchinfo,
      int offset,
      const void *hdr,
      u_int16_t datalen,
      int *hotdrop)
{
	struct iphdr *iph = skb->nh.iph;
	struct udphdr *udph;
	struct tcphdr *tcph;
	const void *data;
	unsigned int len;
	enum ip_conntrack_info ctinfo;
	struct ip_conntrack *ct = ip_conntrack_get((struct sk_buff *)skb, &ctinfo);
	struct in_addr addr;
	struct skypedropsip *skypedrop_curr, *skypedrop_last, **skypedrop_head;
        int skypedrop_hash, skypedrop_count;
	u_int16_t sport;
	struct skypeinfo skypeinfo;
	char *point;

	if (!ct)
		return 0;

//	addr.s_addr = iph->saddr;
	addr.s_addr = ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.ip;
	if (!addr.s_addr)
		return 0;

	spin_lock_bh(&skype_lock);
	if (iph->protocol == IPPROTO_UDP) {
		udph = (void *)iph + iph->ihl * 4;
		data = (void *)((const char *)udph + sizeof(struct udphdr));
		len = skb->len - iph->ihl * 4 - sizeof(struct udphdr);
		if (len < 4) {
			spin_unlock_bh(&skype_lock);
			return 0;
		}
	} else if (iph->protocol == IPPROTO_TCP) {
		tcph = (void *)iph + iph->ihl * 4;
		data = (unsigned char *) tcph + tcph->doff * 4;
		len = skb->len - iph->ihl * 4 - tcph->doff * 4;
	} else {
		spin_unlock_bh(&skype_lock);
		return 0;
	}

	point = (char *)data;
	/* check droplist */
	skypedrop_count = 0;
        skypedrop_last = NULL;
	if ((skypedrop_curr = *(skypedrop_head = &(skypedrop->hash[skypedrop_hash = hashfunc(addr)]))))
		do {
			if (skypedrop_curr->src_addr.s_addr == addr.s_addr) {
				if (iph->protocol == IPPROTO_UDP &&
				    *(point+2) == 0x02) {
					sport = udph->source;
					if (sport == skypedrop_curr->port &&
					   (skypedrop_curr->match + (600 * HZ)) > jiffies) {
						skypedrop_curr->match = jiffies;
						spin_unlock_bh(&skype_lock);
						return 1;
					}
				} else if (iph->protocol == IPPROTO_TCP) {
					if ((skypedrop_curr->match + (600 * HZ)) > jiffies &&
					     check_tcp_skype(ct, tcph, data, len, ctinfo)) {
						skypedrop_curr->match = jiffies;
						spin_unlock_bh(&skype_lock);
						return 1;
					}
				}
			}

			skypedrop_count++;
			if (skypedrop_curr->next) skypedrop_last = skypedrop_curr;
		} while ((skypedrop_curr = skypedrop_curr->next));
	/* check droplist -- */

	if ((iph->protocol == IPPROTO_UDP) && *(point+2) == 0x02) {
		skypeinfo.saddr.s_addr = iph->saddr;
		skypeinfo.daddr.s_addr = iph->daddr;
		skypeinfo.port = udph->source;
		if (find_skype_pattern(skb, &skypeinfo))
			goto skypedrop;
	} else if ((iph->protocol == IPPROTO_UDP) &&
		   (udph->dest == htons(SKYPE_DNS_PORT))) {

		if (find_skype_domain(data, len))
			goto skypedrop;
	} else if (iph->protocol == IPPROTO_TCP) {
		check_tcp_skype(ct, tcph, data, len, ctinfo);
		if (ct->skype.state == 4)
			goto skypedrop;
	}
	
	spin_unlock_bh(&skype_lock);
	return	0;

skypedrop:
	if (skypedrop_count >= SKYPE_SIP_HASH_MAX && skypedrop_last) 
		skypedrop_last->next = NULL;

	if (skypedrop->list[skypedrop->index].src_addr.s_addr)
		skypedrop_head = &skypedrop->hash[hashfunc(skypedrop->list[skypedrop->index].src_addr)];
	else
		skypedrop_head = &skypedrop_last;

	skypedrop_last = NULL;
	if ((skypedrop_curr = *skypedrop_head))
	do {
		if (skypedrop_curr == &skypedrop->list[skypedrop->index]) break;
		skypedrop_last = skypedrop_curr;
	} while ((skypedrop_curr = skypedrop_curr->next));

	if (skypedrop_curr) {
		if (skypedrop_last)
			skypedrop_last->next = skypedrop_last->next->next;
		else if (*skypedrop_head)
			*skypedrop_head = (*skypedrop_head)->next;
	}

	skypedrop_curr = &skypedrop->list[skypedrop->index++];
	if (skypedrop->index >= SKYPE_SIP_LIST_SIZE) skypedrop->index = 0;
	skypedrop_head = &skypedrop->hash[skypedrop_hash];
	skypedrop_curr->next = *skypedrop_head;
	*skypedrop_head = skypedrop_curr;
	skypedrop_curr->src_addr = addr;
	if (iph->protocol == IPPROTO_UDP)
		skypedrop_curr->port = udph->source;
	skypedrop_curr->match = jiffies;

	spin_unlock_bh(&skype_lock);
	return 1;
}

static int
checkentry(const char *tablename,
           const struct ipt_ip *ip,
           void *matchinfo,
           unsigned int matchsize,
           unsigned int hook_mask)
{
	return 1;
}

static struct ipt_match skype_match
= { { NULL, NULL }, "skype", &match, &checkentry, NULL, THIS_MODULE };

static int __init init(void)
{
	skypedrop = (struct skypedroplist *)kmalloc(sizeof(struct skypedroplist), GFP_KERNEL);
	if (!skypedrop)
		return -ENOMEM;

	memset(skypedrop, '\0', sizeof(struct skypedroplist));

	skyperetry = (struct skyperetry *)kmalloc(sizeof(struct skyperetry), GFP_KERNEL);
	if (!skyperetry) {
		kfree(skypedrop);
		return -ENOMEM;
	}

	memset(skyperetry, '\0', sizeof(struct skyperetry));
	return ipt_register_match(&skype_match);
}

static void __exit fini(void)
{
	if (skypedrop)
		kfree(skypedrop);

	if (skyperetry)
		kfree(skyperetry);

	ipt_unregister_match(&skype_match);
}

module_init(init);
module_exit(fini);
