#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/spinlock.h>
#include <linux/interrupt.h>
#include <linux/ip.h>
#include <linux/tcp.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ipt_synlimit.h>
#include <linux/netfilter_ipv4/ip_conntrack.h>

/* The algorithm used is the Simple Token Bucket Filter (TBF)
 * see net/sched/sch_tbf.c in the linux source tree
 */

static spinlock_t synlimit_lock = SPIN_LOCK_UNLOCKED;

/* Rusty: This is my (non-mathematically-inclined) understanding of
   this algorithm.  The `average rate' in jiffies becomes your initial
   amount of credit `credit' and the most credit you can ever have
   `credit_cap'.  The `peak rate' becomes the cost of passing the
   test, `cost'.

   `prev' tracks the last packet hit: you gain one credit per jiffy.
   If you get credit balance more than this, the extra credit is
   discarded.  Every time the match passes, you lose `cost' credits;
   if you don't have that many, the test fails.

   See Alexey's formal explanation in net/sched/sch_tbf.c.

   To avoid underflow, we multiply by 128 (ie. you get 128 credits per
   jiffy).  Hence a cost of 2^32-1, means one pass per 32768 seconds
   at 1024HZ (or one every 9 hours).  A cost of 1 means 12800 passes
   per second at 100HZ.  */

#define CREDITS_PER_JIFFY 128

static inline int hashfunc(struct in_addr addr)
{
        unsigned int value;
        int hash;

        value = addr.s_addr;
        hash = 0;
        do {
                hash ^= value;
        } while ((value >>= SIPSYN_HASH_LOG));

        return hash & (SIPSYN_HASH_SIZE - 1);
}

static int
ipt_synlimit_match(const struct sk_buff *skb,
		   const struct net_device *in,
		   const struct net_device *out,
		   const void *matchinfo,
		   int offset,
		   const void *hdr,
		   u_int16_t datalen,
		   int *hotdrop)
{
	struct ipt_synrateinfo *r = ((struct ipt_synrateinfo *)matchinfo)->master;
	unsigned long now = jiffies;
        struct in_addr addr;
        struct sipsyn_list *curr, *last, **head;
        int hash, count;
        struct iphdr *ip_hdr;
	struct tcphdr *tcph;
	u_int16_t src_port;
	u_int32_t seq;

	spin_lock_bh(&synlimit_lock);
        ip_hdr = skb->nh.iph;
	tcph = (struct tcphdr *)((u_int32_t *)ip_hdr + ip_hdr->ihl);
	src_port = tcph->source;
	seq = tcph->seq;
        addr.s_addr = ip_hdr->saddr;
        if (!addr.s_addr) {
                spin_unlock_bh(&synlimit_lock);
                return 1;
        }

        count = 0;
        last = NULL;
	if ((curr = *(head = &(r->limit->hash[hash = hashfunc(addr)]))))
		do {
			if (curr->src_addr.s_addr == addr.s_addr &&
			    curr->sport == src_port &&
			    curr->seq == seq) {
				break;
			}

			count++;
			if (curr->next) last = curr;
		} while ((curr = curr->next));

	if (curr && now < (curr->last_seen + (10 * HZ)) &&
	    now > (curr->last_seen + (1 * HZ))) {
		curr->last_seen = now;
		spin_unlock_bh(&synlimit_lock);
		return 1;
	} else {
		if (curr)
			curr->last_seen = now;
		else {
			if (count >= SIPSYN_HASH_MAX && last) last->next = NULL;

		        if (r->limit->list[r->limit->index].src_addr.s_addr)
        		        head = &r->limit->hash[hashfunc(r->limit->list[r->limit->index].src_addr)];
        		else
                		head = &last;

        		last = NULL;

        		if ((curr = *head))
        		do {
                	if (curr == &r->limit->list[r->limit->index]) break;
                		last = curr;
        		} while ((curr = curr->next));

		        if (curr) {
                		if (last)
                		        last->next = last->next->next;
                		else if (*head)
                        		*head = (*head)->next;
        		}

		        /* Get our list entry */
        		curr = &r->limit->list[r->limit->index++];
        		if (r->limit->index >= SIPSYN_LIST_SIZE) r->limit->index = 0;

        		/* Link it into the hash table */
        		head = &r->limit->hash[hash];
        		curr->next = *head;
        		*head = curr;
        		curr->src_addr = addr;
			curr->sport = src_port;
			curr->seq = seq;
        		curr->last_seen = now;
		}
	}

	r->credit += (now - xchg(&r->prev, now)) * CREDITS_PER_JIFFY;
	if (r->credit > r->credit_cap)
		r->credit = r->credit_cap;

	if ((r->credit >= r->cost)) {
		if ((now > r->timeout)) {
			/* We're not limited. */
			r->credit -= r->cost;
			spin_unlock_bh(&synlimit_lock);
			return 1;
		}

		r->credit -= r->cost;
		spin_unlock_bh(&synlimit_lock);
		return 0;
	}

	r->credit = 0;
	r->timeout = now + (HZ * 10);
       	spin_unlock_bh(&synlimit_lock);
	return 0;
}

/* Precision saver. */
static u_int32_t
user2credits(u_int32_t user)
{
	/* If multiplying would overflow... */
	if (user > 0xFFFFFFFF / (HZ*CREDITS_PER_JIFFY))
		/* Divide first. */
		return (user / IPT_SYNLIMIT_SCALE) * HZ * CREDITS_PER_JIFFY;

	return (user * HZ * CREDITS_PER_JIFFY) / IPT_SYNLIMIT_SCALE;
}

static void
ipt_synlimit_destroy(void *matchinfo, unsigned int matchsize)
{
	struct ipt_synrateinfo *r = matchinfo;

        if (matchsize != IPT_ALIGN(sizeof(struct ipt_synrateinfo)))
                return;

	spin_lock_bh(&synlimit_lock);
	if (r->limit)
		free_pages((unsigned long)(r->limit), SIPSYN_PAGES);

	spin_unlock_bh(&synlimit_lock);
		return;
}

static int
ipt_synlimit_checkentry(const char *tablename,
		        const struct ipt_ip *ip,
		        void *matchinfo,
		        unsigned int matchsize,
		        unsigned int hook_mask)
{
	struct ipt_synrateinfo *r = matchinfo;

	if (matchsize != IPT_ALIGN(sizeof(struct ipt_synrateinfo)))
		return 0;

	if (ip->proto != IPPROTO_TCP || (ip->invflags & IPT_INV_PROTO)) {
		printk("synlimit: Only works on TCP packets\n");
		return 0;
	}

	/* Check for overflow. */
	if (r->burst == 0
	    || user2credits(r->avg * r->burst) < user2credits(r->avg)) {
		printk("Call rusty: overflow in ipt_limit: %u/%u\n",
		       r->avg, r->burst);
		return 0;
	}

	/* User avg in seconds * IPT_LIMIT_SCALE: convert to jiffies *
	   128. */
	r->prev = jiffies;
	r->credit = user2credits(r->avg * r->burst);	 /* Credits full. */
	r->credit_cap = user2credits(r->avg * r->burst); /* Credits full. */
	r->cost = user2credits(r->avg);

	r->limit = (struct synlimit *)__get_free_pages(GFP_KERNEL, SIPSYN_PAGES);
	if (!r->limit) {
		printk("__get_free_pages fails\n");
		return 0;
	}
	memset(r->limit, '\0', sizeof(struct synlimit));
	/* For SMP, we only want to use one set of counters. */
	r->master = r;

	return 1;
}

static struct ipt_match ipt_synlimit_reg
= { { NULL, NULL }, "synlimit", ipt_synlimit_match, ipt_synlimit_checkentry, ipt_synlimit_destroy,
    THIS_MODULE };

static int __init init(void)
{
	if (ipt_register_match(&ipt_synlimit_reg))
		return -EINVAL;
	return 0;
}

static void __exit fini(void)
{
	ipt_unregister_match(&ipt_synlimit_reg);
}

module_init(init);
module_exit(fini);
