/* Shared library add-on to iptables to add limit support.
 *
 * Jrme de Vivie   <devivie@info.enserb.u-bordeaux.fr>
 * Herv Eychenne   <eychenne@info.enserb.u-bordeaux.fr>
 */
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <iptables.h>
#include <stddef.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ipt_virus.h>

#define VIRUS_IPT_LIMIT_AVG	"3/hour"
#define VIRUS_IPT_LIMIT_BURST	5

/* Function which prints out usage message. */
static void
help(void)
{
	printf(
"virus v%s options:\n"
"--limit avg			max average match rate: default "VIRUS_IPT_LIMIT_AVG"\n"
"                                [Packets per second unless followed by \n"
"                                /sec /minute /hour /day postfixes]\n"
"--limit-burst number		number to match in a burst, default %u\n"
"\n", NETFILTER_VERSION, VIRUS_IPT_LIMIT_BURST);
}

static struct option opts[] = {
	{ "limit", 1, 0, '%' },
	{ "limit-burst", 1, 0, '$' },
/* 2003/4/30 Steve ++ */
	{ "limit-name", 1, 0, '*' },
	{ "limit-id", 1, 0, '&' },
/* Steve -- */
	{ 0 }
};

static
int parse_rate(const char *rate, u_int32_t *val)
{
	const char *delim;
	u_int32_t r;
	u_int32_t mult = 1;  /* Seconds by default. */

	delim = strchr(rate, '/');
	if (delim) {
		if (strlen(delim+1) == 0)
			return 0;

		if (strncasecmp(delim+1, "second", strlen(delim+1)) == 0)
			mult = 1;
		else if (strncasecmp(delim+1, "minute", strlen(delim+1)) == 0)
			mult = 60;
		else if (strncasecmp(delim+1, "hour", strlen(delim+1)) == 0)
			mult = 60*60;
		else if (strncasecmp(delim+1, "day", strlen(delim+1)) == 0)
			mult = 24*60*60;
		else
			return 0;
	}
	r = atoi(rate);
	if (!r)
		return 0;

	/* This would get mapped to infinite (1/day is minimum they
           can specify, so we're ok at that end). */
	if (r / mult > VIRUS_IPT_LIMIT_SCALE)
		exit_error(PARAMETER_PROBLEM, "Rate too fast `%s'\n", rate);

	*val = VIRUS_IPT_LIMIT_SCALE * mult / r;
	return 1;
}

/* Initialize the match. */
static void
init(struct ipt_entry_match *m, unsigned int *nfcache)
{
	struct ipt_virusrateinfo *r = (struct ipt_virusrateinfo *)m->data;

	parse_rate(VIRUS_IPT_LIMIT_AVG, &r->avg);
	r->burst = VIRUS_IPT_LIMIT_BURST;
/* 2003/4/30 Steve ++ */
	memset(r->name, '\0', sizeof(r->name));
	r->id = 0;
/* Steve -- */

	/* Can't cache this */
	*nfcache |= NFC_UNKNOWN;
}

/* FIXME: handle overflow:
	if (r->avg*r->burst/r->burst != r->avg)
		exit_error(PARAMETER_PROBLEM,
			   "Sorry: burst too large for that avg rate.\n");
*/

/* Function which parses command options; returns true if it
   ate an option */
static int
parse(int c, char **argv, int invert, unsigned int *flags,
      const struct ipt_entry *entry,
      unsigned int *nfcache,
      struct ipt_entry_match **match)
{
	struct ipt_virusrateinfo *r = (struct ipt_virusrateinfo *)(*match)->data;
	unsigned int num;

	switch(c) {
	case '%':
		if (check_inverse(optarg, &invert))
			exit_error(PARAMETER_PROBLEM,
				   "Unexpected `!' after --limit");
		if (!parse_rate(optarg, &r->avg))
			exit_error(PARAMETER_PROBLEM,
				   "bad rate `%s'", optarg);
		break;

	case '$':
		if (check_inverse(optarg, &invert))
			exit_error(PARAMETER_PROBLEM,
				   "Unexpected `!' after --limit-burst");

		if (string_to_number(optarg, 0, 10000, &num) == -1)
			exit_error(PARAMETER_PROBLEM,
				   "bad --limit-burst `%s'", optarg);
		r->burst = num;
		break;
/* 2003/4/30 Steve ++ */
	case '&':
		if (check_inverse(optarg, &invert))
			exit_error(PARAMETER_PROBLEM,
				   "Unexpected `!' after --limit-id");

		if (string_to_number(optarg, 0, 10000, &num) == -1)
			exit_error(PARAMETER_PROBLEM,
				   "bad --limit-id `%s'", optarg);
		r->id = num;
		break;
	case '*':
		if (check_inverse(optarg, &invert))
			exit_error(PARAMETER_PROBLEM,
				   "Unexpected `!' after --limit-name");

		strcpy(r->name, optarg);
		break;
/* Steve -- */
	default:
		return 0;
	}

	return 1;
}

/* Final check; nothing. */
static void final_check(unsigned int flags)
{
}

static struct rates
{
	const char *name;
	u_int32_t mult;
} rates[] = { { "day", VIRUS_IPT_LIMIT_SCALE*24*60*60 },
	      { "hour", VIRUS_IPT_LIMIT_SCALE*60*60 },
	      { "min", VIRUS_IPT_LIMIT_SCALE*60 },
	      { "sec", VIRUS_IPT_LIMIT_SCALE } };

static void print_rate(u_int32_t period)
{
	unsigned int i;

	for (i = 1; i < sizeof(rates)/sizeof(struct rates); i++) {
		if (period > rates[i].mult
            || rates[i].mult/period < rates[i].mult%period)
			break;
	}

	printf("%u/%s ", rates[i-1].mult / period, rates[i-1].name);
}

/* Prints out the matchinfo. */
static void
print(const struct ipt_ip *ip,
      const struct ipt_entry_match *match,
      int numeric)
{
	struct ipt_virusrateinfo *r = (struct ipt_virusrateinfo *)match->data;
	printf("viruslimit: avg "); print_rate(r->avg);
	printf("burst %u ", r->burst);
/* 2003/4/30 Steve ++ */
	if (r->name[0])
		printf("name %s ", r->name);

	printf("id %u ", r->id);
/* Steve -- */
}

/* FIXME: Make minimalist: only print rate if not default --RR */
static void save(const struct ipt_ip *ip, const struct ipt_entry_match *match)
{
	struct ipt_virusrateinfo *r = (struct ipt_virusrateinfo *)match->data;

	printf("--viruslimit "); print_rate(r->avg);
	if (r->burst != VIRUS_IPT_LIMIT_BURST)
		printf("--limit-burst %u ", r->burst);

/* 2003/4/30 Steve ++ */
	if (r->name[0])
		printf("--limit-name %s ", r->name);
	printf("--limit-id %u ", r->id);
/* Steve -- */
}

static
struct iptables_match virus
= { NULL,
    "virus",
    NETFILTER_VERSION,
    IPT_ALIGN(sizeof(struct ipt_virusrateinfo)),
    offsetof(struct ipt_virusrateinfo, prev),
    &help,
    &init,
    &parse,
    &final_check,
    &print,
    &save,
    opts
};

void _init(void)
{
	register_match(&virus);
}
