#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/spinlock.h>
#include <linux/interrupt.h>
#include <asm/uaccess.h>
#include <linux/ctype.h>
#include <linux/ip.h>
#include <linux/tcp.h>
#include <linux/udp.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ipt_virus.h>
#include <linux/proc_fs.h>
#include <linux/timer.h>

static spinlock_t virus_lock = SPIN_LOCK_UNLOCKED;
static struct proc_dir_entry *proc_net_virus = NULL; 
static struct virusdroplist *droplistTable;
static struct virusretry *virusretrytable;
static int blocking = 1;
static unsigned int droptimeout = 60;
static struct timer_list virus_timer;

#define VIRUS_CREDITS_PER_JIFFY 128

/* Precision saver. */
static u_int32_t
user2credits(u_int32_t user)
{
	/* If multiplying would overflow... */
	if (user > 0xFFFFFFFF / (HZ*VIRUS_CREDITS_PER_JIFFY))
		/* Divide first. */
		return (user / VIRUS_IPT_LIMIT_SCALE) * HZ * VIRUS_CREDITS_PER_JIFFY;

	return (user * HZ * VIRUS_CREDITS_PER_JIFFY) / VIRUS_IPT_LIMIT_SCALE;
}


static int
ipt_virus_read_info(char *buffer,
		    char **start,
		    off_t offset,
		    int length,
		    int *eof,
		    unsigned int data)
{
	int len = 0, last_len = 0;
	off_t pos = 0;
	off_t begin = 0;
	unsigned int id;
	int i = 0;
	struct virusdroplist *droptable = droplistTable;

	id = data;
	spin_lock_bh(&virus_lock);
	for (i = 0; i < VIRUS_LIST_SIZE; i++) {
		if (droptable->list[i].src_addr.s_addr && id == droptable->list[i].id) {
			last_len = len;
			len += sprintf(buffer+len, "%u.%u.%u.%u %lu %u\n",
					NIPQUAD(droptable->list[i].src_addr.s_addr),
					droptable->list[i].droptime,
					droptable->list[i].alert);

			pos = begin + len;
			if(pos < offset) { len = 0; begin = pos; }
			if(pos > offset + length) { len = last_len; break; }
		}
	}

	*start = buffer + (offset - begin);
	len -= (offset - begin);
	if(len > length) len = length;
	spin_unlock_bh(&virus_lock);
	return len;
}

static int
ipt_virus_write_info(struct file *file,
		     const char *input,
		     unsigned long size,
		     void *data)
{
	char buffer[1024];
	int len = 0, options = 0;
	static const u_int32_t max[4] = { 0xffffffff, 0xffffff, 0xffff, 0xff };
	u_int32_t val;
	int base;
	char c, *cp;
	union iaddr {
		uint8_t bytes[4];
		uint32_t word;
	} res;
	uint8_t *pp = res.bytes;
	int digit;
	u_int32_t saddr = 0;
	int i = 0;
	struct virusdroplist *droptable = droplistTable;

	spin_lock_bh(&virus_lock);
	if (size > (sizeof(buffer) - 1)) {
		spin_unlock_bh(&virus_lock);
		return size;
	} else
		len = size;

	if (copy_from_user(buffer, input, len)) {
		spin_unlock_bh(&virus_lock);
		return -EFAULT;
	}

	buffer[len - 1] = '\0';
	cp = buffer;
	if (*cp == '+')
		options = 1;
	else if (*cp == '-')
		options = 0;
	else {
		if (!memcmp(cp, "block", 5))
			blocking = 1;
		else if (!memcmp(cp, "noblock", 7))
			blocking = 0;
		else if (!memcmp(cp, "time:", 5)) {
			cp += 5;
			droptimeout = simple_strtoul(cp, NULL, 10);
		}
		spin_unlock_bh(virus_lock);
		return len;
	}
	cp++;
	res.word = 0;
	c = *cp;
        for(;;) {
                if(!isdigit(c)) {
			spin_unlock_bh(&virus_lock);
			return len;
		}

                val = 0; base = 10; digit = 0;
                if(c == '0') {
                        c = *++cp;
                        if(c == 'x' || c == 'X') base = 16, c = *++cp;
                        else { base = 8; digit = 1; }
                }
                for(;;) {
                        if(isascii(c) && isdigit(c)) {
                		if(base == 8 && (c == '8' || c == '0')) {
					spin_unlock_bh(&virus_lock);
					return len;
				}
                                val = (val * base) + (c - '0');
                                c = *++cp;
                                digit = 1;
                	} else if(base == 16 && isascii(c) && isxdigit(c)) {
                		val = (val << 4) | (c + 10 - (islower(c) ? 'a' : 'A'));
                                c = *++cp;
                                digit = 1;
                        } else break;
                }
                if(c == '.') {
                        if(pp > res.bytes + 2 || val > 0xff) {
				spin_unlock_bh(&virus_lock);
				return len;
			}
                        *pp++ = val;
                        c = *++cp;
		} else break;
	}
	if(c != '\0' && (!isascii(c) || !isspace(c))) {
		spin_unlock_bh(&virus_lock);
		return len;
	}

	if(!digit) {
		spin_unlock_bh(&virus_lock);
		return len;
	}

	if(val > max[pp - res.bytes]) {
		spin_unlock_bh(&virus_lock); 
		return len;
	}
	saddr = res.word | htonl(val);

	for (i = 0; i < VIRUS_LIST_SIZE; i++) {
		if (droptable->list[i].src_addr.s_addr == saddr) {
			if (!options) {
				droptable->list[i].alert = 0;
				droptable->list[i].droptime = 0;
				droptable->list[i].timeout = 0;
				droptable->list[i].id = 0;
				droptable->list[i].src_addr.s_addr = 0;
			} else
				droptable->list[i].alert = 1;

			break;
		}
	}

	spin_unlock_bh(&virus_lock);
	return size;
}

static inline int hashfunc(struct in_addr addr)
{
        unsigned int value;
        int hash;

        value = addr.s_addr;
        hash = 0;
        do {
                hash ^= value;
        } while ((value >>= VIRUS_HASH_LOG));

        return hash & (VIRUS_HASH_SIZE - 1);
}

static inline int hashfunc_retry(struct in_addr src, struct in_addr dst, unsigned short sport, unsigned short dport)
{
        unsigned int value;
	unsigned short port;
        int hash;

        value = src.s_addr;
        hash = 0;
        do {
                hash ^= value;
        } while ((value >>= VIRUS_RETRY_HASH_LOG));
	value = dst.s_addr;
        do {
                hash ^= value;
        } while ((value >>= VIRUS_RETRY_HASH_LOG));
	port = sport;
        do {
                hash ^= port;
        } while ((port >>= VIRUS_RETRY_HASH_LOG));
	port = dport;
        do {
                hash ^= port;
        } while ((port >>= VIRUS_RETRY_HASH_LOG));

        return hash & (VIRUS_RETRY_HASH_SIZE - 1);
}

static int
retry_packet(struct packetinfo *packet, struct virusretry *virusretry)
{
	struct virus_retry_list *curr, *last, **head;
	int hash, count;
	unsigned long now = jiffies;

	spin_lock_bh(&virus_lock);
        count = 0;
        last = NULL;

	if ((curr = *(head = &(virusretry->hash[hash = hashfunc_retry(packet->saddr.addr, packet->daddr.addr, packet->saddr.port, packet->daddr.port)]))))
		do {
			if (curr->packetinfo.protocol == packet->protocol &&
			    curr->packetinfo.saddr.addr.s_addr == packet->saddr.addr.s_addr &&
			    curr->packetinfo.daddr.addr.s_addr == packet->daddr.addr.s_addr &&
			    curr->packetinfo.saddr.port == packet->saddr.port &&
			    curr->packetinfo.daddr.port == packet->daddr.port &&
			    curr->packetinfo.seq == packet->seq) {
                                break;
                        }

                        count++;
                        if (curr->next) last = curr;
                } while ((curr = curr->next));

        if (curr && now < (curr->last_seen + (10 * HZ)) &&
            now > (curr->last_seen + (1 * HZ))) {
                curr->last_seen = now;
                spin_unlock_bh(&virus_lock);
                return 1;
        } else {
                if (curr)
                        curr->last_seen = now;
		else {
			if (count >= VIRUS_RETRY_HASH_MAX && last) last->next = NULL;

			if (virusretry->list[virusretry->index].packetinfo.saddr.addr.s_addr)
				head = &virusretry->hash[hashfunc_retry(
							virusretry->list[virusretry->index].packetinfo.saddr.addr,
							virusretry->list[virusretry->index].packetinfo.daddr.addr,
							virusretry->list[virusretry->index].packetinfo.saddr.port,
							virusretry->list[virusretry->index].packetinfo.daddr.port)];
			else
				head = &last;

                        last = NULL;

                        if ((curr = *head))
                        do {
                        if (curr == &virusretry->list[virusretry->index]) break;
                                last = curr;
                        } while ((curr = curr->next));

                        if (curr) {
                                if (last)
                                        last->next = last->next->next;
                                else if (*head)
                                        *head = (*head)->next;
                        }

			curr = &virusretry->list[virusretry->index++];

			if (virusretry->index >= VIRUS_RETRY_LIST_SIZE) 
				virusretry->index = 0;

                        head = &virusretry->hash[hash];
                        curr->next = *head;
                        *head = curr;
                        curr->packetinfo.protocol = packet->protocol;
                        curr->packetinfo.saddr.addr.s_addr = packet->saddr.addr.s_addr;
                        curr->packetinfo.saddr.port = packet->saddr.port;
                        curr->packetinfo.daddr.addr.s_addr = packet->daddr.addr.s_addr;
                        curr->packetinfo.daddr.port = packet->daddr.port;
			curr->packetinfo.seq = packet->seq;
                        curr->last_seen = now;
                }
        }

	spin_unlock_bh(&virus_lock);
	return 0;
}

static void
getpacketinfo(const struct sk_buff *skb, struct iphdr *ip_hdr, struct packetinfo *packet)
{
	struct tcphdr *tcph;
	struct udphdr *udph;

	switch (ip_hdr->protocol) {
	case IPPROTO_TCP:
		tcph = (void *)ip_hdr + ip_hdr->ihl * 4;
		packet->protocol = IPPROTO_TCP;
		packet->saddr.addr.s_addr = ip_hdr->saddr;
		packet->daddr.addr.s_addr = ip_hdr->daddr;
		packet->saddr.port = tcph->source;
		packet->daddr.port = tcph->dest;
		packet->seq = tcph->seq;
		break;
	case IPPROTO_UDP:
		udph = (void *)ip_hdr + ip_hdr->ihl * 4;
		packet->protocol = IPPROTO_UDP;
		packet->saddr.addr.s_addr = ip_hdr->saddr;
		packet->daddr.addr.s_addr = ip_hdr->daddr;
		packet->saddr.port = udph->source;
		packet->daddr.port = udph->dest;
		packet->seq = 0;
		break;
	default:
		packet->protocol = 0;
		break;
	}
}

static int
ipt_virus_match(const struct sk_buff *skb,
		const struct net_device *in,
		const struct net_device *out,
		const void *matchinfo,
		int offset,
		const void *hdr,
		u_int16_t datalen,
		int *hotdrop)
{
	struct ipt_virusrateinfo *r;
	unsigned long now = jiffies;
	struct in_addr addr;
	struct virus_host *curr, *last, **head;
	int hash, count;
	struct iphdr *ip_hdr;
	struct virusdropsip *drop_curr, *drop_last, **drop_head;
	int drop_hash, drop_count;
	struct packetinfo packet;
	struct virusretry *virusretry;
	struct timeval currenttvsec;

	spin_lock_bh(&virus_lock);
/* 2003/12/22 Steve ++ */
        if (skb->unreplay) {
                spin_unlock_bh(&virus_lock);
                return 1;
        }
/* Steve -- */

	r = ((struct ipt_virusrateinfo *)matchinfo)->master;
	ip_hdr = skb->nh.iph;
	addr.s_addr = ip_hdr->saddr;
	if (!addr.s_addr) {
		spin_unlock_bh(&virus_lock);
		return 1;
	}

	getpacketinfo(skb, ip_hdr, &packet);
	virusretry = r->virusretry;

	drop_count = 0;
	drop_last = NULL;
	if ((drop_curr = *(drop_head = &(r->drop->hash[drop_hash = hashfunc(addr)]))))
		do {
			if (drop_curr->src_addr.s_addr == addr.s_addr) {
				if (droptimeout && drop_curr->timeout < now) {
					drop_curr->alert = 0;
					drop_curr->droptime = 0;
					drop_curr->timeout = 0;
					drop_curr->id = 0;
					drop_curr->src_addr.s_addr = 0;
					goto overtime;
				}

				if (droptimeout)
					drop_curr->timeout = now + (droptimeout * HZ);

				if (!blocking) {
					spin_unlock_bh(&virus_lock);
					return 1;
				}

				if (packet.protocol && retry_packet(&packet, virusretry)) {
					spin_unlock_bh(&virus_lock);
					return 1;
				}
				spin_unlock_bh(&virus_lock);
				return 0;
			}

			drop_count++;
			if (drop_curr->next) drop_last = drop_curr;
		} while ((drop_curr = drop_curr->next));

overtime:
	count = 0;
	last = NULL;
	if ((curr = *(head = &(r->state->hash[hash = hashfunc(addr)]))))
		do {
			if (curr->src_addr.s_addr == addr.s_addr) break;

			count++;
			if (curr->next) last = curr;
		} while ((curr = curr->next));

	if (curr) {
		curr->credit += (now - xchg(&curr->prev, now)) * VIRUS_CREDITS_PER_JIFFY;
		if (curr->credit > curr->credit_cap)
			curr->credit = curr->credit_cap;

		if (curr->credit >= curr->cost) {
			curr->credit -= curr->cost;
			spin_unlock_bh(&virus_lock);
			return 1;
		}

		goto nomatch;
	}

	if (count >= VIRUS_HASH_MAX && last) last->next = NULL;

	if (r->state->list[r->state->index].src_addr.s_addr)
		head = &r->state->hash[hashfunc(r->state->list[r->state->index].src_addr)];
	else
		head = &last;

	last = NULL;
	if ((curr = *head))
	do {
		if (curr == &r->state->list[r->state->index]) break;
		last = curr;
	} while ((curr = curr->next));

	if (curr) {
		if (last)
			last->next = last->next->next;
		else if (*head)
			*head = (*head)->next;
	}

	/* Get our list entry */
	curr = &r->state->list[r->state->index++];
	if (r->state->index >= VIRUS_LIST_SIZE) r->state->index = 0;

	/* Link it into the hash table */
	head = &r->state->hash[hash];
	curr->next = *head;
	*head = curr;
	curr->src_addr = addr;
	curr->prev = now;
	curr->credit_cap = r->credit_cap;
	curr->credit = r->credit_cap;
	curr->cost = r->cost;

	spin_unlock_bh(&virus_lock);
	return 1;

nomatch:
	drop_count = 0;
	drop_last = NULL;
	if ((drop_curr = *(drop_head = &(r->drop->hash[drop_hash = hashfunc(addr)]))))
		do {
			if (drop_curr->src_addr.s_addr == addr.s_addr)
				break;

			drop_count++;
			if (drop_curr->next) drop_last = drop_curr;
		} while ((drop_curr = drop_curr->next));


	if (drop_count >= VIRUS_HASH_MAX && drop_last) drop_last->next = NULL;

	if (r->drop->list[r->drop->index].src_addr.s_addr)
		drop_head = &r->drop->hash[hashfunc(r->drop->list[r->drop->index].src_addr)];
	else
		drop_head = &drop_last;

	drop_last = NULL;
	if ((drop_curr = *drop_head))
	do {
		if (drop_curr == &r->drop->list[r->drop->index]) break;
		drop_last = drop_curr;
	} while ((drop_curr = drop_curr->next));

	if (drop_curr) {
		if (drop_last)
			drop_last->next = drop_last->next->next;
		else if (*drop_head)
			*drop_head = (*drop_head)->next;
	}

	/* Get our list entry */
	drop_curr = &r->drop->list[r->drop->index++];
	if (r->drop->index >= VIRUS_LIST_SIZE) r->drop->index = 0;

	/* Link it into the hash table */
	drop_head = &r->drop->hash[drop_hash];
	drop_curr->next = *drop_head;
	*drop_head = drop_curr;
	drop_curr->src_addr = addr;
	do_gettimeofday(&currenttvsec);
	drop_curr->droptime = currenttvsec.tv_sec;
	if (droptimeout)
		drop_curr->timeout = now + (droptimeout * HZ);
	drop_curr->id = r->id;
	drop_curr->alert = 0;
	spin_unlock_bh(&virus_lock);
/* 2004/1/19 Steve ++ */
	if (!blocking)
		return 1;
	else
/* Steve -- */
		return 0;
}

static int
ipt_virus_checkentry(const char *tablename,
		     const struct ipt_ip *ip,
		     void *matchinfo,
		     unsigned int matchsize,
		     unsigned int hook_mask)
{
	struct ipt_virusrateinfo *r = matchinfo;

	if (matchsize != IPT_ALIGN(sizeof(struct ipt_virusrateinfo)))
		return 0;

	/* Check for overflow. */
	if (r->burst == 0
	    || user2credits(r->avg * r->burst) < user2credits(r->avg)) {
		printk("Call rusty: overflow in ipt_virus: %u/%u\n",
		       r->avg, r->burst);
		return 0;
	}

	/* User avg in seconds * IPT_LIMIT_SCALE: convert to jiffies *
	   128. */
	r->prev = jiffies;
	r->credit = user2credits(r->avg * r->burst);	 /* Credits full. */
	r->credit_cap = user2credits(r->avg * r->burst); /* Credits full. */
	r->cost = user2credits(r->avg);

	r->state = (struct virusstate *)kmalloc(sizeof(struct virusstate), GFP_KERNEL);
	memset(r->state, '\0', sizeof(struct virusstate));
	r->drop = droplistTable;

	r->virusretry = (struct virusretry *)virusretrytable;

	if (proc_net_virus) {
		r->status_proc = create_proc_entry(r->name, 0644, proc_net_virus);
		if (r->status_proc) {
			r->status_proc->owner = THIS_MODULE;
			r->status_proc->read_proc = ipt_virus_read_info;
			r->status_proc->write_proc = ipt_virus_write_info;
			r->status_proc->data = r->id;
		}
	}

	/* For SMP, we only want to use one set of counters. */
	r->master = r;
	return 1;
}

static void
ipt_virus_destroy(void *matchinfo, unsigned int matchsize)
{
        struct ipt_virusrateinfo *r = matchinfo;

        if (matchsize != IPT_ALIGN(sizeof(struct ipt_virusrateinfo)))
                return;

	spin_lock_bh(&virus_lock);
	if (r->state)
		kfree(r->state);
	if (r->status_proc)
		remove_proc_entry(r->name, proc_net_virus);

	spin_unlock_bh(&virus_lock);
	return;
}

static void
update_virus(unsigned long timeout)
{
	int i = 0;
	struct virusdroplist *droptable = droplistTable;

	for (i = 0; i < VIRUS_LIST_SIZE; i++) {
		if (droptimeout && droptable->list[i].timeout && droptable->list[i].timeout < jiffies) {
			droptable->list[i].alert = 0;
			droptable->list[i].droptime = 0;
			droptable->list[i].timeout = 0;
			droptable->list[i].id = 0;
			droptable->list[i].src_addr.s_addr = 0;
		}
	}

        virus_timer.expires = jiffies + (HZ * 30);
        add_timer(&virus_timer);
}

static struct ipt_match ipt_virus_reg
= { { NULL, NULL }, "virus", ipt_virus_match, ipt_virus_checkentry, ipt_virus_destroy, THIS_MODULE };

static int __init init(void)
{
	if (!proc_net_virus)
		proc_net_virus = proc_mkdir("ipt_virus", proc_net);

	droplistTable = (struct virusdroplist *)__get_free_pages(GFP_KERNEL, 0);
	if (!droplistTable)
		return -ENOMEM;
	memset(droplistTable, '\0', sizeof(struct virusdroplist));
	virusretrytable = (struct virusretry *)__get_free_pages(GFP_KERNEL, 7);
	if (!virusretrytable) {
		free_pages(droplistTable, 0);
		return -ENOMEM;
	}
	memset(virusretrytable, '\0', sizeof(struct virusretry));

	init_timer(&virus_timer);
	virus_timer.function = update_virus;
	virus_timer.expires = jiffies + (HZ * 30);
	add_timer(&virus_timer);


	if (ipt_register_match(&ipt_virus_reg))
		return -EINVAL;

	return 0;
}

static void __exit fini(void)
{
	if (droplistTable)
		free_pages(droplistTable, 0);
	if (virusretrytable)
		free_pages(virusretrytable, 7);

	ipt_unregister_match(&ipt_virus_reg);
}

module_init(init);
module_exit(fini);
